#!/usr/bin/env wish
#
# TGM Minima
# Copyright (C) 2018 Martin Saturka http://minima.tangloid.net/
#
# This is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 2 of the License,
# or (at your option) any later version.
# This program is distributed WITHOUT ANY (even implied) WARRANTY.
#
set app_license "GPL 2+"
set app_version "0.3.0"
set app_website "http://minima.tangloid.net/"
set app_title_base "TGM Minima"

set editor_batch_state ""

# it is easier to locate this than "puts" calls
proc debug_output {mssg} {
  catch {puts $mssg}
  catch {
    upvar 1 $mssg arrvar
    catch {parray arrvar}
  }
}

# older versions of Tcl have the "string" command without "cat" subcommand
proc strcat {args} {return [join $args ""]}

set save_path ""

# these can go via command-line parameter
set last_dir [pwd]
#
# should set "quiet" option too, once it is actually implemented in Maxima;
set with_labels false
# we gather evaluation times regardless of this setting,
# and should have option specifically for the time-info showing;
set bind_labels false
#
set info_line_size [list 90 1000000]
#
set maxima_output_to_split true
set maxima_line_size_min 40
set maxima_line_size_max 200
#
set line_display_mode_current 1
# list items: "label text" "menu text" [long left 1dim inner tex split]
set lambda_char "L"
catch {
  set lambda_char [encoding convertfrom "utf-8" "\xCE\xBB"]
}
# may be to set one another (1d) form that has line split only by Minima;
set line_display_modes_list [list \
  [list "<-1" "1dim (left-adjusted) long lines" [list true true true false false false]] \
  [list "<:1" "1dim (left-adjusted) split lines" [list true true true false false true]] \
  [list "<|1" "1dim (left-adjusted) short lines" [list false true true false false true]] \
  [list "<|2" "2dim left-adjusted (short) lines" [list false true false false false false]] \
  [list "c|2" "2dim centered (short) lines" [list false false false false false false]] \
  [list "($lambda_char)" "$lambda_char: inner Lisp forms shown" [list true true true true false false]] \
]
set chi_letter "X"
catch {
  set chi_letter [encoding convertfrom "utf-8" "\xCE\xA7"]
#  set chi_letter [encoding convertfrom "utf-8" "\xCF\x87"]
}
set line_display_mode_tex [list "\{$chi_letter\}" "TeX: via the Maxima TeX output" [list true true true false true false]]
#
set names_from_functions_to_load true
set names_from_values_to_load true
#
set names_from_files_to_load false
set names_from_files_path_list [list]
#
set lisp_flavor_to_use ""
set maxima_version_to_use ""
set lisp_options_to_set ""
#
set maxima_path "maxima"
#
set maxima_userdir_to_set ""
set maxima_initname_to_set ""
set clipboard_enforcer ""
set minima_clipboard_without_selection false
set minima_clipboard_force_selection false
#
# dealing with 'stderr' in the stack mode only by now;
#
# notice that regarding batch-mode, some exceptions are left without info
# in output and those exceptions do not set the 'error' variable either;
# setting '*debugger-hook*' to 'nil' did not lead to its showing either;
# would need to redefine suprv1.lisp:$errcatch and/or macsys.lisp:maxima-lisp-debugger
# to deal with that (probably via saving the condition into an additional variable);
#
set display_maxima_stderr false
#
# since commonly using the full-fill computation block highlighting,
# setting it as default;
set highlight_text_blocks_mode 0
set show_cmmd_highlight_line_nums false
#
set info_label_output_modes [list "s" "v" "e"]
set info_label_output_index 1
set omit_auto_tildes false
set hide_tag_tildes false
#
set file_path_to_open ""
#
set use_buttons_text false
#
set show_time_elapsed false
set include_maxima_input_default true
#
set test_command_select_first true
set test_command_whole_then true
#
set test_plot_input_select_first true
set test_plot_input_whole_then true
#
set text_trigger_use_text_selection false
set text_trigger_placed_text_selection false
set text_trigger_use_command true
set text_trigger_plot_selection false
#
# if gnuplot is not in PATH env, or if wanna use a different one
set gnuplot_path ""
set spell_path ""
set spch_default_allows_concat false
set spch_ascii_only false
#
set gnuplot_focusing_menu_options [list \
  [list "not at all" 0 false] \
  [list "1x within up to 0.5 s" 5 false] \
  [list "1x within up to 1 s" 10 false] \
  [list "1x within up to 2 s" 20 false] \
  [list "for complete 0.5 s" 5 true] \
  [list "for complete 1 s" 10 true] \
  [list "for complete 2 s" 20 true] \
]
set gnuplot_keep_focused_idx_default 2
set gnuplot_keep_focused_idx_full_inc [expr {int([llength $gnuplot_focusing_menu_options] / 2)}]
set gnuplot_keep_focused_idx $gnuplot_keep_focused_idx_default
set gnuplot_keep_focused_full false
set gnuplot_keep_always_topmost false
#
set command_onset_use_along_commands false
set mxmbatch_input_prepend_use true
set mxmbatch_input_prepend ""
#
set text_selection_action_asked "batch-inout"
set text_stack_action_asked "selection-command"
set text_selection_output_along_asked true
#
set maxima_batch_max_row_count_for_text 20
set make_line_endings false
set make_line_endings_asked false
#
set batch_maxima_end_automatic true
set batch_maxima_end_cleaning false
#
set maxima_lisp_path ""
set maxima_load_path ""
set maxima_exec_path ""
set sbcl_home_path ""
set asked_temporary_path ""
#
# common value for Linux clocks per second is 100,
# it can be gained e.g. by calling Python with command:
# python -c 'import os;print(os.sysconf("SC_CLK_TCK"))'
set system_clock_ticks 100
#
set print_logging_maxima [list]
set print_logging_write_start_only false
set print_logging_write_start_length 300
#
set mxm_display_exp_dstar false
set mxm_display_sorted_ratsimp false
set system_avoid_startup_warning false
set system_startup_warnings [list]
#
set allow_maxima_system_variables ""
#
set import_input_warning false
set plot_input_warning false
set batch_maxima_input_warning false
set batch_maxima_output_warning false
#
set default_tag_link_height 3
set default_tag_link_depth 3
#
set default_ft_minima true
set default_scan_minima_only true
set section_line_border_width 1
#
set backup_on_save_tilde false
#
set show_tags_found_count_info false
set show_tags_found_with_link true
#
set use_tab_length 4
set use_tab_length_max 10
#
proc take_commandline_parameters {} {
  global argv
  global with_labels
  global bind_labels
  global info_line_size
  global line_display_mode_current
  global names_from_functions_to_load
  global names_from_values_to_load
  global names_from_files_to_load
  global names_from_files_path_list
  global lisp_flavor_to_use
  global last_dir
  global app_license
  global app_version
  global app_website
  global maxima_version_to_use
  global maxima_path
  global maxima_userdir_to_set
  global maxima_initname_to_set
  global display_maxima_stderr
  global highlight_text_blocks_mode
  global show_cmmd_highlight_line_nums
  global info_label_output_index
  global info_label_output_modes
  global info_label_output_modes_params
  global omit_auto_tildes
  global hide_tag_tildes
  global clipboard_enforcer
  global editor_batch_state
  global file_path_to_open
  global gnuplot_path
  global spell_path
  global use_buttons_text
  global lisp_options_to_set
  global maxima_line_size_min
  global maxima_line_size_max
  global show_time_elapsed
  global include_maxima_input_default
  global test_command_select_first
  global test_command_whole_then
  global test_plot_input_select_first
  global test_plot_input_whole_then
  global minima_clipboard_without_selection
  global minima_clipboard_force_selection
  global command_onset_use_along_commands
  global mxmbatch_input_prepend_use
  global mxmbatch_input_prepend
  global gnuplot_keep_focused_idx
  global gnuplot_keep_focused_full
  global gnuplot_keep_always_topmost
  global text_selection_action_asked
  global text_stack_action_asked
  global text_selection_output_along_asked
  global make_line_endings_asked
  global maxima_lisp_path
  global maxima_load_path
  global maxima_exec_path
  global sbcl_home_path
  global system_clock_ticks
  global batch_maxima_end_automatic
  global batch_maxima_end_cleaning
  global print_logging_maxima
  global print_logging_write_start_only
  global print_logging_write_start_length
  global app_title_base
  global mxm_display_exp_dstar
  global mxm_display_sorted_ratsimp
  global system_avoid_startup_warning
  global system_startup_warnings
  global asked_temporary_path
  global allow_maxima_system_variables
  global import_input_warning
  global plot_input_warning
  global batch_maxima_input_warning
  global batch_maxima_output_warning
  global default_ft_minima
  global default_scan_minima_only
  global section_line_border_width
  global backup_on_save_tilde
  global show_tags_found_count_info
  global show_tags_found_with_link
  global spch_default_allows_concat
  global spch_ascii_only
  global use_tab_length
  global use_tab_length_max

  set param_linesize_start "--line-size="
  set param_linesize_start_length [string length $param_linesize_start]
  set param_withnames_start "--with-names="
  set param_withnames_start_length [string length $param_withnames_start]
  set param_lastdir_start "--last-dir="
  set param_lastdir_start_length [string length $param_lastdir_start]
  set param_uselisp_start "--use-lisp="
  set param_uselisp_start_length [string length $param_uselisp_start]
  set param_lisp_options_start "--lisp-options="
  set param_lisp_options_start_length [string length $param_lisp_options_start]

  set param_maximauserdir_start "--maxima-userdir="
  set param_maximauserdir_start_length [string length $param_maximauserdir_start]
  set param_maximainit_start "--maxima-init="
  set param_maximainit_start_length [string length $param_maximainit_start]
  set param_maximapath_start "--maxima-path="
  set param_maximapath_start_length [string length $param_maximapath_start]
  set param_maximaversion_start "--maxima-version="
  set param_maximaversion_start_length [string length $param_maximaversion_start]

  set param_outputplacing_start "--output-placing="
  set param_outputplacing_start_length [string length $param_outputplacing_start]

  set param_clipboardforcing_start "--clipboard-forcing="
  set param_clipboardforcing_start_length [string length $param_clipboardforcing_start]
  set param_gnuplotpath_start "--gnuplot-path="
  set param_gnuplotpath_start_length [string length $param_gnuplotpath_start]
  set param_spellpath_start "--spell-path="
  set param_spellpath_start_length [string length $param_spellpath_start]

  set param_fileopen_start "--file-open="
  set param_fileopen_start_length [string length $param_fileopen_start]

  set param_batch_prepend_command "--batch-prepend-command="
  set param_batch_prepend_command_length [string length $param_batch_prepend_command]

  set param_lisppath_start "--lisp-path="
  set param_lisppath_start_length [string length $param_lisppath_start]
  set param_corepath_start "--maxima-load-path="
  set param_corepath_start_length [string length $param_corepath_start]
  set param_imagepath_start "--maxima-exec-path="
  set param_imagepath_start_length [string length $param_imagepath_start]

  set param_sbclhome_start "--sbcl-home="
  set param_sbclhome_start_length [string length $param_sbclhome_start]

  set param_clockticks_start "--system-clock-ticks="
  set param_clockticks_start_length [string length $param_clockticks_start]

  set param_loggingstart_start "--logging-limit="
  set param_loggingstart_start_length [string length $param_loggingstart_start]

  set param_temppath_start "--temp-dir="
  set param_temppath_start_length [string length $param_temppath_start]

  set param_seclineborder_start "--section-line-border="
  set param_seclineborder_start_length [string length $param_seclineborder_start]

  # considering 1 to $use_tab_length_max values only;
  set param_tab_length_start "--tab-size="
  set param_tab_length_start_length [string length $param_tab_length_start]

  foreach param $argv {
    if {$param=="--version"} {
      puts "$app_version"
      exit 0
    }
    if {$param=="--license"} {
      puts "$app_license"
      exit 0
    }
    if {$param=="--help"} {
      puts "see the $app_title_base website for documentation:"
      puts "$app_website"
      puts "--stack-mode for the stack mode (default)"
      puts "--batch-mode for the batch mode (default if renamed to 'minima_batch.tcl')"
      exit 0
    }

    if {$param=="--stack-mode"} {
      set editor_batch_state false
      continue
    }
    if {$param=="--batch-mode"} {
      set editor_batch_state true
      continue
    }

    if {$param=="--flash-stderr"} {
      set display_maxima_stderr true
      continue
    }
    if {$param=="--buttons-text"} {
      set use_buttons_text true
      continue
    }
    if {$param=="--buttons-figs"} {
      set use_buttons_text false
      continue
    }

    if {$param=="--default-ft-twt"} {
      set default_ft_minima true
      continue
    }
    if {$param=="--default-ft-txt"} {
      set default_ft_minima false
      continue
    }
    if {$param=="--scan-txt"} {
      set default_scan_minima_only false
      continue
    }

    if {$param=="--unset-warning-startup"} {
      set system_avoid_startup_warning true
      continue
    }
    if {$param=="--set-warning-startup"} {
      set system_avoid_startup_warning false
      continue
    }
    if {$param=="--unset-warning-import"} {
      set import_input_warning false
      continue
    }
    if {$param=="--set-warning-import"} {
      set import_input_warning true
      continue
    }
    if {$param=="--unset-warning-plot"} {
      set plot_input_warning false
      continue
    }
    if {$param=="--set-warning-plot"} {
      set plot_input_warning true
      continue
    }
    if {$param=="--unset-warning-batch"} {
      set batch_maxima_input_warning false
      set batch_maxima_output_warning false
      continue
    }
    if {$param=="--set-warning-batch"} {
      set batch_maxima_input_warning true
      set batch_maxima_output_warning true
      continue
    }
    if {$param=="--unset-warning-tools"} {
      set import_input_warning false
      set plot_input_warning false
      set batch_maxima_input_warning false
      set batch_maxima_output_warning false
      continue
    }
    if {$param=="--set-warning-tools"} {
      set import_input_warning true
      set plot_input_warning true
      set batch_maxima_input_warning true
      set batch_maxima_output_warning true
      continue
    }

    if {$param=="--logging=none"} {
      set print_logging_maxima [list]
      continue
    }
    if {$param=="--logging=maxima-input"} {
      if {!("maxima-input" in $print_logging_maxima)} {
        lappend print_logging_maxima "maxima-input"
      }
      continue
    }
    if {$param=="--logging=maxima-output"} {
      if {!("maxima-output" in $print_logging_maxima)} {
        lappend print_logging_maxima "maxima-output"
      }
      continue
    }
    if {$param=="--logging=maxima-status"} {
      if {!("maxima-status" in $print_logging_maxima)} {
        lappend print_logging_maxima "maxima-status"
      }
      continue
    }
    if {$param=="--logging=maxima"} {
      if {!("maxima-input" in $print_logging_maxima)} {
        lappend print_logging_maxima "maxima-input"
      }
      if {!("maxima-output" in $print_logging_maxima)} {
        lappend print_logging_maxima "maxima-output"
      }
      if {!("maxima-status" in $print_logging_maxima)} {
        lappend print_logging_maxima "maxima-status"
      }
      continue
    }
    if {$param=="--logging=minima-status"} {
      if {!("minima-status" in $print_logging_maxima)} {
        lappend print_logging_maxima "minima-status"
      }
      continue
    }
    if {$param=="--logging=minima"} {
      if {!("minima-status" in $print_logging_maxima)} {
        lappend print_logging_maxima "minima-status"
      }
      continue
    }
    if {$param=="--logging=all"} {
      if {!("maxima-input" in $print_logging_maxima)} {
        lappend print_logging_maxima "maxima-input"
      }
      if {!("maxima-output" in $print_logging_maxima)} {
        lappend print_logging_maxima "maxima-output"
      }
      if {!("maxima-status" in $print_logging_maxima)} {
        lappend print_logging_maxima "maxima-status"
      }
      if {!("minima-status" in $print_logging_maxima)} {
        lappend print_logging_maxima "minima-status"
      }
      continue
    }
    if {$param=="--logging-limit"} {
      set print_logging_write_start_only true
      continue
    }

    # possibly allowed in batch mode only
    if {$param=="--allow-maxima-setting"} {
      set allow_maxima_system_variables true
      continue
    }
    if {$param=="--disallow-maxima-setting"} {
      set allow_maxima_system_variables false
      continue
    }

    if {$param=="--display-adjust-none"} {
      set mxm_display_exp_dstar false
      set mxm_display_sorted_ratsimp false
      continue
    }
    if {$param=="--display-adjust-exp"} {
      set mxm_display_exp_dstar true
      continue
    }
    if {$param=="--display-adjust-ratsimp"} {
      set mxm_display_sorted_ratsimp true
      continue
    }

    # these two options probably to remove, since Tcl/Tk 8.6.8 was released,
    # containing the clipboard-manager--related changes/fixes;
    if {$param=="--clipboard-selectionless"} {
      set minima_clipboard_without_selection true
      continue
    }
    if {$param=="--clipboard-selection"} {
      set minima_clipboard_force_selection true
      continue
    }

    if {$param=="--batch-end=quit"} {
      set batch_maxima_end_automatic true
      continue
    }
    if {$param=="--batch-end=clean"} {
      set batch_maxima_end_automatic false
      set batch_maxima_end_cleaning true
      continue
    }
    if {$param=="--batch-end=wait"} {
      set batch_maxima_end_automatic false
      set batch_maxima_end_cleaning false
      continue
    }

    # the "--make-line-endings" option used for both modes
    # (in stack-mode used when not in lisp/debug modes only)
    if {$param=="--make-line-endings"} {
      set make_line_endings_asked true
      continue
    }

    # the "--text-batch-action" option for batch-mode only
    if {$param=="--text-batch-action=none"} {
      set text_selection_action_asked "none"
      continue
    }
    if {$param=="--text-batch-action=plot"} {
      set text_selection_action_asked "plot"
      continue
    }
    if {$param=="--text-batch-action=batch"} {
      set text_selection_action_asked "batch"
      continue
    }
    if {$param=="--text-batch-action=batch-inout"} {
      set text_selection_action_asked "batch-inout"
      continue
    }

    # the "--text-stack-action" option for stack-mode only
    if {$param=="--text-stack-action=none"} {
      set text_stack_action_asked "none"
      continue
    }
    if {$param=="--text-stack-action=command-only"} {
      set text_stack_action_asked "command-only"
      continue
    }
    if {$param=="--text-stack-action=selection-only"} {
      set text_stack_action_asked "selection-only"
      continue
    }
    if {$param=="--text-stack-action=selection-command"} {
      set text_stack_action_asked "selection-command"
      continue
    }

    # the "--text-stack-output" option for stack-mode only,
    # since batch-mode only does it this the "along" way;
    if {$param=="--text-stack-output=along"} {
      set text_selection_output_along_asked true
      continue
    }
    if {$param=="--text-stack-output=set"} {
      set text_selection_output_along_asked false
      continue
    }

    # this one option for the second set of plot-wise focusing;
    if {$param=="--plot-focus-full"} {
      set gnuplot_keep_focused_full true
      continue
    }
    if {$param=="--plot-focus=not"} {
      set gnuplot_keep_focused_idx 0
      continue
    }
    if {$param=="--plot-focus=half"} {
      set gnuplot_keep_focused_idx 1
      continue
    }
    if {$param=="--plot-focus=one"} {
      set gnuplot_keep_focused_idx 2
      continue
    }
    if {$param=="--plot-focus=two"} {
      set gnuplot_keep_focused_idx 3
      continue
    }
    if {$param=="--plot-topmost"} {
      set gnuplot_keep_always_topmost true
      continue
    }

    if {$param=="--omit-autotag"} {
      set omit_auto_tildes true
      continue
    }
    if {$param=="--hide-codetags"} {
      set hide_tag_tildes true
      continue
    }

    if {$param=="--block-fill"} {
      set highlight_text_blocks_mode 0
      continue
    }
    if {$param=="--block-line"} {
      set highlight_text_blocks_mode 1
      continue
    }
    if {$param=="--block-side"} {
      set highlight_text_blocks_mode 2
      continue
    }

    if {$param=="--command-lines"} {
      set show_cmmd_highlight_line_nums true
      continue
    }

    if {$param=="--with-eval-time"} {
      set show_time_elapsed true
      continue
    }
    if {$param=="--without-eval-time"} {
      set show_time_elapsed false
      continue
    }
    if {$param=="--include-input"} {
      set include_maxima_input_default true
      continue
    }
    if {$param=="--exclude-input"} {
      set include_maxima_input_default false
      continue
    }

    # these --command-...-... params used for the both stack/batch modes;
    if {$param=="--command-select-only"} {
      set test_command_select_first true
      set test_command_whole_then false
      continue
    }
    if {$param=="--command-whole-only"} {
      set test_command_select_first false
      set test_command_whole_then true
      continue
    }
    if {$param=="--command-select-whole"} {
      set test_command_select_first true
      set test_command_whole_then true
      continue
    }

    if {$param=="--plot-select-only"} {
      set test_plot_input_select_first true
      set test_plot_input_whole_then false
      continue
    }
    if {$param=="--plot-whole-only"} {
      set test_plot_input_select_first false
      set test_plot_input_whole_then true
      continue
    }
    if {$param=="--plot-select-whole"} {
      set test_plot_input_select_first true
      set test_plot_input_whole_then true
      continue
    }

    if {$param=="--batch-prepend-enabled"} {
      set mxmbatch_input_prepend_use true
      continue
    }
    if {$param=="--batch-prepend-disabled"} {
      set mxmbatch_input_prepend_use false
      continue
    }
    if {$param=="--batch-prepend-regular"} {
      set command_onset_use_along_commands true
      continue
    }
    if {$param=="--batch-prepend-ini"} {
      set command_onset_use_along_commands false
      continue
    }

    if {$param=="--backup-on-save=tilde"} {
      set backup_on_save_tilde true
      continue
    }
    if {$param=="--backup-on-save=none"} {
      set backup_on_save_tilde false
      continue
    }

    if {$param=="--with-linked-tags-count"} {
      set show_tags_found_count_info true
      continue
    }
    if {$param=="--without-linked-tags-count"} {
      set show_tags_found_count_info false
      continue
    }

    if {$param=="--linked-tags-with-link"} {
      set show_tags_found_with_link true
      continue
    }
    if {$param=="--linked-tags-without-link"} {
      set show_tags_found_with_link false
      continue
    }

    if {$param=="--spell-concat-permissive"} {
      set spch_default_allows_concat true
      continue
    }
    if {$param=="--spell-concat-restrictive"} {
      set spch_default_allows_concat false
      continue
    }
    if {$param=="--spell-ascii-only"} {
      set spch_ascii_only true
      continue
    }

    if {$param=="--hide-labels"} {
      set with_labels true
      continue
    }
    if {$param=="--show-labels"} {
      set with_labels true
      continue
    }
    if {$param=="--unbind-labels"} {
      set bind_labels false
      continue
    }
    if {$param=="--bind-labels"} {
      set bind_labels true
      continue
    }
    if {$param=="--2dim-center"} {
      set line_display_mode_current 4
      continue
    }
    if {$param=="--2dim-left"} {
      set line_display_mode_current 3
      continue
    }
    if {$param=="--1dim-short"} {
      set line_display_mode_current 2
      continue
    }
    if {$param=="--1dim-split"} {
      set line_display_mode_current 1
      continue
    }
    if {$param=="--1dim-long"} {
      set line_display_mode_current 0
      continue
    }
    if {[string equal -length $param_linesize_start_length $param_linesize_start $param]} {
      set line_size [string range $param "$param_linesize_start_length" "end"]
      set line_size [string trimleft $line_size "0"]
      if {([string length $line_size]<5) && [string is ascii -strict $line_size] && [string is digit -strict $line_size]} {
        if {($line_size>=$maxima_line_size_min) && ($line_size<=$maxima_line_size_max)} {
          lset info_line_size 0 $line_size
        }
      }
      continue
    }

    if {[string equal -length $param_clockticks_start_length $param_clockticks_start $param]} {
      set tick_count [string range $param "$param_clockticks_start_length" "end"]
      set tick_count [string trimleft $tick_count "0"]
      if {([string length $tick_count]<8) && [string is ascii -strict $tick_count] && [string is digit -strict $tick_count]} {
        if {($tick_count>=10)} {
          set system_clock_ticks $tick_count
        }
      }
      continue
    }

    if {[string equal -length $param_loggingstart_start_length $param_loggingstart_start $param]} {
      set logstart_length_filed [string range $param "$param_loggingstart_start_length" "end"]
      set logstart_length [string trimleft $logstart_length_filed "0"]
      if {($logstart_length == "") && ($logstart_length_filed != "")} {
        set print_logging_write_start_only false
        continue
      }
      if {[string is ascii -strict $logstart_length] && [string is digit -strict $logstart_length]} {
        set print_logging_write_start_only true
        if {[string length $logstart_length]>0} {
          if {[string length $logstart_length]>3} {
            set print_logging_write_start_length 1000
          } {
            set print_logging_write_start_length $logstart_length
            if {$logstart_length < 100} {
              set print_logging_write_start_length 100
            }
          }
        }
      }
      continue
    }

    if {[string equal -length $param_seclineborder_start_length $param_seclineborder_start $param]} {
      set secline_width_filed [string range $param "$param_seclineborder_start_length" "end"]
      set secline_width [string trimleft $secline_width_filed "0"]
      if {($secline_width == "") && ($secline_width_filed != "")} {
        set section_line_border_width 0
        continue
      }
      if {[string is ascii -strict $secline_width] && [string is digit -strict $secline_width]} {
        if {[string length $secline_width]>1} {
          set section_line_border_width 10
          continue
        }
        set section_line_border_width $secline_width
      }
      continue
    }

    if {[string equal -length $param_tab_length_start_length $param_tab_length_start $param]} {
      set test_tablen_filed [string range $param "$param_tab_length_start_length" "end"]
      set test_tablen [string trimleft $test_tablen_filed "0"]
      if {($test_tablen == "") && ($test_tablen_filed != "")} {
        # leaving default use_tab_length value for 0 specified
        continue
      }
      if {[string is ascii -strict $test_tablen] && [string is digit -strict $test_tablen]} {
        if {[string length $test_tablen]>3} {
          # tab lengths are allowed for small sizes only
          set use_tab_length $use_tab_length_max
          continue
        }
        if {$test_tablen >= $use_tab_length_max} {
          set use_tab_length $use_tab_length_max
          continue
        }
        set use_tab_length $test_tablen
      }
      continue
    }

    if {$param=="--without-names-functions"} {
      set names_from_functions_to_load false
      continue
    }
    if {$param=="--without-names-values"} {
      set names_from_values_to_load false
      continue
    }
    if {[string equal -length $param_withnames_start_length $param_withnames_start $param]} {
      set with_names [string range $param "$param_withnames_start_length" "end"]
      if {[string length $with_names]>0} {
        lappend names_from_files_path_list $with_names
      }
      continue
    }

    if {[string equal -length $param_fileopen_start_length $param_fileopen_start $param]} {
      set file_open [string range $param "$param_fileopen_start_length" "end"]
      if {[string length $file_open]>0} {
        set file_path_to_open $file_open
      }
      continue
    }

    if {[string equal -length $param_uselisp_start_length $param_uselisp_start $param]} {
      set use_lisp [string range $param "$param_uselisp_start_length" "end"]
      if {[string length $use_lisp]>0} {
        set lisp_flavor_to_use [string tolower $use_lisp]
        if {$lisp_flavor_to_use == "openmcl"} {
          set lisp_flavor_to_use "ccl"
        }
      }
      continue
    }
    if {[string equal -length $param_lisp_options_start_length $param_lisp_options_start $param]} {
      set use_lisp_options [string range $param "$param_lisp_options_start_length" "end"]
      if {[string length $use_lisp_options]>0} {
        set lisp_options_to_set $use_lisp_options
      }
      continue
    }

    if {[string equal -length $param_lastdir_start_length $param_lastdir_start $param]} {
      set last_dir_param [string range $param "$param_lastdir_start_length" "end"]
      if {([string length $last_dir_param]>0) && [file exists $last_dir_param]} {
        if {[file isfile $last_dir_param]} {
          set last_dir_param [file dirname $last_dir_param]
        }
        set last_dir $last_dir_param
      }
      continue
    }

    if {[string equal -length $param_maximauserdir_start_length $param_maximauserdir_start $param]} {
      set mxm_userdir [string range $param "$param_maximauserdir_start_length" "end"]
      if {[string length $mxm_userdir]>0} {
        set maxima_userdir_to_set $mxm_userdir
      }
      continue
    }
    if {[string equal -length $param_maximainit_start_length $param_maximainit_start $param]} {
      set mxm_initname [string range $param "$param_maximainit_start_length" "end"]
      if {[string length $mxm_initname]>0} {
        set maxima_initname_to_set $mxm_initname
      }
      continue
    }
    if {[string equal -length $param_maximapath_start_length $param_maximapath_start $param]} {
      set mxm_runpath [string range $param "$param_maximapath_start_length" "end"]
      if {[string length $mxm_runpath]>0} {
        set maxima_path $mxm_runpath
      }
      continue
    }
    if {[string equal -length $param_maximaversion_start_length $param_maximaversion_start $param]} {
      set mxm_version [string range $param "$param_maximaversion_start_length" "end"]
      if {[string length $mxm_version]>0} {
        set maxima_version_to_use $mxm_version
      }
      continue
    }

    if {[string equal -length $param_lisppath_start_length $param_lisppath_start $param]} {
      set test_path [string range $param "$param_lisppath_start_length" "end"]
      if {[string length $test_path]>0} {
        set maxima_lisp_path $test_path
      }
      continue
    }
    if {[string equal -length $param_corepath_start_length $param_corepath_start $param]} {
      set test_path [string range $param "$param_corepath_start_length" "end"]
      if {[string length $test_path]>0} {
        set maxima_load_path $test_path
      }
      continue
    }
    if {[string equal -length $param_imagepath_start_length $param_imagepath_start $param]} {
      set test_path [string range $param "$param_imagepath_start_length" "end"]
      if {[string length $test_path]>0} {
        set maxima_exec_path $test_path
      }
      continue
    }
    if {[string equal -length $param_sbclhome_start_length $param_sbclhome_start $param]} {
      set test_path [string range $param "$param_sbclhome_start_length" "end"]
      if {[string length $test_path]>0} {
        set sbcl_home_path $test_path
      }
      continue
    }

    if {[string equal -length $param_gnuplotpath_start_length $param_gnuplotpath_start $param]} {
      set test_gnuplotpath [string range $param "$param_gnuplotpath_start_length" "end"]
      if {[string length $test_gnuplotpath]>0} {
        set gnuplot_path $test_gnuplotpath
      }
      continue
    }

    if {[string equal -length $param_spellpath_start_length $param_spellpath_start $param]} {
      set test_spellpath [string range $param "$param_spellpath_start_length" "end"]
      if {[string length $test_spellpath]>0} {
        set spell_path $test_spellpath
      }
      continue
    }

    if {[string equal -length $param_temppath_start_length $param_temppath_start $param]} {
      set test_tmppath [string range $param "$param_temppath_start_length" "end"]
      if {[string length $test_tmppath]>0} {
        if {[file exists $test_tmppath]} {
          if {[file isdirectory $test_tmppath]} {
            if {[file writable $test_tmppath]} {
              set asked_temporary_path $test_tmppath
            } {
              lappend system_startup_warnings [list "the requested TEMP path is not writable: $test_tmppath"]
            }
          } {
            lappend system_startup_warnings [list "the requested TEMP path is not directory: $test_tmppath"]
          }
        } {
          lappend system_startup_warnings [list "the requested TEMP path not found: $test_tmppath"]
        }
      }
      continue
    }

    if {[string equal -length $param_outputplacing_start_length $param_outputplacing_start $param]} {
      set placing_param [string tolower [string range $param "$param_outputplacing_start_length" "end"]]
      foreach placing_info [list [list "insitu" "s"] [list "volatile" "v"] [list "toend" "e"]] {
        set placing_info_name [lindex $placing_info 0]
        if {$placing_param==$placing_info_name} {
          set placing_info_char [lindex $placing_info 1]
          set placing_info_idx [lsearch -exact -ascii $info_label_output_modes $placing_info_char]
          if {$placing_info_idx!=-1} {
            set info_label_output_index $placing_info_idx
          }
          break
        }
      }
      continue
    }

    if {[string equal -length $param_clipboardforcing_start_length $param_clipboardforcing_start $param]} {
      set clp_param [string range $param "$param_clipboardforcing_start_length" "end"]
      if {[string length $clp_param]>0} {
        set clipboard_enforcer $clp_param
      }
      continue
    }

    if {[string equal -length $param_batch_prepend_command_length $param_batch_prepend_command $param]} {
      set prep_comm_param [string range $param "$param_batch_prepend_command_length" "end"]
      if {[string length $prep_comm_param]>0} {
        set mxmbatch_input_prepend $prep_comm_param
      }
      continue
    }
  }

  if {[llength $names_from_files_path_list]>0} {
    set names_from_files_to_load true
  }
}

take_commandline_parameters

set default_menu_offset_x 2
set default_menu_offset_y 2
set menu_selected_background "#d0d0ff"

if {$gnuplot_keep_focused_full} {
  incr gnuplot_keep_focused_idx_default $gnuplot_keep_focused_idx_full_inc
  if {$gnuplot_keep_focused_idx != 0} {
    incr gnuplot_keep_focused_idx $gnuplot_keep_focused_idx_full_inc
  }
}
set gnuplot_keep_focused_last_nonzero_idx \
  [expr {($gnuplot_keep_focused_idx != 0) ? $gnuplot_keep_focused_idx : $gnuplot_keep_focused_idx_default}]

set need_to_set_lisp_path_to_cmucl false
set need_to_set_lisp_path_to_sbcl false
set need_to_set_lisp_path_to_ccl false
set need_to_set_lisp_path_to_ccl64 false
set need_to_set_lisp_path_to_clisp false

if {($maxima_lisp_path!="") || ($maxima_load_path!="")} {
  if {!($lisp_flavor_to_use in [list "cmucl" "sbcl" "clisp" "ccl" "ccl64"])} {
    lappend system_startup_warnings [list "direct startup of Maxima via Lisp and loadable Maxima-core paths is only supported for CMUCL, SBCL, CLISP and CCL(64) now"]
    set maxima_lisp_path ""
    set maxima_load_path ""
  } {
    if {$maxima_load_path==""} {
      lappend system_startup_warnings [list "direct startup of Maxima via Lisp and loadable Maxima core requires the core path to be set:"]
      lappend system_startup_warnings [list "      --maxima-load-path=<path_to_the_maxima_loadable_core>"]
      set maxima_lisp_path ""
    }
  }
}
if {$maxima_load_path!=""} {
  if {$maxima_lisp_path==""} {
    if {$lisp_flavor_to_use == "cmucl"} {
      set need_to_set_lisp_path_to_cmucl true
    } elseif {$lisp_flavor_to_use == "sbcl"} {
      set need_to_set_lisp_path_to_sbcl true
    } elseif {$lisp_flavor_to_use == "clisp"} {
      set need_to_set_lisp_path_to_clisp true
    } elseif {$lisp_flavor_to_use == "ccl64"} {
      set need_to_set_lisp_path_to_ccl64 true
    } {
      set need_to_set_lisp_path_to_ccl true
    }
  }
}

set need_to_set_lisp_path_to_gcl false
set need_to_set_lisp_path_to_ecl false
if {$maxima_exec_path!=""} {
  if {!($lisp_flavor_to_use in [list "gcl" "ecl"])} {
    lappend system_startup_warnings [list "direct startup of Maxima via executable Maxima-image path is only supported for GCL and ECL now"]
    set maxima_exec_path ""
  }
}
if {$maxima_exec_path!=""} {
  if {$maxima_lisp_path==""} {
    if {$lisp_flavor_to_use == "gcl"} {
      set need_to_set_lisp_path_to_gcl true
    } {
      set need_to_set_lisp_path_to_ecl true
    }
  }
}

proc get_batch_ending_label_char {} {
  global batch_maxima_end_automatic
  global batch_maxima_end_cleaning

  if {$batch_maxima_end_automatic} {
    return "x"
  }
  if {$batch_maxima_end_cleaning} {
    return "!"
  }
  return "+"
}

set last_dir_first $last_dir

set minima_started_via_wish true
if {[catch {tk windowingsystem}]} {
  set minima_started_via_wish false
}

package require Tk
package require Ttk

# under Linux and/or related systems
set is_linux_like_known ""
proc is_linux_like {} {
  global is_linux_like_known

  if {$is_linux_like_known!=""} {
    return $is_linux_like_known
  }

  if {[string tolower [tk windowingsystem]]=="x11"} {
    set is_linux_like_known true
    return true
  }
  set is_linux_like_known false
  return false
}

set is_under_linux_known ""
proc is_under_linux {} {
  global is_under_linux_known

  if {$is_under_linux_known!=""} {
    return $is_under_linux_known
  }

  if {[string tolower [tk windowingsystem]]!="x11"} {
    set is_under_linux_known false
    return false
  }

  set at_linux false
  catch {
    if {[string first "linux" [string tolower $tcl_platform(os)]] != -1} {
      set at_linux true
    }
  }

  set is_under_linux_known $at_linux
  return $at_linux
}

set is_under_freebsd_known ""
proc is_under_freebsd {} {
  global is_under_freebsd_known

  if {$is_under_freebsd_known!=""} {
    return $is_under_freebsd_known
  }

  if {[string tolower [tk windowingsystem]]!="x11"} {
    set is_under_freebsd_known false
    return false
  }

  set at_freebsd false
  catch {
    if {[string first "freebsd" [string tolower $tcl_platform(os)]] != -1} {
      set at_freebsd true
    }
  }

  set is_under_freebsd_known $at_freebsd
  return $at_freebsd
}

# under ReactOS and/or related systems
set is_wine_like_known ""
proc is_wine_like {} {
  global is_wine_like_known

  if {$is_wine_like_known!=""} {
    return $is_wine_like_known
  }

  if {[string equal -nocase -length 3 [tk windowingsystem] "win"]} {
    set is_wine_like_known true
    return true
  }

  set is_wine_like_known false
  return false
}

set is_under_wine_known ""
proc is_under_wine {} {
  global env
  global is_under_wine_known

  if {$is_under_wine_known!=""} {
    return $is_under_wine_known
  }

  if {![is_wine_like]} {
    set is_under_wine_known false
    return false
  }

  set env_names [array names env]
  foreach one_env_name [list "XDG_CONFIG_DIRS" "XDG_DATA_DIRS" "XDG_GREETER_DATA_DIR"] {
    foreach one_path_start [list "/etc/" "/usr/" "/var/" "/run/"] {
      if {$one_env_name in $env_names} {
        if {[string equal -nocase -length [string length $one_path_start] $one_path_start $env($one_env_name)]} {
          set is_under_wine_known true
          return true
        }
      }
    }
  }

  set is_under_wine_known false
  return false
}

set is_darwin_like_known ""
proc is_darwin_like {} {
  global is_darwin_like_known

  if {$is_darwin_like_known!=""} {
    return $is_darwin_like_known
  }

  if {[string tolower [tk windowingsystem]]=="aqua"} {
    set is_darwin_like_known true
    return true
  }

  set is_darwin_like_known false
  return false
}

# for wine-like systems to use (show) interrupt (button)
# only if we can catch the ctrl-c signal (needing TWAPI for that);
set minima_uses_twapi false
set to_show_interrupt_button true
set with_common_ctrlc true

if {[is_wine_like]} {
  set minima_uses_twapi true
  # testing presence of TWAPI
  if {[catch {
    package require twapi
  }]} {
    set minima_uses_twapi false
  }
}

# setting the default mode according to platform
if {$editor_batch_state==""} {
  if {[string first "stack" [string tolower $argv0]] != -1} {
    set editor_batch_state false
  }
  if {[string first "batch" [string tolower $argv0]] != -1} {
    if {$editor_batch_state==""} {
      set editor_batch_state true
    } {
      if {[string last "batch" [string tolower $argv0]] > [string last "stack" [string tolower $argv0]]} {
        set editor_batch_state true
      }
    }
  }
}
if {$editor_batch_state==""} {
  set editor_batch_state [expr {[is_wine_like] ? true : false}]
}

proc is_batch_mode {} {
  global editor_batch_state
  return $editor_batch_state
}
proc is_stack_mode {} {
  global editor_batch_state
  return [expr {$editor_batch_state ? false : true}]
}

if {true} {
  if {$make_line_endings_asked} {
    set make_line_endings true
  }
}

if {[is_stack_mode]} {
  set allow_maxima_system_variables false
}
if {[is_batch_mode]} {
  if {$allow_maxima_system_variables == ""} {
    set allow_maxima_system_variables false
  }
}

proc get_tempfile {{asked_suffix ""}} {
  global asked_temporary_path

  set hnd_var ""
  set path_var ""

  set part_access [list RDWR CREAT EXCL TRUNC]
  set part_access_renamed [list RDWR CREAT TRUNC]
  set part_perm 0600

  if {$asked_temporary_path==""} {
    if {[catch {
      if {$asked_suffix==""} {
        set hnd_var [file tempfile path_var]
      } {
        set end_part ".$asked_suffix"
        set end_part_len_test [string length $asked_suffix]
        set hnd_var [file tempfile path_var $end_part]
        if {[string range $path_var "end-$end_part_len_test" "end"]!=$end_part} {
          set path_var_new [strcat $path_var $end_part]
          close $hnd_var
          set hnd_var ""
          file rename -force -- $path_var $path_var_new
          set path_var $path_var_new
          set hnd_var [open $path_var $part_access_renamed $part_perm]
        }
      }
    }]} {
      if {$hnd_var != ""} {
        catch {close $hnd_var}
      }
      if {$path_var != ""} {
        catch {file delete $path_var}
      }
      return [list "" ""]
    }
    return [list $hnd_var $path_var]
  }

  set test_count 0
  while {true} {
    incr test_count
    if {$test_count>10000} {
      break
    }

    set rnd_part [strcat [expr {int(10000*(1+rand()))}] [expr {int(10000*(1+rand()))}]]
    if {[catch {
      set count_in [glob -nocomplain -directory $asked_temporary_path "$rnd_part*"]
    }]} {
      set hnd_var ""
      set path_var ""
      break
    }
    if {[llength $count_in]>0} {
      set hnd_var ""
      set path_var ""
      continue
    }

    set path_var [file join $asked_temporary_path $rnd_part]
    if {$asked_suffix != ""} {
      append path_var "." $asked_suffix
    }
    if {[catch {
      set hnd_var [open $path_var $part_access $part_perm]
    }]} {
      if {$hnd_var != ""} {
        catch {close $hnd_var}
      }
      if {$path_var != ""} {
        catch {file delete $path_var}
      }
      set hnd_var ""
      set path_var ""
    }
    break
  }

  if {($hnd_var=="") || ($path_var=="")} {
    return [list "" ""]
  }

  return [list $hnd_var $path_var]
}

set print_logging_maxima_input [expr {("maxima-input" in $print_logging_maxima) ? true : false}]
set print_logging_maxima_output [expr {("maxima-output" in $print_logging_maxima) ? true : false}]
set print_logging_maxima_status [expr {("maxima-status" in $print_logging_maxima) ? true : false}]
set print_logging_minima_status [expr {("minima-status" in $print_logging_maxima) ? true : false}]

set print_logging_some [expr {$print_logging_maxima_input || $print_logging_maxima_output || $print_logging_minima_status || $print_logging_maxima_status}]
set print_logging_should_clear_console_rest false
set print_logging_num_lines_got_out 0
proc print_logging {txt} {
  global minima_uses_twapi
  global print_logging_should_clear_console_rest
  global print_logging_write_start_only
  global print_logging_write_start_length
  global print_logging_num_lines_got_out

  if {$txt == ""} {
    return
  }

  if {[is_wine_like] && (![is_under_wine])} {
    if {$minima_uses_twapi} {
      set cur_pos_line 0
      catch {
        set cur_pos_info [lindex [twapi::get_console_screen_buffer_info stdout -cursorpos] 1]
        set cur_pos_line [lindex $cur_pos_info 1]
        if {(!$print_logging_write_start_only) && ($cur_pos_line >= 1000)} {
          twapi::fill_console stdout -fillchar " " -mode "line" -position [list 0 0] -numlines 600 -numcols 85
          twapi::set_console_cursor_position stdout [list 0 0]
          set print_logging_should_clear_console_rest true
        } elseif {$print_logging_should_clear_console_rest && ($cur_pos_line > 400)} {
          set line_clean_start 600
          set line_clean_count 600
          if {$cur_pos_line >= 600} {
            set line_clean_diff [expr {1 + $cur_pos_line - 600}]
            incr line_clean_start $line_clean_diff
            incr line_clean_count "-$line_clean_count"
          }
          set print_logging_should_clear_console_rest false
          twapi::fill_console stdout -fillchar " " -mode "line" -position [list 0 600] -numlines 600 -numcols 85
        }
      }
      if {$print_logging_write_start_only && ($print_logging_write_start_length <= $cur_pos_line)} {
        return
      }
      catch {twapi::console_write stdout $txt -newlinemode "line" -restoreposition false}
    }
  } {
    if {$print_logging_write_start_only} {
      if {$print_logging_write_start_length <= $print_logging_num_lines_got_out} {
        return
      } {
        incr print_logging_num_lines_got_out [expr {[llength [split $txt "\n"]] - 1}]
      }
    }
    puts -nonewline stdout $txt
  }
}

#set can_kill_label true
set in_lisp_mode false
set in_debug_mode false

proc set_lisp_mode {value} {
  global in_lisp_mode
  global print_logging_maxima_status

  set in_lisp_mode $value
  if {$print_logging_maxima_status} {
    print_logging "\nXXX in-lisp-mode: $in_lisp_mode\n"
  }

  # in catch since "set_lisp_mode" defined before "show_info_label_endings_state" is defined
  catch {show_info_label_endings_state}
}
proc get_lisp_mode {} {
  global in_lisp_mode
  return $in_lisp_mode
}
proc set_debug_mode {value} {
  global in_debug_mode
  global print_logging_maxima_status

  set in_debug_mode $value
  if {$print_logging_maxima_status} {
    print_logging "\nXXX in-debug-mode: $in_debug_mode\n"
  }

  # in catch since "set_debug_mode" defined before "show_info_label_endings_state" is defined
  catch {show_info_label_endings_state}
}
proc get_debug_mode {} {
  global in_debug_mode
  return $in_debug_mode
}

proc is_stack_mode_common {} {
  global in_lisp_mode
  global in_debug_mode

  if {![is_stack_mode]} {
    return false
  }
  if {$in_lisp_mode || $in_debug_mode} {
    return false
  }

  return true
}

# the "stack_mode_text_based_command_modes_use_idx" variable is reset at "preset_textwise_command_mode"
set stack_mode_text_based_command_modes_use_idx 2
set stack_mode_text_based_command_modes [list \
  [list "only using selection in text (if any), output placed below selection" [list true true false false]] \
  [list "only using selection in text (if any)" [list true false false false]] \
  [list "using selection in text if any (output placed below selection), otherwise using command" [list true true true false]] \
  [list "using selection in text if any, otherwise using command" [list true false true false]] \
  [list "only using command" [list false false true false]] \
  [list "no Maxima action from text" [list false false false false]] \
]

# the "batch_mode_text_based_command_modes_use_idx" variable is reset at "preset_textwise_command_mode"
# the "does_text_batch_input" and "does_text_batch_output" expect the current ordering of the list below;
set batch_mode_text_based_command_modes_use_idx 0
set batch_mode_text_based_command_modes [list \
  [list "selection used for batches, output put both to batch window and beneath selection" [list true true false false]] \
  [list "selection used for batches, output put to batch window only" [list true false false false]] \
  [list "selection used for plotting" [list false false false true]] \
  [list "selection used neither for batches nor for plotting" [list false false false false]] \
]

proc preset_textwise_command_mode {} {
  global text_selection_action_asked
  global text_stack_action_asked
  global text_selection_output_along_asked
  global stack_mode_text_based_command_modes_use_idx
  global batch_mode_text_based_command_modes_use_idx

  if {[is_stack_mode]} {
    if {$text_stack_action_asked=="none"} {
      set stack_mode_text_based_command_modes_use_idx 5
    } elseif {$text_stack_action_asked=="command-only"} {
      set stack_mode_text_based_command_modes_use_idx 4
    } elseif {$text_stack_action_asked=="selection-command"} {
      if {!$text_selection_output_along_asked} {
        set stack_mode_text_based_command_modes_use_idx 3
      } {
        set stack_mode_text_based_command_modes_use_idx 2
      }
    } elseif {$text_stack_action_asked=="selection-only"} {
      if {!$text_selection_output_along_asked} {
        set stack_mode_text_based_command_modes_use_idx 1
      } {
        set stack_mode_text_based_command_modes_use_idx 0
      }
    }
  }

  if {[is_batch_mode]} {
    if {$text_selection_action_asked=="batch-inout"} {
      set batch_mode_text_based_command_modes_use_idx 0
    }
    if {$text_selection_action_asked=="batch"} {
      set batch_mode_text_based_command_modes_use_idx 1
    }
    if {$text_selection_action_asked=="plot"} {
      set batch_mode_text_based_command_modes_use_idx 2
    }
    if {$text_selection_action_asked=="none"} {
      set batch_mode_text_based_command_modes_use_idx 3
    }
  }
}
preset_textwise_command_mode

proc does_text_batch_input {} {
  global batch_mode_text_based_command_modes_use_idx

  if {![is_batch_mode]} {
    return false
  }

  if {$batch_mode_text_based_command_modes_use_idx > 1} {
    return false
  }

  return true
}
proc does_text_batch_output {} {
  global batch_mode_text_based_command_modes_use_idx

  if {![is_batch_mode]} {
    return false
  }

  if {$batch_mode_text_based_command_modes_use_idx > 0} {
    return false
  }

  return true
}

if {[is_batch_mode]} {
  lappend line_display_modes_list $line_display_mode_tex
}

proc set_maxima_env_vars {} {
  global lisp_options_to_set
  global sbcl_home_path
  global env

  if {$lisp_options_to_set!=""} {
    set env_name "MAXIMA_LISP_OPTIONS"
    set env($env_name) $lisp_options_to_set
  }

  if {$sbcl_home_path!=""} {
    set env_name "SBCL_HOME"
    set env($env_name) $sbcl_home_path
  }
}
set_maxima_env_vars

# using ".twt" suffix for "text-with-tags" rather than ".mnm",
# since it can be opened/edited by (mostly) anything, not just Minima;
set minima_file_extension_base "twt"
set minima_file_extension [strcat "." $minima_file_extension_base]

set tag_border_size 1

proc remove_torn_menus {} {
  # the used commands are not defined initially, thus in catch here;
  # probably should use listboxes (with grabbed focus) instead of these;
  catch {remove_tags_insertion_menu}
  catch {remove_keyword_insertion_menu}
  catch {remove_otl_paths_menu_shown}
  catch {remove_tags_menu_shown}
  catch {remove_spch_words_menu_shown}
}

set scr_width 1280
set scr_height 800
catch {set scr_width [winfo screenwidth .]}
catch {set scr_height [winfo screenheight .]}

proc get_screen_width {} {
  global scr_width
  return $scr_width
}
proc get_screen_height {} {
  global scr_height
  return $scr_height
}
proc get_max_menu_entry_count {} {
  if {[get_screen_height] >= 550} {
    return 25
  }
  if {[get_screen_height] >= 440} {
    return 20
  }
  if {[get_screen_height] >= 220} {
    return 10
  }
  return 5
}

tk appname "TGM Minima"
set topw .
if {true} {
  wm title $topw $app_title_base

  if {$scr_width<=1000} {
    set app_pos_y 10
  } {
    set app_pos_y [expr {50 + [::tcl::mathfunc::int [expr {[::tcl::mathfunc::rand] * 11}]] - 5}]
  }
  if {$scr_width<=1000} {
    set app_pos_x 10
  } elseif {$scr_width<=1100} {
    set app_pos_x 25
  } {
    set app_pos_x [expr {125 + [::tcl::mathfunc::int [expr {[::tcl::mathfunc::rand] * 21}]] - 10}]
  }

  wm geometry $topw [strcat \
    [expr {($scr_width>1000) ? "960" : "760"}] \
    "x" [expr {($scr_height>710) ? 640 : (($scr_height<470) ? 400 : ($scr_height-70))}] \
    "+" $app_pos_x "+" $app_pos_y]

  wm minsize $topw \
    [expr {($scr_width>940) ? 900 : (($scr_width>740) ? 700: 400)}] \
    [expr {($scr_height>600) ? 500 : (($scr_height<400) ? 300 : ($scr_height-100))}]

  wm resizable $topw yes yes
  wm protocol $topw WM_DELETE_WINDOW {want_to_exit}
}
set maxima_ready false

if {$need_to_set_lisp_path_to_cmucl} {
  set maxima_lisp_path "lisp"
  if {[is_wine_like]} {
    append maxima_lisp_path ".exe"
  }
}

if {$need_to_set_lisp_path_to_sbcl} {
  set maxima_lisp_path "sbcl"
  if {[is_wine_like]} {
    append maxima_lisp_path ".exe"
  }
}

if {$need_to_set_lisp_path_to_clisp} {
  set maxima_lisp_path "clisp"
  if {[is_wine_like]} {
    append maxima_lisp_path ".exe"
  }
}

if {$need_to_set_lisp_path_to_ccl} {
  if {[is_wine_like]} {
    set maxima_lisp_path "wx86cl.exe"
  } {[is_darwin_like]} {
    set maxima_lisp_path "dx86cl"
  } {[is_under_linux]} {
    set maxima_lisp_path "lx86cl"
  } {[is_under_freebsd]} {
    set maxima_lisp_path "fx86cl"
  } {
    lappend system_startup_warnings [list "unknown form of CCL lisp platform, and path to CCL (--lisp-path=<path>) was not specified"]
    set maxima_lisp_path ""
    set maxima_load_path ""
  }
}
if {$need_to_set_lisp_path_to_ccl64} {
  if {[is_wine_like]} {
    set maxima_lisp_path "wx86cl64.exe"
  } {[is_darwin_like]} {
    set maxima_lisp_path "dx86cl64"
  } {[is_under_linux]} {
    set maxima_lisp_path "lx86cl64"
  } {[is_under_freebsd]} {
    set maxima_lisp_path "fx86cl64"
  } {
    lappend system_startup_warnings [list "unknown form of CCL64 lisp platform, and path to CCL64 (--lisp-path=<path>) was not specified"]
    set maxima_lisp_path ""
    set maxima_load_path ""
  }
}

if {$need_to_set_lisp_path_to_gcl} {
  set maxima_lisp_path "gcl"
  if {[is_wine_like]} {
    append maxima_lisp_path ".exe"
  }
}

if {$need_to_set_lisp_path_to_ecl} {
  set maxima_lisp_path "ecl"
  if {[is_wine_like]} {
    append maxima_lisp_path ".exe"
  }
}

set test_suffixes_maxima [list "mac" "mc" "max" "mxm" "maxima" "mx" "m" "wxm" "demo" "dem"]
set test_suffixes_lisp [list "lisp" "lsp" "lp" "l"]

proc get_maxima_startup_base {} {
  global with_labels
  global lisp_flavor_to_use
  global maxima_lisp_path
  global maxima_load_path
  global lisp_options_to_set
  global maxima_exec_path
  global maxima_path
  global maxima_version_to_use
  global maxima_userdir_to_set
  global maxima_initname_to_set
  global test_suffixes_maxima
  global test_suffixes_lisp

  set maxima_startup ""

  set quiet_mode "--very-quiet"
  if {$with_labels} {
    set quiet_mode "--quiet"
  }
  set without_readline "--disable-readline"

  set start_without_script false

  # regarding startup without maxima script, considering: gcl/ecl (exec) and cmucl/sbcl/clisp/ccl(64) (load);
  if {($maxima_lisp_path!="") && ($maxima_load_path!="")} {
    set lisp_path_param [strcat "\"" [string map [list "\\" "\\\\" "\"" "\\\""] $maxima_lisp_path] "\""]
    set core_path_param [strcat "\"" [string map [list "\\" "\\\\" "\"" "\\\""] $maxima_load_path] "\""]
    if {$lisp_flavor_to_use == "cmucl"} {
      set start_without_script true
      set maxima_startup "|$lisp_path_param"
      if {$lisp_options_to_set != ""} {
        append maxima_startup " $lisp_options_to_set"
      }
      append maxima_startup " -quiet -core $core_path_param"
      append maxima_startup " -eval"
      append maxima_startup " \"(cl-user::run)\""
      append maxima_startup " --"
    }
    if {$lisp_flavor_to_use == "sbcl"} {
      set start_without_script true
      set maxima_startup "|$lisp_path_param --core $core_path_param --noinform"
      if {$lisp_options_to_set != ""} {
        append maxima_startup " $lisp_options_to_set"
      }
      append maxima_startup " --end-runtime-options"
      append maxima_startup " --eval"
      append maxima_startup " \"(cl-user::run)\""
      append maxima_startup " --end-toplevel-options"
      append maxima_startup " --"
    }
    if {$lisp_flavor_to_use in [list "ccl" "ccl64"]} {
      set start_without_script true
      set maxima_startup "|$lisp_path_param -I $core_path_param"
      if {$lisp_options_to_set != ""} {
        append maxima_startup " $lisp_options_to_set"
      }
      append maxima_startup " -e \"(cl-user::run)\""
      append maxima_startup " --"
    }
    if {$lisp_flavor_to_use == "clisp"} {
      set start_without_script true
      set lisp_path_param [strcat "\"" [string map [list "\\" "\\\\" "\"" "\\\""] $maxima_lisp_path] "\""]
      set core_path_param [strcat "\"" [string map [list "\\" "\\\\" "\"" "\\\""] $maxima_load_path] "\""]
      set maxima_startup "|$lisp_path_param"
      if {$lisp_options_to_set != ""} {
        append maxima_startup " $lisp_options_to_set"
      }
      append maxima_startup " -q -M $core_path_param"
      append maxima_startup " \"\""
      append maxima_startup " --"
    }
  }
  if {($maxima_exec_path!="")} {
    set image_path_param [strcat "\"" [string map [list "\\" "\\\\" "\"" "\\\""] $maxima_exec_path] "\""]
    if {$lisp_flavor_to_use == "gcl"} {
      set start_without_script true
      set maxima_startup "|$image_path_param -eval \"(cl-user::run)\""
      if {$lisp_options_to_set != ""} {
        append maxima_startup " $lisp_options_to_set"
      }
      append maxima_startup " -f --"
    }
    if {$lisp_flavor_to_use == "ecl"} {
      set start_without_script true
      set maxima_startup "|$image_path_param"
      if {$lisp_options_to_set != ""} {
        append maxima_startup " $lisp_options_to_set"
      }
      append maxima_startup " --"
    }
  }

  if {!$start_without_script} {
    set maxima_path_param [strcat "\"" [string map [list "\\" "\\\\" "\"" "\\\""] $maxima_path] "\""]
    set maxima_build_param ""
    if {$maxima_version_to_use!=""} {
      set maxima_build_param [strcat "\"--use-version=" [string map [list "\\" "\\\\" "\"" "\\\""] $maxima_version_to_use] "\""]
    }
    set lisp_flavor_param ""
    if {$lisp_flavor_to_use!=""} {
      set lisp_flavor_param [strcat "\"--lisp=" [string map [list "\\" "\\\\" "\"" "\\\""] $lisp_flavor_to_use] "\""]
    }
    set maxima_startup "|"
    if {[is_wine_like]} {
      append maxima_startup " cmd.exe /c "
    }
    append maxima_startup "$maxima_path_param $maxima_build_param $lisp_flavor_param"
  }

  set user_dir_param ""
  set init_file_param ""
  if {$maxima_userdir_to_set != ""} {
    set user_dir_param [strcat "\"--userdir=" [string map [list "\\" "\\\\" "\"" "\\\""] $maxima_userdir_to_set] "\""]
  }
  if {$maxima_initname_to_set != ""} {
    set path_suffix [string tolower [lindex [split $maxima_initname_to_set "."] "end"]]
    set pass_init_type ""
    if {$path_suffix in $test_suffixes_maxima} {
      set pass_init_type "-mac"
    }
    if {$path_suffix in $test_suffixes_lisp} {
      set pass_init_type "-lisp"
    }
    set init_file_param [strcat "\"--init$pass_init_type=" [string map [list "\\" "\\\\" "\"" "\\\""] $maxima_initname_to_set] "\""]
  }
  append maxima_startup " $user_dir_param $init_file_param \"$without_readline\" \"$quiet_mode\""

  return [list $start_without_script $maxima_startup]
}

set minima_may_use_ctrlc_events true
if {[is_wine_like]} {
  if {![lindex [get_maxima_startup_base] 0]} {
    set minima_may_use_ctrlc_events false
    lappend system_startup_warnings [list "Maxima interrupt actions for systems alike ReactOS only available when Maxima started directly without batch"]
  }
}

proc wine_ctrlc_trap {args} {
  return 1
}

if {[is_wine_like]} {
  # to deal with CTRL-C under ReactOS, we need console attached
  if {$minima_uses_twapi} {
    if {$minima_started_via_wish && (![is_under_wine])} {
      if {[catch {
        twapi::allocate_console
      }]} {
        set minima_uses_twapi false
      }
    }
  }
  if {$minima_uses_twapi} {
    if {!$print_logging_some} {
      catch {twapi::set_console_title [strcat "$app_title_base" [expr {[is_batch_mode] ? " (batch mode)" : ""}]]}
      catch {twapi::hide_window [twapi::get_console_window]}
    } {
      catch {twapi::modify_console_output_mode stdout -processedoutput false -wrapoutput true}
      catch {twapi::set_console_title [strcat "$app_title_base" [expr {[is_batch_mode] ? " (batch mode)" : ""}] ", log stream"]}
      catch {twapi::set_console_screen_buffer_size stdout [list 85 1200]}
    }
  }
  # trying to use TWAPI (if present) for CTRL-C handling
  if {$minima_uses_twapi} {
    if {[catch {
      twapi::set_console_control_handler wine_ctrlc_trap
    }]} {
      set minima_uses_twapi false
    }
  }
  # at ReactOS/Wine we only do CTRL-C if TWAPI is present,
  # and only if Maxima not started via batch file, since cmd.exe intercepts CTRL-C signals
  set with_common_ctrlc [expr {$minima_may_use_ctrlc_events && $minima_uses_twapi}]
  set to_show_interrupt_button $with_common_ctrlc
}

if {[is_wine_like] && (!$minima_uses_twapi)} {
  lappend system_startup_warnings [list "TGM Minima under systems alike ReactOS requires TWAPI installed for some features to be available"]
}

if {(!$system_avoid_startup_warning) && ([llength $system_startup_warnings] > 0)} {
    after 1000 [list "show_text_field_message_error" $system_startup_warnings]
}

# not sure if should prevent the CTRL-C signal under Linux
proc attempt_to_ctrlc_catching_at_linux {} {
  global auto_path

  if {[is_linux_like]} {
    set seems_to_be false
    if {[lindex [split [info tclversion] "."] 0]==8} {
      foreach one_dir [list "/usr/local/lib/tclx8.4" "/usr/lib/tclx8.4"] {
        if {[file exists $one_dir]} {
          lappend auto_path $one_dir
          set seems_to_be true
        }
      }
    }
    if {$seems_to_be} {
      catch {package require Tclx}
      catch {signal -restart ignore [list SIGINT]}
    }
  }
}
catch {attempt_to_ctrlc_catching_at_linux}

proc has_common_ctrlc {} {
  global with_common_ctrlc
  return $with_common_ctrlc
}

proc get_null_path {} {
  if {[is_wine_like]} {
    return "nul"
  }
  return "/dev/null"
}

proc get_gnuplot_path {} {
  global gnuplot_path

  if {$gnuplot_path!=""} {
    return $gnuplot_path
  }

  if {[is_wine_like]} {
    #even though the "wgnuplot_pipes.exe" name suggests that it deals with pipes,
    #the "gnuplot.exe" command seems to work under Wine the pipe-wise way by now;
    #the "wgnuplot_pipes.exe" program was used for gnuplot of pre-5 versions;
    return "gnuplot"
  }
  return "gnuplot"
}

proc get_clipboard_enforcer {} {
  global clipboard_enforcer

  if {$clipboard_enforcer!=""} {
    return $clipboard_enforcer
  }

  if {![is_linux_like]} {
    return ""
  }

  foreach path_test [list "/usr/local/bin/store_clipboard_gtk" "/usr/local/bin/store_clipboard" "/usr/bin/store_clipboard_gtk" "/usr/bin/store_clipboard"] {
    if {[file exists $path_test]} {
      return $path_test
    }
  }

  return ""
}

array set spec_file_dialogs [list "save" "" "open" ""]
proc set_file_dialogs {} {
  global spec_file_dialogs
  global auto_path

  set dialog_subdir "fsdialog"
  if {[is_linux_like]} {
    set seems_to_be false
    foreach one_dir [list "/usr/local/share/tcltk" "/usr/share/tcltk"] {
      if {[catch {
        set test_subdirs [glob -nocomplain -directory $one_dir $dialog_subdir]
      }]} {
        set test_subdirs [list]
      }
      if {[llength $test_subdirs]==1} {
        lappend auto_path [lindex $test_subdirs 0]
        set seems_to_be true
      }
    }
    if {$seems_to_be} {
      set spec_file_dialogs(save) {ttk::getSaveFile}
      set spec_file_dialogs(open) {ttk::getOpenFile}
    }
  }
}
catch {set_file_dialogs}

array set spec_base_font [list "search" "" "command" "" "text" ""]
proc set_base_font {} {
  global spec_base_font

  if {[is_linux_like]} {
    set known_font_families [font families]
    foreach one_name [list "Ubuntu Mono" "Liberation Mono" "Linux Libertine Mono O" "Courier"] {
      if {$one_name in $known_font_families} {
        if {$one_name == "Ubuntu Mono"} {
          set spec_base_font(search) [list "Ubuntu Mono" -16]
          set spec_base_font(command) [list "Ubuntu Mono" -16 "bold"]
          set spec_base_font(text) [list "Ubuntu Mono" -18]
        }
        if {$one_name == "Liberation Mono"} {
          set spec_base_font(search) [list "Liberation Mono" -14]
          set spec_base_font(command) [list "Liberation Mono" -15 "bold"]
          set spec_base_font(text) [list "Liberation Mono" -16]
        }
        if {$one_name == "Linux Libertine Mono O"} {
          set spec_base_font(search) [list "Linux Libertine Mono O" -13]
          set spec_base_font(command) [list "Linux Libertine Mono O" -14 "bold"]
          set spec_base_font(text) [list "Linux Libertine Mono O" -16]
        }
        if {$one_name == "Courier"} {
          set spec_base_font(search) [list "Courier" -14]
          set spec_base_font(command) [list "Courier" -15 "bold"]
          set spec_base_font(text) [list "Courier" -17]
        }
        break
      }
    }
  }
}
set_base_font

proc get_maxima_quit {} {
  global in_lisp_mode

  set quit_command "quit();"
  if {$in_lisp_mode} {
    set quit_command "(\$quit)"
  }

  return $quit_command
}

set maxima_command ""
set maxima_command_inner ""
set maxima_command_control ""
set is_ending false
proc want_to_exit {} {
  global is_ending
  global maxima_pid
  global maxima_channel
  global maxima_command
  global maxima_command_inner
  global maxima_command_control
  global save_path
  global text_field
  global state_label
  global maxima_ready
  global has_ever_asked
  global checks_id
  global read_wait_id
  global maxima_stderr_read_id
  global maxima_stderr_write_id
  global maxima_stderr_should_ignore
  global check_maxima_stderr_after
  global newly_at_maxima_outputs
  global editor_batch_state
  global html_tmp_path
  global plot_input_warning
  global plot_input_field
  global plot_data_field
  global batch_maxima_input_warning
  global batch_maxima_output_warning
  global mxmbatch_input_field
  global mxmbatch_output_field
  global store_clipboard_ever_attempted
  global import_input_field
  global import_output_field
  global import_input_warning
  global maxima_stderr_should_check
  global can_deal_with_disconnect
  global should_display_cpu_usage
  global could_display_cpu_usage

  remove_torn_menus

  set to_ask false

  set command_unsaved false
  set command_state [check_stack_save_state]
  if {$command_state == 1} {
    set command_unsaved true
    set to_ask true
  }

  set text_unsaved false
  if {[$text_field edit modified]} {
    set text_unsaved true
    set to_ask true
  }

  set import_input_ask false
  set import_output_ask false
  if {$import_input_warning} {
    if {$import_input_field != ""} {
      if {[$import_input_field index "end -1 chars"] != "1.0"} {
        set to_ask true
        set import_input_ask true
      }
    }
    if {$import_output_field != ""} {
      if {[$import_output_field index "end -1 chars"] != "1.0"} {
        set to_ask true
        set import_output_ask true
      }
    }
  }

  set plot_input_ask false
  set plot_data_ask false
  if {$plot_input_warning} {
    if {$plot_input_field != ""} {
      if {[$plot_input_field index "end -1 chars"] != "1.0"} {
        set to_ask true
        set plot_input_ask true
      }
    }
    if {$plot_input_field != ""} {
      if {[$plot_data_field index "end -1 chars"] != "1.0"} {
        set to_ask true
        set plot_data_ask true
      }
    }
  }

  set batch_maxima_input_ask false
  if {$batch_maxima_input_warning} {
    if {$mxmbatch_input_field != ""} {
      if {[$mxmbatch_input_field index "end -1 chars"] != "1.0"} {
        set to_ask true
        set batch_maxima_input_ask true
      }
    }
  }
  set batch_maxima_output_ask false
  if {$batch_maxima_output_warning} {
    if {$mxmbatch_output_field != ""} {
      if {[$mxmbatch_output_field index "end -1 chars"] != "1.0"} {
        set to_ask true
        set batch_maxima_output_ask true
      }
    }
  }

  set message_detail ""
  set saved_resaved_part " saved"
  if {$save_path!=""} {
    set saved_resaved_part " resaved"
  }
  if {$to_ask} {
    if {$text_unsaved} {
      if {$command_state==-1} {
        set message_detail "Session not"
      } elseif {$command_state==0} {
        set message_detail "Text not"
      } else {
        set message_detail "Neither text nor command"
      }
      append message_detail $saved_resaved_part
    } elseif {$command_unsaved} {
      set message_detail "Command not"
      append message_detail $saved_resaved_part
    }

    if {$import_input_ask || $import_output_ask} {
      set import_nonempty_parts [list]
      if {$import_input_ask} {
        lappend import_nonempty_parts "input"
      }
      if {$import_output_ask} {
        lappend import_nonempty_parts "output"
      }
      if {$message_detail != ""} {
        append message_detail ",\nimport "
      } {
        append message_detail "Import "
      }
      append message_detail [join $import_nonempty_parts "/"] " not empty"
    }
    if {$plot_input_ask || $plot_data_ask} {
      set plot_nonempty_parts [list]
      if {$plot_input_ask} {
        lappend plot_nonempty_parts "input"
      }
      if {$plot_data_ask} {
        lappend plot_nonempty_parts "data"
      }
      if {$message_detail != ""} {
        append message_detail ",\nplot "
      } {
        append message_detail "Plot "
      }
      append message_detail [join $plot_nonempty_parts "/"] " not empty"
    }
    if {$batch_maxima_input_ask || $batch_maxima_output_ask} {
      if {$message_detail != ""} {
        append message_detail ",\nbatch "
      } {
        append message_detail "Batch "
      }
      if {$batch_maxima_input_ask && $batch_maxima_output_ask} {
        append message_detail "fields "
      } elseif {($batch_maxima_input_ask) && (!$batch_maxima_output_ask)} {
        append message_detail "input "
      } elseif {(!$batch_maxima_input_ask) && ($batch_maxima_output_ask)} {
        append message_detail "output "
      }
      append message_detail "not empty"
    }
  }

  set to_exit "ok"
  set message_ask "Discard changes?"
  if {$to_ask} {
    set to_exit [tk_messageBox -message $message_ask -icon question -type okcancel -default cancel -detail $message_detail]
  }

  if {$to_exit=="cancel"} {
    return
  }
  set is_ending true

  # it may be that it does not have enough time to store it (externally),
  # though the varying support for it at current desktop environments
  # forces us to only do it at the application end;
  if {$store_clipboard_ever_attempted} {
    store_clipboard
  }

  set is_ready $maxima_ready
  set maxima_ready false
  set newly_at_maxima_outputs false

  if {$checks_id!=""} {
    catch {after cancel checks_id}
    set checks_id ""
  }

  catch {after cancel $read_wait_id}

  set maxima_stderr_should_check false
  set can_deal_with_disconnect false

  if {$maxima_pid!=""} {
    catch {fileevent $maxima_channel readable ""}
    if {(!$is_ready) && [has_common_ctrlc]} {
      catch {do_kill_maxima false}
    }
    if {true} {
      set maxima_command ""
      set maxima_command_inner ""
      set maxima_command_control [get_maxima_quit]
      catch {fileevent $maxima_channel writable "svwrite 0"}
    }
  }

  clear_read_buffers

  set_button_interrupt_sets
  set_state_disabled true
  set_state_disabled_buttons
  set has_ever_asked false

  clean_outer_pid_onset_files
  reset_cpu_usage
  stop_cpu_usage_updates
  set_ignore_cpu_usage true
  set should_display_cpu_usage false
  set could_display_cpu_usage false
  set_state_label_text "ending"

  if {$check_maxima_stderr_after!=""} {
    catch {after cancel $check_maxima_stderr_after}
  }
  set maxima_stderr_should_ignore true
  catch {close $maxima_stderr_read_id}
  catch {close $maxima_stderr_write_id}

  if {$editor_batch_state} {
    stop_batch_maxima true
  }
  close_gnuplot_direct true

  if {$html_tmp_path!=""} {
    catch {
      file delete -- $html_tmp_path
      set html_tmp_path ""
    }
  }

  if {$editor_batch_state} {
    if {$store_clipboard_ever_attempted} {
      pause_tasks 100
    }
    exit 0
  }

  if {$maxima_pid!=""} {
    after 0 setup_exit
  } {
    force_exit
  }
}

proc setup_exit {} {
  after 200 force_exit
}

proc force_exit {} {
  global maxima_pid
  global maxima_channel
  global svsocket
  global svport
  global temporary_file_paths
  global minima_clipboard_waiting_to_get_stored
  global maxima_channel_hnd

  catch {fileevent $maxima_channel writable ""}

  if {$svsocket!=""} {
    catch {chan close $svsocket}
    set svsocket ""
    set svport -1
  }

  if {$maxima_pid!=""} {
    do_kill_maxima true
    set $maxima_pid ""
    after 200
  }
  if {$maxima_channel_hnd!=""} {
    clear_maxima_channel_hnd
  }

  if {$maxima_channel!=""} {
    if {[llength [file channels $maxima_channel]]>0} {
      catch {fileevent $maxima_channel readable ""}
      catch {fileevent $maxima_channel writable ""}
      catch {chan close $maxima_channel}
      set maxima_channel ""
    }
  }

  foreach one_path $temporary_file_paths {
    catch {file delete $one_path}
  }

  if {$minima_clipboard_waiting_to_get_stored} {
    after 100 "minima_clipboard_reset_waiting_state"
    vwait minima_clipboard_waiting_to_get_stored
  }

  exit 0
}

set icon_data_bm [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAQAAAC1+jfqAAAAAmJLR0QA/4ePzL8AAAAJcEhZcwAA" \
"CxMAAAsTAQCanBgAAAAHdElNRQfhCxsQCiFLpNisAAAACHRFWHRDb21tZW50APbMlr8AAABqSURB" \
"VCjPrZHBDoAgDENf/GrRIwioH10PEhgEb/a0rs3WbPAXVmLHM2svC3FUfiJkLQkhRADgKizZGc1y" \
"lyqOKWIRNKwzCFUOrbkYgyaVge9W+C+5hfRzGZhZ9iH5e4nNznD9Wci4v774AIxBRU1JMPN1AAAA" \
"AElFTkSuQmCC"] ""]
set icon_image_bm [image create photo "icon_image_bm" -data $icon_data_bm]

set icon_data_mm [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAQAAAC1+jfqAAAAAmJLR0QA/4ePzL8AAAAJcEhZcwAA" \
"CxMAAAsTAQCanBgAAAAHdElNRQfhCxsPODTcc4x7AAAACHRFWHRDb21tZW50APbMlr8AAABqSURB" \
"VCjPzdHLDcIwEIThrxBQHqYTkCjbLiVOoI7lgIWQw4ULYk6rnV+affBPOsrmVifZoQeKUM042YTc" \
"A892dXUTNmkfkqxCCOsrrNOlAefP9qQ2oJr29mhpM9yFxfj1FoPydodi+NkHHicyKLtZ2ziSAAAA" \
"AElFTkSuQmCC"] ""]
set icon_image_mm [image create photo "icon_image_mm" -data $icon_data_mm]

set icon_data_qm [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAQAAAC1+jfqAAAAAmJLR0QA/4ePzL8AAAAJcEhZcwAA" \
"CxMAAAsTAQCanBgAAAAHdElNRQfhCAkTGBmkpEt8AAAACHRFWHRDb21tZW50APbMlr8AAABxSURB" \
"VCjPY2AgATgxrGV4yfCd4QpDJgMjNgX/kWARNgVHGAIY+BhEGKYx/Ge4g88qPob/DF9gHCYsCqwY" \
"GBg24tavyPCU4RKDEC5pdYbHDNcZRHHrP8rwn8ESnwN/MfxnYMan4D/DfwbKAIYJTAw0BwALqBzm" \
"yfRo7wAAAABJRU5ErkJggg=="] ""]
set icon_image_qm [image create photo "icon_image_qm" -data $icon_data_qm]

set icon_data_em [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAQAAAC1+jfqAAAAAmJLR0QA/4ePzL8AAAAJcEhZcwAA" \
"CxMAAAsTAQCanBgAAAAHdElNRQfhCAkTHzHeUHVBAAAACHRFWHRDb21tZW50APbMlr8AAAAtSURB" \
"VCjPY2AgABjR+P/RRZkImTAACn4wMDB8x6fgCRSTr+AxugKGQRCSlAMADacJXS87TzYAAAAASUVO" \
"RK5CYII="] ""]
set icon_image_em [image create photo "icon_image_em" -data $icon_data_em]

set icon_data_xm [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAQAAAC1+jfqAAAAAmJLR0QA/4ePzL8AAAAJcEhZcwAA" \
"DdcAAA3XAUIom3gAAAAHdElNRQfhCAkVEgVOx4MLAAAAs0lEQVQoz63PoU6CARTF8R86m1Z1gyjP" \
"QfQZcGqSZNNqdz4JATY2BxaLs2jV4hNAcTgxuDHDsTj2ffixGbzt3PO/957LH2skrhfqVvTLQNNc" \
"7IOOmNpZ3nEpXmxomInD30c2jcW5GzGsznEgvsS7+qqo9yJOiq21EvABnlfNH4mIB7Uqe9dUXHgT" \
"x1XAQNypORUTW8t2W3zaw7oncVW2t72Ksx/VEnPNItATj4WPumLkn+obyCM4m0aVsG4AAAAASUVO" \
"RK5CYII="] ""]
set icon_image_xm [image create photo "icon_image_xm" -data $icon_data_xm]

set icon_data_pg [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAQAAAC1+jfqAAAAAmJLR0QA/4ePzL8AAAAJcEhZcwAA" \
"CxMAAAsTAQCanBgAAAAHdElNRQfhCAwAMyZNDHYzAAAACHRFWHRDb21tZW50APbMlr8AAAClSURB" \
"VCjPldAhTgNhEAXgb5sGSepQS1IHFlwtBtsDwD04R6/AAQghAQWirVmBwy2iohLFso4MBpKfZKdJ" \
"35jJm/dmMo89MPWkEyITvIjfStALF6r/5Kjo12hzP+d6dw6zDXMrD961rof9W2GCW+EmE5zgWPgY" \
"Elz59uxALbwNH5l5da+xcZb9UVnonO6K+0ho8nHlUvgsqXHR/2W4zPwhfHlUl+QPtTYuUBZDmeUA" \
"AAAASUVORK5CYII="] ""]
set icon_image_pg [image create photo "icon_image_pg" -data $icon_data_pg]

set icon_data_as [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAQAAAC1+jfqAAAAAmJLR0QA/4ePzL8AAAAJcEhZcwAA" \
"CxMAAAsTAQCanBgAAAAHdElNRQfhDAUQCg17mkjzAAAACHRFWHRDb21tZW50APbMlr8AAADoSURB" \
"VCjPrdGxK4RxGAfwT3fnTYe44QYUYpQMZLv4B/wRDLJLBhYZ7JR0kjKIDCaLbjMb1BkkujrU1Q2S" \
"1Gt5De5983aLwW97+n2ep2/Pw/+9aUdqQg0VC7Lpz5xddWvGdSmaV3Gj/zfYd6U31bLhTj4uSp70" \
"yNj0rKEsEuHUegzOLGNFpKzgsAVKqjF4NYR7kRGMtkBBGIMvAT5FAgQJeIMMmoqoYwCDSbLreMK5" \
"Ray2MvyEnFI1GYM5D/KytrxoOhB5d2EivYdL3UnV2b7mnB2Plgzr0GfWnpN2NONYTejDrW1jf77h" \
"Nw4VQNLvVmkIAAAAAElFTkSuQmCC"] ""]
set icon_image_as [image create photo "icon_image_as" -data $icon_data_as]

set icon_data_mx [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAQAAAC1+jfqAAAAAmJLR0QA/4ePzL8AAAAJcEhZcwAA" \
"CxMAAAsTAQCanBgAAAAHdElNRQfhChIIKjThoIwXAAAACHRFWHRDb21tZW50APbMlr8AAAB1SURB" \
"VCjPY2CgFDDCWf+xyzARb1Ydw3+GfwzJ+JRUMPxn+MeQjk9JEcN/hn8MOfiU5DL8Z/jPUIAQQHfk" \
"XoYvDAwMMrj0SzA8YPjPsBrJ8yiAi+E0w3+GYwwc2KWZGDYy/Ge4wyCCy/iJDP8Z3jCo0TKocQAA" \
"gxoZz07K6pcAAAAASUVORK5CYII="] ""]
set icon_image_mx [image create photo "icon_image_mx" -data $icon_data_mx]

set icon_data_pl [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAQAAAC1+jfqAAAAAmJLR0QA/4ePzL8AAAAJcEhZcwAA" \
"DdcAAA3XAUIom3gAAAAHdElNRQfhCg8KGiiLEL2HAAABJ0lEQVQoz22QTShEcRTFf++9v68nj2dm" \
"yMdMCm9KkZ1kYaHsyIJiSY00a3vL2ZmdUrKyYKFYoOwsbQZJTSOjDGaFNCMZ6lo889F4Z3fP+XXP" \
"7RqU5Wfev2Yv5wyy5KlSC8cIz1bSnyGLcITlBjoAA7zhI0zn0npkjw4cWnknXARsbthkhBQU1LcB" \
"3DLKFkmUC+xyzkp1JxHSXIGinUn68NIQedCZ4om7oqeJJiXgA5jWfIc/zmK8oNz4YVwCof0/QLZj" \
"zQlVZ+lmg6MMF5DAl232F1foNaYFETLl2mh0NV4aNIQZnRO66fE80gIOdB45ZccTuCDt/mGBMTb+" \
"xTGC9IICXhnkmgmG+UQAjSYuCdJVydeSQpD6l8YcgnDvWcpc21kowWyl9QuK0UyE1pPNHAAAAABJ" \
"RU5ErkJggg=="] ""]
set icon_image_pl [image create photo "icon_image_pl" -data $icon_data_pl]

set icon_data_import [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABmJLR0QA/wD/AP+gvaeTAAAACXBI" \
"WXMAAA3XAAAN1wFCKJt4AAAAB3RJTUUH4QobDBwwFdA+LQAAARBJREFUOMul071KQ0EQBeAvf1gI" \
"IZBKNNilEjsLCXbpbcTK1meQIMoWFlYWqX0FH0JNo00Ea1HLKyJahRiT2FwhxHuTGAeWHc6wZ8/M" \
"ns2EEKo4RMb0OAsh3I0CWSxhF094TlmP2EZlnDEf7x8hhONJV4cQ9pLwrH/GzASFQmHYbrc3cJ7U" \
"wrTYajab5SiKjnD5FwWbaOEqiqLSyJk8FpBLU1DFBdaS1KAX56dpBH2UMUzwxxCDOO+kETxgGau4" \
"xspIrYU6cviaNINhbK4K1ovF4muMD/CJDnqzPuN9o9F4r9frJ7iZywfdbjdbq9VucTC3EzOZ3//t" \
"Z4iLIYT9CYR9lJIKebzFfe1MEdDGyzj4DbC0RJT6A9RAAAAAAElFTkSuQmCC"] ""]
set icon_image_import [image create photo "icon_image_import" -data $icon_data_import]

set icon_data_export [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABmJLR0QA/wD/AP+gvaeTAAAACXBI" \
"WXMAAA3XAAAN1wFCKJt4AAAAB3RJTUUH4QobDB8HhkDI4QAAAP1JREFUOMu10jFKA1EUheFvRjGI" \
"ClaCA9rpPpIN2KhVWmcJForyOgvBwi24h9nDVLqDxFYhKgRMIBNtZkDCTBIUDxwenAv/vdz7ohDC" \
"IS4RWay7EMLTzyDGLk7Rx3ODezjC3ixxtXw/QgjX81qHELp1eeyPWgTYwQNEUfT1G8AKuiiyLNsf" \
"DoeTph3MqlW+a1WjPM/jPM8zBNzjbd4Eo9L9un1igCusNwHeUZSu07Q86bgJkGALBzW1G2wixbQJ" \
"8Fl6VAXtdrtI0/QEF2VtqStMcItWp9PpJUkyXvYKlV5xDkVRxP/yE6sJNkIIZ3OABbabAAPkOF7Q" \
"7BEvs+E3FENFvIalNnwAAAAASUVORK5CYII="] ""]
set icon_image_export [image create photo "icon_image_export" -data $icon_data_export]

set icon_data_blank [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABmJLR0QA/wD/AP+gvaeTAAAACXBI" \
"WXMAAA3XAAAN1wFCKJt4AAAAB3RJTUUH4QobDDks6SKyhQAAAThJREFUOMul07FLW1EUx/FPjEiC" \
"iyIpiBRMh2IJiASngn+CuJhOgosOdtJZkLuVLg4O4iC4ZtfJuQVdoiIp3dTBxWjA4hAw8Fye8Ajx" \
"JdIDl8v9Xfidc7733EwI4TM2kdE7tkMI50lhAOP4hmvcvLGusICPnY6D8f4YQthKSx1CWOqmD/jP" \
"6MfgexftS2cL3SKDAywjymazz1EURZjGBf5iNs1gLAYHu/V6XT6fL+Iw1qbwKc3gPqb+gKFqtSqX" \
"y+0k7udw2YvBEwpoQ6vVetXX8KtfiP/QTJzb2HvPK/zGh47Zue/XYB9foVQqKZfLxwnAVxhMgzgR" \
"jzhcVCqV0UKh8LNWq/3BOiYxm1bBLUbjvzCDdqPRGMYGjlDEyWsFwyGE1Tda+oEVjCS0+SSQJk6x" \
"2IPHGe46xRfs2krCcS5QLQAAAABJRU5ErkJggg=="] ""]
set icon_image_blank [image create photo "icon_image_blank" -data $icon_data_blank]

set icon_data_save [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABmJLR0QA/wD/AP+gvaeTAAAACXBI" \
"WXMAAA3XAAAN1wFCKJt4AAAAB3RJTUUH4QobDQMGFvLPHQAAAQRJREFUOMud07FKA0EUBdATURKS" \
"ytLC1j9ISgu3tBCCWImN+BWKzhfkG4QIUb/AUggI2mgbENRWRRCMBHGNzQbCmmR3vTDM4w5z37tv" \
"3pRCCCvYR0k2WiGEu3FiDkvYwiOepqwHbGA5rTif7O8hhMNZqUMI28Ph0DSBTPR6va9Op3OKS6yP" \
"W8iFOI6hilq6B7lQLpfjJPwuYqGKPrTb7REXYdSM1awKPrE75ayPqzwWjvExgV9DnEfgB0cp7gU3" \
"RZrYSi6NsFP4FRqNxmsS3uOisECz2SxVKhU4+NccDAaDhXq9foKzSXNQCyHszRCMsRhF0Xm32/3z" \
"F95wjc2MIm7xnCZ/AbfqRHfHVCefAAAAAElFTkSuQmCC"] ""]
set icon_image_save [image create photo "icon_image_save" -data $icon_data_save]

set icon_data_recent [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABmJLR0QA/wD/AP+gvaeTAAAACXBI" \
"WXMAAA3XAAAN1wFCKJt4AAAAB3RJTUUH4QobDQg3pNgW7AAAAPRJREFUOMud07FKA0EQgOHvEoOB" \
"dGnEQrDyLSSPoIVanWJnbycBGUsbO0tfIK12B+lNEa1s1VZBsLDJIbE5IRxJLsnAsrP/wj/Lzm4S" \
"ETvoIlEdNxHxPAlq2MQR3vA+Y7xiD1tl41oxf0fE5bzSEZFO4zWLxzjLsuOVBc1mMx8MBoc4WUnQ" \
"6/XGeZ7D1UqC4XC4UaTb2F9W0EV7Yn23jKCOixJrY3dRwRlaU3iGepWghdsZe+voVAl+iieepGn6" \
"UrD+P0N/4S6MRqPaxJ0s38ZGoyFJkk98TPIkIjp4wPkc4S+ucRoR9+XP9IVHHFQc4qlcHf4ATm06" \
"51ixazUAAAAASUVORK5CYII="] ""]
set icon_image_recent [image create photo "icon_image_recent" -data $icon_data_recent]

set icon_data_open [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABmJLR0QA/wD/AP+gvaeTAAAACXBI" \
"WXMAAA3XAAAN1wFCKJt4AAAAB3RJTUUH4QobDQ8uj/Io6wAAANxJREFUOMul0b1KQ0EQBeDP6yUW" \
"eQDBUosIopUkiIXvYGNhm059BGFfIq8jdgEhpBGCYOUj2PiTn2uzFxZZovfmwMKenTkzZ2a3rcc+" \
"unjXEk94ays+RYUV9toUeIgFKrw2FQ8ScX2OcolbIYQDdEIIs+T9EReZfQxqEnVFiRvs4jrGDjNi" \
"6OMM48jv0CmwwHeSeI8JppFX8T7BVZI3x7zMdKqd7OAz/sI5PnI7KGyIck1sgVt8/Rrx3wWWGP3l" \
"YOMRimTTTVHVI6zQCyEMG4pP8FJihmNcNnSwxPMPbFAzZUhdyDIAAAAASUVORK5CYII="] ""]
set icon_image_open [image create photo "icon_image_open" -data $icon_data_open]

set icon_data_save_as [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABmJLR0QA/wD/AP+gvaeTAAAACXBI" \
"WXMAAA3XAAAN1wFCKJt4AAAAB3RJTUUH4QobDRUs0dGzHAAAANZJREFUOMul0T1KBEEQBeDPZhn1" \
"CqYGRnoYz2AmgoIIikLfQE3EYI8lLN7A1MCf3XF3THqgaZ0dZ3xQQXW/evWqim5U+MQLtozAJhp8" \
"YbuLNOnovJEEcrFVirqv811GbFLUycnNX+1/ZMV5VD9GiDHuoooxzrL3W1wW3HMs2iTVhYBjXBfk" \
"36xOi/wEZyHNtig+l7jI8iu8FpwadVizh4dM7H7IGVu84RQ7eB8jAI+YryOEHoF5372Df6IVaEbU" \
"Nu0OVtiLMR4NLD7A8wQz7ONwoIMlnr4Baz83Nw8VG24AAAAASUVORK5CYII="] ""]
set icon_image_save_as [image create photo "icon_image_save_as" -data $icon_data_save_as]

set fmenu [frame $topw.m -relief sunken -bd 1]
set fpaned [panedwindow $topw.p -orient vertical -handlesize 1 -borderwidth 0 -sashpad 0]
set ftext [ttk::frame $fpaned.t -relief groove -borderwidth 2]
set fcmmd [ttk::frame $fpaned.c -relief groove -borderwidth 2]
pack $fmenu -side top -expand no -fill x
$fpaned add $ftext -minsize 100 -stretch always
if {!$editor_batch_state} {
  $fpaned add $fcmmd -minsize 120 -stretch never
}
pack $fpaned -side top -expand yes -fill both

set text_search ""
font create search_font
foreach {font_option_name font_option_value} [font actual TkFixedFont] {
  font configure search_font $font_option_name $font_option_value
}
font configure search_font -size -14

set info_labels_foreground [list "#a0a0a0" "#000000"]

ttk::style configure File.TButton -anchor [ttk::style configure TButton -anchor]
if {[is_linux_like]} {
  ttk::style configure File.TButton -padding [list 3 3 3 3]
}
if {[is_wine_like]} {
  ttk::style configure File.TButton -padding [list 3 1 3 1]
}
set bt_file_open_re [ttk::button $fmenu.open_re -command {draw_open_recent} -takefocus 0]
set bt_file_open [ttk::button $fmenu.open -command {draw_open} -takefocus 0]
set bt_file_import [ttk::button $fmenu.import -command {draw_import} -takefocus 0]
set bt_file_void [ttk::button $fmenu.void -command {draw_blank} -takefocus 0]
set bt_file_export [ttk::button $fmenu.export -command {draw_export_as} -takefocus 0]
set bt_file_save [ttk::button $fmenu.save -command {draw_save} -takefocus 0]
set bt_file_save_as [ttk::button $fmenu.save_as -command {draw_save_as} -takefocus 0]
if {$use_buttons_text} {
  $bt_file_open_re configure -text "R" -width 2 -style File.TButton
  $bt_file_open configure -text "O" -width 2 -style File.TButton
  $bt_file_import configure -text "I" -width 2 -style File.TButton
  $bt_file_void configure -text "V" -width 2 -style File.TButton
  $bt_file_export configure -text "E" -width 2 -style File.TButton
  $bt_file_save configure -text "S" -width 2 -style File.TButton
  $bt_file_save_as configure -text "A" -width 2 -style File.TButton
} {
  $bt_file_open_re configure -image $icon_image_recent
  $bt_file_open configure -image $icon_image_open
  $bt_file_import configure -image $icon_image_import
  $bt_file_void configure -image $icon_image_blank
  $bt_file_export configure -image $icon_image_export
  $bt_file_save configure -image $icon_image_save
  $bt_file_save_as configure -image $icon_image_save_as
}

$bt_file_open_re configure -state "disabled"
$bt_file_open configure -state "disabled"
$bt_file_import configure -state "disabled"
$bt_file_void configure -state "disabled"
$bt_file_export configure -state "disabled"
$bt_file_save configure -state "disabled"

ttk::style configure SearchIn.TEntry -padding [ttk::style configure TEntry -padding]
ttk::style configure SearchIn.TEntry -borderwidth 1
ttk::style configure SearchIn.TEntry -fieldbackground "white"
ttk::style configure SearchOut.TEntry -padding [ttk::style configure TEntry -padding]
ttk::style configure SearchOut.TEntry -borderwidth 1

proc set_search_entry_style_common {} {
  ttk::style configure SearchIn.TEntry -fieldbackground "white"
  ttk::style map SearchOut.TEntry -fieldbackground [list "readonly" "#e0e0e0" "disabled" "#e0e0e0"]
}
proc set_search_entry_style_control {} {
  ttk::style configure SearchIn.TEntry -fieldbackground "#e0e0ff"
  ttk::style map SearchOut.TEntry -fieldbackground [list "readonly" "#c8c8e0" "disabled" "#c8c8e0"]
}
set_search_entry_style_common

set search_entry [ttk::entry $fmenu.search -width [expr {($scr_width>1000) ? 20 : 9}] -textvariable text_search -style SearchOut.TEntry]
$search_entry configure -state "readonly"

if {$spec_base_font(search)==""} {
  $search_entry configure -font search_font
} {
  $search_entry configure -font $spec_base_font(search)
}
set stack_label_saving [ttk::label $fmenu.stack_saving -text " " -padding [list 5 0 5 0] -width [expr {($scr_width>1000) ? 6 : 3}] -anchor "e" -justify "right"]
set stack_label_pos [ttk::label $fmenu.stack_position -text " " -padding [list 5 0 5 0]]

set state_label [label $fmenu.state -text " " -bd 0]

set should_display_cpu_usage false
set could_display_cpu_usage false
set last_cpu_usage_info ""
set last_cpu_usage_high false
set standard_state_label_text ""

proc update_state_label_text {} {
  global should_display_cpu_usage
  global could_display_cpu_usage
  global last_cpu_usage_info
  global last_cpu_usage_high
  global standard_state_label_text
  global state_label

  set displayed_cpu_info false
  if {$last_cpu_usage_info!=""} {
    if {($should_display_cpu_usage) || ($could_display_cpu_usage && $last_cpu_usage_high)} {
      $state_label configure -text $last_cpu_usage_info
      set displayed_cpu_info true
    }
  }

  if {!$displayed_cpu_info} {
    $state_label configure -text $standard_state_label_text
  }
}
proc set_state_label_text {state_text} {
  global standard_state_label_text
  global print_logging_minima_status

  set standard_state_label_text $state_text
  update_state_label_text

  if {$print_logging_minima_status} {
    print_logging "\nXXX minima-status: $state_text\n"
  }
}
proc get_state_label_text {} {
  global standard_state_label_text

  return $standard_state_label_text
}

set batch_maxima_window ""
set mxmbatch_fmenu ""
set mxmbatch_info_label_display ""
set mxmbatch_info_label_verb ""
set mxmbatch_info_label_bind ""
set mxmbatch_info_label_time ""
set mxmbatch_info_label_adjust ""
set mxmbatch_info_label_endings ""
set mxmbatch_info_label_maxcount_text ""

set mxmbatch_info_label_maxcount_text [label $fmenu.rowmaxcount_text -text " " -bd 0 -width 4 -justify "center" -fg "#45456a"]

set info_label_display [label $fmenu.maximadisplay -text " " -bd 0 -width 3]

set info_label_verb [label $fmenu.withlabels -text "%" -bd 0 -width 1]
if {!$with_labels} {
  $info_label_verb configure -foreground [lindex $info_labels_foreground 0]
}
if {[is_wine_like]} {
  $info_label_verb configure -width 2 -justify "right" -anchor "e"
}

set info_label_bind [label $fmenu.labelsbound -text "b" -bd 0 -width 1 -justify "center"]
if {!$bind_labels} {
  $info_label_bind configure -foreground [lindex $info_labels_foreground 0]
}
if {[is_wine_like]} {
  $info_label_bind configure -justify "left" -anchor "w"
}

set info_label_stopping [label $fmenu.batchstopping -text [get_batch_ending_label_char] -bd 0 -width 2 -justify "center"]

set info_label_endings_char ";;"
if {[is_stack_mode]} {
  append info_label_endings_char " "
}
set info_label_endings [label $fmenu.lineendings -text $info_label_endings_char -bd 0 -justify "center"]
if {[is_batch_mode]} {
  $info_label_endings configure -width 2
}
if {!$make_line_endings} {
  $info_label_endings configure -foreground [lindex $info_labels_foreground 0]
}

proc get_display_adjust_char {} {
  global mxm_display_sorted_ratsimp
  global mxm_display_exp_dstar

  if {(!$mxm_display_sorted_ratsimp) && (!$mxm_display_exp_dstar)} {
    return "^"
  }
  if {$mxm_display_sorted_ratsimp && (!$mxm_display_exp_dstar)} {
    return "r^"
  }
  if {(!$mxm_display_sorted_ratsimp) && $mxm_display_exp_dstar} {
    return "**"
  }
  return "r"
}

set info_label_adjust [label $fmenu.displayadjust -text [get_display_adjust_char] -bd 0 -width 2]
$info_label_adjust configure -justify "center"

set info_label_time [label $fmenu.timeelapsed -text "t" -bd 0 -width [expr {[is_stack_mode] ? 1 : 2}]]
if {[is_stack_mode]} {
  $info_label_time configure -anchor "e"
} {
  $info_label_time configure -justify "center"
}

if {!$show_time_elapsed} {
  $info_label_time configure -foreground [lindex $info_labels_foreground 0]
}

set info_label_input [label $fmenu.inputcommands -text "i" -bd 0 -width 1 -justify "center" -anchor "w"]
if {!$include_maxima_input_default} {
  $info_label_input configure -foreground [lindex $info_labels_foreground 0]
}

set info_label_autotag [label $fmenu.autotagging -bd 0 -width [expr {[is_stack_mode] ? 1 : 2}] -justify "center" -text "a"]
if {$omit_auto_tildes} {
  $info_label_autotag configure -foreground [lindex $info_labels_foreground 0]
}

set info_label_output [label $fmenu.outputplacing -bd 0 -width 1 -justify "center"]
$info_label_output configure -text [lindex $info_label_output_modes $info_label_output_index]

set info_label_hidetag [label $fmenu.hidetildes -bd 0 -width 2 -justify "center" -text "~"]

proc draw_plot {} {
  remove_torn_menus
  open_plot_part
}

set bt_plot [ttk::button $fmenu.plot -command {draw_plot} -state normal -takefocus 0]
set bt_stack [ttk::button $fmenu.stack -command {draw_stack} -state normal -takefocus 0]
set bt_history [ttk::button $fmenu.history -command {draw_history} -state normal -takefocus 0]
set bt_maxima_batch [ttk::button $fmenu.maxima_batch_button -command {draw_batch_maxima} -state normal -takefocus 0]
set bt_interrupt [ttk::button $fmenu.interrupt -command {draw_interrupt} -state disabled -takefocus 0]
set bt_restart [ttk::button $fmenu.restart -command {draw_restart} -state normal -takefocus 0]
set bt_kill [ttk::button $fmenu.kill -command {draw_kill_interrupt} -state normal -takefocus 0]
set bt_about [ttk::button $fmenu.about -command {draw_about} -state normal -takefocus 0]

set pargr_char "\$"
catch {
  set pargr_char [encoding convertfrom "utf-8" "\xC2\xA7"]
}

if {$use_buttons_text} {
  $bt_plot configure -text "P" -width 2 -style File.TButton
  $bt_stack configure -text "@" -width 2 -style File.TButton
  $bt_history configure -text "$pargr_char" -width 2 -style File.TButton
  $bt_maxima_batch configure -text "M" -width 2 -style File.TButton
  $bt_interrupt configure -text "X" -width 2 -style File.TButton
  $bt_restart configure -text "!" -width 2 -style File.TButton
  $bt_kill configure -text "K" -width 2 -style File.TButton
  $bt_about configure -text "?" -width 2 -style File.TButton
} {
  $bt_plot configure -image $icon_image_pl
  $bt_stack configure -image $icon_image_as
  $bt_history configure -image $icon_image_pg
  $bt_maxima_batch configure -image $icon_image_mx
  $bt_interrupt configure -image $icon_image_mm
  $bt_restart configure -image $icon_image_em
  $bt_kill configure -image $icon_image_bm
  $bt_about configure -image $icon_image_qm
}

$bt_kill configure -state "disabled"

set btlist_save [list $bt_file_export $bt_file_save $bt_file_save_as]
set btlist_other [list $bt_file_open_re $bt_file_open $bt_file_import $bt_file_void $search_entry $bt_restart $info_label_display $info_label_verb $info_label_bind]

set_state_label_text "starting"

pack $bt_file_open_re $bt_file_open $bt_file_import $bt_file_void $bt_file_export $bt_file_save $bt_file_save_as -side left
if {$scr_width>1000} {
  pack [label $fmenu.spacer1 -width 0 -padx 2] -side left
}
pack $search_entry -side left
if {!$editor_batch_state} {
  pack $stack_label_saving -side left
  pack $stack_label_pos -side left
}
pack $bt_about -side right -expand no -fill none
if {!$editor_batch_state} {
  pack $bt_restart -side right -expand no -fill none
  if {$to_show_interrupt_button} {
    pack $bt_interrupt -side right -expand no -fill none
  }
  pack $bt_history -side right -expand no -fill none
  pack $bt_stack -side right -expand no -fill none
}
if {$editor_batch_state} {
  pack $bt_kill -side right -expand no -fill none
  pack $bt_maxima_batch -side right -expand no -fill none
  pack $bt_history -side right -expand no -fill none
}
pack $bt_plot -side right -expand no -fill none

if {!$to_show_interrupt_button} {
  $bt_interrupt configure -state disabled
}
if {$editor_batch_state} {
  $bt_interrupt configure -state disabled
}

set text_batch_input_labels_shown false
set text_batch_output_labels_shown false

if {!$editor_batch_state} {
  pack $info_label_hidetag $info_label_output $info_label_autotag $info_label_endings $info_label_input $info_label_time $info_label_bind $info_label_verb \
    $info_label_display -side right -expand no -fill y
} else {
  pack $info_label_hidetag $info_label_autotag -side right -expand no -fill y
  if {[does_text_batch_input]} {
    pack $info_label_adjust $info_label_time $info_label_endings $info_label_stopping -side right -expand no -fill y
    set text_batch_input_labels_shown true
  }
  if {[does_text_batch_output]} {
    pack $mxmbatch_info_label_maxcount_text -side right -expand no -fill y
    set text_batch_output_labels_shown true
  }
}

proc show_hide_batch_maxima_labels_in_text {} {
  global info_label_time
  global info_label_adjust
  global info_label_endings
  global info_label_stopping
  global mxmbatch_info_label_maxcount_text
  global text_batch_input_labels_shown
  global text_batch_output_labels_shown

  if {[does_text_batch_input]} {
    if {!$text_batch_input_labels_shown} {
      pack $info_label_adjust $info_label_time $info_label_endings $info_label_stopping -side right -expand no -fill y
      set text_batch_input_labels_shown true
    }
  } {
    if {$text_batch_input_labels_shown} {
      pack forget $info_label_adjust $info_label_time $info_label_endings $info_label_stopping
      set text_batch_input_labels_shown false
    }
  }
  if {[does_text_batch_output]} {
    if {!$text_batch_output_labels_shown} {
      pack $mxmbatch_info_label_maxcount_text -side right -expand no -fill y
      set text_batch_output_labels_shown true
    }
  } {
    if {$text_batch_output_labels_shown} {
      pack forget $mxmbatch_info_label_maxcount_text
      set text_batch_output_labels_shown false
    }
  }
}

proc update_text_enter_mode_menu {mode_idx} {
  global stack_mode_text_based_command_modes_use_idx
  global stack_mode_text_based_command_modes
  global batch_mode_text_based_command_modes_use_idx
  global batch_mode_text_based_command_modes
  global text_trigger_use_text_selection
  global text_trigger_placed_text_selection
  global text_trigger_use_command
  global text_trigger_plot_selection

  if {[is_stack_mode]} {
    set stack_mode_text_based_command_modes_use_idx $mode_idx
    set use_mode [lindex [lindex $stack_mode_text_based_command_modes $mode_idx] 1]
  } {
    set batch_mode_text_based_command_modes_use_idx $mode_idx
    set use_mode [lindex [lindex $batch_mode_text_based_command_modes $mode_idx] 1]
  }

  set text_trigger_use_text_selection [lindex $use_mode 0]
  set text_trigger_placed_text_selection [lindex $use_mode 1]
  set text_trigger_use_command [lindex $use_mode 2]
  set text_trigger_plot_selection [lindex $use_mode 3]

  if {[is_batch_mode]} {
    show_hide_batch_maxima_labels_in_text
  }
}
update_text_enter_mode_menu [expr \
  {[is_stack_mode] ? $stack_mode_text_based_command_modes_use_idx : $batch_mode_text_based_command_modes_use_idx}]

pack $fmenu.state -side left -expand yes -fill x

set app_label_path_part 1
proc set_app_label_path_part {} {
  global app_label_path_part
  global save_path

  if {$save_path==""} {
    return
  }

  set app_label_path_part [expr {($app_label_path_part+1)%3}]
  app_labels_on_modified
}

proc get_app_label_path_part {} {
  global app_label_path_part
  global save_path

  if {$save_path==""} {
    return ""
  }
  set use_part $save_path
  if {$app_label_path_part==1} {
    set use_part [file tail $save_path]
  }
  if {$app_label_path_part==2} {
    set use_part [file dirname $save_path]
  }

  return $use_part
}

proc copy_app_label_path_part {{ending ""}} {
  set use_part [get_app_label_path_part]
  if {$use_part==""} {
    return
  }

  if {$ending!=""} {
    append use_part $ending
  }

  set_clipboard_text $use_part
  store_clipboard_onset
}

bind $state_label <3> {
  set_app_label_path_part
  break
}
bind $state_label <Control-3> {
  copy_app_label_path_part
  break
}
bind $state_label <Control-Shift-3> {
  copy_app_label_path_part "\n"
  break
}

set command_stack_saving_index -1
set command_stack_saving_paused false
set command_stack_saving_last_text ""
set command_stack_saving_last_path ""

proc check_stack_save_state {} {
  global command_stack_saving_last_text
  global command_stack_saving_index
  global command_stack_index
  global command_stack

  if {$command_stack_saving_index == -1} {
    return -1
  }

  if {[is_command_field_active] && ($command_stack_saving_index==$command_stack_index)} {
    save_current_stack
  }

  if {$command_stack_saving_last_text != [lindex $command_stack $command_stack_saving_index]} {
    return 1
  }

  return 0
}

set backgrounds_for_saving [list "#ffffff" "#fafaff" "#f2f2ff"]
proc update_cmmd_colors {} {
  global command_field
  global command_field_aux
  global command_stack_index
  global command_stack_saving_index
  global command_stack_saving_paused
  global backgrounds_for_saving

  set bgcol_idx 0
  if {$command_stack_saving_index==$command_stack_index} {
    if {$command_stack_saving_paused} {
      set bgcol_idx 1
    } {
      set bgcol_idx 2
    }
  }
  set bgcol [lindex $backgrounds_for_saving $bgcol_idx]
  $command_field configure -background $bgcol
  $command_field_aux configure -background $bgcol

  setting_cmmd_highlight_shade 0
}

# the hex-values are taken from:
# binary encode hex [encoding convertto "utf-8" "_the_char_visually_"]
set mdash_char "-"
set sigma_char "Sigma"
catch {
  set mdash_char [encoding convertfrom "utf-8" "\xE2\x80\x94"]
}
catch {
  set sigma_char [encoding convertfrom "utf-8" "\xCE\xA3"]
}

proc update_stack_label {} {
  global command_stack
  global command_stack_index
  global stack_label_pos
  global stack_label_saving
  global command_stack_saving_index
  global command_stack_saving_paused
  global info_labels_foreground
  global mdash_char

  set stack_count [llength $command_stack]
  set stack_position [expr {$stack_count - $command_stack_index}]

  set saving_label $mdash_char
  if {$command_stack_saving_index>-1} {
    set saving_label [expr {$stack_count - $command_stack_saving_index}]
  }
  set saving_label_fg [lindex $info_labels_foreground [expr {$command_stack_saving_paused ? 0 : 1}]]

  $stack_label_saving configure -text "$saving_label" -foreground $saving_label_fg

  $stack_label_pos configure -text "$stack_position/$stack_count"

  update_cmmd_colors
}

proc set_stack_level_saving {} {
  global command_stack_index
  global command_stack_saving_index
  global command_stack_saving_paused

  set command_stack_saving_index $command_stack_index
  set command_stack_saving_paused false

  update_stack_label
  app_labels_on_modified
}

proc unset_stack_level_saving {} {
  global command_stack_saving_index
  global command_stack_saving_paused
  global command_stack_saving_last_path
  global command_stack_saving_last_text

  set command_stack_saving_index -1
  set command_stack_saving_paused false
  set command_stack_saving_last_text ""
  set command_stack_saving_last_path ""

  update_stack_label
  app_labels_on_modified
}

proc update_stack_level_saving {} {
  global command_stack_index
  global command_stack_saving_index

  if {$command_stack_index!=$command_stack_saving_index} {
    set_stack_level_saving
  } {
    unset_stack_level_saving
  }
}

proc update_pause_stack_level_saving {} {
  global command_stack_saving_index
  global command_stack_saving_paused

  if {$command_stack_saving_index==-1} {
    return
  }

  set command_stack_saving_paused [expr {$command_stack_saving_paused ? false : true}]

  update_stack_label
}

bind $stack_label_saving <1> {
  update_stack_level_saving
}
bind $stack_label_saving <3> {
  update_pause_stack_level_saving
}

proc update_info_minima_tags_request {to_show} {
  global hide_tag_tildes

  set to_hide [expr {!$to_show}]
  if {$hide_tag_tildes==$to_hide} {
    return
  }

  set hide_tag_tildes $to_hide
  set_tilde_hiding false
}

set labels_minima_tags_menu ""
proc offer_info_minima_tags {x_abs y_abs x_pos y_pos} {
  global labels_minima_tags_menu
  global topw
  global info_label_hidetag
  global hide_tag_tildes
  global editor_batch_state
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  if {!$editor_batch_state} {
    if {![is_text_field_active]} {
      return
    }
    if {![is_command_field_active]} {
      return
    }
  }

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  catch {destroy $labels_minima_tags_menu}
  set labels_minima_tags_menu [menu "$topw.labels_minima_tags_menu" -tearoff false]
  $labels_minima_tags_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $labels_minima_tags_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $labels_minima_tags_menu configure -activeborderwidth 1

  $labels_minima_tags_menu add command -label "TGM Minima tags options" \
    -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
  $labels_minima_tags_menu add command -label "~-lines shown" \
    -command [list "update_info_minima_tags_request" true]
  $labels_minima_tags_menu add command -label "~-lines hidden" \
    -command [list "update_info_minima_tags_request" false]

  set cur_idx 1
  if {$hide_tag_tildes} {
    incr cur_idx 1
  }

  $labels_minima_tags_menu entryconfigure $cur_idx -background $menu_selected_background -activebackground "#e0e0ff"

  tk_popup $labels_minima_tags_menu $use_x $use_y
}

set cursor_was_at_block_border false
proc set_tilde_hiding {change} {
  global info_label_hidetag
  global hide_tag_tildes
  global text_field
  global text_inner_codes_start_tag
  global text_inner_codes_end_tag
  global info_labels_foreground
  global cursor_was_at_block_border

  if {(![is_text_field_active]) && $change} {
    return
  }

  if {$change} {
    set hide_tag_tildes [expr {$hide_tag_tildes ? false : true}]
  }

  $info_label_hidetag configure -foreground [lindex $info_labels_foreground [expr {$hide_tag_tildes ? 0 : 1}]]

  $text_field tag configure $text_inner_codes_start_tag -elide $hide_tag_tildes
  $text_field tag configure $text_inner_codes_end_tag -elide $hide_tag_tildes

  if {$cursor_was_at_block_border && $hide_tag_tildes} {
    update_text_properties
  }
}

bind $info_label_hidetag <1> {
  if {[is_text_field_active]} {
    offer_info_minima_tags %X %Y %x %y
  }
}

proc update_info_outputs_tags_enforcing_request {enf_val} {
  global info_label_autotag
  global omit_auto_tildes

  set omit_auto_tildes [expr {$enf_val ? false : true}]

  set_auto_tagging false
}

set labels_outputs_tags_enforcing_menu ""
proc offer_info_outputs_tags_enforcing {x_abs y_abs x_pos y_pos} {
  global labels_outputs_tags_enforcing_menu
  global topw
  global info_label_autotag
  global omit_auto_tildes
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  if {$labels_outputs_tags_enforcing_menu != ""} {
    catch {destroy $labels_outputs_tags_enforcing_menu}
  }
  set labels_outputs_tags_enforcing_menu [menu "$topw.labels_outputs_tags_enforcing_menu" -tearoff false]
  $labels_outputs_tags_enforcing_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $labels_outputs_tags_enforcing_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $labels_outputs_tags_enforcing_menu configure -activeborderwidth 1

  $labels_outputs_tags_enforcing_menu add command -label "Block insertion:" \
    -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  $labels_outputs_tags_enforcing_menu add command -label "~-lines enforced" \
    -command [list "update_info_outputs_tags_enforcing_request" true]
  $labels_outputs_tags_enforcing_menu add command -label "~-lines manual only" \
    -command [list "update_info_outputs_tags_enforcing_request" false]

  set cur_idx 1
  if {$omit_auto_tildes} {
    incr cur_idx 1
  }

  $labels_outputs_tags_enforcing_menu entryconfigure $cur_idx -background $menu_selected_background -activebackground "#e0e0ff"

  tk_popup $labels_outputs_tags_enforcing_menu $use_x $use_y
}

proc update_info_outputs_tags_placing_request {to_enforce to_place} {
  global omit_auto_tildes
  global info_label_output_index

  set omit_auto_tildes [expr {!$to_enforce}]
  set info_label_output_index $to_place

  set_auto_tagging false
  set_output_placing 0
}

set labels_outputs_tags_placing_menu ""
proc offer_info_outputs_tags_placing {x_abs y_abs x_pos y_pos} {
  global labels_outputs_tags_placing_menu
  global topw
  global info_label_autotag
  global omit_auto_tildes
  global info_label_output_modes
  global info_label_output_index
  global info_label_output
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  # ensuring that we still have info_label_output_modes==[list "s" "v" "e"]
  # if it would (for whatever reason) changed, this would need update then;
  if {[llength $info_label_output_modes]!=3} {
    return false
  }
  if {[lindex $info_label_output_modes 0]!="s"} {
    return false
  }
  if {[lindex $info_label_output_modes 1]!="v"} {
    return false
  }
  if {[lindex $info_label_output_modes 2]!="e"} {
    return false
  }

  set use_x [expr {$x_abs - 120}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  catch {destroy $labels_outputs_tags_placing_menu}
  set labels_outputs_tags_placing_menu [menu "$topw.labels_outputs_tags_placing_menu" -tearoff false]
  $labels_outputs_tags_placing_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $labels_outputs_tags_placing_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $labels_outputs_tags_placing_menu configure -activeborderwidth 1

  $labels_outputs_tags_placing_menu add command -label "Result insertion: TGM Minima tags and placing" \
    -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  $labels_outputs_tags_placing_menu add command -label "~-lines enforced | output inserted by cursor" \
    -command [list "update_info_outputs_tags_placing_request" true 0]
  $labels_outputs_tags_placing_menu add command -label "~-lines enforced | output inserted at free place" \
    -command [list "update_info_outputs_tags_placing_request" true 1]
  $labels_outputs_tags_placing_menu add command -label "~-lines enforced | output pushed over to end" \
    -command [list "update_info_outputs_tags_placing_request" true 2]
  $labels_outputs_tags_placing_menu add command -label "  no ~ enforcing | output inserted by cursor" \
    -command [list "update_info_outputs_tags_placing_request" false 0]
  $labels_outputs_tags_placing_menu add command -label "  no ~ enforcing | output inserted at free place" \
    -command [list "update_info_outputs_tags_placing_request" false 1]
  $labels_outputs_tags_placing_menu add command -label "  no ~ enforcing | output pushed over to end" \
    -command [list "update_info_outputs_tags_placing_request" false 2]

  set cur_idx 1
  if {$omit_auto_tildes} {
    incr cur_idx 3
  }
  incr cur_idx $info_label_output_index

  $labels_outputs_tags_placing_menu entryconfigure $cur_idx -background $menu_selected_background -activebackground "#e0e0ff"

  tk_popup $labels_outputs_tags_placing_menu $use_x $use_y
}

proc set_auto_tagging {{to_change true}} {
  global omit_auto_tildes
  global info_label_autotag
  global info_labels_foreground

  if {$to_change} {
    set omit_auto_tildes [expr {$omit_auto_tildes ? false : true}]
  }

  $info_label_autotag configure -foreground [lindex $info_labels_foreground [expr {$omit_auto_tildes ? 0 : 1}]]
}

bind $info_label_autotag <1> {
  if {[is_stack_mode]} {
    offer_info_outputs_tags_placing %X %Y %x %y
  } {
    offer_info_outputs_tags_enforcing %X %Y %x %y
  }
}

set force_output_insitu false
set force_output_to_end false
proc set_output_placing {change} {
  global info_label_output_modes
  global info_label_output_index
  global info_label_output
  global force_output_insitu
  global force_output_to_end

  incr info_label_output_index $change
  set info_label_output_index [expr {$info_label_output_index % [llength $info_label_output_modes]}]

  set curr_mode [lindex $info_label_output_modes $info_label_output_index]

  $info_label_output configure -text $curr_mode

  set force_output_insitu false
  set force_output_to_end false

  if {$curr_mode=="s"} {
    set force_output_insitu true
  }
  if {$curr_mode=="e"} {
    set force_output_to_end true
  }
}
set_output_placing 0

bind $info_label_output <1> {
  offer_info_outputs_tags_placing %X %Y %x %y
}

set info_line_long true
set info_line_left true
set info_line_1dim true
set info_line_inner false
set info_line_tex false

set control_output_onset_shibboleth [strcat \
  "//" [expr {int(10000*rand()+20000)}] [expr {int(10000*rand()+20000)}] "//"]

set waiting_for_control_answer false
set maxima_in_setting false
set maxima_in_setting_start false
proc call_maxima_line_setting {{to_onset true}} {
  global maxima_in_setting
  global maxima_in_setting_start
  global maxima_command
  global maxima_command_control
  global control_output_onset_shibboleth
  global state_label
  global bt_restart
  global info_line_size_asked
  global info_line_left_asked
  global info_line_1dim_asked
  global info_line_long_asked
  global info_line_inner_asked
  global info_tag_start
  global info_tag_middle
  global info_tag_end
  global bind_labels_asked
  global with_labels_asked
  global maxima_setting_done
  global has_ever_asked
  global after_restart_fresh
  global newly_at_maxima_outputs
  #global can_kill_label
  global waiting_for_control_answer
  global should_display_cpu_usage
  global could_display_cpu_usage
  global mxm_display_exp_dstar
  global asked_temporary_path

  if {$maxima_command!=""} {
    # should not interfere with common commands;
    # i.e. if a common command gets canceled,
    # its "maxima_command" has tobe cleared;
    return
  }

  # disable widgets
  set_state_disabled
  $bt_restart configure -state normal

  # interrupt-button is set once the shibboleth is read
  #set_button_interrupt_sets

  set use_line_left $info_line_left_asked
  set use_line_long $info_line_long_asked
  set use_line_1dim $info_line_1dim_asked

  if {$info_line_1dim_asked} {
    set use_line_left true
    # some forms get split by Maxima, some are dealt with Minima;
    #set use_line_long true
  } {
    set use_line_long false
  }

  if {$info_line_inner_asked} {
    set use_line_1dim true
    set use_line_left true
    set use_line_long true
  }

  # prepare command
  set maxima_command_control ""

  if {$to_onset} {
    append maxima_command_control ":lisp(format t \"~&$control_output_onset_shibboleth~%\");\n"
  }

  if {(!$has_ever_asked) && $after_restart_fresh} {
    if {$asked_temporary_path!=""} {
      append maxima_command_control [strcat \
        "maxima_tempdir:\"" [string map [list "\\" "\\\\" "\"" "\\\""] $asked_temporary_path] "\"\$\n" \
      ]
    }
    if {$mxm_display_exp_dstar} {
      append maxima_command_control ":lisp(setf (get 'mexpt 'grind) 'msize-infix);\n"
    }
  }

  if {!$has_ever_asked} {
    append maxima_command_control "(linel:79,inchar:%i,outchar:%o,linechar:%t,ttyoff:false,"
  } {
    append maxima_command_control ":lisp(progn (setf \$linel 79)(setf linel 79)(setf #.ttyoff nil)"
  }

  if {!$has_ever_asked} {
    append maxima_command_control "?\\*maxima\\-quiet\\*:false,"
  } {
    append maxima_command_control "(setf *maxima-quiet* nil)"
  }

  if {$with_labels_asked} {
    if {!$has_ever_asked} {
      append maxima_command_control "?\\*display\\-labels\\-p\\*:true,"
    } {
      append maxima_command_control "(setf *display-labels-p* t)"
    }
  } {
    if {!$has_ever_asked} {
      append maxima_command_control "?\\*display\\-labels\\-p\\*:false,"
    } {
      append maxima_command_control "(setf *display-labels-p* nil)"
    }
  }
  if {$use_line_left} {
    if {!$has_ever_asked} {
      append maxima_command_control "leftjust:true,"
    } {
      append maxima_command_control "(setf \$leftjust t)"
    }
  } {
    if {!$has_ever_asked} {
      append maxima_command_control "leftjust:false,"
    } {
      append maxima_command_control "(setf \$leftjust nil)"
    }
  }
  if {$use_line_1dim} {
    if {!$has_ever_asked} {
      append maxima_command_control "display2d:false,"
    } {
      append maxima_command_control "(setf \$display2d nil)"
    }
  } {
    if {!$has_ever_asked} {
      append maxima_command_control "display2d:true,"
    } {
      append maxima_command_control "(setf \$display2d t)"
    }
  }
  set maxima_line_length [lindex $info_line_size_asked [expr {$use_line_long ? 1 : 0}]]
  if {!$has_ever_asked} {
    append maxima_command_control "linel:$maxima_line_length,"
  } {
    append maxima_command_control "(setf \$linel $maxima_line_length)(setf linel $maxima_line_length)"
  }

  if {!$has_ever_asked} {
    set nolabels_value [expr {$bind_labels_asked ? "false" : "true"}]
    append maxima_command_control "nolabels:$nolabels_value,"
  } {
    set nolabels_value [expr {$bind_labels_asked ? "nil" : "t"}]
    append maxima_command_control "(setf \$nolabels $nolabels_value)"
  }

  set alt_disp_mode_wl "(format nil \"~*\" (setf *alt-display1d* (lambda (form) (format t \"(~a) ~a~%\" (string-downcase (subseq (format nil \"~a\" (cadr form)) 1)) (caddr form)))))"
  set alt_disp_mode_wol "(format nil \"~*\" (setf *alt-display1d* (lambda (form) (format t \"~a~%\" (caddr form)))))"
  set alt_disp_mode [expr {$with_labels_asked ? $alt_disp_mode_wl : $alt_disp_mode_wol}]

  if {!$has_ever_asked} {
    if {$info_line_inner_asked} {
      append maxima_command_control "1);\n:lisp$alt_disp_mode;\n("
    } {
      append maxima_command_control "?\\*alt\\-display1d\\*:false,"
    }
  } {
    if {$info_line_inner_asked} {
      append maxima_command_control $alt_disp_mode
    } {
      append maxima_command_control "(setf *alt-display1d* nil)"
    }
  }

  if {!$has_ever_asked} {
    append maxima_command_control "?\\*alt\\-display2d\\*:false,"
  } {
    append maxima_command_control "(setf *alt-display2d* nil)"
  }

  if {!$has_ever_asked} {
    append maxima_command_control "showtime:true,"
  } {
    append maxima_command_control "(setf \$showtime t)"
  }

  set asking_grind_part "\"$info_tag_start$info_tag_end\""
  if {!$has_ever_asked} {
    if {$after_restart_fresh} {
      set asking_grind_part "concat(\"$info_tag_start\",build_info()@version,\"$info_tag_middle\",build_info()@lisp_name,\"$info_tag_end\")"
    }
    append maxima_command_control "grind($asking_grind_part),kill(labels))\$"
  } {
    append maxima_command_control "(mgrind $asking_grind_part t)"
    if {$with_labels_asked} {
      append maxima_command_control "(terpri)(\$disp (\$concat \"(\" \$inchar \$linenum \")\"))(terpri)"
    }
    append maxima_command_control "(force-output) #\\newline);"
  }

  # these are set via "consume_read_parts"
  # once the "control_output_onset_shibboleth" value
  # comes from Maxima, notifying that Maxima actually
  # got it into taking the control command;
  #set maxima_in_setting true
  #set maxima_in_setting_start true
  #set maxima_setting_done false
  #set newly_at_maxima_outputs false

  set should_display_cpu_usage false
  set could_display_cpu_usage false
  set_state_label_text "setting"

  #set can_kill_label true
  if {$to_onset} {
    set waiting_for_control_answer true
  }

  svwrite_onset
}

proc set_info_batchstopping_request {to_stop {to_clean ""}} {
  global batch_maxima_end_automatic
  global batch_maxima_end_cleaning
  global info_labels_foreground
  global info_label_stopping
  global mxmbatch_info_label_stopping

  set batch_maxima_end_automatic $to_stop
  if {$to_clean!=""} {
    set batch_maxima_end_cleaning $to_clean
  }

  $info_label_stopping configure -text [get_batch_ending_label_char]

  if {![is_batch_mode]} {
    return
  }
  if {$mxmbatch_info_label_stopping==""} {
    return
  }

  $mxmbatch_info_label_stopping configure -text [get_batch_ending_label_char]
}

set labels_make_batchstopping_menu ""
proc mxmbatch_offer_info_batch_stopping {x_abs y_abs x_pos y_pos {at_batch true}} {
  global labels_make_batchstopping_menu
  global topw
  global batch_maxima_end_automatic
  global batch_maxima_end_cleaning
  global mxmbatch_info_label_stopping
  global info_label_stopping
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  if {![is_batch_mode]} {
    return
  }
  if {$at_batch && ($mxmbatch_info_label_stopping=="")} {
    return
  }

  if {$at_batch} {
    set x_offset $default_menu_offset_x
  } {
    set x_offset "-100"
  }
  set use_x [expr {$x_abs + $x_offset}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  catch {destroy $labels_make_batchstopping_menu}
  set labels_make_batchstopping_menu [menu "$topw.labels_make_batchstopping_menu" -tearoff false]
  $labels_make_batchstopping_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $labels_make_batchstopping_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $labels_make_batchstopping_menu configure -activeborderwidth 1

  set label_text "Maxima quits automatically at batch end"
  $labels_make_batchstopping_menu add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  set label_yes "yes (after the commands are executed)"
  set label_no "no (let Maxima running, e.g. for plotting)"
  set label_clean "no (let Maxima running) but do cleaning"

  $labels_make_batchstopping_menu add command -label $label_yes \
    -command [list "set_info_batchstopping_request" true]
  $labels_make_batchstopping_menu add command -label $label_no \
    -command [list "set_info_batchstopping_request" false false]
  $labels_make_batchstopping_menu add command -label $label_clean \
    -command [list "set_info_batchstopping_request" false true]

  set cur_idx 1
  if {!$batch_maxima_end_automatic} {
    incr cur_idx 1
    if {$batch_maxima_end_cleaning} {
      incr cur_idx 1
    }
  }

  $labels_make_batchstopping_menu entryconfigure $cur_idx -background $menu_selected_background -activebackground "#e0e0ff"

  $labels_make_batchstopping_menu add command -label "running batches admit line-wise answers" -state "disabled" -background "#e0f0e0" -activebackground "#e0f0e0"

  tk_popup $labels_make_batchstopping_menu $use_x $use_y
}

proc show_info_label_endings_state {} {
  global info_labels_foreground
  global info_label_endings
  global make_line_endings

  if {[is_stack_mode] && (![is_stack_mode_common])} {
    $info_label_endings configure -foreground [lindex $info_labels_foreground 0]
    return
  }

  $info_label_endings configure -foreground [lindex $info_labels_foreground [expr {$make_line_endings ? 1 : 0}]]
}

proc set_info_lineendings_request {to_make} {
  global info_labels_foreground
  global mxmbatch_info_label_endings
  global make_line_endings

  set make_line_endings $to_make

  show_info_label_endings_state

  if {![is_batch_mode]} {
    return
  }
  if {$mxmbatch_info_label_endings==""} {
    return
  }

  $mxmbatch_info_label_endings configure -foreground [lindex $info_labels_foreground [expr {$make_line_endings ? 1 : 0}]]
}

set labels_make_lineending_menu ""
proc offer_info_make_lineendings {x_abs y_abs x_pos y_pos {at_batch false}} {
  global labels_make_lineending_menu
  global topw
  global make_line_endings
  global mxmbatch_info_label_endings
  global info_label_endings
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  if {(![is_batch_mode]) && (![is_stack_mode_common])} {
    return
  }
  if {$at_batch && ($mxmbatch_info_label_endings=="")} {
    return
  }

  if {$at_batch} {
    set x_offset $default_menu_offset_x
  } {
    set x_offset "-100"
  }
  set use_x [expr {$x_abs + $x_offset}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  catch {destroy $labels_make_lineending_menu}
  set labels_make_lineending_menu [menu "$topw.labels_make_lineending_menu" -tearoff false]
  $labels_make_lineending_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $labels_make_lineending_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $labels_make_lineending_menu configure -activeborderwidth 1

  $labels_make_lineending_menu add command -label "End (inner) command lines with ;" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  set label_no "no (just overall command end ensured)"
  set label_yes "yes (for line-wise expressions only)"

  $labels_make_lineending_menu add command -label $label_no \
    -command [list "set_info_lineendings_request" false]
  $labels_make_lineending_menu add command -label $label_yes \
    -command [list "set_info_lineendings_request" true]

  set cur_idx 1
  if {$make_line_endings} {
    incr cur_idx 1
  }

  $labels_make_lineending_menu entryconfigure $cur_idx -background $menu_selected_background -activebackground "#e0e0ff"

  $labels_make_lineending_menu add command -label "for sets of simple lines only!" -state "disabled" -background "#f0e9e0" -activebackground "#f0e9e0"

  tk_popup $labels_make_lineending_menu $use_x $use_y
}

proc set_info_display_adjust_request {{to_dstar ""} {to_ratsimp ""}} {
  global mxm_display_exp_dstar
  global mxm_display_sorted_ratsimp
  global info_label_adjust
  global mxmbatch_info_label_adjust

  if {![is_batch_mode]} {
    return
  }

  if {$to_dstar != ""} {
    set mxm_display_exp_dstar $to_dstar
  }

  if {$to_ratsimp != ""} {
    set mxm_display_sorted_ratsimp $to_ratsimp
  }

  $info_label_adjust configure -text [get_display_adjust_char]

  if {$mxmbatch_info_label_adjust==""} {
    return
  }

  $mxmbatch_info_label_adjust configure -text [get_display_adjust_char]
}

set labels_display_adjust_menu ""
proc mxmbatch_offer_info_display_adjust {x_abs y_abs x_pos y_pos {at_batch true}} {
  global labels_display_adjust_menu
  global topw
  global mxm_display_exp_dstar
  global mxm_display_sorted_ratsimp
  global info_label_adjust
  global mxmbatch_info_label_adjust
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  if {![is_batch_mode]} {
    return
  }
  if {$at_batch && ($mxmbatch_info_label_adjust=="")} {
    return
  }

  if {$at_batch} {
    set x_offset $default_menu_offset_x
  } {
    set x_offset "-100"
  }
  set use_x [expr {$x_abs + $x_offset}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  catch {destroy $labels_display_adjust_menu}
  set labels_display_adjust_menu [menu "$topw.labels_display_adjust_menu" -tearoff false]
  $labels_display_adjust_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $labels_display_adjust_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $labels_display_adjust_menu configure -activeborderwidth 1

  $labels_display_adjust_menu add command -label "Display adjusting for Maxima expressions" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  set label_none "no adjusting (keep a^b, without ratsimp)"
  set label_dexp "output a**b instead of a^b (in 1D displays)"
  set label_ratsimp "add ratsimp (to 1D displays, e.g. input)"
  set label_both "both a**b and ratsimp"

  $labels_display_adjust_menu add command -label $label_none \
    -command [list "set_info_display_adjust_request" false false]
  $labels_display_adjust_menu add command -label $label_dexp \
    -command [list "set_info_display_adjust_request" true false]
  $labels_display_adjust_menu add command -label $label_ratsimp \
    -command [list "set_info_display_adjust_request" false true]
  $labels_display_adjust_menu add command -label $label_both \
    -command [list "set_info_display_adjust_request" true true]

  set cur_idx 1
  if {$mxm_display_exp_dstar} {
    incr cur_idx 1
  }
  if {$mxm_display_sorted_ratsimp} {
    incr cur_idx 2
  }

  $labels_display_adjust_menu entryconfigure $cur_idx -background $menu_selected_background -activebackground "#e0e0ff"

  tk_popup $labels_display_adjust_menu $use_x $use_y
}

proc set_info_elapsed_time_request {to_show} {
  global show_time_elapsed
  global info_labels_foreground
  global info_label_time
  global mxmbatch_info_label_time

  set show_time_elapsed $to_show

  $info_label_time configure -foreground [lindex $info_labels_foreground [expr {$show_time_elapsed ? 1 : 0}]]

  if {![is_batch_mode]} {
    return
  }
  if {$mxmbatch_info_label_time==""} {
    return
  }

  $mxmbatch_info_label_time configure -foreground [lindex $info_labels_foreground [expr {$show_time_elapsed ? 1 : 0}]]
}

proc set_info_input_commands_request {to_show} {
  global info_labels_foreground
  global info_label_input
  global include_maxima_input_default

  set include_maxima_input_default $to_show

  if {![is_stack_mode]} {
    return
  }

  $info_label_input configure -foreground [lindex $info_labels_foreground [expr {$include_maxima_input_default ? 1 : 0}]]
}

set labels_show_time_menu ""
proc mxmbatch_offer_info_show_time {x_abs y_abs x_pos y_pos {at_batch true}} {
  global labels_show_time_menu
  global topw
  global show_time_elapsed
  global mxmbatch_info_label_time
  global info_label_time
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  if {![is_batch_mode]} {
    return
  }
  if {$at_batch && ($mxmbatch_info_label_time=="")} {
    return
  }

  if {$at_batch} {
    set x_offset $default_menu_offset_x
  } {
    set x_offset "-100"
  }
  set use_x [expr {$x_abs + $x_offset}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  catch {destroy $labels_show_time_menu}
  set labels_show_time_menu [menu "$topw.labels_show_time_menu" -tearoff false]
  $labels_show_time_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $labels_show_time_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $labels_show_time_menu configure -activeborderwidth 1

  $labels_show_time_menu add command -label "\"Elapsed time\" as returned by Maxima" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  set label_without "not put into results"
  set label_with "is included within results"

  $labels_show_time_menu add command -label $label_without \
    -command [list "set_info_elapsed_time_request" false]
  $labels_show_time_menu add command -label $label_with \
    -command [list "set_info_elapsed_time_request" true]

  set cur_idx 1
  if {$show_time_elapsed} {
    incr cur_idx 1
  }

  $labels_show_time_menu entryconfigure $cur_idx -background $menu_selected_background -activebackground "#e0e0ff"

  tk_popup $labels_show_time_menu $use_x $use_y
}

proc update_info_elapsed_time_input_setting {time_info input_commands} {
  set_info_elapsed_time_request $time_info
  set_info_input_commands_request $input_commands
}

set labels_show_time_menu ""
proc offer_info_show_time_input {x_abs y_abs x_pos y_pos} {
  global labels_show_time_menu
  global in_lisp_mode
  global in_debug_mode
  global topw
  global info_label_time
  global show_time_elapsed
  global include_maxima_input_default
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  if {![is_stack_mode]} {
    return
  }

  set x_offset "-100"
  set use_x [expr {$x_abs + $x_offset}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  catch {destroy $labels_show_time_menu}
  set labels_show_time_menu [menu "$topw.labels_show_time_menu" -tearoff false]
  $labels_show_time_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $labels_show_time_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $labels_show_time_menu configure -activeborderwidth 1

  $labels_show_time_menu add command -label "Output with input and/or elapsed-time info:" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  $labels_show_time_menu add command -label "with none" \
    -command [list "update_info_elapsed_time_input_setting" false false]
  $labels_show_time_menu add command -label "with elapsed-time info" \
    -command [list "update_info_elapsed_time_input_setting" true false]
  $labels_show_time_menu add command -label "with input commands" \
    -command [list "update_info_elapsed_time_input_setting" false true]
  $labels_show_time_menu add command -label "with both" \
    -command [list "update_info_elapsed_time_input_setting" true true]

  set cur_idx 1
  if {$show_time_elapsed} {
    incr cur_idx 1
  }
  if {$include_maxima_input_default} {
    incr cur_idx 2
  }

  $labels_show_time_menu entryconfigure $cur_idx -background $menu_selected_background -activebackground "#e0e0ff"

  tk_popup $labels_show_time_menu $use_x $use_y
}

# these functions used setting via shortcuts
proc change_info_batch_stopping_request {} {
  global batch_maxima_end_automatic

  set_info_batchstopping_request [expr {$batch_maxima_end_automatic ? false : true}]
}

proc change_info_make_lineendings_request {} {
  global make_line_endings

  set_info_lineendings_request [expr {$make_line_endings ? false : true}]
}

proc change_info_display_adjust_request {on_dstar on_ratsimp} {
  global mxm_display_exp_dstar
  global mxm_display_sorted_ratsimp

  set new_exp_dstar [expr {$on_dstar ? (!$mxm_display_exp_dstar) : ""}]
  set new_sorted_ratsimp [expr {$on_ratsimp ? (!$mxm_display_sorted_ratsimp) : ""}]

  set_info_display_adjust_request $new_exp_dstar $new_sorted_ratsimp
}
proc change_info_show_time_request {} {
  global show_time_elapsed

  set_info_elapsed_time_request [expr {$show_time_elapsed ? false : true}]
}
proc change_info_include_input_request {} {
  global include_maxima_input_default

  set_info_input_commands_request [expr {$include_maxima_input_default ? false : true}]
}

bind $info_label_stopping <1> {
  remove_torn_menus
  if {[is_batch_mode]} {
    mxmbatch_offer_info_batch_stopping %X %Y %x %y false
  }
}

bind $info_label_endings <1> {
  remove_torn_menus
  if {[is_batch_mode] || [is_stack_mode_common]} {
    offer_info_make_lineendings %X %Y %x %y false
  }
}

bind $info_label_adjust <1> {
  remove_torn_menus
  mxmbatch_offer_info_display_adjust %X %Y %x %y false
}

bind $info_label_time <1> {
  remove_torn_menus
  if {[is_stack_mode]} {
    offer_info_show_time_input %X %Y %x %y
  } {
    mxmbatch_offer_info_show_time %X %Y %x %y false
  }
}
bind $info_label_input <1> {
  remove_torn_menus
  offer_info_show_time_input %X %Y %x %y
}

proc mxmbatch_update_info_labels_shown_bound_request {change {to_display_labels false} {to_bind_labels false}} {
  global mxmbatch_info_label_verb
  global mxmbatch_info_label_bind
  global bind_labels
  global with_labels
  global info_labels_foreground

  if {![is_batch_mode]} {
    return
  }
  if {($mxmbatch_info_label_verb=="") || ($mxmbatch_info_label_bind=="")} {
    return
  }

  if {$change} {
    set with_labels $to_display_labels
    set bind_labels $to_bind_labels
  }

  $mxmbatch_info_label_verb configure -foreground [lindex $info_labels_foreground [expr {$with_labels ? 1 : 0}]]
  $mxmbatch_info_label_bind configure -foreground [lindex $info_labels_foreground [expr {$bind_labels ? 1 : 0}]]
}

set bind_labels_asked $bind_labels
set with_labels_asked $with_labels
proc update_info_labels_shown_bound_request {change {to_display_labels false} {to_bind_labels false}} {
  global info_label_verb
  global info_label_bind
  global bind_labels
  global with_labels
  global bind_labels_asked
  global with_labels_asked
  global info_labels_foreground

  if {[is_batch_mode]} {
    mxmbatch_update_info_labels_shown_bound_request $change $to_display_labels $to_bind_labels
    return
  }

  if {$change} {
    set with_labels_asked $to_display_labels
    set bind_labels_asked $to_bind_labels
  } {
    set with_labels_asked $with_labels
    set bind_labels_asked $bind_labels
  }

  if {!$change} {
    $info_label_verb configure -foreground [lindex $info_labels_foreground [expr {$with_labels_asked ? 1 : 0}]]
    $info_label_bind configure -foreground [lindex $info_labels_foreground [expr {$bind_labels_asked ? 1 : 0}]]
    return
  }

  call_maxima_line_setting
}

set labels_shown_bound_menu ""
proc offer_info_labels_shown_bound {x_abs y_abs x_pos y_pos} {
  global labels_shown_bound_menu
  global in_lisp_mode
  global in_debug_mode
  global topw
  global info_label_verb
  global info_label_bind
  global bind_labels
  global with_labels
  global mxmbatch_info_label_verb
  global mxmbatch_info_label_bind
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  if {[is_batch_mode] && (($mxmbatch_info_label_verb=="") || ($mxmbatch_info_label_bind==""))} {
    return
  }

  if {[is_stack_mode]} {
    if {$in_lisp_mode || $in_debug_mode} {
      return
    }
    if {![is_command_field_active]} {
      return
    }
  }

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  catch {destroy $labels_shown_bound_menu}
  set labels_shown_bound_menu [menu "$topw.labels_shown_bound_menu" -tearoff false]
  $labels_shown_bound_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $labels_shown_bound_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $labels_shown_bound_menu configure -activeborderwidth 1

  $labels_shown_bound_menu add command -label "Maxima labels options" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  $labels_shown_bound_menu add command -label "labels not displayed, nor bound" \
    -command [list "update_info_labels_shown_bound_request" true false false]
  $labels_shown_bound_menu add command -label "labels not displayed, are bound" \
    -command [list "update_info_labels_shown_bound_request" true false true]
  $labels_shown_bound_menu add command -label "labels are displayed, not bound" \
    -command [list "update_info_labels_shown_bound_request" true true false]
  $labels_shown_bound_menu add command -label "labels are displayed and bound" \
    -command [list "update_info_labels_shown_bound_request" true true true]

  set cur_idx 1
  if {$with_labels} {
    incr cur_idx 2
  }
  if {$bind_labels} {
    incr cur_idx 1
  }

  $labels_shown_bound_menu entryconfigure $cur_idx -background $menu_selected_background -activebackground "#e0e0ff"

  tk_popup $labels_shown_bound_menu $use_x $use_y
}

bind $info_label_bind <1> {
  if {[is_command_field_active]} {
    offer_info_labels_shown_bound %X %Y %x %y
  }
}
# this for calls separate from the labels-showing menu;
# used via shortcuts binding;
# it can be set via "offer_info_labels_shown_bound"
# (that is called by mouse clicks) too;
proc mxmbatch_update_info_labels_bound_request {{change true}} {
  global bind_labels
  global info_labels_foreground
  global mxmbatch_info_label_bind

  if {$mxmbatch_info_label_bind==""} {
    return
  }

  if {$change} {
    set bind_labels [expr {$bind_labels ? false : true}]
  }

  $mxmbatch_info_label_bind configure -foreground [lindex $info_labels_foreground [expr {$bind_labels ? 1 : 0}]]
}
proc update_info_labels_bound_request {{change true}} {
  global bind_labels
  global info_labels_foreground
  global info_label_bind
  global in_lisp_mode
  global in_debug_mode
  global bind_labels_asked

  if {$change && ($in_lisp_mode || $in_debug_mode)} {
    return
  }

  if {$change} {
    set bind_labels_asked [expr {$bind_labels ? false : true}]
  }

  $info_label_bind configure -foreground [lindex $info_labels_foreground [expr {$bind_labels ? 1 : 0}]]

  if {$change} {
    call_maxima_line_setting
  }
}

bind $info_label_display <1> {
  if {[is_command_field_active]} {
    offer_info_display_request %X %Y %x %y
  }
}

set display_modes_menu ""
proc offer_info_display_request {x_abs y_abs x_pos y_pos} {
  global display_modes_menu
  global in_lisp_mode
  global in_debug_mode
  global topw
  global info_label_display
  global line_display_modes_list
  global line_display_mode_current
  global mxmbatch_info_label_display
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  if {[is_batch_mode] && ($mxmbatch_info_label_display=="")} {
    return
  }

  if {[is_stack_mode]} {
    if {$in_lisp_mode || $in_debug_mode} {
      return
    }
  }

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  catch {destroy $display_modes_menu}
  set display_modes_menu [menu "$topw.display_modes_menu" -tearoff false]
  $display_modes_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $display_modes_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $display_modes_menu configure -activeborderwidth 1

  $display_modes_menu add command -label "Maxima display options" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  set mode_idx 0
  set mode_len [llength $line_display_modes_list]
  while {$mode_idx<$mode_len} {
    set curr_idx $mode_idx
    set curr_mode [lindex $line_display_modes_list $curr_idx]
    set label_text [lindex $curr_mode 1]
    incr mode_idx

    $display_modes_menu add command -label $label_text -command [list "update_info_display_request" true $curr_idx]
    if {$line_display_mode_current==$curr_idx} {
      $display_modes_menu entryconfigure $mode_idx -background $menu_selected_background -activebackground "#e0e0ff"
    }
  }
  tk_popup $display_modes_menu $use_x $use_y
}

set line_display_mode_current_asked $line_display_mode_current
proc update_info_display_request {change {idx -1}} {
  global display_modes_menu
  global in_lisp_mode
  global in_debug_mode
  global info_label_display
  global line_display_modes_list
  global line_display_mode_current
  global line_display_mode_current_asked
  global info_line_long
  global info_line_left
  global info_line_1dim
  global info_line_inner
  global maxima_output_to_split
  global info_line_long_asked
  global info_line_left_asked
  global info_line_1dim_asked
  global info_line_inner_asked
  global maxima_output_to_split_asked
  global info_line_size

  if {[is_batch_mode]} {
    mxmbatch_update_info_display_request $change $idx
    return
  }

  if {($in_lisp_mode || $in_debug_mode) && $change} {
    return
  }

  if {$change} {
    set line_display_mode_current_asked $idx
  } {
    set line_display_mode_current_asked $line_display_mode_current
  }

  set asked_mode [lindex $line_display_modes_list $line_display_mode_current_asked]
  set mode_values [lindex $asked_mode 2]
  # the flags set there: long left 1dim inner
  set info_line_long_asked [lindex $mode_values 0]
  set info_line_left_asked [lindex $mode_values 1]
  set info_line_1dim_asked [lindex $mode_values 2]
  set info_line_inner_asked [lindex $mode_values 3]

  set maxima_output_to_split_asked [lindex $mode_values 5]
  # Tk has issues with (too) long lines, thus trying to avoid them

  if {!$change} {
    set info_line_long $info_line_long_asked
    set info_line_left $info_line_left_asked
    set info_line_1dim $info_line_1dim_asked
    set info_line_inner $info_line_inner_asked
    set maxima_output_to_split $maxima_output_to_split_asked

    $info_label_display configure -text [lindex $asked_mode 0]
  }

  if {$change} {
    call_maxima_line_setting
  }
}
if {[is_stack_mode]} {
  update_info_display_request false
}

proc mxmbatch_update_info_display_request {change {idx -1}} {
  global display_modes_menu
  global line_display_modes_list
  global line_display_mode_current
  global info_line_long
  global info_line_left
  global info_line_1dim
  global info_line_inner
  global info_line_tex
  global maxima_output_to_split
  global mxmbatch_info_label_display

  if {![is_batch_mode]} {
    return
  }
  if {$mxmbatch_info_label_display==""} {
    return
  }

  if {$change} {
    set line_display_mode_current $idx
  }

  set asked_mode [lindex $line_display_modes_list $line_display_mode_current]
  set mode_values [lindex $asked_mode 2]
  # the flags set there: long left 1dim inner
  set info_line_long [lindex $mode_values 0]
  set info_line_left [lindex $mode_values 1]
  set info_line_1dim [lindex $mode_values 2]
  set info_line_inner [lindex $mode_values 3]
  set info_line_tex [lindex $mode_values 4]

  set maxima_output_to_split [lindex $mode_values 5]
  # Tk has issues with (too) long lines, thus trying to avoid them

  $mxmbatch_info_label_display configure -text [lindex $asked_mode 0]
}

bind $info_label_verb <1> {
  if {[is_command_field_active]} {
    offer_info_labels_shown_bound %X %Y %x %y
  }
}

# this for calls separate from the labels-binding;
# used via shortcuts binding;
# it can be set via "offer_info_labels_shown_bound"
# (that is called by mouse clicks) too;
proc mxmbatch_update_info_labels_request {change} {
  global with_labels
  global mxmbatch_info_label_verb
  global info_labels_foreground

  if {$mxmbatch_info_label_verb==""} {
    return
  }

  if {$change} {
    set with_labels [expr {$with_labels ? false : true}]
  }

  $mxmbatch_info_label_verb configure -foreground [lindex $info_labels_foreground [expr {$with_labels ? 1 : 0}]]
}
proc update_info_labels_request {change} {
  global with_labels
  global with_labels_asked
  global info_label_verb
  global in_lisp_mode
  global in_debug_mode
  global info_labels_foreground

  if {$change && ($in_lisp_mode || $in_debug_mode)} {
    return
  }

  if {$change} {
    set with_labels_asked [expr {$with_labels ? false : true}]
  }

  $info_label_verb configure -foreground [lindex $info_labels_foreground [expr {$with_labels ? 1 : 0}]]

  if {$change} {
    call_maxima_line_setting
  }

  return
}
update_info_labels_request false

bind $topw <Control-f> {
  focus $search_entry
  break
}
bind $topw <Control-F> {
  focus $search_entry
  break
}
bind $search_entry <1> {
  remove_torn_menus
  if {[is_text_field_active]} {
    focus $search_entry
  }
  if {![is_text_field_active]} {
    break
  }
}

set info_line_size_asked [list]
foreach info_line_size_item $info_line_size {
  lappend info_line_size_asked $info_line_size_item
}
proc set_maxima_line_length {value} {
  global info_line_size
  global info_line_size_asked
  global maxima_line_size_min
  global maxima_line_size_max
  global in_lisp_mode
  global in_debug_mode

  if {[is_stack_mode]} {
    if {![is_text_field_active]} {
      return false
    }
    if {$in_lisp_mode || $in_debug_mode} {
      return false
    }
  }

  set value [string trim $value]
  if {[regexp -- "^\[0-9\]+\$" $value]==0} {
    return false
  }

  set mssg ""
  set mssg_start "length of (short) output lines used:"

  set value [string trimleft $value "0"]
  if {[string length $value]<1} {
    set value 0
  }
  if {[string length $value]>6} {
    set value [expr {$maxima_line_size_max + 1}]
  }

  if {$value<$maxima_line_size_min} {
    set value $maxima_line_size_min
    set mssg "min $mssg_start $maxima_line_size_min"
  } elseif {$value>$maxima_line_size_max} {
    set value $maxima_line_size_max
    set mssg "max $mssg_start $maxima_line_size_max"
  }

  if {[is_stack_mode]} {
    lset info_line_size_asked 0 $value
    after 0 "call_maxima_line_setting"
  } {
    lset info_line_size 0 $value
  }

  if {$mssg!=""} {
    show_text_field_message_error [list [list $mssg]]
  }

  return true
}

set search_entry_control_mode false
proc search_entry_on_ctrl_click {} {
  global search_entry_control_mode
  global search_entry
  global text_search
  global info_line_size
  global in_lisp_mode
  global in_debug_mode

  if {$search_entry_control_mode} {
    set search_entry_control_mode false
    set_search_entry_style_common
    set text_search ""
    focus $search_entry
    return
  }

  if {![is_text_field_active]} {
    return false
  }
  if {$in_lisp_mode || $in_debug_mode} {
    return false
  }

  set search_entry_control_mode true
  set_search_entry_style_control
  set text_search [lindex $info_line_size 0]
  focus $search_entry
}

proc search_entry_on_return {with_ctrl with_shift with_alt} {
  global search_entry_control_mode
  global text_search

  if {$search_entry_control_mode} {
    if {$with_ctrl || $with_shift || $with_alt} {
      return
    }
    if {![set_maxima_line_length $text_search]} {
      return
    }
    set search_entry_control_mode false
    set_search_entry_style_common
    set text_search ""
    return
  }

  # leaving ctrl-based events for search in the command field,
  # that may get implemented at some later;
  if {$with_ctrl} {
    return
  }
  if {$with_shift && $with_alt} {
    return
  }
  if {$with_alt} {
    find_regexp
    return
  }
  if {$with_shift} {
    find_exact
    return
  }
  find_nocase
}

bind $search_entry <Control-1> {
  if {[is_text_field_active]} {
    search_entry_on_ctrl_click
  }
  break
}

bind $search_entry <Return> {
  if {[is_text_field_active]} {
    search_entry_on_return false false false
  }
  break
}
bind $search_entry <Shift-Return> {
  if {[is_text_field_active]} {
    search_entry_on_return false true false
  }
  break
}
bind $search_entry <Alt-Return> {
  if {[is_text_field_active]} {
    search_entry_on_return false false true
  }
  break
}
bind $search_entry <Shift-Alt-Return> {
  break
}
bind $search_entry <Control-Return> {
  break
}

proc destroy_replace {} {
  global display_replace
  global text_focus_in_real_mode
  global text_field

  set text_focus_in_real_mode true
  catch {grab release $display_replace}
  catch {destroy $display_replace}
  focus $text_field
}

set search_replace ""
set display_replace ""
proc draw_replace {start_forwards} {
  global display_replace
  global search_tag_all
  global search_replace
  global text_field
  global spec_base_font
  global text_focus_in_real_mode

  remove_torn_menus

  if {[llength [$text_field tag ranges $search_tag_all]]==0} {
    return
  }
  set_tags_shown_state false

  set curr_sel [list]
  foreach {sel_start sel_end} [$text_field tag ranges "sel"] {
    lappend curr_sel [$text_field get $sel_start $sel_end]
  }
  set curr_sel [join $curr_sel " "]
  # is something got read from "sel"
  if {$curr_sel!=""} {
    # this puts the seeked region in by default, and that is not good,
    # thus disabling it (for now);
    #set search_replace $curr_sel
  }

  $text_field tag remove "sel" "1.0" "end"
  if {$start_forwards} {
    seek_next
  } {
    seek_prev
  }

  set replace_char "R"

  catch {destroy $display_replace}

  set display_replace [toplevel ".replace" -width 200 -height 200]
  wm title $display_replace "Replace tool"
  wm minsize $display_replace 300 50
  wm resizable $display_replace false false

  set top_pos [get_top_window_position]
  set x_pos [expr {225 + [lindex $top_pos 0]}]
  set y_pos [expr {75+ [lindex $top_pos 1]}]
  wm geometry $display_replace "+$x_pos+$y_pos"

  set lframe [labelframe $display_replace.labelframe -text "Replace ($replace_char) / Seek to (<>) a found range" -bd 1 -padx 10 -pady 10]
  pack $lframe -expand yes -fill both

  entry $lframe.entry -width 30 -bd 1 -textvariable search_replace
  base_bindings_for_entry $lframe.entry "search_replace"

  if {$spec_base_font(search)==""} {
    $lframe.entry configure -font search_font
  } {
    $lframe.entry configure -font $spec_base_font(search)
  }

  pack $lframe.entry -side top -expand yes -fill x

  frame $lframe.buttons -pady 5
  pack $lframe.buttons -side top
  button $lframe.buttons.backwards -text "<" -width 4 -command "command_replace -2"
  button $lframe.buttons.replace_and_backwards -text "$replace_char <" -width 4 -command "command_replace -1"
  button $lframe.buttons.replace -text "$replace_char" -width 4 -command "command_replace 0"
  button $lframe.buttons.replace_and_forwards -text "$replace_char >" -width 4 -command "command_replace 1"
  button $lframe.buttons.forwards -text ">" -width 4 -command "command_replace 2"
  pack $lframe.buttons.backwards $lframe.buttons.replace_and_backwards \
    $lframe.buttons.replace $lframe.buttons.replace_and_forwards $lframe.buttons.forwards -side left -expand no -fill none

  bind $display_replace <Escape> {
    destroy_replace
  }

  wm attributes $display_replace -topmost true
  wm protocol $display_replace WM_DELETE_WINDOW {destroy_replace}

  catch {grab $display_replace}
  focus $lframe.entry

  set_text_tags_view_focused true
  set text_focus_in_real_mode false
}

proc command_replace {seq_idx} {
  global display_replace

  if {![is_text_field_active]} {
    return false
  }

  if {$seq_idx==-2} {
    seek_prev
    focus $display_replace
    return true
  }
  if {$seq_idx==-1} {
    make_replace
    seek_prev
    focus $display_replace
    return true
  }
  if {$seq_idx==0} {
    make_replace
    return true
  }
  if {$seq_idx==1} {
    make_replace
    seek_next
    focus $display_replace
    return true
  }
  if {$seq_idx==2} {
    seek_next
    focus $display_replace
    return true
  }

  return false
}

proc make_replace {} {
  global text_field
  global search_replace
  global search_highlights_max_id
  global search_tag_all

  set sel_region [$text_field tag ranges "sel"]
  if {[llength $sel_region]<2} {
    return
  }
  set sel_start_idx [lindex $sel_region 0]
  set sel_start [split $sel_start_idx "."]
  set sel_start_row [lindex $sel_start 0]
  set sel_start_col [lindex $sel_start 1]
  set sel_end_idx [lindex $sel_region 1]
  set sel_end [split $sel_end_idx "."]
  set sel_end_row [lindex $sel_end 0]
  set sel_end_col [lindex $sel_end 1]

  $text_field tag remove $search_tag_all "1.0" "end"
  $text_field tag remove "sel" "1.0" "end"

  set tag_id 1
  while {$tag_id<=$search_highlights_max_id} {
    set tag_name [strcat $search_tag_all $tag_id]
    set tag_locs [$text_field tag ranges $tag_name]
    incr tag_id
    if {$tag_locs==""} {
      continue
    }
    set tag_start_idx [lindex $tag_locs 0]
    set tag_start_info [split $tag_start_idx "."]
    set tag_start_row [lindex $tag_start_info 0]
    set tag_start_col [lindex $tag_start_info 1]
    set tag_end_idx [lindex $tag_locs "end"]
    set tag_end_info [split $tag_end_idx "."]
    set tag_end_row [lindex $tag_end_info 0]
    set tag_end_col [lindex $tag_end_info 1]
    # if tag starts before the sel end, and ends after the sel start,
    # it has non-empty intersection with sel, and is removed here;
    if {($tag_start_row<$sel_end_row) || (($tag_start_row==$sel_end_row)&&($tag_start_col<$sel_end_col))} {
      if {($tag_end_row>$sel_start_row) || (($tag_end_row==$sel_start_row)&&($tag_end_col>$sel_start_col))} {
        $text_field tag delete $tag_name
      }
    }
  }

  $text_field edit separator
  $text_field replace $sel_start_idx $sel_end_idx $search_replace
  $text_field edit separator

  set tag_id 1
  while {$tag_id<=$search_highlights_max_id} {
    set tag_name [strcat $search_tag_all $tag_id]
    set tag_locs [$text_field tag ranges $tag_name]
    incr tag_id
    if {$tag_locs==""} {
      continue
    }
    set tag_start_idx [lindex $tag_locs 0]
    set tag_end_idx [lindex $tag_locs "end"]
    $text_field tag add $search_tag_all $tag_start_idx $tag_end_idx
  }

  update_text_properties
}

set search_tag_all "search_res_all"

proc set_seek_position {wdg range_start range_end} {
  $wdg tag remove "sel" "1.0" "end"
  $wdg mark set "insert" $range_start
  focus $wdg
  $wdg see $range_start
  $wdg see $range_end
  $wdg tag add "sel" $range_start $range_end
  set_anchor_position $wdg $range_start
  target_within_tag_entries $range_start
  after 0 "update_text_highlight_lines"
  after 0 "update_text_bracket_pairing"
}

proc seek_next {} {
  global search_tag_all
  global text_field
  global search_highlights_max_id

  set cursor_pos [split [$text_field index "insert"] "."]
  set cursor_row [lindex $cursor_pos 0]
  set cursor_col [lindex $cursor_pos 1]

  set sel_region [$text_field tag nextrange "sel" "insert" "end"]
  if {$sel_region!=""} {
    set sel_start [split [lindex $sel_region 0] "."]
    set sel_start_row [lindex $sel_start 0]
    set sel_start_col [lindex $sel_start 1]
    set sel_end [split [lindex $sel_region 1] "."]
    set sel_end_row [lindex $sel_end 0]
    set sel_end_col [lindex $sel_end 1]
  }

  set got_first_pos false
  set use_first_pos_start ""
  set use_first_pos_end ""
  set got_next_pos false
  set use_next_pos_start ""
  set use_next_pos_end ""
  set tag_id 1
  while {$tag_id<=$search_highlights_max_id} {
    if {$got_first_pos&&$got_next_pos} {
      break
    }
    set tag_locs [$text_field tag ranges [strcat $search_tag_all $tag_id]]
    incr tag_id
    if {$tag_locs==""} {
      continue
    }
    set tag_start_idx [lindex $tag_locs 0]
    set tag_start_info [split $tag_start_idx "."]
    set tag_start_row [lindex $tag_start_info 0]
    set tag_start_col [lindex $tag_start_info 1]
    set tag_end_idx [lindex $tag_locs "end"]
    set tag_end_info [split $tag_end_idx "."]
    set tag_end_row [lindex $tag_end_info 0]
    set tag_end_col [lindex $tag_end_info 1]
    if {!$got_first_pos} {
      set use_first_pos_start $tag_start_idx
      set use_first_pos_end $tag_end_idx
      set got_first_pos true
    }
    if {!$got_next_pos} {
      # if a region (starting at cursor or later) is selected, taking the first tag after it, excl. the selection;
      # otherwise taking the first tag starting after cursor, incl. the cursor position;
      set take_the_tag false
      if {$sel_region!=""} {
        if {$tag_start_row>$sel_start_row} {
          set take_the_tag true
        } elseif {$tag_start_row==$sel_start_row} {
          if {$tag_start_col>$sel_start_col} {
            set take_the_tag true
          } elseif {$tag_start_col==$sel_start_col} {
            if {$tag_end_row<$sel_end_row} {
              set take_the_tag true
            } elseif {($tag_end_row==$sel_end_row)&&($tag_end_col<$sel_end_col)} {
              set take_the_tag true
            }
          }
        }
      } {
        if {$tag_start_row>$cursor_row} {
          set take_the_tag true
        } elseif {($tag_start_row==$cursor_row)&&($tag_start_col>=$cursor_col)} {
          set take_the_tag true
        }
      }
      if {$take_the_tag} {
        set use_next_pos_start $tag_start_idx
        set use_next_pos_end $tag_end_idx
        set got_next_pos true
      }
    }
  }

  if {$got_next_pos} {
    set_seek_position $text_field $use_next_pos_start $use_next_pos_end
    return true
  }
  if {$got_first_pos} {
    set_seek_position $text_field $use_first_pos_start $use_first_pos_end
    return true
  }

  after 0 "update_text_highlight_lines"
  after 0 "update_text_bracket_pairing"
  return false
}

proc seek_prev {} {
  global search_tag_all
  global text_field
  global search_highlights_max_id

  set use_pos [split [$text_field index "insert"] "."]
  set use_row [lindex $use_pos 0]
  set use_col [lindex $use_pos 1]

  set sel_region [$text_field tag prevrange "sel" "insert" "1.0"]
  if {$sel_region!=""} {
    set sel_start [split [lindex $sel_region 0] "."]
    set sel_start_row [lindex $sel_start 0]
    set sel_start_col [lindex $sel_start 1]
    if {($sel_start_row<$use_row) || (($sel_start_row==$use_row)&&($sel_start_col<$use_col))} {
      set use_row $sel_start_row
      set use_col $sel_start_col
    }
  }

  # taking the prev tag starting before cursor, excl. the cursor position;
  # or if a region is selected starting before the cursor, taking the prev tag
  # starting before start of the sel region (that is again excl. the start sel position);

  set got_last_pos false
  set use_last_pos_start ""
  set use_last_pos_end ""
  set got_prev_pos false
  set use_prev_pos_start ""
  set use_prev_pos_end ""
  set tag_id $search_highlights_max_id
  while {$tag_id>0} {
    if {$got_last_pos&&$got_prev_pos} {
      break
    }
    set tag_locs [$text_field tag ranges [strcat $search_tag_all $tag_id]]
    incr tag_id -1
    if {$tag_locs==""} {
      continue
    }
    set tag_start_idx [lindex $tag_locs 0]
    set tag_start_info [split $tag_start_idx "."]
    set tag_start_row [lindex $tag_start_info 0]
    set tag_start_col [lindex $tag_start_info 1]
    set tag_end_idx [lindex $tag_locs 1]
    if {!$got_last_pos} {
      set use_last_pos_start $tag_start_idx
      set use_last_pos_end $tag_end_idx
      set got_last_pos true
    }
    if {!$got_prev_pos} {
      set take_the_tag false
      if {$tag_start_row<$use_row} {
        set take_the_tag true
      } elseif {($tag_start_row==$use_row)&&($tag_start_col<$use_col)} {
        set take_the_tag true
      }
      if {$take_the_tag} {
        set use_prev_pos_start $tag_start_idx
        set use_prev_pos_end $tag_end_idx
        set got_prev_pos true
      }
    }
  }

  if {$got_prev_pos} {
    set_seek_position $text_field $use_prev_pos_start $use_prev_pos_end
    return true
  }
  if {$got_last_pos} {
    set_seek_position $text_field $use_last_pos_start $use_last_pos_end
    return true
  }

  after 0 "update_text_highlight_lines"
  after 0 "update_text_bracket_pairing"
  return false
}

set onset_unset_search_highlights_delay 1000
set onset_unset_search_highlights_id false
proc onset_unset_search_highlights {} {
  global search_tag_all
  global text_field
  global onset_unset_search_highlights_id
  global onset_unset_search_highlights_delay
  global search_highlights_max_id

  if {$onset_unset_search_highlights_id!=false} {
    catch {after cancel $onset_unset_search_highlights_id}
    set onset_unset_search_highlights_id false
  }

  if {$search_highlights_max_id==0} {
    return
  }

  set onset_unset_search_highlights_id [after $onset_unset_search_highlights_delay \
    [list "unset_search_highlights" $text_field $search_tag_all]]
}

set search_highlights_max_id 0
proc unset_search_highlights {wdg search_tag_all} {
  global search_highlights_max_id
  global onset_unset_search_highlights_id

  if {$onset_unset_search_highlights_id!=false} {
    catch {after cancel $onset_unset_search_highlights_id}
    set onset_unset_search_highlights_id false
  }

  $wdg tag remove $search_tag_all "1.0" "end"

  set tag_id 1
  while {$tag_id<=$search_highlights_max_id} {
    $wdg tag delete [strcat $search_tag_all $tag_id]
    incr tag_id
  }

  set search_highlights_max_id 0
}

proc set_search_highlights {wdg search_tag_all idx_starts idx_lens} {
  global search_highlights_max_id

  unset_search_highlights $wdg $search_tag_all
  set search_highlights_max_id [llength $idx_starts]

  set tag_id 0
  foreach one_start $idx_starts one_len $idx_lens {
    incr tag_id
    $wdg tag add $search_tag_all "$one_start" "$one_start + $one_len chars"
    set curr_tag_name [strcat $search_tag_all $tag_id]
    $wdg tag add $curr_tag_name "$one_start" "$one_start + $one_len chars"
    $wdg tag lower $curr_tag_name
  }
}

proc find_all {mode text_search} {
  global text_field
  global search_tag_all

  set search_idx_starts [list]
  set search_idx_lengths [list]

  unset_search_highlights $text_field $search_tag_all
  $text_field tag remove "sel" "1.0" "end"

  if {$text_search==""} {
    return 0
  }

  set search_idx_starts [$text_field search $mode -overlap -all -count search_idx_lengths -forwards -- $text_search "1.0"]
  if {[llength $search_idx_starts]>0} {
    set_search_highlights $text_field $search_tag_all $search_idx_starts $search_idx_lengths
  }

  return [llength $search_idx_starts]
}

proc find_nocase {} {
  global text_search
  set found_count [find_all "-nocase" $text_search]
  if {$found_count==0} {
    return
  }
  seek_next
}
proc find_exact {} {
  global text_search
  if {[catch {
    set found_count [find_all "-exact" $text_search]
  }]} {
    show_text_field_message_error [list [list "can not do the search"]]
    return
  }

  if {$found_count==0} {
    return
  }
  seek_next
}
proc find_regexp {} {
  global text_search

  if {[catch {
    set found_count [find_all "-regexp" $text_search]
  }]} {
    show_text_field_message_error [list [list "malformed regular expression"]]
    return
  }

  if {$found_count==0} {
    return
  }
  seek_next
}

proc clear_temporary_highlights {} {
  onset_unset_search_highlights
}

set dot_char "*"
catch {set dot_char [encoding convertfrom "utf-8" "\xE2\x80\xA2"]}

proc destroy_about {} {
  global going_into_displaying_tag_help
  global display_about
  global text_field

  if {$display_about==""} {
    return
  }

  catch {grab release $display_about}

  set should_try_to_focus false
  set window_to_focus ""

  if {[is_wine_like]} {
    if {!$going_into_displaying_tag_help} {
      set should_try_to_focus true
      # here we would get just the $display_about window
      #catch {set window_to_focus [focus -lastfor $display_about]}
    }
  }

  catch {destroy $display_about}
  set display_about ""

  if {!$should_try_to_focus} {
    return
  }

  if {$window_to_focus==""} {
    set window_to_focus $text_field
  }

  update
  if {[focus -displayof $window_to_focus]==""} {
    catch {focus -force $window_to_focus}
  }
}

set display_about ""
set maxima_info_version ""
set maxima_info_lisp_name ""
set maxima_info_website "http://maxima.sourceforge.net/"
set gnuplot_info_website "http://gnuplot.info/"
set going_into_displaying_tag_help false
proc draw_about {} {
  global display_about
  global maxima_pid
  global app_license
  global app_version
  global app_website
  global gnuplot_version_read
  global gnuplot_info_website
  global maxima_info_website
  global maxima_info_version
  global maxima_info_lisp_name
  global editor_batch_state
  global maxima_pid_batch
  global sigma_char
  global use_buttons_text
  global dot_char
  global going_into_displaying_tag_help

  remove_torn_menus

  catch {destroy $display_about}
  set display_about [toplevel ".about" -width 380 -height 350]
  catch {wm protocol $display_about WM_DELETE_WINDOW "destroy_about"}

  set going_into_displaying_tag_help false
  set entry_link_length [expr {max([string length $app_website], [string length $maxima_info_website], [string length $gnuplot_info_website])}]

  set display_about_part [frame $display_about.frame -bd 25]
  pack $display_about_part -side top

  pack [frame $display_about_part.min1a] -side "top"
  pack [label $display_about_part.min1a.txt1 -text "Minimalistic "] -side "left"
  pack [label $display_about_part.min1a.txt2 -text " TGM " -foreground "#000000" -background "#ffffff" -cursor question_arrow] -side "left"
  pack [label $display_about_part.min1a.txt3 -text " interface to Maxima"] -side "left"

  bind $display_about_part.min1a.txt2 <1> {
    set going_into_displaying_tag_help true
    catch {grab release $display_about}
    catch {destroy $display_about}
    after 10 [list "draw_help_view_for_tag_lines" false]
  }

  pack [label $display_about_part.min1b -text "version: $app_version, license: $app_license"] -side "top"
  pack [label $display_about_part.min1c -text " "] -side "top"

  pack [label $display_about_part.tgm1a -text [strcat $dot_char " Text editor with Tags " $dot_char]] -side "top"
  pack [label $display_about_part.tgm1b -text [strcat $dot_char " Graphics with Gnuplot " $dot_char]] -side "top"
  pack [label $display_about_part.tgm1c -text [strcat $dot_char " Mathematics with Maxima " $dot_char]] -side "top"
  pack [label $display_about_part.tgm1d -text " "] -side "top"

  if {$editor_batch_state} {
    set mxbatch_sign $sigma_char
    if {$use_buttons_text} {
      set mxbatch_sign "M"
    }
    pack [label $display_about_part.run1a -text "   Batch mode: click the $mxbatch_sign button   " -background "#ffffff"] -side "top"
  } {
    pack [label $display_about_part.run1a -text " Control/Shift+Enter for running commands " -background "#ffffff"] -side "top"
  }

  pack [entry $display_about_part.min2a -textvariable app_website -width $entry_link_length -justify center -state "readonly" -relief flat -borderwidth 0] -side "top"
  base_bindings_for_entry $display_about_part.min2a "app_website"
  pack [label $display_about_part.min2b -text " "] -side "top"

  set gnuplot_version_string ""
  set gnuplot_patchlevel_string ""
  foreach one_gpl_line $gnuplot_version_read {
    set one_gpl_line [string tolower $one_gpl_line]
    if {[string first "version" $one_gpl_line]!=-1} {
      set getting_gnuplot_version_string false
      set getting_gnuplot_patchlevel_string false
      foreach one_gpl_item [split $one_gpl_line] {
        set one_gpl_item [string trim $one_gpl_item]
        if {$one_gpl_item==""} {
          continue
        }
        if {[string equal -nocase $one_gpl_item "version"]} {
          set getting_gnuplot_version_string true
          set getting_gnuplot_patchlevel_string false
          continue
        }
        if {[string equal -nocase $one_gpl_item "patchlevel"]} {
          set getting_gnuplot_version_string false
          set getting_gnuplot_patchlevel_string true
          continue
        }
        if {$getting_gnuplot_version_string} {
          set gnuplot_version_string $one_gpl_item
        }
        if {$getting_gnuplot_patchlevel_string} {
          set gnuplot_patchlevel_string $one_gpl_item
        }
        set getting_gnuplot_version_string false
        set getting_gnuplot_patchlevel_string false
      }
    }
  }

  set gnuplot_label_text "Gnuplot"
  if {$gnuplot_version_string!=""} {
    append gnuplot_label_text " " $gnuplot_version_string
  }
  if {$gnuplot_patchlevel_string!=""} {
    append gnuplot_label_text " pl " $gnuplot_patchlevel_string
  }

  pack [label $display_about_part.gpl1a -text $gnuplot_label_text -justify "center"] -side "top"
  pack [entry $display_about_part.gpl1b -textvariable gnuplot_info_website -width $entry_link_length -justify center -state "readonly" -relief flat -borderwidth 0] -side "top"
  base_bindings_for_entry $display_about_part.gpl1b "gnuplot_info_website"
  pack [label $display_about_part.gpl1c -text " "] -side "top"

  set max_first_line "Maxima"
  if {[string length $maxima_info_version]>0} {
    append max_first_line " " $maxima_info_version
  }
  pack [label $display_about_part.max1a -text $max_first_line] -side "top"
  pack [entry $display_about_part.max1b -textvariable maxima_info_website -width $entry_link_length -justify center -state "readonly" -relief flat -borderwidth 0] -side "top"
  base_bindings_for_entry $display_about_part.max1b "maxima_info_website"

  set is_focused false
  if {([string length $maxima_info_lisp_name]>0) || ($maxima_pid != "")} {
    pack [frame $display_about_part.max2a] -side "top"
    set lisp_pid_label_part_list [list]
    if {[string length $maxima_info_lisp_name]>0} {
      lappend lisp_pid_label_part_list $maxima_info_lisp_name
    }
    set to_show_pid false
    if {((!$editor_batch_state) && ($maxima_pid != "")) || (($editor_batch_state) && ($maxima_pid_batch != ""))} {
      lappend lisp_pid_label_part_list "PID:"
      set to_show_pid true
    }
    set lisp_pid_label_part [join $lisp_pid_label_part_list ", "]
    pack [label $display_about_part.max2a.label -text $lisp_pid_label_part] -side "left"
    if {$to_show_pid} {
      set pid_entry [entry $display_about_part.max2a.entry -justify center -state "readonly" -relief flat -borderwidth 0]
      if {!$editor_batch_state} {
        $pid_entry configure -textvariable maxima_pid -width [expr {[string length $maxima_pid] + 1}]
        base_bindings_for_entry $pid_entry "maxima_pid"
      } {
        $pid_entry configure -textvariable maxima_pid_batch -width [expr {[string length $maxima_pid_batch] + 1}]
        base_bindings_for_entry $pid_entry "maxima_pid_batch"
      }
      pack $pid_entry -side "left"
      set is_focused true
      focus $pid_entry
    }
  }

  wm title $display_about "TGM Minima: About"

  wm minsize $display_about 380 350
  wm attributes $display_about -topmost true
  wm resizable $display_about false false

  set top_pos [get_top_window_position]
  set x_pos [expr {225 + [lindex $top_pos 0]}]
  set y_pos [expr {75+ [lindex $top_pos 1]}]
  wm geometry $display_about "+$x_pos+$y_pos"

  catch {grab $display_about}
  if {!$is_focused} {
    focus $display_about
  }

  bind $display_about <Escape> {
    destroy_about
  }
}

proc draw_kill_interrupt {} {
  remove_torn_menus

  if {[is_stack_mode]} {
    return
  }

  if {![has_common_ctrlc]} {
    stop_batch_maxima
    return
  }
  if {[is_batch_maxima_running] && (![is_batch_maxima_waiting])} {
    break_batch_maxima
    return
  }
  stop_batch_maxima
}

proc draw_restart {} {
  global maxima_command
  global maxima_command_inner
  global maxima_command_control
  global maxima_pid
  global has_ever_asked
  global maxima_ready
  global maxima_channel
  global state_label
  global checks_id
  global read_wait_id
  global maxima_stderr_read_id
  global maxima_stderr_write_id
  global maxima_stderr_should_ignore
  global newly_at_maxima_outputs
  global command_history
  global maxima_stderr_should_check
  global can_deal_with_disconnect
  global should_display_cpu_usage
  global could_display_cpu_usage

  remove_torn_menus

  if {[is_batch_mode]} {
    return
  }

  set is_ready $maxima_ready
  set maxima_ready false
  set newly_at_maxima_outputs false

  if {$checks_id!=""} {
    catch {after cancel checks_id}
    set checks_id ""
  }

  catch {after cancel $read_wait_id}

  set maxima_stderr_should_check false
  set can_deal_with_disconnect false

  if {$maxima_pid!=""} {
    catch {fileevent $maxima_channel readable ""}
    if {(!$is_ready) && [has_common_ctrlc]} {
      catch {do_kill_maxima false}
    }
    if {true} {
      set maxima_command ""
      set maxima_command_inner ""
      set maxima_command_control [get_maxima_quit]
      catch {fileevent $maxima_channel writable "svwrite 0"}
    }
  }

  clear_read_buffers
  clean_control_parts

  if {$has_ever_asked} {
    lappend command_history [list "\[ restarted \]"]
  }

  set_button_interrupt_sets
  set_state_disabled
  set has_ever_asked false
  set_ignore_cpu_usage true
  set should_display_cpu_usage false
  set could_display_cpu_usage false
  set_state_label_text "restarting"

  set maxima_stderr_should_ignore true
  catch {close $maxima_stderr_read_id}
  catch {close $maxima_stderr_write_id}

  if {$maxima_pid!=""} {
    after 0 setup_restart
  } {
    force_restart
  }
}

proc setup_restart {} {
  after 200 force_restart
}

set after_restart_fresh false
proc force_restart {} {
  global maxima_pid
  global maxima_info_version
  global maxima_info_lisp_name
  global maxconn
  global maxima_channel
  global after_restart
  global after_restart_fresh
  global svsocket
  global svport
  global maxima_setting_done
  #global can_kill_label
  global maxima_command
  global maxima_command_inner
  global maxima_command_control
  global maxima_channel_hnd

  catch {fileevent $maxima_channel writable ""}

  if {$svsocket!=""} {
    catch {chan close $svsocket}
    set svsocket ""
    set svport -1
  }

  if {$maxima_channel!=""} {
    if {[llength [file channels $maxima_channel]]>0} {
      catch {fileevent $maxima_channel readable ""}
      catch {fileevent $maxima_channel writable ""}
      catch {chan close $maxima_channel}
      set maxima_channel ""
    }
  }

  clear_read_buffers

  if {$maxima_pid!=""} {
    do_kill_maxima true
    after 200
  }
  if {$maxima_channel_hnd!=""} {
    clear_maxima_channel_hnd
  }

  set maxima_pid ""
  set maxconn 0
  set maxima_channel ""

  set maxima_info_version ""
  set maxima_info_lisp_name ""

  set after_restart true
  set after_restart_fresh true
  set maxima_setting_done false

  #set can_kill_label true
  set_labels_to_asked_values
  set maxima_command_control ""
  set maxima_command_inner ""
  set maxima_command ""

  setup_maxima
}

proc draw_interrupt {} {
  global maxima_pid
  global command_history
  global has_ever_asked
  #global curr_stack_mode

  remove_torn_menus

  if {![is_stack_mode]} {
    return
  }
  if {![has_common_ctrlc]} {
    return
  }
  if {$maxima_pid==""} {
    return
  }

  if {$has_ever_asked} {
    lappend command_history [list "\[ interrupted \]"]
  }

  # not sure if to deal with the curr_stack_mode here:
  # may be should keep the input stack level so that it is easier to check it;
  # though command parts that are after the part that is gonna interrupted
  # are (commonly) executed after the interrupt (since were already sent),
  # thus it is necessary to consider it that way;
  # and if there was already some output (with $curr_stack_mode>0)
  # it is moved all anyway (with Minima not knowing about any late part);
  #set curr_stack_mode 0
  do_kill_maxima false
}

set recently_opened_files [list]

proc update_recent_index {} {
  global recently_opened_files
  global save_path
  global text_field

  if {$save_path==""} {
    return
  }
  if {[llength $recently_opened_files]==0} {
    return
  }

  set newest_recent [lindex [lindex $recently_opened_files "end"] 0]
  if {$newest_recent!=$save_path} {
    return
  }

  lset recently_opened_files "end" [list $newest_recent [$text_field index "insert"]]
}

proc update_recent_files {save_path save_index} {
  global recently_opened_files
  global last_dir

  set prev_used [lsearch -exact -index 0 $recently_opened_files $save_path]
  if {$prev_used>-1} {
    set recently_opened_files [lreplace $recently_opened_files $prev_used $prev_used]
  }
  lappend recently_opened_files [list $save_path $save_index]

  set last_dir [file dirname $save_path]
}

proc save_into_file {} {
  global save_path

  set is_saved [save_into_file_inner]
  if {!$is_saved} {
    set save_path ""
  }
  set_state_for_save
  return $is_saved
}

proc save_into_file_inner {} {
  global save_path
  global text_field
  global topw
  global command_stack_saving_last_text
  global command_stack_saving_last_path
  global command_stack_saving_paused
  global command_stack_saving_index
  global command_stack_index
  global command_stack
  global text_inner_code_blocks
  global min_tilde_count
  global omit_auto_tildes
  global backup_on_save_tilde

  set channel -1

  if {$backup_on_save_tilde} {
    if {[file exists $save_path]} {
      if {[catch {
        file rename -force -- $save_path [strcat $save_path "~"]
      }]} {
        show_text_field_message_error [list [list "can not make the backup file"]]
        return false
      }
    }
  }

  set bad [catch {set channel [open $save_path "w" "0664"]}]
  if {$bad} {
    show_text_field_message_error [list [list "can not open file for writing"]]
    return false
  }

  if {$command_stack_saving_paused && ($command_stack_saving_last_path==$save_path)} {
    set command_stack_saving_last_text ""
  }

  set command_too false
  if {($command_stack_saving_index!=-1) && (!$command_stack_saving_paused)} {
    if {[is_command_field_active] && ($command_stack_saving_index==$command_stack_index)} {
      save_current_stack
    }
    set command_too true
    set command_stack_saving_last_path $save_path
    set command_stack_saving_last_text [lindex $command_stack $command_stack_saving_index]
  }

  if {$command_too && (!$omit_auto_tildes)} {
    # to know whether to put ~-line above the saved stack level
    seek_inner_code
  }

  set intron_string ""
  if {$command_too} {
    set intron_string "\n/* command */"
    if {!$omit_auto_tildes} {
      if {[lindex [lindex $text_inner_code_blocks "end"] 1]!="end"} {
        set block_delimiter [string repeat "~" $min_tilde_count]
        set intron_string "\n$block_delimiter$intron_string"
      }
    }
  }

  set bad [catch {
    puts -nonewline $channel [$text_field get 1.0 "end - 1 chars"]
    if {$command_too} {
      puts $channel $intron_string
      puts $channel $command_stack_saving_last_text
    }
  }]
  if {$bad} {
    set command_stack_saving_last_path ""
    set command_stack_saving_last_text ""
    show_text_field_message_error [list [list "can not write into file"]]
    catch {chan close $channel}
    app_labels_on_modified
    return false
  }

  catch {chan close $channel}

  $text_field edit separator
  $text_field edit modified false
  set file_name [file tail $save_path]

  update_recent_files $save_path [$text_field index "insert"]

  app_labels_on_modified
  set_reopen_button_actual_state

  return true
}

proc draw_save {} {
  global save_path

  remove_torn_menus

  if {$save_path==""} {
    draw_save_as
    return
  }

  return [save_into_file]
}

set common_file_types [list [list "Text-with-tags files" [list $minima_file_extension]] [list "Text files" [list ".txt"]]]
if {!$default_ft_minima} {
  set common_file_types [lreverse $common_file_types]
}
lappend common_file_types [list "All files" [list "*"]]

proc draw_save_as {} {
  global last_dir
  global save_path
  global spec_file_dialogs
  global common_file_types
  global tags_decls_list
  global tags_decls_state_clean
  global minima_file_extension

  remove_torn_menus

  set file_types_save $common_file_types

  set initial_dir $last_dir
  set initial_file ""

  if {$save_path!=""} {
    set initial_dir [file dirname $save_path]
    set initial_file [file tail $save_path]
  } {
    # if already having saved the file under a name,
    # here it may be about just backup-like saving,
    # with an altered name, though not with a completely different name;
    # thus only trying to use the tag-based name if no name by now at all;
    set sug_name_base [get_tag_suggested_name_base]
    if {$sug_name_base!=""} {
      set initial_file [strcat $sug_name_base $minima_file_extension]
    }
  }

  set save_path_tmp ""
  if {$initial_file != ""} {
    set use_common_save_dialog true
    if {$spec_file_dialogs(save) != ""} {
      set use_common_save_dialog false
      set bad [catch {
        set save_path_tmp [$spec_file_dialogs(save) -details true -initialdir $initial_dir -confirmoverwrite true \
          -initialfile $initial_file -filetypes $file_types_save -defaultextension $minima_file_extension]
      }]
      if {$bad} {
        set use_common_save_dialog true
        set save_path_tmp ""
      }
    }
    if {$use_common_save_dialog} {
      set save_path_tmp [tk_getSaveFile -initialdir $initial_dir -confirmoverwrite true \
        -initialfile $initial_file -filetypes $file_types_save -defaultextension $minima_file_extension]
    }
  } {
    set use_common_save_dialog true
    if {$spec_file_dialogs(save) != ""} {
      set use_common_save_dialog false
      set bad [catch {
        set save_path_tmp [$spec_file_dialogs(save) -details true -initialdir $initial_dir -confirmoverwrite true \
          -filetypes $file_types_save -defaultextension $minima_file_extension]
      }]
      if {$bad} {
        set use_common_save_dialog true
        set save_path_tmp ""
      }
    }
    if {$use_common_save_dialog} {
      set save_path_tmp [tk_getSaveFile -initialdir $initial_dir -confirmoverwrite true \
        -filetypes $file_types_save -defaultextension $minima_file_extension]
    }
  }

  if {$save_path_tmp==""} {
    return
  }
  update_recent_index
  set save_path $save_path_tmp

  return [save_into_file]
}

proc close_import {} {
  global import_window

  catch {wm withdraw $import_window}
}

proc set_import_title {} {
  global import_window
  global import_input_warning

  if {$import_window==""} {
    return
  }

  set title_string "TGM Minima: Data import"
  if {$import_input_warning} {
    append title_string " *"
  }

  wm title $import_window $title_string
}

proc draw_import {} {
  global import_window
  global import_input_field

  remove_torn_menus

  if {$import_window==""} {
    setup_import
  }
  update_import_warning_label

  catch {wm manage $import_window}

  wm protocol $import_window WM_DELETE_WINDOW {close_import}
  set_import_title
  wm deiconify $import_window
  wm minsize $import_window 600 300
  wm resizable $import_window true true

  raise $import_window
  focus $import_input_field
}

proc get_html_export_suggested_name {} {
  set sug_name [get_tag_suggested_name_base]
  if {$sug_name!=""} {
    return [strcat $sug_name ".html"]
  }
  return ""
}

proc get_tag_suggested_name_base {} {
  global tag_type_document_ids
  global tags_decls_list
  global tags_decls_state_clean
  global tag_border_size

  if {!$tags_decls_state_clean} {
    set tags_decls_list [extract_tag_list false]
    set tags_decls_state_clean true
  }

  set doc_id_list_use [list]
  foreach one_type_doc $tag_type_document_ids {
    set one_type_doc_part "$one_type_doc:"
    lappend doc_id_list_use $one_type_doc_part [string length $one_type_doc_part]
  }

  foreach one_tag_info $tags_decls_list {
    set one_tag_string_inner [string range [lindex $one_tag_info 0] "$tag_border_size+1" "end-$tag_border_size"]
    foreach {tag_doc_id_part tag_doc_id_part_len} $doc_id_list_use {
      if {[string equal -nocase -length $tag_doc_id_part_len $one_tag_string_inner $tag_doc_id_part]} {
        set tag_id_test [string range $one_tag_string_inner $tag_doc_id_part_len "end"]
        if {$tag_id_test==""} {
          continue
        }
        set tag_based_id [tag_suggested_urlize_document_spec $tag_id_test]
        if {$tag_based_id==""} {
          continue
        }
        return $tag_based_id
      }
    }
  }

  return ""
}

set last_dir_html ""
set save_path_html ""
proc draw_export_html {{to_show false}} {
  global last_dir
  global last_dir_html
  global save_path
  global save_path_html
  global spec_file_dialogs

  set use_last_dir $last_dir_html
  if {$use_last_dir==""} {
    set use_last_dir $last_dir
  }

  if {$save_path_html!=""} {
    if {[file dirname $save_path_html]!=$use_last_dir} {
      set save_path_html ""
    }
  }
  if {$save_path_html==""} {
    if {$save_path!=""} {
      set save_path_html [strcat [file rootname $save_path] ".html"]
      #if {![file exists $save_path_html]} {
      #  set save_path_html ""
      #}
    }
  }

  set initial_dir $use_last_dir
  if {$save_path_html!=""} {
    set initial_dir [file dirname $save_path_html]
  }

  set initial_file [get_html_export_suggested_name]
  if {$initial_file==""} {
    if {$save_path_html!=""} {
      set initial_file [file tail $save_path_html]
    }
  }

  set exp_label "Export content to HTML"
  set save_path_tmp ""
  if {$initial_file!=""} {
    set use_common_save_dialog true
    if {$spec_file_dialogs(save) != ""} {
      set use_common_save_dialog false
      set bad [catch {
        set save_path_tmp [$spec_file_dialogs(save) -details true -initialdir $initial_dir -confirmoverwrite true \
          -initialfile $initial_file -defaultextension ".html" -title $exp_label]
      }]
      if {$bad} {
        set use_common_save_dialog true
        set save_path_tmp ""
      }
    }
    if {$use_common_save_dialog} {
      set save_path_tmp [tk_getSaveFile -initialdir $initial_dir -confirmoverwrite true \
        -initialfile $initial_file -defaultextension ".html" -title $exp_label]
    }
  } {
    set use_common_save_dialog true
    if {$spec_file_dialogs(save) != ""} {
      set use_common_save_dialog false
      set bad [catch {
        set save_path_tmp [$spec_file_dialogs(save) -details true -initialdir $initial_dir -confirmoverwrite true \
          -defaultextension ".html" -title $exp_label]
      }]
      if {$bad} {
        set use_common_save_dialog true
        set save_path_tmp ""
      }
    }
    if {$use_common_save_dialog} {
      set save_path_tmp [tk_getSaveFile -initialdir $initial_dir -confirmoverwrite true \
        -defaultextension ".html" -title $exp_label]
    }
  }

  if {$save_path_tmp==""} {
    return
  }
  set save_path_html $save_path_tmp

  set is_saved [do_text_export_html false $save_path_html]
  if {!$is_saved} {
    set save_path_html ""
    return
  }

  set last_dir_tmp [file dirname $save_path_html]
  if {$last_dir_tmp==""} {
    set last_dir_html ""
  } {
    if {$last_dir_tmp==$last_dir} {
      set last_dir_html ""
    } {
      set last_dir_html $last_dir_tmp
    }
  }

  if {$to_show && ($save_path_html != "")} {
    open_html_path $save_path_html
  }
}

set export_modes_menu ""
proc draw_export_as {} {
  global text_field
  global topw
  global export_modes_menu
  global bt_file_export

  if {[$text_field index "end -1 chars"]=="1.0"} {
    return
  }

  remove_torn_menus

  # currently just a single export mode, thus without the menu now;
  draw_export_html
  return

  set abs_pos_button [get_menu_button_position $bt_file_export]
  set abs_pos_field [get_text_field_position]
  set use_x [expr {0 + [lindex $abs_pos_button 0]}]
  set use_y [expr {0 + [lindex $abs_pos_field 1]}]

  catch {destroy $export_modes_menu}
  set export_modes_menu [menu "$topw.export_modes_menu" -tearoff false]
  $export_modes_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $export_modes_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $export_modes_menu configure -activeborderwidth 1

  $export_modes_menu add command -label "Content export modes:" \
    -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
  $export_modes_menu add command -label "export into html file" \
    -command [list "draw_export_html"]

  tk_popup $export_modes_menu $use_x $use_y
}

proc load_from_file {open_path {open_index ""}} {
  global text_field

  set res [load_from_file_inner $open_path $open_index]
  set_state_for_save
  clear_temporary_highlights

  if {$res} {
    $text_field mark set "insert" "1.0"
    $text_field see "insert"

    if {$open_index != ""} {
      $text_field mark set "insert" $open_index
      $text_field see "insert"
    } {
      if {[lindex [split [$text_field index "end -1 chars"] "."] 0]>1} {
        if {[$text_field search -regexp -- [get_sec_regexp] "1.0" "2.0"]=="1.0"} {
          $text_field mark set "insert" "2.0"
        }
      }
    }
  }

  return $res
}

proc load_from_file_inner {open_path open_index} {
  global text_field
  global save_path
  global topw
  global common_insert_mark_name
  global mxmbatch_text_insert_mark_name
  global maxima_batch_asked_with_selections

  if {![is_text_field_active]} {
    return false
  }

  set channel -1

  # if can not read it, not dealing with the path at all
  set bad [catch {set channel [open $open_path "r"]}]
  if {$bad} {
    show_text_field_message_error [list [list "can not open file for reading"]]
    app_labels_on_modified
    return false
  }

  set save_path ""

  $text_field delete 1.0 end
  $text_field mark unset $common_insert_mark_name
  $text_field mark unset $mxmbatch_text_insert_mark_name
  $text_field edit reset
  $text_field edit modified false
  set maxima_batch_asked_with_selections [list]

  clear_tag_info

  set bad [catch {$text_field insert end [read $channel]}]
  if {$bad} {
    show_text_field_message_error [list [list "can not read from file"]]
    catch {chan close $channel}
    app_labels_on_modified
    return false
  }

  catch {chan close $channel}

  $text_field edit reset
  $text_field edit separator
  $text_field edit modified false
  $text_field see "end"
  update_text_properties

  set file_name [file tail $open_path]
  app_labels_on_modified

  set save_path $open_path
  update_recent_files $save_path $open_index

  update_text_whole_tag

  app_labels_on_modified
  set_reopen_button_actual_state

  return true
}

proc draw_open {{initial_path ""}} {
  global save_path
  global text_field
  global last_dir
  global spec_file_dialogs
  global common_file_types

  remove_torn_menus

  if {![is_text_field_active]} {
    return
  }

  set file_types_open $common_file_types

  set to_load "ok"
  set message_ask "Discard changes?"
  set message_detail "Session not saved"
  if {$save_path!=""} {
    set message_detail "Session not resaved"
  }
  if {[$text_field edit modified]} {
    set to_load [tk_messageBox -message $message_ask -icon question -type okcancel -default cancel -detail $message_detail]
  }

  if {$to_load=="cancel"} {
    return
  }
  update_recent_index

  set open_path ""
  if {$initial_path==""} {
    set use_common_open_dialog true
    if {$spec_file_dialogs(open) != ""} {
      set use_common_open_dialog false
      set bad [catch {
        set open_path [$spec_file_dialogs(open) -details true -initialdir $last_dir -multiple false -filetypes $file_types_open]
      }]
      if {$bad} {
        set use_common_open_dialog true
        set open_path ""
      }
    }
    if {$use_common_open_dialog} {
      set open_path [tk_getOpenFile -initialdir $last_dir -multiple false -filetypes $file_types_open]
    }
  } {
    set use_common_open_dialog true
    if {$spec_file_dialogs(open) != ""} {
      set use_common_open_dialog false
      set bad [catch {
        set open_path [$spec_file_dialogs(open) -details true -initialdir [file dirname $initial_path] -initialfile [file tail $initial_path] \
          -multiple false -filetypes $file_types_open]
      }]
      if {$bad} {
        set use_common_open_dialog true
        set open_path ""
      }
    }
    if {$use_common_open_dialog} {
      set open_path [tk_getOpenFile -initialfile $initial_path -multiple false -filetypes $file_types_open]
    }
  }

  if {$open_path==""} {
    return
  }

  load_from_file $open_path
}

proc draw_open_recent {} {
  global recently_opened_files
  global save_path
  global text_field

  remove_torn_menus

  if {![is_text_field_active]} {
    return
  }

  set recent_count [llength $recently_opened_files]

  if {$recent_count==0} {
    draw_open
    return
  }
  #if {$recent_count==1} {
  #  draw_open [lindex $recently_opened_files 0]
  #  return
  #}

  # ask if not saved
  set to_load "ok"
  set message_ask "Discard changes?"
  set message_detail "Session not saved"
  if {$save_path!=""} {
    set message_detail "Session not resaved"
  }
  if {[$text_field edit modified]} {
    set to_load [tk_messageBox -message $message_ask -icon question -type okcancel -default cancel -detail $message_detail]
  }

  if {$to_load=="cancel"} {
    return
  }

  show_menu_recent_files
}

proc load_from_recent_file {level} {
  global recently_opened_files

  update_recent_index
  set use_recent_info [lindex $recently_opened_files $level]

  catch {load_from_file [lindex $use_recent_info 0] [lindex $use_recent_info 1]}
}

set recently_open_menu ""
set recently_open_show_max [get_max_menu_entry_count]
proc show_menu_recent_files {} {
  global recently_opened_files
  global recently_open_show_max
  global recently_open_menu
  global display_recent
  global topw
  global save_path
  global menu_selected_background

  set abs_pos [get_text_field_position]
  set pos_x [expr {5 + [lindex $abs_pos 0]}]
  set pos_y [expr {0 + [lindex $abs_pos 1]}]

  set recently_opened_files_length [llength $recently_opened_files]

  catch {destroy $recently_open_menu}
  set recently_open_menu [menu "$topw.recentlyopenmenu" -tearoff false]
  $recently_open_menu configure -relief solid -borderwidth 1 -background "#f0f0f0" -disabledforeground "#000000"
  $recently_open_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $recently_open_menu configure -activeborderwidth 1

  set label_text "Recently open files"
  $recently_open_menu add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  set curr_idx 0
  while {($curr_idx < $recently_open_show_max) && ($curr_idx < $recently_opened_files_length)} {
    incr curr_idx
    set curr_level [expr {$recently_opened_files_length - $curr_idx}]
    set show_path [lindex [lindex $recently_opened_files $curr_level] 0]
    if {[string length $show_path] > 55} {
      set show_path_tmp [string range $show_path "0" "15"]
      append show_path_tmp "..."
      append show_path_tmp [string range $show_path "end-35" "end"]
      set show_path $show_path_tmp
    }
    $recently_open_menu add command -label $show_path -command [list "load_from_recent_file" $curr_level]
    $recently_open_menu entryconfigure $curr_idx -background "#ffffff"
  }
  if {$save_path!=""} {
    $recently_open_menu entryconfigure 1 -background $menu_selected_background -activebackground "#e0e0ff"
  }

  tk_popup $recently_open_menu $pos_x $pos_y

  return
}

proc draw_blank {} {
  global save_path
  global text_field
  global topw
  global common_insert_mark_name
  global mxmbatch_text_insert_mark_name
  global maxima_batch_asked_with_selections

  if {![is_text_field_active]} {
    return
  }

  remove_torn_menus

  set to_pass "ok"
  set message_ask "Discard changes?"
  set message_detail "Session not saved"
  if {$save_path!=""} {
    set message_detail "Session not resaved"
  }
  if {[$text_field edit modified]} {
    set to_pass [tk_messageBox -message $message_ask -icon question -type okcancel -default cancel -detail $message_detail]
  }

  if {$to_pass=="cancel"} {
    return
  }
  update_recent_index

  set save_path ""
  app_labels_on_modified

  $text_field delete 1.0 end
  $text_field mark unset $common_insert_mark_name
  $text_field mark unset $mxmbatch_text_insert_mark_name
  $text_field edit reset
  $text_field edit modified false
  set maxima_batch_asked_with_selections [list]

  clear_tag_info
  clear_temporary_highlights
  set_state_for_save

  return true
}

set tag_stripes "line_stripes"
set tag_numbers "line_numbers"

proc prepare_command_aux {} {
  global command_field_aux
  global tag_stripes
  global tag_numbers
  $command_field_aux configure -state "normal"
  $command_field_aux delete "1.0" "end"
  $command_field_aux insert "1.0" "1\n"
  $command_field_aux tag configure $tag_numbers -justify "right"
  $command_field_aux configure -state "disabled"
}

proc adjust_command_aux_count_pos {} {
  global command_field_aux
  global command_field
  global tag_stripes
  global tag_numbers

  set last_line [lindex [split [$command_field index "end - 1 chars"] "."] 0]
  set last_line_aux [lindex [split [$command_field_aux index "end - 1 chars"] "."] 0]

  $command_field_aux configure -state "normal"

  set ins_line [lindex [split [$command_field index "insert"] "."] 0]
  set ins_line_plus $ins_line
  incr ins_line_plus

  if {$last_line == $last_line_aux} {
    $command_field_aux tag remove $tag_stripes "1.0" "end"
    $command_field_aux tag add $tag_stripes "$ins_line.0" "$ins_line_plus.0"
    $command_field_aux configure -state "disabled"
    return
  }

  if {$last_line > $last_line_aux} {
    set ins_lines ""
    set aux_idx $last_line_aux
    while {$aux_idx<$last_line} {
      incr aux_idx
      append ins_lines "\n" [expr {$aux_idx%1000}]
    }
    $command_field_aux insert "end" $ins_lines
  } {
    set del_from $last_line
    incr del_from
    $command_field_aux delete "$del_from.0" "end"
  }

  $command_field_aux tag remove $tag_numbers "1.0" "end"
  $command_field_aux tag add $tag_numbers "1.0" "end"
  $command_field_aux tag remove $tag_stripes "1.0" "end"
  $command_field_aux tag add $tag_stripes "$ins_line.0" "$ins_line_plus.0"
  $command_field_aux configure -state "disabled"
  update
}

set select_bg_col "#c3c3c3"
set select_bg_col_alt "#e0e0e0"
set select_fg_col "#000000"

set command_field_holder [frame $fcmmd.holder]
set command_field_aux [text $fcmmd.holder.command_aux -height 8 -exportselection true -wrap none -takefocus 0 -width 3 -state disabled]
set command_field [text $fcmmd.holder.command -height 8 -undo true -exportselection true -maxundo 1000 -wrap none]
pack $command_field -side right -expand yes -fill both
pack $command_field_aux -side left -expand no -fill both
if {!$show_cmmd_highlight_line_nums} {
  pack forget $command_field_aux
}
$command_field configure -selectbackground $select_bg_col -selectforeground $select_fg_col
$command_field_aux configure -selectbackground $select_bg_col -selectforeground $select_fg_col
$command_field configure -inactiveselectbackground $select_bg_col
$command_field_aux configure -inactiveselectbackground $select_bg_col

prepare_command_aux

proc adjust_command_windows {adj_form {param1 ""} {param2 ""}} {
  global command_field
  global command_field_aux

  if {$param2==""} {
    $command_field yview $adj_form $param1
  } {
    $command_field yview $adj_form $param1 $param2
  }

  adjust_command_aux_count_pos

  if {$param2==""} {
    $command_field_aux yview $adj_form $param1
  } {
    $command_field_aux yview $adj_form $param1 $param2
  }
}

proc adjust_command_scroll_aux {param1 param2} {
  global command_field_aux
  global fcmmd_scroll_vert

  $fcmmd_scroll_vert set $param1 $param2

  adjust_command_aux_count_pos
  $command_field_aux yview moveto $param1
}

set fcmmd_scroll_vert [ttk::scrollbar $fcmmd.s_v -orient vertical -command "adjust_command_windows"]
ttk::scrollbar $fcmmd.s_h -orient horizontal -command "$command_field xview"

$command_field configure -yscrollcommand "adjust_command_scroll_aux"
$command_field configure -xscrollcommand "$fcmmd.s_h set"

set size_grip ""

if {[is_stack_mode]} {
  set size_grip [ttk::sizegrip $fcmmd.size_grip]

  grid $command_field_holder $fcmmd.s_v -sticky nsew
  grid $fcmmd.s_h $size_grip -sticky nsew
  grid rowconfigure $fcmmd $command_field_holder -weight 1
  grid columnconfigure $fcmmd $command_field_holder -weight 1
}

set text_field [text $ftext.content -undo true -exportselection true -maxundo 1000 -wrap char]
ttk::scrollbar $ftext.s_v -orient vertical -command "$text_field yview"
$text_field configure -yscrollcommand "$ftext.s_v set"
$text_field configure -selectbackground $select_bg_col -selectforeground $select_fg_col
$text_field configure -inactiveselectbackground $select_bg_col

set tags_holder [frame $ftext.tags_holder]

set tags_decls_show_default_text "declared tags"
set tags_decls_show $tags_decls_show_default_text
set tags_decls_line [entry $tags_holder.tags_decls_line -textvariable tags_decls_show -state "readonly" \
  -exportselection true -takefocus 1 -justify "left" -relief "sunken" -borderwidth 0 \
  -selectforeground $select_fg_col -selectbackground $select_bg_col]

set tags_links_show_default_text "links to tags"
set tags_links_show $tags_links_show_default_text
set tags_links_line [entry $tags_holder.tags_links_line -textvariable tags_links_show -state "readonly" \
  -exportselection true -takefocus 1 -justify "left" -relief "sunken" -borderwidth 0 \
  -selectforeground $select_fg_col -selectbackground $select_bg_col]

grid $tags_decls_line $tags_links_line -sticky nsew
grid rowconfigure $tags_holder "all" -uniform true
grid columnconfigure $tags_holder "all" -uniform true
grid rowconfigure $tags_holder "all" -weight 1
grid columnconfigure $tags_holder "all" -weight 1

if {[is_batch_mode]} {
  set size_grip [ttk::sizegrip $ftext.size_grip]
}

grid $text_field $ftext.s_v -sticky nsew
if {[is_batch_mode]} {
  grid $tags_holder $size_grip -sticky nsew
}
grid rowconfigure $ftext $text_field -weight 1
grid columnconfigure $ftext $text_field -weight 1

font create text_font
foreach {font_option_name font_option_value} [font actual TkFixedFont] {
  font configure text_font $font_option_name $font_option_value
}
font configure text_font -size 11

if {$spec_base_font(text)==""} {
  $text_field configure -font text_font
} {
  $text_field configure -font $spec_base_font(text)
}

font create command_font
foreach {font_option_name font_option_value} [font actual TkFixedFont] {
  font configure command_font $font_option_name $font_option_value
}
font configure command_font -size -14 -weight bold

if {$spec_base_font(command)==""} {
  $command_field configure -font command_font
  $command_field_aux configure -font command_font
  $tags_decls_line configure -font command_font
  $tags_links_line configure -font command_font
} {
  $command_field configure -font $spec_base_font(command)
  $command_field_aux configure -font $spec_base_font(command)
  $tags_decls_line configure -font $spec_base_font(command)
  $tags_links_line configure -font $spec_base_font(command)
}

if {[is_wine_like]} {
  focus -force $topw
}

proc entry_on_ctrl_a {wdg} {
  $wdg selection clear
  $wdg selection range "0" "end"
}

proc entry_on_ctrl_c {wdg text_var_name} {
  global $text_var_name

  if {![$wdg selection present]} {
    return
  }

  set text_var [set $text_var_name]
  set sel_start [$wdg index sel.first]
  set sel_end [$wdg index sel.last]
  incr sel_end -1
  set_clipboard_text [string range $text_var $sel_start $sel_end]
  store_clipboard_onset
}

proc entry_on_ctrl_v {wdg text_var_name} {
  global $text_var_name

  if {[string tolower [$wdg cget -state]] != "normal"} {
    return
  }

  set string_to_put_in [get_clipboard_text]
  set text_var [set $text_var_name]
  set pos_idx [$wdg index insert]

  if {[$wdg selection present]} {
    set sel_start [$wdg index sel.first]
    set sel_end [$wdg index sel.last]
    incr sel_end -1
    $wdg selection clear
    set $text_var_name [string replace $text_var $sel_start $sel_end $string_to_put_in]
    if {$sel_start != $pos_idx} {
      $wdg icursor [expr {$sel_start + [string length $string_to_put_in]}]
    }
    return
  }

  set result_string [string range $text_var 0 [expr {$pos_idx - 1}]]
  append result_string $string_to_put_in
  append result_string [string range $text_var $pos_idx "end"]
  set $text_var_name $result_string
  $wdg icursor [expr {$pos_idx + [string length $string_to_put_in]}]
}

proc entry_on_ctrl_x {wdg text_var_name} {
  global $text_var_name

  if {![$wdg selection present]} {
    return
  }

  set text_var [set $text_var_name]
  set sel_start [$wdg index sel.first]
  set sel_end [$wdg index sel.last]
  incr sel_end -1
  set_clipboard_text [string range $text_var $sel_start $sel_end]
  store_clipboard_onset

  if {[string tolower [$wdg cget -state]] != "normal"} {
    return
  }

  $wdg selection clear
  set $text_var_name [string replace $text_var $sel_start $sel_end]
}

proc base_bindings_for_entry {wdg text_var_name} {
  set proc_on_ctrl_a [list entry_on_ctrl_a $wdg]
  set proc_on_ctrl_c [list entry_on_ctrl_c $wdg $text_var_name]
  set proc_on_ctrl_v [list entry_on_ctrl_v $wdg $text_var_name]
  set proc_on_ctrl_x [list entry_on_ctrl_x $wdg $text_var_name]

  foreach keyspressed [list "Control-slash" "Control-backslash" "Control-d" "Alt-d" "Control-k" "Control-t" "Control-h" \
    "Control-w" "Alt-w" "Control-y" "Control-z" "Control-Z" "Control-e" "Control-E"] {
    bind $wdg <$keyspressed> {
      break
    }
  }
  bind $wdg <Control-a> "$proc_on_ctrl_a; break"
  bind $wdg <Control-c> "$proc_on_ctrl_c; break"
  bind $wdg <Control-v> "$proc_on_ctrl_v; break"
  bind $wdg <Control-x> "$proc_on_ctrl_x; break"
}

base_bindings_for_entry $search_entry "text_search"
if {[is_batch_mode]} {
  base_bindings_for_entry $tags_decls_line "tags_decls_show"
  base_bindings_for_entry $tags_links_line "tags_links_show"
}

bind $tags_decls_line <Control-f> {
  focus $search_entry
  break
}
bind $tags_decls_line <Control-F> {
  if {[$tags_decls_line selection present]} {
    entry_on_ctrl_c $tags_decls_line "tags_decls_show"
    $search_entry selection clear
    set text_search [get_clipboard_text]
    focus $search_entry
  } {
    focus $search_entry
  }
  break
}
bind $tags_links_line <Control-f> {
  focus $search_entry
  break
}
bind $tags_links_line <Control-F> {
  if {[$tags_links_line selection present]} {
    entry_on_ctrl_c $tags_links_line "tags_links_show"
    $search_entry selection clear
    set text_search [get_clipboard_text]
    focus $search_entry
  } {
    focus $search_entry
  }
  break
}

set tags_text_help_view ""
set display_help_view_for_tag_lines ""
set display_help_view_for_tag_lines_grabbed false
set display_help_view_for_tag_lines_grab_id ""
set display_help_view_for_tag_lines_grab_rnk 0
proc draw_help_view_for_tag_lines {{from_bars true}} {
  global display_help_view_for_tag_lines_grabbed
  global display_help_view_for_tag_lines_grab_id
  global display_help_view_for_tag_lines_grab_rnk
  global display_help_view_for_tag_lines
  global select_bg_col
  global select_fg_col
  global tags_text_help_view

  remove_torn_menus

  catch {destroy $display_help_view_for_tag_lines}
  set display_help_view_for_tag_lines [toplevel ".display_help_view_for_tag_lines" -width 600 -height 400]

  set ftext [frame $display_help_view_for_tag_lines.ftext]
  pack $ftext -side top -expand yes -fill both

  set tags_text_help_view [text $ftext.tags_text_help_view -wrap none -width 100 -height 30]
  ttk::scrollbar $ftext.vscr -orient vertical -command "$ftext.tags_text_help_view yview"
  ttk::scrollbar $ftext.hscr -orient horizontal -command "$ftext.tags_text_help_view xview"
  $ftext.tags_text_help_view configure -xscroll "$ftext.hscr set" -yscroll "$ftext.vscr set"
  ttk::sizegrip $ftext.size_grip
  $ftext.tags_text_help_view configure -selectbackground $select_bg_col -selectforeground $select_fg_col -inactiveselectbackground $select_bg_col

  grid $ftext.tags_text_help_view $ftext.vscr -sticky nsew
  grid $ftext.hscr $ftext.size_grip -sticky nsew

  grid rowconfigure $ftext $ftext.tags_text_help_view -weight 1
  grid columnconfigure $ftext $ftext.tags_text_help_view -weight 1

  set hlp [list]
  lappend hlp "Organizing the TGM Minima documents can be underwent in two ways:"
  lappend hlp ""
  lappend hlp "1) by section lines (for text parts) and tilde lines (for computation blocks),"
  lappend hlp "   where the sections/blocks are partitioned in a schematic way;"
  lappend hlp "2) semantics is added by regular tags and links to them, where the tagging"
  lappend hlp "   is for marking particular parts of text according to its meaning;"
  lappend hlp ""
  lappend hlp "Tag forms: declared tags, links to local tags, links to remote tags"
  lappend hlp "* declared tags: \[#tag_type:tag value\]"
  lappend hlp "tag types: simple word-like, allowed chars: a-zA-Z0-9_"
  lappend hlp "tag values: a more-free form, with next restrictions:"
  lappend hlp "            without any nesting of (links to) tags, without \\ \{ \} letters,"
  lappend hlp "            and \[\] () \"\" only allowed when simply (without nesting) pair-matched."
  lappend hlp "* local links: \[@tag_type:(part of) tag value\]"
  lappend hlp "* remote links: \[@...dir_part.../...glob_part...\\tag_type:(part of) tag value\]"
  lappend hlp ""
  lappend hlp ""
  lappend hlp ""
  lappend hlp "Use of tags:"
  lappend hlp ""
  lappend hlp "Section lines (*/=/#-lines) are mainly for schematic delimiting of parts of documents,"
  lappend hlp "and tilde-lines alike schematically delimit parts with code and computation outputs."
  lappend hlp "Tags add contextual labeling into the TGM Minima documents."
  lappend hlp ""
  lappend hlp "The 'doc(ument)' and 'top(ic)' types are used as helpers for document saving and exporting."
  lappend hlp "Then the 'sec' type is supposed to serve for (sub)sections, like:"
  lappend hlp "\[#sec:/part/subpart/subsubpart\] etc."
  lappend hlp ""
  lappend hlp "Links to tags can be both within documents:"
  lappend hlp "using @ instead of # (and without the directory/glob part put in),"
  lappend hlp "and between documents:"
  lappend hlp "with 'dir/glob\\' part added in between link start (again by @) and the linked tag;"
  lappend hlp "here dir can be both relative and absolute ('~/subdir' can be used for starting the dir"
  lappend hlp "from home directory), and glob can contain the '*' wildcard character."
  lappend hlp "If using drive spec in the dir part, it has to be preceded by slash, like: /Z:/dir"
  lappend hlp ""
  lappend hlp "When ':' char is put at the dir part start, it tries several upward (three by default)"
  lappend hlp "directories when looking for the files (with tags); the testing count can be specified."
  lappend hlp "E.g. \[@:2/name_start\\sec:some\] looks for files with names starting by \"name_start\""
  lappend hlp "in the current and two upwards directories (checking \[#sec:some*\] tag presence there)."
  lappend hlp ""
  lappend hlp "The ':' or ':number' specifier can be used by the end of the glob part (before \\) too."
  lappend hlp "There it is for testing files in subdirectories (':' is for three subdirs by default)."
  lappend hlp "In such a case the first subdirectories (along the files there) are filtered by the glob"
  lappend hlp "(if the glob itself is specified)."
  lappend hlp ""
  lappend hlp "Matching is done case-insensitive for tag names (filesystem part in remote links"
  lappend hlp "depends on the ways of operating systems). And comparisons among links and tags"
  lappend hlp "is (by default) limited to the extents put into the links: \[@sec:some\] links to"
  lappend hlp "\[#sec:some\], \[#sec:some section\], etc. This can be changed by adding '=' to link"
  lappend hlp "ends (where the added '=' is only used for declaring the non-limited comparisons),"
  lappend hlp "and similarly a trailing '*' is considered to explicitly state the limited"
  lappend hlp "(that is default) way of comparisons."
  lappend hlp ""
  lappend hlp "The '*' wildcard can be put at the tag type end too (that is just before the \":\" char)."
  lappend hlp "Thus e.g. both \[@*:*\] and \[@*:\] links look for all tags within the loaded document."
  lappend hlp ""
  lappend hlp "The '*' wildcard is put automatically to the end of glob part, thus e.g. \[@./\\sec:some\]"
  lappend hlp "looks for \[#sec:some...\] at all TGM Minima documents in the directory where the current"
  lappend hlp "document is located. If '=' is put to the glob end, the '*' is not added there."
  lappend hlp ""
  lappend hlp "Up to two tag parts can be put in links, separated by \\ or \\\\: requiring tags of both"
  lappend hlp "specs to be present in the searched document, and tags of the second tag spec are shown."
  lappend hlp "The \\\\ inter-tag connective allows tags of the two specs being wherever in documents,"
  lappend hlp "while the \\ inter-tag connective requires occurrences within the same (sub)sections;"
  lappend hlp "(sub)section nesting goes as \"# > = > *\" and by section-line lengths then;"
  lappend hlp ""
  lappend hlp "Thus e.g. the \[@tag:some=\\*:*\] link looks for all tags at all document sections"
  lappend hlp "(delimited by section lines) where the \[#tag:some\] tag is located."
  lappend hlp "Recall that sections for tag pairs are defined by the section/tilde lines,"
  lappend hlp "that is not by possibly present tags with \"sec\" type."
  lappend hlp ""
  lappend hlp "Notice that Control-\[ Control-\] are for lists of tags, links to tags, respectively,"
  lappend hlp "Control-@ provides list of possible (local) links to tags present in the document."
  lappend hlp ""
  lappend hlp "Highlighting is by Control-\{ Control-\} for links to tags, tags, respectively,"
  lappend hlp "Control-# is for highlighting all the links/tags in the document."
  lappend hlp ""
  lappend hlp "Following a link to its target is done by Control-Enter and Control-Shift-Enter"
  lappend hlp "if it is not bound for command executing."
  lappend hlp "Control-Double-click and Control-Shift-Double-click (when at text) can be used too."
  lappend hlp "Use of Shift at local links is for starting the search upwards."
  lappend hlp "Regarding remote links, the case without/with Shift looks (by default) at files"
  lappend hlp "with '.twt' suffix, with '.twt' and '.txt' suffixes, respectively."
  # have been using double-click, not !, thus rather saving Control-! for other actions;
  #lappend hlp "Control-! (Control-Alt-!) are provided as alternative keys for tag-link following,"
  #lappend hlp "e.g. for situations when Enter-based actions are set for command executions."
  lappend hlp ""
  lappend hlp ""
  lappend hlp ""

  set bar_hlp [list]
  lappend bar_hlp "Actions on the tag bar (at the bottom of the main window) when it is focused,"
  lappend bar_hlp "for declared tags (links to tags) when at left-hand (right-hand)"
  lappend bar_hlp "part of the bottom bar:"
  lappend bar_hlp ""
  lappend bar_hlp "Forwards and leftwards arrows are used for sliding over tags just within the bar,"
  lappend bar_hlp "that is without any actions within the document."
  lappend bar_hlp "Tab (Shift-Tab) slides over tags forwards (backwards) along with highlighting"
  lappend bar_hlp "the tags at document."
  lappend bar_hlp "Control (Shift) Tab are for lists of tags, alike Control-\[ Control-\] at document."
  lappend bar_hlp ""
  lappend bar_hlp "'Enter' locates and highlights the current (link to) tag at the document,"
  lappend bar_hlp "'Shift-Enter' locates tags in a complementary way: from document location to the bar."
  lappend bar_hlp "'Control-Enter' (and similarly 'Control-Shift-Enter') locates tags at document,"
  lappend bar_hlp "with focusing the text (and putting the cursor at the located link/tag)."
  lappend bar_hlp "'Alt-Enter' highlights all (links to) tags in the document; Control-g/G can slide"
  lappend bar_hlp "over the highlighted (links to) tags then."
  lappend bar_hlp ""
  lappend bar_hlp "Notice that passing between the main window and the bar parts can be done by"
  lappend bar_hlp "Control-comma, Control-period (alike at other parts of the TGM Minima interface)."
  lappend bar_hlp ""
  lappend bar_hlp ""
  lappend bar_hlp ""

  if {[is_batch_mode] && $from_bars} {
    $ftext.tags_text_help_view insert "end" [join $bar_hlp "\n"]
    $ftext.tags_text_help_view insert "end" "\n"
  }

  $ftext.tags_text_help_view insert "end" [join $hlp "\n"]
  $ftext.tags_text_help_view insert "end" "\n"

  if {[is_batch_mode] && (!$from_bars)} {
    $ftext.tags_text_help_view insert "end" [join $bar_hlp "\n"]
    $ftext.tags_text_help_view insert "end" "\n"
  }

  $ftext.tags_text_help_view insert "end" "\n"

  $ftext.tags_text_help_view configure -state disabled
  focus $display_help_view_for_tag_lines

  bind $display_help_view_for_tag_lines <Control-a> {
    $tags_text_help_view tag remove "sel" "1.0" "end"
    $tags_text_help_view tag add "sel" "1.0" "end -1 chars"
    set_anchor_position $tags_text_help_view "1.0"
  }
  bind $tags_text_help_view <Control-a> {
    $tags_text_help_view tag remove "sel" "1.0" "end"
    $tags_text_help_view tag add "sel" "1.0" "end -1 chars"
    set_anchor_position $tags_text_help_view "1.0"
  }
  bind $display_help_view_for_tag_lines <Control-c> {
    set_clipboard $tags_text_help_view
    store_clipboard_onset
  }
  bind $tags_text_help_view <Control-c> {
    set_clipboard $tags_text_help_view
    store_clipboard_onset
  }

  bind $display_help_view_for_tag_lines <Return> {
    focus $tags_text_help_view
  }
  bind $tags_text_help_view <1> {
    focus $tags_text_help_view
  }

  wm title $display_help_view_for_tag_lines "TGM Minima: Tags"

  wm minsize $display_help_view_for_tag_lines 600 400
  wm attributes $display_help_view_for_tag_lines -topmost true
  wm resizable $display_help_view_for_tag_lines true true

  set top_pos [get_top_window_position]
  set x_pos [expr {100 + [lindex $top_pos 0]}]
  set y_pos [expr {50+ [lindex $top_pos 1]}]
  wm geometry $display_help_view_for_tag_lines "+$x_pos+$y_pos"

  if {$display_help_view_for_tag_lines_grab_id != ""} {
    catch {after cancel $display_help_view_for_tag_lines_grab_id}
    set display_help_view_for_tag_lines_grab_id ""
  }

  set display_help_view_for_tag_lines_grab_rnk 0
  set display_help_view_for_tag_lines_grabbed false
  if {[catch {
    grab $display_help_view_for_tag_lines
    set display_help_view_for_tag_lines_grabbed true
  }]} {
    set display_help_view_for_tag_lines_grab_id [after 10 try_to_grab_for_tag_help_view]
  }

  focus $display_help_view_for_tag_lines

  bind $display_help_view_for_tag_lines <Escape> {
    catch {grab release $display_help_view_for_tag_lines}
    catch {destroy $display_help_view_for_tag_lines}
  }

  catch {focus $tags_text_help_view}
  return
}

proc try_to_grab_for_tag_help_view {} {
  global display_help_view_for_tag_lines_grabbed
  global display_help_view_for_tag_lines_grab_id
  global display_help_view_for_tag_lines_grab_rnk
  global display_help_view_for_tag_lines

  if {$display_help_view_for_tag_lines_grab_id != ""} {
    catch {after cancel $display_help_view_for_tag_lines_grab_id}
    set display_help_view_for_tag_lines_grab_id ""
  }

  if {$display_help_view_for_tag_lines_grabbed} {
    return
  }

  if {[catch {
    grab $display_help_view_for_tag_lines
    set display_help_view_for_tag_lines_grabbed true
  }]} {
    if {$display_help_view_for_tag_lines_grab_rnk < 10} {
      set display_help_view_for_tag_lines_grab_id [after 50 try_to_grab_for_tag_help_view]
    }
  }
}

bind $tags_decls_line <3> {
  after 10 draw_help_view_for_tag_lines true
  break
}
bind $tags_links_line <3> {
  after 10 draw_help_view_for_tag_lines true
  break
}

set text_input_tags_name "input_start_tags"
set text_output_tags_name "output_start_tags"
set text_lisp_tags_name "lisp_start_tags"
$text_field tag configure $text_input_tags_name -background "#f0eed3"
$text_field tag configure $text_output_tags_name -background "#f0ecc0"
$text_field tag configure $text_lisp_tags_name -background "#e0d0e0"

catch {
  $text_field tag configure $text_input_tags_name -selectbackground $select_bg_col -selectforeground "#606060"
  $text_field tag configure $text_output_tags_name -selectbackground $select_bg_col -selectforeground "#606060"
  $text_field tag configure $text_lisp_tags_name -selectbackground $select_bg_col -selectforeground "#606060"
}

foreach wdg [list $topw $text_field $command_field $tags_decls_line $tags_links_line] {
  bind $wdg <Alt-T> {
    set_app_label_path_part
    break
  }
}

# removing some (yet unused) default text bindings that look strange to me
# control-d/D currently getting used for dealing with data import;
# control-"<>" should go for (un)indenting, control/alt-"+-=" for font sizes;
# control-(semi)colon for showing history/stack, '?' for help/about;
foreach keyspressed [list "Alt-d"] {
  bind $text_field <$keyspressed> {
    break
  }
}
foreach keyspressed [list "Alt-d"] {
  bind $command_field <$keyspressed> {
    break
  }
}


# Home key should act reasonably,
# here the basic (without Shift, etc. modifiers) only;
proc home_key_reasonable {wdg} {
  $wdg tag remove "sel" "1.0" "end"

  set insert_pos [$wdg index "insert"]
  set insert_info [split $insert_pos "."]
  set insert_row [lindex $insert_info 0]
  set insert_col [lindex $insert_info 1]

  set row_end_col [lindex [split [$wdg index "$insert_row.end"] "."] 1]

  set col_home_idx -1
  while {$col_home_idx < $row_end_col} {
    incr col_home_idx
    if {[$wdg get -displaychars "--" "$insert_row.$col_home_idx"] != " "} {
      break
    }
  }

  if {$insert_col != $col_home_idx} {
    $wdg mark set "insert" "$insert_row.$col_home_idx"
  } {
    $wdg mark set "insert" "$insert_row.0"
  }

  set_anchor_insert $wdg
}

bind $text_field <KeyPress-Home> {
  home_key_reasonable $text_field
  update_text_bracket_pairing
  break
}
bind $text_field <Control-KeyPress-Home> {
  update_text_highlight_lines_onset
}
bind $text_field <Shift-KeyPress-Home> {
  update_text_highlight_lines_onset
}
bind $text_field <Alt-KeyPress-Home> {
  update_text_highlight_lines_onset
}

bind $command_field <KeyPress-Home> {
  home_key_reasonable $command_field
  update_cmmd_bracket_pairing
  break
}
bind $command_field <Control-KeyPress-Home> {
  after 0 "update_cmmd_highlight_lines"
  after 0 "update_cmmd_bracket_pairing"
}
bind $command_field <Shift-KeyPress-Home> {
  after 0 "update_cmmd_highlight_lines"
  after 0 "update_cmmd_bracket_pairing"
}
bind $command_field <Alt-KeyPress-Home> {
  after 0 "update_cmmd_highlight_lines"
  after 0 "update_cmmd_bracket_pairing"
}

proc insert_tag_empty_declared {} {
  global text_field
  global tags_decls_state_clean
  global tags_links_state_clean

  $text_field edit separator
  $text_field insert "insert" "\[#:\]"
  $text_field edit separator

  set tags_decls_state_clean false
  set tags_links_state_clean false

  update_text_properties
  update_text_whole_tag
}
proc insert_tag_empty_link {remote} {
  global text_field
  global tags_decls_state_clean
  global tags_links_state_clean

  $text_field edit separator
  $text_field insert "insert" [expr {$remote ? "\[@./\\:\]" : "\[@:\]"}]
  $text_field edit separator

  set tags_decls_state_clean false
  set tags_links_state_clean false

  update_text_properties
  update_text_whole_tag
}
proc offer_tag_list_on_key {} {
  global text_field

  set ins_pos [$text_field index "insert"]
  set pos_list [$text_field bbox "insert"]
  if {[llength $pos_list]==0} {
    set pos_list [list 0 0]
  }

  offer_tag_insertion [lindex $pos_list 0] [lindex $pos_list 1] $ins_pos
}

foreach wdg [list $topw $text_field $command_field $search_entry $tags_decls_line $tags_links_line] {

  bind $wdg <Control-braceleft> {
    if {[is_text_field_active]} {
      show_all_tags true
    }
    break
  }
  bind $wdg <Control-Alt-braceleft> {
    break
  }

  bind $wdg <Control-braceright> {
    if {[is_text_field_active]} {
      show_all_tags false
    }
    break
  }
  bind $wdg <Control-Alt-braceright> {
    break
  }

  bind $wdg <Control-asterisk> {
    # this looked strange, and Control-P/N is available for section lines now;
    #if {[is_text_field_active]} {
    #  show_all_sections
    #}
    break
  }
  bind $wdg <Control-Alt-asterisk> {
    break
  }

  bind $wdg <Control-numbersign> {
    if {[is_text_field_active]} {
      show_all_tags true true
    }
    break
  }
  bind $wdg <Control-Alt-numbersign> {
    if {[is_text_field_active]} {
      insert_tag_empty_link true
      clear_temporary_highlights
    }
    break
  }

  bind $wdg <Control-at> {
    if {[is_text_field_active]} {
      offer_tag_list_on_key
    }
    break
  }

}

foreach wdg [list $topw $text_field $command_field] {
  bind $wdg <Control-question> {
    draw_about
    break
  }
  bind $wdg <Control-quoteright> {
    draw_history
    break
  }
  bind $wdg <Control-quotedbl> {
    if {[is_stack_mode]} {
      draw_stack
    }
    break
  }
}

proc topwise_page_up {wdg} {
  global text_inner_codes_start_tag
  global hide_tag_tildes

  if {[$wdg cget -wrap] != "none"} {
    if {[llength [$wdg bbox "1.0"]]>0} {
      $wdg mark set "insert" "1.0"
      $wdg see "insert"
      return true
    }
    if {[llength [$wdg bbox "2.0"]]>0} {
      if {$hide_tag_tildes && ([$wdg tag nextrange $text_inner_codes_start_tag "1.0" "1.end"]!="")} {
        $wdg mark set "insert" "1.0"
        $wdg see "insert"
        return true
      }
    }
  } {
    set first_idx [$wdg index "@0,0"]
    set first_line [lindex [split $first_idx "."] 0]
      if {$first_line==1} {
      $wdg mark set "insert" $first_idx
      $wdg see "insert"
      return true
    }
  }

  return false
}

proc bottomwise_page_down {wdg} {

  if {[$wdg cget -wrap] != "none"} {
    if {[llength [$wdg bbox "end - 1 chars"]]>0} {
      $wdg mark set "insert" "end"
      $wdg see "insert"
      return true
    }
  } {
    set y_part [lindex [$wdg yview] 1]
    if {$y_part>0.999999} {
      set insert_col [lindex [split [$wdg index "insert"] "."] 1]
      set last_line [lindex [split [$wdg index "end - 1 chars"] "."] 0]
      $wdg mark set "insert" "$last_line.$insert_col"
      $wdg see "insert"
      return true
    }
  }

  return false
}

bind $text_field <Prior> {
  topwise_page_up $text_field
  set_anchor_insert $text_field
  update_text_highlight_lines_onset
}
bind $text_field <Shift-Prior> {
}
bind $text_field <Next> {
  bottomwise_page_down $text_field
  set_anchor_insert $text_field
  update_text_highlight_lines_onset
}
bind $text_field <Shift-Next> {
}
bind $command_field <Prior> {
  topwise_page_up $command_field
  set_anchor_insert $command_field
  update_cmmd_highlight_lines_onset
}
bind $command_field <Shift-Prior> {
}
bind $command_field <Next> {
  bottomwise_page_down $command_field
  set_anchor_insert $command_field
  update_cmmd_highlight_lines_onset
}
bind $command_field <Shift-Next> {
}

foreach wdg [list $topw $text_field $command_field $search_entry $tags_decls_line $tags_links_line] {
  bind $wdg <Control-q> {
    want_to_exit
    break
  }
  bind $wdg <Control-s> {
    draw_save
    break
  }
  bind $wdg <Control-S> {
    draw_save_as
    break
  }
  bind $wdg <Control-Alt-s> {
    draw_save_as
    break
  }
}

foreach wdg [list $topw $text_field $command_field $search_entry $tags_decls_line $tags_links_line] {
  bind $wdg <Control-m> {
    if {!$editor_batch_state} {
      set_stack_level_saving
    } {
      draw_batch_maxima
    }
    break
  }
  bind $wdg <Control-M> {
    if {!$editor_batch_state} {
      unset_stack_level_saving
    } {
      draw_batch_maxima
    }
    break
  }
  bind $command_field <Alt-m> {
    update_pause_stack_level_saving
    break
  }
}

foreach wdg [list $topw $text_field $command_field $search_entry $tags_decls_line $tags_links_line] {
  bind $wdg <Control-d> {
    draw_import
    break
  }
  bind $wdg <Control-D> {
    break
  }
  bind $wdg <Control-bar> {
    if {[is_batch_mode]} {
      draw_batch_maxima
    }
    break
  }
}

foreach wdg [list $topw $search_entry $tags_decls_line $tags_links_line] {
  bind $wdg <Control-r> {
    open_plot_part
    break
  }
}

$text_field tag configure $search_tag_all -background "#e0ff80"
catch {
  $text_field tag configure $search_tag_all -selectbackground "#a0e0a0" -selectforeground "#000000"
}

set keysyms_moves [list "Home" "End" "Left" "Right" "Up" "Down" "Prior" "Next"]

set delayed_text_update_id ""
set delayed_text_update_lag 500
proc update_text_properties_delayed {} {
  global delayed_text_update_id
  global delayed_text_update_lag
  global text_field_needs_update_var
  global tags_decls_state_clean
  global tags_links_state_clean

  set tags_decls_state_clean false
  set tags_links_state_clean false

  catch {after cancel $delayed_text_update_id}

  set use_delay_lag $delayed_text_update_lag
  if {$text_field_needs_update_var} {
    set use_delay_lag 0
  }

  if {$use_delay_lag!=0} {
    # the current highlights look strange if not updated online;
    after 0 update_text_highlight_lines_onset
  }
  # dealing with more involved stuff with delays
  set delayed_text_update_id [after $use_delay_lag "update_text_properties"]
}

proc update_text_properties {} {
  global text_field_needs_update_var
  set text_field_needs_update_var false
  app_labels_on_modified
  after 0 "update_text_highlight"
}

proc update_command_properties {} {
  global command_stack_saving_index

  save_current_stack

  after 0 "update_cmmd_states_onset"
}

proc update_text_whole_tag {} {
  global text_field
  global tag_text_all
  $text_field tag remove $tag_text_all "1.0" "end"
  $text_field tag add $tag_text_all "1.0" "end"
}

proc update_text_highlight {{propagate true}} {
  global text_field
  global text_input_tags_name
  global text_output_tags_name
  global text_lisp_tags_name
  global highlight_text_blocks_mode

  $text_field tag remove $text_input_tags_name "1.0" "end"
  $text_field tag remove $text_output_tags_name "1.0" "end"
  $text_field tag remove $text_lisp_tags_name "1.0" "end"

  # tags of the input/output prompts are put below tags of inner computation blocks
  # when filled-block mode is set on;
  # we have to seek these tags in those situations too, since input/output prompt
  # would not be highlighted even outside the inner blocks otherwise;
  if {true} {

    set input_tag_lengths [list]
    set input_tag_starts [$text_field search -all -count input_tag_lengths -regexp -- "\\(\\%i\[\\d\]\{1,20\}\\)" "1.0" "end"]

    set output_tag_lengths [list]
    set output_tag_starts [$text_field search -all -count output_tag_lengths -regexp -- "\\(\\%o\[\\d\]\{1,20\}\\)" "1.0" "end"]

    set lisp_tag_lengths [list]
    set lisp_tag_starts [$text_field search -all -count lisp_tag_lengths -regexp -- "MAXIMA\[\\>\]+" "1.0" "end"]

    set max_offset 10
    foreach one_tag_start $input_tag_starts one_tag_length $input_tag_lengths {
      if {[lindex [split $one_tag_start "."] 1]<= $max_offset} {
        $text_field tag add $text_input_tags_name "$one_tag_start" "$one_tag_start + $one_tag_length chars"
      }
    }

    foreach one_tag_start $output_tag_starts one_tag_length $output_tag_lengths {
      if {[lindex [split $one_tag_start "."] 1]<= $max_offset} {
        $text_field tag add $text_output_tags_name "$one_tag_start" "$one_tag_start + $one_tag_length chars"
      }
    }

    foreach one_tag_start $lisp_tag_starts one_tag_length $lisp_tag_lengths {
      if {[lindex [split $one_tag_start "."] 1]<= $max_offset} {
        $text_field tag add $text_lisp_tags_name "$one_tag_start" "$one_tag_start + $one_tag_length chars"
      }
    }

  }

  update_text_whole_tag

  if {!$propagate} {
    return
  }

  # need to have positions of inner computation blocks before line highlighting
  seek_inner_code_onset
  update_text_highlight_lines_onset
}

proc is_text_field_active {} {
  global text_field
  global editor_batch_state

  if {$editor_batch_state} {
    return true
  }

  if {[string equal -nocase [$text_field cget -state] "normal"]} {
    return true
  }
  return false
}

proc is_command_field_active {} {
  global command_field
  global editor_batch_state

  if {$editor_batch_state} {
    return false
  }

  if {[string equal -nocase [$command_field cget -state] "normal"]} {
    return true
  }
  return false
}

proc can_do_interrupt {} {
  global editor_batch_state
  global bt_interrupt

  if {$editor_batch_state} {
    return false
  }
  if {[string tolower [$bt_interrupt cget -state]]=="normal"} {
    return true
  }
  return false
}
proc can_do_restart {} {
  global editor_batch_state
  global bt_restart

  if {$editor_batch_state} {
    return false
  }
  if {[string tolower [$bt_restart cget -state]]=="normal"} {
    return true
  }
  return false
}

foreach wdg [list $topw $text_field $command_field $search_entry $tags_decls_line $tags_links_line] {
  bind $wdg <Control-i> {
    if {[has_common_ctrlc]} {
      if {![is_batch_mode]} {
        if {[can_do_interrupt]} {
          draw_interrupt
        }
      } {
        if {[is_batch_maxima_running] && (![is_batch_maxima_waiting])} {
          break_batch_maxima
        } {
          stop_batch_maxima
        }
      }
    } {
      if {![is_batch_mode]} {
        if {[can_do_restart]} {
          draw_restart
        }
      } {
        stop_batch_maxima
      }
    }
    break
  }
  bind $wdg <Control-I> {
    if {![is_batch_mode]} {
      if {[can_do_restart]} {
        draw_restart
      }
    } {
      stop_batch_maxima
    }
    break
  }
  bind $wdg <Control-Alt-i> {
    close_gnuplot_direct
    break
  }
}

foreach wdg [list $topw $text_field $command_field $search_entry $tags_decls_line $tags_links_line] {
  bind $wdg <Control-o> {
    if {[is_text_field_active]} {
      draw_open
    }
    break
  }
  foreach keyspressed [list "Control-Alt-o" "Control-O"] {
    bind $wdg <$keyspressed> {
      if {[is_text_field_active]} {
        draw_open_recent
      }
      break
    }
  }
}

proc setup_search {wdg} {
  global search_entry
  global text_search

  set test_sel [$wdg tag nextrange "sel" "insert" "insert + 1 chars"]
  if {($test_sel=="") || ([llength $test_sel]!=2)} {
    set test_sel [$wdg tag prevrange "sel" "insert" "insert - 4000 chars"]
  }

  if {($test_sel!="") || ([llength $test_sel]==2)} {
    set selected_part [$wdg get [lindex $test_sel 0] [lindex $test_sel 1]]
    set selected_part [string range $selected_part 0 120]
    if {[string length $selected_part]>0} {
      set text_search $selected_part
    }
  }

  focus $search_entry
}

bind $text_field <Control-f> {
  focus $search_entry
  break
}
bind $text_field <Control-F> {
  setup_search $text_field
  break
}
bind $command_field <Control-f> {
  focus $search_entry
  break
}
bind $command_field <Control-F> {
  setup_search $command_field
  break
}

foreach wdg [list $topw $text_field $command_field $search_entry $tags_decls_line $tags_links_line] {
  bind $wdg <Control-g> {
    if {[is_text_field_active]} {
      seek_next
    }
    break
  }
  bind $wdg <Control-G> {
    if {[is_text_field_active]} {
      seek_prev
    }
    break
  }
  bind $wdg <Control-h> {
    if {[is_text_field_active]} {
      draw_replace true
    }
    break
  }
  bind $wdg <Control-H> {
    if {[is_text_field_active]} {
      draw_replace false
    }
    break
  }
}

proc get_prev_section_line {pos} {
  global text_field

  set reg_parts [list [strcat "(?:" [get_sec_regexp] ")"]]
  set use_reg [join $reg_parts "|"]

  set use_pos [$text_field index $pos]

  set sought_pos [$text_field search -backwards -regexp -- $use_reg "$use_pos linestart"]
  if {$sought_pos==""} {
    return ""
  }

  return [lindex [split $sought_pos "."] 0]
}
proc get_next_section_line {pos} {
  global text_field

  set reg_parts [list [strcat "(?:" [get_sec_regexp] ")"]]
  set use_reg [join $reg_parts "|"]

  set use_pos [$text_field index $pos]

  set sought_pos [$text_field search -forwards -regexp -- $use_reg "$use_pos +1 chars"]
  if {$sought_pos==""} {
    return ""
  }

  return [lindex [split $sought_pos "."] 0]
}

proc get_prev_block_start {pos} {
  global text_field
  global text_inner_code_blocks

  seek_inner_code

  set cur_pos_info [split [$text_field index $pos] "."]
  set cur_pos_row [lindex $cur_pos_info 0]
  set cur_pos_col [lindex $cur_pos_info 1]

  set last_block ""
  set prev_block ""

  foreach one_block [lreverse $text_inner_code_blocks] {
    set block_start_row [lindex $one_block 0]
    if {$last_block==""} {
      set test_row $block_start_row
      set is_elided [expr {[lindex [$text_field count -displaychars "$test_row.0" "$test_row.1"] 0] == 0}]
      if {$is_elided} {
        incr test_row
      }
      set last_block $test_row
    }
    if {$block_start_row < $cur_pos_row} {
      set test_row $block_start_row
      set is_elided [expr {[lindex [$text_field count -displaychars "$test_row.0" "$test_row.1"] 0] == 0}]
      if {$is_elided} {
        incr test_row
      }
      if {$test_row < $cur_pos_row} {
        set prev_block $test_row
        break
      }
    }
  }

  set sought_row $prev_block
  if {$sought_row==""} {
    set sought_row $last_block
  }
  if {$sought_row==""} {
    return ""
  }

  return $sought_row
}
proc get_next_block_start {pos} {
  global text_field
  global text_inner_code_blocks

  seek_inner_code

  set cur_pos_info [split [$text_field index "insert"] "."]
  set cur_pos_row [lindex $cur_pos_info 0]
  set cur_pos_col [lindex $cur_pos_info 1]

  seek_inner_code

  set first_block ""
  set next_block ""

  foreach one_block $text_inner_code_blocks {
    set block_start_row [lindex $one_block 0]
    if {$first_block==""} {
      set first_block $block_start_row
    }
    if {$block_start_row > $cur_pos_row} {
      set next_block $block_start_row
      break
    }
  }

  set sought_row $next_block
  if {$sought_row==""} {
    set sought_row $first_block
  }
  if {$sought_row==""} {
    return ""
  }
  set is_elided [expr {[lindex [$text_field count -displaychars "$sought_row.0" "$sought_row.1"] 0] == 0}]
  if {$is_elided} {
    incr sought_row
  }

  return $sought_row
}

proc slide_to_prev_section_line {} {
  global text_field

  set cur_pos [$text_field index "insert"]
  set sought_row [get_prev_section_line $cur_pos]

  if {$sought_row==""} {
    return false
  }

  $text_field mark set "insert" "$sought_row.0"
  $text_field see "insert"
  update_text_highlight_lines
  update_text_bracket_pairing

  return true
}
proc slide_to_next_section_line {} {
  global text_field

  set cur_pos [$text_field index "insert"]
  set sought_row [get_next_section_line $cur_pos]

  if {$sought_row==""} {
    return false
  }

  $text_field mark set "insert" "$sought_row.0"
  $text_field see "insert"
  update_text_highlight_lines
  update_text_bracket_pairing

  return true
}

proc slide_to_prev_block_start {} {
  global text_field

  set cur_pos [$text_field index "insert"]
  set sought_row [get_prev_block_start $cur_pos]

  if {$sought_row==""} {
    return false
  }

  $text_field mark set "insert" "$sought_row.0"
  $text_field see "insert"
  update_text_highlight_lines
  update_text_bracket_pairing

  return true
}

proc slide_to_next_block_start {} {
  global text_field

  set cur_pos [$text_field index "insert"]
  set sought_row [get_next_block_start $cur_pos]

  if {$sought_row==""} {
    return false
  }

  $text_field mark set "insert" "$sought_row.0"
  $text_field see "insert"
  update_text_highlight_lines
  update_text_bracket_pairing

  return true
}

proc slide_to_prev_block_start_or_section_line {} {
  global text_field

  set cur_pos [$text_field index "insert"]
  set cur_row [lindex [split $cur_pos "."] 0]

  set sought_section [get_prev_section_line $cur_pos]
  set sought_block [get_prev_block_start $cur_pos]

  if {($sought_section=="") && ($sought_block=="")} {
    return false
  }

  if {$sought_section==""} {
    set sought_row $sought_block
  } elseif {$sought_block==""} {
    set sought_row $sought_section
  } {
    if {($sought_block < $cur_row) && ($sought_section < $cur_row)} {
      set sought_row [expr {max($sought_block,$sought_section)}]
    } elseif {$sought_block < $cur_row} {
      set sought_row $sought_block
    } elseif {$sought_section < $cur_row} {
      set sought_row $sought_section
    } {
      set sought_row [expr {max($sought_block,$sought_section)}]
    }
  }

  $text_field mark set "insert" "$sought_row.0"
  $text_field see "insert"
  update_text_highlight_lines
  update_text_bracket_pairing

  return true
}
proc slide_to_next_block_start_or_section_line {} {
  global text_field

  set cur_pos [$text_field index "insert"]
  set cur_row [lindex [split $cur_pos "."] 0]

  set sought_section [get_next_section_line $cur_pos]
  set sought_block [get_next_block_start $cur_pos]

  if {($sought_section=="") && ($sought_block=="")} {
    return false
  }

  if {$sought_section==""} {
    set sought_row $sought_block
  } elseif {$sought_block==""} {
    set sought_row $sought_section
  } {
    if {($sought_block > $cur_row) && ($sought_section > $cur_row)} {
      set sought_row [expr {min($sought_block,$sought_section)}]
    } elseif {$sought_block > $cur_row} {
      set sought_row $sought_block
    } elseif {$sought_section > $cur_row} {
      set sought_row $sought_section
    } {
      set sought_row [expr {min($sought_block,$sought_section)}]
    }
  }

  $text_field mark set "insert" "$sought_row.0"
  $text_field see "insert"
  update_text_highlight_lines
  update_text_bracket_pairing

  return true
}

foreach wdg [list $topw $text_field $command_field $search_entry $tags_decls_line $tags_links_line] {
  bind $wdg <Control-e> {
    if {[is_text_field_active]} {
      draw_blank
    }
    break
  }
  bind $wdg <Control-E> {
    if {[is_text_field_active]} {
      draw_blank
    }
    break
  }

  bind $wdg <Control-P> {
    if {[is_text_field_active]} {
      slide_to_prev_section_line
    }
    break
  }
  bind $wdg <Control-N> {
    if {[is_text_field_active]} {
      slide_to_next_section_line
    }
    break
  }

  bind $wdg <Control-Alt-p> {
    if {[is_text_field_active]} {
      slide_to_prev_block_start
    }
    break
  }
  bind $wdg <Control-Alt-n> {
    if {[is_text_field_active]} {
      slide_to_next_block_start
    }
    break
  }
}

bind $info_label_autotag <Control-1> {
  remove_torn_menus
  set_auto_tagging
  break
}
bind $info_label_output <Control-1> {
  remove_torn_menus
  set_auto_tagging
  break
}
bind $info_label_hidetag <Control-1> {
  remove_torn_menus
  if {[is_text_field_active]} {
    set_tilde_hiding true
  }
  break
}

bind $info_label_adjust <Shift-1> {
  remove_torn_menus
  change_info_display_adjust_request true false
  break
}
bind $info_label_adjust <Control-1> {
  remove_torn_menus
  change_info_display_adjust_request false true
  break
}
bind $info_label_adjust <Control-Shift-1> {
  remove_torn_menus
  change_info_display_adjust_request true true
  break
}

bind $info_label_time <Control-1> {
  remove_torn_menus
  change_info_show_time_request
  break
}
bind $info_label_input <Control-1> {
  remove_torn_menus
  change_info_show_time_request
  break
}
bind $info_label_endings <Control-1> {
  remove_torn_menus
  if {[is_batch_mode] || [is_stack_mode_common]} {
    change_info_make_lineendings_request
  }
  break
}
bind $info_label_stopping <Control-1> {
  remove_torn_menus
  change_info_batch_stopping_request
  break
}

foreach wdg [list $topw $text_field $command_field $search_entry $tags_decls_line $tags_links_line] {
  bind $wdg <Control-asciitilde> {
    set_auto_tagging
    break
  }
  bind $wdg <Alt-a> {
    set_auto_tagging
    break
  }
  bind $wdg <Control-quoteleft> {
    if {[is_text_field_active]} {
      set_tilde_hiding true
    }
    break
  }
  bind $wdg <Alt-l> {
    if {[is_text_field_active]} {
      set_tilde_hiding true
    }
    break
  }
  bind $wdg <Alt-asciitilde> {
    set_output_placing -1
    break
  }
  bind $wdg <Alt-quoteleft> {
    set_output_placing 1
    break
  }
}

foreach wdg [list $topw $text_field $command_field $search_entry $tags_decls_line $tags_links_line] {
  bind $wdg <Alt-percent> {
    if {[is_command_field_active]} {
      update_info_labels_request true
    }
    break
  }
  bind $wdg <Alt-b> {
    if {[is_command_field_active]} {
      update_info_labels_bound_request
    }
    break
  }
  bind $wdg <Alt-t> {
    if {[is_stack_mode] || ([is_batch_mode] && [does_text_batch_input])} {
      change_info_show_time_request
    }
    break
  }
  bind $wdg <Alt-r> {
    if {[is_batch_mode]} {
      change_info_display_adjust_request false true
    }
    break
  }
  bind $wdg <Alt-asciicircum> {
    if {[is_batch_mode]} {
      change_info_display_adjust_request true false
    }
    break
  }
  bind $wdg <Alt-i> {
    if {[is_stack_mode]} {
      change_info_include_input_request
    }
    break
  }
  bind $wdg <Alt-semicolon> {
    if {[is_batch_mode] && [does_text_batch_input]} {
      change_info_make_lineendings_request
    }
    if {[is_stack_mode_common]} {
      change_info_make_lineendings_request
    }
    break
  }
  bind $wdg <Alt-plus> {
    if {[is_batch_mode] && [does_text_batch_input]} {
      change_info_batch_stopping_request
    }
    break
  }
}

proc get_command_content {{selections false}} {
  global command_field
  global mxmbatch_input_field

  set use_wdg [expr {[is_stack_mode] ? $command_field : $mxmbatch_input_field}]
  if {$use_wdg==""} {
    return ""
  }

  if {!$selections} {
    return [$use_wdg get "1.0" "end - 1 chars"]
  }

  set command_text_list [list]
  foreach {sel_start sel_after} [$use_wdg tag ranges "sel"] {
    lappend command_text_list [$use_wdg get $sel_start $sel_after]
  }
  return [join $command_text_list "\n"]

}

set minima_clipboard_using_selection false
set minima_clipboard_waiting_to_get_stored false
set minima_clipboard_owner_state false
set minima_clipboard_text_content ""

set minima_clipboard_text_loose_outer_tests false
set minima_clipboard_text_loose_outer_setup 0
proc minima_clipboard_text_loose_outer {} {
  global minima_clipboard_text_loose_outer_tests
  global minima_clipboard_waiting_to_get_stored
  global minima_clipboard_text_loose_outer_setup

  if {$minima_clipboard_text_loose_outer_tests != false} {
    catch {
      after cancel $minima_clipboard_text_loose_outer_tests
      set minima_clipboard_text_loose_outer_tests false
    }
  }

  set may_wait true
  if {$minima_clipboard_text_loose_outer_setup==0} {
    set minima_clipboard_text_loose_outer_setup [clock milliseconds]
  } {
    if {[expr {$minima_clipboard_text_loose_outer_setup + 1000} < [clock milliseconds]]} {
      set minima_clipboard_text_loose_outer_setup 0
      set may_wait false
    }
  }

  if {(!$minima_clipboard_waiting_to_get_stored) || (!$may_wait)} {
    minima_clipboard_text_loose
  } {
    set minima_clipboard_text_loose_outer_tests [after 100 "minima_clipboard_text_loose_outer"]
  }
}

proc minima_clipboard_text_loose {} {
  global minima_clipboard_owner_state
  global minima_clipboard_text_content
  global minima_clipboard_text_loose_outer_setup
  global is_ending

  set minima_clipboard_text_loose_outer_setup 0

  set minima_clipboard_owner_state false
  if {!$is_ending} {
    set minima_clipboard_text_content ""
  }
}

proc minima_clipboard_text_onset {} {
  global minima_clipboard_owner_state

  set minima_clipboard_owner_state true
  selection own -command minima_clipboard_text_loose_outer -selection CLIPBOARD .
}

proc minima_clipboard_reset_waiting_state {} {
  global minima_clipboard_waiting_to_get_stored
  set minima_clipboard_waiting_to_get_stored false
}

proc minima_clipboard_text_provide {offset max_chars} {
  global minima_clipboard_owner_state
  global minima_clipboard_waiting_to_get_stored
  global is_ending

  if {(!$minima_clipboard_owner_state) && (!$is_ending)} {
    set minima_clipboard_waiting_to_get_stored false
    return ""
  }

  # apparently have to set "minima_clipboard_waiting_to_get_stored" to "true" by default,
  # since (at least during ending) getting multiply requests on clipboard;
  set minima_clipboard_waiting_to_get_stored true
  if {[string length [get_clipboard_text]] <= [expr {$offset+$max_chars}]} {
    after 0 "minima_clipboard_reset_waiting_state"
  }

  return [string range [get_clipboard_text] $offset [expr {$max_chars+$offset-1}]]
}

proc minima_clipboard_text_setup {} {
  global minima_clipboard_using_selection
  global minima_clipboard_without_selection
  global minima_clipboard_force_selection
  global tk_version
  global tk_patchLevel

  if {$minima_clipboard_without_selection} {
    return
  }

  if {!$minima_clipboard_force_selection} {

    if {![is_linux_like]} {
      return
    }
    set current_tk_version [split $tk_version "."]
    set current_tk_version_major [lindex $current_tk_version 0]
    set current_tk_version_minor [lindex $current_tk_version 1]
    if {![string is integer -strict $current_tk_version_major]} {
      set current_tk_version_major 0
    }
    if {![string is integer -strict $current_tk_version_minor]} {
      set current_tk_version_minor 0
    }
    set current_tk_version_patch [lindex [split $tk_patchLevel "."] 2]
    if {![string is integer -strict $current_tk_version_patch]} {
      set current_tk_version_patch 0
    }

    if {$current_tk_version_major<8} {
      return
    }
    if {($current_tk_version_major==8) && ($current_tk_version_minor<6)} {
      return
    }
    if {($current_tk_version_major==8) && ($current_tk_version_minor==6) && ($current_tk_version_patch<8)} {
      return
    }

  }

  set can_setup false
  if {[catch {
    set mng_targets [selection get -selection CLIPBOARD_MANAGER -type TARGETS]
    if {"SAVE_TARGETS" in $mng_targets} {
      set can_setup true
    }
  }]} {
    set can_setup false
  }
  if {!$can_setup} {
    return
  }

  set minima_clipboard_using_selection true

  if {[catch {
    selection handle -selection CLIPBOARD . minima_clipboard_text_provide
  }]} {
    set minima_clipboard_using_selection false
  }
}
minima_clipboard_text_setup

proc clear_clipboard {} {
  global minima_clipboard_using_selection
  global minima_clipboard_text_content

  if {$minima_clipboard_using_selection} {
    set minima_clipboard_text_content ""
    minima_clipboard_text_onset
    return
  }

  catch {
    clipboard clear
  }
}

# Tcl/Tk seems to put selection-start to selection-end region
# into clipboard and that means not-only selected region if selection
# consists of more than one region;
proc set_clipboard {wdg} {
  global minima_clipboard_using_selection
  global minima_clipboard_text_content

  set content [list]
  foreach {one_range_start one_range_end} [$wdg tag ranges "sel"] {
    lappend content [$wdg get $one_range_start $one_range_end]
  }
  set content [join $content "\n"]
  if {[string length $content]==0} {
    return
  }

  if {$minima_clipboard_using_selection} {
    set minima_clipboard_text_content $content
    minima_clipboard_text_onset
    return
  }

  clipboard clear

  if {[is_wine_like]} {
    clipboard append -- $content
  } {
    # works bad with "-type UTF8_STRING"
    #clipboard append -type UTF8_STRING -- $content
    clipboard append -- $content
  }
}
proc set_clipboard_text {txt} {
  global minima_clipboard_using_selection
  global minima_clipboard_text_content

  if {[string length $txt]==0} {
    return
  }

  if {$minima_clipboard_using_selection} {
    set minima_clipboard_text_content $txt
    minima_clipboard_text_onset
    return
  }

  clipboard clear

  if {[is_wine_like]} {
    clipboard append -- $txt
  } {
    # works bad with "-type UTF8_STRING"
    #clipboard append -type UTF8_STRING -- $txt
    clipboard append -- $txt
  }
}
proc append_clipboard_text {txt} {
  global minima_clipboard_using_selection
  global minima_clipboard_text_content

  if {[string length $txt]==0} {
    return
  }

  if {$minima_clipboard_using_selection} {
    append minima_clipboard_text_content $txt
    minima_clipboard_text_onset
    return
  }

  if {[is_wine_like]} {
    clipboard append -- $txt
  } {
    # works bad with "-type UTF8_STRING"
    #clipboard append -type UTF8_STRING -- $txt
    clipboard append -- $txt
  }
}
proc get_clipboard_text {} {
  global minima_clipboard_using_selection
  global minima_clipboard_text_content
  global minima_clipboard_owner_state
  global is_ending

  # if not both using-and-owning selection, getting for clipboard
  if {$minima_clipboard_using_selection} {
    if {$minima_clipboard_owner_state || $is_ending} {
      return $minima_clipboard_text_content
    }
  }

  set string_return ""
  if {[catch {
    if {[is_wine_like]} {
      set string_return [clipboard get]
    } {
      if {[catch {set string_return [clipboard get -type UTF8_STRING]}]} {
        set string_return [clipboard get]
      }
    }
  }]} {
    return ""
  }
  return $string_return
}

# Tcl/Tk does not enforce clipboard preservation after its exit
# thus providing an auxiliary script (in gtk) for that;
# it is not necessary for clipboard work during Minima running,
# it's just for that after-Minima-is-closed clipboard preservation;
# tested this (with provided program) under XFCE Ubunutu 16.04;
# since support from desktop environments is not reliable,
# this should only be called at the application end;
proc store_clipboard {} {
  # direct saving into GNOME clipboard manager (when present) can be used since Tcl/Tk 8.6.8;
  global minima_clipboard_using_selection
  global minima_clipboard_owner_state
  global minima_clipboard_waiting_to_get_stored

  if {$minima_clipboard_using_selection} {
    if {!$minima_clipboard_owner_state} {
      return
    }

    set minima_clipboard_waiting_to_get_stored true
    if {[catch {
      selection get -selection CLIPBOARD_MANAGER -type SAVE_TARGETS
    }]} then {
      # if failing in the SAVE_TARGETS request, reverting the form here;
      set minima_clipboard_waiting_to_get_stored false
    } else {
      # that is only ending here if the request (here above) did not fail;
      # notice that this function is only called when Minima ends,
      # thus this reverting is just for the attempt of the duration storage;
      return
    }
  }

  set use_clipboard_enforcer [get_clipboard_enforcer]

  if {$use_clipboard_enforcer==""} {
    return ""
  }

  if {[catch {
    # can not run it without "&" since we have to answer its clipboard request
    set ret_val [exec $use_clipboard_enforcer "&"]
  }]} {
    return false
  }
  return ret_val
}
set store_clipboard_ever_attempted false
proc store_clipboard_onset {} {
  global store_clipboard_ever_attempted
  # since having issues (under some desktop environments)
  # with clipboard preservation, leaving it to the application end,
  # thus only making a mark here;

  set store_clipboard_ever_attempted true
  return

  after 0 {
    store_clipboard
  }
}

set command_history [list]
set display_history ""
set display_history_text ""
proc draw_history {} {
  global command_history
  global display_history
  global display_history_text
  global select_fg_col
  global select_bg_col

  remove_torn_menus

  catch {destroy $display_history}

  set display_history [toplevel ".history" -width 600 -height 400]
  wm title $display_history "TGM Minima: History"
  wm minsize $display_history 300 200
  wm resizable $display_history true true

  set top_pos [get_top_window_position]
  set x_pos [expr {60 + [lindex $top_pos 0]}]
  set y_pos [expr {40 + [lindex $top_pos 1]}]
  wm geometry $display_history "+$x_pos+$y_pos"

  set ftext [frame $display_history.ftext]
  pack $ftext -side top -expand yes -fill both

  set display_history_text [text $ftext.history -wrap none]
  ttk::scrollbar $ftext.vscr -orient vertical -command "$ftext.history yview"
  ttk::scrollbar $ftext.hscr -orient horizontal -command "$ftext.history xview"
  $ftext.history configure -xscroll "$ftext.hscr set" -yscroll "$ftext.vscr set"
  ttk::sizegrip $ftext.size_grip
  $ftext.history configure -selectbackground $select_bg_col -selectforeground $select_fg_col -inactiveselectbackground $select_bg_col

  grid $ftext.history $ftext.vscr -sticky nsew
  grid $ftext.hscr $ftext.size_grip -sticky nsew

  grid rowconfigure $ftext $ftext.history -weight 1
  grid columnconfigure $ftext $ftext.history -weight 1

  set export_string ""

  set info_count 0
  foreach one_info $command_history {
    set one_info_len [llength $one_info]
    if {$one_info_len==0} {
      continue
    }
    incr info_count
    if {$one_info_len==1} {
      # info on computation time, etc.
      append export_string "/* " [string trim [lindex $one_info 0]] " */\n"
    } elseif {$one_info_len==2} {
      # datetime of command issuing and the command itself
      append export_string "/* " [string trim [lindex $one_info 0]] " */\n" [string trim [lindex $one_info 1]] "\n"
    } elseif {$one_info_len==3} {
      # stack-level deleted, etc.
      append export_string "/* " [string trim [lindex $one_info 1]] " */\n" [string trim [lindex $one_info 2]] "\n"
    }
  }

  if {$info_count>0} {
    $ftext.history insert "1.0" $export_string
  }

  $ftext.history configure -state disabled
  focus $display_history

  bind $display_history <Control-a> {
    $display_history_text tag remove "sel" "1.0" "end"
    $display_history_text tag add "sel" "1.0" "end -1 chars"
    set_anchor_position $display_history_text "1.0"
  }
  bind $display_history_text <Control-a> {
    $display_history_text tag remove "sel" "1.0" "end"
    $display_history_text tag add "sel" "1.0" "end -1 chars"
    set_anchor_position $display_history_text "1.0"
  }
  bind $display_history <Control-c> {
    set_clipboard $display_history_text
    store_clipboard_onset
  }
  bind $display_history_text <Control-c> {
    set_clipboard $display_history_text
    store_clipboard_onset
  }

  bind $display_history <Return> {
    focus $display_history_text
  }
  bind $display_history_text <1> {
    focus $display_history_text
  }

  bind $display_history <Escape> {
    catch {destroy $display_history}
  }

  catch {focus $display_history_text}
  return
}

set command_stack [list ""]
set command_stack_index 0

set display_stack ""
set display_stack_text ""
proc draw_stack {} {
  global command_stack
  global display_stack
  global display_stack_text
  global select_fg_col
  global select_bg_col

  remove_torn_menus

  catch {destroy $display_stack}

  set display_stack [toplevel ".stack" -width 600 -height 400]
  wm title $display_stack "TGM Minima: Stack"
  wm minsize $display_stack 300 200
  wm resizable $display_stack true true

  set top_pos [get_top_window_position]
  set x_pos [expr {40 + [lindex $top_pos 0]}]
  set y_pos [expr {20 + [lindex $top_pos 1]}]
  wm geometry $display_stack "+$x_pos+$y_pos"

  set ftext [frame $display_stack.ftext]
  pack $ftext -side top -expand yes -fill both

  set display_stack_text [text $ftext.stack -wrap none]
  ttk::scrollbar $ftext.vscr -orient vertical -command "$ftext.stack yview"
  ttk::scrollbar $ftext.hscr -orient horizontal -command "$ftext.stack xview"
  $ftext.stack configure -xscroll "$ftext.hscr set" -yscroll "$ftext.vscr set"
  ttk::sizegrip $ftext.size_grip
  $ftext.stack configure -selectbackground $select_bg_col -selectforeground $select_fg_col -inactiveselectbackground $select_bg_col

  grid $ftext.stack $ftext.vscr -sticky nsew
  grid $ftext.hscr $ftext.size_grip -sticky nsew

  grid rowconfigure $ftext $ftext.stack -weight 1
  grid columnconfigure $ftext $ftext.stack -weight 1

  save_current_stack

  set export_string ""

  set command_count 0
  foreach one_command $command_stack {
    set one_command [string trim $one_command]
    if {$one_command == ""} {
      continue
    }
    append one_command "\n"
    append export_string "/* */\n" $one_command
    incr command_count
  }

  if {$command_count>0} {
    $ftext.stack insert "1.0" $export_string
  }

  $ftext.stack configure -state disabled
  focus $display_stack

  bind $display_stack <Control-a> {
    $display_stack_text tag remove "sel" "1.0" "end"
    $display_stack_text tag add "sel" "1.0" "end -1 chars"
    set_anchor_position $display_stack_text "1.0"
  }
  bind $display_stack_text <Control-a> {
    $display_stack_text tag remove "sel" "1.0" "end"
    $display_stack_text tag add "sel" "1.0" "end -1 chars"
    set_anchor_position $display_stack_text "1.0"
  }
  bind $display_stack <Control-c> {
    set_clipboard $display_stack_text
    store_clipboard_onset
  }
  bind $display_stack_text <Control-c> {
    set_clipboard $display_stack_text
    store_clipboard_onset
  }

  bind $display_stack <Return> {
    focus $display_stack_text
  }
  bind $display_stack_text <1> {
    focus $display_stack_text
  }

  bind $display_stack <Escape> {
    catch {destroy $display_stack}
  }

  catch {focus $display_stack_text}
  return
}

proc stack_on_exec_start {} {
  save_current_stack
}

proc stack_on_exec_end {} {
  global command_stack
  global command_stack_index
  global command_stack_saving_index
  global command_field

  if {$command_stack_saving_index>-1} {
    if {$command_stack_index==$command_stack_saving_index} {
      unset_stack_level_saving
    } elseif {$command_stack_index<$command_stack_saving_index} {
      incr command_stack_saving_index -1
    }
  }

  set command_stack_length [llength $command_stack]
  while {$command_stack_index < [expr {$command_stack_length - 1}]} {
    if {[string trim [lindex $command_stack "end"]] == ""} {
      set command_stack [lreplace $command_stack "end" "end"]
      set command_stack_length [llength $command_stack]
    } {
      break
    }
  }

  # checking whether saving_index is not topwise with empty field here;
  set to_put_saving_atop false
  if {$command_stack_saving_index >= [expr {$command_stack_length - 1}]} {
    set to_put_saving_atop true
  }

  set current_part [lindex $command_stack $command_stack_index]

  if {$command_stack_index < [expr {$command_stack_length - 1}]} {
    set prev_part [string trim [lindex $command_stack "end"]]
    set command_stack [lreplace $command_stack $command_stack_index $command_stack_index]
    if {[string trim $current_part] == $prev_part} {
      set command_stack [lreplace $command_stack "end" "end" $current_part]
    } {
      lappend command_stack $current_part
    }
  } {
    if {$command_stack_length > 1} {
      set prev_part [string trim [lindex $command_stack "end-1"]]
      if {[string trim $current_part] == $prev_part} {
        set command_stack [lreplace $command_stack "end-1" "end" $current_part]
      } {
        set command_stack [lreplace $command_stack "end" "end" $current_part]
      }
    } {
      set command_stack [list $current_part]
    }
  }

  lappend command_stack ""
  set command_stack_index [llength $command_stack]
  incr command_stack_index -1

  if {$to_put_saving_atop} {
    set command_stack_saving_index $command_stack_index
  }

  update_stack_label

  return
}

proc blank_command_field {reset} {
  global command_field

  $command_field delete "1.0" "end"

  if {$reset} {
    $command_field edit reset
    $command_field edit modified false
  }

  update_cmmd_highlight_lines
  update_cmmd_strings_comments
}

proc remove_current_from_command {upwards do_save} {
  global command_stack
  global command_stack_index
  global command_stack_saving_index
  global command_field
  global command_history

  if {$command_stack_saving_index==$command_stack_index} {
    show_text_field_message_error [list [list "Removal of the saving stack level is blocked"]]
    return
  }

  if {$do_save} {
    set curr_content [string trim [$command_field get "1.0" "end - 1 chars"]]
    if {$curr_content!=""} {
      lappend command_history [list [clock milliseconds] "stack level deleted" $curr_content]
    }
  }

  if {($command_stack_saving_index!=-1) && ($command_stack_saving_index>$command_stack_index)} {
    incr command_stack_saving_index -1
  }

  set command_stack_length [llength $command_stack]

  if {$command_stack_length==1} {
    set command_stack [list ""]
    set command_stack_index 0
    blank_command_field true
    update_stack_label
    update_cmmd_highlight_lines
    update_cmmd_strings_comments
    return
  }

  set top_index [expr {$command_stack_length - 1}]

  # at top, going upwards
  if {$upwards && ($command_stack_index==$top_index)} {
    set command_stack [lreplace $command_stack "end" "end" ""]
    blank_command_field true
    update_stack_label
    update_cmmd_highlight_lines
    update_cmmd_strings_comments
    return
  }

  # not being at top here, moving upwards
  if {$upwards || ($command_stack_index==0)} {
    set command_stack [lreplace $command_stack $command_stack_index $command_stack_index]
    $command_field replace "1.0" "end" [lindex $command_stack $command_stack_index]
    $command_field edit reset
    $command_field edit modified false

    if {[string trim [lindex $command_stack "end"]] != ""} {
      lappend command_stack ""
    }

    update_stack_label
    update_cmmd_highlight_lines
    update_cmmd_strings_comments
    return
  }

  # not being at bottom, moving downwards
  set command_stack [lreplace $command_stack $command_stack_index $command_stack_index]
  incr command_stack_index -1

  $command_field replace "1.0" "end" [lindex $command_stack $command_stack_index]
  $command_field edit reset
  $command_field edit modified false

  if {[string trim [lindex $command_stack "end"]] != ""} {
    lappend command_stack ""
  }

  update_stack_label
  update_cmmd_highlight_lines
  update_cmmd_strings_comments
  return
}

bind $stack_label_pos <3> {
  if {[is_command_field_active]} {
    pass_in_stack -1
  }
}
bind $stack_label_pos <1> {
  if {[is_command_field_active]} {
    pass_in_stack_one_up
  }
}
bind $stack_label_pos <Shift-1> {
  if {[is_command_field_active]} {
    move_in_stack_to_top
  }
}
bind $stack_label_pos <Control-1> {
  if {[is_command_field_active]} {
    move_in_stack_to_saving
  }
}

foreach wdg [list $topw $text_field $command_field $search_entry $tags_decls_line $tags_links_line] {
  bind $wdg <Control-p> {
    if {[is_stack_mode]} {
      if {[is_command_field_active]} {
        pass_in_stack -1
      }
    } {
      slide_to_prev_block_start_or_section_line
    }
    break
  }
  bind $wdg <Control-n> {
    if {[is_stack_mode]} {
      if {[is_command_field_active]} {
        pass_in_stack_one_up
      }
    } {
      slide_to_next_block_start_or_section_line
    }
    break
  }
  bind $wdg <Control-l> {
    if {[is_command_field_active]} {
      move_in_stack_to_top
    }
    break
  }
  bind $wdg <Control-j> {
    if {[is_command_field_active]} {
      move_in_stack_to_saving
    }
    break
  }
  bind $wdg <Control-k> {
    if {[is_command_field_active]} {
      remove_current_from_command false true
    }
    break
  }
  bind $wdg <Control-K> {
    if {[is_command_field_active]} {
      remove_current_from_command true true
    }
    break
  }
  bind $wdg <Control-Alt-k> {
    if {[is_command_field_active]} {
      remove_current_from_command false false
    }
    break
  }
  bind $wdg <Control-Alt-K> {
    if {[is_command_field_active]} {
      remove_current_from_command true false
    }
    break
  }
}

proc cmmd_got_modified {} {
  remove_torn_menus
  if {[is_command_field_active]} {
    save_current_stack
    update_cmmd_highlight_lines
    update_cmmd_strings_comments
  }
}

bind $command_field <<Modified>> {
  cmmd_got_modified
}

proc get_stack_size {} {
  global command_stack

  if {![is_stack_mode]} {
    return 0
  }
  return [llength $command_stack]
}

proc save_current_stack {} {
  global command_stack
  global command_stack_index
  global command_field

  if {![is_command_field_active]} {
    return
  }

  lset command_stack $command_stack_index [get_command_content]

  return
}

proc move_in_stack_to_top {} {
  global command_stack
  pass_in_stack [llength $command_stack]
}

proc move_in_stack_to_saving {} {
  global command_stack_index
  global command_stack_saving_index

  if {$command_stack_saving_index==-1} {
    show_text_field_message_error [list [list "No stack level marked for saving (Control+m/M)"]]
    return
  }

  pass_in_stack [expr {$command_stack_saving_index - $command_stack_index}]
}

proc pass_in_stack_one_up {} {
  global command_stack_index
  global command_stack

  # allowing to make a free space at (non-empty) stack top
  if {$command_stack_index == [expr {[llength $command_stack] - 1}]} {
    if {[string trim [lindex $command_stack "end"]]!=""} {
      lappend command_stack ""
    }
  }

  pass_in_stack 1
}

proc pass_in_stack {step} {
  global command_stack
  global command_stack_index
  global command_field

  save_current_stack

  set new_stack_index $command_stack_index
  incr new_stack_index $step
  if {$new_stack_index<0} {
    set new_stack_index 0
  }
  if {$new_stack_index>=[llength $command_stack]} {
    set new_stack_index [llength $command_stack]
    incr new_stack_index -1
  }
  if {$command_stack_index==$new_stack_index} {
    update_cmmd_highlight_lines
    update_cmmd_strings_comments
    return false
  }

  set command_stack_index $new_stack_index

  $command_field replace "1.0" "end" [lindex $command_stack $command_stack_index]
  $command_field edit reset
  $command_field edit modified false

  update_stack_label
  update_cmmd_highlight_lines
  update_cmmd_strings_comments
  return true
}

proc set_reopen_button_actual_state {} {
  global recently_opened_files
  global bt_file_open_re
  global text_field

  if {![is_text_field_active]} {
    $bt_file_open_re configure -state "disabled"
    return
  }

  if {[llength $recently_opened_files]==0} {
    $bt_file_open_re configure -state "disabled"
    return
  }

  $bt_file_open_re configure -state "normal"
}

proc set_void_button_actual_state {} {
  global bt_file_void
  global bt_file_export
  global save_path
  global text_field

  set state_void "disabled"
  set state_export "disabled"

  if {[$text_field index "end -1 chars"]!="1.0"} {
    set state_void "normal"
    set state_export "normal"
  } elseif {[$text_field edit modified]} {
    set state_void "normal"
  } elseif {$save_path!=""} {
    set state_void "normal"
  }

  if {![is_text_field_active]} {
    set state_void "disabled"
  }

  $bt_file_void configure -state $state_void
  $bt_file_export configure -state $state_export
}

proc app_labels_on_modified {} {
  global text_field
  global save_path
  global topw
  global app_title_base

  set has_file false
  set file_path_part ""
  if {$save_path != ""} {
    set has_file true
    set file_path_part [get_app_label_path_part]

    if {[string length $file_path_part] > 70} {
      set file_path_start [string range $file_path_part 0 20]
      set file_path_end [string range $file_path_part "end-45" "end"]
      set file_path_middle "..."
      set file_path_part "$file_path_start$file_path_middle$file_path_end"
    }
  }

  set use_base "$app_title_base"

  set text_modified_yes " *"
  set text_modified_yes_file " *"
  set text_modified_no ""
  set text_modified_no_file ""
  set command_diff [check_stack_save_state]
  if {$command_diff == 1} {
    set text_modified_yes " *(c,t)"
    set text_modified_yes_file " *(c,t)"
    set text_modified_no " *(c)"
    set text_modified_no_file " *(c)"
  }
  if {$command_diff == 0} {
    set text_modified_yes " *(t)"
    set text_modified_yes_file " *(t)"
    set text_modified_no ""
    set text_modified_no_file ""
  }

  if {$has_file} {
    if {[$text_field edit modified]} {
      wm title $topw "$file_path_part - $use_base$text_modified_yes_file"
    } {
      wm title $topw "$file_path_part - $use_base$text_modified_no_file"
    }
  } {
    if {[$text_field edit modified]} {
      wm title $topw "$use_base$text_modified_yes"
    } {
      wm title $topw "$use_base$text_modified_no"
    }
  }

  set_void_button_actual_state
}

proc remove_selected_if_by_cursor {wdg} {
  set ins_pos [split [$wdg index "insert"] "."]
  set inc_line [lindex $ins_pos 0]
  set inc_column [lindex $ins_pos 1]

  foreach {sel_rgn_start sel_rgn_after} [$wdg tag ranges "sel"] {
    set sel_rgn_start_info [split $sel_rgn_start "."]
    set sel_rgn_start_line [lindex $sel_rgn_start_info 0]
    set sel_rgn_start_column [lindex $sel_rgn_start_info 1]
    if {$sel_rgn_start_line>$inc_line} {
      break
    }
    if {$sel_rgn_start_line==$inc_line} {
      if {$sel_rgn_start_column>$inc_column} {
        break
      }
    }
    set sel_rgn_after_info [split $sel_rgn_after "."]
    set sel_rgn_after_line [lindex $sel_rgn_after_info 0]
    set sel_rgn_after_column [lindex $sel_rgn_after_info 1]
    if {$sel_rgn_after_line<$inc_line} {
      continue
    }
    if {$sel_rgn_after_line==$inc_line} {
      if {$sel_rgn_after_column<$inc_column} {
        continue
      }
    }
    # if here, we are within a sel region
    $wdg delete $sel_rgn_start $sel_rgn_after
    break
  }
}

proc delete_selected_text {wdg} {
  set sel_parts [$wdg tag ranges "sel"]
  if {[llength $sel_parts]==0} {
    return
  }

  set curr_state [$wdg cget -state]
  $wdg configure -state "normal"
  $wdg tag remove "sel" "1.0" "end"
  foreach {region_end region_start} [lreverse $sel_parts] {
    $wdg delete $region_start $region_end
  }
  $wdg configure -state $curr_state
}

bind $command_field <Control-x> {
  if {[is_command_field_active]} {
    $command_field edit separator
    set_clipboard $command_field
    store_clipboard_onset
    #tk_textCut $command_field
    delete_selected_text $command_field
    $command_field edit separator
    update_command_properties
  }
  break
}
bind $command_field <Control-c> {
  if {[is_command_field_active]} {
    $command_field edit separator
    #tk_textCopy $command_field
    set_clipboard $command_field
    store_clipboard_onset
    $command_field edit separator
    update_command_properties
  }
  break
}

# the tk_textPaste calls apparently do not lead to loosing the selection,
# thus using the default "tk_textPaste" method here (at least by now);
# otherwise would need to reimplement it (alike other tk_text..... methods);

bind $command_field <Control-V> {
  if {[is_command_field_active]} {
    $command_field edit separator
    #tk_textPaste $command_field
    $command_field insert "insert" [get_clipboard_text]
    $command_field edit separator
    $command_field see "insert"
    update_command_properties
  }
  break
}
bind $command_field <Control-v> {
  if {[is_command_field_active]} {
    $command_field edit separator
    remove_selected_if_by_cursor $command_field
    $command_field edit separator
    #tk_textPaste $command_field
    $command_field insert "insert" [get_clipboard_text]
    $command_field edit separator
    $command_field see "insert"
    update_command_properties
  }
  break
}

proc get_selected_parts {wdg {without_hidden false} {without_tags false}} {
  global text_field
  global tags_decls_list
  global tags_decls_state_clean
  global tags_links_list
  global tags_links_state_clean
  global use_tab_length

  set tab_spacer [string repeat " " $use_tab_length]

  # the simpler case, i.e. without any subtracting
  if {!$without_tags} {
    set sel_parts ""
    set end_row 0
    # regions that are without newlines between themselves should be connected by spaces;
    foreach {region_start region_end} [$wdg tag ranges "sel"] {
      if {$sel_parts==""} {
        if {$without_hidden} {
          set sel_parts [$wdg get -displaychars -- $region_start $region_end]
        } {
          set sel_parts [$wdg get -- $region_start $region_end]
        }
      } {
        set connector "\n"
        if {[lindex [split $region_start "."] 0] == $end_row} {
          set connector $tab_spacer
        }
        append sel_parts $connector
        if {$without_hidden} {
          append sel_parts [$wdg get -displaychars -- $region_start $region_end]
        } {
          append sel_parts [$wdg get -- $region_start $region_end]
        }
      }
      set end_row [lindex [split $region_end "."] 0]
    }
    return $sel_parts
  }

  # below dealing with tag subtraction

  set tags_use_list [list]

  set wdg_tag_lengths [list]
  set wdg_tag_regexp [strcat "(?:" [get_tag_regexp false] ")|(?:" [get_tag_regexp true] \
    ")|(?:" [get_sec_regexp] ")|(?:" [get_til_regexp] ")"]
  set wdg_tag_starts [$wdg search -all -count wdg_tag_lengths -regexp $wdg_tag_regexp "1.0"]

  set wdg_tag_idx 0
  set wdg_tag_count [llength $wdg_tag_starts]
  while {$wdg_tag_idx < $wdg_tag_count} {
    lappend tags_use_list [list [lindex $wdg_tag_starts $wdg_tag_idx] [lindex $wdg_tag_lengths $wdg_tag_idx]]
    incr wdg_tag_idx
  }

  set use_regions [list]

  foreach {region_start region_end} [$wdg tag ranges "sel"] {
    set region_start_info [split $region_start "."]
    set region_start_row [lindex $region_start_info 0]
    set region_start_col [lindex $region_start_info 1]
    set region_end_info [split $region_end "."]
    set region_end_row [lindex $region_end_info 0]
    set region_end_col [lindex $region_end_info 1]

    set use_remainder true
    foreach one_tag $tags_use_list {
      set one_tag_pos_info [split [lindex $one_tag 0] "."]
      set one_tag_pos_row [lindex $one_tag_pos_info 0]
      set one_tag_pos_col [lindex $one_tag_pos_info 1]
      set one_tag_end_col [expr {$one_tag_pos_col + [lindex $one_tag 1]}]

      # if a tag ends before selection start, it does not intersect the selection
      if {$one_tag_pos_row < $region_start_row} {
        continue
      }
      if {$one_tag_pos_row == $region_start_row} {
        if {$one_tag_end_col <= $region_start_col} {
          continue
        }
      }
      # if a tag starts after the selection, we are done with the subtracting
      if {$one_tag_pos_row > $region_end_row} {
        break
      }
      if {$one_tag_pos_row == $region_end_row} {
        if {$one_tag_pos_col >= $region_end_col} {
          break
        }
      }

      # here we are left with tags that intersect the selection

      # cases where tag starts before selection or at selection start: moving the selection start forwards
      if {($one_tag_pos_row == $region_start_row) && ($one_tag_pos_col <= $region_start_col)} {
        set region_start_col [expr {$one_tag_end_col + 0}]
        if {($region_start_row == $region_end_row) && ($region_start_col >= $region_end_col)} {
          # here the whole selection was removed (by the from-start removing)
          set use_remainder false
          break
        }
      }

      # cases where tag ends after selection or at selection end: moving the selection end backwards
      if {($one_tag_pos_row == $region_end_row) && ($one_tag_end_col >= $region_end_col)} {
        set region_end_col [expr {$one_tag_pos_col - 0}]
        if {($region_start_row == $region_end_row) && ($region_start_col >= $region_end_col)} {
          # here the whole selection was removed (by the from-start removing)
          set use_remainder false
          break
        }
      }

      # here we are left with tags that split the selection into two non-empty regions
      lappend use_regions [list [join [list $region_start_row $region_start_col] "."] [join [list $one_tag_pos_row $one_tag_pos_col] "."]]
      set region_start_row $one_tag_pos_row
      set region_start_col [expr {$one_tag_end_col + 0}]
    }

    if {$use_remainder} {
      lappend use_regions [list [join [list $region_start_row $region_start_col] "."] [join [list $region_end_row $region_end_col] "."]]
    }
  }

  set sel_parts ""
  set end_row 0
  # regions that are without newlines between themselves should be connected by spaces;
  foreach one_region $use_regions {
    set region_start [lindex $one_region 0]
    set region_end [lindex $one_region 1]
    if {$sel_parts==""} {
      if {$without_hidden} {
        set sel_parts [$wdg get -displaychars -- $region_start $region_end]
      } {
        set sel_parts [$wdg get -- $region_start $region_end]
      }
    } {
      set connector "\n"
      if {[lindex [split $region_start "."] 0] == $end_row} {
        set connector $tab_spacer
      }
      append sel_parts $connector
      if {$without_hidden} {
        append sel_parts [$wdg get -displaychars -- $region_start $region_end]
      } {
        append sel_parts [$wdg get -- $region_start $region_end]
      }
    }
    set end_row [lindex [split $region_end "."] 0]
  }
  return $sel_parts
}

proc command_to_text_do_paste {{to_trim true}} {
  global command_field
  global mxmbatch_input_field
  global text_field

  set use_field [expr {[is_batch_mode] ? $mxmbatch_input_field : $command_field}]
  if {$use_field==""} {
    return
  }
  set use_show_err [expr {[is_batch_mode] ? "show_batch_maxima_message_error" : "show_command_field_message_error"}]

  if {![is_text_field_active]} {
    $use_show_err [list [list "main text locked"]]
    return
  }
  set sel_text [get_selected_parts $use_field true false]
  if {$to_trim} {
    set sel_text [string trim $sel_text]
  }
  if {$sel_text==""} {
    $use_show_err [list [list "no command content to paste"]]
    return
  }

  $text_field edit separator
  remove_selected_if_by_cursor $text_field
  $text_field edit separator
  $text_field insert "insert" $sel_text
  $text_field edit separator
  $text_field see "insert"
  update_text_properties
  clear_temporary_highlights
  return
}

proc command_to_text_do_copy {at_end {to_trim true}} {
  global command_field
  global mxmbatch_input_field
  global text_field

  set use_field [expr {[is_batch_mode] ? $mxmbatch_input_field : $command_field}]
  if {$use_field==""} {
    return
  }
  set use_show_err [expr {[is_batch_mode] ? "show_batch_maxima_message_error" : "show_command_field_message_error"}]

  if {![is_text_field_active]} {
    $use_show_err [list [list "main text locked"]]
    return
  }
  set sel_text [get_selected_parts $use_field true false]
  if {$to_trim} {
    set sel_text [string trim $sel_text]
  }
  if {$sel_text==""} {
    $use_show_err [list [list "no command content to copy"]]
    return
  }
  if {[string range $sel_text "end" "end"]!="\n"} {
    append sel_text "\n"
  }

  if {!$at_end} {
    set insert_info [prepare_cursor_wise_insertion_point_other false]
  } {
    set insert_info [prepare_end_wise_insertion_point_other]
  }
  set insert_position [lindex $insert_info 0]
  set prepare_insert_info [lindex $insert_info 1]
  if {[llength $prepare_insert_info]==2} {
    $text_field edit separator
    $text_field insert [lindex $prepare_insert_info 0] [lindex $prepare_insert_info 1]
  }
  set real_index [$text_field index [expr {$insert_position=="end" ? "end -1 chars" : $insert_position}]]
  $text_field edit separator
  $text_field insert $insert_position $sel_text
  $text_field edit separator
  $text_field mark set "insert" $real_index
  $text_field see $real_index

  update_text_properties
}

proc command_to_text_do_insert {{to_trim true}} {
  global command_field
  global mxmbatch_input_field
  global text_field

  set use_field [expr {[is_batch_mode] ? $mxmbatch_input_field : $command_field}]
  if {$use_field==""} {
    return
  }
  set use_show_err [expr {[is_batch_mode] ? "show_batch_maxima_message_error" : "show_command_field_message_error"}]

  if {![is_text_field_active]} {
    $use_show_err [list [list "main text locked"]]
    return
  }
  set sel_text [get_selected_parts $use_field true false]
  if {$to_trim} {
    set sel_text [string trim $sel_text]
  }
  if {$sel_text==""} {
    $use_show_err [list [list "no command content to insert"]]
    return
  }
  if {[string range $sel_text "end" "end"]!="\n"} {
    append sel_text "\n"
  }

  set insert_pos [split [$text_field index "insert lineend"] "."]
  set insert_pos_row [lindex $insert_pos 0]
  if {[lindex $insert_pos 1]!=0} {
    incr insert_pos_row
  }
  if {[lindex [split [$text_field index "end -1 chars"] "."] 0] < $insert_pos_row} {
    $text_field edit separator
    $text_field insert "end" "\n"
  }
  $text_field edit separator
  $text_field insert "$insert_pos_row.0" $sel_text
  $text_field edit separator
  $text_field see "$insert_pos_row.0"
  update_text_properties
  clear_temporary_highlights
  return
}

proc command_to_text_do_append {{to_trim true}} {
  global command_field
  global mxmbatch_input_field
  global text_field

  set use_field [expr {[is_batch_mode] ? $mxmbatch_input_field : $command_field}]
  if {$use_field==""} {
    return
  }
  set use_show_err [expr {[is_batch_mode] ? "show_batch_maxima_message_error" : "show_command_field_message_error"}]

  if {![is_text_field_active]} {
    $use_show_err [list [list "main text locked"]]
    return
  }
  set sel_text [get_selected_parts $use_field true false]
  if {$to_trim} {
    set sel_text [string trim $sel_text]
  }
  if {$sel_text==""} {
    $use_show_err [list [list "no command content to append"]]
    return
  }
  if {[string range $sel_text "end" "end"]!="\n"} {
    append sel_text "\n"
  }

  if {[lindex [split [$text_field index "end -1 chars"] "."] 1] != 0} {
    $text_field edit separator
    $text_field insert "end" "\n"
  }
  $text_field edit separator
  $text_field insert "end" $sel_text
  $text_field edit separator
  $text_field see "end"
  update_text_properties
  clear_temporary_highlights
  return
}

# eventually should provide menu for command-to-text copy actions:
# copy/paste/insert/append (both with and without ~-lines enforcing)

bind $command_field <Control-w> {
  if {[is_text_field_active]} {
    command_to_text_do_copy false true
  }
  break
}
bind $command_field <Control-Alt-w> {
  if {[is_text_field_active]} {
    #command_to_text_do_paste true
    command_to_text_do_insert true
  }
  break
}
bind $command_field <Control-W> {
  if {[is_text_field_active]} {
    command_to_text_do_copy true true
  }
  break
}
bind $command_field <Control-Alt-W> {
  if {[is_text_field_active]} {
    command_to_text_do_append true
  }
  break
}

bind $text_field <Control-x> {
  if {[is_text_field_active]} {
    $text_field edit separator
    set_clipboard $text_field
    store_clipboard_onset
    #tk_textCut $text_field
    delete_selected_text $text_field
    $text_field edit separator
    update_text_properties
    clear_temporary_highlights
  }
  break
}
bind $text_field <Control-c> {
  if {[is_text_field_active]} {
    #tk_textCopy $text_field
    set_clipboard $text_field
    store_clipboard_onset
    $text_field edit separator
    update_text_properties
  }
  break
}
bind $text_field <Control-V> {
  if {[is_text_field_active]} {
    $text_field edit separator
    #tk_textPaste $text_field
    $text_field insert "insert" [get_clipboard_text]
    $text_field edit separator
    $text_field see "insert"
    update_text_properties
    clear_temporary_highlights
  }
  break
}
bind $text_field <Control-v> {
  if {[is_text_field_active]} {
    $text_field edit separator
    remove_selected_if_by_cursor $text_field
    $text_field edit separator
    #tk_textPaste $text_field
    $text_field insert "insert" [get_clipboard_text]
    $text_field edit separator
    $text_field see "insert"
    update_text_properties
    clear_temporary_highlights
  }
  break
}

proc text_to_command_do_paste {{reread false}} {
  global command_field
  global text_field
  global text_part_actions_last_content

  if {$reread} {
    set text_part_actions_last_content [get_selected_parts $text_field true false]
  }
  if {$text_part_actions_last_content==""} {
    return
  }
  if {![is_command_field_active]} {
    set text_part_actions_last_content ""
    return
  }

  $command_field edit separator
  remove_selected_if_by_cursor $command_field
  $command_field edit separator
  $command_field insert "insert" $text_part_actions_last_content
  $command_field edit separator
  $command_field see "insert"
  update_command_properties

  set text_part_actions_last_content ""
  return
}

proc text_to_command_do_append {{reread false}} {
  global command_field
  global text_field
  global text_part_actions_last_content

  if {$reread} {
    set text_part_actions_last_content [get_selected_parts $text_field true false]
  }
  if {$text_part_actions_last_content==""} {
    return
  }
  if {![is_command_field_active]} {
    set text_part_actions_last_content ""
    return
  }

  $command_field edit separator
  $command_field insert "end" $text_part_actions_last_content
  $command_field edit separator
  $command_field see "end"
  update_command_properties

  set text_part_actions_last_content ""
  return
}

proc text_to_command_do_overwrite {{reread false}} {
  global command_field
  global text_field
  global text_part_actions_last_content

  if {$reread} {
    set text_part_actions_last_content [get_selected_parts $text_field true false]
  }
  if {$text_part_actions_last_content==""} {
    return
  }
  if {![is_command_field_active]} {
    set text_part_actions_last_content ""
    return
  }

  $command_field edit separator
  $command_field delete "1.0" "end"
  $command_field edit separator
  $command_field insert "insert" $text_part_actions_last_content
  $command_field edit separator
  $command_field see "insert"
  update_command_properties

  set text_part_actions_last_content ""
  return
}
proc text_to_command_do_copytotop {{reread false}} {
  global command_field
  global text_field
  global command_stack
  global command_stack_index
  global text_part_actions_last_content

  if {$reread} {
    set text_part_actions_last_content [get_selected_parts $text_field true false]
  }
  if {$text_part_actions_last_content==""} {
    return
  }
  if {![is_command_field_active]} {
    set text_part_actions_last_content ""
    return
  }

  save_current_stack

  set trimmed_last_command [string trim [lindex $command_stack "end"]]
  if {$trimmed_last_command != ""} {
    lappend command_stack ""
  }
  set command_stack_index [expr {[llength $command_stack] - 1}]

  $command_field delete "1.0" "end"
  $command_field edit reset
  $command_field edit modified false
  $command_field insert "insert" $text_part_actions_last_content
  $command_field edit separator
  $command_field see "insert"
  update_command_properties

  update_stack_label
  set text_part_actions_last_content ""
  return
}

proc text_to_command_do_copytoempty {{reread false}} {

  set trimmed_current_command [string trim [get_command_content]]
  if {$trimmed_current_command != ""} {
    text_to_command_do_copytotop $reread
    return
  }

  text_to_command_do_overwrite $reread
  return
}

bind $text_field <Control-w> {
  offer_text_selection_actions_on_key
  break
}
bind $text_field <Control-W> {
  if {[is_command_field_active]} {
    text_to_command_do_copytoempty true
  }
  break
}
bind $text_field <Control-Alt-w> {
  break
}

bind $text_field <Control-z> {
  if {[is_text_field_active]} {
    catch {
      $text_field edit undo
      update_text_properties
      clear_temporary_highlights
    }
  }
  break
}
foreach redo_keys [list "Control-y" "Control-Z"] {
  bind $text_field <$redo_keys> {
    if {[is_text_field_active]} {
      catch {
        $text_field edit redo
        update_text_properties
        clear_temporary_highlights
      }
    }
    break
  }
}

bind $command_field <Control-z> {
  if {[is_command_field_active]} {
    catch {
      $command_field edit undo
      update_command_properties
    }
  }
  break
}
foreach redo_keys [list "Control-y" "Control-Z"] {
  bind $command_field <$redo_keys> {
    if {[is_command_field_active]} {
      catch {
        $command_field edit redo
        update_command_properties
      }
    }
    break
  }
}

bind $text_field <Control-Alt-z> {
  if {[is_command_field_active]} {
    catch {
      $command_field edit undo
      update_command_properties
    }
  }
  break
}
bind $text_field <Control-Alt-y> {
  if {[is_command_field_active]} {
    catch {
      $command_field edit redo
      update_command_properties
    }
  }
  break
}

bind $command_field <Control-Alt-z> {
  if {[is_text_field_active]} {
    catch {
      $text_field edit undo
      update_text_properties
      clear_temporary_highlights
    }
  }
  break
}
bind $command_field <Control-Alt-y> {
  if {[is_text_field_active]} {
    catch {
      $text_field edit redo
      update_text_properties
      clear_temporary_highlights
    }
  }
  break
}

set focused_part -1
proc set_focused {} {
  global focused_part
  global text_field
  global command_field

  if {![is_stack_mode]} {
    focus $text_field
    return
  }
  set use_command $command_field

  if {$focused_part != -1} {
    if {([focus] != $text_field) && ([focus] != $use_command)} {
      set focused_part [expr {1 - $focused_part}]
    }
  }

  if {$focused_part != 0} {
    set focused_part 0
    focus $use_command
  } {
    set focused_part 1
    focus $text_field
  }
}

foreach wdg [list $topw $text_field $command_field $search_entry $tags_decls_line $tags_links_line] {
  bind $wdg <Shift-space> {
    # using Control-comma/period now
    #set_focused
    break
  }
  bind $wdg <Control-Key-2> {
    # using Control-comma/period now
    #set_focused
    break
  }
  bind $wdg <Control-Prior> {
    if {[is_stack_mode]} {
      if {[focus]==$search_entry} {
        focus $command_field
      } {
        focus $text_field
      }
    } {
      if {[focus]==$search_entry} {
        focus $tags_links_line
      } elseif {[focus]==$tags_links_line} {
        focus $tags_decls_line
      } {
        focus $text_field
      }
    }
    break
  }
  bind $wdg <Control-comma> {
    if {[is_stack_mode]} {
      if {[focus]==$search_entry} {
        focus $command_field
      } {
        focus $text_field
      }
    } {
      if {[focus]==$search_entry} {
        focus $tags_links_line
      } elseif {[focus]==$tags_links_line} {
        focus $tags_decls_line
      } {
        focus $text_field
      }
    }
    break
  }
  bind $wdg <Control-Next> {
    if {[is_stack_mode]} {
      if {[focus]==$search_entry} {
        focus $text_field
      } {
        focus $command_field
      }
    } {
      if {[focus]==$search_entry} {
        focus $text_field
      } elseif {[focus]!=$tags_links_line} {
        if {[focus]==$tags_decls_line} {
          focus $tags_links_line
        } {
          focus $tags_decls_line
        }
      }
    }
    break
  }
  bind $wdg <Control-period> {
    if {[is_stack_mode]} {
      if {[focus]==$search_entry} {
        focus $text_field
      } {
        focus $command_field
      }
    } {
      if {[focus]==$search_entry} {
        focus $text_field
      } elseif {[focus]!=$tags_links_line} {
        if {[focus]==$tags_decls_line} {
          focus $tags_links_line
        } {
          focus $tags_decls_line
        }
      }
    }
    break
  }
}

bind $text_field <Tab> {
  if {[is_text_field_active]} {
    clear_temporary_highlights
    string_action_on_tab $text_field true false
    update_text_properties
  }
  break
}
bind $command_field <Tab> {
  if {[is_command_field_active]} {
    string_action_on_tab $command_field false false
    update_command_properties
  }
  break
}

proc string_action_on_tab_simple {wdg} {
  global use_tab_length

  set tab_width $use_tab_length
  $wdg edit separator

  set ins_pos [split [$wdg index "insert"] "."]
  set ins_col [lindex $ins_pos 1]
  set space_count [expr {$tab_width-($ins_col%$tab_width)}]
  set space_string [string repeat " " $space_count]

  set curr_state [$wdg cget -state]
  $wdg configure -state "normal"
  $wdg insert "insert" $space_string
  $wdg configure -state $curr_state
  $wdg edit separator
}

set min_tilde_count 3
set min_other_count 10
proc string_action_on_tab {wdg is_text fnc_only} {
  global min_tilde_count
  global min_other_count
  global use_tab_length

  $wdg edit separator

  if {$is_text && ([lindex [split [$wdg index "insert"] "."] 1]!=0)} {
    set line_whole [$wdg get "insert linestart" "insert lineend"]
    set line_whole_len [string length $line_whole]
    if {$line_whole_len>0} {
      foreach one_char [list "~" "*" "=" "#"] {
        if {$line_whole==[string repeat $one_char $line_whole_len]} {
          if {($line_whole_len<$min_tilde_count) && ($one_char=="~")} {
            set len_to_add [expr {
              $min_tilde_count - $line_whole_len
            }]
          } {
            set len_to_add [expr {
              $min_other_count-($line_whole_len%$min_other_count)
            }]
          }

          $wdg mark set "insert" "insert lineend"
          $wdg insert "insert" [string repeat $one_char $len_to_add]
          $wdg edit separator
          update_text_highlight
          return
        }
      }
    }
  }

  set to_compare ""
  set line_start [$wdg get "insert linestart" "insert"]

  set function_params_only false
  set line_start_reversed [split [string reverse $line_start] {}]
  if {[lindex $line_start_reversed 0]=="("} {
    set fnc_only true
    set function_params_only true
    set line_start_reversed [lrange $line_start_reversed 1 "end"]
  }

  foreach one_char $line_start_reversed {
    if {![string is ascii -strict $one_char]} {
      break
    }
    if {(![string is alnum -strict $one_char]) && ("_" != $one_char)} {
      break
    }
    set to_compare $one_char$to_compare
  }

  set to_compare_length [string length $to_compare]
  if {$to_compare_length<1} {
    $wdg insert "insert" [string repeat " " $use_tab_length]
  } {
    set paren_pos -1
    if {$function_params_only} {
      set paren_pos $to_compare_length
    }
    set found_part [get_matching_prolongation $to_compare $fnc_only $paren_pos]
    if {$function_params_only} {
      if {$found_part != ""} {
        set found_part [string range $found_part 1 "end"]
      }
    }
    if {$found_part != ""} {
      $wdg insert "insert" $found_part
    }
  }

  $wdg edit separator
}

bind $search_entry <Tab> {
  if {[is_text_field_active]} {
    entry_action_on_tab $search_entry "text_search" false
  }
  break
}
foreach keyspressed [list "Shift-Key-Tab" "Shift-Key-KP_Tab" "Shift-Key-ISO_Left_Tab" "Key-ISO_Left_Tab"] {
  catch {
    bind $search_entry <$keyspressed> {
      if {[is_text_field_active]} {
        entry_action_on_tab $search_entry "text_search" true
      }
      break
    }
  }
}

# alike the text-wise ways, just assuming entry instead of text widget
proc entry_action_on_tab {wdg text_var_name fnc_only} {
  global $text_var_name
  global use_tab_length

  set text_var [set $text_var_name]

  set to_compare ""
  set pos_idx [$wdg index "insert"]
  set line_start [string range $text_var 0 [expr {$pos_idx - 1}]]

  set function_params_only false
  set line_start_reversed [split [string reverse $line_start] {}]
  if {[lindex $line_start_reversed 0]=="("} {
    set fnc_only true
    set function_params_only true
    set line_start_reversed [lrange $line_start_reversed 1 "end"]
  }

  foreach one_char $line_start_reversed {
    if {![string is ascii -strict $one_char]} {
      break
    }
    if {(![string is alnum -strict $one_char]) && ("_" != $one_char)} {
      break
    }
    set to_compare $one_char$to_compare
  }

  set to_compare_length [string length $to_compare]

  set string_to_put_in ""
  if {$to_compare_length<1} {
    set string_to_put_in [string repeat " " $use_tab_length]
  } {
    set paren_pos -1
    if {$function_params_only} {
      set paren_pos $to_compare_length
    }
    set found_part [get_matching_prolongation $to_compare $fnc_only $paren_pos]
    if {$function_params_only} {
      if {$found_part != ""} {
        set found_part [string range $found_part 1 "end"]
      }
    }
    if {$found_part != ""} {
      set string_to_put_in $found_part
    }
  }

  if {$string_to_put_in !=""} {
    set result_string [string range $text_var 0 [expr {$pos_idx - 1}]]
    append result_string $string_to_put_in
    append result_string [string range $text_var $pos_idx "end"]
    set $text_var_name $result_string
    $wdg icursor [expr {$pos_idx + [string length $string_to_put_in]}]
  }
}

# workaround on an issue in Tk where anchor for text selections
# is (lazily) defined too late (only after selections grow from empty one);
# since not sure whether the tk::TextAnchor method will remain defined
# in the future, putting it into catch block;
proc assure_anchor {wdg} {
  catch {
    set anchor_name [tk::TextAnchor $wdg]
    if {[catch {$wdg index $anchor_name}]} {
      $wdg mark set $anchor_name "1.0"
    }
  }
}
assure_anchor $text_field
assure_anchor $command_field

# eventually should do set_anchor_... via:
# "event generate $wdg <Control-space>"
# with suitably positioned insert cursor

# prefering to have selection mark in such a way that under all-selection event,
# we do not lose most of the selected region when moving up/down
proc set_anchor_opposite {wdg} {
  assure_anchor $wdg
  catch {
    set anchor_name [tk::TextAnchor $wdg]
    set end_line [lindex [split [$wdg index "end - 1 chars"] "."] 0]
    set ins_line [lindex [split [$wdg index "insert"] "."] 0]
    if {[expr {2*$ins_line}]>$end_line} {
      $wdg mark set $anchor_name "1.0"
    } {
      $wdg mark set $anchor_name "end - 1 chars"
    }
  }
}

# selections by mouse needs updated anchors on clicks
proc set_anchor_insert {wdg} {
  assure_anchor $wdg
  catch {
    set anchor_name [tk::TextAnchor $wdg]
    set ins_index [$wdg index "insert"]
    $wdg mark set $anchor_name $ins_index
  }  
}
proc set_anchor_insert_line_start {wdg} {
  assure_anchor $wdg
  catch {
    set anchor_name [tk::TextAnchor $wdg]
    set ins_index [$wdg index "insert linestart"]
    $wdg mark set $anchor_name $ins_index
  }  
}
proc set_anchor_position {wdg pos} {
  assure_anchor $wdg
  catch {
    set anchor_name [tk::TextAnchor $wdg]
    set ins_index [$wdg index $pos]
    $wdg mark set $anchor_name $ins_index
  }
}

bind $text_field <Control-Shift-space> {
  event generate $text_field <Control-space>
  break
}
bind $command_field <Control-Shift-space> {
  event generate $command_field <Control-space>
  break
}

bind $text_field <Control-a> {
  $text_field tag remove "sel" "1.0" "end"
  $text_field tag add "sel" "1.0" "end -1 chars"
  set_anchor_opposite $text_field
  $text_field edit separator
  break
}
bind $command_field <Control-a> {
  $command_field tag remove "sel" "1.0" "end"
  $command_field tag add "sel" "1.0" "end -1 chars"
  set_anchor_opposite $command_field
  $command_field edit separator
  break
}

proc follow_tag_link_on_key {with_mod} {

  target_within_tag_entries_on_key

  if {![on_tag_link_try_to_jump_key true $with_mod]} {
    show_text_field_message_error [list [list "no tag link under cursor"]]
  }

}

proc spell_check_on_key {with_alt} {
  global spch_default_allows_concat
  global spch_ascii_only
  global text_field

  if {[$text_field tag nextrange "sel" "1.0"]==""} {
    return
  }

  set with_concat $spch_default_allows_concat
  if {$with_alt} {
    set with_concat [expr {$with_concat ? false : true}]
  }

  spell_check_selected_text $with_concat $spch_ascii_only
}

# have not been using Ctrl-! for links, thus using it for spell-checking actions by now;
bind $text_field <Control-exclam> {
  remove_torn_menus
  spell_check_on_key false
  #if {[is_batch_mode] || ([is_text_field_active] && [is_command_field_active])} {
  #  follow_tag_link_on_key false
  #}
  break
}
bind $text_field <Control-Alt-exclam> {
  remove_torn_menus
  spell_check_on_key true
  #if {[is_batch_mode] || ([is_text_field_active] && [is_command_field_active])} {
  #  follow_tag_link_on_key true
  #}
  break
}

# aimed at basic cases of (several) lines of simple (line-wise) expressions;
# this function only should be used once the system of brackets is checked;
proc put_maxima_command_ends_to_lines {command comments strings brackets} {
  if {$command==""} {
    return [list true ""]
  }

  set in_batch [is_batch_mode]

  set comment_idx 0
  set comment_cnt [llength $comments]
  set comment_curr ""
  if {$comment_cnt>0} {
    set comment_curr [lindex $comments 0]
    set comment_start [lindex $comment_curr 0]
    set comment_start_row [lindex $comment_start 0]
    set comment_start_col [lindex $comment_start 1]
    set comment_end [lindex $comment_curr 1]
    set comment_end_row [lindex $comment_end 0]
    set comment_end_col [lindex $comment_end 1]
  }

  # blanking comments first;
  set lines [list]
  set line_idx 0
  foreach one_line [split $command "\n"] {
    incr line_idx
    set line_got_put false
    while {$comment_curr!=""} {
      if {$line_idx<$comment_start_row} {
        lappend lines $one_line
        set line_got_put true
        break
      } elseif {$line_idx==$comment_start_row} {
        if {$line_idx<$comment_end_row} {
          set one_line [string replace $one_line $comment_start_col "end" \
            [string repeat " " [expr {[string length $one_line] - $comment_start_col}]] \
          ]
          lappend lines $one_line
          set line_got_put true
          break
        } {
          set one_line [string replace $one_line $comment_start_col $comment_end_col \
            [string repeat " " [expr {1 + $comment_end_col - $comment_start_col}]] \
          ]
          incr comment_idx
          set comment_curr ""
          if {$comment_cnt>$comment_idx} {
            set comment_curr [lindex $comments $comment_idx]
            set comment_start [lindex $comment_curr 0]
            set comment_start_row [lindex $comment_start 0]
            set comment_start_col [lindex $comment_start 1]
            set comment_end [lindex $comment_curr 1]
            set comment_end_row [lindex $comment_end 0]
            set comment_end_col [lindex $comment_end 1]
          }
          continue
        }
      } {
        if {$line_idx<$comment_end_row} {
          lappend lines [string repeat " " [string length $one_line]]
          set line_got_put true
          break
        } {
          set one_line [string replace $one_line 0 $comment_end_col \
            [string repeat " " [expr {1 + $comment_end_col}]] \
          ]
          incr comment_idx
          set comment_curr ""
          if {$comment_cnt>$comment_idx} {
            set comment_curr [lindex $comments $comment_idx]
            set comment_start [lindex $comment_curr 0]
            set comment_start_row [lindex $comment_start 0]
            set comment_start_col [lindex $comment_start 1]
            set comment_end [lindex $comment_curr 1]
            set comment_end_row [lindex $comment_end 0]
            set comment_end_col [lindex $comment_end 1]
          }
          continue
        }
      }
    }
    if {!$line_got_put} {
      lappend lines $one_line
    }
  }

  # taking outermost brackets (considering already checked system)
  set outer_pairs [list]

  set curr_pair_start ""
  set bracket_stack_size 0
  foreach one_bracket $brackets {
    set one_bracket_char [lindex $one_bracket 2]
    if {$one_bracket_char in [list "\{" "\[" "("]} {
      if {$bracket_stack_size==0} {
        set curr_pair_start [list [lindex $one_bracket 0] [lindex $one_bracket 1]]
      }
      incr bracket_stack_size
    } {
      incr bracket_stack_size -1
      if {$bracket_stack_size==0} {
        lappend outer_pairs [list $curr_pair_start [list [lindex $one_bracket 0] [lindex $one_bracket 1]]]
      }
    }
  }

  # putting in strings that are outside the pairs of outermost brackets

  set delimiters [list]

  set pair_idx 0
  set pair_cnt [llength $outer_pairs]
  set pair_curr ""
  if {$pair_cnt>0} {
    set pair_curr [lindex $outer_pairs 0]
    set pair_start [lindex $pair_curr 0]
    set pair_start_row [lindex $pair_start 0]
    set pair_start_col [lindex $pair_start 1]
    set pair_end [lindex $pair_curr 1]
    set pair_end_row [lindex $pair_end 0]
    set pair_end_col [lindex $pair_end 1]
  }

  foreach one_string $strings {
    set one_string_start [lindex $one_string 0]
    set one_string_start_row [lindex $one_string_start 0]
    set one_string_start_col [lindex $one_string_start 1]
    set one_string_end [lindex $one_string 1]
    set one_string_end_row [lindex $one_string_end 0]
    set one_string_end_col [lindex $one_string_end 1]

    set to_put_string_in true

    # here taking as granted that (the used) brackets vs. strings either do not intersect or one contains the other completely
    while {$pair_curr!=""} {
      if {($one_string_start_row < $pair_start_row) || (($one_string_start_row == $pair_start_row) && ($one_string_start_col < $pair_start_col))} {
        if {$one_string_start_row < $one_string_end_row} {
          lappend delimiters [list $one_string_start_row $one_string_end_row]
        }
        set to_put_string_in false
        break
      }
      if {($one_string_end_row < $pair_end_row) || (($one_string_start_row == $pair_start_row) && ($one_string_end_col < $pair_end_col))} {
        set to_put_string_in false
        break
      }

      if {$pair_start_row < $pair_end_row} {
        lappend delimiters [list $pair_start_row $pair_end_row]
      }
      incr pair_idx
      set pair_curr ""
      if {$pair_cnt>$pair_idx} {
        set pair_curr [lindex $outer_pairs $pair_idx]
        set pair_start [lindex $pair_curr 0]
        set pair_start_row [lindex $pair_start 0]
        set pair_start_col [lindex $pair_start 1]
        set pair_end [lindex $pair_curr 1]
        set pair_end_row [lindex $pair_end 0]
        set pair_end_col [lindex $pair_end 1]
      }
    }

    if {$to_put_string_in} {
      if {$one_string_start_row < $one_string_end_row} {
        lappend delimiters [list $one_string_start_row $one_string_end_row]
      }
    }

  }

  if {$pair_curr != ""} {
    if {$pair_start_row < $pair_end_row} {
      lappend delimiters [list $pair_start_row $pair_end_row]
    }
    incr pair_idx
    while {$pair_cnt>$pair_idx} {
      set pair_curr [lindex $outer_pairs $pair_idx]
      set pair_start [lindex $pair_curr 0]
      set pair_start_row [lindex $pair_start 0]
      set pair_end [lindex $pair_curr 1]
      set pair_end_row [lindex $pair_end 0]

      if {$pair_start_row < $pair_end_row} {
        lappend delimiters [list $pair_start_row $pair_end_row]
      }
      incr pair_idx
    }
  }

  set blocks [list]

  set delm_idx 0
  set delm_cnt [llength $delimiters]
  set delm_curr ""
  if {$delm_cnt>0} {
    set delm_curr [lindex $delimiters 0]
    set delm_start_row [lindex $delm_curr 0]
    set delm_end_row [lindex $delm_curr 1]
  }

  set curr_block [list]

  set line_idx 0
  set line_cnt [llength $lines]
  while {$line_cnt>$line_idx} {
    set one_line [lindex $lines $line_idx]
    # comparisons are for lines numbered from 1 (while lists count from 0);
    incr line_idx
    lappend curr_block $one_line

    if {$delm_curr != ""} {
      while {$delm_end_row<=$line_idx} {
        incr delm_idx
        set delm_curr ""
        if {$delm_cnt>$delm_idx} {
          set delm_curr [lindex $delimiters $delm_idx]
          set delm_start_row [lindex $delm_curr 0]
          set delm_end_row [lindex $delm_curr 1]
        } {
          break
        }
      }
    }

    if {$delm_curr != ""} {
      if {($delm_start_row<=$line_idx) && ($delm_end_row>$line_idx)} {
        continue
      }
    }

    set block_may_end true
    if {[string range $one_line "end" "end"]=="\\"} {
      set block_may_end false
      foreach one_char [lreverse [split [string range $one_line 0 "end-1"] {}]] {
        if {$one_char=="\\"} {
          set block_may_end [expr {$block_may_end ? false : true}]
        } {
          break
        }
      }
    }

    if {!$block_may_end} {
      continue
    }

    lappend blocks $curr_block
    set curr_block [list]

  }

  if {$curr_block != [list]} {
    lappend blocks $curr_block
  }

  # to trim the blocks

  set use_blocks [list]

  foreach one_block $blocks {
    set one_size [llength $one_block]

    set one_idx [expr {$one_size - 1}]
    while {$one_idx>=0} {
      if {[string trimleft [lindex $one_block $one_idx]] in [list "" "\\"]} {
        incr one_idx -1
        continue
      }
      break
    }

    if {0>$one_idx} {
      continue
    }

    set one_block [lrange $one_block 0 $one_idx]
    set one_size [llength $one_block]

    set one_idx 0
    while {$one_idx < $one_size} {
      if {[string trimleft [lindex $one_block $one_idx]] in [list "" "\\"]} {
        incr one_idx
        continue
      }
      break
    }

    if {$one_size==$one_idx} {
      continue
    }

    set one_block [lrange $one_block $one_idx "end"]
    set last_line [string trimright [lindex $one_block "end"]]

    if {[string range $last_line "end" "end"]=="\\"} {
      set to_add_space true
      foreach one_char [lreverse [split [string $last_line 0 "end-1"] {}]] {
        if {$one_char=="\\"} {
          set to_add_space [expr {$to_add_space ? false : true}]
        } {
          break
        }
      }
      if {$to_add_space} {
        append last_line " "
      }
    }

    lset one_block "end" $last_line
    set one_block [string trimleft [join $one_block "\n"]]

    if {$one_block==""} {
      continue
    }

    lappend use_blocks $one_block
  }

  set block_cnt [llength $use_blocks]
  if {$block_cnt==0} {
    return [list true ""]
  }

  set result [list]

  set curr_idx 0
  while {$curr_idx < $block_cnt} {
    set curr_block [lindex $use_blocks $curr_idx]
    if {$curr_block==""} {
      incr curr_idx
      continue
    }

    if {!([string range $curr_block "end" "end"] in [list "\$" ";"])} {
      set next_idx [expr {$curr_idx + 1}]
      set next_block [lindex $use_blocks $next_idx]
      set next_char [string range $next_block 0 0]
      if {$next_char in [list "\$" ";"]} {
        append curr_block $next_char

        set next_block [string trimleft [string range $next_block 1 "end"]]
        lset use_blocks $next_idx $next_block
      } {
        # ":lisp(...)" lines at batches tend to dislike semicolon endings
        set avoid_ending false
        if {$in_batch} {
          set aux_block [string trimleft $curr_block]
          if {[string range $aux_block 0 0] == ":"} {
            set aux_block [string trimleft [string range $aux_block 1 "end"]]
            if {[string tolower [string range $aux_block 0 3]]=="lisp"} {
              set avoid_ending true
            }
          }
        }
        if {!$avoid_ending} {
          append curr_block ";"
        }
      }
    }

    lappend result $curr_block
    incr curr_idx
  }

  if {[llength $result]==0} {
    return [list true ""]
  }

  return [list true [join $result "\n"]]
}

proc prepare_end_wise_insertion_point_other {} {
  global text_field
  global omit_auto_tildes
  global min_tilde_count
  global text_inner_code_blocks

  # if the last block is not loose-ending,
  # checking block ends for several lines up from the end
  # while those lines are (effectively) empty;

  seek_inner_code
  set within_block false
  set prepend_stuff [list]
  set insertion_point "end"

  if {[llength $text_inner_code_blocks]>0} {
    set last_block_end [lindex [lindex $text_inner_code_blocks "end"] 1]
    if {$last_block_end=="end"} {
      set within_block true
      if {[lindex [split [$text_field index "end -1 chars"] "."] 1] != 0} {
        set prepend_stuff [list "end" "\n"]
      }
      return [list "end" $prepend_stuff]
    }

    set end_row [lindex [split [$text_field index "end -1 chars"] "."] 0]
    set is_within_limits false
    if {[expr {$end_row - 3}] < $last_block_end} {
      set is_within_limits true
      set test_row $end_row
      while {$test_row > $last_block_end} {
        if {[string trim [$text_field get "$test_row.0" "$test_row.end"]] != ""} {
          set is_within_limits false
          break
        }
        incr test_row -1
      }
    }

    if {$is_within_limits} {
      set prepend_stuff [list "$last_block_end.0" "\n"]
      return [list "$last_block_end.0" $prepend_stuff]
    }
  }

  # not a block near the end enough

  if {$omit_auto_tildes} {
    if {[lindex [split [$text_field index "end -1 chars"] "."] 1] != 0} {
      set prepend_stuff [list "end" "\n"]
    }
    return [list "end" $prepend_stuff]
  }

  set to_insert [strcat [string repeat "~" $min_tilde_count] "\n"]
  if {[lindex [split [$text_field index "end -1 chars"] "."] 1] != 0} {
    set to_insert [strcat "\n" $to_insert]
  }
  set prepend_stuff [list "end" $to_insert]
  return [list "end" $prepend_stuff]
}

proc prepare_cursor_wise_insertion_point_other {is_batch} {
  global text_field
  global omit_auto_tildes
  global min_tilde_count
  global text_inner_code_blocks
  global maxima_batch_asked_with_selections

  set prepare_info [split [$text_field index "insert"] "."]
  set selection_is_above false

  if {$is_batch && ([llength $maxima_batch_asked_with_selections]>0)} {
    set sel_start_pos_info [split [lindex $maxima_batch_asked_with_selections 0] "."]
    set sel_end_pos_info [split [lindex $maxima_batch_asked_with_selections "end"] "."]
    if {[llength $sel_end_pos_info]==2} {
      if {[lindex $sel_start_pos_info 0] < [lindex $sel_end_pos_info 0]} {
        set selection_is_above true
      }
      set prepare_info $sel_end_pos_info
    }
  }

  set prepare_row [lindex $prepare_info 0]
  set prepare_col [lindex $prepare_info 1]

  set insert_stuff [list]

  set to_move_below false
  if {$prepare_col!=0} {
    set to_move_below true
  }
  # if there was no above-selection used, and cursor at start of a section line,
  # should move line below even if at line start;
  if {(!$to_move_below) && (!$selection_is_above)} {
    if {[regexp -- [get_sec_regexp] [$text_field get "$prepare_row.0" "$prepare_row.end"]]} {
      set to_move_below true
    }
  }
  if {$to_move_below} {
    incr prepare_row
  }

  seek_inner_code

  set prev_pos_in_block false
  set curr_pos_in_block false
  set next_pos_in_block false
  set nextnext_pos_in_block false

  set prev_row [expr {$prepare_row - 1}]
  set next_row [expr {$prepare_row + 1}]
  set nextnext_row [expr {$prepare_row + 2}]

  foreach one_block $text_inner_code_blocks {
    if {$curr_pos_in_block} {
      break
    }

    set one_block_start [lindex $one_block 0]
    set one_block_end [lindex $one_block 1]

    if {($one_block_start < $prepare_row) && (($one_block_end=="end") || ($one_block_end >= $prepare_row))} {
      set curr_pos_in_block true
    }
    if {($one_block_start < $prev_row) && (($one_block_end=="end") || ($one_block_end >= $prev_row))} {
      set prev_pos_in_block true
    }
    if {($one_block_start < $next_row) && (($one_block_end=="end") || ($one_block_end >= $next_row))} {
      set next_pos_in_block true
    }
    if {($one_block_start < $nextnext_row) && (($one_block_end=="end") || ($one_block_end >= $nextnext_row))} {
      set nextnext_pos_in_block true
    }
  }

  set end_line_row [lindex [split [$text_field index "end -1 chars"] "."] 0]

  # if not in a block but a block nearby, switching into it;
  # if not ending in a block, plus if with ~-line enforcing, making a new block;
  if {$curr_pos_in_block} {
    if {$end_line_row < $prepare_row} {
      set insert_stuff [list "end" "\n"]
    } {
      if {[lindex [split [$text_field index "$prepare_row.end"] "."] 1]!=0} {
        set insert_stuff [list "$prepare_row.0" "\n"]
      }
    }
  } elseif {$next_pos_in_block} {
    set prepare_row $next_row
    if {$end_line_row < $prepare_row} {
      set insert_stuff [list "end" "\n"]
    } {
      if {[lindex [split [$text_field index "$prepare_row.end"] "."] 1]!=0} {
        set insert_stuff [list "$prepare_row.0" "\n"]
      }
    }
  } elseif {$prev_pos_in_block} {
    set prepare_row $prev_row
    if {$end_line_row < $prepare_row} {
      set insert_stuff [list "end" "\n"]
    } {
      if {[lindex [split [$text_field index "$prepare_row.end"] "."] 1]!=0} {
        set insert_stuff [list "$prepare_row.0" "\n"]
      }
    }
  } elseif {$nextnext_pos_in_block} {
    set prepare_row $nextnext_row
    if {$end_line_row < $prepare_row} {
      set insert_stuff [list "end" "\n"]
    } {
      if {[lindex [split [$text_field index "$prepare_row.end"] "."] 1]!=0} {
        set insert_stuff [list "$prepare_row.0" "\n"]
      }
    }
  } elseif {$omit_auto_tildes} {
    if {$end_line_row < $prepare_row} {
      set insert_stuff [list "end" "\n"]
    } {
      if {[lindex [split [$text_field index "$prepare_row.end"] "."] 1]!=0} {
        set insert_stuff [list "$prepare_row.0" "\n"]
      }
    }
  } {
    set insert_prepend ""
    set insert_stuff_index "$prepare_row.0"
    if {$end_line_row < $prepare_row} {
      set insert_prepend "\n"
      set insert_stuff_index "$prev_row.end"
    }
    set tilde_row [string repeat "~" $min_tilde_count]
    # putting in an empty line, since the output may end without newline;
    set insert_stuff [list $insert_stuff_index \
      [strcat $insert_prepend $tilde_row "\n\n" $tilde_row "\n"] \
    ]
    set prepare_row $next_row
  }

  return [list "$prepare_row.0" $insert_stuff]
}

set gnuplot_got_new_command false
proc exec_plot_from_text_field {with_control with_shift with_alt} {
  global text_field
  global gnuplot_handle
  global gnuplot_commands_to_write
  global gnuplot_got_new_command

  if {($with_control && $with_shift) || ($with_shift && $with_alt) || ($with_alt && $with_control)} {
    return
  }

  set command_text [string trim [get_selected_parts $text_field true true]]
  if {$command_text==""} {
    show_text_field_message_error [list [list "no text selected for plotting"]]
  }

  # checking alike at the "take_plot_command" function
  set plot_comments_strings_brackets [get_comments_strings_brackets_in_gnuplot $command_text]

  # check if sane text content
  set brackets_use [lindex $plot_comments_strings_brackets 2]
  set err_messages [test_brackets_sane $brackets_use false]

  if {[llength $err_messages]>0} {
    show_text_field_message_error $err_messages
    return
  }

  if {$with_alt} {
    show_text_field_message_error [list [list "simple parsing: correct"]]
  }

  # making sure the plot stuff is initialized

  if {![is_plot_setup]} {
    setup_plot_part
    close_plot_part true
  }

  # putting it into Gnuplot alike at the "exec_plot_command" function

  set command_text_info [insert_data_into_plot_command $command_text]
  if {([lindex $command_text_info 0]==false) || ([lindex $command_text_info 1]=="")} {
    show_text_field_message_error [list [list "can not use data for plotting"]]
    return
  }
  set command_text [lindex $command_text_info 1]

  set gnuplot_got_new_command true

  set to_restart $with_control
  if {![check_gnuplot_connected]} {
    set to_restart true
  }
  if {$to_restart} {
    # starting closes it too, thus without explicit closing here;
    start_gnuplot_direct
  }

  if {![check_gnuplot_connected]} {
    show_text_field_message_error [list [list "can not start gnuplot"]]
    return
  }

  lappend gnuplot_commands_to_write $command_text
  set bad [catch {
    fileevent $gnuplot_handle writable "write_to_gnuplot"
  }]
  if {$bad} {
    show_text_field_message_error [list [list "error at gnuplot running"]]
    return
  }

  return
}

proc text_field_on_return_for_tags {with_control with_shift with_alt} {
  # shift (without control) only locates,
  # control tries to do linked-tag jump;

  set failed_action_mssg ""

  if {($with_control) && (!$with_shift) && (!$with_alt)} {
    if {[is_batch_mode]} {
      target_within_tag_entries_on_key
    }
    if {[on_tag_link_try_to_jump_key true false]} {
      return [list true]
    }
    set failed_action_mssg "no tag link under cursor"
  }
  if {($with_control) && ($with_shift) && (!$with_alt)} {
    if {[is_batch_mode]} {
      target_within_tag_entries_on_key
    }
    if {[on_tag_link_try_to_jump_key true true]} {
      return [list true]
    }
    set failed_action_mssg "no tag link under cursor"
  }

  if {[is_batch_mode]} {
    if {(!$with_control) && ($with_shift) && (!$with_alt)} {
      if {[target_within_tag_entries_on_key]} {
        return [list true]
      }
      set failed_action_mssg "no (link to a) tag under cursor"
    }
  }

  return [list false $failed_action_mssg]
}

set mxmbatch_text_insert_mark_name "mxmbatch_text_insert"
set mxmbatch_text_command_is_one_line false
set mxmbatch_text_command_one_line_string ""
set maxima_batch_asked_with_selections [list]

proc text_field_on_return_in_batch_mode {with_control with_shift with_alt} {
  global text_trigger_plot_selection
  global text_trigger_use_text_selection
  global text_trigger_placed_text_selection
  global text_trigger_use_command
  global mxmbatch_input_field
  global batch_maxima_asks_names
  global batch_maxima_last_command
  global mxmbatch_text_insert_mark_name
  global text_field
  global mxmbatch_text_command_is_one_line
  global mxmbatch_text_command_one_line_string
  global maxima_batch_asked_to_write_to_text
  global maxima_batch_asked_with_selections
  global make_line_endings

  if {(!$with_control) && (!$with_shift) && (!$with_alt)} {
    return false
  }

  set command_string [string trim [get_selected_parts $text_field true true]]

  if {((!$text_trigger_use_text_selection) && (!$text_trigger_plot_selection)) || ($command_string=="")} {
    # when not using text selection for command
    set res_info [text_field_on_return_for_tags $with_control $with_shift $with_alt]

    if {[lindex $res_info 0]} {
      return true
    }

    set err_mssg [lindex $res_info 1]
    if {$text_trigger_use_text_selection || $text_trigger_plot_selection} {
      if {$err_mssg!=""} {
        append err_mssg ", and "
      }
      append err_mssg "no command text selected"
    }
    if {$err_mssg!=""} {
      show_text_field_message_error [list [list $err_mssg]]
    }
    return false
  }

  if {$text_trigger_plot_selection} {
    exec_plot_from_text_field $with_control $with_shift $with_alt
    return true
  }

  # actions regarding commands/batches

  if {![is_batch_maxima_set]} {
    setup_batch_maxima
    close_batch_maxima
  }

  set is_checking_only false
  if {(!$with_control) && (!$with_shift) && ($with_alt)} {
    set is_checking_only true
  }

  if {[is_batch_maxima_running] && (![is_batch_maxima_waiting]) && (!$is_checking_only)} {
    show_text_field_message_error [list [list "a Maxima batch is already running"]]
    return true
  }
  if {![is_mxmbatch_input_field_active]} {
    # this should not occur (when no batch is running)
    return true
  }

  if {$command_string==""} {
    show_text_field_message_error [list [list "no text selected"]]
    return true
  }

  set maxima_batch_asked_with_selections [$text_field tag ranges "sel"]
  set sel_end_pos_info [split [lindex $maxima_batch_asked_with_selections "end"] "."]
  if {[llength $sel_end_pos_info]!=2} {
    # this should not occur (when there is a selection)
    return true
  }

  set command_string_visible ""
  if {$make_line_endings} {
    set inner_part [trigger_command_prepare $command_string false false false]
    # this preserves the initial command form for being put into batch input (when requested);
    # notice that overwriting it below (for now) so that the processed form is visible somewhere;
    set command_string_visible $command_string
  } {
    set inner_part [trigger_command_prepare $command_string false false]
    set command_string [lindex $inner_part 1]
  }

  if {![lindex $inner_part 0]} {
    set err_mssgs [lindex $inner_part 1]
    if {!($err_mssgs in [list false "" [list]])} {
      show_text_field_message_error $err_mssgs
    }
    return true
  }

  if {$make_line_endings} {
    set comments_strings_brackets [lindex $inner_part 3]
    set command_string_info [put_maxima_command_ends_to_lines $command_string \
      [lindex $comments_strings_brackets 0] \
      [lindex $comments_strings_brackets 1] \
      [lindex $comments_strings_brackets 2] \
    ]

    if {(![lindex $command_string_info 0]) || ([lindex $command_string_info 1]=="")} {
      show_text_field_message_error [list [list "void command"]]
      return true
    }
    set command_string [lindex $command_string_info 1]
  }

  # checking only
  if {$is_checking_only} {
    show_text_field_message_error [list [list "simple parsing: correct"]]
    return true
  }

  # preparing real run

  set mxmbatch_text_command_is_one_line false
  set mxmbatch_text_command_one_line_string ""
  if {$with_alt} {
    if {[string first "\n" $command_string]==-1} {
      set mxmbatch_text_command_is_one_line true
      set mxmbatch_text_command_one_line_string $command_string
      if {[string range $mxmbatch_text_command_one_line_string "end" "end"] in [list ";" "\$"]} {
        set mxmbatch_text_command_one_line_string [string range $mxmbatch_text_command_one_line_string 0 "end-1"]
      }
    }
  }

  if {!$make_line_endings} {
    set command_string_visible [lindex $inner_part 1]
    set command_string $command_string_visible

    if {([llength $inner_part]>2) && ([lindex $inner_part 2]!="")} {
      set command_string [lindex $inner_part 2]
    }
  } {
    # this sets the processed command form for being put into batch input (when requested);
    # and since it is a way to actually see it, using it here now;
    set command_string_visible $command_string
  }

  # to set the running mode;

  set batch_maxima_asks_names false

  if {$with_control} {
    set batch_maxima_last_command [strcat $command_string_visible "\n"]
    if {$with_shift} {
      $mxmbatch_input_field replace "1.0" "end" [strcat $command_string_visible "\n"]
    }
    update_batch_maxima_properties
  }

  $text_field mark unset $mxmbatch_text_insert_mark_name
  set maxima_batch_asked_to_write_to_text false

  if {$text_trigger_placed_text_selection} {
    # since it may end up with no actual output, should do the initialization stuff
    # only if (eventually) having something to write in;
    # ...having to keep the mark unset, so that we (later) know that we should
    #    do the position finding/preparation then;
    # but we loose the initial selection,
    # thus having to preserve the info on it
    # (though vanishing it on doc load/empty);
    set maxima_batch_asked_to_write_to_text true
  }

  set use_trigger [expr {[is_batch_maxima_waiting] ? "continue_batch_maxima" : "start_batch_maxima"}]
  if {[$use_trigger false false $command_string]} {
    $text_field tag remove "sel" "1.0" "end"
  }

  after 0 "update_text_properties"
  return true

}

proc can_use_stack_command_from_text {} {
  global text_trigger_use_command
  global test_command_whole_then

  if {![is_stack_mode]} {
    return false
  }

  if {!$text_trigger_use_command} {
    return false
  }

  if {[string trim [get_command_content [expr {!$test_command_whole_then}]]] == ""} {
    return false
  }

  return true
}

# control/shift-based ways when using text-wise selections in the stack mode:
# control = copy-selection,
# shift = stack-remains-at-its-current-level (after the copy, if it happens);
# that is:
# shift+return: without putting it in command (stack without change alike command-wise shift+return),
# control+return: putting it at the top of stack (ending with empty top stack slot, input beneath, i.e.~as if command-wise),
# control+shift+return: should be stack-top and remaining there (more similar to command-wise control+shift+return);

set current_command_placed_beneath false
proc text_field_on_return_in_stack_mode {with_control with_shift with_alt} {
  global text_trigger_use_text_selection
  global text_trigger_placed_text_selection
  global text_trigger_use_command
  global current_command_placed_beneath
  global include_maxima_input_default
  global include_maxima_input
  global curr_stack_mode
  global command_field
  global text_field
  global make_line_endings
  global in_lisp_mode
  global in_debug_mode

  set use_make_line_endings $make_line_endings
  if {$in_lisp_mode || $in_debug_mode} {
    set use_make_line_endings false
  }

  set can_delegate_to_stack [can_use_stack_command_from_text]

  # without removed (links to) tags from the selected region;
  set command_string [string trim [get_selected_parts $text_field true true]]

  if {(!$can_delegate_to_stack) && ((!$text_trigger_use_text_selection) || ($command_string==""))} {
    # if first, not delegating to command, and second, no/nothing-for text-wise trigger,
    # trying action on (links to) tags;
    set res_info [text_field_on_return_for_tags $with_control $with_shift $with_alt]

    if {[lindex $res_info 0]} {
      return true
    }

    set err_mssg [lindex $res_info 1]
    if {$text_trigger_use_text_selection || $text_trigger_use_command} {
      if {$err_mssg!=""} {
        append err_mssg ", and "
      }
      if {$text_trigger_use_text_selection && $text_trigger_use_command} {
        append err_mssg "neither (command) text selected nor command available"
      } elseif {$text_trigger_use_text_selection} {
        append err_mssg "no (command) text selected"
      } {
        append err_mssg "no command available"
      }
    }
    if {$err_mssg!=""} {
      show_text_field_message_error [list [list $err_mssg]]
    }

    # here always returning "true" so that it is not re-delegated to command-wise actions;
    return true
  }

  if {!$text_trigger_use_text_selection} {
    # not dealing with text parts if asked to avoid it
    return false
  }

  # whether should put it to the command-wise way
  if {$command_string==""} {
    if {!$text_trigger_use_command} {
      # this sub-case should not occur, since (above) going into tag-wise actions in such cases;
      show_text_field_message_error [list [list "no text selected"]]
      return true
    }
    # though this case can occur;
    return false
  }

  # doing it the text-wise ways

  # first checking if the selected text can be used at all;
  set command_string_visible ""
  if {$use_make_line_endings} {
    set inner_part [trigger_command_prepare $command_string false false false]
    set command_string_visible $command_string
  } {
    set inner_part [trigger_command_prepare $command_string false false]
    set command_string_visible [lindex $inner_part 1]
    set command_string $command_string_visible
  }

  if {![lindex $inner_part 0]} {
    set err_mssgs [lindex $inner_part 1]
    if {!($err_mssgs in [list false "" [list]])} {
      show_text_field_message_error $err_mssgs
    }
    return true
  }

  if {$use_make_line_endings} {
    set comments_strings_brackets [lindex $inner_part 3]
    set command_string_info [put_maxima_command_ends_to_lines $command_string \
      [lindex $comments_strings_brackets 0] \
      [lindex $comments_strings_brackets 1] \
      [lindex $comments_strings_brackets 2] \
    ]

    if {(![lindex $command_string_info 0]) || ([lindex $command_string_info 1]=="")} {
      show_text_field_message_error [list [list "void command"]]
      return true
    }
    set command_string [lindex $command_string_info 1]
  }

  # checking mode only does the checking;
  if {(!$with_control) && (!$with_shift) && ($with_alt)} {
    show_text_field_message_error [list [list "simple parsing: correct"]]
    return true
  }

  if {!$use_make_line_endings} {
    if {([llength $inner_part]>2) && ([lindex $inner_part 2]!="")} {
      set command_string [lindex $inner_part 2]
    }
  }

  # in-place (that is without copying the command into stack) mode;
  if {(!$with_control) && ($with_shift)} {
    set include_maxima_input $include_maxima_input_default
    if {$with_alt} {
      set include_maxima_input [expr {$include_maxima_input_default ? false : true}]
    }
    # leaving stack intact;
    set curr_stack_mode 0
    # deciding on where to place output;
    set current_command_placed_beneath [expr {$text_trigger_placed_text_selection ? true : false}]
    if {[exec_max_command $command_string_visible $command_string]} {
      $text_field tag remove "sel" "1.0" "end"
    }
    return true
  }

  # mode with putting command at the top of stack,
  # and trying to avoid over-growth (that would occur under repeating control+shift action with copy without checks);
  if {$with_control} {
    # passing to the top where the text-wise command is to be put;
    move_in_stack_to_top
    # checking if have to put the command into new slot (that is if the top parts are different);
    if {[string trim [get_command_content false]]==""} {
      # looking below (if there is some below)
      if {[get_stack_size]>1} {
        # empty-like top, going to the level below the top, checking it;
        pass_in_stack -1
        if {[string trim [get_command_content false]]!=$command_string} {
          # if having a different command, having to put it into the top level (that was empty-like here);
          pass_in_stack 1
        }
      }
    } {
      if {[string trim [get_command_content false]]!=$command_string} {
        # if having a different command, having to put it into a new level (since the top one is not empty-like here);
        pass_in_stack_one_up
      }
    }

    # copying the command in there
    $command_field replace "1.0" "end" [strcat $command_string "\n"] [list]
    # updating the info presentating
    update_stack_label
    update_cmmd_highlight_lines
    update_cmmd_strings_comments
    # setting the stack mode for the command
    set curr_stack_mode [expr {$with_shift ? 2 : 1}]
    # deciding on where to place output;
    set current_command_placed_beneath [expr {$text_trigger_placed_text_selection ? true : false}]
    if {[exec_max_command $command_string_visible $command_string]} {
      $text_field tag remove "sel" "1.0" "end"
    }
    return true
  }

  # here only if (!$with_control) && (!$with_shift) && (!$with_alt)
  # this function should not get called without any of them;
  return false
}

# this for (quite simple) checking only
bind $command_field <Alt-Return> {
  if {[is_command_field_active]} {
    trigger_command_check
  }
  break
}
bind $text_field <Alt-Return> {
  if {[is_stack_mode]} {
    if {[is_command_field_active]} {
      if {![text_field_on_return_in_stack_mode false false true]} {
        trigger_command_check
      }
    }
  } {
    text_field_on_return_in_batch_mode false false true
  }
  break
}

# regular command trigger, the "move" way;
bind $command_field <Control-Return> {
  if {[is_command_field_active]} {
    trigger_command_stack_wise false
  }
  break
}
bind $text_field <Control-Return> {
  if {[is_stack_mode]} {
    if {[is_command_field_active]} {
      if {![text_field_on_return_in_stack_mode true false false]} {
        trigger_command_stack_wise false
      }
    }
  } {
    text_field_on_return_in_batch_mode true false false
  }
  break
}
# Alt reverts the (default) way of input outputting;
bind $command_field <Control-Alt-Return> {
  if {[is_command_field_active]} {
    trigger_command_stack_wise false true
  }
  break
}
bind $text_field <Control-Alt-Return> {
  if {[is_stack_mode]} {
    if {[is_command_field_active]} {
      if {![text_field_on_return_in_stack_mode true false true]} {
        trigger_command_stack_wise false true
      }
    }
  } {
    text_field_on_return_in_batch_mode true false true
  }
  break
}

# <Shift-Return> command triggers for action without move in stack
bind $command_field <Shift-Return> {
  if {[is_command_field_active]} {
    trigger_command_local_wise
  }
  break
}
# may be to alter text_field-based trigger for considering selections within text_field;
bind $text_field <Shift-Return> {
  if {[is_stack_mode]} {
    if {[is_command_field_active]} {
      if {![text_field_on_return_in_stack_mode false true false]} {
        trigger_command_local_wise
      }
    }
  } {
    text_field_on_return_in_batch_mode false true false
  }
  break
}
# Alt reverts the (default) way of input outputting;
bind $command_field <Shift-Alt-Return> {
  if {[is_command_field_active]} {
    trigger_command_local_wise true
  }
  break
}
bind $text_field <Shift-Alt-Return> {
  if {[is_stack_mode]} {
    if {[is_command_field_active]} {
      if {![text_field_on_return_in_stack_mode false true true]} {
        trigger_command_local_wise true
      }
    }
  } {
    text_field_on_return_in_batch_mode false true true
  }
  break
}

# regular command trigger, the "copy" way;
bind $command_field <Control-Shift-Return> {
  if {[is_command_field_active]} {
    trigger_command_stack_wise true
  }
  break
}
bind $text_field <Control-Shift-Return> {
  if {[is_stack_mode]} {
    if {[is_command_field_active]} {
      if {![text_field_on_return_in_stack_mode true true false]} {
        trigger_command_stack_wise true
      }
    }
  } {
    text_field_on_return_in_batch_mode true true false
  }
  break
}

# Alt reverts the (default) way of input outputting;
bind $command_field <Control-Shift-Alt-Return> {
  if {[is_command_field_active]} {
    trigger_command_stack_wise true true
  }
  break
}
bind $text_field <Control-Shift-Alt-Return> {
  if {[is_stack_mode]} {
    if {[is_command_field_active]} {
      if {![text_field_on_return_in_stack_mode true true true]} {
        trigger_command_stack_wise true true
      }
    }
  } {
    text_field_on_return_in_batch_mode true true true
  }
  break
}

foreach wdg [list $topw $text_field $command_field $search_entry $tags_decls_line $tags_links_line] {
  # simple text export: into regular file (thus with asking)
  bind $wdg <Control-t> {
    draw_export_as
    break
  }
  # simple text-to-html export,
  # the (regular) exported html file is loaded at browser
  bind $wdg <Control-T> {
    if {[$text_field index "end -1 chars"]!="1.0"} {
      draw_export_html true
    }
    break
  }
}

# simple text-to-html export: into clipboard or tmp file
bind $text_field <Control-Alt-t> {
  if {[is_wine_like]} {
    trigger_text_export_tmp true false
  } {
    trigger_text_export_tmp true false
  }
  break
}
bind $text_field <Control-Alt-T> {
  if {[is_wine_like]} {
    trigger_text_export_tmp true false
  } {
    trigger_text_export_tmp true false
  }
  break
}
bind $command_field <Control-Alt-t> {
  if {[is_wine_like]} {
    trigger_text_export_tmp true false
  } {
    trigger_text_export_tmp true false
  }
  break
}
bind $command_field <Control-Alt-T> {
  if {[is_wine_like]} {
    trigger_text_export_tmp true false
  } {
    trigger_text_export_tmp true false
  }
  break
}

set plot_field_message_regions [list]
proc show_plot_select_on_message {idx} {
  global plot_field_message_regions
  global plot_input_field
  global gnuplot_at_data_view
  global gnuplot_at_output_view

  if {($idx<0) || ($idx>=[llength $plot_field_message_regions])} {
    return
  }

  set region_info [lindex $plot_field_message_regions $idx]
  if {[llength $region_info]!=2} {
    return
  }

  if {$gnuplot_at_data_view || $gnuplot_at_output_view} {
    present_gnuplot_input
    update
  }

  $plot_input_field tag remove "sel" "1.0" "end"
  $plot_input_field tag add "sel" [lindex $region_info 0] [lindex $region_info 1]
  # since not setting the insert mark, not dealing with anchor either here
  #set_anchor_position $plot_input_field [lindex $region_info 0]
  set_anchor_insert $plot_input_field
  $plot_input_field see [lindex $region_info 0]
  if {[lindex $region_info 1]!="end"} {
    $plot_input_field see [lindex $region_info 1]
  }
  focus $plot_input_field
}

set plot_field_message ""
proc show_plot_field_message_error {mssg_list} {
  global plot_put_window
  global plot_field_message
  global topw
  global plot_field_message_regions
  global plot_input_field
  global plot_data_field
  global plot_output_field
  global gnuplot_at_data_view
  global gnuplot_at_output_view

  set use_plot_field $plot_input_field
  if {$gnuplot_at_data_view} {
    set use_plot_field $plot_data_field
  }
  if {$gnuplot_at_output_view} {
    set use_plot_field $plot_output_field
  }

  set plot_field_message_regions [list]

  if {[llength $mssg_list]==0} {
    return
  }

  catch {destroy $plot_field_message}
  catch {show_plot_part}
  catch {wm deiconify $plot_put_window}
  update

  set abs_pos [get_plot_field_position]
  set pos_x [expr {10 + [lindex $abs_pos 0]}]
  set pos_y [expr {5 + [lindex $abs_pos 1]}]

  set bgcolor "#ffc080"
  set bgcolor_act "#f0a070"
  set plot_field_message [menu "$topw.plotmessagemenu" -tearoff false]

  $plot_field_message configure -relief solid -borderwidth 1 -background $bgcolor -disabledforeground "#000000"
  $plot_field_message configure -activeforeground "#000000" -foreground "#000000" -activebackground $bgcolor_act
  $plot_field_message configure -activeborderwidth 1

  set curr_idx 0
  foreach mssg_info $mssg_list {
    set mssg [lindex $mssg_info 0]
    if {[llength $mssg_info]==1} {
      lappend plot_field_message_regions [list]
      $plot_field_message add command -label $mssg -state normal -background $bgcolor -activebackground $bgcolor_act \
        -command [list focus $use_plot_field]
    } {
      lappend plot_field_message_regions [lindex $mssg_info 1]
      $plot_field_message add command -label $mssg -state normal -background $bgcolor -activebackground $bgcolor_act \
        -command [list "show_plot_select_on_message" $curr_idx]
    }
    incr curr_idx
  }

  catch {
    tk_popup $plot_field_message $pos_x $pos_y
  }
}

proc dispatch_message_targets {target} {
  if {$target=="gnuplot-output"} {
    catch {show_plot_part}
    catch {present_gnuplot_output}
  }
}

set text_field_message ""
proc show_text_field_message_error {mssg_list} {
  global text_field
  global text_field_message
  global topw

  if {[llength $mssg_list]==0} {
    return
  }

  catch {destroy $text_field_message}
  catch {wm deiconify $topw}
  update

  set offset_y_sep 5
  set offset_y_ini 20
  set offset_y $offset_y_ini
  set offset_y_check [expr {$offset_y + $offset_y_sep}]
  set cursor_pos [$text_field bbox "insert"]
  if {[llength $cursor_pos] == 4} {
    set cursor_top [lindex $cursor_pos 1]
    set cursor_bottom [expr {[lindex $cursor_pos 1] + [lindex $cursor_pos 3]}]
    set cursor_line_height [lindex $cursor_pos 3]
    if {$cursor_top < [expr {$offset_y_check + $cursor_line_height}]} {
      if {$cursor_top > [expr {(2*$offset_y_sep) + $cursor_line_height}]} {
        set offset_y $offset_y_sep
      } {
        set offset_y [expr {$cursor_bottom + $offset_y_sep}]
        if {$offset_y < $offset_y_ini} {
          set offset_y $offset_y_ini
        }
      }
    }
  }

  set abs_pos [get_text_field_position]
  set pos_x [expr {120 + [lindex $abs_pos 0]}]
  set pos_y [expr {$offset_y + [lindex $abs_pos 1]}]

  set bgcolor "#ffc080"
  set bgcolor_act "#f0a070"
  set text_field_message [menu "$topw.textmessagemenu" -tearoff false]

  $text_field_message configure -relief solid -borderwidth 1 -background $bgcolor -disabledforeground "#000000"
  $text_field_message configure -activeforeground "#000000" -foreground "#000000" -activebackground $bgcolor_act
  $text_field_message configure -activeborderwidth 1

  foreach mssg $mssg_list {
    set set_command [list focus $text_field]
    if {[llength $mssg] > 1} {
      set set_command [list "dispatch_message_targets" [lindex $mssg 1]]
    }
    $text_field_message add command -label [lindex $mssg 0] -state normal -background $bgcolor -activebackground $bgcolor_act \
      -command $set_command
  }

  catch {
    tk_popup $text_field_message $pos_x $pos_y
  }
}

set command_field_message_regions [list]
proc show_command_select_on_message {idx} {
  global command_field_message_regions
  global command_field

  if {($idx<0) || ($idx>=[llength $command_field_message_regions])} {
    return
  }

  set region_info [lindex $command_field_message_regions $idx]
  if {[llength $region_info]!=2} {
    return
  }

  $command_field tag remove "sel" "1.0" "end"
  $command_field tag add "sel" [lindex $region_info 0] [lindex $region_info 1]
  # since not setting the insert mark, not dealing with anchor either here
  #set_anchor_position $command_field [lindex $region_info 0]
  set_anchor_insert $command_field
  $command_field see [lindex $region_info 0]
  if {[lindex $region_info 1]!="end"} {
    $command_field see [lindex $region_info 1]
  }
  focus $command_field
}

set command_field_message ""
proc show_command_field_message_error {mssg_list} {
  global command_field_message
  global topw
  global command_field
  global command_field_message_regions

  set command_field_message_regions [list]

  if {[llength $mssg_list]==0} {
    return
  }

  catch {destroy $command_field_message}
  catch {wm deiconify $topw}
  update

  set abs_pos [get_command_field_position]
  set pos_x [expr {5 + [lindex $abs_pos 0]}]
  set pos_y [expr {5 + [lindex $abs_pos 1]}]

  set bgcolor "#ffc080"
  set bgcolor_act "#f0a070"
  set command_field_message [menu "$topw.commandmessagemenu" -tearoff false]

  $command_field_message configure -relief solid -borderwidth 1 -background $bgcolor -disabledforeground "#000000"
  $command_field_message configure -activeforeground "#000000" -foreground "#000000" -activebackground $bgcolor_act
  $command_field_message configure -activeborderwidth 1

  set curr_idx 0
  foreach mssg_info $mssg_list {
    set mssg [lindex $mssg_info 0]
    if {[llength $mssg_info]==1} {
      lappend command_field_message_regions [list]
      $command_field_message add command -label $mssg -state normal -background $bgcolor -activebackground $bgcolor_act \
        -command [list focus $command_field]
    } {
      lappend command_field_message_regions [lindex $mssg_info 1]
      $command_field_message add command -label $mssg -state normal -background $bgcolor -activebackground $bgcolor_act \
        -command [list "show_command_select_on_message" $curr_idx]
    }
    incr curr_idx
  }

  catch {
    tk_popup $command_field_message $pos_x $pos_y
  }
}

proc show_batch_maxima_command_select_on_message {idx} {
  global batch_maxima_message_regions
  global mxmbatch_input_field

  if {($idx<0) || ($idx>=[llength $batch_maxima_message_regions])} {
    return
  }

  set region_info [lindex $batch_maxima_message_regions $idx]
  if {[llength $region_info]!=2} {
    return
  }

  $mxmbatch_input_field tag remove "sel" "1.0" "end"
  $mxmbatch_input_field tag add "sel" [lindex $region_info 0] [lindex $region_info 1]
  # since not setting the insert mark, not dealing with anchor either here
  #set_anchor_position $mxmbatch_input_field [lindex $region_info 0]
  set_anchor_insert $mxmbatch_input_field
  $mxmbatch_input_field see [lindex $region_info 0]
  if {[lindex $region_info 1]!="end"} {
    $mxmbatch_input_field see [lindex $region_info 1]
  }
  focus $mxmbatch_input_field
}

set batch_maxima_message ""
set batch_maxima_message_regions [list]
proc show_batch_maxima_message_error {mssg_list} {
  global batch_maxima_window
  global mxmbatch_input_field
  global batch_maxima_message
  global topw
  global batch_maxima_message_regions

  if {$mxmbatch_input_field==""} {
    return false
  }

  set batch_maxima_message_regions [list]

  if {[llength $mssg_list]==0} {
    return
  }

  catch {destroy $batch_maxima_message}
  catch {draw_batch_maxima}
  catch {wm deiconify $batch_maxima_window}
  update

  set abs_pos [get_maxima_batch_input_position]
  set pos_x [expr {5 + [lindex $abs_pos 0]}]
  set pos_y [expr {5 + [lindex $abs_pos 1]}]

  set bgcolor "#ffc080"
  set bgcolor_act "#f0a070"
  set batch_maxima_message [menu "$topw.batchmaximamessagemenu" -tearoff false]

  $batch_maxima_message configure -relief solid -borderwidth 1 -background $bgcolor -disabledforeground "#000000"
  $batch_maxima_message configure -activeforeground "#000000" -foreground "#000000" -activebackground $bgcolor_act
  $batch_maxima_message configure -activeborderwidth 1

  set curr_idx 0
  foreach mssg_info $mssg_list {
    set mssg [lindex $mssg_info 0]
    if {[llength $mssg_info]==1} {
      lappend batch_maxima_message_regions [list]
      $batch_maxima_message add command -label $mssg -state normal -background $bgcolor -activebackground $bgcolor_act \
        -command [list focus $mxmbatch_input_field]
    } {
      lappend batch_maxima_message_regions [lindex $mssg_info 1]
      $batch_maxima_message add command -label $mssg -state normal -background $bgcolor -activebackground $bgcolor_act \
        -command [list "show_batch_maxima_command_select_on_message" $curr_idx]
    }
    incr curr_idx
  }

  catch {
    tk_popup $batch_maxima_message $pos_x $pos_y
  }
}

proc show_import_input_command_select_on_message {idx} {
  global import_input_message_regions
  global import_input_field

  if {($idx<0) || ($idx>=[llength $import_input_message_regions])} {
    return
  }

  set region_info [lindex $import_input_message_regions $idx]
  if {[llength $region_info]!=2} {
    return
  }

  $import_input_field tag remove "sel" "1.0" "end"
  $import_input_field tag add "sel" [lindex $region_info 0] [lindex $region_info 1]
  # since not setting the insert mark, not dealing with anchor either here
  #set_anchor_position $import_input_field [lindex $region_info 0]
  set_anchor_insert $import_input_field
  $import_input_field see [lindex $region_info 0]
  if {[lindex $region_info 1]!="end"} {
    $import_input_field see [lindex $region_info 1]
  }
  focus $import_input_field
}

set import_input_message ""
set import_input_message_regions [list]
proc show_import_input_message_error {mssg_list} {
  global import_window
  global import_input_field
  global import_input_message
  global topw
  global import_input_message_regions

  if {$import_input_field==""} {
    return false
  }

  set import_input_message_regions [list]

  if {[llength $mssg_list]==0} {
    return
  }

  catch {destroy $import_input_message}
  catch {draw_import}
  catch {wm deiconify $import_window}
  update

  set abs_pos [get_import_input_position]
  set pos_x [expr {5 + [lindex $abs_pos 0]}]
  set pos_y [expr {5 + [lindex $abs_pos 1]}]

  set bgcolor "#ffc080"
  set bgcolor_act "#f0a070"
  set import_input_message [menu "$topw.importinputmessagemenu" -tearoff false]

  $import_input_message configure -relief solid -borderwidth 1 -background $bgcolor -disabledforeground "#000000"
  $import_input_message configure -activeforeground "#000000" -foreground "#000000" -activebackground $bgcolor_act
  $import_input_message configure -activeborderwidth 1

  set curr_idx 0
  foreach mssg_info $mssg_list {
    set mssg [lindex $mssg_info 0]
    if {[llength $mssg_info]==1} {
      lappend import_input_message_regions [list]
      $import_input_message add command -label $mssg -state normal -background $bgcolor -activebackground $bgcolor_act \
        -command [list focus $import_input_field]
    } {
      lappend import_input_message_regions [lindex $mssg_info 1]
      $import_input_message add command -label $mssg -state normal -background $bgcolor -activebackground $bgcolor_act \
        -command [list "show_import_input_command_select_on_message" $curr_idx]
    }
    incr curr_idx
  }

  catch {
    tk_popup $import_input_message $pos_x $pos_y
  }
}

proc test_brackets_sane {brackets_use {mssgs_with_locations true} {mssg_start ""}} {
  global bracket_complements

  set err_messages [list]

  set brackets_matched true
  foreach bracket_pair [list [list "\{" "\}"] [list "\[" "\]"] [list "(" ")"]] {
    set test_stack [list]
    set bracket_open [lindex $bracket_pair 0]
    set bracket_close [lindex $bracket_pair 1]
    set bracket_correct true
    set unmatched_form ""
    set unmatched_start ""
    set unmatched_end ""
    foreach one_bracket_info $brackets_use {
      set one_bracket [lindex $one_bracket_info 2]
      if {!($one_bracket in $bracket_pair)} {
        continue
      }
      if {$one_bracket == $bracket_open} {
        lappend test_stack [lrange $one_bracket_info 0 1]
      } else {
        if {[llength $test_stack]>0} {
          set test_stack [lreplace $test_stack "end" "end"]
        } {
          set bracket_correct false
          set unmatched_form $bracket_close
          set unmatched_start "1.0"
          set unmatched_end [join [list [lindex $one_bracket_info 0] [expr {[lindex $one_bracket_info 1]+1}]] "."]
          break
        }
      }
    }
    if {$bracket_correct && ([llength $test_stack]>0)} {
      set bracket_correct false
      set unmatched_form $bracket_open
      set unmatched_start [join [lindex $test_stack "end"] "."]
      set unmatched_end "end"
    }
    if {!$bracket_correct} {
      if {$mssgs_with_locations} {
        lappend err_messages [list [strcat $mssg_start "unmatched bracket \"$unmatched_form\""] [list $unmatched_start $unmatched_end]]
      } {
        lappend err_messages [list [strcat $mssg_start "unmatched bracket \"$unmatched_form\""]]
      }
      set brackets_matched false
    }
  }

  # test overall mismatches
  if {$brackets_matched} {
    set brackets_open [list "\{" "\[" "("]
    set test_stack [list]
    set bracket_correct true
    set bracket_mismatch_form ""
    set bracket_mismatch_start ""
    set bracket_mismatch_end ""
    foreach one_bracket_info $brackets_use {
      set one_bracket [lindex $one_bracket_info 2]
      if {$one_bracket in $brackets_open} {
        lappend test_stack $one_bracket_info
      } else {
        set one_bracket_opposite $bracket_complements($one_bracket)
        set facing_bracket_info [lindex $test_stack "end"]
        if {([llength $test_stack]>0) && ([lindex $facing_bracket_info 2]==$one_bracket_opposite)} {
          set test_stack [lreplace $test_stack "end" "end"]
        } {
          set bracket_correct false
          append mismatch_form [lindex $facing_bracket_info 2] $one_bracket
          set mismatch_start [join [lrange $facing_bracket_info 0 1] "."]
          set mismatch_end [join [list [lindex $one_bracket_info 0] [expr {[lindex $one_bracket_info 1]+1}]] "."]
          if {$mssgs_with_locations} {
            lappend err_messages [list [strcat $mssg_start "mismatched brackets: \"$mismatch_form\""] [list $mismatch_start $mismatch_end]]
          } {
            lappend err_messages [list [strcat $mssg_start "mismatched brackets: \"$mismatch_form\""]]
          }
          break
        }
      }
    }
    if {$bracket_correct && ([llength $test_stack]>0)} {
      # should not get here, it should already be due to a mismatch above
      set bracket_correct false
      lappend err_messages [list [strcat $mssg_start "mismatched brackets"]]
    }
  }

  return $err_messages
}

set include_maxima_input $include_maxima_input_default

# 0: stay(shift), 1: move(ctrl), 2: copy(ctrl/shift)
set curr_stack_mode 1

proc trigger_command_stack_wise {{with_copy false} {altered_input_shown false}} {
  global include_maxima_input_default
  global include_maxima_input
  global curr_stack_mode
  global test_command_select_first
  global test_command_whole_then
  global current_command_placed_beneath
  global make_line_endings
  global in_lisp_mode
  global in_debug_mode

  set use_make_line_endings $make_line_endings
  if {$in_lisp_mode || $in_debug_mode} {
    set use_make_line_endings false
  }

  set current_command_placed_beneath false

  set include_maxima_input $include_maxima_input_default
  if {$altered_input_shown} {
    set include_maxima_input [expr {$include_maxima_input_default ? false : true}]
  }

  set curr_stack_mode [expr {$with_copy ? 2 : 1}]

  set to_check_borders [expr {$use_make_line_endings ? false : true}]
  set inner_part [trigger_command_prepare "" $test_command_select_first $test_command_whole_then $to_check_borders]
  if {![lindex $inner_part 0]} {
    set err_mssgs [lindex $inner_part 1]
    if {!($err_mssgs in [list false "" [list]])} {
      show_command_field_message_error $err_mssgs
    }
    return false
  }

  if {$use_make_line_endings} {
    set command_string_base [lindex $inner_part 1]
    set comments_strings_brackets [lindex $inner_part 3]
    set command_string_info [put_maxima_command_ends_to_lines $command_string_base \
      [lindex $comments_strings_brackets 0] \
      [lindex $comments_strings_brackets 1] \
      [lindex $comments_strings_brackets 2] \
    ]

    if {(![lindex $command_string_info 0]) || ([lindex $command_string_info 1]=="")} {
      show_text_field_message_error [list [list "void command"]]
      return false
    }
    set inner_part [list true $command_string_base [lindex $command_string_info 1]]
  }

  return [exec_max_command [lindex $inner_part 1] [lindex $inner_part 2]]
}

proc trigger_command_local_wise {{altered_input_shown false}} {
  global include_maxima_input_default
  global include_maxima_input
  global curr_stack_mode
  global test_command_select_first
  global test_command_whole_then
  global current_command_placed_beneath
  global make_line_endings
  global in_lisp_mode
  global in_debug_mode

  set use_make_line_endings $make_line_endings
  if {$in_lisp_mode || $in_debug_mode} {
    set use_make_line_endings false
  }

  set current_command_placed_beneath false

  set include_maxima_input $include_maxima_input_default
  if {$altered_input_shown} {
    set include_maxima_input [expr {$include_maxima_input_default ? false : true}]
  }

  set curr_stack_mode 0

  set inner_part [trigger_command_prepare "" $test_command_select_first $test_command_whole_then]
  if {![lindex $inner_part 0]} {
    set err_mssgs [lindex $inner_part 1]
    if {!($err_mssgs in [list false "" [list]])} {
      show_command_field_message_error $err_mssgs
    }
    return false
  }

  if {$use_make_line_endings} {
    set command_string_base [lindex $inner_part 1]
    set comments_strings_brackets [lindex $inner_part 3]
    set command_string_info [put_maxima_command_ends_to_lines $command_string_base \
      [lindex $comments_strings_brackets 0] \
      [lindex $comments_strings_brackets 1] \
      [lindex $comments_strings_brackets 2] \
    ]

    if {(![lindex $command_string_info 0]) || ([lindex $command_string_info 1]=="")} {
      show_text_field_message_error [list [list "void command"]]
      return false
    }
    set inner_part [list true $command_string_base [lindex $command_string_info 1]]
  }

  return [exec_max_command [lindex $inner_part 1] [lindex $inner_part 2]]
}

proc trigger_command_check {} {
  global test_command_select_first
  global test_command_whole_then
  global current_command_placed_beneath
  global make_line_endings
  global in_lisp_mode
  global in_debug_mode

  set use_make_line_endings $make_line_endings
  if {$in_lisp_mode || $in_debug_mode} {
    set use_make_line_endings false
  }

  set current_command_placed_beneath false

  set inner_part [trigger_command_prepare "" $test_command_select_first $test_command_whole_then]
  if {![lindex $inner_part 0]} {
    set err_mssgs [lindex $inner_part 1]
    if {!($err_mssgs in [list false "" [list]])} {
      show_command_field_message_error $err_mssgs
    }
    return false
  }

  if {$use_make_line_endings} {
    set command_string_base [lindex $inner_part 1]
    set comments_strings_brackets [lindex $inner_part 3]
    set command_string_info [put_maxima_command_ends_to_lines $command_string_base \
      [lindex $comments_strings_brackets 0] \
      [lindex $comments_strings_brackets 1] \
      [lindex $comments_strings_brackets 2] \
    ]

    if {(![lindex $command_string_info 0]) || ([lindex $command_string_info 1]=="")} {
      show_text_field_message_error [list [list "void command"]]
      return false
    }
    set inner_part [list true $command_string_base [lindex $command_string_info 1]]
  }

  show_command_field_message_error [list [list "simple parsing: correct"]]
  return true
}

# at some point we should try to split the provided input into distinct commands
# (or at least to evaluate the expected count of eval-time info that we should read),
# so that we do not offer to deal with Maxima input/output when it computes some parts yet;
# it would be quite involved though, since we would need to guess whether it is/will be
# in Lisp mode too (where we get "Maxima>" prompt instead of eval-time info);
proc trigger_command_prepare {{command_string ""} {test_selects false} {test_whole true} {check_borders true}} {
  global bracket_complements
  global in_lisp_mode
  global in_debug_mode

  set use_command ""
  set using_selections false

  if {$command_string!=""} {
    set use_command $command_string
  }
  if {$use_command!=""} {
    set using_selections true
  } {
    if {$test_selects} {
      set use_command [get_command_content true]
      if {$use_command!=""} {
        set using_selections true
      }
    }
    if {!$using_selections} {
      if {$test_whole} {
        set use_command [get_command_content false]
      }
    }
  }

  if {[string trim $use_command]==""} {
    return [list false [list [list "empty command"]]]
  }

  set is_admissible true
  # when in lisp mode, comments are based on semicolon,
  # and the brackets are ordinary chars when after backslash;
  if {(!$in_lisp_mode) && (!$in_debug_mode)} {
    set comments_strings_brackets [get_comments_strings_brackets $use_command]
  } {
    set comments_strings_brackets [get_comments_strings_brackets_in_lisp $use_command]
  }

  set brackets_use [lindex $comments_strings_brackets 2]
  set err_messages [test_brackets_sane $brackets_use [expr {!$using_selections}]]
  if {[llength $err_messages]>0} {
    set is_admissible false
  }

  set comment_last [lindex [lindex $comments_strings_brackets 0] "end"]
  if {[llength $comment_last]>1} {
    if {[lindex [lindex $comment_last 1] 0]=="end"} {
      set is_admissible false
      if {!$using_selections} {
        lappend err_messages [list "unfinished comment" [list [join [lindex $comment_last 0] "."] "end"]]
      } {
        lappend err_messages [list "unfinished comment"]
      }
    }
  }
  set string_last [lindex [lindex $comments_strings_brackets 1] "end"]
  if {[llength $string_last]>1} {
    if {[lindex [lindex $string_last 1] 0]=="end"} {
      set is_admissible false
      if {!$using_selections} {
        lappend err_messages [list "unfinished string" [list [join [lindex $string_last 0] "."] "end"]]
      } {
        lappend err_messages [list "unfinished string"]
      }
    }
  }

  if {!$is_admissible} {
    return [list false $err_messages]
  }

  set disallowed_warns [check_setting_presence_in_command $use_command $comments_strings_brackets [expr {$in_lisp_mode || $in_debug_mode}]]
  if {[llength $disallowed_warns]>0} {
    return [list false $disallowed_warns]
  }
  if {(!$in_lisp_mode) && (!$in_debug_mode)} {
    # since stuff could be set/executed via :lisp(...) constructs
    set disallowed_warns [check_setting_presence_in_command $use_command $comments_strings_brackets true]
    if {[llength $disallowed_warns]>0} {
      return [list false $disallowed_warns]
    }
  }

  if {!$check_borders} {
    return [list true $use_command $use_command $comments_strings_brackets]
  }

# * checking whether there is something left after removing flanking comments;
#   possibly added an ending " ;"/";" (for Maxima), " " (for Lisp) char in there;
#   the ";" ending is commonly reasonable, the " " ending commonly will not fix it;
# * could check for ";" at ends of all lines that are not at regions within brackets,
#   not within comments, not at strings, and with ends not escaped by backslashes;
#   ...this as a possible non-default feature for some later (it will be involved);
  if {!$in_debug_mode} {
    set end_check [check_command_ending $in_lisp_mode $use_command [lindex $comments_strings_brackets 0]]
    if {[lindex $end_check 0]==false} {
      # taking error message in the result (if any), to trigger its display;
      set err_mssg [lindex $end_check 1]
      return [expr {($err_mssg!="") ? [list false [list [list $err_mssg]]] : [list false ""]}]
    }
    if {[llength $end_check]>1} {
      return [list true $use_command [lindex $end_check 1] $comments_strings_brackets]
    } {
      return [list true $use_command $use_command $comments_strings_brackets]
    }
  }

  return [list true $use_command $use_command $comments_strings_brackets]
}

# checking whether the command is not effectively empty
# (that is just comments and space strings),
# and for presence of effectively ending ;/$ chars;
# this should be used both for stack-mode and batch mode cases;
proc check_command_ending {in_lisp_mode command_string cmmd_comments} {
  set err_mssg "Effectively empty command"

  set in_batch [is_batch_mode]

  set command_string [string trimright $command_string]
  if {$command_string==""} {
    return [list false $err_mssg]
  }
  set command_list [split $command_string "\n"]

  # when a line ends by \ in Maxima mode, it is line continuation,
  # and thus by itself should not mean the (effective) end of a command;
  set did_a_change false
  set command_list_tmp [list]

  if {!$in_lisp_mode} {

    set command_list_aux [list]

    foreach one_line $command_list {
      if {[string range $one_line "end" "end"]=="\\"} {
        set end_escaped true
        foreach one_char [lreverse [split [string range $one_line "0" "end-1"] {}]] {
          if {$one_char=="\\"} {
            set end_escaped [expr {$end_escaped ? false : true}]
          } {
            break
          }
        }
        if {$end_escaped} {
          lappend command_list_aux [string range $one_line "0" "end-1"]
          set did_a_change true
        } {
          lappend command_list_aux $one_line
        }
      } {
        lappend command_list_aux $one_line
      }
    }

    if {$did_a_change} {
      set command_string_aux [string trimright [join $command_list_aux "\n"]]
      if {$command_string_aux==""} {
        return [list false $err_mssg]
      }
      set command_list_tmp $command_list
      set command_list [split $command_string_aux "\n"]
    }
  }

  # common end-wise trimming:

  set cmd_idx [llength $cmmd_comments]
  while {0<$cmd_idx} {
    incr cmd_idx -1
    set curr_comment [lindex $cmmd_comments $cmd_idx]
    set curr_comment_end [lindex $curr_comment 1]
    set curr_comment_end_row [lindex $curr_comment_end 0]
    set curr_comment_end_col [lindex $curr_comment_end 1]

    if {[llength $command_list]!=$curr_comment_end_row} {
      break
    }
    if {[string length [lindex $command_list "end"]]!=[expr {$curr_comment_end_col+1}]} {
      break
    }

    set curr_comment_start [lindex $curr_comment 0]
    set curr_comment_start_row [lindex $curr_comment_start 0]
    set curr_comment_start_col [lindex $curr_comment_start 1]

    if {$curr_comment_start_col==0} {
      if {$curr_comment_start_row==1} {
        return [list false $err_mssg]
      }
      set command_list [lrange $command_list 0 [expr {$curr_comment_start_row-2}]]
    } {
      set last_row [string range [lindex $command_list [expr {$curr_comment_start_row-1}]] 0 [expr {$curr_comment_start_col-1}]]
      set last_row [string trimright $last_row]
      if {$curr_comment_start_row==1} {
        set command_list [list]
      } {
        set command_list [lrange $command_list 0 [expr {$curr_comment_start_row-2}]]
      }
      if {$last_row!=""} {
        lappend command_list $last_row
      }
    }

    while {true} {
      if {[llength $command_list]==0} {
        return [list false $err_mssg]
      }
      set last_row [string trimright [lindex $command_list "end"]]
      if {$last_row==""} {
        set command_list [lrange $command_list 0 "end-1"]
      } {
        lset command_list "end" $last_row
        break
      }
    }
  }

  if {[llength $command_list]==0} {
    return [list false $err_mssg]
  }

  # to use the original command lines (if a line-ending backslash had been removed)
  if {$did_a_change} {
    set last_row [lindex $command_list "end"]

    set use_len_base [expr {[llength $command_list] - 2}]
    if {$use_len_base>=0} {
      set command_list [lrange $command_list_tmp 0 $use_len_base]
    } {
      set command_list [list]
    }
    lappend command_list $last_row
  }

  # going to trim comments from left-top too;

  set cmt_idx 0
  set has_a_comment [expr {[llength $cmmd_comments]>0}]
  if {$has_a_comment} {
    set curr_comment [lindex $cmmd_comments $cmt_idx]
    set curr_comment_start [lindex $curr_comment 0]
    set curr_comment_start_row [lindex $curr_comment_start 0]
    set curr_comment_start_col [lindex $curr_comment_start 1]
    set curr_comment_end [lindex $curr_comment 1]
    set curr_comment_end_row [lindex $curr_comment_end 0]
    set curr_comment_end_col [lindex $curr_comment_end 1]
  }

  set command_list_aux [list]

  set row_idx 0
  foreach one_line $command_list {
    incr row_idx
    set cur_line_offset 0
    if {!$has_a_comment} {
      lappend command_list_aux [string range $one_line $cur_line_offset "end"]
      continue
    }
    while {true} {
      if {$curr_comment_start_row==$row_idx} {
        set check_col_upto [expr {$curr_comment_start_col - 1}]
        if {[string trimleft [string range $one_line $cur_line_offset $check_col_upto]]!=""} {
          set has_a_comment false
          lappend command_list_aux [string trimleft [string range $one_line $cur_line_offset "end"]]
          break
        }
        if {$curr_comment_end_row>$row_idx} {
          # here we need 'continue' for the outer cycle,
          # and since inner/outer ones have the same/aligned ends, we can just 'break' here;
          break
        }
        set cur_line_offset [expr {$curr_comment_end_col + 1}]
        incr cmt_idx
        set has_a_comment [expr {[llength $cmmd_comments]>$cmt_idx}]
        if {!$has_a_comment} {
          lappend command_list_aux [string range $one_line $cur_line_offset "end"]
          break
        }
        set curr_comment [lindex $cmmd_comments $cmt_idx]
        set curr_comment_start [lindex $curr_comment 0]
        set curr_comment_start_row [lindex $curr_comment_start 0]
        set curr_comment_start_col [lindex $curr_comment_start 1]
        set curr_comment_end [lindex $curr_comment 1]
        set curr_comment_end_row [lindex $curr_comment_end 0]
        set curr_comment_end_col [lindex $curr_comment_end 1]
        continue
      } elseif {$curr_comment_start_row<$row_idx} {
        if {$curr_comment_end_row>$row_idx} {
          # here we need 'continue' for the outer cycle,
          # and since inner/outer ones have the same/aligned ends, we can just 'break' here;
          break
        }
        set cur_line_offset [expr {$curr_comment_end_col + 1}]
        incr cmt_idx
        set has_a_comment [expr {[llength $cmmd_comments]>$cmt_idx}]
        if {!$has_a_comment} {
          lappend command_list_aux [string range $one_line $cur_line_offset "end"]
          break
        }
        set curr_comment [lindex $cmmd_comments $cmt_idx]
        set curr_comment_start [lindex $curr_comment 0]
        set curr_comment_start_row [lindex $curr_comment_start 0]
        set curr_comment_start_col [lindex $curr_comment_start 1]
        set curr_comment_end [lindex $curr_comment 1]
        set curr_comment_end_row [lindex $curr_comment_end 0]
        set curr_comment_end_col [lindex $curr_comment_end 1]
      } {
        set start_trimmed_line_part [string trimleft [string range $one_line $cur_line_offset "end"]]
        set allowed_line_ends [expr {$in_lisp_mode ? [list ""] : [list "" "\\"]}]
        if {!($start_trimmed_line_part in $allowed_line_ends)} {
          lappend command_list_aux $start_trimmed_line_part
          set has_a_comment false
        }
        break
      }
    }
  }

  set command_list $command_list_aux

  # looking at the ending itself

  set last_row [string trimright [lindex $command_list "end"]]

  if {!$in_lisp_mode} {
    set last_char [string range $last_row "end" "end"]
    if {!($last_char in [list ";" "\$"])} {
      if {$last_char=="\\"} {
        append last_row " "
      }
      # ":lisp(...)" lines at batches tend to dislike semicolon endings
      set avoid_ending false
      if {$in_batch} {
        set aux_block [string trimleft $last_row]
        if {[string range $aux_block 0 0] == ":"} {
          set aux_block [string trimleft [string range $aux_block 1 "end"]]
          if {[string tolower [string range $aux_block 0 3]]=="lisp"} {
            set avoid_ending true
          }
        }
      }
      if {!$avoid_ending} {
        append last_row ";"
      }
    }
    lset command_list "end" $last_row
    return [list true [string trimleft [join $command_list "\n"]]]
  }

  # by now, this checking is only effective for the lisp mode
  set check_odd false
  foreach check_char [lreverse [split $last_row {}]] {
    if {$check_char=="\\"} {
      set check_odd [expr {$check_odd ? false : true}]
      continue
    }
    break
  }
  if {$check_odd} {
    append last_row " "
  }

  lset command_list "end" $last_row
  return [list true [string trimleft [join $command_list "\n"]]]
}

# this only checks reasonably when in maxima mode
proc check_setting_presence_in_command {command_string comments_strings_brackets lisp_mode} {
  set content_list [split $command_string "\n"]
  set replace_idx 0
  while {$replace_idx<2} {
    set replace_regions [lindex $comments_strings_brackets $replace_idx]
    set replace_char [lindex [list " " "X"] $replace_idx]
    incr replace_idx
    foreach curr_item $replace_regions {
      set curr_start [lindex $curr_item 0]
      set curr_start_line [expr {[lindex $curr_start 0] - 1}]
      set curr_start_char [lindex $curr_start 1]
      set curr_end [lindex $curr_item 1]
      set curr_end_line [expr {[lindex $curr_end 0] - 1}]
      set curr_end_char [lindex $curr_end 1]
      if {$curr_start_line==$curr_end_line} {
        lset content_list $curr_start_line \
          [string replace [lindex $content_list $curr_start_line] $curr_start_char $curr_end_char \
            [string repeat $replace_char [expr {1 + $curr_end_char - $curr_start_char}]]]
      } {
        lset content_list $curr_start_line \
          [string replace [lindex $content_list $curr_start_line] $curr_start_char "end" $replace_char]
        set curr_line_idx [expr {$curr_start_line + 1}]
        while {$curr_line_idx < $curr_end_line} {
          lset content_list $curr_line_idx $replace_char
          incr curr_line_idx
        }
        lset content_list $curr_end_line \
          [string replace [lindex $content_list $curr_end_line] 0 $curr_end_char \
            [string repeat $replace_char [expr {1 + $curr_end_char}]]]
      }
    }
  }

  set content_replaced [join $content_list "\n"]
  while {true} {
    set replace_linebreak [string last "\\\n" $content_replaced]
    if {$replace_linebreak==-1} {
      break
    }
    set content_replaced [string replace $content_replaced $replace_linebreak "$replace_linebreak+1"]
  }

  set warn_list [list \
    [list "the variable \"" [strcat "\" should only be set via the " [expr {[is_stack_mode] ? "settings above" : "control settings"}]]] \
    [list "the variable \"" "\" should not be reset here"] \
    [list "the function \"" "\" should not be used here"]]

  set found_terms [list]

  if {!$lisp_mode} {
    set seek_terms [list [list "showtime" "nolabels" "display2d" "leftjust" "linel" "?linel" "?\\*maxima\\-quiet\\*" "?\\*display\\-labels\\-p\\*" "?\\*alt\\-display1d\\*"] \
      [list "ttyoff" "inchar" "outchar" "linechar" "?\\*alt\\-display2d\\*"] [list "reset" "reset_displays" "set_alt_display"]]
  } {
    set seek_terms [list [list "\$showtime" "\$nolabels" "\$display2d" "\$leftjust" "\$linel" "linel" "*maxima-quiet*" "*display-labels-p*" "*alt-display1d*"] \
      [list "#.ttyoff" "\$inchar" "\$outchar" "\$linechar" "*alt-display2d*"] [list "\$reset" "\$reset_displays" "\$set_alt_display"]]
  }

  set space "\\ \\t"
  set setidx 0
  while {$setidx<3} {
    set form_warn [lindex [list 0 1 2] $setidx]
    set form_term [lindex $seek_terms $setidx]
    set form_func [lindex [list false false true] $setidx]
    incr setidx
    foreach setterm $form_term {
      set regterm [join [split $setterm "\\"] "\\\\"]
      set regterm [join [split $regterm "?"] "\\?"]
      set regterm [join [split $regterm "\$"] "\\\$"]
      set regterm [join [split $regterm "*"] "\\*"]
      set regterm [join [split $regterm "-"] "\\-"]
      set regterm [join [split $regterm "."] "\\."]
      set regterm [join [split $regterm "#"] "\\#"]
      set curr_offset 0
      while {true} {
        set idx_pair [list]
        if {!$form_func} {
          if {!$lisp_mode} {
            set found_count [regexp -start $curr_offset -indices "$regterm\[$space\]*:" $content_replaced idx_pair]
          } {
            set found_count [regexp -start $curr_offset -indices "\\(\[$space\]*set\[qf\]\{0,1\}\[$space\]+\['\]\{0,1\}$regterm\[$space\]+" $content_replaced idx_pair]
          }
        } {
          if {!$lisp_mode} {
            set found_count [regexp -start $curr_offset -indices "$regterm\[$space\]*\\(" $content_replaced idx_pair]
          } {
            set found_count [regexp -start $curr_offset -indices "\\(\[$space\]*$regterm\[$space\]+" $content_replaced idx_pair]
          }
        }
        if {$found_count==0} {
          break
        }
        set idx_start [lindex $idx_pair 0]
        set idx_end [lindex $idx_pair 1]
        set prev_char [string index $content_replaced [expr {$idx_start - 1}]]
        set next_char [string index $content_replaced [expr {$idx_end + 1}]]
        if {$lisp_mode || (($prev_char!="?") && (![string is wordchar -strict $prev_char]) && ($form_func || ($next_char!="=")))} {
          set warn_parts [lindex $warn_list $form_warn]
          set show_message [lindex $warn_parts 0]
          append show_message $setterm
          append show_message [lindex $warn_parts 1]
          lappend found_terms [list $show_message]
          break
        } {
          set curr_offset [expr {$idx_start + 1}]
        }
      }
    }
  }

  return $found_terms
}

set info_tag_start "|\{\{\{"
set info_tag_end "\}\}\}|"
set info_tag_middle "\}||||\{"

proc set_labels_to_asked_values {} {
  global bind_labels
  global with_labels
  global bind_labels_asked
  global with_labels_asked
  global line_display_mode_current
  global line_display_mode_current_asked
  global info_line_size
  global info_line_size_asked

  lset info_line_size 0 [lindex $info_line_size_asked 0]

  set bind_labels $bind_labels_asked
  set with_labels $with_labels_asked
  update_info_labels_shown_bound_request false

  set line_display_mode_current $line_display_mode_current_asked
  update_info_display_request false

}

set maxima_setting_done false
set maxima_setting_answers ""
proc setup_information_update {onset answer} {
  global maxima_in_setting
  global maxima_in_setting_start
  global maxima_setting_answers
  global info_tag_start
  global info_tag_middle
  global info_tag_end
  global maxima_command
  global maxima_command_inner
  global maxima_command_control
  global maxima_ready
  global state_label
  global maxima_setting_done
  global after_restart_fresh
  global maxima_info_version
  global maxima_info_lisp_name
  global with_control_output
  global should_display_cpu_usage
  global could_display_cpu_usage

  if {$maxima_in_setting_start} {
    set maxima_setting_answers ""
  }
  set maxima_in_setting_start false

  if {($onset=="") && ([llength $answer]==0)} {
    return
  }

  # setup sometimes gets superfluous backslashes,
  # thus trying to deal with that here too;
  set overall_answer [string map [list "\\" ""] $onset]
  append overall_answer [string map [list "\\" ""] [join $answer ""]]
  append maxima_setting_answers $overall_answer
  set loc_tag_end [string last $info_tag_end $maxima_setting_answers]
  if {$loc_tag_end<0} {
    return
  }

  # if got version/lisp info, set it in
  set loc_tag_start [string first $info_tag_start $maxima_setting_answers]
  set loc_tag_middle [string first $info_tag_middle $maxima_setting_answers]
  if {($loc_tag_start>-1) && ($loc_tag_middle>-1)} {
    set info_part_first [string range $maxima_setting_answers [expr {
      $loc_tag_start + [string length $info_tag_start]
    }] [expr {$loc_tag_middle - 1}]]
    set loc_tag_middle [string last $info_tag_middle $maxima_setting_answers]
    set info_part_second [string range $maxima_setting_answers [expr {
      $loc_tag_middle + [string length $info_tag_middle]
    }] [expr {$loc_tag_end - 1}]]

    if {[string length $info_part_first]>0} {
      set maxima_info_version $info_part_first
    }
    if {[string length $info_part_second]>0} {
      set maxima_info_lisp_name $info_part_second
    }
  }
  set maxima_setting_answers ""

  set maxima_in_setting false
  set maxima_setting_done true

  set_labels_to_asked_values

  # enable stuff
  set maxima_command ""
  set maxima_command_inner ""
  set maxima_command_control ""
  set_state_ready
  set maxima_ready true
  set_button_interrupt_idle
  set should_display_cpu_usage false
  set could_display_cpu_usage true
  set_state_label_text "ready"

  set with_control_output false
  # notice: should check whether the received data contain some another part;
  # if it does contain, it should be pushed back to "consume read parts"
  # for another processing (or that function should do more general splitting
  # of the received data by itself);
  # ...this may be a stuff to do at some later;

  if {$after_restart_fresh} {
    set after_restart_fresh false
    # this had to be used when the restart button was taking focus
    #set_focused
  }
}

proc set_functions_info {info_string} {
  global info_func_list
  global info_tag_start
  global info_tag_end
  global within_names_update

  # taking the last occurrence to avoid (sometimes excessive)
  # debug output from the used lisp;
  set found_loc [string last $info_tag_start $info_string]
  set found_loc_end [expr {$found_loc+[string length $info_tag_start]-1}]
  set info_string [string replace $info_string 0 $found_loc_end]
  set found_loc [string last $info_tag_end $info_string]
  set info_string [string replace $info_string $found_loc "end"]
  set info_string [string map [list "\\" ""] $info_string]

  if {$info_string=="false"} {
    set info_string ""
  }
  set info_func_list [split $info_string ";"]

  set func_count [llength $info_func_list]

  set curr_size [create_keyword_vocabulary $info_func_list]
  set within_names_update false
}

proc set_values_info {info_string} {
  global info_vals_list
  global info_tag_start
  global info_tag_end
  global within_names_update

  # taking the last occurrence to avoid (sometimes excessive)
  # debug output from the used lisp;
  set found_loc [string last $info_tag_start $info_string]
  set found_loc_end [expr {$found_loc+[string length $info_tag_start]-1}]
  set info_string [string replace $info_string 0 $found_loc_end]
  set found_loc [string last $info_tag_end $info_string]
  set info_string [string replace $info_string $found_loc "end"]
  set info_string [string map [list "\\" ""] $info_string]

  if {$info_string=="false"} {
    set info_string ""
  }
  set info_vals_list [split $info_string ";"]

  set curr_size [create_keyword_vocabulary $info_vals_list]
  set within_names_update false
}

proc get_maxima_command_startup_setting_end {} {
  global info_line_left
  global info_line_1dim
  global info_line_long
  global info_line_size
  global bind_labels
  global info_tag_start
  global info_tag_end
  global with_labels

  # not setting the global maxima_command directly here

  set use_line_left $info_line_left
  set use_line_long $info_line_long

  if {$info_line_1dim} {
    set use_line_left true
    # some forms get split by Maxima, some are dealt with Minima;
    #set use_line_long true
  } {
    set use_line_long false
  }

  set maxima_line_length [lindex $info_line_size [expr {$use_line_long ? 1 : 0}]]

  set use_maxima_command "(?\\*maxima\\-quiet\\*:false,"
  append use_maxima_command "linel:$maxima_line_length,"
  if {$use_line_left} {
    append use_maxima_command "leftjust:true,"
  } {
    append use_maxima_command "leftjust:false,"
  }
  if {$info_line_1dim} {
    append use_maxima_command "display2d:false,"
  } {
    append use_maxima_command "display2d:true,"
  }
  if {$with_labels} {
    append use_maxima_command "?\\*display\\-labels\\-p\\*:true,"
  } {
    append use_maxima_command "?\\*display\\-labels\\-p\\*:false,"
  }

  set nolabels_value [expr {$bind_labels ? "false" : "true"}]
  append use_maxima_command "nolabels:$nolabels_value,"

  append use_maxima_command "showtime:true,"

  append use_maxima_command "grind(\"$info_tag_start$info_tag_end\"),kill(labels))\$"

  return $use_maxima_command
}

foreach wdg [list $topw $text_field $command_field] {
  bind $wdg <Control-u> {
    if {[is_command_field_active]} {
      ask_for_current_names "functions"
    }
  }
  bind $wdg <Control-U> {
    if {[is_command_field_active]} {
      ask_for_current_names "values"
    }
  }
}

set in_name_asking false
set in_name_asking_form ""
set within_names_update false
set control_names_updates false
proc ask_for_current_names {form} {
  global in_name_asking
  global in_name_asking_form
  global maxima_command
  global maxima_command_control
  global in_lisp_mode
  global in_debug_mode
  global overall_names_asked
  global state_label
  global bt_restart
  global info_tag_start
  global info_tag_end
  global newly_at_maxima_outputs
  #global can_kill_label
  global has_ever_asked
  global within_names_update
  global control_output_onset_shibboleth
  global control_names_updates
  global waiting_for_control_answer

  if {$maxima_command!=""} {
    # should not push both common commands and control commands at the same time;
    return
  }

  if {$in_lisp_mode || $in_debug_mode} {
    return
  }

  if {!($form in [list "functions" "values"])} {
    return
  }

  # disable widgets
  set_state_disabled
  $bt_restart configure -state normal
  # interrupt-button is set once the shibboleth is read
  #set_button_interrupt_sets

  # avoiding use of simplode, since its package need not be loaded
  set form_part "(\$tree_reduce (lambda (x y) (\$sconcat x \";\" y)) \$$form \"\")"

  set in_name_asking_form $form
  set overall_names_asked ""
  # this is set once shibboleth is read
  #set in_name_asking true

  set should_display_cpu_usage false
  set could_display_cpu_usage false
  set_state_label_text "updating"

  # this is set once shibboleth is read
  #set newly_at_maxima_outputs false

  # prepare command
  set maxima_command_control ":lisp(format t \"~&$control_output_onset_shibboleth~%\");\n"

  append maxima_command_control ":lisp(progn (setf \$leftjust t)(setf \$display2d nil)(setf \$linel 1000000)(setf linel 1000000)(setf #.ttyoff nil)"
  append maxima_command_control "(mgrind (\$concat \"$info_tag_start\" $form_part \"$info_tag_end\") t)(force-output) #\\newline);"
  # this set once shibboleth is read
  #set within_names_update true

  # if had to switch to Maxima-based commands without lisp:
  #set can_kill_label true
  set control_names_updates true
  set waiting_for_control_answer true

  svwrite_onset
  return
}

set overall_names_asked ""
proc wait_for_current_names {onset answer} {
  global in_name_asking
  global in_name_asking_form
  global overall_names_asked
  global info_tag_end
  global with_control_output

  if {($onset=="") && ([llength $answer]==0)} {
    return
  }

  # setup sometimes gets superfluous backslashes,
  # thus trying to deal with that here too;
  set overall_answer [string map [list "\\" ""] $onset]
  append overall_answer [string map [list "\\" ""] [join $answer ""]]
  append overall_names_asked $overall_answer

  set with_end [string last $info_tag_end $overall_names_asked]
  if {$with_end>-1} {
    if {$in_name_asking_form=="functions"} {
      set_functions_info $overall_names_asked
    }
    if {$in_name_asking_form=="values"} {
      set_values_info $overall_names_asked
    }
    set overall_names_asked ""
    set in_name_asking false
    set in_name_asking_form ""
    set with_control_output false
    call_maxima_line_setting true
  }
  return
}

set info_state "ini0"
set info_versions_string ""
set info_func_string ""
set info_func_list [list]
set info_vals_string ""
set info_vals_list [list]
set info_setup_inside false
proc setup_information {onset answer} {
  global maxima_command
  global maxima_command_inner
  global maxima_command_control
  global maxima_ready
  global info_state
  global info_versions_string
  global info_func_string
  global info_vals_string
  global info_setup_inside
  global state_label
  global focused_part
  global info_tag_start
  global info_tag_middle
  global info_tag_end
  global names_from_functions_to_load
  global names_from_values_to_load
  global maxima_info_version
  global maxima_info_lisp_name
  global should_display_cpu_usage
  global could_display_cpu_usage
  global mxm_display_exp_dstar
  global asked_temporary_path

  if {$info_state=="done"} {
    return
  }

  if {($onset=="") && ([llength $answer]==0)} {
    if {$info_state!="ini0"} {
      return
    }
  }
  # when getting into setup after some Maxima ini stuff outputs,
  # we get superfluous backslashes here; 
  set overall_answer [string map [list "\\" ""] $onset]
  append overall_answer [string map [list "\\" ""] [join $answer ""]]

  set info_setup_inside true
  set should_display_cpu_usage false
  set could_display_cpu_usage false
  set_state_label_text "setup"

  # avoiding simplode use, since its package need not be loaded;
  set common_funval_part_start "tree_reduce(lambda(\[x,y\],sconcat(x,\";\",y)),"
  set common_funval_part_end ",\"\")"
  set functions_part [strcat $common_funval_part_start "functions" $common_funval_part_end]
  set values_part [strcat $common_funval_part_start "values" $common_funval_part_end]

  set functions_to_load_command "grind(concat(\"$info_tag_start\",$functions_part,\"$info_tag_end\"));"
  set values_to_load_command "grind(concat(\"$info_tag_start\",$values_part,\"$info_tag_end\"));"

  if {$info_state=="ini0"} {
    set maxima_info_version ""
    set maxima_info_lisp_name ""
    set info_versions_string ""

    set info_state "ini1"
    set maxima_command ""
    set maxima_command_inner ""
    set maxima_command_control "(leftjust:true,display2d:false,linel:1000000,nolabels:true,"
    # we already expect shown labels when with such an option given, thus dealing with it here too;
    #append maxima_command_control "?\\*display\\-labels\\-p\\*:false,?\\*maxima\\-quiet\\*:false,"
    append maxima_command_control "?\\*maxima\\-quiet\\*:false,"
    append maxima_command_control "?\\*alt\\-display1d\\*:false,"
    append maxima_command_control "?\\*alt\\-display2d\\*:false,"
    append maxima_command_control "inchar:%i,outchar:%o,linechar:%t,ttyoff:false,"
    if {$mxm_display_exp_dstar} {
      append maxima_command_control "1);\n:lisp(setf (get 'mexpt 'grind) 'msize-infix)\n("
    }
    if {$asked_temporary_path!=""} {
      append maxima_command_control [strcat \
        "1);maxima_tempdir:\"" [string map [list "\\" "\\\\" "\"" "\\\""] $asked_temporary_path] "\"\$\n(" \
      ]
    }
    append maxima_command_control "grind(concat(\"$info_tag_start\",build_info()@version,\"$info_tag_middle\",build_info()@lisp_name,\"$info_tag_end\")));"
    svwrite_onset
    return
  }
  if {$info_state=="ini1"} {
    append info_versions_string $overall_answer
    if {[string last $info_tag_end $info_versions_string]<0} {
      return
    }
    set found_loc_start [string first $info_tag_start $info_versions_string]
    set found_loc_middle [string first $info_tag_middle $info_versions_string]
    if {($found_loc_start>-1)&&($found_loc_middle>-1)} {
      set maxima_info_version [string range $info_versions_string [expr {
        $found_loc_start + [string length $info_tag_start]
      }] [expr {$found_loc_middle - 1}]]
    }
    set found_loc_middle [string last $info_tag_middle $info_versions_string]
    set found_loc_end [string last $info_tag_end $info_versions_string]
    if {($found_loc_middle>-1)&&($found_loc_end>-1)} {
      set maxima_info_lisp_name [string range $info_versions_string [expr {
        $found_loc_middle + [string length $info_tag_middle]
      }] [expr {$found_loc_end - 1}]]
    }
    set info_versions_string ""

    if {$names_from_functions_to_load} {
      set info_state "fun0"
      set maxima_command ""
      set maxima_command_inner ""
      set maxima_command_control $functions_to_load_command
    } {
      if {$names_from_values_to_load} {
        set info_state "val0"
        set maxima_command ""
        set maxima_command_inner ""
        set maxima_command_control $values_to_load_command
      } {
        set info_state "res0"
        set maxima_command ""
        set maxima_command_inner ""
        set maxima_command_control [get_maxima_command_startup_setting_end]
      }
    }
    svwrite_onset
    return
  }
  if {($info_state=="fun0") || ($info_state=="funN")} {
    set info_state "funN"
    append info_func_string $overall_answer
    set with_end [string last $info_tag_end $info_func_string]
    if {$with_end>-1} {
      set_functions_info $info_func_string
      if {$names_from_values_to_load} {
        set info_state "val0"
        set maxima_command ""
        set maxima_command_inner ""
        set maxima_command_control $values_to_load_command
      } {
        set info_state "res0"
        set maxima_command ""
        set maxima_command_inner ""
        set maxima_command_control [get_maxima_command_startup_setting_end]
      }
      svwrite_onset
    }
    return
  }
  if {($info_state=="val0") || ($info_state=="valN")} {
    set info_state "valN"
    append info_vals_string $overall_answer
    set with_end [string last $info_tag_end $info_vals_string]
    if {$with_end>-1} {
      set_values_info $info_vals_string
      set info_state "res0"
      set maxima_command ""
      set maxima_command_inner ""
      set maxima_command_control [get_maxima_command_startup_setting_end]
      svwrite_onset
    }
    return
  }
  if {$info_state=="res0"} {
    if {[string last $info_tag_end $overall_answer]<0} {
      return
    }
    set info_state "resN"
  }

  set maxima_command ""
  set maxima_command_inner ""
  set maxima_command_control ""
  set_state_ready
  set maxima_ready true
  set_button_interrupt_idle

  set should_display_cpu_usage false
  set could_display_cpu_usage true
  set_state_label_text "ready"

  if {$focused_part==-1} {
    set_focused
  }

  set info_setup_inside false
  set info_state "done"
}

set after_restart false

proc ensure_empty_line {wdg line} {
  # if 'line' is over the text end, we get 0 as the end of such line,
  # thus we do not need to check it here;
  if {$line<1} {
    return false
  }
  set line_end [lindex [split [$wdg index "$line.end"] "."] 1]
  if {$line_end!=0} {
    $wdg insert "$line.0" "\n"
  }
  return true
}

set shibboleth_onset_read false
set shibboleth_onset_postponed false
set was_in_debug_state false
proc process_answer {onset answer} {
  global text_field
  global command_field
  global maxima_pid
  global maxima_command
  global maxima_command_inner
  #global maxima_command_control
  global has_ever_asked
  global maxima_insert_line
  global state_label
  global maxima_ready
  global in_lisp_mode
  global in_debug_mode
  global focused_part
  global with_labels
  global with_labels_asked
  global info_setup_inside
  global after_restart
  global maxima_in_setting
  global maxima_in_setting_start
  global maxima_setting_done
  global in_name_asking
  global shibboleth_onset
  global shibboleth_onset_read
  global newly_at_maxima_outputs
  global include_maxima_input
  global curr_stack_mode
  global with_control_output
  global waiting_for_control_answer
  global common_insert_mark_name
  global should_display_cpu_usage
  global could_display_cpu_usage
  global shibboleth_onset_postponed
  global was_in_debug_state

  set_tags_shown_state false

  # this should process correctly in several distinct situations:
  # * after Maxima (re)start, with lesser/larger information update after restart/start;
  #   ...and (most of) the info update gets delayed if Maxima is in Lisp mode
  #      (either intentionally due to init setting, or due to an issue in the inits);
  #   ...plus the info on PID may be (somewhat) delayed if the display resetting
  #      functions were active (used to make sure we are in a common display mode);
  # * during Maxima common running:
  #   commands in Maxima mode and in Lisp mode,
  #   in Maxima settings update and/or asking for (user-defined) function/value names;
  #
  # since Maxima can output mostly anything, in any way,
  # this processing may guess wrongly on the current processing state;
  # still it was tested in a multitude of situations on the cases above;

  if {$in_name_asking} {
    wait_for_current_names $onset $answer
    return
  }

  if {$in_lisp_mode || $in_debug_mode} {
    set maxima_in_setting false
    set maxima_in_setting_start false
    set with_control_output false
  }
  if {$maxima_in_setting} {
    setup_information_update $onset $answer
    return
  }

  set main_answer ""
  set reduced_answer [list]
  foreach one_line $answer {
    set is_output_line true
    if {($with_labels || $with_labels_asked) && [string equal -nocase -length 3 "(%i" $one_line]} {
      if {[string equal -nocase ") " [string range $one_line end-1 end]]} {
        if {[string is digit -strict [string range $one_line 3 end-2]]} {
          set is_output_line false
        }
      }
    }
    if {$is_output_line} {
      lappend reduced_answer $one_line
    }
  }

  if {$info_setup_inside} {
    setup_information $onset $reduced_answer
    return
  }

  if {$maxima_pid==""} {
    set pid_line $onset
    set onset ""
    if {[string first "pid=" $pid_line]==-1} {
      set pid_line [lindex $reduced_answer 0]
      set reduced_answer [lrange $reduced_answer 1 "end"]
      set_maxima_pid $pid_line
      set has_ever_asked false
    }
    set reduced_answer_cut [list]
    foreach one_reduced_part $reduced_answer {
      lappend reduced_answer_cut $one_reduced_part
      # this already tested in the "consume_read_parts" function
      #if {[string first "MAXIMA>" $one_reduced_part]==0} {
      #  set_lisp_mode true
      #}
      if {[string first $shibboleth_onset $one_reduced_part]!=-1} {
        set shibboleth_onset_read true
        set reduced_answer_cut [list]
      }
    }
    set one_reduced_part $reduced_answer_cut
    if {!$in_lisp_mode} {
      if {!$shibboleth_onset_read} {
        return
      }
    }
  }

  set to_call_maxima_line_setting false
  if {(!$shibboleth_onset_read) && (!$in_lisp_mode) && (!$in_debug_mode)} {
    set reduced_answer_new [list]
    foreach one_reduced_part $reduced_answer {
      if {[string first $shibboleth_onset $one_reduced_part]!=-1} {
        set shibboleth_onset_read true
      } {
        lappend reduced_answer_new $one_reduced_part
      }
      # this already tested in the "consume_read_parts" function
      #if {[string first "MAXIMA>" $one_reduced_part]==0} {
      #  set_lisp_mode true
      #}
    }
    set reduced_answer $reduced_answer_new
    if {(!$in_lisp_mode) && (!$in_debug_mode)} {
      if {!$shibboleth_onset_read} {
        return
      } {
        set to_call_maxima_line_setting true
      }
    }
  }

  # setting the "maxima_insert_line" variable;
  # sets it by the first outputs on a command only
  # (meant as if a command output has some delayed parts)
  set use_line -1
  if {$shibboleth_onset_read && (!$to_call_maxima_line_setting)} {
    if {(!($common_insert_mark_name in [$text_field mark names])) && $has_ever_asked} {
      set newly_at_maxima_outputs true
    }
    if {$newly_at_maxima_outputs} {
      set_output_insertion_point
      set newly_at_maxima_outputs false
    } {
      if {($maxima_insert_line!=-1) && $has_ever_asked} {
        # have to adjust the insert point according to the "common_insert_mark_name" mark,
        # since the text may had get altered (if this is not right by the mark setting)
        # if dealing with a later-coming parts (thus the insertions would be mistargeted);
        set mark_based_index ""
        # if $has_ever_asked, then the mark should be there, since we deal with it above;
        # the mark "$common_insert_mark_name" is set via "set_output_insertion_point";
        if {$common_insert_mark_name in [$text_field mark names]} {
          set mark_based_index [$text_field index $common_insert_mark_name]
        }
        if {$mark_based_index!=""} {
          set mark_based_info [split $mark_based_index "."]
          set mark_based_row [lindex $mark_based_info 0]
          if {[lindex $mark_based_info 1]!=0} {
            # if some editing had put some text in front of the mark,
            # put it at the start of the next line;
            incr mark_based_row
            set curr_text_state [$text_field cget -state]
            $text_field configure -state "normal"
            $text_field mark set $common_insert_mark_name [strcat $mark_based_row ".0"]
            $text_field configure -state $curr_text_state
          }
          set maxima_insert_line $mark_based_row
          set end_row_num [lindex [split [$text_field index "end -1 chars"] "."] 0]
          if {$end_row_num < $maxima_insert_line} {
            set curr_text_state [$text_field cget -state]
            $text_field configure -state "normal"
            $text_field edit separator
            $text_field insert "end" "\n"
            $text_field configure -state $curr_text_state
          }
        }
      }
    }
    set use_line $maxima_insert_line
  }

  set reduced_answer_length [llength $reduced_answer]
  if {$reduced_answer_length>0} {
    set main_answer [join $reduced_answer "\n"]
  }

  if {!$has_ever_asked} {
    # if error occurred during startup, avoiding ini-info reading
    if {$in_lisp_mode || $in_debug_mode} {
      # though getting here (with empty stuff) when resetting display too
      if {[string length $main_answer]>0} {
        set has_ever_asked true
      }
      if {!$in_debug_mode} {
        set should_display_cpu_usage false
        set could_display_cpu_usage true
        if {!$has_ever_asked} {
          set_state_label_text "startup"
        } {
          # may be that here should unlock;
          set_state_label_text "issue"
          set shibboleth_onset_postponed true
        }
      }
      if {$in_debug_mode || $has_ever_asked} {
        set_state_ready
        set maxima_ready true
        set_button_interrupt_idle
      }
      if {$focused_part==-1} {
        set_focused
      }
    }
  }

  if {!$has_ever_asked && (!$shibboleth_onset_read)} {
    if {((!$in_lisp_mode) && (!$in_debug_mode)) || ([string length $main_answer]==0)} {
      return
    }
  }

  if {!$has_ever_asked} {
    if {(!$after_restart) && ($maxima_pid!="")} {
      # "going into info setup"
      setup_information $onset $reduced_answer
      return
    }
    if {($after_restart) && ($maxima_pid!="")} {
      # "going into info setting"
      if {!$maxima_setting_done} {
        call_maxima_line_setting
      }
      return
    }
    return
  }

  if {$onset=="" && $maxima_command=="" && $main_answer==""} {
    if {$to_call_maxima_line_setting} {
      if {$was_in_debug_state} {
        call_maxima_line_setting
      } {
        setup_information "" ""
      }
    }
    return
  }

  set to_reset_fields false
  if {(![is_maxima_command_pending]) && (!$waiting_for_control_answer)} {
    if {!$in_debug_mode} {
      set should_display_cpu_usage false
      set could_display_cpu_usage true
      if {$shibboleth_onset_read} {
        set_state_label_text "ready"
      } {
        # here should unlock (too)
        if {!$in_debug_mode} {
          set_state_label_text "issue"
        }
        set shibboleth_onset_postponed true
      }
    }
    set_state_ready
    set_button_interrupt_idle
    set maxima_ready true
  } {
    set to_reset_fields true
    set text_field_state [$text_field cget -state]
    set cmmd_field_state [$command_field cget -state]
    $text_field configure -state "normal"
    $command_field configure -state "normal"
  }

  set to_insert_first $onset
  append to_insert_first $maxima_command
  # to not use late-coming answer parts with (%iN)
  if {$maxima_command==""} {
    if {[string equal -nocase -length 3 "(%i" $onset]} {
      if {[string equal -nocase ") " [string range $onset end-1 end]]} {
        if {[string is digit -strict [string range $onset 3 end-2]]} {
          set to_insert_first ""
        }
      }
    }
  }
  if {!$include_maxima_input} {
    set to_insert_first ""
  }

  # setting place for scrolling to, and for possible late-coming answer parts
  set line_increase 0
  if {$include_maxima_input} {
    set to_insert_first [string trimright $to_insert_first]
    if {$to_insert_first!=""} {
      incr line_increase [llength [split $to_insert_first "\n"]]
      append to_insert_first "\n"
    }
    set line_increase_for_second_part $line_increase
  }

  set to_insert ""
  if {[string length $main_answer]>0} {
    append to_insert $main_answer "\n"
    incr line_increase $reduced_answer_length
  }

  # some results (in multi-command) may come with a delay
  # thus not to delete new parts in the command field
  if {$maxima_command!=""} {
    if {[string trim [get_command_content]]!=""} {
      save_current_stack
    }
    if {$curr_stack_mode!=0} {
      stack_on_exec_end
      if {$curr_stack_mode==1} {
        $command_field delete 1.0 end
        $command_field edit reset
        $command_field edit modified false
      }
    }
  }
  # to target correctly (in the text field)
  # the possibly late-coming parts of results
  if {$use_line != -1} {
    incr maxima_insert_line $line_increase
  }

  if {$use_line != -1} {
    # can not do replacing, since late-coming parts would overwrite earlier ones
    # and without replacing we still have an empty line for next command output;
    ensure_empty_line $text_field $use_line
    $text_field edit separator
    set use_line_next $use_line
    if {$to_insert_first!=""} {
      $text_field insert "$use_line.0" $to_insert_first
      $text_field edit separator
      incr use_line_next $line_increase_for_second_part
    }
    $text_field insert "$use_line_next.0" $to_insert
    $text_field edit separator
    $text_field see "$maxima_insert_line.0"
    text_outer_modified
  } {
    # tk always puts newline at end, thus testing line before it
    set check_line_index [expr {[lindex [split [$text_field index "end"] "."] 0] - 1}]
    set check_line [$text_field get "$check_line_index.0" "$check_line_index.end"]
    set needs_newline false
    if {[string length $check_line]>0} {
      if {[string range $check_line end end] != "\n"} {
        set needs_newline true
      }
    }
    $text_field edit separator
    if {$needs_newline} {
      $text_field insert end "\n"
      $text_field edit separator
    }
    if {$to_insert_first!=""} {
      $text_field insert end $to_insert_first
      $text_field edit separator
    }
    $text_field insert end $to_insert
    $text_field edit separator
    $text_field yview end
    text_outer_modified
  }

  $text_field edit separator
  set maxima_command ""
  set maxima_command_inner ""

  if {$to_call_maxima_line_setting} {
    if {$was_in_debug_state} {
      call_maxima_line_setting
    } {
      setup_information "" ""
    }
  }

  if {$to_reset_fields} {
    $text_field configure -state $text_field_state
    $command_field configure -state $cmmd_field_state
  }
  return
}

set can_deal_with_disconnect false
proc set_disconnect {} {
  global maxima_channel
  global checks_id
  global maxima_ready
  global has_ever_asked
  global state_label
  global bt_restart
  global text_field
  global maxima_pid
  global maxconn
  global read_wait_id
  global maxima_channel_hnd
  global can_deal_with_disconnect
  global should_display_cpu_usage
  global could_display_cpu_usage

  if {!$can_deal_with_disconnect} {
    return
  }

  catch {after cancel $read_wait_id}
  clear_read_buffers

  set maxima_ready false

  if {$checks_id!=""} {
    catch {after cancel $checks_id}
  }

  if {$maxima_channel!=""} {
    catch {fileevent $maxima_channel readable ""}
    catch {fileevent $maxima_channel writable ""}
    catch {chan close $maxima_channel}
    set maxima_channel ""
  }

  if {$maxima_pid!=""} {
    do_kill_maxima true
    set maxima_pid ""
  }
  if {$maxima_channel_hnd!=""} {
    clear_maxima_channel_hnd
  }

  set maxconn 0
  set has_ever_asked false

  set_button_interrupt_sets
  set_state_disabled
  $bt_restart configure -state normal
  $text_field configure -state normal

  set should_display_cpu_usage false
  set could_display_cpu_usage false
  set_state_label_text "disconnected"
}

set checks_id ""
proc setup_checks {} {
  global maxima_channel
  global checks_id

  if {$checks_id!=""} {
    catch {after cancel $checks_id}
  }
  set checks_id ""

  if {$maxima_channel!=""} {
    set checks_id [after 1000 check_connected_after]
  }
}

proc check_connected_after {} {
  global checks_id

  if {$checks_id!=""} {
    catch {after cancel $checks_id}
  }
  set checks_id ""

  set is_connected [check_connected]

  if {$is_connected} {
    set checks_id [after 0 setup_checks]
  }
}

proc check_connected {{soft_check false}} {
  global maxima_channel

  set is_connected false
  if {$maxima_channel != ""} {
    if {[llength [file channels $maxima_channel]]>0} {
      set is_connected true
    }
    set at_eof 0
    if {$is_connected} {
      if {[catch {set at_eof [eof $maxima_channel]}]} {
        set is_connected false
      }
    }
    if {$at_eof} {
      set is_connected false
    }
  }

  if {(!$is_connected) && (!$soft_check)} {
    set_disconnect
  }

  return $is_connected
}

proc update_focus_info {state} {
  global focused_part
  set focused_part $state
}

set text_focus_in_real_mode true
proc set_text_tags_view_focused {is_focused} {
  global text_field
  global tag_text_line
  global tag_text_line_block
  global tag_text_line_part
  global text_focus_in_real_mode

  if {!$text_focus_in_real_mode} {
    return
  }

  set stipple_form ""
  if {!$is_focused} {
    set stipple_form gray75
  }
  $text_field tag configure "sel" -bgstipple $stipple_form
  $text_field tag configure $tag_text_line -bgstipple $stipple_form
  $text_field tag configure $tag_text_line_block -bgstipple $stipple_form
  $text_field tag configure $tag_text_line_part -bgstipple $stipple_form
}

proc set_cmmd_tags_view_focused {is_focused} {
  global command_field
  global tag_stripes

  set stipple_form ""
  if {!$is_focused} {
    set stipple_form gray75
  }
  $command_field tag configure "sel" -bgstipple $stipple_form
  $command_field tag configure $tag_stripes -bgstipple $stipple_form
}

bind $text_field <FocusIn> {
  set_text_tags_view_focused true
  update_focus_info 1
  set_tags_shown_state false
  break
}
bind $command_field <FocusIn> {
  set_cmmd_tags_view_focused true
  update_focus_info 0
  set_tags_shown_state false
  break
}
bind $command_field_aux <FocusIn> {
  remove_torn_menus
  break
}

bind $search_entry <FocusIn> {
  if {[string tolower [$search_entry cget -state]]!="disabled"} {
    $search_entry configure -state "normal"
    $search_entry configure -style SearchIn.TEntry
  }
  remove_torn_menus
  break
}

bind $tags_decls_line <FocusIn> {
  if {[is_wine_like]} {
    $tags_decls_line configure -borderwidth 1 -relief sunken
  }
  remove_torn_menus
  catch {slide_tags_declared 0}
  break
}
bind $tags_decls_line <1> {
  catch {slide_tags_declared 0}
}

bind $tags_links_line <FocusIn> {
  if {[is_wine_like]} {
    $tags_links_line configure -borderwidth 1 -relief sunken
  }
  remove_torn_menus
  catch {slide_tags_links 0}
  break
}
bind $tags_links_line <1> {
  catch {slide_tags_links 0}
}

bind $text_field <FocusOut> {
  set_text_tags_view_focused false
}
bind $command_field <FocusOut> {
  set_cmmd_tags_view_focused false
}
bind $search_entry <FocusOut> {
  $search_entry configure -state "readonly"
  $search_entry configure -style SearchOut.TEntry
}
bind $tags_decls_line <FocusOut> {
  if {[is_wine_like]} {
    $tags_decls_line configure -borderwidth 1 -relief flat
  }
}
bind $tags_links_line <FocusOut> {
  if {[is_wine_like]} {
    $tags_links_line configure -borderwidth 1 -relief flat
  }
}
if {[is_batch_mode]} {
  if {[is_wine_like]} {
    $tags_decls_line configure -borderwidth 1 -relief flat
    $tags_links_line configure -borderwidth 1 -relief flat
  }
}

proc position_effectively_at_end {wdg line_idx char_idx} {
  # not considering too-far-from-end lines as effectively at the end
  if {[lindex [$wdg count -lines "$line_idx.$char_idx" "end"] 0] > 10} {
    return false
  }
  if {[lindex [$wdg count -chars "$line_idx.$char_idx" "end"] 0] > 100} {
    return false
  }
  # otherwise space-like characters not considered for preventing end-like state
  if {[string trim [$wdg get "$line_idx.$char_idx" "end"]]==""} {
    return true
  }

  return false
}

set common_insert_mark_name "common_insert_mark"
proc set_output_insertion_point {} {
  global maxima_insert_line
  global common_insert_mark_name
  global text_field

  set maxima_insert_line [set_output_insertion_point_inner]

  if {$maxima_insert_line!=-1} {
    set mark_pos [strcat $maxima_insert_line ".0"]
  } {
    set mark_pos "end"
  }

  set curr_text_state [$text_field cget -state]
  $text_field configure -state "normal"
  $text_field mark set $common_insert_mark_name $mark_pos
  $text_field configure -state $curr_text_state
}

# it is sort-of arbitrary how to deal with situations where the last inner block ends by the end;
# dealing with it partially different ways for auto-tagging switched on vs. switched off;
proc set_output_insertion_point_inner {} {
  global force_output_insitu
  global force_output_to_end
  # without direct setting of the maxima-wise point, so that this function can be used for other insertions too;
  #global maxima_insert_line
  global text_field
  global text_inner_code_blocks
  global omit_auto_tildes
  global min_tilde_count
  global current_command_placed_beneath

  set use_force_output_insitu $force_output_insitu
  set use_force_output_to_end $force_output_to_end
  if {$current_command_placed_beneath} {
    set use_force_output_insitu true
    set use_force_output_to_end false
  }

  # to have current locations of inner-computation blocks
  seek_inner_code

  set intron_prepend_part ""
  set end_index_info [split [$text_field index "end -1 chars"] "."]
  set ins_index_info [split [$text_field index "insert"] "."]

  # when "in situ" and text ends there, and that ending is without newline,
  # having to put "\n" at intron start, since otherwise it gets appended onto that line;
  if {([lindex $end_index_info 1]!=0) && ([lindex $end_index_info 0]==[lindex $ins_index_info 0])} {
    if {$use_force_output_insitu} {
      # if not at (the last) block, add trailing "\n"
      set intron_prepend_part "\n"
    }
  }

  set maxima_insert_line -1
  set tilde_line_min [string repeat "~" $min_tilde_count]

  set placed_at_end false
  if {[lindex $end_index_info 0] == [lindex $ins_index_info 0]} {
    if {[string trim [$text_field get "insert linestart" "insert lineend"]]==""} {
      set placed_at_end true
    }
  }

  if {(!$use_force_output_to_end) && (!$placed_at_end)} {
    set use_insitu $use_force_output_insitu
    # the cases that end up with "maxima_insert_line!=-1" have tildes dealt with here
    set cur_pos [split [$text_field index "insert"] "."]
    set cur_line [lindex $cur_pos 0]
    set cur_line_plus $cur_line
    incr cur_line_plus
    set cur_line_plus_plus $cur_line_plus
    incr cur_line_plus_plus
    set cur_line_minus $cur_line
    incr cur_line_minus -1
    set is_empty false
    set is_effectively_empty false
    set is_tilde_line false
    set is_space_line false
    if {[lindex $cur_pos 1]==0} {
      set end_pos [split [$text_field index "$cur_line.end"] "."]
      if {[lindex $end_pos 1]==0} {
        set is_empty true
        set is_effectively_empty true
      }
    }
    # effectively empty lines are supposed to be: empty, with space-like chars, ~-lines
    if {!$is_effectively_empty} {
      set insert_line_content [$text_field get "insert linestart" "insert lineend"]
      if {[string trim $insert_line_content]==""} {
        set is_effectively_empty true
        set is_space_line true
      } {
        if {[string trim $insert_line_content "~"] in [list "" "\n"]} {
          set is_effectively_empty true
          set is_tilde_line true
        }
      }
    }
    if {$is_effectively_empty} {
      if {!$is_empty} {
        set use_insitu true
      } {
        set maxima_insert_line $cur_line
        # at empty line, yet to check if being within inner blocks
        if {!$omit_auto_tildes} {
          set is_within false
          set ends_above false
          set starts_below false
          foreach one_block $text_inner_code_blocks {
            set one_block_start [lindex $one_block 0]
            set one_block_end [lindex $one_block 1]
            if {$one_block_start==$cur_line_plus} {
              set starts_below true
              break
            }
            if {$one_block_start>$cur_line} {
              break
            }
            if {$one_block_end=="end"} {
              set is_within true
              break
            }
            if {$one_block_end>=$cur_line} {
              set is_within true
              break
            }
            if {$one_block_end==$cur_line_minus} {
              set ends_above true
            }
          }
          if {!$is_within} {
            if {$ends_above} {
              # put above, to not make pair of ~-lines
              set intron_index "$cur_line_minus.0"
              set intron_part "\n"
              set settle_line $cur_line_minus
            } elseif {$starts_below} {
              # put below, to not make pair of ~-lines
              set intron_index "$cur_line_plus_plus.0"
              set intron_part "\n"
              set settle_line $cur_line_plus_plus
            } {
              # put tildes in
              set intron_index "$cur_line.0"
              set intron_part "$tilde_line_min\n"
              set settle_line $cur_line_plus
              if {![position_effectively_at_end $text_field $cur_line 0]} {
                set intron_part "$tilde_line_min\n\n$tilde_line_min"
              }
            }
            set curr_txt_state [$text_field cget -state]
            $text_field configure -state "normal"
            $text_field insert $intron_index "$intron_prepend_part$intron_part"
            $text_field mark set "insert" "$settle_line.0"
            $text_field configure -state $curr_txt_state
            set maxima_insert_line $settle_line
          }
        }
      }
    }
    if {($maxima_insert_line==-1) && $use_insitu} {
      # check the line position within inner-computation blocks;
      # if the line is tildes-based block end, put stuff above,
      # otherwise putting it below the current line;
      set is_within false
      set put_above false
      set ends_above false
      set starts_below false
      foreach one_block $text_inner_code_blocks {
        set one_block_start [lindex $one_block 0]
        set one_block_end [lindex $one_block 1]
        if {$one_block_start==$cur_line_plus} {
          set starts_below true
          break
        }
        if {$one_block_start>$cur_line} {
          break
        }
        if {$one_block_end=="end"} {
          set is_within true
          break
        }
        if {$one_block_end>$cur_line} {
          set is_within true
          break
        }
        if {$one_block_end==$cur_line_minus} {
          set ends_above true
        }
        if {$one_block_end==$cur_line} {
          set is_within true
          if {[string trim [$text_field get "$cur_line.0" "$cur_line.end"] "~"] in [list "" "\n"]} {
            set put_above true
          }
          break
        }
      }
      set delete_current false
      # not within => possible ~-line there is short, not a block separator => can be replaced
      if {(!$is_within) && ($is_tilde_line || $is_space_line) && (!$omit_auto_tildes)} {
        # would create ~-line pair otherwise
        if {!$ends_above} {
          set delete_current true
        }
      }
      set intron_index "$cur_line_plus.0"
      set intron_part "\n"
      set settle_line $cur_line_plus
      if {$put_above} {
        # notice that "put_above" being true implies "is_within" is true
        set intron_index "$cur_line.0"
        set intron_part "\n"
        set settle_line $cur_line
      }
      if {(!$is_within) && (!$omit_auto_tildes)} {
        set settle_line [expr {$cur_line_plus+1}]
        # check whether the line is effectively at the end
        if {(!$put_above) && (![position_effectively_at_end $text_field $cur_line_plus 0])} {
          set intron_part "$tilde_line_min\n\n$tilde_line_min\n"
          if {$starts_below} {
            append intron_part "\n"
          }
        } {
          set intron_part "$tilde_line_min\n\n"
        }
      }
      set curr_txt_state [$text_field cget -state]
      $text_field configure -state "normal"
      $text_field insert $intron_index "$intron_prepend_part$intron_part"
      $text_field mark set "insert" "$settle_line.0"
      if {$delete_current} {
        $text_field delete "$cur_line.0" "$cur_line_plus.0"
      }
      $text_field configure -state $curr_txt_state
      set maxima_insert_line $settle_line
      if {$delete_current} {
        incr maxima_insert_line -1
      }
    }
  }

  if {$maxima_insert_line==-1} {
    set settle_line "end"
    set intron_index "end"
    set intron_part ""
    if {[$text_field get "end - 2 chars"]!="\n"} {
      set intron_part "\n"
    }
    set last_block_end [lindex [lindex $text_inner_code_blocks "end"] 1]
    set is_within false
    if {$last_block_end=="end"} {
      set is_within true
    }
    # if the last inner-computation block ends at the end-line via a ~-line,
    # should put it just above it (would be strange otherwise)
    set end_line [lindex [split [$text_field index "end - 1 chars"] "."] 0]
    if {$last_block_end==$end_line} {
      if {[string trim [$text_field get "$end_line.0" "$end_line.end"] "~"] in [list "" "\n"]} {
        set is_within true
        set intron_index "$end_line.0"
        set intron_part "\n"
        set settle_line $end_line
      }
    }
    # testing a possibility that the last block ends with ~-line and a new line,
    # with the last line in the text being empty then;
    # here putting it within the block too
    set end_line_minus [expr {$end_line - 1}]
    if {$last_block_end==$end_line_minus} {
      # if auto-tagging switched off and explicitely placed below, leaving it there though;
      if {(!$omit_auto_tildes) || (!$placed_at_end)} {
        if {[string trim [$text_field get "$end_line_minus.0" "$end_line_minus.end"] "~"] in [list "" "\n"]} {
          if {[string trim [$text_field get "$end_line.0" "$end_line.end"]]==""} {
            set is_within true
            set intron_index "$end_line_minus.0"
            set intron_part "\n"
            set settle_line $end_line_minus
          }
        }
      }
    }

    if {(!$is_within) && (!$omit_auto_tildes)} {
      # adding a single ~-line to text end
      append intron_part "$tilde_line_min\n"
    }

    set curr_txt_state [$text_field cget -state]
    $text_field configure -state "normal"
    if {$intron_part!=""} {
      $text_field insert $intron_index "$intron_prepend_part$intron_part"
    }
    if {$settle_line=="end"} {
      $text_field mark set "insert" "end"
    } {
      $text_field mark set "insert" "$settle_line.0"
    }
    $text_field configure -state $curr_txt_state
    # insert line might had changed from -1 into inner one
    if {$settle_line!="end"} {
      set maxima_insert_line $settle_line
    }

  }

  return $maxima_insert_line
}

set last_cpu_usage_value ""
set last_cpu_usage_time ""

proc reset_cpu_usage {} {
  global last_cpu_usage_value
  global last_cpu_usage_time
  global last_cpu_usage_info
  global last_cpu_usage_high

  set last_cpu_usage_value ""
  set last_cpu_usage_time ""

  set last_cpu_usage_info ""
  set last_cpu_usage_high false
}

set to_avoid_updates_for_cpu_usage true
proc set_ignore_cpu_usage {value} {
  global to_avoid_updates_for_cpu_usage

  set to_avoid_updates_for_cpu_usage $value
  if {$to_avoid_updates_for_cpu_usage} {
    reset_cpu_usage
  }
}
proc get_ignore_cpu_usage {} {
  global to_avoid_updates_for_cpu_usage
  return $to_avoid_updates_for_cpu_usage
}

set outer_tmp_dir_path_in_shell "/tmp/"
set outer_tmp_dir_path [strcat "Z:" $outer_tmp_dir_path_in_shell]

set outer_proc_stat_path_base "Z:/proc/"

set outer_tmp_dir_path_input ""
set outer_tmp_dir_path_output ""
set outer_tmp_dir_path_done ""

set get_outer_pid_final_id ""
set outer_tmp_dir_dealing_onset_time ""

set maxima_batch_pid_outer ""

proc clean_outer_pid_onset_files {} {
  global outer_tmp_dir_path_input
  global outer_tmp_dir_path_output
  global outer_tmp_dir_path_done

  if {$outer_tmp_dir_path_input != ""} {
    catch {file delete $outer_tmp_dir_path_input}
  }
  if {$outer_tmp_dir_path_output != ""} {
    catch {file delete $outer_tmp_dir_path_output}
  }
  if {$outer_tmp_dir_path_done != ""} {
    catch {file delete $outer_tmp_dir_path_done}
  }

  set outer_tmp_dir_path_input ""
  set outer_tmp_dir_path_output ""
  set outer_tmp_dir_path_done ""
}

proc get_outer_pid_final {} {
  global get_outer_pid_final_id
  global outer_tmp_dir_path_input
  global outer_tmp_dir_path_output
  global outer_tmp_dir_path_done
  global outer_tmp_dir_dealing_onset_time
  global maxima_batch_pid_outer

  if {$get_outer_pid_final_id != ""} {
    catch {after cancel $get_outer_pid_final_id}
    set get_outer_pid_final_id ""
  }

  if {($outer_tmp_dir_path_output=="") || ($outer_tmp_dir_path_done=="") || ($outer_tmp_dir_dealing_onset_time=="")} {
    return
  }

  set got_pid ""
  set fh ""
  if {[catch {
    set fh [open $outer_tmp_dir_path_done "r"]
    chan configure $fh -translation "lf"
    set line [string trim [gets $fh]]
    close $fh
    set fh ""
    if {[string length $line]>0} {
      set fh [open $outer_tmp_dir_path_output "r"]
      chan configure $fh -translation "lf"
      set line_idx 0
      while {true} {
        incr line_idx
        if {$line_idx>100} {
          break
        }
        set got_line [string trim [gets $fh]]
        if {$got_line!=""} {
          set got_pid $got_line
        } {
          break
        }
      }
      close $fh
    }
  }]} {
    if {$fh != ""} {
      catch {close $fh}
    }
    clean_outer_pid_onset_files
    return
  }

  if {$got_pid != ""} {
    if {[string is digit -strict $got_pid] && [string is ascii -strict $got_pid]} {
      set maxima_batch_pid_outer $got_pid
      set cpu_usage_updates_id [after 100 "update_current_cpu_usage"]
    }
    clean_outer_pid_onset_files
    return
  }

  if {[clock milliseconds] < [expr {$outer_tmp_dir_dealing_onset_time + 10000}]} {
    set get_outer_pid_final_id [after 100 "get_outer_pid_final"]
    return
  }

  clean_outer_pid_onset_files
  set outer_tmp_dir_dealing_onset_time ""
}

proc get_outer_pid_onset {} {
  global outer_tmp_dir_path_in_shell
  global outer_tmp_dir_path
  global outer_tmp_dir_path_input
  global outer_tmp_dir_path_output
  global outer_tmp_dir_path_done
  global outer_rnd_part_identifier
  global get_outer_pid_final_id
  global outer_tmp_dir_dealing_onset_time
  global maxima_batch_pid_outer

  set maxima_batch_pid_outer ""

  if {$outer_tmp_dir_path_input != ""} {
    catch {file delete $outer_tmp_dir_path_input}
  }
  if {$outer_tmp_dir_path_output != ""} {
    catch {file delete $outer_tmp_dir_path_output}
  }
  if {$outer_tmp_dir_path_done != ""} {
    catch {file delete $outer_tmp_dir_path_done}
  }

  clean_outer_pid_onset_files

  if {$get_outer_pid_final_id != ""} {
    catch {after cancel $get_outer_pid_final_id}
    set get_outer_pid_final_id ""
  }

  if {![is_under_wine]} {
    return
  }
  if {$outer_rnd_part_identifier == ""} {
    return
  }

  set fn_start "wmp"
  set rnd_part ""

  set test_rank 0
  while {true} {
    set rnd_part ""
    incr test_rank
    if {$test_rank > 1000} {
      break
    }

    set rnd_part [expr {int(10000*(rand()+1))}]
    if {[catch {
      set count_in [llength [glob -nocomplain -directory $outer_tmp_dir_path [strcat $fn_start $rnd_part "*"]]]
    }]} {
      set rnd_part ""
      break
    }
    if {$count_in > 0} {
      set rnd_part ""
      continue
    }
    break
  }
  if {$rnd_part == ""} {
    return
  }

  set txt_fh ""
  set shell_fh ""
  set done_fh ""
  if {[catch {
    set outer_tmp_dir_path_input [strcat $outer_tmp_dir_path $fn_start $rnd_part ".sh"]
    set outer_tmp_dir_path_input_in_shell [strcat $outer_tmp_dir_path_in_shell $fn_start $rnd_part ".sh"]
    set outer_tmp_dir_path_output [strcat $outer_tmp_dir_path $fn_start $rnd_part ".txt"]
    set outer_tmp_dir_path_output_in_shell [strcat $outer_tmp_dir_path_in_shell $fn_start $rnd_part ".txt"]
    set outer_tmp_dir_path_done [strcat $outer_tmp_dir_path $fn_start $rnd_part ".set"]
    set outer_tmp_dir_path_done_in_shell [strcat $outer_tmp_dir_path_in_shell $fn_start $rnd_part ".set"]

    set txt_fh [open $outer_tmp_dir_path_output "w"]
    close $txt_fh
    set shell_fh [open $outer_tmp_dir_path_input "w"]
    chan configure $shell_fh -translation "lf"
    puts $shell_fh "#!/bin/sh"
    puts $shell_fh ""
    puts $shell_fh "find -L /proc -type f -maxdepth 2 -mindepth 2 -executable 2>/dev/null |xargs ls -l 2>/dev/null |grep wine |sed -E 's/.*\\/proc\\/(\[0-9\]+)\\/exe.*\$/\\/proc\\/\\1\\/cmdline/' |xargs grep \"$outer_rnd_part_identifier\" |sed -E 's/.*\\/proc\\/(\[0-9\]+)\\/cmdline.*\$/\\/proc\\/\\1\\/cmdline/' |xargs grep -a -c -l -v \"cmd.exe\" |sed  -E 's/.*\\/proc\\/(\[0-9\]+)\\/cmdline.*\$/\\1/' > $outer_tmp_dir_path_output_in_shell"
    puts $shell_fh ""
    puts $shell_fh "echo \"done\" > $outer_tmp_dir_path_done_in_shell"
    puts $shell_fh ""
    close $shell_fh
    set done_fh [open $outer_tmp_dir_path_done "w"]
    close $done_fh

    exec "cmd.exe" "/c" "start" "/unix" "/bin/sh" "$outer_tmp_dir_path_input_in_shell" ">nul" "2>nul" "&"
  }]} {
    if {$shell_fh != ""} {
      catch {close $shell_fh}
    }
    if {$txt_fh != ""} {
      catch {close $txt_fh}
    }
    if {$done_fh != ""} {
      catch {close $done_fh}
    }
    clean_outer_pid_onset_files
    return
  }

  set outer_tmp_dir_dealing_onset_time [clock milliseconds]
  set get_outer_pid_final_id [after 100 "get_outer_pid_final"]
}

proc get_current_cpu_usage {} {
  global minima_uses_twapi
  global last_cpu_usage_value
  global last_cpu_usage_time
  global maxima_pid
  global maxima_pid_batch
  global system_clock_ticks
  global maxima_batch_pid_outer
  global outer_proc_stat_path_base

  set use_outer_proc false
  set use_pid [expr {[is_stack_mode] ? $maxima_pid : $maxima_pid_batch}]
  if {[is_under_wine] && ($maxima_batch_pid_outer != "")} {
    set use_pid $maxima_batch_pid_outer
    set use_outer_proc true
  }

  if {$use_pid==""} {
    set last_cpu_usage_value ""
    set last_cpu_usage_time ""
    return ""
  }

  set taking_info_first_time false
  if {($last_cpu_usage_value=="") || ($last_cpu_usage_time=="")} {
    set taking_info_first_time true
  }
  set current_cpu_usage_time [clock milliseconds]
  set time_delta 0
  if {!$taking_info_first_time} {
    set time_delta [expr {$current_cpu_usage_time - $last_cpu_usage_time}]
    if {$time_delta < 0} {
      set time_delta 0
    }
  }
  set last_cpu_usage_time $current_cpu_usage_time

  if {[is_wine_like] && (!$use_outer_proc)} {
    if {!$minima_uses_twapi} {
      return ""
    }
    set cpu_usage ""
    if {[catch {
      set current_cpu_usage_value 0
      foreach one_time_info [twapi::get_process_info $use_pid -usertime -privilegedtime] {
        if {[string is digit -strict $one_time_info]} {
          incr current_cpu_usage_value $one_time_info
        }
      }
      set current_cpu_usage_value [expr {$current_cpu_usage_value / 10000.0}]
      set usage_delta 0
      if {!$taking_info_first_time} {
        set usage_delta [expr {$current_cpu_usage_value - $last_cpu_usage_value}]
        if {$usage_delta<0} {
          set usage_delta 0
        }
      }
      set last_cpu_usage_value $current_cpu_usage_value

      if {(!$taking_info_first_time) && ($time_delta!=0)} {
        set cpu_usage [expr {$usage_delta / $time_delta}]
      }
    }]} {
      set last_cpu_usage_value ""
      set last_cpu_usage_time ""
      return ""
    }
    return $cpu_usage
  }

  if {[is_linux_like] || $use_outer_proc} {
    set statfh ""
    set stat_line ""

    if {[catch {
      set path_for_stat [strcat [expr {$use_outer_proc ? $outer_proc_stat_path_base : "/proc/"}] "$use_pid/stat"]
      set statfh [open $path_for_stat "r"]
      set stat_line [gets $statfh]
      close $statfh
    }]} {
      if {$statfh != ""} {
        catch {close $statfh}
      }
      set last_cpu_usage_value ""
      set last_cpu_usage_time ""
      return ""
    }
    if {$stat_line==""} {
      set last_cpu_usage_value ""
      set last_cpu_usage_time ""
      return ""
    }
    set stat_info [list]
    foreach one_part [split $stat_line " "] {
      if {$one_part!=""} {
        lappend stat_info $one_part
      }
    }
    set stat_info [lrange $stat_info 1 "end"]
    set stat_info_idx 1
    if {[string range [lindex $stat_info 0] 0 0]=="("} {
      foreach one_part $stat_info {
        if {[string range $one_part "end" "end"]==")"} {
          break
        }
        incr stat_info_idx
      }
    }
    set stat_info [lrange $stat_info $stat_info_idx "end"]

    if {[llength $stat_info]<15} {
      set last_cpu_usage_value ""
      set last_cpu_usage_time ""
      return ""
    }
    set current_cpu_usage_value 0
    if {[catch {
      # man 5 proc: utime, stime, cutime, cstime (here after two first fields removed)
      incr current_cpu_usage_value [lindex $stat_info 11]
      incr current_cpu_usage_value [lindex $stat_info 12]
      incr current_cpu_usage_value [lindex $stat_info 13]
      incr current_cpu_usage_value [lindex $stat_info 14]
      set current_cpu_usage_value [expr {($current_cpu_usage_value * 1000.0) / $system_clock_ticks}]
    }]} {
      set last_cpu_usage_value ""
      set last_cpu_usage_time ""
      return ""
    }

    set usage_delta 0
    if {!$taking_info_first_time} {
      set usage_delta [expr {$current_cpu_usage_value - $last_cpu_usage_value}]
      if {$usage_delta<0} {
        set usage_delta 0
      }
    }
    set last_cpu_usage_value $current_cpu_usage_value

    if {$taking_info_first_time || ($time_delta==0)} {
      return ""
    }
    return [expr {$usage_delta / $time_delta}]
  }
}

set cpu_usage_updates_id ""
set cpu_usage_high_threshold 0.1
proc update_current_cpu_usage {} {
  global last_cpu_usage_info
  global last_cpu_usage_high
  global cpu_usage_high_threshold
  global cpu_usage_updates_id

  if {$cpu_usage_updates_id!=""} {
    catch {after cancel $cpu_usage_updates_id}
    set cpu_usage_updates_id ""
  }

  if {[get_ignore_cpu_usage]} {
    set last_cpu_usage_info ""
    set last_cpu_usage_high false

    set cpu_usage_updates_id [after 1000 "update_current_cpu_usage"]
    return
  }

  set last_cpu_usage_info [get_current_cpu_usage]
  set last_cpu_usage_high false
  if {$last_cpu_usage_info!=""} {
    if {$last_cpu_usage_info >= $cpu_usage_high_threshold} {
      set last_cpu_usage_high true
    }
    set last_cpu_usage_info [strcat "CPU " [expr {int(100 * $last_cpu_usage_info + 0.5)}] "%"]
  }
  update_state_label_text
  if {[is_batch_mode]} {
    state_batch_maxima
  }

  set cpu_usage_updates_id [after 1000 "update_current_cpu_usage"]
}

proc stop_cpu_usage_updates {} {
  global cpu_usage_updates_id
  global last_cpu_usage_info
  global last_cpu_usage_high
  global maxima_batch_pid_outer

  if {$cpu_usage_updates_id!=""} {
    catch {after cancel $cpu_usage_updates_id}
    set cpu_usage_updates_id ""
  }

  # can not reset the rnd-part, since using this stop-wise cleaning for starts too;
  set maxima_batch_pid_outer ""

  set last_cpu_usage_info ""
  set last_cpu_usage_high false
}

# should avoid cpu-dealing at non-procfs/non-reactos systems
proc start_cpu_usage_updates {} {
  global minima_uses_twapi
  global cpu_usage_updates_id

  stop_cpu_usage_updates

  if {[is_wine_like]} {
    if {[is_under_wine]} {
      return
    }
    if {$minima_uses_twapi} {
      set cpu_usage_updates_id [after 200 "update_current_cpu_usage"]
    }
    return
  }
  set proc_dir "/proc"
  if {[file exists $proc_dir] && [file isdirectory $proc_dir]} {
    if {[catch {
      set dcount [llength [glob -nocomplain -directory $proc_dir -types [list "d"] [pid]]]
    }]} {
      set dcount 0
    }
    # if active procfs, at least this process should be in $proc_dir
    if {$dcount>0} {
      set cpu_usage_updates_id [after 200 "update_current_cpu_usage"]
    }
  }
}
catch {
  if {(![is_batch_mode]) && (![is_under_wine])} {
    # batch mode should only have this active when a batch runs
    # and under wine, this should wait until the outer PID can be got;
    after 200 "start_cpu_usage_updates"
  }
}

proc start_cpu_usage_updates_under_wine {} {
  if {![is_under_wine]} {
    return
  }
  stop_cpu_usage_updates

  after 200 "get_outer_pid_onset"
}

set has_ever_asked false
set maxima_insert_line -1
set newly_at_maxima_outputs false
set maxima_command_at_debug false
proc exec_max_command {command {command_inner ""}} {
  global maxima_channel
  global text_field
  global command_field
  global maxima_command
  global maxima_command_inner
  global has_ever_asked
  global state_label
  global bt_restart
  global maxima_ready
  global in_lisp_mode
  global in_debug_mode
  global command_history
  global newly_at_maxima_outputs
  #global can_kill_label
  global bind_labels
  global maxima_command_at_debug
  global curr_stack_mode
  global maxima_pid
  global priority_wine_common
  global priority_wine_idle_set
  global minima_uses_twapi
  global should_display_cpu_usage
  global could_display_cpu_usage

  if {!$maxima_ready} {
    return false
  }
  if {$maxima_channel==""} {
    return false
  }
  if {![check_connected]} {
    return false
  }

  set maxima_command $command
  set maxima_command_inner $command_inner

  set maxima_command [string trim $maxima_command]
  if {[string length $maxima_command]==0} {
    return false
  }

  if {!$in_debug_mode} {
    set should_display_cpu_usage true
  }
  set could_display_cpu_usage true

  if {[is_wine_like] && $priority_wine_idle_set && $minima_uses_twapi} {
    if {$maxima_pid != ""} {
      catch {
        twapi::set_priority_class $maxima_pid $priority_wine_common
        set priority_wine_idle_set false
      }
    }
  }

  $text_field edit separator
  $command_field edit separator

  # * the add_end part is left false when in_debug_mode is true;
  #   and since at some cases we do not detect end of debug mode
  #   fast enough (e.g. on ":top"-based debug end) thus adding ";"
  #   there too, though it is done at the writing-to-maxima phase;
  # * notice that debugger commands when issued at common maxima mode
  #   (regardless whether debugger is enabled by debugmode(true));
  #   do not lead to emitting of any text (neither eval-time info),
  #   and thus unless labels are se to be shown, it gets stalled on that
  #   (would probably need to redefine some maxima stuff to solve it);

  set curr_time [clock format [clock seconds] -gmt true -format "%Y-%m-%d %H:%M:%S"]
  lappend command_history [list "$curr_time GMT" $maxima_command]

  stack_on_exec_start

  if {!$in_debug_mode} {
    set maxima_ready false
    # should be able to do save during command running,
    # if e.g. not having saved doc before starting a long-time run command;
    set_state_disabled false
    set_button_interrupt_exec
    $bt_restart configure -state normal
    set_state_label_text "computing"
    set maxima_command_at_debug false
  } {
    if {$curr_stack_mode!=0} {
      stack_on_exec_end
      if {$curr_stack_mode==1} {
        set curr_cmmd_state [$command_field cget -state]
        $command_field configure -state "normal"
        $command_field delete 1.0 end
        $command_field edit reset
        $command_field edit modified false
        $command_field configure -state $curr_cmmd_state
      }
    }
    set maxima_command_at_debug true
  }

  #if {$bind_labels} {
  #  set can_kill_label false
  #}

  set has_ever_asked true
  set newly_at_maxima_outputs true
  svwrite_onset

  return true
}

proc make_debug_mode_start {} {
  global shibboleth_onset_read
  global newly_at_maxima_outputs
  global in_debug_mode
  global state_label
  global maxima_ready
  global should_display_cpu_usage
  global could_display_cpu_usage
  global was_in_debug_state

  set_debug_mode true
  set was_in_debug_state true

  if {$shibboleth_onset_read} {
    if {$newly_at_maxima_outputs} {
      set_output_insertion_point
      set newly_at_maxima_outputs false
    }
  }

  set_state_ready true
  set_button_interrupt_idle

  set should_display_cpu_usage false
  set could_display_cpu_usage true
  set_state_label_text "debug"
  set maxima_ready true
}
proc make_debug_mode_end {} {
  global in_debug_mode
  global state_label
  global maxima_ready
  global should_display_cpu_usage
  global could_display_cpu_usage

  set_debug_mode false
  set_state_ready true
  set_button_interrupt_idle

  set should_display_cpu_usage false
  set could_display_cpu_usage true
  set_state_label_text "ready"
  set maxima_ready true
}

proc clean_control_parts {} {
  global maxima_in_setting
  global maxima_in_setting_start
  global maxima_setting_done
  global waiting_for_control_answer
  global control_names_updates
  global in_name_asking
  global within_names_update

  set maxima_in_setting false
  set maxima_in_setting_start false
  set maxima_setting_done false
  set waiting_for_control_answer false
  set control_names_updates false
  set in_name_asking false
  set within_names_update false
}

# when (finally) got some answer (part) on a control command
proc make_setup_on_new_control_output {} {
  global maxima_in_setting
  global maxima_in_setting_start
  global maxima_setting_done
  global newly_at_maxima_outputs
  global waiting_for_control_answer
  global control_names_updates
  global in_name_asking
  global within_names_update
  global control_names_updates

  set_button_interrupt_sets
  set waiting_for_control_answer false
  set newly_at_maxima_outputs false

  if {$control_names_updates} {
    # when control action starts with name asking
    set in_name_asking true
    set within_names_update true
    set control_names_updates false
    return
  }

  set maxima_in_setting true
  set maxima_in_setting_start true
  set maxima_setting_done false
}

set shibboleth_colon_spacer [strcat \
    [expr {int(1000+1000*rand())}] \
    [expr {int(1000+1000*rand())}] \
    [expr {int(1000+1000*rand())}] \
    [expr {int(1000+1000*rand())}] \
  ]

set read_wait_id 0
set next_with_empty_line false
# here the comparisons go by lines, thus can use "\\s" that would otherwise match newlines too;
# the seen regexps are "Evaluation took [N]+\.NNNN seconds \([N]+\.NNNN elapsed\) ..."
set time_elapsed_pattern "Evaluation*took*\[0-9\]*seconds*(\[0-9\]*elapsed)*"
set time_elapsed_regexps "^Evaluation\[\\s\]*took\[\\s\]*\[0-9\\.,\]\[^\$\]*seconds\[^\$\]*\\(\[0-9\\.,\]\[^\$\]*elapsed\\)\[^\$\]*\$"
set time_elapsed_regexps_start "^Evaluation\[\\s\]*took\[\\s\]*\[0-9\\.,\]\[^\$\]*seconds\[^\$\]*\$"
set time_elapsed_regexps_involved "^(?:\\(\\%i\[0-9\]+\\)\\s*)?(Evaluation\[\\s\]*took\[\\s\]*\[0-9\\.,\]\[^\$\]*seconds\[^\$\]*\\(\[0-9\\.,\]\[^\$\]*elapsed\\)\[^\$\]*)\$"
set time_elapsed_regexps_comb "^(\[^\$\]*)(Evaluation\[\\s\]*took\[\\s\]*\[0-9\\.,\]\[^\$\]*seconds\[^\$\]*\\(\[0-9\\.,\]\[^\$\]*elapsed\\)\[^\$\]*)\$"
set input_line_label_regexps "^\\(\\%i\[0-9\]+\\)\\s*\$"

# putting in here the regexps that seem to belong to debuggers
#set debug_mode_line_start_re "^\\(dbm\\:\[0-9\]+\\)"
set debug_mode_line_start_re "^(\\(dbm\\:\[0-9\]+\\))|(dbl:MAXIMA\[>\]+)"
set with_control_output false
proc consume_read_parts {} {
  global read_buffer_onset
  global read_buffer_all
  global read_wait_id
  # not sure how to decide which of with_labels vs. with_labels_asked to use here,
  # thus (until it causes an issue) using both of them;
  global with_labels
  global with_labels_asked
  global show_time_elapsed
  global has_ever_asked
  global in_lisp_mode
  global in_debug_mode
  global command_history
  global in_name_asking
  global maxima_in_setting
  global info_setup_inside
  global within_names_update
  global next_with_empty_line
  global time_elapsed_pattern
  global maxima_output_to_split
  global info_line_long
  global info_line_inner
  global info_line_size
  global debug_mode_line_start_re
  global control_output_onset_shibboleth
  global with_control_output
  global time_elapsed_regexps_comb
  global input_line_label_regexps
  #global shibboleth_colon_spacer

  catch {after cancel $read_wait_id}

  if {$within_names_update} {
    process_answer "" [list $read_buffer_all]
    set read_buffer_all ""
    set next_with_empty_line true
    return
  }
  if {$maxima_in_setting && $has_ever_asked} {
    set read_buffer_onset ""
    if {$with_labels || $with_labels_asked} {
      foreach one_set_out [lreverse [split $read_buffer_all "\n"]] {
        set one_set_out [string trim $one_set_out]
        set one_set_out [string trim $one_set_out "\""]
        if {[string equal -length 3 "(%i" $one_set_out]} {
          set read_buffer_onset $one_set_out
          append read_buffer_onset " "
          if {$in_debug_mode} {
            make_debug_mode_end
            #set_debug_mode false
          }
          break
        }
      }
    }
    process_answer "" [list $read_buffer_all]
    set read_buffer_all ""
    set next_with_empty_line true
    return
  }

  set onset ""
  if {$with_labels || $with_labels_asked || $in_lisp_mode || $in_debug_mode} {
    set onset $read_buffer_onset
  }

  set maxima_output_to_split_len 0
  if {$maxima_output_to_split} {
    set maxima_output_to_split_len [lindex $info_line_size 0]
  }

  set with_control_output_newly false
  set has_to_process_in_parts false

  set answer [list]
  set answer_second [list]
  set dummy_regexp_match ""
  # can not put "update" here, since new (fast-issued) commands would destroy these data
  foreach answer_line [split $read_buffer_all "\n"] {
    # it seems that at wine-like systems we get lines ended by "\r\n" from Maxima
    # while Tcl/Tk apparently wants just "\n" ending even at the wine-like system
    set answer_line [string trim $answer_line "\r"]
    # ignoring auxiliary parts (if used)
    #if {$answer_line==$shibboleth_colon_spacer} {
    #  continue;
    #}

    # while startup errors in GCL lead to lisp-like prompts, the other cases are left without that
    if {(!$in_debug_mode) && (!$has_ever_asked) && ([string length $answer_line]<400)} {
      set answer_line_lc [string tolower $answer_line]
      foreach debug_test_string [list "debugger invoked" "type help for debugger help" "maxima top-level" "maxima-lisp-debugger"] {
        if {[string first $debug_test_string $answer_line_lc]>-1} {
          make_debug_mode_start
          break
        }
      }
    }

    if {!$with_control_output} {
      set answer_line_trimmed [string trim $answer_line]
      # can not compare directly here, since such a comparison with a long line in there takes infinite time
      if {([string length $answer_line_trimmed] == [string length $control_output_onset_shibboleth])} {
        if {($answer_line_trimmed==$control_output_onset_shibboleth)} {
          set with_control_output true
          set with_control_output_newly true
          set_lisp_mode false
          set_debug_mode false

          if {[llength $answer]!=0} {
            # has to regularly process the before-control part first
            set has_to_process_in_parts true
          }
          continue
        }
      }
    }

    set eval_regexp_match ""
    set onset_regexp_match ""
    if {[regexp -nocase -- $debug_mode_line_start_re $answer_line]!=0} {
      if {!$in_debug_mode} {
        make_debug_mode_start
        #set_debug_mode true
      }
    # the time-elapsed info is only shown when (finally) in maxima mode
    } elseif {[regexp -- $time_elapsed_regexps_comb $answer_line dummy_regexp_match onset_regexp_match eval_regexp_match]} {
      # we tend to get the time-eval info put over other stuff;
      # not using $shibboleth_colon_spacer by now;
      #if {($onset_regexp_match!="") && ($onset_regexp_match!=$shibboleth_colon_spacer)} {}
      if {$onset_regexp_match!=""} {
        if {![regexp -- $input_line_label_regexps $onset_regexp_match]} {
          if {$has_to_process_in_parts} {
            lappend answer_second $onset_regexp_match
          } {
            lappend answer $onset_regexp_match
          }
        }
      }
      if {$eval_regexp_match!=""} {
        set answer_line $eval_regexp_match
      }
      if {$has_ever_asked && (!$in_name_asking) && (!$maxima_in_setting) && (!$info_setup_inside) && (!$with_control_output)} {
        lappend command_history [list $answer_line]
      }
      set_lisp_mode false
      if {$in_debug_mode} {
        make_debug_mode_end
        #set_debug_mode false
      }
      if {(!$has_ever_asked) || (!$show_time_elapsed) || ($with_control_output)} {
        continue
      } {
        set answer_line [strcat "/* " $answer_line " */"]
      }
    } {
      if {$maxima_output_to_split && (!$maxima_in_setting) && (!$with_control_output) && $has_ever_asked} {
        set answer_line_len [string length $answer_line]
        set curr_line_offset 0
        while {true} {
          set curr_line_part_end [expr {$curr_line_offset+$maxima_output_to_split_len-1}]
          if {$curr_line_part_end<$answer_line_len} {
            lappend answer [strcat [string range $answer_line $curr_line_offset $curr_line_part_end] "\\"]
            set curr_line_offset [expr {$curr_line_part_end+1}]
          } {
            set answer_line [string range $answer_line $curr_line_offset "end"]
            break
          }
        }
      }
    }

    # if got control output after some late-coming regular output,
    # using the control output as a next part;
    if {$has_to_process_in_parts} {
      lappend answer_second $answer_line
      continue
    } {
      lappend answer $answer_line
    }
  }

  clear_read_buffers

  set has_regular_answer true
  if {$with_control_output && (!$has_to_process_in_parts)} {
    set has_regular_answer false
  }
  if {!$has_regular_answer} {
    if {$with_control_output_newly} {
      make_setup_on_new_control_output
    }
    # to save the (%iN) label
    set last_part [lindex $answer "end"]
    if {$with_labels || $with_labels_asked} {
      if {[regexp {^\(%i[\d]{1,10}\)$} [string trim $last_part]]} {
        set read_buffer_onset $last_part
        set answer [lrange $answer 0 end-1]
        set_lisp_mode false
        if {$in_debug_mode} {
          make_debug_mode_end
          #set_debug_mode false
        }
      }
    }
    process_answer $onset $answer
    return
  }

  set last_part [lindex $answer "end"]

  # when in Lisp mode, we get "MAXIMA[>]+" as a prompt
  # it seems that it may be "Maxima restarted." too;
  if {(!$in_lisp_mode) || $in_debug_mode} {
    foreach lisp_test_string [list "MAXIMA>" "Maxima restarted"] {
      if {[string equal -length [string length $lisp_test_string] $last_part $lisp_test_string]} {
        set_lisp_mode true
        if {$in_debug_mode} {
          make_debug_mode_end
          #set_debug_mode false
        }
      }
    }
  }

  # we (commonly) get "(%iN)" as the last (non-full) line when with labels
  if {$with_labels || $with_labels_asked} {
    if {[regexp {^\(%i[\d]{1,10}\)$} [string trim $last_part]]} {
      set read_buffer_onset $last_part
      set answer [lrange $answer 0 end-1]
      set_lisp_mode false
      if {$in_debug_mode} {
        make_debug_mode_end
        #set_debug_mode false
      }
    }
  } {
    if {([llength $answer]>0) && ($last_part=="")} {
      set answer [lrange $answer 0 "end-1"]
    }
    if {$in_lisp_mode || $in_debug_mode} {
      if {[string first "Returning to Maxima" [lindex $answer "end"]] > -1} {
        set_lisp_mode false
        if {$in_debug_mode} {
          make_debug_mode_end
          #set_debug_mode false
        }
      }
    }
  }

  if {$in_lisp_mode} {
    set last_part [lindex $answer "end"]
    if {[string trimright [string trim $last_part] ">"]=="MAXIMA"} {
      set read_buffer_onset $last_part
      set answer [lrange $answer 0 end-1]
      if {$in_debug_mode} {
        make_debug_mode_end
        #set_debug_mode false
      }
    }
  }

  process_answer $onset $answer

  if {!$has_to_process_in_parts} {
    return
  }

  if {$with_control_output_newly} {
    make_setup_on_new_control_output
  }

  set last_part [lindex $answer_second "end"]
  if {$with_labels || $with_labels_asked} {
    # to save the (%iN) label
    if {[regexp {^\(%i[\d]{1,10}\)$} [string trim $last_part]]} {
      set read_buffer_onset $last_part
      set answer_second [lrange $answer_second 0 end-1]
      set_lisp_mode false
      if {$in_debug_mode} {
        make_debug_mode_end
        #set_debug_mode false
      }
    }
  }

  process_answer "" $answer_second
}

proc clear_read_buffers {} {
  global read_buffer_onset
  global read_buffer_all

  set read_buffer_all ""
  set read_buffer_onset ""
}

set could_not_read_times 0
set could_not_read_times_max_mssgs 3
set read_buffer_onset ""
set read_buffer_all ""
set read_wait_time [expr {[is_wine_like] ? 500 : 100}]
proc svhear {} {
  global maxima_channel
  global state_label
  global read_buffer_onset
  global has_ever_asked
  global read_wait_id
  global read_buffer_all
  global next_with_empty_line
  global could_not_read_times
  global could_not_read_times_max_mssgs
  global read_wait_time
  global print_logging_maxima_output

  catch {after cancel $read_wait_id}

  set channel $maxima_channel
  catch {fileevent $channel readable ""}

  set bad [catch {
    set to_break false
    set curr_line_without_nl false
    while {true} {
      set one_part [chan read $channel]
      if {$print_logging_maxima_output} {
        print_logging $one_part
      }
      if {$next_with_empty_line} {
        set next_with_empty_line false
        set did_subtraction false
        foreach test_prefix [list "\r\n" "\n"] {
          set test_prefix_len [string length $test_prefix]
          if {[string equal -length $test_prefix_len $one_part $test_prefix]} {
            set one_part [string range $one_part $test_prefix_len "end"]
            set did_subtraction true
            break
          }
        }
        if {$did_subtraction} {
          if {[string length $one_part]==0} {
            continue
          }
        }
      }
      if {[string length $one_part]==0} {
        break
      }
      append read_buffer_all $one_part
      if {[chan eof $channel]} {
        break
      }

      continue
    }
  }]
  if {$bad} {
    if {$could_not_read_times < $could_not_read_times_max_mssgs} {
      incr could_not_read_times
      show_text_field_message_error [list [list "can not read data from maxima"]]
    }
    clear_read_buffers
    return
  }

  set to_process_answer false
  set remains_connected true

  if {![check_connected true]} {
    # if got disconnected, process what was received anyway
    set to_process_answer true
    set remains_connected false
  }

  if {$to_process_answer} {
    consume_read_parts
  }

  if {$remains_connected} {
    catch {fileevent $channel readable "svhear"}
    if {!$to_process_answer} {
      if {$has_ever_asked} {
        set read_wait_id [after $read_wait_time "consume_read_parts"]
      } {
        set read_wait_id [after 10 "consume_read_parts"]
      }
    }
  }

  check_connected
}

proc set_maxima_pid {answer} {
  global maxima_pid
  global priority_wine_common
  global priority_wine_idle_set
  global minima_uses_twapi

  set maxima_pid_prev $maxima_pid
  set maxima_pid ""

  set pid_chars_end 0
  set pid_negative false
  if {[string equal -nocase -length 4 $answer "pid="]} {
    set pid_chars_end 4
  }
  if {[string equal -nocase -length 5 $answer "pid=-"]} {
    set pid_chars_end 5
    set pid_negative true
  }
  if {$pid_chars_end>0} {
    while {[string is digit -strict [string range $answer $pid_chars_end $pid_chars_end]]} {
      incr pid_chars_end
    }
    set maxima_pid [string range $answer 4 [expr {$pid_chars_end-1}]]
    if {$pid_negative} {
      set maxima_pid "-$maxima_pid"
    }
    set answer [string trimleft [string range $answer $pid_chars_end end]]
  }

  if {[is_wine_like] && ($maxima_pid!="") && $minima_uses_twapi} {
    catch {
      twapi::set_priority_class $maxima_pid $priority_wine_common
      set priority_wine_idle_set false
    }
  }

  if {$maxima_pid_prev != $maxima_pid} {
    # have to reset cpu-usage info here, since a new value means that we have been tracking the enveloping batch
    if {![is_under_wine]} {
      reset_cpu_usage
    }
  }

  return $maxima_pid
}

proc is_maxima_command_pending {} {
  global maxima_channel

  if {$maxima_channel==""} {
    return false
  }

  set pending_amount 0
  if {[catch {
    set pending_amount [chan pending output $maxima_channel]
  }]} {
    return false
  }

  if {$pending_amount>0} {
    return true
  }

  return false
}

proc svwrite_onset {} {
  global maxima_channel

  if {$maxima_channel==""} {
    return false
  } {
    catch {fileevent $maxima_channel writable ""}
  }

  if {[catch {
    fileevent $maxima_channel writable "svwrite"
  }]} {
    return false
  }

  return true
}

proc svwrite {{to_read 1}} {
  global maxima_command
  global maxima_command_inner
  global maxima_command_control
  global maxima_channel
  global maxima_command_at_debug
  global with_labels
  #global shibboleth_colon_spacer
  global in_debug_mode
  global in_lisp_mode
  global print_logging_maxima_input

  set channel $maxima_channel

  catch {fileevent $channel writable ""}

  set use_command $maxima_command
  set use_command_common true
  if {$use_command!=""} {
    if {$maxima_command_inner!=""} {
      set use_command $maxima_command_inner
    }
    if {$maxima_command_at_debug && (!$with_labels)} {
      # at these cases it may go out of debug mode silently
      set use_command [strcat $use_command ";"]
    }
  } {
    set use_command $maxima_command_control
    set use_command_common false
  }

  catch {
    puts $channel $use_command
    if {$print_logging_maxima_input} {
      print_logging [strcat $use_command "\n"]
    }
    # (an attempt to) workarounds for delayed/none outputs on lisp-wise/debug-wise commands
    # ...it seems that the issues are present even with labels to be displayed;
    if {(!$in_debug_mode) && (!$in_lisp_mode) && ($use_command_common)} {
      # should parse it to commands, this simple way is not general;
      set last_command_line [string trim [lindex [split $use_command "\n"] "end"]]
      if {[string equal -nocase -length 5 ":lisp" $last_command_line]} {
        # for :lisp(...) the format-based form seems to work, though not much for :top, etc. commands
        set output_enforcer1 ":lisp(format nil \"~&\" (force-output));"
        puts $channel $output_enforcer1
        if {$print_logging_maxima_input} {
          print_logging [strcat $output_enforcer1 "\n"]
        }
      } elseif {[string equal -nocase -length 5 ":next" $last_command_line]} {
        set output_enforcer1 ":lisp(format nil \"~%\");"
        set output_enforcer2 ":lisp(format nil \"~*\" (force-output));"
        puts $channel $output_enforcer1
        puts $channel $output_enforcer2
        if {$print_logging_maxima_input} {
          print_logging [strcat $output_enforcer1 "\n"]
          print_logging [strcat $output_enforcer2 "\n"]
        }
      } elseif {[string equal -nocase -length 1 ":" $last_command_line]} {
        set output_enforcer1 [strcat ":lisp(format nil \"~" [expr {$with_labels ? "&" : "%"}] "\");"]
        set output_enforcer2 ":lisp(format nil \"~*\" (force-output));"
        puts $channel $output_enforcer1
        puts $channel $output_enforcer2
        if {$print_logging_maxima_input} {
          print_logging [strcat $output_enforcer1 "\n"]
          print_logging [strcat $output_enforcer2 "\n"]
        }
      }
    }
    flush $channel
  }

  if {!$to_read} {
    set maxima_command ""
    set maxima_command_inner ""
  }

  if {$use_command_common && $maxima_command_at_debug && ($maxima_command!="")} {
    set process_text $maxima_command
    set maxima_command ""
    set maxima_command_inner ""
    process_answer "$process_text" [list]
  }
  set maxima_command_at_debug false
  set maxima_command_control ""

  if {$to_read==1} {
    catch {fileevent $channel readable "svhear"}
  } {
    catch {fileevent $channel readable ""}
  }

}

set maxima_channel ""
proc service {channel address port} {
  global maxima_channel
  global svsocket
  global svport
  global in_lisp_mode
  global in_debug_mode
  global can_deal_with_disconnect
  global could_not_read_times
  global shibboleth_onset_postponed
  global was_in_debug_state

  set was_in_debug_state false
  set could_not_read_times 0
  set shibboleth_onset_postponed false

  set maxima_channel $channel

  # not sure which default setting of lisp_mode leads to lesser amount of issues
  #set_lisp_mode true
  set_lisp_mode false
  set_debug_mode false

  catch {
    fconfigure $channel -blocking false
    fconfigure $channel -encoding "utf-8"
    fconfigure $channel -translation lf
  }
  # we do explicit flushing when writing to maxima
  #fconfigure $channel -buffering line

  clear_read_buffers

  set can_deal_with_disconnect true
  reset_cpu_usage
  set_ignore_cpu_usage false

  catch {fileevent $channel readable "svhear"}

  setup_checks

  if {$svsocket!=""} {
    catch {chan close $svsocket}
    set svsocket ""
    set svport -1
  }
}

set maxconn 0
set svsocket ""
set svport -1
proc server_start {} {
  global svport
  global svsocket
  global maxconn
  global state_label
  global after_restart
  global should_display_cpu_usage
  global could_display_cpu_usage

  set svport [::tcl::mathfunc::int [expr {20000 + 20000*[::tcl::mathfunc::rand]}]]

  set bad 1
  set attempts 0
  while {$bad} {
    set bad [catch {set svsocket [socket -server service -myaddr "127.0.0.1" $svport]}]
    if {$bad} {
      set svport [::tcl::mathfunc::int [::tcl::mathfunc::fmod [expr {$svport+1}] 60000]]
      if {$svport<10000} {set svport 10000}
    }
    incr attempts
    if {$attempts>100000} {
      set should_display_cpu_usage false
      set could_display_cpu_usage false
      set_state_label_text "not connected"
      if {$after_restart} {
        show_text_field_message_error [list [list "can not start server for maxima"]]
      }
      break
    }
  }

  if {$bad} {
    set maxconn 0
    set svsocket ""
  } {
    set maxconn 1
  }

  return $maxconn
}

proc do_kill_maxima_done {} {
  global maxima_pid
  global maxima_got_killed
  global maxima_pid_batch

  set use_pid $maxima_pid
  if {[is_batch_mode]} {
    set use_pid $maxima_pid_batch
  }

  if {[is_wine_like]} {
    catch {exec "taskkill" "/f" "/pid" $use_pid}
  } {
    catch {exec "kill" "-s" "SIGKILL" $use_pid}
  }

  set maxima_got_killed true
}

# Priority classes for Maxima running at ReactOS
# using "below_normal" priority class to make ui usable
set priority_wine_common [expr {0 + 0x04000}]
#set priority_wine_common [expr {0 + 0x020}]
set priority_wine_idle [expr {0 + 0x040}]
set priority_wine_idle_set false

set maxima_got_killed false
proc do_kill_maxima {really} {
  global maxima_pid
  global in_lisp_mode
  global in_debug_mode
  global maxima_got_killed
  global maxima_pid_batch
  global maxima_channel_hnd
  global priority_wine_idle
  global priority_wine_idle_set
  global minima_uses_twapi
  global batch_maxima_handle

  # it chokes ReactOS when CTRL-C triggered while Maxima run over batch
  if {(!$really) && (![has_common_ctrlc])} {
    retrun false
  }

  set use_pid $maxima_pid
  if {[is_batch_mode]} {
    set use_pid $maxima_pid_batch
  }

  if {$use_pid==""} {
    if {$maxima_channel_hnd!=""} {
      clear_maxima_channel_hnd
    }
    return false
  }

  if {[is_wine_like] && $minima_uses_twapi} {
    catch {
      twapi::set_priority_class $use_pid $priority_wine_idle
      set priority_wine_idle_set true
    }
  }

  if {[is_wine_like]} {
    if {$really} {
      # under wine we apparently need to kill it without "/t" that is not implemented there
      # and not sure if should use the "/t" option at first;
      set bad [catch {exec "taskkill" "/pid" $use_pid}]
      set maxima_got_killed false
      after 100 "do_kill_maxima_done"
      if {!$maxima_got_killed} {
        vwait maxima_got_killed
      }
      set_lisp_mode false
      if {$maxima_channel_hnd!=""} {
        clear_maxima_channel_hnd
      }
      return [expr {!$bad}]
    } {
      if {$minima_uses_twapi} {
        catch {twapi::set_console_control_handler wine_ctrlc_trap}
        update
        set bad [catch {twapi::generate_console_control_event ctrl-c}]
        update
        if {$batch_maxima_handle!=""} {
          catch {puts $batch_maxima_handle ""}
        }
        return [expr {!$bad}]
      } {
        return false
      }
    }
  } {
    if {$really} {
      set bad [catch {exec "kill" "-s" "SIGTERM" $use_pid}]
      set maxima_got_killed false
      after 100 "do_kill_maxima_done"
      if {!$maxima_got_killed} {
        vwait maxima_got_killed
      }
      set_lisp_mode false
      if {$maxima_channel_hnd!=""} {
        clear_maxima_channel_hnd
      }
    } {
      set bad [catch {exec "kill" "-s" "SIGINT" $use_pid}]
      if {$batch_maxima_handle!=""} {
        catch {puts $batch_maxima_handle ""}
      }
    }
    return [expr {!$bad}]
  }
}

set maxima_stderr_check_after ""
set maxima_stderr_should_check false
set maxima_stderr_should_ignore false
proc check_maxima_stderr {} {
  global maxima_stderr_path
  global maxima_stderr_should_check
  global maxima_stderr_should_ignore
  global maxima_stderr_path
  global maxima_stderr_read_id
  global check_maxima_stderr_after

  set read_parts []
  if {$maxima_stderr_should_check && (!$maxima_stderr_should_ignore)} {
    if {($maxima_stderr_read_id=="") && ($maxima_stderr_path!="")} {
      if [catch {set maxima_stderr_read_id [open $maxima_stderr_path "r"]}] {
        set maxima_stderr_read_id ""
        set maxima_stderr_should_ignore true
      }
    }

    if {$maxima_stderr_read_id!=""} {
      set read_count_max 20
      set read_count_max_skip 100
      set read_count_other 0
      set read_data_max 1000000
      catch {
        set read_line_idx 0
        foreach one_part [split [read $maxima_stderr_read_id $read_data_max] "\n"] {
          set one_part [string trim $one_part]
          if {$one_part!=""} {
            if {$read_line_idx < $read_count_max} {
              lappend read_parts [list $one_part]
            } {
              set maxima_stderr_should_ignore true
              incr read_count_other
              if {$read_count_other>$read_count_max_skip} {
                set "more than $read_count_max_skip"
                break
              }
            }
            incr read_line_idx
          }
        }
      }
      if {[llength $read_parts]>0} {
        if {$read_count_other!=0} {
          set part_word [lindex [list "line" "lines"] [expr {$read_count_other==1 ? 0 : 1}]]
          lappend read_parts [list "($read_count_other $part_word skipped)"]
        }
        show_text_field_message_error $read_parts
        set read_parts []
      }
    }
  }

  set check_maxima_stderr_after [after 1000 {check_maxima_stderr}]
}
set check_maxima_stderr_after ""
if {$display_maxima_stderr} {
  set check_maxima_stderr_after [after 1000 {check_maxima_stderr}]
}

proc set_editor_batch_state {} {
  global editor_batch_state
  global bt_file_open
  global bt_file_import
  global text_field
  global search_entry
  global state_label
  global bt_interrupt

  if {!$editor_batch_state} {
    return false
  }

  set_state_disabled

  foreach wdg [list $text_field $bt_file_open $bt_file_import $search_entry] {
    $wdg configure -state "normal"
  }
  event generate $search_entry <FocusOut>
  
  $bt_interrupt configure -state disabled

  set_state_label_text "batch mode"
  focus $text_field
}

set maxima_channel_hnd ""
set maxima_channel_hnd_pids ""
proc clear_maxima_channel_hnd {} {
  global maxima_channel_hnd
  global maxima_channel_hnd_pids
  global maxima_stderr_should_check

  set maxima_stderr_should_check false

  if {$maxima_channel_hnd_pids != ""} {
    if {![is_wine_like]} {
      foreach one_pid $maxima_channel_hnd_pids {
        catch {exec "kill" "-s" "SIGKILL" $one_pid}
      }
    } {
      foreach one_pid $maxima_channel_hnd_pids {
        catch {exec "taskkill" "/pid" $one_pid}
        catch {exec "taskkill" "/f" "/pid" $one_pid}
      }
    }
    set maxima_channel_hnd_pids ""
  }

  if {$maxima_channel_hnd!=""} {
    catch {chan close $maxima_channel_hnd}
    set maxima_channel_hnd ""
  }
}

set shibboleth_onset ""
set temporary_file_paths [list]
set maxima_pid ""
set maxima_stderr_read_id ""
set maxima_stderr_write_id ""
set maxima_stderr_path ""
set outer_rnd_part_identifier ""
proc setup_maxima {} {
  global maxconn
  global svsocket
  global svport
  global state_label
  global bt_restart
  global has_ever_asked
  global after_restart
  global info_labels_foreground
  global maxima_stderr_read_id
  global maxima_stderr_write_id
  global maxima_stderr_path
  global temporary_file_paths
  global maxima_stderr_should_ignore
  global display_maxima_stderr
  global shibboleth_onset
  global shibboleth_onset_read
  global editor_batch_state
  global include_maxima_input_default
  global include_maxima_input
  global curr_stack_mode
  global maxima_channel_hnd
  global maxima_channel_hnd_pids
  global maxima_stderr_should_check
  global should_display_cpu_usage
  global could_display_cpu_usage
  global print_logging_maxima_input
  global outer_rnd_part_identifier

  if {$editor_batch_state} {
    set_editor_batch_state
    return -1
  }

  set started 1

  if {$maxconn==0} {
    if {$maxima_channel_hnd!=""} {
      clear_maxima_channel_hnd
    }
    if {[is_under_wine]} {
      stop_cpu_usage_updates
    }
    reset_cpu_usage

    set has_ever_asked false
    set started 0
    if {[server_start] == 1} {
      set started 1
      set null_path [get_null_path]
      set maxima_stderr_should_ignore false
      set std_err_redir "2>$null_path"
      set shibboleth_onset_read false
      set shibboleth_onset "so"
      append shibboleth_onset [expr {int(10000*rand()+1000)}]

      if {$maxima_stderr_read_id!=""} {
        catch {close $maxima_stderr_read_id}
        set maxima_stderr_read_id ""
      }
      if {$maxima_stderr_path!=""} {
        catch {file delete "--" $maxima_stderr_path}
        set maxima_stderr_path ""
      }

      set can_continue true
      if {[catch {
        if {$display_maxima_stderr} {
          set maxima_stderr_write_info [get_tempfile]
          set maxima_stderr_write_id [lindex $maxima_stderr_write_info 0]
          set maxima_stderr_path [lindex $maxima_stderr_write_info 1]
          if {($maxima_stderr_write_id=="")||($maxima_stderr_path=="")} {
            set can_continue false
          } {
            lappend temporary_file_paths $maxima_stderr_path
            set std_err_redir "2>@$maxima_stderr_write_id"
          }
        }
      }]} {
        set can_continue false
      }
      if {$can_continue} {
        set outer_rnd_part_identifier "$shibboleth_onset"
        if {[catch {
          set maxima_startup_info [get_maxima_startup_base]
          set maxima_startup_direct [lindex $maxima_startup_info 0]
          set maxima_startup [lindex $maxima_startup_info 1]
          # since ini-commands lead to mess when maxima encounters startup error,
          # we put "-r shibboleth;" here so that we have clear prompt from maxima when it starts correctly,
          # and the lisp-wise prompt when it has an initial/startup issue;
          # getting maxima pid as the sole line here "-s $svport", though it may vary among environments;
          append maxima_startup [strcat " -s $svport -r \"(reset_displays(),linel:100,showtime:true,$shibboleth_onset);\" >$null_path $std_err_redir" [expr {$maxima_startup_direct ? " &" : ""}]]
  
          set maxima_channel_hnd [open $maxima_startup "w"]
          if {$print_logging_maxima_input} {
            print_logging [strcat $maxima_startup "\n"]
          }
  
          chan configure $maxima_channel_hnd -blocking false
          set maxima_channel_hnd_pids [pid $maxima_channel_hnd]
  
          set maxima_stderr_should_check true
          if {[is_under_wine]} {
            start_cpu_usage_updates_under_wine
            set_ignore_cpu_usage false
          }
        }]} {
          set can_continue false
        }
      }
      if {!$can_continue} {
        set outer_rnd_part_identifier ""
        if {$maxima_channel_hnd!=""} {
          clear_maxima_channel_hnd
        }
        if {$maxima_stderr_path != ""} {
          catch {file delete "--" $maxima_stderr_path}
          set maxima_stderr_path ""
        }
        set has_ever_asked false
        set started 0
        # if getting here, maxima (probably) was not started
        set should_display_cpu_usage false
        set could_display_cpu_usage false
        set_state_label_text "not connected"
        if {$after_restart} {
          show_text_field_message_error [list [list "can not start maxima"]]
        }
        catch {chan close $svsocket}
        set svsocket ""
        set svport -1
      }
    }
  }

  $bt_restart configure -state normal

  set include_maxima_input $include_maxima_input_default
  set curr_stack_mode 1
  return $started
}

proc set_state_disabled_buttons {} {
  global bt_plot
  global bt_maxima_batch
  global bt_history
  global bt_stack
  global bt_interrupt
  global bt_restart
  global bt_about
  global search_entry
  global tags_decls_line
  global tags_links_line
  global stack_label_saving
  global stack_label_pos
  global info_label_display
  global info_label_verb
  global info_label_bind
  global info_label_time
  global info_label_autotag
  global info_label_output
  global info_label_hidetag

  $bt_plot configure -state "disabled"
  $bt_maxima_batch configure -state "disabled"
  $bt_history configure -state "disabled"
  $bt_stack configure -state "disabled"
  $bt_interrupt configure -state "disabled"
  $bt_restart configure -state "disabled"
  $bt_about configure -state "disabled"

  catch {
    $stack_label_saving configure -foreground ""
    $stack_label_pos configure -foreground ""
  }

  $stack_label_saving configure -state "disabled"
  $stack_label_pos configure -state "disabled"
  $info_label_display configure -state "disabled"
  $info_label_verb configure -state "disabled"
  $info_label_bind configure -state "disabled"
  $info_label_time configure -state "disabled"
  $info_label_autotag configure -state "disabled"
  $info_label_output configure -state "disabled"
  $info_label_hidetag configure -state "disabled"
  $search_entry configure -state "disabled"
  $tags_decls_line configure -state "disabled"
  $tags_links_line configure -state "disabled"
}

proc set_state_for_save {} {
  global bt_file_save
  global save_path

  if {$save_path==""} {
    $bt_file_save configure -state "disabled"
  } {
    $bt_file_save configure -state "normal"
  }
}

proc set_state_disabled {{save_too false}} {
  global btlist_save
  global btlist_other
  global text_field
  global command_field

  remove_torn_menus

  $text_field configure -state disabled
  $command_field configure -state disabled

  if {$save_too} {
    foreach onebt $btlist_save {
      $onebt configure -state disabled
    }
  } {
    foreach onebt $btlist_save {
      $onebt configure -state normal
      set_state_for_save
      set_void_button_actual_state
    }
  }
  foreach onebt $btlist_other {
    $onebt configure -state disabled
  }
}

proc set_state_ready {{force false}} {
  global btlist_save
  global btlist_other
  global text_field
  global command_field
  global with_common_ctrlc
  global bt_interrupt
  global search_entry
  global shibboleth_onset_read
  global shibboleth_onset_postponed

  if {(!$shibboleth_onset_read) && (!$shibboleth_onset_postponed) && (!$force)} {
    return
  }

  $text_field configure -state normal
  $command_field configure -state normal

  foreach onebt $btlist_save {
    $onebt configure -state normal
    set_state_for_save
  }
  foreach onebt $btlist_other {
    $onebt configure -state normal
  }

  if {!$with_common_ctrlc} {
    $bt_interrupt configure -state disabled
  }

  if {[focus] != $search_entry} {
    event generate $search_entry <FocusOut>
  } {
    event generate $search_entry <FocusIn>
  }

  set_reopen_button_actual_state
  set_void_button_actual_state
}

set_state_disabled

# do not want to deal with amounts of ifs here,
# thus coding two similar functions for (detected) states
# in maxima mode vs. in lisp mode
proc get_comments_strings_brackets_in_lisp {parse_text} {
  set found_comments [list]
  set found_strings [list]
  set found_brackets [list]

  set in_string false
  set cur_string_start [list]

  set in_block_comment false
  set cur_block_comment_start [list]

  set after_backslash false
  set bracket_list [list "\{" "\}" "\[" "\]" "(" ")"]

  set line_idx 0
  foreach one_line [split $parse_text "\n"] {
    incr line_idx
    set char_idx 0
    set one_line_list [split $one_line {}]
    set one_line_list_length [llength $one_line_list]
    set after_backslash false
    if {$one_line_list_length==0} {
      continue
    }
    while {$char_idx < $one_line_list_length} {
      set one_char [lindex $one_line_list $char_idx]
      set next_char [lindex $one_line_list [expr {$char_idx + 1}]]
      if {$in_string} {
        if {$after_backslash} {
          set after_backslash false
        } {
          if {$one_char=="\""} {
            set in_string false
            lappend found_strings [list $cur_string_start [list $line_idx $char_idx]]
            set cur_string_start [list]
          } elseif {$one_char=="\\"} {
            set after_backslash true
          }
        }
        incr char_idx
        continue
      } elseif {$in_block_comment} {
        if {($one_char=="|") && ($next_char=="#")} {
          set in_block_comment false
          lappend found_comments [list $cur_block_comment_start [list $line_idx [expr {$char_idx + 1}]]]
          set cur_block_comment_start [list]
          incr char_idx 2
          continue
        }
        incr char_idx
        continue
      } else {
        if {!$after_backslash} {
          if {$one_char in $bracket_list} {
            lappend found_brackets [list $line_idx $char_idx $one_char]
            incr char_idx
            continue
          }
          if {$one_char=="\""} {
            set in_string true
            set cur_string_start [list $line_idx $char_idx]
            incr char_idx
            continue
          }
          # structure of comments in lisp should be one of:
          # blocks between #| and |# pairs
          # semicolon to end_of_line
          if {$one_char==";"} {
            lappend found_comments [list [list $line_idx $char_idx] [list $line_idx [expr {$one_line_list_length - 1}]]]
            break
          }
          # only considering simple char pairs (without backslash-newline interventions) here,
          # since this is not an editor for lisp programming
          # (and not sure about validity of the interventions for the #| |# pairs either);
          if {($one_char=="#") && ($next_char=="|")} {
            set in_block_comment true
            set cur_block_comment_start [list $line_idx $char_idx]
            incr char_idx 2
            continue
          }
        }
        if {$after_backslash} {
          set after_backslash false
          incr char_idx
          continue
        }
        if {$one_char == "\\"} {
          set after_backslash true
        }
        incr char_idx
        continue
      }
    }
  }

  if {$in_string} {
    lappend found_strings [list $cur_string_start [list "end" "end"]]
  }
  if {$in_block_comment} {
    lappend found_comments [list $cur_block_comment_start [list "end" "end"]]
  }

  return [list $found_comments $found_strings $found_brackets]
}

proc get_comments_strings_brackets {parse_text} {
  set found_comments [list]
  set found_strings [list]
  set found_brackets [list]

  set in_string false
  set in_comment false
  set comment_depth 0

  set line_continues false
  set after_backslash false
  set after_slash false
  set after_star false
  set last_slash_line -1
  set last_slash_char -1

  set cur_string_start [list]
  set cur_comment_start [list]

  set bracket_list [list "\{" "\}" "\[" "\]" "(" ")"]

  set line_idx 0
  foreach one_line [split $parse_text "\n"] {
    incr line_idx
    set char_idx 0
    set one_line_list [split $one_line {}]
    set one_line_list_length [llength $one_line_list]
    if {$one_line_list_length==0} {
      set after_backslash false
      set after_slash false
      set after_star false
      continue
    }
    if {$after_star && (!$after_backslash)} {
      set after_star false
    }
    if {$after_slash && (!$after_backslash)} {
      set after_slash false
    }
    set after_backslash false
    while {$char_idx < $one_line_list_length} {
      set one_char [lindex $one_line_list $char_idx]
      set next_char [lindex $one_line_list "$char_idx+1"]
      if {$one_char=="/"} {
        set last_slash_line $line_idx
        set last_slash_char $char_idx
      }
      if {$in_string} {
        if {$after_backslash} {
          set after_backslash false
        } {
          if {$one_char=="\""} {
            set in_string false
            lappend found_strings [list $cur_string_start [list $line_idx $char_idx]]
            set cur_string_start [list]
          } elseif {$one_char=="\\"} {
            set after_backslash true
          }
        }
        incr char_idx
        continue
      # going (more deep) into comments can be within comments too;
      } else {
        if {(!$in_comment) && ($one_char in $bracket_list)} {
          lappend found_brackets [list $line_idx $char_idx $one_char]
        }
        if {(!$in_comment) && (!$after_backslash) && ($one_char=="\"")} {
          set in_string true
          set cur_string_start [list $line_idx $char_idx]
          set after_star false
          set after_slash false
          incr char_idx
          continue
        }
        if {$after_star && $in_comment} {
          if {$one_char=="/"} {
            set after_star false
            set after_backslash false
            incr comment_depth -1
            if {$comment_depth == 0} {
              set in_comment false
              lappend found_comments [list $cur_comment_start [list $line_idx $char_idx]]
              set cur_comment_start [list]
            }
          } elseif {$one_char=="*"} {
            set after_backslash false
          } elseif {($one_char=="\\") && ($next_char=="")} {
            set after_backslash true
          } else {
            set after_star false
            set after_backslash false
          }
          incr char_idx
          continue
        }
        if {$after_slash} {
          if {$one_char=="*"} {
            set after_slash false
            incr comment_depth 1
            if {$comment_depth == 1} {
              set in_comment true
              # position of respective "/" may be way up
              set cur_comment_start [list $last_slash_line $last_slash_char]
            }
          } elseif {$one_char=="/"} {
            set after_backslash false
          } elseif {($one_char=="\\") && ($next_char=="")} {
            set after_backslash true
          } else {
            set after_slash false
            set after_backslash false
          }
          incr char_idx
          continue
        }
        if {$after_backslash} {
          if {$in_comment && ($one_char=="*")} {
            set after_star true
          }
          if {$in_comment && ($one_char=="/")} {
            set after_slash true
          }
          set after_backslash false
          incr char_idx
          continue
        }
        if {$one_char == "\\"} {
          set after_backslash true
        }
        if {$one_char == "/"} {
          set after_slash true
        }
        if {$one_char == "*"} {
          set after_star true
        }
        incr char_idx
        continue
      }
    }
  }

  if {$in_comment} {
    lappend found_comments [list $cur_comment_start [list "end" "end"]]
  }
  if {$in_string} {
    lappend found_strings [list $cur_string_start [list "end" "end"]]
  }

  return [list $found_comments $found_strings $found_brackets]
}

proc check_name_is_word {name} {
  foreach name_part [split $name "_"] {
    if {![string is ascii $name_part]} {
      return false
    }
    if {![string is alnum $name_part]} {
      return false
    }
  }

  return true
}

set key_block_size 4
set key_min_size 1
array set keyword_vocabulary [list]
# considering strings containing _AZaz09 only;
# names can have trailing parts starting with "(" too;
proc create_keyword_vocabulary {name_list} {
  global keyword_vocabulary
  global key_block_size
  global key_min_size

  foreach one_name $name_list {
    set use_part $one_name
    set add_trailing ""
    set bar_pos [string first "(" $one_name]
    if {$bar_pos>-1} {
      set use_part [string range $one_name 0 $bar_pos-1]
      set add_trailing "("
    }
    set len_part [string length $use_part]
    if {$len_part<$key_min_size} {
      continue
    }
    if {![check_name_is_word $use_part]} {
      continue
    }

    set arr_key [string range $use_part 0 $key_block_size-1]
    if {[string length $arr_key]<$key_block_size} {
      if {[catch {set use_list $keyword_vocabulary($arr_key)}]} {
        set use_list [list]
      }
      set item_to_add [list $use_part $one_name $add_trailing]
      if {([info exists use_list]) && ($item_to_add in $use_list)} {
        continue
      }
      lappend use_list $item_to_add
      array set keyword_vocabulary [list $arr_key $use_list]
    } {
      set dct_key [string range $use_part $key_block_size [expr {2*$key_block_size - 1}]]
      catch {set use_dict $keyword_vocabulary($arr_key)}
      set item_to_add [list $use_part $one_name $add_trailing]
      if {([info exists use_dict]) && ([dict exists $use_dict $dct_key]) && ($item_to_add in [dict get $use_dict $dct_key])} {
        unset use_dict
        continue
      }
      dict lappend use_dict $dct_key $item_to_add
      array set keyword_vocabulary [list $arr_key $use_dict]
      unset use_dict
    }
  }

  return [array size keyword_vocabulary]
}

proc get_matching_list {start} {
  global keyword_vocabulary
  global key_block_size

  set found_names [list]

  set start_fin_idx0 [expr {2*$key_block_size}]
  set start_fin_idx2 [string length $start]
  set start_fin_idx1 [expr {$start_fin_idx2 - 1}]
  set start_fin [string range $start $start_fin_idx0 end]

  set within_keys true
  if {[string length $start]>[expr {2*$key_block_size}]} {
    set within_keys false
  }

  set arr_key [string range $start 0 $key_block_size-1]
  if {[string length $arr_key]<$key_block_size} {
    append arr_key "*"
  }
  set dct_key [string range $start $key_block_size [expr {2*$key_block_size - 1}]]
  if {[string length $dct_key]<$key_block_size} {
    append dct_key "*"
  }

  foreach {one_arr_key one_arr_val} [array get keyword_vocabulary $arr_key] {
    if {[string length $one_arr_key]<$key_block_size} {
      foreach one_name_pair $one_arr_val {
        set name_end [string range [lindex $one_name_pair 0] $start_fin_idx2 "end"]
        lappend found_names [list [strcat $name_end [lindex $one_name_pair 2]] [lindex $one_name_pair 1]]
      }
      continue
    }
    # search in dicts
    foreach {one_dct_key one_dct_val} [dict filter $one_arr_val key $dct_key] {
      if {$within_keys} {
        foreach one_name_pair $one_dct_val {
          set name_end [string range [lindex $one_name_pair 0] $start_fin_idx2 "end"]
          lappend found_names [list [strcat $name_end [lindex $one_name_pair 2]] [lindex $one_name_pair 1]]
        }
        continue
      }
      foreach one_name_pair $one_dct_val {
        set one_name_comp [lindex $one_name_pair 0]
        if {[string range $one_name_comp $start_fin_idx0 $start_fin_idx1] == $start_fin} {
          set name_end [string range [lindex $one_name_pair 0] $start_fin_idx2 "end"]
          lappend found_names [list [strcat $name_end [lindex $one_name_pair 2]] [lindex $one_name_pair 1]]
        }
      }
    }
  }

  return $found_names
}

proc get_common_part_length {clist1 clist2 clist2_length} {
  set test_length [expr {min([llength $clist1],$clist2_length)}]
  if {$test_length==0} {
    return -1
  }
  set curr_idx 0
  while {$curr_idx<$test_length} {
    if {[lindex $clist2 $curr_idx]!=[lindex $clist1 $curr_idx]} {
      incr curr_idx -1
      break
    }
    incr curr_idx
  }
  return $curr_idx
}

proc get_matching_prolongation {start fnc_only paren_pos} {
  global keyword_vocabulary
  global key_block_size

  set found_part [list]
  set found_part_length [llength $found_part]
  set first_found true

  set start_fin_idx0 [expr {2*$key_block_size}]
  set start_fin_idx1 [expr {[string length $start] - 1}]
  set start_fin [string range $start $start_fin_idx0 end]
  set start_fin_idx2 $start_fin_idx1
  incr start_fin_idx2

  set within_keys true
  if {[string length $start]>[expr {2*$key_block_size}]} {
    set within_keys false
  }

  set arr_key [string range $start 0 $key_block_size-1]
  if {[string length $arr_key]<$key_block_size} {
    append arr_key "*"
  }
  set dct_key [string range $start $key_block_size [expr {2*$key_block_size - 1}]]
  if {[string length $dct_key]<$key_block_size} {
    append dct_key "*"
  }

  foreach {one_arr_key one_arr_val} [array get keyword_vocabulary $arr_key] {
    if {[string length $one_arr_key]<$key_block_size} {
      foreach one_name_pair $one_arr_val {
        set pair_idx_use 0
        if {$fnc_only} {
          if {$paren_pos>-1} {
            if {[string first "(" [lindex $one_name_pair 1]]!=$paren_pos} {
              continue
            }
          } {
            if {[string first "(" [lindex $one_name_pair 1]]==-1} {
              continue
            }
          }
          set pair_idx_use 1
        }
        set test_part [split [string range [lindex $one_name_pair $pair_idx_use] $start_fin_idx2 "end"] {}]
        if {$first_found} {
          set found_part $test_part
          set found_part_length [llength $found_part]
          set first_found false
          if {$found_part_length==0} {
            return ""
          }
        } {
          set curr_idx [get_common_part_length $test_part $found_part $found_part_length]
          if {$curr_idx<0} {
            return ""
          }
          set found_part [lrange $found_part 0 $curr_idx]
          set found_part_length [llength $found_part]
        }
      }
      continue
    }
    # search in dicts
    foreach {one_dct_key one_dct_val} [dict filter $one_arr_val key $dct_key] {
      if {$within_keys} {
        foreach one_name_pair $one_dct_val {
          set pair_idx_use 0
          if {$fnc_only} {
            if {$paren_pos>-1} {
              if {[string first "(" [lindex $one_name_pair 1]]!=$paren_pos} {
                continue
              }
            } {
              if {[string first "(" [lindex $one_name_pair 1]]==-1} {
                continue
              }
            }
            set pair_idx_use 1
          }
          set test_part [split [string range [lindex $one_name_pair $pair_idx_use] $start_fin_idx2 "end"] {}]
          if {$first_found} {
            set found_part $test_part
            set found_part_length [llength $found_part]
            set first_found false
            if {$found_part_length==0} {
              return ""
            }
          } {
            set curr_idx [get_common_part_length $test_part $found_part $found_part_length]
            if {$curr_idx<0} {
              return ""
            }
            set found_part [lrange $found_part 0 $curr_idx]
            set found_part_length [llength $found_part]
          }
        }
        continue
      }
      foreach one_name_pair $one_dct_val {
        set one_name_comp [lindex $one_name_pair 0]
        if {[string range $one_name_comp $start_fin_idx0 $start_fin_idx1] == $start_fin} {
          set pair_idx_use 0
          if {$fnc_only} {
            if {$paren_pos>-1} {
              if {[string first "(" [lindex $one_name_pair 1]]!=$paren_pos} {
                continue
              }
            } {
              if {[string first "(" [lindex $one_name_pair 1]]==-1} {
                continue
              }
            }
            set pair_idx_use 1
          }
          set test_part [split [string range [lindex $one_name_pair $pair_idx_use] $start_fin_idx2 "end"] {}]
          if {$first_found} {
            set found_part $test_part
            set found_part_length [llength $found_part]
            set first_found false
            if {$found_part_length==0} {
              return ""
            }
          } {
            set curr_idx [get_common_part_length $test_part $found_part $found_part_length]
            if {$curr_idx<0} {
              return ""
            }
            set found_part [lrange $found_part 0 $curr_idx]
            set found_part_length [llength $found_part]
          }
        }
      }
    }
  }

  return [join $found_part ""]
}

catch {
  bind $text_field <Control-Tab> {
    if {[is_text_field_active]} {
      clear_temporary_highlights
      show_menu_key_names $text_field "get_text_field_position" false
    }
    break
  }
}
foreach keyspressed [list "Shift-Key-Tab" "Shift-Key-KP_Tab" "Shift-Key-ISO_Left_Tab" "Key-ISO_Left_Tab"] {
  catch {
    bind $text_field <$keyspressed> {
      if {[is_text_field_active]} {
        clear_temporary_highlights
        string_action_on_tab $text_field true true
        update_text_properties
      }
      break
    }
  }
  catch {
    bind $text_field <Control-$keyspressed> {
      if {[is_text_field_active]} {
        clear_temporary_highlights
        show_menu_key_names $text_field "get_text_field_position" true
      }
      break
    }
  }
}

catch {
  bind $command_field <Control-Tab> {
    if {[is_command_field_active]} {
      show_menu_key_names $command_field "get_command_field_position" false
    }
    break
  }
}
foreach keyspressed [list "Shift-Key-Tab" "Shift-Key-KP_Tab" "Shift-Key-ISO_Left_Tab" "Key-ISO_Left_Tab"] {
  catch {
    bind $command_field <$keyspressed> {
      if {[is_command_field_active]} {
        string_action_on_tab $command_field false true
        update_command_properties
      }
      break
    }
  }
  catch {
    bind $command_field <Control-$keyspressed> {
      if {[is_command_field_active]} {
        show_menu_key_names $command_field "get_command_field_position" true
      }
      break
    }
  }
}

proc update_keyword_menu_entries {forwards shift_size} {
  global keyword_menu
  global keyword_menu_offset
  global keyword_set_last_found
  global keyword_menu_break_point
  global keyword_menu_shown_items

  set keyword_menu_shown_items 0

  set name_count [llength $keyword_set_last_found]

  if {!$forwards} {
    set keyword_menu_offset [expr {$keyword_menu_offset - 2*$shift_size}]
    if {$keyword_menu_offset<0} {
      set keyword_menu_offset [expr {
        [llength $keyword_set_last_found] - ((([llength $keyword_set_last_found]-1) % $shift_size) + 1)
      }]
    }
  }

  if {$keyword_menu_offset >= $name_count} {
    set keyword_menu_offset 0
  }
  set count_to_show [expr {$name_count - $keyword_menu_offset}]
  if {$count_to_show > $shift_size} {
    set count_to_show $shift_size
  }
  set keyword_menu_shown_items $count_to_show

  set name_list_length [llength $keyword_set_last_found]
  set first_entry_to_show [expr {$keyword_menu_offset + 1}]
  set last_entry_to_show [expr {$keyword_menu_offset + $shift_size}]
  if {$last_entry_to_show > $name_list_length} {
    set last_entry_to_show $name_list_length
  }

  set menu_offset 0
  if {$shift_size==$keyword_menu_break_point} {
    set label_text "shown part: $first_entry_to_show - $last_entry_to_show / $name_list_length"
    $keyword_menu entryconfigure 0 -label $label_text
    set menu_offset 1
  }

  set curr_idx 0
  while {$curr_idx<$count_to_show} {
    set stack_idx [expr {$curr_idx+$keyword_menu_offset}]
    set one_name [lindex [lindex $keyword_set_last_found $stack_idx] 1]
    set bgcolor "#f0f0f0"
    if {[expr {$curr_idx%2}]==0} {
      set bgcolor "#ffffff"
    }
    $keyword_menu entryconfigure [expr {$curr_idx+$menu_offset}] -label $one_name -state normal -background $bgcolor
    $keyword_menu entryconfigure [expr {$curr_idx+$menu_offset}] -command [list "keyword_add_word_end" $stack_idx]
    incr curr_idx
  }

  while {$curr_idx<$shift_size} {
    $keyword_menu entryconfigure [expr {$curr_idx+$menu_offset}] -label "   " -state disabled -background "#e0e0e0"
    $keyword_menu entryconfigure [expr {$curr_idx+$menu_offset}] -command [list "keyword_add_word_end" "none"]
    incr curr_idx
  }

  incr keyword_menu_offset $curr_idx
}

proc remove_keyword_insertion_menu {} {
  global keyword_menu
  global keyword_menu_torn
  global keyword_menu_tears
  global keyword_use_wdg
  global keyword_menu_use

  if {!$keyword_menu_tears} {
    return
  }
  set keyword_menu_tears false
  set keyword_menu_use ""

  if {$keyword_menu_torn!=""} {
    catch {destroy $keyword_menu_torn}
    set keyword_menu_torn ""
  }
  if {$keyword_menu!=""} {
    catch {destroy $keyword_menu}
    set keyword_menu ""
  }

  catch {focus $keyword_use_wdg}
}

proc get_keyword_menu_use {} {
  global keyword_menu_use
  return $keyword_menu_use
}

set keyword_menu ""
set keyword_menu_torn ""
set keyword_menu_use ""
set keyword_menu_tears false
set keyword_set_last_found [list]
set keyword_min_start_length 1
set keyword_menu_break_point 25
# see below the "shown_by_small_parts_request" resetting on small screens
set keyword_column_width_min 375
if {[get_screen_width]>=(4*$keyword_column_width_min)} {
  set max_size_of_keyword_menu [expr {4*$keyword_menu_break_point}]
} elseif {[get_screen_width]>=(3*$keyword_column_width_min)} {
  set max_size_of_keyword_menu [expr {3*$keyword_menu_break_point}]
} elseif {[get_screen_width]>=(2*$keyword_column_width_min)} {
  set max_size_of_keyword_menu [expr {2*$keyword_menu_break_point}]
} {
  set max_size_of_keyword_menu $keyword_menu_break_point
}

if {$keyword_menu_break_point < [get_max_menu_entry_count]} {
  set keyword_menu_break_point [get_max_menu_entry_count]
  set max_size_of_keyword_menu $keyword_menu_break_point
}

set keyword_menu_offset 0
set keyword_use_wdg ""
set keyword_menu_shown_items 0
proc show_menu_key_names {use_wdg use_field_position_method shown_by_small_parts_request} {
  global text_field
  global command_field
  global keyword_use_wdg
  global keyword_set_last_found
  global keyword_min_start_length
  global keyword_menu_break_point
  global max_size_of_keyword_menu
  global keyword_menu_offset
  global keyword_menu
  global keyword_menu_torn
  global keyword_menu_use
  global keyword_menu_tears
  global topw
  global keyword_menu_shown_items

  # see above the "max_size_of_keyword_menu" valuation according to screen sizes
  if {$max_size_of_keyword_menu == $keyword_menu_break_point} {
    set shown_by_small_parts_request true
  }

  set keyword_use_wdg $use_wdg
  set keyword_menu_offset 0
  set keyword_menu_shown_items 0

  set pos_list [$use_wdg bbox "insert"]
  if {[llength $pos_list]==0} {
    set pos_list [list 0 0]
  }

  set abs_pos [$use_field_position_method]
  set use_x [expr {[lindex $pos_list 0] + [lindex $abs_pos 0]}]
  set use_y [expr {[lindex $pos_list 1] + [lindex $abs_pos 1]}]

  set to_compare ""

  set line_start [$use_wdg get "insert linestart" "insert"]
  foreach one_char [split [string reverse $line_start] {}] {
    if {![string is ascii -strict $one_char]} {
      break
    }
    if {(![string is alnum -strict $one_char]) && ("_" != $one_char)} {
      break
    }
    set to_compare $one_char$to_compare
  }

  set to_compare_length [string length $to_compare]
  if {$to_compare_length<$keyword_min_start_length} {
    return
  }

  set name_list [get_matching_list $to_compare]
  set name_list [lsort -dictionary -index 1 $name_list]

  set keyword_set_last_found $name_list
  set name_list_length [llength $name_list]
  if {$name_list_length == 0} {
    return
  }

  if {[string equal -nocase -length 4 "aqua" [tk windowingsystem]]} {
    set shown_by_small_parts_request true
  }
  if {[is_wine_like] && ($name_list_length>$max_size_of_keyword_menu)} {
    set shown_by_small_parts_request true
  }

  set part_size $keyword_menu_break_point
  set keyword_menu_shown_items $name_list_length
  set shown_by_parts false
  set shown_by_small_parts false
  if {$shown_by_small_parts_request} {
    if {$name_list_length > $keyword_menu_break_point} {
      set keyword_menu_shown_items $keyword_menu_break_point
      set shown_by_small_parts true
      set shown_by_parts true
    }
  } {
    set part_size $max_size_of_keyword_menu
    if {$name_list_length > $max_size_of_keyword_menu} {
      set keyword_menu_shown_items $max_size_of_keyword_menu
      set shown_by_parts true
    }
  }

  set keyword_menu_tears [expr {$shown_by_parts ? true : false}]
  # when using "torn" menus, binding for Key-p, Key-n are broken;
  if {[is_linux_like]} {
    set keyword_menu_tears false
  }

  if {$keyword_menu_torn!=""} {
    catch {destroy $keyword_menu_torn}
  }
  catch {destroy $keyword_menu}
  set keyword_menu [menu "$topw.keywordmenu" -tearoff false]
  $keyword_menu configure -relief solid -borderwidth 1 -background "#f0f0f0" -disabledforeground "#000000"
  $keyword_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f8ffc0"
  $keyword_menu configure -relief "solid" -borderwidth 1 -activeborderwidth 1
  set keyword_menu_use $keyword_menu

  if {$shown_by_parts && $shown_by_small_parts} {
    set backwards_list [list "2" "Control-Left" "Control-p" "Left" "Key-p"]
    foreach keyspressed $backwards_list {
      bind $keyword_menu <$keyspressed> {
        catch {
          update_keyword_menu_entries false $keyword_menu_break_point
        }
        break
      }
    }
    set forwards_list [list "3" "Control-Right" "Control-n" "Right" "Key-n"]
    foreach keyspressed $forwards_list {
      bind $keyword_menu <$keyspressed> {
        catch {
          update_keyword_menu_entries true $keyword_menu_break_point
        }
        break
      }
    }
  }
  if {$shown_by_parts && (!$shown_by_small_parts)} {
    foreach keyspressed [list "Key-Left" "Key-comma" "Key-p"] {
      bind $keyword_menu <$keyspressed> {
        catch {
          if {[[get_keyword_menu_use] index "active"]!="none"} {
            [get_keyword_menu_use] activate [expr { \
              (([[get_keyword_menu_use] index "active"] - $keyword_menu_break_point) > 0) ? \
               ([[get_keyword_menu_use] index "active"] - $keyword_menu_break_point) : 0 \
            }]
          } {
            [get_keyword_menu_use] activate [expr {$keyword_menu_shown_items-1}]
          }
        }
        break
      }
    }
    foreach keyspressed [list "Key-Right" "Key-period" "Key-n"] {
      bind $keyword_menu <$keyspressed> {
        catch {
          if {[[get_keyword_menu_use] index "active"]!="none"} {
            [get_keyword_menu_use] activate [expr { \
              (($keyword_menu_shown_items-1) < ([[get_keyword_menu_use] index "active"] + $keyword_menu_break_point)) ? \
               ($keyword_menu_shown_items-1) : ([[get_keyword_menu_use] index "active"] + $keyword_menu_break_point) \
            }]
          } {
            [get_keyword_menu_use] activate 0
          }
        }
        break
      }
    }
    set backwards_list [list "2" "Control-Key-Left" "Control-Key-p" "Control-Key-comma"]
    foreach keyspressed $backwards_list {
      bind $keyword_menu <$keyspressed> {
        catch {
          update_keyword_menu_entries false $max_size_of_keyword_menu
        }
        break
      }
    }
    set forwards_list [list "3" "Control-Key-Right" "Control-Key-n" "Control-Key-period"]
    foreach keyspressed $forwards_list {
      bind $keyword_menu <$keyspressed> {
        catch {
          update_keyword_menu_entries true $max_size_of_keyword_menu
        }
        break
      }
    }
  }
  if {!$shown_by_parts} {
    bind $keyword_menu <Left> {
      catch {
        if {[$keyword_menu index "active"]!="none"} {
          $keyword_menu activate [expr {max(0,[expr {[$keyword_menu index "active"] - $keyword_menu_break_point}])}]
        } {
          $keyword_menu activate [expr {[llength $keyword_set_last_found]-1}]
        }
      }
      break
    }
    bind $keyword_menu <Right> {
      catch {
        if {[$keyword_menu index "active"]!="none"} {
          $keyword_menu activate [expr {min([expr {[llength $keyword_set_last_found]-1}],[expr {[$keyword_menu index "active"] + $keyword_menu_break_point}])}]
        } {
          $keyword_menu activate 0
        }
      }
      break
    }
  }

  if {$shown_by_parts && $shown_by_small_parts} {
    set label_text "shown part: 1 - $keyword_menu_break_point / $name_list_length"
    $keyword_menu add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
  }

  set curr_idx 0
  foreach one_name_pair $name_list {
    set one_name [lindex $one_name_pair 1]
    $keyword_menu add command -label $one_name -command [list "keyword_add_word_end" $curr_idx]
    incr curr_idx
    if {[expr {$curr_idx % 2}]==0} {
      $keyword_menu entryconfigure $curr_idx -background "#f0f0f0"
    } {
      $keyword_menu entryconfigure $curr_idx -background "#ffffff"
    }
    if {($keyword_menu_break_point<$part_size) && ($curr_idx>$keyword_menu_break_point)} {
      if {[expr {($curr_idx-1) % $keyword_menu_break_point}]==0} {
        $keyword_menu entryconfigure $curr_idx -columnbreak 1
      }
    }
    if {$shown_by_parts} {
      if {[expr {$curr_idx % $part_size}]==0} {
        break
      }
    }
  }
  set keyword_menu_offset $curr_idx

  # to finish a column if it is not the first one,
  # regardless whether there are (not) other parts
  if {$curr_idx>$keyword_menu_break_point} {
    while {[expr {$curr_idx % $keyword_menu_break_point}]!=0} {
      $keyword_menu add command -label "   " -state disabled -background "#e0e0e0" -command [list "keyword_add_word_end" "none"]
      incr curr_idx
    }
  }

  set current_menu $keyword_menu
  set pos_vert [expr {$use_y - ([$current_menu yposition [$current_menu index last]] + (($shown_by_parts || [is_linux_like]) ? 20 : 2))}]
  set pos_vert_min [expr {24 + [winfo vrooty $current_menu]}]
  if {$pos_vert < $pos_vert_min} {
    set pos_vert $pos_vert_min
  }
  set pos_horz [expr {$use_x + 10}]

  if {$keyword_menu_tears} {
    $keyword_menu configure -title "Defined names"
    set keyword_menu_torn [::tk::TearOffMenu $keyword_menu $pos_horz $pos_vert]
    set keyword_menu_use $keyword_menu_torn
    catch {wm resizable $keyword_menu false false}
    catch {wm resizable $keyword_menu_torn false false}
    catch {focus $keyword_menu_torn}
    catch {
      bind $keyword_menu_torn <Escape> {
        remove_keyword_insertion_menu
        break
      }
    }
  } {
    tk_popup $keyword_menu $pos_horz $pos_vert
  }
}

proc keyword_add_word_end {index} {
  global keyword_set_last_found
  global keyword_use_wdg

  if {$index=="none"} {
    return
  }

  remove_keyword_insertion_menu

  if {[catch {set word_end [lindex [lindex $keyword_set_last_found $index] 0]}]} {
    return
  }

  $keyword_use_wdg edit separator
  $keyword_use_wdg insert "insert" $word_end
  $keyword_use_wdg edit separator
}

proc get_top_window_position {} {
  global topw

  set top_pos [winfo geometry $topw]
  set top_pos [string range $top_pos [string first "+" $top_pos]+1 end]

  return [split $top_pos "+"]
}

proc get_command_field_position {} {
  global command_field
  global command_field_holder
  global fcmmd
  global fpaned
  global topw

  set pos_x 0
  set pos_y 0
  foreach wnd [list $topw $fpaned $fcmmd $command_field_holder $command_field] {
    set one_pos [winfo geometry $wnd]
    set one_pos [string range $one_pos [string first "+" $one_pos]+1 end]
    set one_pos [split $one_pos "+"]
    set pos_x [expr {$pos_x + [lindex $one_pos 0]}]
    set pos_y [expr {$pos_y + [lindex $one_pos 1]}]
  }

  return [list $pos_x $pos_y]
}

proc get_plot_field_position {} {
  global plot_put_window
  global plot_put_holder
  global plot_input_frame
  global plot_input_field

  set pos_x 0
  set pos_y 0
  foreach wnd [list $plot_put_window $plot_put_holder $plot_input_frame $plot_input_field] {
    set one_pos [winfo geometry $wnd]
    set one_pos [string range $one_pos [string first "+" $one_pos]+1 end]
    set one_pos [split $one_pos "+"]
    set pos_x [expr {$pos_x + [lindex $one_pos 0]}]
    set pos_y [expr {$pos_y + [lindex $one_pos 1]}]
  }

  return [list $pos_x $pos_y]
}

proc get_text_field_position {} {
  global text_field
  global ftext
  global fpaned
  global topw

  set pos_x 0
  set pos_y 0
  foreach wnd [list $topw $fpaned $ftext $text_field] {
    set one_pos [winfo geometry $wnd]
    set one_pos [string range $one_pos [string first "+" $one_pos]+1 end]
    set one_pos [split $one_pos "+"]
    set pos_x [expr {$pos_x + [lindex $one_pos 0]}]
    set pos_y [expr {$pos_y + [lindex $one_pos 1]}]
  }

  return [list $pos_x $pos_y]
}

proc get_menu_button_position {menu_button} {
  global topw
  global fmenu

  set pos_x 0
  set pos_y 0
  foreach wnd [list $topw $fmenu $menu_button] {
    set one_pos [winfo geometry $wnd]
    set one_pos [string range $one_pos [string first "+" $one_pos]+1 end]
    set one_pos [split $one_pos "+"]
    set pos_x [expr {$pos_x + [lindex $one_pos 0]}]
    set pos_y [expr {$pos_y + [lindex $one_pos 1]}]
  }

  return [list $pos_x $pos_y]
}

proc get_mxmbatch_menu_button_position {menu_button} {
  global batch_maxima_window
  global mxmbatch_fmenu

  if {($batch_maxima_window=="") || ($mxmbatch_fmenu=="")} {
    # this should not occur
    return [list 0 0]
  }

  set pos_x 0
  set pos_y 0
  foreach wnd [list $batch_maxima_window $mxmbatch_fmenu $menu_button] {
    set one_pos [winfo geometry $wnd]
    set one_pos [string range $one_pos [string first "+" $one_pos]+1 end]
    set one_pos [split $one_pos "+"]
    set pos_x [expr {$pos_x + [lindex $one_pos 0]}]
    set pos_y [expr {$pos_y + [lindex $one_pos 1]}]
  }

  return [list $pos_x $pos_y]
}

proc get_import_menu_button_position {menu_button} {
  global import_window
  global import_fmenu

  if {($import_window=="") || ($import_fmenu=="")} {
    # this should not occur
    return [list 0 0]
  }

  set pos_x 0
  set pos_y 0
  foreach wnd [list $import_window $import_fmenu $menu_button] {
    set one_pos [winfo geometry $wnd]
    set one_pos [string range $one_pos [string first "+" $one_pos]+1 end]
    set one_pos [split $one_pos "+"]
    set pos_x [expr {$pos_x + [lindex $one_pos 0]}]
    set pos_y [expr {$pos_y + [lindex $one_pos 1]}]
  }

  return [list $pos_x $pos_y]
}

proc get_import_input_position {} {
  global import_window
  global import_pane_holder
  global import_input_frame
  global import_input_field

  set pos_x 0
  set pos_y 0
  foreach wnd [list $import_window $import_pane_holder $import_input_frame $import_input_field] {
    set one_pos [winfo geometry $wnd]
    set one_pos [string range $one_pos [string first "+" $one_pos]+1 end]
    set one_pos [split $one_pos "+"]
    set pos_x [expr {$pos_x + [lindex $one_pos 0]}]
    set pos_y [expr {$pos_y + [lindex $one_pos 1]}]
  }

  return [list $pos_x $pos_y]
}

proc get_maxima_batch_output_position {} {
  global batch_maxima_window
  global mxmbatch_pane_holder
  global mxmbatch_output_frame
  global mxmbatch_output_field

  set pos_x 0
  set pos_y 0
  foreach wnd [list $batch_maxima_window $mxmbatch_pane_holder $mxmbatch_output_frame $mxmbatch_output_field] {
    set one_pos [winfo geometry $wnd]
    set one_pos [string range $one_pos [string first "+" $one_pos]+1 end]
    set one_pos [split $one_pos "+"]
    set pos_x [expr {$pos_x + [lindex $one_pos 0]}]
    set pos_y [expr {$pos_y + [lindex $one_pos 1]}]
  }

  return [list $pos_x $pos_y]
}

proc get_maxima_batch_input_position {} {
  global batch_maxima_window
  global mxmbatch_pane_holder
  global mxmbatch_input_frame
  global mxmbatch_input_field

  set pos_x 0
  set pos_y 0
  foreach wnd [list $batch_maxima_window $mxmbatch_pane_holder $mxmbatch_input_frame $mxmbatch_input_field] {
    set one_pos [winfo geometry $wnd]
    set one_pos [string range $one_pos [string first "+" $one_pos]+1 end]
    set one_pos [split $one_pos "+"]
    set pos_x [expr {$pos_x + [lindex $one_pos 0]}]
    set pos_y [expr {$pos_y + [lindex $one_pos 1]}]
  }

  return [list $pos_x $pos_y]
}

proc get_maxima_batch_prepend_entry_position {} {
  global batch_maxima_window
  global mxmbatch_pane_holder
  global mxmbatch_input_frame
  global mxmbatch_input_entry_holder
  global mxmbatch_input_entry

  set pos_x 0
  set pos_y 0
  foreach wnd [list $batch_maxima_window $mxmbatch_pane_holder $mxmbatch_input_frame $mxmbatch_input_entry_holder $mxmbatch_input_entry] {
    set one_pos [winfo geometry $wnd]
    set one_pos [string range $one_pos [string first "+" $one_pos]+1 end]
    set one_pos [split $one_pos "+"]
    set pos_x [expr {$pos_x + [lindex $one_pos 0]}]
    set pos_y [expr {$pos_y + [lindex $one_pos 1]}]
  }

  return [list $pos_x $pos_y]
}

proc get_gnuplot_field_position {} {
  global plot_put_window
  global plot_put_holder
  global plot_input_frame
  global plot_input_field
  global plot_data_frame
  global plot_data_field
  global plot_output_frame
  global plot_output_field
  global gnuplot_at_data_view
  global gnuplot_at_output_view

  set use_plot_frame $plot_input_frame
  set use_plot_field $plot_input_field
  if {$gnuplot_at_data_view} {
    set use_plot_frame $plot_data_frame
    set use_plot_field $plot_data_field
  }
  if {$gnuplot_at_output_view} {
    set use_plot_frame $plot_output_frame
    set use_plot_field $plot_output_field
  }

  set pos_x 0
  set pos_y 0
  foreach wnd [list $plot_put_window $plot_put_holder $use_plot_frame $use_plot_field] {
    set one_pos [winfo geometry $wnd]
    set one_pos [string range $one_pos [string first "+" $one_pos]+1 end]
    set one_pos [split $one_pos "+"]
    set pos_x [expr {$pos_x + [lindex $one_pos 0]}]
    set pos_y [expr {$pos_y + [lindex $one_pos 1]}]
  }

  return [list $pos_x $pos_y]
}

proc get_gnuplot_menu_position {wdg} {
  global plot_put_window
  global plot_menu_holder

  set pos_x 0
  set pos_y 0
  foreach wnd [list $plot_put_window $plot_menu_holder $wdg] {
    set one_pos [winfo geometry $wnd]
    set one_pos [string range $one_pos [string first "+" $one_pos]+1 end]
    set one_pos [split $one_pos "+"]
    set pos_x [expr {$pos_x + [lindex $one_pos 0]}]
    set pos_y [expr {$pos_y + [lindex $one_pos 1]}]
  }

  return [list $pos_x $pos_y]
}

set last_button_interrupt_state "idle"

proc set_button_interrupt_sets {} {
  global icon_image_mm
  global bt_interrupt
  global use_buttons_text
  global last_button_interrupt_state

  set last_button_interrupt_state "sets"

  if {$use_buttons_text} {
    $bt_interrupt configure -text "X" -state disabled
  } {
    $bt_interrupt configure -image $icon_image_mm -state disabled
  }
}

proc set_button_interrupt_idle {} {
  global icon_image_mm
  global bt_interrupt
  global with_common_ctrlc
  global use_buttons_text
  global last_button_interrupt_state

  set last_button_interrupt_state "idle"

  if {$with_common_ctrlc} {
    if {$use_buttons_text} {
      $bt_interrupt configure -text "X" -state normal
    } {
      $bt_interrupt configure -image $icon_image_mm -state normal
    }
  } {
    if {$use_buttons_text} {
      $bt_interrupt configure -text "X" -state disabled
    } {
      $bt_interrupt configure -image $icon_image_mm -state disabled
    }
  }
}

proc set_button_interrupt_exec {} {
  global icon_image_hg
  global icon_image_mm
  global bt_interrupt
  global with_common_ctrlc
  global use_buttons_text
  global last_button_interrupt_state

  set last_button_interrupt_state "exec"

  if {$with_common_ctrlc} {
    if {$use_buttons_text} {
      $bt_interrupt configure -text "W" -state normal
    } {
      $bt_interrupt configure -image $icon_image_hg -state normal
    }
  } {
    if {$use_buttons_text} {
      $bt_interrupt configure -text "X" -state disabled
    } {
      $bt_interrupt configure -image $icon_image_mm -state disabled
    }
  }
}

set icon_data_hg [join [list \
"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAQAAAC1+jfqAAAAAmJLR0QA/4ePzL8AAAAJcEhZcwAA" \
"CxMAAAsTAQCanBgAAAAHdElNRQfhCAsVAgWuDFnRAAAACHRFWHRDb21tZW50APbMlr8AAAD1SURB" \
"VCjPY2AgABihdDIDO4bcT4a5CAUnWDQ5/yDLfmf5c5XBCqGAged1rwiygpLXn8UYGBgYmGACLM/e" \
"I0m/Y2B6DmHBFfzedeMvQsHNv392oin4OmPHh/9Q9j+G7R+/zkRTwHD3y8pdP34zMDD8Ztj1/esK" \
"hruYXuaQ+ST3P+u/3H+ZTwwcMEFmhDzPRDuTGBYpBkuGH/+eiP3ahm6FMk+4GwcrAwMDK4M7J3cE" \
"gzKaAu4MDwFYoDAxePJzp6MpYHXTQLJOnZnFHU3BHylBJPcKMfyTRI2skywanH9R4oL5z3UGCwYG" \
"Fih/5h+2z5ixycDAAACULUPVgn3f2AAAAABJRU5ErkJggg=="] ""]
set icon_image_hg [image create photo "icon_image_hg" -data $icon_data_hg]

set icon_data [join [list \
"iVBORw0KGgoAAAANSUhEUgAAAEEAAABBCAYAAACO98lFAAAABmJLR0QA/wD/AP+gvaeTAAAACXBI" \
"WXMAAA9hAAAPYQGoP6dpAAAAB3RJTUUH4QgDFRQ6wUbp1AAABhlJREFUeNrtmmtoXEUUx///eze7" \
"W60m2odt1ZiKGpNuqjRYbfFRUESQ+kIUQXxUmn4URWuzSXDIY29rFawFkVBE/KBIfWArJSD4QVEr" \
"NAl20/hq06g1RdJIsNbuTXfn+KWKhjQ7c3f3qnDPx52ZO2d+8z9nzp27QGSRhWJrlHL/6z7Gyv3A" \
"q5WqLiSTNwmwGiKrjwENABaYAvslHm8W8hYBbiSwWID5BOaBFIhMgjwkWg8K0OcfPfrhwe3bp0r1" \
"mWVZuVJMVVXdDMd5lOQ9AJJ/b/ZzufnfKjUx2+LH4/GHHfJZkLUWM/8sIi9r339hWKkT/wqENUq5" \
"44nEIyTbCdSdcZJCYcX+trbBmdqaPG+pALtJLgvqh4iMUevHsm1tfUHGO0EnbvC8qycSiX6H3DEb" \
"AADQZ9jdVE9PE8hPSwEAACSXwHH2NHlee2gQUp73aIz8AuRVhkNqZ1BADRynD8DisoQkSZBdTZ7n" \
"VRxCyvM2kXwVQNzCv9oZfnyB5JKyp3pyUyqTaakYhFQm00LSC+Ba7TQVrACwrlJHHh1nW0NnZ2PZ" \
"IaS6u5tJvmSZsE4AeKWgdWbabq03Gg/0FfL5lTx27OyC1o0AXjWcOhmLx3eU9XRYo5Q7kUgMgFxu" \
"DADod7V+4Mt0+ru//37RE08kaxYuPEqypgjBD7IDA3di5079j83YvPk5Ak8bOaH1vdl0+p2yKGE8" \
"Hl9vBUBkbyKXu2E6AAA4b/781cUAiMgpp1DYMB0AACRGRtIADprp3HmmLOHQ3NISo+HDTivgJ8f3" \
"7+hX6uQZElezQXLb9WV7+9hMTf29vXkRyRi6c83yTGZlyRD8urrbi9UB//BfZON+pcZn2Z1mgxj9" \
"aLb2nO+/LSKnjCKCfKD0cCAftAiDgWxr6xtFlHKFwaOyszUeVOo4gC8Mk97dJUFoVKqK5G0WB0Kv" \
"AamaYl0KudyowVz7DOuGS5o8b2lwJcRiKwHMNVTBqd8nJ980KJzOLdZnCpg0mPAr053R5PWBIbiu" \
"e41FVbhvZMuWXw1gFYVQPTZ2sqiEye8t6oCmUnJCyiIffGIIq6pYn/7e3nzRkCGPWiTrhlIgXGpR" \
"HmbDvA1iPv+LsWvkhcEhkMZveFrrw6FCcN3fLMJhUSlKWGBMG/gxTAgnx8Z+t+h+TmAIInKWhTyP" \
"hwnhonnzChYblAwMgWTCdKKpiYkT+J+aU4SgNn1Qte/rMB2fNKxfTp8OJ4MrAcgZK2Hp0rlhQpBY" \
"7GyLcAgOAYCxxAu+f064GnaMIRCYCA5BxLggcaqqasJkoG2+T5A/lqKEI8bzOE59yPnsSotw+KGU" \
"Yukbi4oxFTKEeotwGCwlHAYsStMVoSZGcrUxBK33BYag8/m9Fn7dUr9xYygnRJNSFxJoNhOoTOan" \
"poIr4UBHxwiArw3fDufEq6vXhiKDROIui957hpU6VUpihBZ511iirvt4pdffqFSVkE9ahE3Rix7H" \
"oMMOGFaOBK5d1tNzf0XLg0RiHQ1f8QUYOdDfv6dkCNnW1sMA3jevYZwXl3d11VUCQENnZyOBrcYq" \
"0HrrTN8urCEAQEHrNhExe2sjF+lYrK9eqcVlDoOL3Xh8N0jTyvTr5Oio0ac4IwjD6fRXBLaZF2is" \
"jyeT2ZTn3VeW0yCTWesmk4M0v+nSovUGk2u602FsZnVKJeYmk/toce94Oi6/BfAatP40BgwXHGe8" \
"2Jh4LndWLpm8nCKrALTQsgYRke6h1tYOi2LK3JZ3ddVJLPY5yEWVrYREQDLg0DeGfP9BKCVluU+Y" \
"bvs7OkZF61tFZKyiEEoAoH3/ERsA1kr4K0t3d9e6rvseQy6VZ1l8AUDnkO932QIIDOHPooWJhHLI" \
"p2Dx150KAMhCpGUond4bWHilOrGsq+tSJxZTIO8PE4aIHIDI80ODg6+b1AIVhfCnpXp6ForjPOSQ" \
"90DkuqBxXWThRyCyC8BbQ+n0x2VLQRW57chkzneBVQ65EmS9AJdB5AIA80jOMVjtD0KOQ+QIgG+E" \
"HBKRz4ZbWw9VJA+HHcNNmzcXTVzZTZtC9ctBZBGECEIEIYIQQYggRBAiCBGECEIEIYIQQYggRBaZ" \
"tf0BIvYtBEMhkNQAAAAASUVORK5CYII="] ""]

set icon_data_batch [join [list \
"iVBORw0KGgoAAAANSUhEUgAAAEEAAABBCAYAAACO98lFAAAABmJLR0QA/wD/AP+gvaeTAAAACXBI" \
"WXMAAA9hAAAPYQGoP6dpAAAAB3RJTUUH4gEOFiY7IJn7aAAABntJREFUeNrtml1sFFUUx8+5d7a7" \
"9IsiRWkJFVCLVFBDQ5XGjweJMVGqEpRgCIXWdgsGtNCWltaAbWFtdysCsVBSEDQCJiABDGli4oMG" \
"hYTSh6pA5csCRW1LCv3uztzjQ8VgLZ07s7ujJnMeZ87MPfO7/3PuuXcXwDbbLLHDWWn8vx6jEuwX" \
"Vq9bNLonavSzGlEqQ0ptApgGAONkgV18IC5ZcJwjEJ9hRHECMZYRjSXkhKR1EPILINQGQqxTb7R+" \
"VeQ5OBBozBiMD0+uWY5vdGjPCc6WIrB5AOC6877wd8cWFO1uH+njmxLj0xnAOkKeYGDo3whEtdLX" \
"U5Vbsqf7X4FwOCuNn02csAQASzjipLv5cb9/Zm7Rjobh7lV5cyYLxKMI7BGzcRCIFq5qmasKd9SZ" \
"eZ6ZHbi8Mufx84nx9RxZ7UgAAAA0jsPOblWFewYhPx4IgMGZZPHE2bEqb06JZRC8vmVLXYyfJOSP" \
"ST6SMIwCYgRndQAQF4yUJORIyMuqvDmekEPw+pYVIrBdABAmPVP4TyUQ8ioEFh/swkzIC32V7uyQ" \
"QfBVurMRmMdEbAlDVDATADJCtuYxZbPPk5kUdAgfeN5MRsQtBgtWNwBsF0Abh8xWlszzGlEdaQMp" \
"Y65dilBJTQKAXZJDu8Dhqg3q6nA4K42fT4w/TcgflX2xRlTv1LSF7xTW/Hzn9dLS11yjwmOvI7CY" \
"EQMj7cu9bezl+opqcef1Cu/ySo6YLzXDQp2/qqDmYFCU0JQYn2UEAIE4MebWzaeHAgAAiAwbk6oH" \
"gED4HQN+91AAAAC8+fe1AHBeJg7BlDVBSQffivmKALbGgAKuubq70rLLPusdfkRM1g2e8MjK4p0t" \
"w93L23pAJRAbJcOZtel9d0rgSpg49kW9PuDvfbgoWLHu09YRZkcXAkf8eqT7fn/fAQLhl4lHVdjC" \
"gCEQY4sMpMHp1fnb9+ooJVHiVY0j3Vxb9HEnAJyUkjrRqwFB2FSe7kBgLxhYEHboDogUo6uEvq7L" \
"EmOdkuwb7q/y5kw2DUF1hKUAQKSkCvxK56/7JFrcaD2fXs46dN8jxBnZmRkA9pRpCIzxWQb691O5" \
"7x26JQFLF0JYS2evRN34xUAzNCOQmjDdQD34VhKWQ88nb+sBVc+nD9h1A8V6WiAQpkgrQYhGK0+D" \
"nOrADVlfFdgE0xAIufQOj0hcshICKUqXrC9HHB+IEsZJB8XwipUQBm609hhwjzKvBBDhsqM4+vs7" \
"rYSQ1ObXDHSxLtMQEJhTdqC+7q5u+J8a0yEopJXQhcLKwK/FR0fK+iogek1D4Ih90qtDQmyklRDU" \
"MGeErK9ANA8BAKQlzvp7o6yEILgiDYERtZuvCaRJNyTC4YqxEsLdTrDvstRfCUQJV+WrC061uJ49" \
"bGB1aA6kWTonXROIplsMQRo6R2wIJB1OS9NmbKalhRFYqjQEVT1lPh1U/wkDcc3ZVviSJSvE1tLF" \
"EzjqH9P92fB1gNpnXgmri3ZeBICzkrvDUb0x8XMtaZlHhb9iwP1YbskefyCFETQSX8iO5ufK26EG" \
"sKk83aECWyWdNoS6Bz26EBSgWgAQkgXoCW+le0FIa4ErPIMjSm3xNaKLB9rgWMAQVudvvwQAh+VX" \
"Svxwy4aMSaEA4PNkJjEir/yeQHiH++3CMIRBSanFBEJq10bIx/c7nXVez5K4IKfBRE1xHiXksp3p" \
"WWxuq5WDJWGF+TVnGNFm6Z4B2FR0RDR6fcteDwaADyrdczVXZINsGgCAAKG6ZY7pBuOVtOKN6c5o" \
"R8QpjmioKdKImgBgNyftuEOjn/wOR6veM9E3O8Lbo0c/pCDNBoBsBGM9CIEoz8/b9q78pBmwLRsy" \
"JvnDwr4n5ONDWfyQNCLkpv5KRCD27o/CRfXuagrKecJQW1m86zIIep5AtIQSQiAAlL6eJUYAGFbC" \
"bav0Zicg8kNGZRoyaINFu3R/FJYZBWAawu2mZcAZvp4jywMDf90JAYBGFCI7r6DmhOn0CzSIKk/m" \
"FFAc6wn5AithEIgfUQjfvnb+iUwvEFIIf8GocN+rMraYI5uHpD1pNq91PvwqCnEEAD7PK6j5JmiF" \
"OBSzVFbhvsfFcDYDSCHkUzWiBxnSfQAwFoGNklgdmlVgrQzpKgCcUwl/AKDvivK3XQjJamR1Dvt8" \
"b+kWrry8jyyNi4FtNgQbgg3BhmBDsCHYEGwINgQbgg3BhmBDsCHYZpth+wOnklW5nlFgrwAAAABJ" \
"RU5ErkJggg=="] ""]

catch {
  image create photo icon_photo -data [expr {([is_stack_mode]) ? $icon_data : $icon_data_batch}] -format "png"
  wm iconphoto $topw -default icon_photo
}

##########################################

set select_foreground_inner_block "#000080"
set select_foreground_line_separs "#808000"

set text_inner_line_parts [list]
set text_inner_code_blocks [list]

set tag_text_line "line_text"
set tag_text_line_block "line_text_block"
set tag_text_line_part "line_text_part"

set text_highlight_line_shades_list [list "#e8e8e8" "#f0f0f0" "#f4f4f4"]
set text_highlight_line_shades_block_list [list "#d8c8ff" "#d0d0ff" "#e0e0ff"]
set text_highlight_line_shades_part_list [list "#feffc0" "#feffd0" "#feffda"]
set text_highlight_line_shades_index 1

$text_field tag configure $tag_text_line -background [lindex $text_highlight_line_shades_list $text_highlight_line_shades_index]
$text_field tag configure $tag_text_line_block -background [lindex $text_highlight_line_shades_block_list $text_highlight_line_shades_index]
$text_field tag configure $tag_text_line_part -background [lindex $text_highlight_line_shades_part_list $text_highlight_line_shades_index]

catch {
  $text_field tag configure $tag_text_line -selectbackground $select_bg_col -selectforeground $select_fg_col
  $text_field tag configure $tag_text_line_block -selectbackground $select_bg_col -selectforeground $select_foreground_inner_block
  $text_field tag configure $tag_text_line_part -selectbackground $select_bg_col -selectforeground $select_foreground_line_separs
}

set tag_text_bracket_pair "text_bracket_pair"
set tag_text_bracket "text_bracket"

$text_field tag configure $tag_text_bracket_pair -foreground "#4040ff"
$text_field tag configure $tag_text_bracket -foreground "#ff8000"
catch {
  # -selectbackground according to the other properties in the place
  $text_field tag configure $tag_text_bracket_pair -selectforeground "#4040ff"
  $text_field tag configure $tag_text_bracket -selectforeground "#ff8000"
}

set text_inner_codes_start_tag "codes_start"
set text_inner_codes_end_tag "codes_end"
set text_inner_codes_start_vis_tag "codes_start_vis"
set text_inner_codes_end_vis_tag "codes_end_vis"
$text_field tag configure $text_inner_codes_start_tag -background "#ffffff"
$text_field tag configure $text_inner_codes_end_tag -background "#ffffff"
$text_field tag configure $text_inner_codes_start_vis_tag -background "#ffffff"
$text_field tag configure $text_inner_codes_end_vis_tag -background "#ffffff"

catch {
  $text_field tag configure $text_inner_codes_start_tag -selectforeground $select_foreground_inner_block -selectbackground $select_bg_col
  $text_field tag configure $text_inner_codes_end_tag -selectforeground $select_foreground_inner_block -selectbackground $select_bg_col
  $text_field tag configure $text_inner_codes_start_vis_tag -selectforeground $select_foreground_inner_block -selectbackground $select_bg_col
  $text_field tag configure $text_inner_codes_end_vis_tag -selectforeground $select_foreground_inner_block -selectbackground $select_bg_col
}

$text_field tag configure $text_inner_codes_start_tag -lmargin1 0 -lmargin2 5
$text_field tag configure $text_inner_codes_end_tag -lmargin1 0 -lmargin2 5
$text_field tag configure $text_inner_codes_start_vis_tag -lmargin1 0 -lmargin2 5
$text_field tag configure $text_inner_codes_end_vis_tag -lmargin1 0 -lmargin2 5

set text_inner_codes_tag "codes"
if {$spec_base_font(command)==""} {
  $text_field tag configure $text_inner_codes_tag -font command_font
  $text_field tag configure $text_inner_codes_start_tag -font command_font
  $text_field tag configure $text_inner_codes_end_tag -font command_font
  $text_field tag configure $text_inner_codes_start_vis_tag -font command_font
  $text_field tag configure $text_inner_codes_end_vis_tag -font command_font
} {
  $text_field tag configure $text_inner_codes_tag -font $spec_base_font(command)
  $text_field tag configure $text_inner_codes_start_tag -font $spec_base_font(command)
  $text_field tag configure $text_inner_codes_end_tag -font $spec_base_font(command)
  $text_field tag configure $text_inner_codes_start_vis_tag -font $spec_base_font(command)
  $text_field tag configure $text_inner_codes_end_vis_tag -font $spec_base_font(command)
}

$text_field tag configure $text_inner_codes_tag -lmargin1 0 -lmargin2 5
$text_field tag configure $text_inner_codes_tag -background "#ffffff"
catch {
  $text_field tag configure $text_inner_codes_tag -selectforeground $select_foreground_inner_block -selectbackground $select_bg_col
}
catch {
  $text_field tag configure $text_inner_codes_tag -lmargincolor "#a0a0e0"
}

set text_inner_star_line_tag "inner_sep_star"
set text_inner_hash_line_tag "inner_sep_hash"
set text_inner_equa_line_tag "inner_sep_equals"
$text_field tag configure $text_inner_star_line_tag -background "#f0f0d8" -foreground "#604000"
$text_field tag configure $text_inner_star_line_tag -relief sunken -borderwidth $section_line_border_width
$text_field tag configure $text_inner_equa_line_tag -background "#f0f0b0" -foreground "#503000"
$text_field tag configure $text_inner_equa_line_tag -relief solid -borderwidth $section_line_border_width
$text_field tag configure $text_inner_hash_line_tag -background "#e4d460" -foreground "#402000"
$text_field tag configure $text_inner_hash_line_tag -relief raised -borderwidth $section_line_border_width

set known_selectforeground_selectbackground_for_tags true
if {[catch {
  $text_field tag configure $text_inner_star_line_tag -selectforeground $select_foreground_line_separs -selectbackground $select_bg_col
  $text_field tag configure $text_inner_equa_line_tag -selectforeground $select_foreground_line_separs -selectbackground $select_bg_col
  $text_field tag configure $text_inner_hash_line_tag -selectforeground $select_foreground_line_separs -selectbackground $select_bg_col
}]} {
  set known_selectforeground_selectbackground_for_tags false
}

set tag_text_all "whole_text"
$text_field tag configure $tag_text_all -lmargin1 0 -lmargin2 5
catch {
  $text_field tag configure $tag_text_all -lmargincolor "#808080"
}

# tag visual priorities: text part

$text_field tag raise "sel"

$text_field tag raise $text_inner_codes_tag

$text_field tag raise $text_output_tags_name
$text_field tag raise $text_input_tags_name
$text_field tag raise $text_lisp_tags_name

$text_field tag raise $text_inner_codes_start_tag
$text_field tag raise $text_inner_codes_end_tag
$text_field tag raise $text_inner_codes_start_vis_tag
$text_field tag raise $text_inner_codes_end_vis_tag

$text_field tag raise $text_inner_star_line_tag
$text_field tag raise $text_inner_equa_line_tag
$text_field tag raise $text_inner_hash_line_tag

$text_field tag raise $tag_text_line
$text_field tag raise $tag_text_line_block
$text_field tag raise $tag_text_line_part

$text_field tag raise $search_tag_all

$text_field tag raise $tag_text_bracket_pair
$text_field tag raise $tag_text_bracket

$text_field tag lower $tag_text_all
$text_field tag add $tag_text_all "1.0" "end"

if {!$known_selectforeground_selectbackground_for_tags} {
  $text_field tag raise "sel"
}

set_tilde_hiding false

##########################################

set cmmd_highlight_line_shades_list [list "#e8e8e8" "#f0f0f0" "#f4f4f4"]
set cmmd_highlight_line_shades_list_saving_paused [list "#e4e4e4" "#e8e8e8" "#f0f0f0"]
set cmmd_highlight_line_shades_list_saving [list "#e0e0e0" "#e4e4e4" "#e8e8e8"]
set cmmd_highlight_line_shades_index 1

$command_field tag configure $tag_stripes -background [lindex $cmmd_highlight_line_shades_list $cmmd_highlight_line_shades_index]
$command_field_aux tag configure $tag_stripes -background [lindex $cmmd_highlight_line_shades_list $cmmd_highlight_line_shades_index]

set tag_command_strings "command_strings"
set tag_command_comments "command_comments"
set tag_command_bracket_pair "command_bracket_pair"
set tag_command_bracket "command_bracket"

$command_field tag configure $tag_command_strings -foreground "#805020"
$command_field tag configure $tag_command_comments -foreground "#605060"
$command_field tag configure $tag_command_bracket_pair -foreground "#00a0f0"
$command_field tag configure $tag_command_bracket -foreground "#f03000"

# tag visual priorities: command part

$command_field tag lower $tag_stripes
$command_field_aux tag lower $tag_stripes

$command_field tag raise "sel"

$command_field tag raise $tag_command_strings
$command_field tag raise $tag_command_comments

$command_field tag raise $tag_command_bracket_pair
$command_field tag raise $tag_command_bracket

if {!$known_selectforeground_selectbackground_for_tags} {
  $command_field tag raise "sel"
}

##########################################

bind $text_field <1> {
  remove_torn_menus
  if {[is_text_field_active]} {
    place_cursor_on_mouse $text_field %X %Y %x %y
    $text_field tag remove "sel" "1.0" "end"
    set_anchor_insert $text_field
    update_text_highlight_lines_onset
    break
  }
}
bind $text_field <Control-1> {
  remove_torn_menus
  if {[is_text_field_active]} {
    place_cursor_on_mouse $text_field %X %Y %x %y
    set_anchor_insert $text_field
    update_text_highlight_lines_onset
  }
  if {[is_batch_mode]} {
    target_within_tag_entries_on_click %X %Y %x %y
  }
  break
}
bind $text_field <Control-ButtonRelease-1> {
  break
}

bind $command_field <1> {
  remove_torn_menus
  if {[is_command_field_active]} {
    place_cursor_on_mouse $command_field %X %Y %x %y
    $command_field tag remove "sel" "1.0" "end"
    set_anchor_insert $command_field
    update_cmmd_highlight_lines_onset
    break
  }
}
bind $command_field <Control-1> {
  remove_torn_menus
  break
}
bind $command_field <Control-ButtonRelease-1> {
  break
}

bind $text_field <3> {
  remove_torn_menus
  focus $text_field
  catch {
    text_field_on_right_click %X %Y %x %y false false false
  }
  if {[is_text_field_active]} {
    set_anchor_insert $text_field
  }
  break
}
bind $text_field <Control-3> {
  remove_torn_menus
  if {[is_text_field_active]} {
    focus $text_field
    catch {
      text_field_on_right_click %X %Y %x %y true false false
    }
    set_anchor_insert $text_field
  }
  break
}
bind $text_field <Shift-3> {
  remove_torn_menus
  if {[is_text_field_active]} {
    focus $text_field
    catch {
      text_field_on_right_click %X %Y %x %y false true false
    }
    set_anchor_insert $text_field
  }
  break
}
bind $text_field <Alt-3> {
  break
}
bind $text_field <Control-Shift-3> {
  remove_torn_menus
  focus $text_field
  catch {
    text_field_on_right_click %X %Y %x %y true true false
  }
  if {[is_text_field_active]} {
    set_anchor_insert $text_field
  }
  break
}
bind $text_field <Control-Alt-3> {
  break
}
bind $text_field <Shift-Alt-3> {
  break
}
bind $text_field <Control-Shift-Alt-3> {
  break
}

bind $command_field <Button-3> {
  remove_torn_menus
  if {[is_stack_mode]} {
    focus $command_field
    catch {
      command_field_on_right_click %X %Y %x %y false false
    }
  }
  break
}
bind $command_field <Control-Button-3> {
  remove_torn_menus
  if {[is_command_field_active]} {
    offer_command_maxima_run_action %X %Y %x %y
  }
  break
}
bind $command_field <Shift-3> {
  break
}
bind $command_field <Alt-3> {
  break
}

proc field_on_double_click {wdg} {
  focus $wdg
  $wdg tag remove "sel" "1.0" "end"

  set index_pos [$wdg index "insert"]
  set index_line [lindex [split $index_pos "."] 0]
  set word_start_pos [$wdg index "insert wordstart"]
  set word_start_line [lindex [split $word_start_pos "."] 0]
  if {$index_line!=$word_start_line} {
    set word_start_pos "$index_line.0"
  }

  $wdg tag add "sel" $word_start_pos "insert wordend"
  set_anchor_position $wdg $word_start_pos
}

bind $text_field <Double-1> {
  if {[is_text_field_active]} {
    focus $text_field
    set acted_on_bind false
    catch {
      set acted_on_bind [text_field_on_double_click %X %Y %x %y false false]
    }
    if {!$acted_on_bind} {
      field_on_double_click $text_field
    }
    break
  }
}
bind $text_field <Control-Double-1> {
  if {[is_text_field_active]} {
    focus $text_field
    set acted_on_bind false
    catch {
      set acted_on_bind [text_field_on_double_click %X %Y %x %y true false]
    }
    if {!$acted_on_bind} {
      set_anchor_insert $text_field
      #field_on_double_click $text_field
    }
    break
  }
}
bind $text_field <Control-Shift-Double-1> {
  if {[is_text_field_active]} {
    focus $text_field
    set acted_on_bind false
    catch {
      set acted_on_bind [text_field_on_double_click %X %Y %x %y true true]
    }
    if {!$acted_on_bind} {
      set_anchor_insert $text_field
      #field_on_double_click $text_field
    }
    break
  }
}

bind $command_field <Double-1> {
  field_on_double_click $command_field
  break
}

bind $command_field <Control-Double-1> {
  if {[is_command_field_active]} {
    ask_for_current_names "functions"
  }
  break
}
bind $command_field <Control-Shift-Double-1> {
  if {[is_command_field_active]} {
    ask_for_current_names "values"
  }
  break
}

proc field_on_multi_click {wdg} {
  focus $wdg
  $wdg tag remove "sel" "1.0" "end"
  $wdg tag add "sel" "insert linestart" "insert lineend"
  set_anchor_insert_line_start $wdg
}

foreach keyspressed [list "Triple-1" "Quadruple-1"] {
  bind $text_field <$keyspressed> {
    field_on_multi_click $text_field
    break
  }
}
foreach keyspressed [list "Triple-1" "Quadruple-1"] {
  bind $command_field <$keyspressed> {
    field_on_multi_click $command_field
    break
  }
}
foreach keyspressed [list "Control-Triple-1" "Control-Quadruple-1"] {
  bind $text_field <$keyspressed> {
    break
  }
}
foreach keyspressed [list "Control-Triple-1" "Control-Quadruple-1"] {
  bind $command_field <$keyspressed> {
    break
  }
}

set text_bracket_jump_position ""

proc text_jump_bracket {{to_select false}} {
  global text_bracket_jump_position
  global text_field

  if {$text_bracket_jump_position==""} {
    return
  }

  if {$to_select} {
    set start_pos [$text_field index "insert"]
    set start_pos_info [split $start_pos "."]
    set start_pos_row [lindex $start_pos_info 0]
    set start_pos_col [lindex $start_pos_info 1]
    set_anchor_insert $text_field
  }

  $text_field mark set "insert" $text_bracket_jump_position
  $text_field see $text_bracket_jump_position

  if {$to_select} {
    set end_pos [$text_field index "insert"]
    set end_pos_info [split $end_pos "."]
    set end_pos_row [lindex $end_pos_info 0]
    set end_pos_col [lindex $end_pos_info 1]
    if {($start_pos_row<$end_pos_row) || (($start_pos_row==$end_pos_row)&&($start_pos_col<$end_pos_col))} {
      set sel_from $start_pos
      set sel_to $end_pos
    } {
      set sel_from $end_pos
      set sel_to $start_pos
    }
    $text_field tag remove "sel" "1.0" "end"
    $text_field tag add "sel" "$sel_from" "$sel_to"
  } {
    set_anchor_insert $text_field
  }

  after 0 "update_text_highlight_lines"
  after 0 "update_text_bracket_pairing"
}

proc update_text_bracket_pairing {} {
  global text_field
  global bracket_complements
  global tag_text_bracket_pair
  global tag_text_bracket
  global text_bracket_jump_position

  set text_bracket_jump_position ""

  set use_any false
  set use_bracket ""

  set insert_index_info [split [$text_field index "insert"] "."]
  set use_position_line [lindex $insert_index_info 0]
  set use_position_column [lindex $insert_index_info 1]
  if {$use_position_column>0} {
    set use_bracket [$text_field get "insert - 1 chars" "insert"]
    if {$use_bracket in [array names bracket_complements]} {
      incr use_position_column -1
      set use_any true
    }
  }

  if {!$use_any} {
    set use_bracket [$text_field get "insert" "insert + 1 chars"]
    if {$use_bracket in [array names bracket_complements]} {
      set use_any true
    }
  }

  $text_field tag remove $tag_text_bracket_pair "1.0" "end"
  $text_field tag remove $tag_text_bracket "1.0" "end"

  if {!$use_any} {
    return
  }

  set brackets_open [list "\{" "\[" "("]

  if {$use_bracket in $brackets_open} {
    # going forwards
    brackets_seek_text_forwards $use_bracket $use_position_line $use_position_column
    return
  }

  # going backwards
  brackets_seek_text_backwards $use_bracket $use_position_line $use_position_column
}

proc brackets_seek_text_forwards {use_bracket use_position_line use_position_column} {
  global text_field
  global tag_text_bracket_pair
  global tag_text_bracket
  global bracket_complements
  global text_bracket_jump_position

  set max_line_count_seek 250
  set main_position_index "$use_position_line.$use_position_column"

  set use_line_part [string trimright [$text_field get "$main_position_index + 1 chars" "$use_position_line.end"]]
  set use_line_chars [split $use_line_part {}]

  set curr_offset [expr {$use_position_column + 1}]
  set opposing_bracket $bracket_complements($use_bracket)
  set bracket_stack 1

  set pair_found false
  set pair_line ""
  set pair_column ""

  # simplified-only guessing on strings and comments here
  set in_string false
  set in_comment false
  set comment_depth 0

  set end_position_line [lindex [split [$text_field index "end"] "."] 0]

  set line_count 0
  while {true} {
    set use_line_length [llength $use_line_chars]
    set char_idx_max [expr {$use_line_length - 1}]
    set char_idx -1
    while {true} {
      incr char_idx 1
      if {$char_idx>=$use_line_length} {
        break
      }
      set curr_char [lindex $use_line_chars $char_idx]
      set prev_char ""
      if {$char_idx>0} {
        set prev_char [lindex $use_line_chars [expr {$char_idx-1}]]
      }
      if {$in_string} {
        if {($curr_char=="\"") && ($prev_char!="\\")} {
          set in_string false
        }
        continue
      }
      set next_char ""
      if {$char_idx<$char_idx_max} {
        set next_char [lindex $use_line_chars [expr {$char_idx+1}]]
      }
      if {($curr_char=="/") && ($next_char=="*")} {
        incr char_idx 1
        set in_comment true
        incr comment_depth
        continue
      }
      if {($curr_char=="*") && ($next_char=="/")} {
        incr char_idx 1
        incr comment_depth -1
        if {$comment_depth==0} {
          set in_comment false
        }
        if {$comment_depth<0} {
          # getting here, we probably started inside a comment,
          # then not doing bracket pairing into outside;
          break
        }
        continue
      }
      if {$in_comment} {
        continue
      }
      if {($curr_char=="\"") && ($prev_char!="\\")} {
        set in_string true
        continue
      }
      if {$curr_char==$use_bracket} {
        incr bracket_stack
        continue
      }
      if {$curr_char==$opposing_bracket} {
        incr bracket_stack -1
        if {$bracket_stack==0} {
          set pair_found true
          set pair_line $use_position_line
          set pair_column $char_idx
          if {$line_count==0} {
            incr pair_column $curr_offset
          }
          break
        }
        continue
      }
    }

    if {$pair_found} {
      break
    }
    incr use_position_line 1
    if {$use_position_line>$end_position_line} {
      break
    }
    incr line_count
    if {$line_count > $max_line_count_seek} {
      break
    }
    set use_line_part [string trimright [$text_field get "$use_position_line.0" "$use_position_line.end"]]
    set use_line_chars [split $use_line_part {}]
  }

  if {$pair_found} {
    $text_field tag add $tag_text_bracket_pair $main_position_index "$main_position_index + 1 chars"
    set found_index "$pair_line.$pair_column"
    $text_field tag add $tag_text_bracket_pair $found_index "$found_index + 1 chars"
    set text_bracket_jump_position "$found_index + 1 chars"
  } {
    $text_field tag add $tag_text_bracket $main_position_index "$main_position_index + 1 chars"
  }

  return
}

proc brackets_seek_text_backwards {use_bracket use_position_line use_position_column} {
  global text_field
  global tag_text_bracket_pair
  global tag_text_bracket
  global bracket_complements
  global text_bracket_jump_position

  set max_line_count_seek 250
  set main_position_index "$use_position_line.$use_position_column"

  if {($use_position_line==1) && ($use_position_column==0)} {
    $text_field tag add $tag_text_bracket $main_position_index "$main_position_index + 1 chars"
    return
  }

  set use_line_part [string trimright [$text_field get "$use_position_line.0" "$main_position_index"]]
  set use_line_chars [split $use_line_part {}]

  set opposing_bracket $bracket_complements($use_bracket)
  set bracket_stack 1

  set pair_found false
  set pair_line ""
  set pair_column ""

  # simplified-only guessing on strings and comments here
  set in_string false
  set in_comment false
  set comment_depth 0

  set line_count 0
  while {true} {
    set use_line_length [llength $use_line_chars]
    set char_idx $use_line_length
    while {true} {
      incr char_idx -1
      if {$char_idx<0} {
        break
      }
      set curr_char [lindex $use_line_chars $char_idx]
      set prev_char ""
      if {$char_idx>0} {
        set prev_char [lindex $use_line_chars [expr {$char_idx-1}]]
      }
      if {$in_string} {
        if {($curr_char=="\"") && ($prev_char!="\\")} {
          set in_string false
        }
        continue
      }
      if {($curr_char=="/") && ($prev_char=="*")} {
        incr char_idx -1
        set in_comment true
        incr comment_depth
        continue
      }
      if {($curr_char=="*") && ($prev_char=="/")} {
        incr char_idx -1
        incr comment_depth -1
        if {$comment_depth==0} {
          set in_comment false
        }
        if {$comment_depth<0} {
          # getting here, we probably started inside a comment,
          # then not doing bracket pairing into outside;
          break
        }
        continue
      }
      if {$in_comment} {
        continue
      }
      if {($curr_char=="\"") && ($prev_char!="\\")} {
        set in_string true
        continue
      }
      if {$curr_char==$use_bracket} {
        incr bracket_stack
        continue
      }
      if {$curr_char==$opposing_bracket} {
        incr bracket_stack -1
        if {$bracket_stack==0} {
          set pair_found true
          set pair_line $use_position_line
          set pair_column $char_idx
          break
        }
        continue
      }
    }

    if {$pair_found} {
      break
    }
    incr use_position_line -1
    if {$use_position_line<1} {
      break
    }
    incr line_count
    if {$line_count > $max_line_count_seek} {
      break
    }
    set use_line_part [string trimright [$text_field get "$use_position_line.0" "$use_position_line.end"]]
    set use_line_chars [split $use_line_part {}]
  }

  if {$pair_found} {
    $text_field tag add $tag_text_bracket_pair $main_position_index "$main_position_index + 1 chars"
    set found_index "$pair_line.$pair_column"
    $text_field tag add $tag_text_bracket_pair $found_index "$found_index + 1 chars"
    set text_bracket_jump_position "$found_index + 1 chars"
  } {
    $text_field tag add $tag_text_bracket $main_position_index "$main_position_index + 1 chars"
  }

  return
}

proc update_text_highlight_lines_onset {} {
  after 0 "update_text_highlight_lines"
  after 0 "update_text_bracket_pairing"
}
proc update_text_highlight_lines_onset_23 {} {
  global highlight_text_blocks_mode

  set after_ms 0
  if {$highlight_text_blocks_mode==0} {
    set after_ms 100
  }

  after $after_ms "update_text_highlight_lines"
  after $after_ms "update_text_bracket_pairing"
  clear_temporary_highlights
}

proc update_cmmd_states_onset {} {
  global command_stack_saving_index

  after 0 update_cmmd_highlight_lines
  if {$command_stack_saving_index != -1} {
    after 0 app_labels_on_modified
  }
  after 0 update_cmmd_strings_comments
}
proc update_cmmd_states_onset_23 {} {
  global command_stack_saving_index
  global show_cmmd_highlight_line_nums

  set after_ms 0
  if {$show_cmmd_highlight_line_nums} {
    set after_ms 100
  }
  after $after_ms update_cmmd_highlight_lines

  if {$command_stack_saving_index != -1} {
    after 100 app_labels_on_modified
  }
  after 100 update_cmmd_strings_comments
}

bind $text_field <Control-b> {
  if {[is_text_field_active]} {
    text_jump_bracket
  }
  break
}
bind $text_field <Control-B> {
  if {[is_text_field_active]} {
    text_jump_bracket true
  }
  break
}

bind $command_field <Control-b> {
  if {[is_command_field_active]} {
    command_jump_bracket
  }
  break
}
bind $command_field <Control-B> {
  if {[is_command_field_active]} {
    command_jump_bracket true
  }
  break
}

proc command_jump_bracket {{to_select false}} {
  global command_bracket_jump_position
  global command_field

  if {$command_bracket_jump_position==""} {
    return
  }

  if {$to_select} {
    set start_pos [$command_field index "insert"]
    set start_pos_info [split $start_pos "."]
    set start_pos_row [lindex $start_pos_info 0]
    set start_pos_col [lindex $start_pos_info 1]
    set_anchor_insert $command_field
  }

  $command_field mark set "insert" $command_bracket_jump_position
  $command_field see $command_bracket_jump_position

  if {$to_select} {
    set end_pos [$command_field index "insert"]
    set end_pos_info [split $end_pos "."]
    set end_pos_row [lindex $end_pos_info 0]
    set end_pos_col [lindex $end_pos_info 1]
    if {($start_pos_row<$end_pos_row) || (($start_pos_row==$end_pos_row)&&($start_pos_col<$end_pos_col))} {
      set sel_from $start_pos
      set sel_to $end_pos
    } {
      set sel_from $end_pos
      set sel_to $start_pos
    }
    $command_field tag remove "sel" "1.0" "end"
    $command_field tag add "sel" "$sel_from" "$sel_to"
  } {
    set_anchor_insert $command_field
  }

  after 0 "update_cmmd_highlight_lines"
  after 0 "update_cmmd_bracket_pairing"
}

proc batch_maxima_jump_bracket {{to_select false}} {
  global mxmbatch_input_field
  global command_bracket_jump_position

  if {$mxmbatch_input_field==""} {
    return
  }
  if {$command_bracket_jump_position==""} {
    return
  }

  if {$to_select} {
    set start_pos [$mxmbatch_input_field index "insert"]
    set start_pos_info [split $start_pos "."]
    set start_pos_row [lindex $start_pos_info 0]
    set start_pos_col [lindex $start_pos_info 1]
    set_anchor_insert $mxmbatch_input_field
  }

  $mxmbatch_input_field mark set "insert" $command_bracket_jump_position
  $mxmbatch_input_field see $command_bracket_jump_position

  if {$to_select} {
    set end_pos [$mxmbatch_input_field index "insert"]
    set end_pos_info [split $end_pos "."]
    set end_pos_row [lindex $end_pos_info 0]
    set end_pos_col [lindex $end_pos_info 1]
    if {($start_pos_row<$end_pos_row) || (($start_pos_row==$end_pos_row)&&($start_pos_col<$end_pos_col))} {
      set sel_from $start_pos
      set sel_to $end_pos
    } {
      set sel_from $end_pos
      set sel_to $start_pos
    }
    $mxmbatch_input_field tag remove "sel" "1.0" "end"
    $mxmbatch_input_field tag add "sel" "$sel_from" "$sel_to"
  } {
    set_anchor_insert $mxmbatch_input_field
  }

  after 0 "update_cmmd_bracket_pairing"
}

array set bracket_complements {
  "\{" "\}"
  "\}" "\{"
  "\[" "\]"
  "\]" "\["
  "(" ")"
  ")" "("
}

set command_bracket_stack [list]
set command_bracket_jump_position ""

proc update_cmmd_bracket_pairing {} {
  global command_bracket_stack
  global bracket_complements
  global command_bracket_jump_position
  global tag_command_bracket_pair
  global tag_command_bracket
  global command_field
  global mxmbatch_input_field

  set use_field [expr {[is_stack_mode] ? $command_field : $mxmbatch_input_field}]
  if {$use_field==""} {
    return
  }
  $use_field tag remove $tag_command_bracket_pair "1.0" "end"
  $use_field tag remove $tag_command_bracket "1.0" "end"

  set command_bracket_jump_position ""

  set insert_pos [split [$use_field index "insert"] "."]
  set insert_line [lindex $insert_pos 0]
  set insert_column [lindex $insert_pos 1]
  set insert_text [$use_field get "insert" "insert + 1 chars"]
  set use_bracket ""
  set use_bracket_line -1
  set use_bracket_column -1

  # for (position of) the seen bracket in the bracket stack
  set curr_idx 0
  set one_bracket [list]
  set bracket_count [llength $command_bracket_stack]
  set having_bracket false

  if {$insert_column>0} {
    set insert_prev_text [$use_field get "insert - 1 chars" "insert"]
    if {$insert_prev_text in [array names bracket_complements]} {
      set having_bracket true
      set use_bracket $insert_prev_text
      set use_bracket_line $insert_line
      set use_bracket_column [expr {$insert_column-1}]
      # checking whether this bracket is (not) acting as a real bracket
      while {$curr_idx < $bracket_count} {
        set one_bracket [lindex $command_bracket_stack $curr_idx]
        if {($use_bracket_line==[lindex $one_bracket 0]) && ($use_bracket_column==[lindex $one_bracket 1])} {
          break
        }
        incr curr_idx
      }
      if {$curr_idx >= $bracket_count} {
        set curr_idx 0
        set use_bracket ""
        set use_bracket_line -1
        set use_bracket_column -1
        set having_bracket false
      }
    }
  }
  if {$use_bracket==""} {
    if {$insert_text in [array names bracket_complements]} {
      set having_bracket true
      set use_bracket $insert_text
      set use_bracket_line $insert_line
      set use_bracket_column $insert_column
      # checking whether this bracket is (not) acting as a real bracket
      while {$curr_idx < $bracket_count} {
        set one_bracket [lindex $command_bracket_stack $curr_idx]
        if {($use_bracket_line==[lindex $one_bracket 0]) && ($use_bracket_column==[lindex $one_bracket 1])} {
          break
        }
        incr curr_idx
      }
      if {$curr_idx >= $bracket_count} {
        set curr_idx 0
        set use_bracket ""
        set use_bracket_line -1
        set use_bracket_column -1
        set having_bracket false
      }
    }
  }

  if {!$having_bracket} {
    return
  }

  set brackets_open [list "\{" "\[" "("]
  set brackets_close [list "\}" "\]" ")"]

  set bracket_index_char [lindex $one_bracket 2]
  set bracket_index_char_opposite $bracket_complements($bracket_index_char)
  set bracket_index_start_main [join [lrange $one_bracket 0 1] "."] 
  set is_paired false
  set local_stack_size 1
  if {$bracket_index_char in $brackets_open} {
    # search forwards
    incr curr_idx
    while {$curr_idx < $bracket_count} {
      set one_bracket [lindex $command_bracket_stack $curr_idx]
      set one_bracket_char [lindex $one_bracket 2]
      if {$one_bracket_char==$bracket_index_char} {
        incr local_stack_size 1
      } elseif {$one_bracket_char==$bracket_index_char_opposite} {
        incr local_stack_size -1
        if {$local_stack_size==0} {
          break
        }
      }
      incr curr_idx
    }
    if {$curr_idx < $bracket_count} {
      set bracket_index_start [join [lrange $one_bracket 0 1] "."]
      $use_field tag add $tag_command_bracket_pair $bracket_index_start "$bracket_index_start + 1 char"
      set command_bracket_jump_position "$bracket_index_start + 1 chars"
      set is_paired true
    }
  } else {
    # search backwards
    incr curr_idx -1
    while {$curr_idx >= 0} {
      set one_bracket [lindex $command_bracket_stack $curr_idx]
      set one_bracket_char [lindex $one_bracket 2]
      if {$one_bracket_char==$bracket_index_char} {
        incr local_stack_size 1
      } elseif {$one_bracket_char==$bracket_index_char_opposite} {
        incr local_stack_size -1
        if {$local_stack_size==0} {
          break
        }
      }
      incr curr_idx -1
    }
    if {$curr_idx >= 0} {
      set bracket_index_start [join [lrange $one_bracket 0 1] "."]
      $use_field tag add $tag_command_bracket_pair $bracket_index_start "$bracket_index_start + 1 char"
      set command_bracket_jump_position "$bracket_index_start + 1 chars"
      set is_paired true
    }
  }
  set tag_name_use [expr {$is_paired ? $tag_command_bracket_pair : $tag_command_bracket}]
  $use_field tag add $tag_name_use $bracket_index_start_main "$bracket_index_start_main + 1 char"

}

proc update_cmmd_strings_comments {} {
  global command_field
  global tag_command_strings
  global tag_command_comments
  global command_bracket_stack
  global in_lisp_mode
  global in_debug_mode
  global mxmbatch_input_field

  set use_field [expr {[is_stack_mode] ? $command_field : $mxmbatch_input_field}]
  if {$use_field==""} {
    return
  }
  $use_field tag remove $tag_command_strings "1.0" "end"
  $use_field tag remove $tag_command_comments "1.0" "end"

  if {((!$in_lisp_mode) && (!$in_debug_mode)) || [is_batch_mode]} {
    set comments_strings [get_comments_strings_brackets [get_command_content]]
  } {
    set comments_strings [get_comments_strings_brackets_in_lisp [get_command_content]]
  }

  foreach one_comment [lindex $comments_strings 0] {
    set one_start [lindex $one_comment 0]
    set index_start [join $one_start "."]

    set one_end [lindex $one_comment 1]
    set index_end "end"
    if {[lindex $one_end 0]!="end"} {
      set index_end [join $one_end "."]
      append index_end " + 1 chars"
    }
    $use_field tag add $tag_command_comments $index_start $index_end
  }

  foreach one_string [lindex $comments_strings 1] {
    set one_start [lindex $one_string 0]
    set index_start [join $one_start "."]

    set one_end [lindex $one_string 1]
    set index_end "end"
    if {[lindex $one_end 0]!="end"} {
      set index_end [join $one_end "."]
      append index_end " + 1 chars"
    }
    $use_field tag add $tag_command_strings $index_start $index_end
  }

  set command_bracket_stack [lindex $comments_strings 2]
  update_cmmd_bracket_pairing
  return
}

foreach wdg [list $topw $text_field $command_field] {
  bind $wdg <Alt-slash> {
    setting_cmmd_highlight_shade
    break
  }
  bind $wdg <Alt-backslash> {
    setting_text_highlight_shade
    break
  }
}

bind $state_label <Shift-1> {
  setting_text_highlight_shade
  break
}
bind $state_label <Control-Shift-1> {
  setting_cmmd_highlight_shade
  break
}

proc setting_cmmd_highlight_shade {{step 1}} {
  global command_field
  global command_field_aux
  global cmmd_highlight_line_shades_list
  global cmmd_highlight_line_shades_list_saving
  global cmmd_highlight_line_shades_list_saving_paused
  global cmmd_highlight_line_shades_index
  global tag_stripes
  global command_stack_index
  global command_stack_saving_index
  global command_stack_saving_paused

  incr cmmd_highlight_line_shades_index $step
  if {$cmmd_highlight_line_shades_index==[llength $cmmd_highlight_line_shades_list]} {
    set cmmd_highlight_line_shades_index 0
  }

  if {$command_stack_saving_index!=$command_stack_index} {
    set shades_use $cmmd_highlight_line_shades_list
  } {
    if {$command_stack_saving_paused} {
      set shades_use $cmmd_highlight_line_shades_list_saving_paused
    } {
      set shades_use $cmmd_highlight_line_shades_list_saving
    }
  }

  set shade_value [lindex $shades_use $cmmd_highlight_line_shades_index]
  $command_field tag configure $tag_stripes -background $shade_value
  $command_field_aux tag configure $tag_stripes -background $shade_value
}

proc setting_text_highlight_shade {} {
  global text_field
  global text_highlight_line_shades_list
  global text_highlight_line_shades_block_list
  global text_highlight_line_shades_part_list
  global text_highlight_line_shades_index
  global tag_text_line
  global tag_text_line_block
  global tag_text_line_part

  incr text_highlight_line_shades_index
  if {$text_highlight_line_shades_index==[llength $text_highlight_line_shades_list]} {
    set text_highlight_line_shades_index 0
  }

  $text_field tag configure $tag_text_line -background [lindex $text_highlight_line_shades_list $text_highlight_line_shades_index]
  $text_field tag configure $tag_text_line_block -background [lindex $text_highlight_line_shades_block_list $text_highlight_line_shades_index]
  $text_field tag configure $tag_text_line_part -background [lindex $text_highlight_line_shades_part_list $text_highlight_line_shades_index]
}

foreach wdg [list $topw $text_field $command_field] {
  bind $wdg <Control-slash> {
    if {[is_stack_mode]} {
      setting_cmmd_highlight_lines
    }
    break
  }
  bind $wdg <Control-backslash> {
    setting_text_highlight_blocks 1
    break
  }
}

bind $state_label <1> {
  setting_text_highlight_blocks 1
  break
}
bind $state_label <Control-1> {
  setting_cmmd_highlight_lines
  break
}

proc setting_text_highlight_blocks {inc} {
  global highlight_text_blocks_mode
  global text_inner_star_line_tag
  global text_inner_equa_line_tag
  global text_inner_hash_line_tag
  global text_field
  global text_inner_codes_start_tag
  global text_inner_codes_end_tag
  global text_inner_codes_start_vis_tag
  global text_inner_codes_end_vis_tag
  global text_inner_codes_tag
  global text_input_tags_name
  global text_output_tags_name
  global text_lisp_tags_name

  set text_inner_code_block_color "#e8f0ff"

  incr highlight_text_blocks_mode $inc
  if {$highlight_text_blocks_mode==3} {
    set highlight_text_blocks_mode 0
  }

  if {$highlight_text_blocks_mode==0} {
    $text_field tag lower $text_input_tags_name $text_inner_codes_tag
    $text_field tag lower $text_output_tags_name $text_inner_codes_tag
    $text_field tag lower $text_lisp_tags_name $text_inner_codes_tag
  } {
    $text_field tag raise $text_input_tags_name $text_inner_codes_tag
    $text_field tag raise $text_output_tags_name $text_inner_codes_tag
    $text_field tag raise $text_lisp_tags_name $text_inner_codes_tag
  }

  $text_field tag configure $text_inner_star_line_tag -justify [lindex [list "center" "left" "left"] $highlight_text_blocks_mode]
  $text_field tag configure $text_inner_equa_line_tag -justify [lindex [list "center" "left" "left"] $highlight_text_blocks_mode]
  $text_field tag configure $text_inner_hash_line_tag -justify [lindex [list "center" "left" "left"] $highlight_text_blocks_mode]

  $text_field tag configure $text_inner_codes_tag -background [lindex [list $text_inner_code_block_color "#ffffff" "#ffffff"] $highlight_text_blocks_mode]

  $text_field tag configure $text_inner_codes_tag -rmargin [lindex [list 0 0 10] $highlight_text_blocks_mode]

  catch {
    $text_field tag configure $text_inner_codes_tag -rmargincolor $text_inner_code_block_color
  }

  $text_field tag configure $text_inner_codes_start_tag -background [lindex [list $text_inner_code_block_color "#d0e0ff" "#ffffff"] $highlight_text_blocks_mode]
  $text_field tag configure $text_inner_codes_end_tag -background [lindex [list $text_inner_code_block_color "#d0d0d0" "#ffffff"] $highlight_text_blocks_mode]
  $text_field tag configure $text_inner_codes_start_vis_tag -background [lindex [list $text_inner_code_block_color "#d0e0ff" "#ffffff"] $highlight_text_blocks_mode]
  $text_field tag configure $text_inner_codes_end_vis_tag -background [lindex [list $text_inner_code_block_color "#d0d0d0" "#ffffff"] $highlight_text_blocks_mode]

  $text_field tag configure $text_inner_codes_start_tag -rmargin [lindex [list 0 0 10] $highlight_text_blocks_mode]
  $text_field tag configure $text_inner_codes_end_tag -rmargin [lindex [list 0 0 10] $highlight_text_blocks_mode]
  $text_field tag configure $text_inner_codes_start_vis_tag -rmargin [lindex [list 0 0 10] $highlight_text_blocks_mode]
  $text_field tag configure $text_inner_codes_end_vis_tag -rmargin [lindex [list 0 0 10] $highlight_text_blocks_mode]

  catch {
    $text_field tag configure $text_inner_codes_start_tag -rmargincolor $text_inner_code_block_color
    $text_field tag configure $text_inner_codes_end_tag -rmargincolor $text_inner_code_block_color
    $text_field tag configure $text_inner_codes_start_vis_tag -rmargincolor $text_inner_code_block_color
    $text_field tag configure $text_inner_codes_end_vis_tag -rmargincolor $text_inner_code_block_color
  }

  update_text_highlight
  return
}

# need to know whether we are within block-highlighted inner code part,
# so that the highlight color is more visible color than grey there;
proc update_text_highlight_lines {} {
  global text_field
  global tag_text_line
  global tag_text_line_block
  global tag_text_line_part
  global text_inner_code_blocks
  global text_inner_line_parts

  $text_field tag remove $tag_text_line "1.0" "end"
  $text_field tag remove $tag_text_line_block "1.0" "end"
  $text_field tag remove $tag_text_line_part "1.0" "end"

  set curr_line [lindex [split [$text_field index "insert"] "."] 0]
  set curr_line_start "$curr_line.0"

  set next_line [expr {$curr_line + 1}]
  set curr_line_end "$next_line.0"

  set use_tag $tag_text_line
  set form_found false
  # check to be at a part-wise line
  if {$curr_line in $text_inner_line_parts} {
    set use_tag $tag_text_line_part
    set form_found true
  }
  if {!$form_found} {
    # check to be within highlighted block
    foreach block_info $text_inner_code_blocks {
      set block_start [lindex $block_info 0]
      set block_end [lindex $block_info 1]
      if {$curr_line>=$block_start} {
        if {($curr_line<=$block_end) || ("end"==$block_end)} {
          set use_tag $tag_text_line_block
          break
        }
      }
    }
  }

  $text_field tag add $use_tag "$curr_line_start" "$curr_line_end"

  update_text_bracket_pairing

  return
}

update_text_highlight_lines
update_text_bracket_pairing

proc insert_primary {wdg x_abs y_abs x_pos y_pos} {
  set curr_index [$wdg index "@$x_pos,$y_pos"]
  $wdg mark set "insert" $curr_index
  set_anchor_insert $wdg
  if {[catch {
    if {[is_wine_like]} {
      set text_data [selection get -selection "PRIMARY"]
    } {
      set text_data [selection get -selection "PRIMARY" -type "UTF8_STRING"]
    }
  }]} {
    set text_data ""
  }
  if {$text_data!=""} {
    $wdg edit separator
    $wdg insert "insert" $text_data
    $wdg edit separator
  }
}

bind $text_field <ButtonRelease-2> {
  update_text_properties
  break
}
bind $command_field <ButtonRelease-2> {
  update_command_properties
  break
}

bind $text_field <2> {
  insert_primary $text_field %X %Y %x %y
  update_text_highlight_lines_onset_23
  break
}
bind $command_field <2> {
  insert_primary $command_field %X %Y %x %y
  update_cmmd_states_onset_23
  break
}

set empty_keys_for_updates [list "control*" "shift*" "alt*" "super*" "menu*" "iso_next_group"]

proc general_onset_on_text_keys {{keysym false} {with_control false}} {
  global delayed_text_update_id
  global delayed_text_update_lag
  global text_field_needs_update_var
  global keysyms_moves
  global tags_decls_state_clean
  global tags_links_state_clean
  global empty_keys_for_updates
  global cursor_was_at_block_border
  global hide_tag_tildes

  foreach empty $empty_keys_for_updates {
    if {[string match -nocase $empty $keysym]==1} {
      return
    }
  }

  if {$keysym in $keysyms_moves} {
    # onset of brackets && lines settings only;
    # though they are dealt with via another bindings:
    update_text_highlight_lines_onset
    if {!($cursor_was_at_block_border && $hide_tag_tildes)} {
      return
    }
  }

  set tags_decls_state_clean false
  set tags_links_state_clean false

  if {$with_control} {
    # events with control held are generally done by the events;
    # just the movements (dealt above) are done their default ways;
    # if some relevant control-type events left, should put them here;
    return
  }

  # clicks/control-clicks ($keysym==false) are handled by button press events
  if {$keysym!=false} {
    clear_temporary_highlights
    remove_torn_menus
  }

  update_text_properties_delayed
  return
}

# can not use <KeyRelease> here, since numerical state of keypad gives non-numerical
# (that is KP_Left, etc.) events for the KeyRelease event types.
bind $text_field <KeyPress> "+[list general_onset_on_text_keys %K false]"
bind $text_field <Control-KeyPress> "+[list general_onset_on_text_keys %K true]"
bind $text_field <ButtonRelease> "+{general_onset_on_text_keys}"

# this is with an amount of idles-more-delays, since it is too slow on Enter-pressing otherwise
set update_text_properties_delayed_onset_id ""
set after_idle_clear_temporary_highlights_id ""
set text_field_needs_update_var false
proc text_field_needs_update_set {} {
  global text_field_needs_update_var
  global tags_decls_state_clean
  global tags_links_state_clean
  global update_text_properties_delayed_onset_id
  global after_idle_clear_temporary_highlights_id
  global text_field

  set tags_decls_state_clean false
  set tags_links_state_clean false

  set text_field_needs_update_var true

  if {$update_text_properties_delayed_onset_id!=""} {
    catch {after cancel $update_text_properties_delayed_onset_id}
    set update_text_properties_delayed_onset_id ""
  }

  set end_row [lindex [split [$text_field index "end"] "."] 0]
  set delay_int [expr {($end_row<100) ? 20 : (($end_row<400) ? 60 : 200)}]

  set update_text_properties_delayed_onset_id [after $delay_int update_text_properties_delayed_onset]

  if {$after_idle_clear_temporary_highlights_id!=""} {
    catch {after cancel $after_idle_clear_temporary_highlights_id}
    set after_idle_clear_temporary_highlights_id ""
  }
  set after_idle_clear_temporary_highlights_id [after idle clear_temporary_highlights]
}

set after_idle_update_text_properties_delayed_id ""
proc update_text_properties_delayed_onset {} {
  global after_idle_update_text_properties_delayed_id

  if {$after_idle_update_text_properties_delayed_id!=""} {
    catch {after cancel $after_idle_update_text_properties_delayed_id}
    set after_idle_update_text_properties_delayed_id ""
  }

  set after_idle_update_text_properties_delayed_id [after idle update_text_properties_delayed]
}

# this avoids the "general_onset_on_text_keys" function for "Enter" pressing
bind $text_field <Return> "+{text_field_needs_update_set}"

proc text_got_modified {} {
  remove_torn_menus
  update_text_properties
}

proc text_outer_modified {} {
  clear_temporary_highlights
  text_got_modified
}

bind $text_field <1> "+[list update_text_highlight_lines_onset]"
bind $text_field <Control-1> "+[list update_text_highlight_lines_onset]"
bind $text_field <Control-ButtonRelease-1> "+[list update_text_highlight_lines_onset]"

foreach keyspressed [list "3" "ButtonRelease-3"] {
  bind $text_field <$keyspressed> "+update_text_highlight_lines_onset_23"
}

proc set_cmmd_highlight_lines_mode {to_show} {
  global show_cmmd_highlight_line_nums

  set show_cmmd_highlight_line_nums $to_show
  setting_cmmd_highlight_lines false
}

proc setting_cmmd_highlight_lines {{change true}} {
  global show_cmmd_highlight_line_nums
  global command_field_aux

  if {$change} {
    set show_cmmd_highlight_line_nums [expr {
      $show_cmmd_highlight_line_nums ? false : true
    }]
  }

  if {$show_cmmd_highlight_line_nums} {
    pack $command_field_aux -side left -expand no -fill both
  } {
    pack forget $command_field_aux
  }

  update_cmmd_highlight_lines
}

proc update_cmmd_highlight_lines {} {
  global command_field
  global show_cmmd_highlight_line_nums
  global tag_stripes

  adjust_command_aux_count_pos

  $command_field tag remove $tag_stripes "1.0" "end"

  set curr_line [lindex [split [$command_field index "insert"] "."] 0]
  set curr_line_start "$curr_line.0"

  set next_line [expr {$curr_line + 1}]
  set curr_line_end "$next_line.0"

  $command_field tag add $tag_stripes "$curr_line_start" "$curr_line_end"
}

update_cmmd_highlight_lines

proc update_cmmd_highlight_lines_onset {} {
  after 0 update_cmmd_highlight_lines
  after 0 update_cmmd_bracket_pairing
}

foreach keyspressed [list "Key" "1" "ButtonRelease-1"] {
  bind $command_field <$keyspressed> "+update_cmmd_states_onset"
}
foreach keyspressed [list "3" "ButtonRelease-3"] {
  bind $command_field <$keyspressed> "+update_cmmd_states_onset_23"
}

proc load_names_from_file {path_list} {
  set err_list [list]
  set used_list [list]

  foreach path_term $path_list {
    if {[file isdirectory $path_term]} {
      if {[catch {
        set curr_paths [glob -nocomplain [file join $path_term "*"]]
      }]} {
        set curr_paths [list]
      }
    } {
      set curr_paths [list $path_term]
    }

    foreach path $curr_paths {
      if {$path in $used_list} {
        continue
      }
      lappend used_list $path

      set bad [catch {set channel [open $path "r"]}]
      if {$bad} {
        lappend err_list [list "can not open file with names: \"$path\""]
        continue
      }

      set bad [catch {set fnames_content [read $channel]}]
      if {$bad} {
        lappend err_list [list "can not read from file with names: \"$path\""]
        catch {chan close $channel}
        continue
      }

      catch {chan close $channel}

      set names_list [list]
      foreach one_line [split $fnames_content "\n"] {
        set one_line [string trim $one_line]
        if {$one_line==""} {
          continue
        }
        set first_hash_pos [string first "#" $one_line]
        if {$first_hash_pos==0} {
          continue
        }
        if {$first_hash_pos>-1} {
          set one_line [string trim [string range $one_line 0 "$first_hash_pos-1"]]
        }
        lappend names_list $one_line
      }

      set curr_size [create_keyword_vocabulary $names_list]
    }
  }

  if {[llength $err_list]>0} {
    after 500 [list show_text_field_message_error $err_list]
  }
}

if {$names_from_files_to_load} {
  load_names_from_file $names_from_files_path_list
}

proc get_tag1_tag2_props {tags_search} {

  set tags_search_cnt 0
  set tags_search_len [llength $tags_search]
  if {$tags_search_len>2} {
    return [list false "Too many tag groups at link: $tags_search_len"]
  }

  foreach one_tag_group $tags_search {
    incr tags_search_cnt [llength $one_tag_group]
  }
  if {$tags_search_cnt==0} {
    return [list false "No tag spec within link"]
  }
  if {$tags_search_cnt>2} {
    return [list false "Too many tag specs within link: $tags_search_cnt"]
  }

  set tags_close false
  set has_two_tags false

  set tag_group1 [lindex $tags_search 0]
  set tag_group2 [lindex $tags_search 1]

  if {[llength $tag_group1]>0} {
    set tag_first_info [lindex $tag_group1 0]
    if {[llength $tag_group1]>1} {
      set tag_second_info [lindex $tag_group1 1]
      set tags_close true
      set has_two_tags true
    } {
      if {[llength $tag_group2]>0} {
        set tag_second_info [lindex $tag_group2 0]
        set has_two_tags true
      }
    }
  } {
    if {[llength $tag_group2]>0} {
      set tag_first_info [lindex $tag_group2 0]
      if {[llength $tag_group2]>1} {
        set tag_second_info [lindex $tag_group2 1]
        set tags_close true
        set has_two_tags true
      }
    }
  }

  set tag_first_type [lindex $tag_first_info 0]
  set tag_first_type_lim [lindex $tag_first_info 1]
  set tag_first_value [lindex $tag_first_info 2]
  set tag_first_value_lim [lindex $tag_first_info 3]

  set tag_second_type ""
  set tag_second_type_lim false
  set tag_second_value ""
  set tag_second_value_lim false
  if {$has_two_tags} {
    set tag_second_type [lindex $tag_second_info 0]
    set tag_second_type_lim [lindex $tag_second_info 1]
    set tag_second_value [lindex $tag_second_info 2]
    set tag_second_value_lim [lindex $tag_second_info 3]
  }

  return [list true $has_two_tags \
    [list $tag_first_type $tag_first_type_lim $tag_first_value $tag_first_value_lim] \
    [list $tag_second_type $tag_second_type_lim $tag_second_value $tag_second_value_lim] \
    $tags_close]
}

proc translate_wine_drive {drive_letter} {
  global env

  if {[is_wine_like]} {
    return [strcat $drive_letter ":"]
  }

  if {[file isdirectory [strcat "/" $drive_letter ":"]]} {
    return [strcat "/" $drive_letter ":"]
  }

  if {[file isdirectory [strcat "/" [string tolower $drive_letter] ":"]]} {
    return [strcat "/" [string tolower $drive_letter] ":"]
  }

  if {[file isdirectory [strcat "/" [string toupper $drive_letter] ":"]]} {
    return [strcat "/" [string toupper $drive_letter] ":"]
  }

  set wine_prefix ""
  if {[catch {
    set wine_prefix $env(WINEPREFIX)
    if {![file isdirectory $wine_prefix]} {
      set wine_prefix ""
    }
  }]} {
    set wine_prefix ""
  }
  if {$wine_prefix==""} {
    set wine_prefix [file join "~" ".wine"]
  }

  if {![file isdirectory $wine_prefix]} {
    if {[string tolower $drive_letter]=="z"} {
      return "/"
    }
    return [strcat "/" $drive_letter ":"]
  }

  set aux_wine_dir [file join $wine_prefix [strcat "drive_" [string tolower $drive_letter]]]
  if {[file isdirectory $aux_wine_dir]} {
    return $aux_wine_dir
  }
  set aux_wine_dir [file join $wine_prefix [strcat "drive_" [string toupper $drive_letter]]]
  if {[file isdirectory $aux_wine_dir]} {
    return $aux_wine_dir
  }

  if {[string tolower $drive_letter]=="z"} {
    return "/"
  }
  return [strcat "/" $drive_letter ":"]
}

proc on_tag_link_try_to_find_paths {up_heights seeked_dir seeked_glob sub_depths seeked_tags mnm_only} {
  global last_dir
  global save_path
  global minima_file_extension_base
  global tag_border_size

  set tags_info [get_tag1_tag2_props $seeked_tags]
  if {![lindex $tags_info 0]} {
    return [list false [lindex $tags_info 1]]
  }

  set onset_dirs [list]

  set up_dirs_height 0
  if {[lindex $up_heights 0]} {
    set up_dirs_height [lindex $up_heights 1]
  }

  if {$up_dirs_height>0} {
    set aux_dir_list [list]
    # preparing for the upwards going dirs
    set aux_idx 0
    set aux_dir $last_dir
    set aux_dir_last $aux_dir
    while {$aux_idx <= $up_dirs_height} {
      incr aux_idx
      lappend aux_dir_list $aux_dir
      set aux_dir [file dirname $aux_dir]
      if {$aux_dir_last == $aux_dir} {
        break
      }
      set aux_dir_last $aux_dir
    }
    foreach one_aux_dir $aux_dir_list {
      foreach one_part [split $seeked_dir "/"] {
        if {($one_part=="")||($one_part==".")} {
          continue
        }
        if {$one_part==".."} {
          set one_aux_dir [file dirname $one_aux_dir]
          continue
        }
        set one_aux_dir [file join $one_aux_dir $one_part]
      }
      if {!($one_aux_dir in $onset_dirs)} {
        lappend onset_dirs $one_aux_dir
      }
    }

  } {
    # preparing a single starting dir
    set use_dir $last_dir

    set dir_char_first [string range $seeked_dir 0 0]
    set dir_char_second [string range $seeked_dir 1 1]
    set dir_char_third [string range $seeked_dir 2 2]
    set dir_char_fourth [string range $seeked_dir 3 3]
    if {$dir_char_first=="/"} {
      if {[string is ascii $dir_char_second] && [string is alpha $dir_char_second] && ($dir_char_third==":") && ($dir_char_fourth=="/")} {
        set use_dir [translate_wine_drive $dir_char_second]
        set seeked_dir [string range $seeked_dir 3 "end"]
      } {
        set use_dir "/"
        set seeked_dir [string range $seeked_dir 1 "end"]
      }
      while {[string range $seeked_dir 0 0]=="/"} {
        set seeked_dir [string range $seeked_dir 1 "end"]
      }
    } elseif {$dir_char_first=="~"} {
      set use_dir "~"
      set seeked_dir [string range $seeked_dir 1 "end"]
      while {[string range $seeked_dir 0 0]=="/"} {
        set seeked_dir [string range $seeked_dir 1 "end"]
      }
    }

    foreach one_part [split $seeked_dir "/"] {
      if {($one_part=="")||($one_part==".")} {
        continue
      }
      if {$one_part==".."} {
        set use_dir [file dirname $use_dir]
        continue
      }
      set use_dir [file join $use_dir $one_part]
    }

    lappend onset_dirs $use_dir
  }

  set visiting_files [list]
  set visited_files_norm [list]
  set visiting_dirs [list]
  set visited_dirs_norm [list]

  set sub_dirs_depth 0
  if {[lindex $sub_depths 0]} {
    set sub_dirs_depth [lindex $sub_depths 1]
  }

  # the glob may be (in principle) for onset-wise dirs/files or for final files;
  # as for now, only having the onset-wise case;
  set use_glob_at_start true

  if {($sub_dirs_depth==0) || $use_glob_at_start} {
    foreach use_dir $onset_dirs {
      # if without downwards traversing or the glob used at onset (for files too)

      set use_glob $seeked_glob
      append use_glob [expr {$mnm_only ? ".$minima_file_extension_base" : ".\{$minima_file_extension_base,txt\}"}]

      set seeked_types [list "f"]
      if {[catch {
        set paths_to_seek [lsort -dictionary [glob -nocomplain -types $seeked_types -directory $use_dir -- $use_glob]]
      }]} {
        set paths_to_seek [list]
      }

      foreach one_path $paths_to_seek {
        set one_path_norm [file normalize $one_path]
        if {!($one_path_norm in $visited_files_norm)} {
          lappend visiting_files $one_path
          lappend visited_files_norm $one_path_norm
        }
      }

    }
  }

  if {$sub_dirs_depth>0} {
    if {$use_glob_at_start} {
      incr sub_dirs_depth -1
      foreach use_dir $onset_dirs {
        # if downwards traversing with glob at onset (for dirs too)
        set use_glob $seeked_glob
        set seeked_types [list "d"]
        if {[catch {
          set paths_to_seek [lsort -dictionary [glob -nocomplain -types $seeked_types -directory $use_dir -- $use_glob]]
        }]} {
          set paths_to_seek [list]
        }
        foreach one_path $paths_to_seek {
          set one_path_norm [file normalize $one_path]
          if {!($one_path_norm in $visited_dirs_norm)} {
            lappend visited_dirs_norm $one_path_norm
            # to set the current onset dirs with the depths to go yet from them, so that we can start regular (depth-first) traversing
            lappend visiting_dirs [list $one_path $sub_dirs_depth]
          }
        }
      }
    } {
      foreach use_dir $onset_dirs {
        set one_path $use_dir
        set one_path_norm [file normalize $one_path]
        if {!($one_path_norm in $visited_dirs_norm)} {
          lappend visited_dirs_norm $one_path_norm
          # to set the current onset dirs with the depths to go yet from them, so that we can start regular (depth-first) traversing
          lappend visiting_dirs [list $one_path $sub_dirs_depth]
        }
      }
    }
  }


  # when $use_glob_at_start==true then below dealing with deeper-located files only
  set use_glob [expr {$use_glob_at_start ? "*" : $seeked_glob}]
  append use_glob [expr {$mnm_only ? ".$minima_file_extension_base" : ".\{$minima_file_extension_base,txt\}"}]

  while {true} {
    if {[llength $visiting_dirs]==0} {
      break
    }

    set use_dir_info [lindex $visiting_dirs 0]
    set use_dir [lindex $use_dir_info 0]
    set sub_dirs_depth [lindex $use_dir_info 1]
    set visiting_dirs [lrange $visiting_dirs 1 "end"]

    if {$sub_dirs_depth > 0} {
      incr sub_dirs_depth -1

      # taking all directories below start dirs (if glob used at bottom) or below the filtered dirs (if glob used at top)
      set seeked_types [list "d"]
      if {[catch {
        set paths_to_seek [lsort -dictionary [glob -nocomplain -types $seeked_types -directory $use_dir -- "*"]]
      }]} {
        set paths_to_seek [list]
      }

      foreach one_path $paths_to_seek {
        set one_path_norm [file normalize $one_path]
        if {!($one_path_norm in $visited_dirs_norm)} {
          lappend visited_dirs_norm $one_path_norm
          # to set the current onset dirs with the depths to go yet from them, so that we can start regular (depth-first) traversing
          lappend visiting_dirs [list $one_path $sub_dirs_depth]
        }
      }
    }

    set seeked_types [list "f"]
    if {[catch {
      set paths_to_seek [lsort -dictionary [glob -nocomplain -types $seeked_types -directory $use_dir -- $use_glob]]
    }]} {
      set paths_to_seek [list]
    }

    foreach one_path $paths_to_seek {
      set one_path_norm [file normalize $one_path]
      if {!($one_path_norm in $visited_files_norm)} {
        lappend visiting_files $one_path
        lappend visited_files_norm $one_path_norm
      }
    }

  }

  set has_two_tags [lindex $tags_info 1]

  set tag_first_info [lindex $tags_info 2]
  set tag_first_type [lindex $tag_first_info 0]
  set tag_first_type_lim [lindex $tag_first_info 1]
  set tag_first_value [lindex $tag_first_info 2]
  set tag_first_value_lim [lindex $tag_first_info 3]

  set tag_second_info [lindex $tags_info 3]
  set tag_second_type [lindex $tag_second_info 0]
  set tag_second_type_lim [lindex $tag_second_info 1]
  set tag_second_value [lindex $tag_second_info 2]
  set tag_second_value_lim [lindex $tag_second_info 3]

  set tags_close [lindex $tags_info 4]

  set files_found [list]

  foreach one_path $visiting_files {
    # to do the tag search itself
    if {$has_two_tags} {
      if {$tags_close} {
        set tags_found [find_tag_pair_in_file false \
          $tag_first_type $tag_first_type_lim $tag_first_value $tag_first_value_lim \
          $tag_second_type $tag_second_type_lim $tag_second_value $tag_second_value_lim \
          false $one_path]
      } {
        set tags_found [find_tag_single_in_file false \
          $tag_first_type $tag_first_type_lim $tag_first_value $tag_first_value_lim \
          false $one_path]
        if {[lindex $tags_found 0] && ([llength [lindex $tags_found 1]]>0)} {
          set tags_found [find_tag_single_in_file false \
            $tag_second_type $tag_second_type_lim $tag_second_value $tag_second_value_lim \
            false $one_path]
        }
      }
    } {
      set tags_found [find_tag_single_in_file false \
        $tag_first_type $tag_first_type_lim $tag_first_value $tag_first_value_lim \
        false $one_path]
    }

    if {![lindex $tags_found 0]} {
      # this could be e.g. due to not having access rights to the file
      continue
    }
    if {[llength [lindex $tags_found 1]]==0} {
      # tag was not found
      continue
    }

    set tag_found_info [lindex [lindex $tags_found 1] 0]
    lappend files_found [list $one_path [lindex $tag_found_info 0] [lindex $tag_found_info 1]]
  }

  return [list true $files_found]
}

proc get_tag_decl_index {row_idx column_idx} {
  global tags_decls_list
  global tags_decls_state_clean

  if {!$tags_decls_state_clean} {
    set tags_decls_list [extract_tag_list false]
  }

  # searching for a declared tag at the given place
  set tag_idx -1
  foreach one_tag_info $tags_decls_list {
    incr tag_idx
    set one_tag_pos_info [split [lindex $one_tag_info 1] "."]
    set one_tag_row [lindex $one_tag_pos_info 0]
    if {$row_idx>$one_tag_row} {
      continue
    }
    if {$row_idx<$one_tag_row} {
      break
    }
    set one_tag_len [lindex $one_tag_info 2]
    set one_tag_col_start [lindex $one_tag_pos_info 1]
    set one_tag_col_end [expr {$one_tag_col_start + $one_tag_len}]
    if {($column_idx<$one_tag_col_start) || ($column_idx>$one_tag_col_end)} {
      continue
    }
    return $tag_idx
  }

  return false
}

proc get_tag_link_index {row_idx column_idx} {
  global tags_links_list
  global tags_links_state_clean

  if {!$tags_links_state_clean} {
    set tags_links_list [extract_tag_list true]
  }

  # searching for a link to a tag at the given place
  set tag_idx -1
  foreach one_tag_info $tags_links_list {
    incr tag_idx
    set one_tag_pos_info [split [lindex $one_tag_info 1] "."]
    set one_tag_row [lindex $one_tag_pos_info 0]
    if {$row_idx>$one_tag_row} {
      continue
    }
    if {$row_idx<$one_tag_row} {
      break
    }
    set one_tag_len [lindex $one_tag_info 2]
    set one_tag_col_start [lindex $one_tag_pos_info 1]
    set one_tag_col_end [expr {$one_tag_col_start + $one_tag_len}]
    if {($column_idx<$one_tag_col_start) || ($column_idx>$one_tag_col_end)} {
      continue
    }
    return $tag_idx
  }

  return false
}

proc target_within_tag_entry {form pos_idx} {
  global text_field
  global tags_decls_id
  global tags_decls_list
  global tags_decls_state_clean
  global tags_links_id
  global tags_links_list
  global tags_links_state_clean

  if {[is_stack_mode]} {
    return false
  }
  if {[catch {
    set pos_idx [$text_field index $pos_idx]
  }]} {
    return false
  }

  if {$form=="decls"} {
    if {$tags_decls_state_clean} {
      set use_list $tags_decls_list
    } {
      set use_list [extract_tag_list false]
      set tags_decls_list $use_list
      set tags_decls_state_clean true
    }
    set with_links false
  } elseif {$form=="links"} {
    if {$tags_links_state_clean} {
      set use_list $tags_links_list
    } {
      set use_list [extract_tag_list true]
      set tags_links_list $use_list
      set tags_links_state_clean true
    }
    set use_list $tags_links_list
    set with_links true
  } {
    return false
  }

  set pos_info [split $pos_idx "."]
  set row_idx [lindex $pos_info 0]
  set col_idx [lindex $pos_info 1]

  if {$with_links} {
    set tag_found [get_tag_link_index $row_idx $col_idx]
    if {$tag_found!=false} {
      set tags_links_id $tag_found
      slide_tags_links 0
      return true
    }
  } {
    set tag_found [get_tag_decl_index $row_idx $col_idx]
    if {$tag_found!=false} {
      set tags_decls_id $tag_found
      slide_tags_declared 0
      return true
    }
  }

  return false
}

proc target_within_tag_entries_on_click {x_abs y_abs x_pos y_pos} {
  global text_field

  set curr_index [$text_field index "@$x_pos,$y_pos"]
  target_within_tag_entries $curr_index
}
proc target_within_tag_entries_on_key {} {
  global text_field

  set curr_index [$text_field index "insert"]
  target_within_tag_entries $curr_index
}

proc target_within_tag_entries {pos_idx} {
  global tags_decls_id
  global tags_decls_list
  global tags_decls_state_clean
  global tags_links_id
  global tags_links_list
  global tags_links_state_clean

  if {![is_batch_mode]} {
    return false
  }

  if {!$tags_decls_state_clean} {
    set tags_decls_list [extract_tag_list false]
    set tags_decls_state_clean true
  }
  if {!$tags_links_state_clean} {
    set tags_links_list [extract_tag_list true]
    set tags_links_state_clean true
  }

  # here if between a declared tag and a link to a tag, locating both of them;
  set targeted_decls [target_within_tag_entry "decls" $pos_idx]
  set targeted_links [target_within_tag_entry "links" $pos_idx]

  return [expr {$targeted_decls || $targeted_links}]
}

# due to increased complexity, better to do actual tag-searches in separate functions
proc find_tag_pair_in_file {search_all_tags \
  tag_first_type tag_first_type_lim tag_first_value tag_first_value_lim \
  tag_second_type tag_second_type_lim tag_second_value tag_second_value_lim \
  in_current_file {file_name ""}} {
  # here meant a close tag pair; search for non-close tag pair is done via two calls for single-tag search
  global text_field
  global tag_border_size

  if {(!$in_current_file) && ($file_name=="")} {
    return [list false "empty file name"]
  }

  # the way here is to do concurrent search for both tag1 and tag2,
  # during that keeping info on the outer-most found tag1 (that has the largest scope covered)
  # and on all the (yet not-reset) tag2 occurrences, since can not know where (if at all) tag1 occurs then;
  # if tag1 is found, it is reset on the-same-or-greater section (incl. computation blocks);
  # tag2 locations are added to the current tag1 if tag1 is found, or reset on any section,
  # and if located at a computation block, put at within-section list on leaving its block;

  # we need the tag completed for [@...] borders if searched full,
  # and to put it e.g. lowercase in both cases;

  set tag_first_type [string tolower $tag_first_type]
  set tag_first_value [string tolower $tag_first_value]
  set first_with_second_part true
  set tag_first_lim true
  if {(!$tag_first_type_lim) && (!$tag_first_value_lim)} {
    set tag_first [strcat [string repeat "\[" $tag_border_size] "#" $tag_first_type ":" $tag_first_value [string repeat "\]" $tag_border_size]]
    set first_with_second_part false
    set tag_first_lim false
  } {
    set tag_first $tag_first_type
    if {!$tag_first_type_lim} {
      append tag_first ":" $tag_first_value
      set first_with_second_part false
    } {
      if {($tag_first_value=="") && $tag_first_value_lim} {
        set first_with_second_part false
      }
    }
  }

  set tag_first_len [string length $tag_first]
  if {$first_with_second_part} {
    set tag_first_sec $tag_first_value
    if {!$tag_first_value_lim} {
      append tag_first_sec [string repeat "\]" $tag_border_size]
    }
    set tag_first_sec_len [string length $tag_first_sec]
  }

  set tag_second_type [string tolower $tag_second_type]
  set tag_second_value [string tolower $tag_second_value]
  set second_with_second_part true
  set tag_second_lim true
  if {(!$tag_second_type_lim) && (!$tag_second_value_lim)} {
    set tag_second [strcat [string repeat "\[" $tag_border_size] "#" $tag_second_type ":" $tag_second_value [string repeat "\]" $tag_border_size]]
    set second_with_second_part false
    set tag_second_lim false
  } {
    set tag_second $tag_second_type
    if {!$tag_second_type_lim} {
      append tag_second ":" $tag_second_value
      set second_with_second_part false
    } {
      if {($tag_second_value=="") && $tag_second_value_lim} {
        set second_with_second_part false
      }
    }
  }

  set tag_second_len [string length $tag_second]
  if {$second_with_second_part} {
    set tag_second_sec $tag_second_value
    if {!$tag_second_value_lim} {
      append tag_second_sec [string repeat "\]" $tag_border_size]
    }
    set tag_second_sec_len [string length $tag_second_sec]
  }

  array set sec_weights [list "#" 3 "=" 2 "*" 1 "~" 0]

  set last_row 1000000000
  set file_hnd ""
  if {!$in_current_file} {
    if {[catch {
      set file_hnd [open $file_name "r"]
    }]} {
      return [list false "can not open file"]
    }
  } {
    set last_row [lindex [split [$text_field index "end -1 chars"] "."] 0]
  }

  set tag_decl_regexp [get_tag_regexp false]
  set sec_line_regexp [get_sec_regexp]
  set block_line_regexp [get_til_regexp]

  # if initial last_section_line_weight is set to:
  # -1 then tags above sections do not chain with the in-sections tags,
  #  4 then tags above sections do chain with all the in-sections tags;
  # ...not sure which case is more reasonable;
  set last_section_line_weight 4
  set last_section_line_len 0
  set last_within_code_block false

  set having_first_tag false
  set tag2_within_section [list]
  set tag2_within_code_block [list]
  set found_positions [list]
  set within_code_block false

  set line_idx 0
  while {true} {
    incr line_idx
    # both checking current-text end, and not dealing with effectively infinite outer files
    if {$line_idx>$last_row} {
      break
    }
    # outer-file end checking
    if {!$in_current_file} {
      if {[catch {
        set at_file_end [eof $file_hnd]
      }]} {
        set at_file_end true
      }
      if {$at_file_end} {
        catch {close $file_hnd}
        set file_hnd ""
        break
      }
    }

    if {$in_current_file} {
      set curr_line [string tolower [$text_field get "$line_idx.0" "$line_idx.end"]]
    } {
      if {[catch {
        set curr_line [string tolower [gets $file_hnd]]
      }]} {
        catch {close $file_hnd}
        set file_hnd ""
        return [list false "can not read from file"]
      }
    }
    if {$curr_line==""} {
      continue
    }

    if {[regexp $sec_line_regexp $curr_line]} {
      # setting only if not at second-tag search or if current section line with greater/same weight
      set test_section_line_type [string range $curr_line 0 0]
      set test_section_line_weight $sec_weights($test_section_line_type)
      set test_section_line_len [string length $curr_line]
      if {(!$having_first_tag) || $last_within_code_block || ($test_section_line_weight>$last_section_line_weight) || \
        (($test_section_line_weight==$last_section_line_weight) && ($test_section_line_len>=$last_section_line_len))} {
        set having_first_tag false
        set last_within_code_block false
        set last_section_line_weight $test_section_line_weight
        set last_section_line_len $test_section_line_len
        set tag2_within_section [list]
        set tag2_within_code_block [list]
      }
      set within_code_block false
      continue
    } elseif {[regexp $block_line_regexp $curr_line]} {
      if {$within_code_block} {
        # if the first tag was at computation block, it is reset at its end
        if {$having_first_tag} {
          if {$last_within_code_block} {
            set having_first_tag false
            set tag2_within_section [list]
            set tag2_within_code_block [list]
          }
        } {
          set tag2_within_section [concat $tag2_within_section $tag2_within_code_block]
          set tag2_within_code_block [list]
        }
        set within_code_block false
      } {
        set within_code_block true
      }
      continue
    }

    # if first_tag_found, searching for the second only;
    # otherwise have to search for both of them, since the first one can be located later in a section;

    set reg_for_tag1 false
    if {!$having_first_tag} {
      if {!$tag_first_lim} {
        if {[string first $tag_first $curr_line]>-1} {
          set having_first_tag true
          if {!$within_code_block} {
            set found_positions [concat $found_positions $tag2_within_section]
          } {
            set found_positions [concat $found_positions $tag2_within_code_block]
          }
          if {(!$search_all_tags) && ([llength $found_positions]>0)} {
            return [list true $found_positions]
          }
          set tag2_within_section [list]
          set tag2_within_code_block [list]
        }
      } {
        set reg_for_tag1 true
      }
    }

    set reg_for_tag2 true
    if {!$tag_second_lim} {
      set reg_for_tag2 false
      set tag2_offset 0
      while {true} {
        set tag2_col [string first $tag_second [string range $curr_line $tag2_offset "end"]]
        if {$tag2_col==-1} {
          break
        }
        incr tag2_offset $tag2_col
        set cur_pos [list $line_idx $tag2_offset $tag_second_len]
        if {$having_first_tag} {
          lappend found_positions $cur_pos
          if {!$search_all_tags} {
            return [list true $found_positions]
          }
        } {
          if {$within_code_block} {
            lappend tag2_within_code_block $cur_pos
          } {
            lappend tag2_within_section $cur_pos
          }
        }
        incr tag2_offset $tag_second_len
      }
    }

    if {$reg_for_tag1 || $reg_for_tag2} {
      set start_idx 0
      set var_idx ""
      while {true} {
        set has_found [regexp -start $start_idx -indices $tag_decl_regexp $curr_line var_idx]
        if {$has_found==0} {
          break
        }
        set start_idx [lindex $var_idx 1]
        set offset_len [lindex $var_idx 0]
        set cmp_idx [expr {$offset_len + $tag_border_size + 1}]

        if {$reg_for_tag1} {
          if {[string equal -nocase -length $tag_first_len $tag_first [string range $curr_line $cmp_idx "end"]]} {
            set sec_part_ok true
            if {$first_with_second_part} {
              # since at a tag, $tag_value_start should be within it
              set tag_value_start [string first ":" $curr_line $offset_len]
              incr tag_value_start
              if {![string equal -nocase -length $tag_first_sec_len $tag_first_sec [string range $curr_line $tag_value_start "end"]]} {
                set sec_part_ok false
              }
            }
            if {$sec_part_ok} {
              set having_first_tag true
              if {(!$search_all_tags) && ([llength $found_positions]>0)} {
                return [list true $found_positions]
              }
              # setting reg_for_tag1 to false, since got found (and not needing it more times),
              # and if (!$reg_for_tag2) doing break, since then nothing to do more;
              set reg_for_tag1 false
              if {!$reg_for_tag2} {
                break
              }
            }
          }
        }
        if {$reg_for_tag2} {
          if {[string equal -nocase -length $tag_second_len $tag_second [string range $curr_line $cmp_idx "end"]]} {
            set sec_part_ok true
            if {$second_with_second_part} {
              # since at a tag, $tag_value_start should be within it
              set tag_value_start [string first ":" $curr_line $offset_len]
              incr tag_value_start
              if {![string equal -nocase -length $tag_second_sec_len $tag_second_sec [string range $curr_line $tag_value_start "end"]]} {
                set sec_part_ok false
              }
            }
            if {$sec_part_ok} {
              set cur_pos [list $line_idx $offset_len [expr 1 + $start_idx - $offset_len]]
              if {$having_first_tag} {
                lappend found_positions $cur_pos
                if {!$search_all_tags} {
                  return [list true $found_positions]
                }
              } {
                if {$within_code_block} {
                  lappend tag2_within_code_block $cur_pos
                } {
                  lappend tag2_within_section $cur_pos
                }
              }
            }
          }
        }
      }
    }

  }

  if {$file_hnd!=""} {
    catch {close $file_hnd}
  }

  return [list true $found_positions]

}

# due to increased complexity, better to do actual tag-searches in separate functions
proc find_tag_single_in_file {search_all_tags tag_spec_type tag_spec_type_lim tag_spec_value tag_spec_value_lim \
  in_current_file {file_name ""}} {
  global text_field
  global tag_border_size

  if {(!$in_current_file) && ($file_name=="")} {
    return [list false "empty file name"]
  }

  # we need the tag completed for [@...] borders if searched full,
  # and to put it e.g. lowercase in both cases;
  set tag_spec_type [string tolower $tag_spec_type]
  set tag_spec_value [string tolower $tag_spec_value]
  set with_second_part true
  set tag_spec_lim true
  if {(!$tag_spec_type_lim) && (!$tag_spec_value_lim)} {
    set tag_spec [strcat [string repeat "\[" $tag_border_size] "#" $tag_spec_type ":" $tag_spec_value [string repeat "\]" $tag_border_size]]
    set with_second_part false
    set tag_spec_lim false
  } {
    set tag_spec $tag_spec_type
    if {!$tag_spec_type_lim} {
      append tag_spec ":" $tag_spec_value
      set with_second_part false
    } {
      if {($tag_spec_value=="") && $tag_spec_value_lim} {
        set with_second_part false
      }
    }
  }

  set tag_spec_len [string length $tag_spec]
  if {$with_second_part} {
    set tag_spec_sec $tag_spec_value
    if {!$tag_spec_value_lim} {
      append tag_spec_sec [string repeat "\]" $tag_border_size]
    }
    set tag_spec_sec_len [string length $tag_spec_sec]
  }

  set last_row 1000000000
  set file_hnd ""
  if {!$in_current_file} {
    if {[catch {
      set file_hnd [open $file_name "r"]
    }]} {
      return [list false "can not open file"]
    }
  } {
    set last_row [lindex [split [$text_field index "end -1 chars"] "."] 0]
  }

  set tag_decl_regexp [get_tag_regexp false]

  set found_positions [list]

  set within_code_block false

  set line_idx 0
  while {true} {
    incr line_idx
    # both checking current-text end, and not dealing with effectively infinite outer files
    if {$line_idx>$last_row} {
      break
    }
    # outer-file end checking
    if {!$in_current_file} {
      if {[catch {
        set at_file_end [eof $file_hnd]
      }]} {
        set at_file_end true
      }
      if {$at_file_end} {
        catch {close $file_hnd}
        set file_hnd ""
        break
      }
    }

    if {$in_current_file} {
      set curr_line [string tolower [$text_field get "$line_idx.0" "$line_idx.end"]]
    } {
      if {[catch {
        set curr_line [string tolower [gets $file_hnd]]
      }]} {
        catch {close $file_hnd}
        set file_hnd ""
        return [list false "can not read from file"]
      }
    }
    if {$curr_line==""} {
      continue
    }

    if {!$tag_spec_lim} {

      set tag_offset 0
      while {true} {
        set tag_col [string first $tag_spec [string range $curr_line $tag_offset "end"]]
        if {$tag_col==-1} {
          break
        }
        incr tag_offset $tag_col
        set cur_pos [list $line_idx $tag_offset $tag_spec_len]
        lappend found_positions $cur_pos
        if {!$search_all_tags} {
          return [list true $found_positions]
        }
        incr tag_offset $tag_spec_len

      }
    } {

      set start_idx 0
      set var_idx ""
      while {true} {
        set has_found [regexp -start $start_idx -indices $tag_decl_regexp $curr_line var_idx]
        if {$has_found==0} {
          break
        }
        set start_idx [lindex $var_idx 1]
        set offset_len [lindex $var_idx 0]
        set cmp_idx [expr {$offset_len + $tag_border_size + 1}]
        if {[string equal -nocase -length $tag_spec_len $tag_spec [string range $curr_line $cmp_idx "end"]]} {
          if {$with_second_part} {
            # since at a tag, $tag_value_start should be within it
            set tag_value_start [string first ":" $curr_line $offset_len]
            incr tag_value_start
            if {![string equal -nocase -length $tag_spec_sec_len $tag_spec_sec [string range $curr_line $tag_value_start "end"]]} {
              continue
            }
          }
          set cur_pos [list $line_idx $offset_len [expr 1 + $start_idx - $offset_len]]
          lappend found_positions $cur_pos
          if {!$search_all_tags} {
            return [list true $found_positions]
          }
        }
      }
    }

  }

  if {$file_hnd!=""} {
    catch {close $file_hnd}
  }

  return [list true $found_positions]

}

proc on_tag_link_target_within_file {tags_search {start_idx "1.0"} {forwards true} {link_info ""}} {
  global text_field
  global search_tag_all
  global tags_decls_id
  global tags_decls_list
  global tags_decls_state_clean
  global tags_links_id
  global tags_links_list
  global tags_links_state_clean
  global tag_border_size
  global show_tags_found_count_info
  global show_tags_found_with_link

  if {$text_field==""} {
    return
  }

  set tags_info [get_tag1_tag2_props $tags_search]
  if {![lindex $tags_info 0]} {
    show_text_field_message_error [list [list [lindex $tags_info 1]]]
    return
  }
  set has_two_tags [lindex $tags_info 1]

  set tag_first_info [lindex $tags_info 2]
  set tag_first_type [lindex $tag_first_info 0]
  set tag_first_type_lim [lindex $tag_first_info 1]
  set tag_first_value [lindex $tag_first_info 2]
  set tag_first_value_lim [lindex $tag_first_info 3]

  set tag_second_info [lindex $tags_info 3]
  set tag_second_type [lindex $tag_second_info 0]
  set tag_second_type_lim [lindex $tag_second_info 1]
  set tag_second_value [lindex $tag_second_info 2]
  set tag_second_value_lim [lindex $tag_second_info 3]

  set tags_close [lindex $tags_info 4]

  set tags_found [list false]
  if {$has_two_tags} {
    if {$tags_close} {
      set tags_found [find_tag_pair_in_file true \
        $tag_first_type $tag_first_type_lim $tag_first_value $tag_first_value_lim \
        $tag_second_type $tag_second_type_lim $tag_second_value $tag_second_value_lim \
        true]
    } {
      set tags_found [find_tag_single_in_file false \
        $tag_first_type $tag_first_type_lim $tag_first_value $tag_first_value_lim \
        true]
      if {[lindex $tags_found 0] && ([llength [lindex $tags_found 1]]>0)} {
        set tags_found [find_tag_single_in_file true \
          $tag_second_type $tag_second_type_lim $tag_second_value $tag_second_value_lim \
          true]
      }
    }
  } {
    set tags_found [find_tag_single_in_file true \
      $tag_first_type $tag_first_type_lim $tag_first_value $tag_first_value_lim \
      true]
  }

  if {![lindex $tags_found 0]} {
    set err_message "Can not search tag"
    set err_expl [lindex $tags_found 1]
    if {$err_expl!=""} {
      append err_message ": $err_expl"
    }
    show_text_field_message_error [list [list $err_message]]
    return
  }

  set tags_found [lindex $tags_found 1]
  set tag_count [llength $tags_found]

  if {$tag_count==0} {
    if {[is_batch_mode]} {
      target_within_tag_entries [$text_field index "insert"]
    }
    show_text_field_message_error [list [list "Linked tag was not found within the document"]]
    return
  }

  set with_tag_link false
  set tag_link_set false
  set tag_link_row ""
  set tag_link_col ""
  set tag_link_start ""
  set tag_link_len ""
  if {$show_tags_found_with_link && ($tag_count>0) && ($link_info != "")} {
    set with_tag_link true
    set tag_link_pos [lindex $link_info 1]
    set tag_link_pos_info [split $tag_link_pos "."]
    set tag_link_row [lindex $tag_link_pos_info 0]
    set tag_link_col [lindex $tag_link_pos_info 1]
    set tag_link_start [strcat $tag_link_pos " +$tag_border_size chars"]
    set tag_link_len_out [lindex $link_info 2]
    set tag_link_len [expr {$tag_link_len_out-(2*$tag_border_size)}]
    if {$forwards} {
      set start_idx [strcat $tag_link_pos " +" $tag_link_len_out " chars"]
    } {
      set start_idx $tag_link_pos
    }
  }

  set tags_found_start [list]
  set tags_found_len [list]
  foreach one_tag_info $tags_found {
    if {$with_tag_link && (!$tag_link_set)} {
      if {([lindex $one_tag_info 0]>$tag_link_row) || (([lindex $one_tag_info 0]==$tag_link_row) && ([lindex $one_tag_info 1]>$tag_link_col))} {
        set tag_link_set true
        lappend tags_found_start $tag_link_start
        lappend tags_found_len $tag_link_len
      }
    }
    lappend tags_found_start [strcat [lindex $one_tag_info 0] "." [lindex $one_tag_info 1] " +$tag_border_size chars"]
    lappend tags_found_len [expr {[lindex $one_tag_info 2]-(2*$tag_border_size)}]
  }
  if {$with_tag_link && (!$tag_link_set)} {
    set tag_link_set true
    lappend tags_found_start $tag_link_start
    lappend tags_found_len $tag_link_len
  }

  set got_focused false
  # gonna reuse the search/seek functions;
  unset_search_highlights $text_field $search_tag_all
  $text_field tag remove "sel" "1.0" "end"
  if {$tag_count>0} {
    set_search_highlights $text_field $search_tag_all $tags_found_start $tags_found_len
    # have to locate one of the found tag locations
    $text_field mark set "insert" $start_idx
    # seeks should do targets too (if at batch mode)
    if {$forwards} {
      seek_next
    } {
      seek_prev
    }
    if {($tag_count>1) && $show_tags_found_count_info} {
      set got_focused true
      show_text_field_message_error [list [list "Found $tag_count tag locations within the document"]]
    }
  } {
    # when here the searched (inner) tag was not empty
    set got_focused true
    show_text_field_message_error [list [list "Linked tag was not found within the document"]]
  }

  # this should be done within seeks,
  # though if nothing found but got remote loading,
  # we have to do the locating outside seeks;
  # remote loading should only be done when something is in,
  # though it should not hurt it to call the locating here;
  if {[is_batch_mode]} {
    target_within_tag_entries [$text_field index "insert"]
  }

  $text_field see [$text_field index "insert"]
  update_text_highlight_lines_onset
  set_anchor_insert $text_field
  if {!$got_focused} {
    focus $text_field
  }
  return
}

proc on_tag_link_slip_into_file {idx} {
  global otl_last_found_linked_files
  global otl_last_used_inner_tag
  global text_field
  global tag_border_size

  if {($idx==-1)||($idx=="none")||([llength $otl_last_found_linked_files]<=$idx)} {
    return false
  }

  remove_otl_paths_menu_shown

  set linked_path_info [lindex $otl_last_found_linked_files $idx]
  set linked_path [lindex $linked_path_info 0]
  set linked_id_row [lindex $linked_path_info 1]
  set linked_id_col [lindex $linked_path_info 2]

  update_recent_index
  if {![load_from_file $linked_path]} {
    return true
  }

  # here just the (previously) first found tag seeked
  set set_position [join [list $linked_id_row $linked_id_col] "."]
  append set_position " +$tag_border_size chars"
  $text_field tag remove "sel" "1.0" "end"
  $text_field mark set "insert" $set_position
  $text_field see $set_position

  set tags_to_search [lindex $otl_last_used_inner_tag 0]

  on_tag_link_target_within_file $tags_to_search
}

proc on_tag_link_try_to_jump_key {with_control with_shift} {
  global text_field

  if {$text_field==""} {
    return false
  }

  set ins_info [split [$text_field index "insert"] "."]
  set use_row [lindex $ins_info 0]
  set use_col [lindex $ins_info 1]

  set pos_list [$text_field bbox "insert"]
  if {[llength $pos_list]==0} {
    set pos_list [list 0 0]
  }

  set x_pos [lindex $pos_list 0]
  set y_pos [lindex $pos_list 1]

  set abs_pos [get_text_field_position]
  set x_abs [expr {$x_pos + [lindex $abs_pos 0]}]
  set y_abs [expr {$y_pos + [lindex $abs_pos 1]}]

  return [on_tag_link_try_to_jump $use_row $use_col $x_abs $y_abs $x_pos $y_pos $with_control $with_shift]
}

# returns if correct:
#   [true / is_remote / tag_parts (incl. is_lim) / dir_part / glob_part / up_heights / sub_depths]
# or if an error occurs:
#   [false "reason"]
proc tag_link_split_into_dir_glob_tag {tag_link_spec} {
  global tag_def_lc
  global tag_def_uc
  global default_tag_link_height
  global default_tag_link_depth

  if {[string range $tag_link_spec 0 0]=="@"} {
    # considering both cases with "@" still present and already removed
    set tag_link_spec [string range $tag_link_spec 1 "end"]
  }

  if {$tag_link_spec==""} {
    return [list false "empty link"]
  }

  set cmp_endings [list $tag_def_lc $tag_def_uc]

  set dir_glob_part ""
  set dir_part ""
  set glob_part ""

  set aux_backslash_pos [string first "\\" $tag_link_spec]
  if {$aux_backslash_pos > 0} {
    # testing dir-glob presence if there is any place for that at first
    set dir_glob_part [string range $tag_link_spec 0 [expr {$aux_backslash_pos - 1}]]
    set aux_colon_pos [string first ":" $dir_glob_part]
    set aux_slash_pos [string first "/" $dir_glob_part]
    set tmp_is_dir_glob false
    if {$aux_colon_pos < 1} {
      # if without colon (common dir) or colon as the first char (for up-dirs height),
      # it means it should be dir-glob part;
      set tmp_is_dir_glob true
    } elseif {($aux_slash_pos > -1) && ($aux_slash_pos < $aux_colon_pos)} {
      # if there is slash, and it is before the first colon, it should be dir-glob part
      # (it can be for device-spec e.g. "/z:" and/or sub-dir depth spec e.g. ":3")
      set tmp_is_dir_glob true
    }
    if {$tmp_is_dir_glob} {
      set tag_link_spec [string trim [string range $tag_link_spec $aux_backslash_pos "end"] "\\"]
    } {
      set dir_glob_part ""
    }
  }

  set tag_groups [list]
  set curr_tag_group [list]
  foreach one_tag_part [split $tag_link_spec "\\"] {
    if {$one_tag_part==""} {
      if {[llength $curr_tag_group]>0} {
        lappend tag_groups $curr_tag_group
        set curr_tag_group [list]
      }
      continue
    }
    set aux_tag_split_pos [string first ":" $one_tag_part]
    if {$aux_tag_split_pos < 1} {
      return [list false "no tag type in"]
    }
    # getting tag type vs. tag value separately, since both of them can be limited
    set one_tag_part_type [string range $one_tag_part 0 [expr {$aux_tag_split_pos - 1}]]
    set one_tag_part_value [string range $one_tag_part [expr {$aux_tag_split_pos + 1}] "end"]
    set aux_last_type_char [string range $one_tag_part_type "end" "end"]
    set one_tag_part_type_is_lim false
    if {$aux_last_type_char == "*"} {
      set one_tag_part_type [string range $one_tag_part_type 0 "end-1"]
      set one_tag_part_type_is_lim true
    }
    set one_tag_part_value_is_lim true
    set aux_last_char [string range $one_tag_part_value "end" "end"]
    if {$aux_last_char in $cmp_endings} {
      set one_tag_part_value_is_lim [expr {($aux_last_char==$tag_def_lc) ? true : false}]
      set one_tag_part_value [string range $one_tag_part_value 0 "end-1"]
    }
    lappend curr_tag_group [list $one_tag_part_type $one_tag_part_type_is_lim $one_tag_part_value $one_tag_part_value_is_lim]
  }
  if {[llength $curr_tag_group]>0} {
    lappend tag_groups $curr_tag_group
    set curr_tag_group [list]
  }

  if {[llength $tag_groups]==0} {
    return [list false "no tag specs"]
  }

  set has_up_heights false
  set up_heights_part ""
  set has_sub_depths false
  set sub_depths_part ""

  if {$dir_glob_part != ""} {
    set aux_slash_pos [string last "/" $dir_glob_part]
    if {$aux_slash_pos==-1} {
      return [list false "dir without slash"]
    }
    set dir_part [string range $dir_glob_part 0 $aux_slash_pos]
    set glob_part [string range $dir_glob_part [expr {$aux_slash_pos + 1}] "end"]

    if {[string range $dir_part 0 0]==":"} {
      set has_up_heights true
      # since having a slash found by "last", there must be found one by "first" too;
      set aux_slash_pos [string first "/" $dir_part]
      set up_heights_part [string range $dir_part 1 [expr {$aux_slash_pos - 1}]]
      set dir_part [string range $dir_part $aux_slash_pos "end"]
      if {$up_heights_part==""} {
        set up_heights_part $default_tag_link_height
      } {
        if {(![string is digit -strict $up_heights_part]) || (![string is ascii -strict $up_heights_part])} {
          return [list false "unknown up-dir height form"]
        }
        set up_heights_part [string trimleft $up_heights_part "0"]
        if {$up_heights_part==""} {
          set up_heights_part 0
        }
      }
    }

    set aux_colon_pos [string last ":" $glob_part]
    if {$aux_colon_pos > -1} {
      set has_sub_depths true
      set sub_depths_part [string range $glob_part [expr {$aux_colon_pos + 1}] "end"]
      set glob_part [string range $glob_part 0 [expr {$aux_colon_pos - 1}]]
      if {$sub_depths_part==""} {
        set sub_depths_part $default_tag_link_depth
      } {
        if {(![string is digit -strict $sub_depths_part]) || (![string is ascii -strict $sub_depths_part])} {
          return [list false "unknown sub-dir depth form"]
        }
        set sub_depths_part [string trimleft $sub_depths_part "0"]
        if {$sub_depths_part==""} {
          set sub_depths_part 0
        }
      }
    }

    if {!([string range $glob_part "end" "end"] in $cmp_endings)} {
      append glob_part $tag_def_lc
    }
  }

  return [list true [expr {($dir_part=="") ? false : true}] $tag_groups $dir_part $glob_part \
    [list $has_up_heights $up_heights_part] [list $has_sub_depths $sub_depths_part]]

}

set otl_last_found_linked_files [list]
set otl_last_used_inner_tag [list]
proc on_tag_link_try_to_jump {use_row use_col x_abs y_abs x_pos y_pos with_control with_shift} {
  global otl_last_found_linked_files
  global otl_last_used_inner_tag
  global tags_links_state_clean
  global tags_decls_state_clean
  global text_field
  global tags_links_id
  global tags_decls_id
  global tags_links_list
  global tags_decls_list
  global save_path
  global tag_border_size
  global default_scan_minima_only

  set otl_last_found_linked_files [list]
  set otl_last_used_inner_tag [list]

  if {!$tags_links_state_clean} {
    set tags_links_list [extract_tag_list true]
  }

  # searching for a link to a tag at the given place
  set on_link_info [list]
  set on_link_id [get_tag_link_index $use_row $use_col]

  if {$on_link_id!=false} {
    set on_link_info [lindex $tags_links_list $on_link_id]
    set on_link_start [lindex $on_link_info 1]
    # if between two (links to) tags, the former gets its outermost bracket highligted,
    # thus we consider that one (link to a) tag to be the active (reached) one;
    if {!$tags_decls_state_clean} {
      set tags_decls_list [extract_tag_list false]
    }
    set on_decl_id [get_tag_decl_index $use_row $use_col]
    if {$on_decl_id!=false} {
      set on_decl_start [lindex [lindex $tags_decls_list $on_decl_id] 1]
      if {$on_decl_start<$on_link_start} {
        set on_link_start false
      }
    }
  }
  if {$on_link_id==false} {
    return false
  }
  set tags_links_id $on_link_id
  slide_tags_links 0

  set one_tag_link_string [lindex $on_link_info 0]
  set one_tag_link_inner [string range $one_tag_link_string [expr {1+$tag_border_size}] "end-$tag_border_size"]

  set link_parts [tag_link_split_into_dir_glob_tag $one_tag_link_inner]
  if {[lindex $link_parts 0]==false} {
    set curr_err_message "Could not parse tag link"
    if {[lindex $link_parts 1]!=""} {
      append curr_err_message ": " [lindex $link_parts 1]
    }
    show_text_field_message_error [list [list $curr_err_message]]
    return false
  }
  set is_remote [lindex $link_parts 1]
  set tag_parts [lindex $link_parts 2]

  if {!$is_remote} {
    # local search
    on_tag_link_target_within_file $tag_parts [join [list $use_row $use_col] "."] [expr {!$with_shift}] $on_link_info
    return true
  }

  set remote_dir_part [lindex $link_parts 3]
  set remote_glob_part [lindex $link_parts 4]
  set up_heights [lindex $link_parts 5]
  set sub_depths [lindex $link_parts 6]

  set mnm_only [expr {$with_shift ? false : true}]
  if {!$default_scan_minima_only} {
    set mnm_only [expr {$mnm_only ? false : true}]
  }

  # should ask the user on non-saved content before making the search,
  # since the search may take some time, thus not to do it again then;
  set to_load "ok"
  set message_ask "Discard changes?"
  set message_detail "Session not saved"
  if {$save_path!=""} {
    set message_detail "Session not resaved"
  }
  if {[$text_field edit modified]} {
    set to_load [tk_messageBox -message $message_ask -icon question -type okcancel -default cancel -detail $message_detail]
  }
  if {$to_load=="cancel"} {
    return true
  }

  set found_path_info [on_tag_link_try_to_find_paths \
    $up_heights $remote_dir_part $remote_glob_part $sub_depths $tag_parts $mnm_only]

  if {![lindex $found_path_info 0]} {
    set err_mssg "Can not search files"
    if {[lindex $found_path_info 1]!=""} {
      append err_msg [lindex $found_path_info 1]
    }
    show_text_field_message_error [list [list $err_mssg]]
    return true
  }

  set found_path_list [lindex $found_path_info 1]
  if {[llength $found_path_list]==0} {
    show_text_field_message_error [list [list "No file with requested tag found"]]
    return true
  }

  set otl_last_found_linked_files $found_path_list
  # prepare info for a possible attempt to find inner location;
  set otl_last_used_inner_tag [list $tag_parts]

  # if more than one path found, have to provide list with them;
  # this should be alike it is at the "recent open files" case
  # (though possibly with sublist-sliding for this);
  if {[llength $found_path_list]==1} {
    on_tag_link_slip_into_file 0
  } {
    on_tag_link_offer_found_paths $x_abs $y_abs $x_pos $y_pos
  }
  return true
}

proc on_tag_link_update_menu_entries {forwards step} {
  global otl_last_found_linked_files
  global otl_max_size_of_paths_menu
  global otl_paths_menu_offset
  global otl_paths_menu_shown
  global otl_paths_menu_label_start
  global save_path
  global menu_selected_background

  if {$otl_paths_menu_shown==""} {
    return
  }

  set use_list $otl_last_found_linked_files
  set use_list_len [llength $use_list]
  if {$use_list_len==0} {
    return
  }

  if {$forwards} {
    set offset_idx $otl_paths_menu_offset
    if {$offset_idx>=$use_list_len} {
      set offset_idx 0
    }
  } {
    set offset_idx [expr {$otl_paths_menu_offset - (2*$step)}]
    if {$offset_idx<0} {
      set offset_idx [expr {$use_list_len - ($use_list_len % $otl_max_size_of_paths_menu)}]
    }
  }

  set label_text $otl_paths_menu_label_start
  append label_text "[expr {1 + $offset_idx}] - [expr {min($use_list_len, $otl_max_size_of_paths_menu + $offset_idx)}]"
  append label_text " / $use_list_len"
  $otl_paths_menu_shown entryconfigure 0 -label $label_text

  set save_path_norm [file normalize $save_path]
  set curr_idx 0
  set list_idx $offset_idx
  while {$list_idx<$use_list_len} {
    set one_path_info [lindex $use_list $list_idx]
    set show_path [lindex $one_path_info 0]
    set show_highlight false
    if {[file normalize $show_path]==$save_path_norm} {
      set show_highlight true
    }
    if {[string length $show_path] > 55} {
      set show_path_tmp [string range $show_path "0" "15"]
      append show_path_tmp "..."
      append show_path_tmp [string range $show_path "end-35" "end"]
      set show_path $show_path_tmp
    }
    incr curr_idx
    $otl_paths_menu_shown entryconfigure $curr_idx -label $show_path -command [list "on_tag_link_slip_into_file" $list_idx]
    $otl_paths_menu_shown entryconfigure $curr_idx -background "#ffffff" -state "normal"
    incr list_idx
    if {$show_highlight} {
      $otl_paths_menu_shown entryconfigure $curr_idx -background $menu_selected_background -activebackground "#e0e0ff"
    }
    if {$curr_idx==$otl_max_size_of_paths_menu} {
      break
    }
  }

  while {$curr_idx<$otl_max_size_of_paths_menu} {
    incr curr_idx
    $otl_paths_menu_shown entryconfigure $curr_idx -label "   " -state disabled -background "#f0f0f0" -command [list "on_tag_link_slip_into_file" "none"]
  }

  set otl_paths_menu_offset [expr {$offset_idx + $curr_idx}]
}

proc remove_otl_paths_menu_shown {} {
  global otl_paths_menu_shown_tears
  global otl_paths_menu_shown_torn
  global otl_paths_menu_shown

  if {!$otl_paths_menu_shown_tears} {
    return
  }
  set otl_paths_menu_shown_tears false

  if {$otl_paths_menu_shown_torn!=""} {
    catch {destroy $otl_paths_menu_shown_torn}
    set otl_paths_menu_shown_torn ""
  }
  if {$otl_paths_menu_shown!=""} {
    catch {destroy $otl_paths_menu_shown}
    set otl_paths_menu_shown ""
  }

  catch {focus $text_field}
}

set otl_max_size_of_paths_menu [get_max_menu_entry_count]
set otl_paths_menu_offset 0
set otl_paths_menu_shown ""
set otl_paths_menu_label_start ""
set otl_paths_menu_shown_torn ""
set otl_paths_menu_shown_tears false
proc on_tag_link_offer_found_paths {x_abs y_abs x_pos y_pos} {
  global topw
  global otl_last_found_linked_files
  global otl_max_size_of_paths_menu
  global otl_paths_menu_offset
  global otl_paths_menu_shown
  global otl_paths_menu_label_start
  global save_path
  global otl_paths_menu_shown_tears
  global otl_paths_menu_shown_torn
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  set use_list $otl_last_found_linked_files
  set use_list_len [llength $use_list]
  if {$use_list_len==0} {
    return
  }

  if {($x_abs=="") || ($y_abs=="")} {
    set abs_pos [get_text_field_position]
    set use_x [expr {$x_pos + [lindex $abs_pos 0] + $default_menu_offset_x}]
    set use_y [expr {$y_pos + [lindex $abs_pos 1] + $default_menu_offset_y}]
  } {
    set use_x [expr {$x_abs + $default_menu_offset_x}]
    set use_y [expr {$y_abs + $default_menu_offset_y}]
  }

  if {$otl_paths_menu_shown_torn!=""} {
    catch {destroy $otl_paths_menu_shown_torn}
    set otl_paths_menu_shown_torn ""
  }
  catch {destroy $otl_paths_menu_shown}
  set otl_paths_menu_shown [menu "$topw.otlpathsmenushown" -tearoff false]
  $otl_paths_menu_shown configure -relief solid -borderwidth 1 -background "#f0f0f0" -disabledforeground "#000000"
  $otl_paths_menu_shown configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f8ffc0"
  $otl_paths_menu_shown configure -relief "solid" -borderwidth 1 -activeborderwidth 1

  set shown_by_parts false
  set otl_paths_menu_shown_tears false
  if {$use_list_len>$otl_max_size_of_paths_menu} {
    set shown_by_parts true
    set otl_paths_menu_shown_tears true
  }
  if {[is_linux_like]} {
    set otl_paths_menu_shown_tears false
  }

  if {$shown_by_parts} {
    set backwards_list [list "2" "Control-Left" "Control-P" "Left"]
    foreach keyspressed $backwards_list {
      bind $otl_paths_menu_shown <$keyspressed> {
        catch {
          on_tag_link_update_menu_entries false $otl_max_size_of_paths_menu
        }
        break
      }
    }
    set forwards_list [list "3" "Control-Right" "Control-N" "Right"]
    foreach keyspressed $forwards_list {
      bind $otl_paths_menu_shown <$keyspressed> {
        catch {
          on_tag_link_update_menu_entries true $otl_max_size_of_paths_menu
        }
        break
      }
    }
  }

  set offset_idx 0
  set label_text "Found files, "
  if {$shown_by_parts} {
    append label_text "shown part: "
    set otl_paths_menu_label_start $label_text
    append label_text "[expr {1 + $offset_idx}] - [expr {min($use_list_len, $otl_max_size_of_paths_menu + $offset_idx)}]"
    append label_text " / $use_list_len"
  } {
    append label_text "all shown: "
    append label_text "1 - $use_list_len"
  }
  $otl_paths_menu_shown add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  set save_path_norm [file normalize $save_path]
  set curr_idx 0
  set list_idx $offset_idx
  while {$list_idx<$use_list_len} {
    set one_path_info [lindex $use_list $list_idx]
    set show_path [lindex $one_path_info 0]
    set show_highlight false
    if {[file normalize $show_path]==$save_path_norm} {
      set show_highlight true
    }
    if {[string length $show_path] > 55} {
      set show_path_tmp [string range $show_path "0" "15"]
      append show_path_tmp "..."
      append show_path_tmp [string range $show_path "end-35" "end"]
      set show_path $show_path_tmp
    }
    $otl_paths_menu_shown add command -label $show_path -command [list "on_tag_link_slip_into_file" $list_idx]
    incr curr_idx
    incr list_idx
    $otl_paths_menu_shown entryconfigure $curr_idx -background "#ffffff"
    if {$show_highlight} {
      $otl_paths_menu_shown entryconfigure $curr_idx -background $menu_selected_background -activebackground "#e0e0ff"
    }
    if {$curr_idx==$otl_max_size_of_paths_menu} {
      break
    }
  }

  # to finish a column if it is at an offset part;
  # this case should not happen, since only using one-column menus here;
  if {$shown_by_parts} {
    while {$curr_idx<$otl_max_size_of_paths_menu} {
      $otl_paths_menu_shown add command -label "   " -state disabled -background "#f0f0f0" -command [list "on_tag_link_slip_into_file" "none"]
      incr curr_idx
    }
  }

  set otl_paths_menu_offset [expr {$offset_idx + $curr_idx}]

  set current_menu $otl_paths_menu_shown
  set pos_vert [expr {$use_y - ([$current_menu yposition [$current_menu index last]] + ([is_linux_like] ? 20 : 2))}]
  set pos_vert_min [expr {24 + [winfo vrooty $current_menu]}]
  if {$pos_vert < $pos_vert_min} {
    set pos_vert $pos_vert_min
  }

  if {$otl_paths_menu_shown_tears} {
    $otl_paths_menu_shown configure -title "Found files"
    set otl_paths_menu_shown_torn [::tk::TearOffMenu $otl_paths_menu_shown $use_x $pos_vert]
    catch {wm resizable $otl_paths_menu_shown false false}
    catch {wm resizable $otl_paths_menu_shown_torn false false}
    catch {focus $otl_paths_menu_shown_torn}
    catch {
      bind $otl_paths_menu_shown_torn <Escape> {
        remove_otl_paths_menu_shown
        break
      }
    }
  } {
    tk_popup $otl_paths_menu_shown $use_x $use_y
  }
}

set text_sel_actions_menu ""
proc offer_sel_text_actions {x_abs y_abs x_pos y_pos curr_index} {
  global topw
  global text_field
  global text_sel_actions_menu
  global default_menu_offset_x
  global default_menu_offset_y
  global spch_in_work

  $text_field mark set "insert" $curr_index
  update_text_highlight_lines
  update

  set line_num [lindex [split $curr_index "."] 0]

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  catch {destroy $text_sel_actions_menu}
  set text_sel_actions_menu [menu "$topw.text_sel_actions_menu" -tearoff false]
  $text_sel_actions_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $text_sel_actions_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $text_sel_actions_menu configure -activeborderwidth 1

  if {$spch_in_work} {
    $text_sel_actions_menu add command -label "spell checker is working" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
  } {
    $text_sel_actions_menu add command -label "spell-checking:" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
    $text_sel_actions_menu add command -label "restrictive on concatenated words" \
      -command [list "spell_check_selected_text" false false] \
      -background "#fafaff" -activebackground "#eaeafc"
    $text_sel_actions_menu add command -label "permissive for concatenated words" \
      -command [list "spell_check_selected_text" true false] \
      -background "#fafaff" -activebackground "#eaeafc"
    $text_sel_actions_menu add command -label "check ascii only (restrictive on concatenated words)" \
      -command [list "spell_check_selected_text" false true] \
      -background "#fafaff" -activebackground "#eaeafc"
    $text_sel_actions_menu add command -label "check ascii only (permissive for concatenated words)" \
      -command [list "spell_check_selected_text" true true] \
      -background "#fafaff" -activebackground "#eaeafc"
  }

  tk_popup $text_sel_actions_menu $use_x $use_y
}

proc make_line_insertion {line_num line_length line_char} {
  global text_field
  global last_insertion_request_x
  global last_insertion_request_y

  append line_insertion [string repeat $line_char $line_length] "\n"

  $text_field edit separator
  $text_field insert "$line_num.0" $line_insertion
  $text_field edit separator

  update
  place_cursor_on_mouse $text_field  "" "" $last_insertion_request_x $last_insertion_request_y
  update_text_properties
  update_text_whole_tag
}

set line_insertion_lengths [list]
proc prepare_line_insertion_lengths {} {
  global min_other_count
  global line_insertion_lengths

  foreach multiple [list 8 4 2 1] {
    lappend line_insertion_lengths [expr {$multiple*$min_other_count}]
  }
}
prepare_line_insertion_lengths

set line_insertion_chars [list "*" "=" "#"]
set line_insertion_chars_index 0
set text_line_insertion_menu ""
set last_insertion_request_x ""
set last_insertion_request_y ""

set menu_sign_times "x"
catch {set menu_sign_times [encoding convertfrom "utf-8" "\xC3\x97"]}
proc offer_line_insertion {x_abs y_abs x_pos y_pos curr_index} {
  global topw
  global text_field
  global text_line_insertion_menu
  global line_insertion_chars
  global line_insertion_chars_index
  global line_insertion_lengths
  global last_insertion_request_x
  global last_insertion_request_y
  global menu_sign_times
  global min_tilde_count
  global default_menu_offset_x
  global default_menu_offset_y

  set last_insertion_request_x $x_pos
  set last_insertion_request_y $y_pos

  $text_field mark set "insert" $curr_index
  update_text_highlight_lines
  update

  set line_num [lindex [split $curr_index "."] 0]

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  catch {destroy $text_line_insertion_menu}
  set text_line_insertion_menu [menu "$topw.text_line_insertion_menu" -tearoff false]
  $text_line_insertion_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $text_line_insertion_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $text_line_insertion_menu configure -activeborderwidth 1

  $text_line_insertion_menu add command -label "computation-block line:" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
  $text_line_insertion_menu add command -label "~-line ($min_tilde_count $menu_sign_times ~)" \
    -command [list "make_line_insertion" $line_num $min_tilde_count "~"] \
    -background "#fafaff" -activebackground "#eaeafc"
  $text_line_insertion_menu add command -label "slide: Control+Alt+p/n" -state "disabled" -background "#e0f0e0" -activebackground "#e0f0e0"

  $text_line_insertion_menu add command -label "section lines:" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  # "#f8ffc0" is the common menu_active_background_common color
  set bglist [list \
    [list "#ffffff" "#f8ffc0"] \
    [list "#fafaf0" "#f0f7b8"] \
    [list "#f2f2e0" "#e8efb0"] \
  ]

  set use_char_count [llength $line_insertion_chars]
  set use_char_idx 0
  while {$use_char_idx<$use_char_count} {
    set use_char [lindex $line_insertion_chars $use_char_idx]
    set bg_set [lindex $bglist $use_char_idx]
    incr use_char_idx 1
    foreach line_length $line_insertion_lengths {
      $text_line_insertion_menu add command -label "$use_char-line ($line_length $menu_sign_times $use_char)" \
        -command [list "make_line_insertion" $line_num $line_length $use_char] \
        -background [lindex $bg_set 0] -activebackground [lindex $bg_set 1]
    }
  }

  $text_line_insertion_menu add command -label "slide: Control+P/N" -state "disabled" -background "#e0f0e0" -activebackground "#e0f0e0"

  tk_popup $text_line_insertion_menu $use_x $use_y
}

proc make_offered_tag_insertion {tag_offered} {
  global tags_insertion_index
  global text_field
  global tags_insertion_menu_tears
  global tags_insertion_menu
  global tags_insertion_menu_torn

  if {$tags_insertion_menu_tears} {
    remove_tags_insertion_menu
  }

  if {![is_text_field_active]} {
    return
  }
  if {$tag_offered==""} {
    return
  }
  clear_temporary_highlights

  $text_field edit separator
  $text_field insert $tags_insertion_index $tag_offered
  $text_field edit separator
  focus $text_field
  update_text_properties
  update_text_whole_tag
}

set menu_active_background_common "#f8ffc0"

set menu_sign_backwards "<-"
set menu_sign_forwards "->"
catch {set menu_sign_backwards [encoding convertfrom "utf-8" "\xE2\x86\x90"]}
catch {set menu_sign_forwards [encoding convertfrom "utf-8" "\xE2\x86\x92"]}
proc offer_tag_insertion_update {forwards} {
  global tags_decls_list
  global tags_decls_state_clean
  global tags_insertion_menu
  global tags_insertion_menu_offset
  global menu_sign_backwards
  global menu_sign_forwards
  global tags_insertion_menu_tears
  global tags_insertion_menu_max_count
  global menu_active_background_common
  global tag_border_size

  if {!$tags_decls_state_clean} {
    set tags_decls_list [extract_tag_list false]
    set tags_decls_state_clean true
  }

  set tag_count [llength $tags_decls_list]
  if {$tag_count==0} {
    # there should be plentiful tags to come to this paging
    return
  }
  set set_count $tags_insertion_menu_max_count
  if {$set_count<2} {
    # this should not occur
    return
  }

  if {$forwards} {
    incr tags_insertion_menu_offset $set_count
    if {$tags_insertion_menu_offset>=$tag_count} {
      set tags_insertion_menu_offset 0
    }
  } {
    incr tags_insertion_menu_offset -$set_count
    if {$tags_insertion_menu_offset<0} {
      set tags_insertion_menu_offset [expr {$tag_count - ((($tag_count-1)%$set_count)+1)}]
    }
  }

  set link_rank_first [expr {$tags_insertion_menu_offset+1}]
  set link_rank_last [expr {min($tag_count,$tags_insertion_menu_offset+$set_count)}]
  set label_text "links to local tags: $link_rank_first - $link_rank_last / $tag_count"
  append label_text "    $menu_sign_backwards $menu_sign_forwards"

  $tags_insertion_menu entryconfigure 0 -state "disabled" -label "$label_text"

  set count_idx 0
  set tag_idx 0
  set link_idx $tags_insertion_menu_offset
  while {$count_idx<$set_count} {
    incr count_idx
    if {[expr {$count_idx % 2}]==0} {
      set use_background "#f0f0f0"
    } {
      set use_background "#ffffff"
    }
    set state_use true
    if {$tag_count>$link_idx} {
      set one_offer [string replace [lindex [lindex $tags_decls_list $link_idx] 0] $tag_border_size $tag_border_size "@"]
      set one_label "$one_offer"
      incr link_idx
    } {
      set state_use false
    }
    incr tag_idx
    if {$state_use} {
      $tags_insertion_menu entryconfigure $tag_idx -state "normal" -label $one_label \
        -background $use_background -activebackground $menu_active_background_common \
        -command [list "make_offered_tag_insertion" $one_offer]
    } {
      $tags_insertion_menu entryconfigure $tag_idx -state "disabled" -label "" \
        -background "#e8e8e8" -activebackground "#e8e8e8" \
        -command [list "make_offered_tag_insertion" ""]
    }
  }
}

proc remove_tags_insertion_menu {} {
  global tags_insertion_menu_tears
  global tags_insertion_menu_torn
  global tags_insertion_menu

  if {!$tags_insertion_menu_tears} {
    return
  }
  set tags_insertion_menu_tears false

  if {$tags_insertion_menu_torn!=""} {
    catch {destroy $tags_insertion_menu_torn}
    set tags_insertion_menu_torn ""
  }
  if {$tags_insertion_menu!=""} {
    catch {destroy $tags_insertion_menu}
    set tags_insertion_menu ""
  }

  catch {focus $text_field}
}

set tags_insertion_menu ""
set tags_insertion_menu_max_count [get_max_menu_entry_count]
set tags_insertion_index ""
set tags_insertion_menu_tears false
set tags_insertion_menu_torn ""
proc offer_tag_insertion {x_pos y_pos curr_index} {
  global topw
  global text_field
  global tags_insertion_menu
  global tags_insertion_menu_torn
  global tags_insertion_index
  global last_insertion_request_x
  global last_insertion_request_y
  global tags_decls_list
  global tags_decls_state_clean
  global tags_insertion_menu_offset
  global menu_sign_backwards
  global menu_sign_forwards
  global tags_insertion_menu_tears
  global tags_insertion_menu_max_count
  global menu_active_background_common
  global tag_border_size

  if {!$tags_decls_state_clean} {
    set tags_decls_list [extract_tag_list false]
    set tags_decls_state_clean true
  }
  set tag_count [llength $tags_decls_list]
  if {$tag_count==0} {
    return
  }

  set tags_insertion_menu_tears false
  set by_parts false
  if {$tag_count>$tags_insertion_menu_max_count} {
    set by_parts true
    set tags_insertion_menu_tears true
  }
  if {[is_linux_like]} {
    set tags_insertion_menu_tears false
  }

  set tags_insertion_menu_offset 0
  set tags_insertion_index $curr_index

  set last_insertion_request_x $x_pos
  set last_insertion_request_y $y_pos

  $text_field mark set "insert" $curr_index
  update_text_highlight_lines
  update

  set use_x $x_pos
  set use_y $y_pos

  set abs_pos [get_text_field_position]
  set use_x [expr {$use_x + [lindex $abs_pos 0]}]
  set use_y [expr {$use_y + [lindex $abs_pos 1]}]

  if {$tags_insertion_menu!=""} {
    catch {destroy $tags_insertion_menu}
  }
  if {$tags_insertion_menu_torn!=""} {
    catch {destroy $tags_insertion_menu_torn}
  }
  set tags_insertion_menu_torn ""
  set tags_insertion_menu [menu "$topw.tags_insertion_menu" -tearoff false]
  $tags_insertion_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $tags_insertion_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground $menu_active_background_common
  $tags_insertion_menu configure -activeborderwidth 1
  $tags_insertion_menu configure -title "links to tags"

  if {$by_parts} {
    foreach one_key [list "less" "comma" "Left"] {
      bind $tags_insertion_menu <$one_key> {
        catch {
          offer_tag_insertion_update false
        }
        break
      }
    }
    foreach one_key [list "greater" "period" "Right"] {
      bind $tags_insertion_menu <$one_key> {
        catch {
          offer_tag_insertion_update true
        }
        break
      }
    }
  }

  set tag_use_count [expr {$by_parts ? $tags_insertion_menu_max_count : $tag_count}]
  set label_text "links to local tags: 1 - $tag_use_count / $tag_count"
  if {$by_parts} {
    append label_text "    $menu_sign_backwards $menu_sign_forwards"
  }
  $tags_insertion_menu add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  set link_idx 0
  while {$link_idx<$tag_use_count} {
    set one_offer [string replace [lindex [lindex $tags_decls_list $link_idx] 0] $tag_border_size $tag_border_size "@"]
    set one_label "$one_offer"
    incr link_idx
    if {[expr {$link_idx % 2}]==0} {
      set use_background "#f0f0f0"
    } {
      set use_background "#ffffff"
    }
    $tags_insertion_menu add command -label $one_label -command [list "make_offered_tag_insertion" $one_offer] \
       -background $use_background
  }

  set current_menu $tags_insertion_menu
  set pos_vert [expr {$use_y - ([$current_menu yposition [$current_menu index last]] + ([is_linux_like] ? 20 : 2))}]
  set pos_vert_min [expr {24 + [winfo vrooty $current_menu]}]
  if {$pos_vert < $pos_vert_min} {
    set pos_vert $pos_vert_min
  }
  set pos_horz [expr {$use_x + 10}]

  if {$tags_insertion_menu_tears} {
    set tags_insertion_menu_torn [::tk::TearOffMenu $tags_insertion_menu $pos_horz $pos_vert]
    catch {wm resizable $tags_insertion_menu false false}
    catch {wm resizable $tags_insertion_menu_torn false false}
    catch {focus $tags_insertion_menu_torn}
    catch {
      bind $tags_insertion_menu_torn <Escape> {
        remove_tags_insertion_menu
        break
      }
    }
  } {
    tk_popup $tags_insertion_menu $pos_horz $pos_vert
  }
}

set tags_insertion_offers_declared [list \
  "\[#:\]" "\[#exp:\]" "\[#sec:\]" "\[#top:\]" "\[#doc:\]"]
set tags_insertion_offers_local_links [list \
  "\[@exp:\]" "\[@sec:\]" "\[@top:\]"]
set tags_insertion_offers_remote_links [list \
  "\[@./\\exp:\]" "\[@./\\sec:\]" "\[@./\\doc:\]" \
  "\[@./\\doc:\\\\sec:\]" "\[@./\\doc:\\\\exp:\]"]
proc offer_tag_template_insertion {x_abs y_abs x_pos y_pos curr_index} {
  global topw
  global text_field
  global tags_insertion_menu
  global tags_insertion_menu_torn
  global tags_insertion_offers_declared
  global tags_insertion_offers_local_links
  global tags_insertion_offers_remote_links
  global tags_insertion_index
  global last_insertion_request_x
  global last_insertion_request_y
  global tags_insertion_menu_tears
  global default_menu_offset_x
  global default_menu_offset_y

  set tags_insertion_index $curr_index

  set tags_insertion_menu_tears false
  if {$tags_insertion_menu_torn!=""} {
    catch {destroy $tags_insertion_menu_torn}
    set tags_insertion_menu_torn ""
  }

  set last_insertion_request_x $x_pos
  set last_insertion_request_y $y_pos

  $text_field mark set "insert" $curr_index
  update_text_highlight_lines
  update

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  if {$tags_insertion_menu!=""} {
    catch {destroy $tags_insertion_menu}
  }
  set tags_insertion_menu [menu "$topw.tags_insertion_menu" -tearoff false]
  $tags_insertion_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $tags_insertion_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $tags_insertion_menu configure -activeborderwidth 1
  $tags_insertion_menu configure -title "tag templates"

  set label_text "tag templates:"
  $tags_insertion_menu add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
  foreach one_offer $tags_insertion_offers_declared {
    set one_label "$one_offer"
    $tags_insertion_menu add command -label $one_label -command [list "make_offered_tag_insertion" $one_offer]
  }

  $tags_insertion_menu add command -label "see: Control + \[\{#" -state "disabled" -background "#e0f0e0" -activebackground "#e0f0e0"

  $tags_insertion_menu add command -label "for local links:" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
  foreach one_offer $tags_insertion_offers_local_links {
    set one_label "$one_offer"
    $tags_insertion_menu add command -label $one_label -command [list "make_offered_tag_insertion" $one_offer]
  }

  $tags_insertion_menu add command -label "insert: Control+@" -state "disabled" -background "#e0f0e0" -activebackground "#e0f0e0"

  $tags_insertion_menu add command -label "for remote links:" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
  foreach one_offer $tags_insertion_offers_remote_links {
    set one_label "$one_offer"
    $tags_insertion_menu add command -label $one_label -command [list "make_offered_tag_insertion" $one_offer]
  }

  $tags_insertion_menu add command -label "see: Control + \]\}#" -state "disabled" -background "#e0f0e0" -activebackground "#e0f0e0"

  tk_popup $tags_insertion_menu $use_x $use_y
}

proc set_text_highlight_blocks_mode {mode_idx} {
  global highlight_text_blocks_mode

  set highlight_text_blocks_mode $mode_idx
  setting_text_highlight_blocks 0
}

set text_enter_mode_menu ""
proc offer_text_enter_mode_menu {x_abs y_abs x_pos y_pos} {
  global topw
  global text_field
  global text_enter_mode_menu
  global stack_mode_text_based_command_modes_use_idx
  global stack_mode_text_based_command_modes
  global batch_mode_text_based_command_modes_use_idx
  global batch_mode_text_based_command_modes
  global highlight_text_blocks_mode
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  if {$text_enter_mode_menu!=""} {
    catch {destroy $text_enter_mode_menu}
  }
  set text_enter_mode_menu [menu "$topw.text_enter_mode_menu" -tearoff false]
  $text_enter_mode_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $text_enter_mode_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $text_enter_mode_menu configure -activeborderwidth 1
  $text_enter_mode_menu configure -title "text command mode"

  $text_enter_mode_menu add command -label "Display of computation blocks (started by ~-lines, ended by ~-lines or section lines):" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
  $text_enter_mode_menu add command -label "block sides highlighted" -command [list "set_text_highlight_blocks_mode" 2]
  $text_enter_mode_menu add command -label "~-lines (when not set hidden) highlighted" -command [list "set_text_highlight_blocks_mode" 1]
  $text_enter_mode_menu add command -label "whole blocks highlighted" -command [list "set_text_highlight_blocks_mode" 0]
  incr use_idx [expr {3 - $highlight_text_blocks_mode}]
  $text_enter_mode_menu entryconfigure $use_idx -background $menu_selected_background -activebackground "#e0e0ff"
  set idx_offset 4

  if {[is_stack_mode]} {
    $text_enter_mode_menu add command -label "use Control+f/g/h for search/seek/replace, Control+comma/period for sliding up/down" -state "disabled" -background "#e0f0e0" -activebackground "#e0f0e0"
    incr idx_offset
  }

  set label_text "Control/Shift+Enter action in text:"
  if {[is_batch_mode]} {
    set label_text "Control/Shift+Enter action on selections:"
  }
  $text_enter_mode_menu add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  if {[is_stack_mode]} {
    set current_mode_idx $stack_mode_text_based_command_modes_use_idx
    set available_modes $stack_mode_text_based_command_modes
  } {
    set current_mode_idx $batch_mode_text_based_command_modes_use_idx
    set available_modes $batch_mode_text_based_command_modes
  }

  set cur_idx 0
  foreach one_mode $available_modes {
    set one_label [lindex $one_mode 0]
    set is_used false
    if {$cur_idx==$current_mode_idx} {
      set is_used true
    }
    $text_enter_mode_menu add command -label $one_label -command [list "update_text_enter_mode_menu" $cur_idx]
    incr cur_idx
    if {$is_used} {
      $text_enter_mode_menu entryconfigure [expr {$idx_offset + $cur_idx}] -background $menu_selected_background -activebackground "#e0e0ff"
    }
  }

  $text_enter_mode_menu add command -label "other menus: Control(outside of text)/Shift+Mouse-3, and Control+w (on selection)" -state "disabled" -background "#e0f0e0" -activebackground "#e0f0e0"

  tk_popup $text_enter_mode_menu $use_x $use_y
}

proc is_plot_setup {} {
  global plot_put_window
  return [expr {$plot_put_window!=""}]
}

set text_part_actions_last_content ""
proc copy_text_part_into {to_where copy_mode} {
  global text_part_actions_last_content
  global plot_input_field
  global plot_data_field
  global mxmbatch_input_field

  if {$text_part_actions_last_content==""} {
    return
  }
  if {$copy_mode=="paste"} {
    set do_paste true
    set do_append false
    set do_write false
  } elseif {$copy_mode=="append"} {
    set do_paste false
    set do_append true
    set do_write false
  } elseif {$copy_mode=="write"} {
    set do_paste false
    set do_append false
    set do_write true
  } else {
    set text_part_actions_last_content ""
    return
  }

  set wdg_use ""
  if {[is_batch_mode]} {
    if {$to_where=="batch-input"} {
      if {![is_batch_maxima_set]} {
        setup_batch_maxima
      } {
        if {![is_mxmbatch_input_field_active]} {
          show_text_field_message_error [list [list "Maxima batch input is locked"]]
          set text_part_actions_last_content ""
          return
        }
      }
      draw_batch_maxima
      set wdg_use $mxmbatch_input_field
    }
  }
  if {$to_where=="plot-input"} {
    if {$plot_input_field==""} {
      setup_plot_part
    } {
      if {[$plot_input_field cget -state]!="normal"} {
        show_text_field_message_error [list [list "Gnuplot input is locked"]]
        set text_part_actions_last_content ""
        return
      }
    }
    set wdg_use $plot_input_field
    open_plot_part
    present_gnuplot_input
  }
  if {$to_where=="plot-data"} {
    if {$plot_data_field==""} {
      setup_plot_part
    } {
      if {[$plot_data_field cget -state]!="normal"} {
        show_text_field_message_error [list [list "Gnuplot data is locked"]]
        set text_part_actions_last_content ""
        return
      }
    }
    set wdg_use $plot_data_field
    open_plot_part
    present_gnuplot_data
  }
  if {$wdg_use==""} {
    set text_part_actions_last_content ""
    return
  }

  if {$do_paste} {
    $wdg_use edit separator
    remove_selected_if_by_cursor $wdg_use
    $wdg_use edit separator
    $wdg_use insert "insert" $text_part_actions_last_content
    $wdg_use edit separator
    $wdg_use see "insert"
  } elseif {$do_append} {
    if {[lindex [split [$wdg_use index "end -1 chars"] "."] 1] != 0} {
      $wdg_use edit separator
      $wdg_use insert "end" "\n"
    }
    $wdg_use edit separator
    $wdg_use insert "end" $text_part_actions_last_content
    if {[lindex [split [$wdg_use index "end -1 chars"] "."] 1] != 0} {
      $wdg_use edit separator
      $wdg_use insert "end" "\n"
    }
    $wdg_use edit separator
    $wdg_use see "end"
  } elseif {$do_write} {
    $wdg_use edit separator
    $wdg_use replace "1.0" "end" $text_part_actions_last_content
    if {[lindex [split [$wdg_use index "end -1 chars"] "."] 1] != 0} {
      $wdg_use edit separator
      $wdg_use insert "end" "\n"
    }
    $wdg_use edit separator
    $wdg_use see "insert"
  }

  if {($to_where=="plot-input") || ($to_where=="plot-data")} {
    update_plot_properties
  }
  if {($to_where=="batch-input")} {
    update_batch_maxima_properties
  }

  focus $wdg_use
  set text_part_actions_last_content ""
  return
}

proc offer_text_selection_actions_on_key {} {
  global text_field

  set pos_list [$text_field bbox "insert"]
  if {[llength $pos_list]==0} {
    set pos_list [list 0 0]
  }

  offer_text_selection_actions_menu "" "" [lindex $pos_list 0] [lindex $pos_list 1]
}

# this menu available via <Control-Shift-3> and via <Control-w> at text_field when there is a selection in text;
set text_selection_actions_menu ""
proc offer_text_selection_actions_menu {x_abs y_abs x_pos y_pos} {
  global text_field
  global text_selection_actions_menu
  global topw
  global text_field
  global text_part_actions_last_content
  global default_menu_offset_x
  global default_menu_offset_y

  set text_part_actions_last_content [get_selected_parts $text_field true false]
  if {$text_part_actions_last_content==""} {
    return false
  }

  if {($x_abs=="") || ($y_abs=="")} {
    set abs_pos [get_text_field_position]
    set use_x [expr {$x_pos + [lindex $abs_pos 0] + $default_menu_offset_x}]
    set use_y [expr {$y_pos + [lindex $abs_pos 1] + $default_menu_offset_y}]
  } {
    set use_x [expr {$x_abs + $default_menu_offset_x}]
    set use_y [expr {$y_abs + $default_menu_offset_y}]
  }

  if {$text_selection_actions_menu!=""} {
    catch {destroy $text_selection_actions_menu}
  }
  set text_selection_actions_menu [menu "$topw.text_selection_actions_menu" -tearoff false]
  $text_selection_actions_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $text_selection_actions_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $text_selection_actions_menu configure -activeborderwidth 1
  $text_selection_actions_menu configure -title "text selection actions"

  $text_selection_actions_menu add command -label "To put selection at:" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  if {[is_batch_mode]} {
    if {(![is_batch_maxima_set]) || ([is_mxmbatch_input_field_active])} {
      $text_selection_actions_menu add command -label "batch input, paste into" -command [list "copy_text_part_into" "batch-input" "paste"]
      $text_selection_actions_menu add command -label "batch input, append to" -command [list "copy_text_part_into" "batch-input" "append"]
      $text_selection_actions_menu add command -label "batch input, overwrite" -command [list "copy_text_part_into" "batch-input" "write"]
    }
  } {
    if {[is_command_field_active]} {
      $text_selection_actions_menu add command -label "command, paste into" -command [list "text_to_command_do_paste"]
      $text_selection_actions_menu add command -label "command, append to" -command [list "text_to_command_do_append"]
      $text_selection_actions_menu add command -label "command, overwrite" -command [list "text_to_command_do_overwrite"]
      $text_selection_actions_menu add command -label "command, copy to stack top" -command [list "text_to_command_do_copytotop"]
    }
  }
  $text_selection_actions_menu add command -label "gnuplot input, paste into" -command [list "copy_text_part_into" "plot-input" "paste"]
  $text_selection_actions_menu add command -label "gnuplot input, append to" -command [list "copy_text_part_into" "plot-input" "append"]
  $text_selection_actions_menu add command -label "gnuplot input, overwrite" -command [list "copy_text_part_into" "plot-input" "write"]
  $text_selection_actions_menu add command -label "gnuplot data, paste into" -command [list "copy_text_part_into" "plot-data" "paste"]
  $text_selection_actions_menu add command -label "gnuplot data, append to" -command [list "copy_text_part_into" "plot-data" "append"]
  $text_selection_actions_menu add command -label "gnuplot data, overwrite" -command [list "copy_text_part_into" "plot-data" "write"]

  tk_popup $text_selection_actions_menu $use_x $use_y
}

# currently line-insertion and tag-template insertion (at allowed places) is implemented
proc text_field_on_right_click {x_abs y_abs x_pos y_pos with_control with_shift with_alt} {
  global text_field
  global tags_decls_list
  global tags_decls_state_clean
  global tags_links_list
  global tags_links_state_clean
  global min_tilde_count
  global min_other_count

  if {$with_alt} {
    # this is apparently not available for actions
    return false
  }

  if {($with_control) && ($with_shift)} {
    offer_text_selection_actions_menu $x_abs $y_abs $x_pos $y_pos
    return true
  }

  if {(!$with_control) && (!$with_shift)} {
    offer_text_enter_mode_menu $x_abs $y_abs $x_pos $y_pos
    return true
  }

  if {![is_text_field_active]} {
    return false
  }

  set curr_index [$text_field index "@$x_pos,$y_pos"]
  set curr_index_parts [split $curr_index "."]
  set curr_index_row [lindex $curr_index_parts 0]
  set curr_index_col [lindex $curr_index_parts 1]

  # spell-checking on shift-menu if some text selected, otherwise putting lines in;
  if {(!$with_control)&&($with_shift)} {
    if {[$text_field tag nextrange "sel" "1.0"]!=""} {
      offer_sel_text_actions $x_abs $y_abs $x_pos $y_pos $curr_index
    } {
      offer_line_insertion $x_abs $y_abs $x_pos $y_pos $curr_index
    }
    return true
  }

  # by now without ctrl-actions on selected text
  if {[$text_field tag nextrange "sel" "1.0"]!=""} {
    return false
  }

  # not putting tags onto block/section lines
  set curr_line [$text_field get -- "$curr_index linestart" "$curr_index lineend"]
  set curr_line_len [string length $curr_line]
  if {$curr_line_len>0} {
    foreach one_test_char [list "~" "*" "=" "#"] {
      if {($one_test_char=="~") && ($curr_line_len<$min_tilde_count)} {
        continue
      }
      if {($one_test_char!="~") && ($curr_line_len<$min_other_count)} {
        continue
      }
      if {[string trim $curr_line $one_test_char]==""} {
        return false
      }
    }
  }

  # only offering the tag-related insertions if there is an empty-like line path
  # between the click index and either line start/end or another (link to a) tag;

  set line_empty_like false
  if {[string trim [$text_field get -- "$curr_index linestart" "$curr_index lineend"]]==""} {
    set line_empty_like true
  }

  set can_offer_tag $line_empty_like
  if {!$can_offer_tag} {
    if {$curr_index_col==0} {
      set can_offer_tag true
    }
  }
  if {!$can_offer_tag} {
    set line_end_col [lindex [split [$text_field index "$curr_index lineend"] "."] 1]
    if {$curr_index_col==$line_end_col} {
      set can_offer_tag true
    }
  }
  if {!$can_offer_tag} {
    if {[string trim [$text_field get -- "$curr_index linestart" "$curr_index"]]==""} {
      set can_offer_tag true
    }
  }
  if {!$can_offer_tag} {
    if {[string trim [$text_field get -- "$curr_index" "$curr_index lineend"]]==""} {
      set can_offer_tag true
    }
  }
  if {!$can_offer_tag} {
    if {!$tags_decls_state_clean} {
      set tags_decls_list [extract_tag_list false]
      set tags_decls_state_clean true
    }
    if {!$tags_links_state_clean} {
      set tags_links_list [extract_tag_list true]
      set tags_links_state_clean true
    }
    set all_tags [lsort -dictionary -index 1 [concat $tags_decls_list $tags_links_list]]
    foreach one_tag $all_tags {
      set one_tag_start [lindex $one_tag 1]
      set one_tag_start_parts [split $one_tag_start "."]
      set one_tag_len [lindex $one_tag 2]
      set one_tag_row [lindex $one_tag_start_parts 0]
      set one_tag_col [lindex $one_tag_start_parts 1]
      if {$one_tag_row>$curr_index_row} {
        break
      }
      if {$one_tag_row!=$curr_index_row} {
        continue
      }
      if {$one_tag_col==$curr_index_col} {
        # click right by a tag start
        set can_offer_tag true
        break
      }
      if {$one_tag_col>$curr_index_col} {
        if {[string trim [$text_field get -- "$curr_index" "$one_tag_start"]]==""} {
          # click before a tag, with nothing/spaces in between;
          set can_offer_tag true
        }
        # next tags on the line (if any) are more distant
        break
      }
      if {[expr {$one_tag_col+$one_tag_len}]>$curr_index_col} {
        # click within a tag; not inserting nested tags
        break
      }
      if {[string trim [$text_field get -- "$one_tag_start +$one_tag_len chars" "$curr_index"]]==""} {
        # click after a tag, with nothing/spaces in between;
        set can_offer_tag true
        break
      }
    }
  }

  if {!$can_offer_tag} {
    return false
  }

  if {($with_control) && (!$with_shift)} {
    # offers empty (links to) tags
    offer_tag_template_insertion $x_abs $y_abs $x_pos $y_pos $curr_index
    return true
  }

  # (($with_control) && ($with_shift))
  # ...currently set for cases with a selection;
  # otherwise possibly to offer tag links to declared tags in $minima_file_extension files in the current directory;
  # not sure on possibility of filters here; development of this at some later;
  return false
}

# risky stuff that may be occasionally necessary
proc do_unlock_fields {do_text do_cmmd} {
  global command_field
  global text_field
  global state_label
  global maxima_ready
  global last_button_interrupt_state
  global bt_interrupt
  global should_display_cpu_usage
  global could_display_cpu_usage

  # not sure if everything (incl. buttons) to set as unlocked here;

  if {$do_text && (![is_text_field_active])} {
    set state_text [get_state_label_text]
    set mxm_state $maxima_ready
    set cmmd_state [$command_field cget -state]
    set intr_state [$bt_interrupt cget -state]
    set intr_phase $last_button_interrupt_state
    set_state_ready true
    if {"idle"==$intr_phase} {
      set_button_interrupt_idle
    } elseif {"exec"==$intr_phase} {
      set_button_interrupt_exec
    } elseif {"sets"==$intr_phase} {
      set_button_interrupt_sets
    }
    $bt_interrupt configure -state $intr_state
    $command_field configure -state $cmmd_state
    set maxima_ready $mxm_state

    set should_display_cpu_usage false
    set could_display_cpu_usage true
    set_state_label_text $state_text
    $text_field configure -state "normal"
    set_anchor_insert $text_field
  }
  if {$do_cmmd && (![is_command_field_active])} {
    set intr_state [$bt_interrupt cget -state]
    set_state_ready true
    set_button_interrupt_idle
    $bt_interrupt configure -state $intr_state
    $command_field configure -state "normal"
    set_anchor_insert $command_field
    set should_display_cpu_usage false
    set could_display_cpu_usage true
    set_state_label_text "unlocked"
    set maxima_ready true
  }
}

set cmmd_unlocking_menu ""
proc offer_field_unlocking_via_cmmd {x_abs y_abs x_pos y_pos for_text for_cmmd} {
  global topw
  global cmmd_unlocking_menu
  global default_menu_offset_x
  global default_menu_offset_y

  if {(!$for_text) && (!$for_cmmd)} {
    return
  }

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  if {$cmmd_unlocking_menu!=""} {
    catch {destroy $cmmd_unlocking_menu}
  }
  set cmmd_unlocking_menu [menu "$topw.cmmd_unlocking_menu" -tearoff false]
  $cmmd_unlocking_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $cmmd_unlocking_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $cmmd_unlocking_menu configure -activeborderwidth 1
  $cmmd_unlocking_menu configure -title "unlocking"

  set label_text "unlocking, if TGM Minima has not detected an input to Maxima:"
  $cmmd_unlocking_menu add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  if {$for_text} {
    $cmmd_unlocking_menu add command -label "unlock text (may mistarget answers from Maxima)" -command [list "do_unlock_fields" true false]
  }
  if {$for_cmmd} {
    if {$for_text} {
      $cmmd_unlocking_menu add command -label "unlock both text and command (quite risky!)" -command [list "do_unlock_fields" true true]
    } {
      $cmmd_unlocking_menu add command -label "unlock the command input (quite risky!)" -command [list "do_unlock_fields" false true]
    }
  } {
    $cmmd_unlocking_menu add command -label "without command unlocking: Maxima not available" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
  }

  tk_popup $cmmd_unlocking_menu $use_x $use_y
}

proc set_command_select_mode {do_first do_then} {
  global test_command_select_first
  global test_command_whole_then

  set test_command_select_first $do_first
  set test_command_whole_then $do_then
}

set cmmd_selectmode_menu ""
proc offer_command_select_modes {x_abs y_abs x_pos y_pos} {
  global topw
  global cmmd_selectmode_menu
  global test_command_select_first
  global test_command_whole_then
  global show_cmmd_highlight_line_nums
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  if {$cmmd_selectmode_menu!=""} {
    catch {destroy $cmmd_selectmode_menu}
  }
  set cmmd_selectmode_menu [menu "$topw.cmmd_selectmode_menu" -tearoff false]
  $cmmd_selectmode_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $cmmd_selectmode_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $cmmd_selectmode_menu configure -activeborderwidth 1
  $cmmd_selectmode_menu configure -title "select mode"

  set idx_offset 0

  if {[is_stack_mode]} {
    set label_text "Show command line numbers:"
    $cmmd_selectmode_menu add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
    $cmmd_selectmode_menu add command -label "yes" -command [list "set_cmmd_highlight_lines_mode" true]
    $cmmd_selectmode_menu add command -label "no" -command [list "set_cmmd_highlight_lines_mode" false]
    set use_idx [expr {$show_cmmd_highlight_line_nums ? 1 : 2}]
    $cmmd_selectmode_menu entryconfigure $use_idx -background $menu_selected_background -activebackground "#e0e0ff"

    $cmmd_selectmode_menu add command -label "Control+m/M to (un)set stack level for saving" -state "disabled" -background "#e0f0e0" -activebackground "#e0f0e0"
    set idx_offset 4
  }

  set label_text "Usage of command selections for input to Maxima:"
  $cmmd_selectmode_menu add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  $cmmd_selectmode_menu add command -label "to only use selected regions" -command [list "set_command_select_mode" true false]
  $cmmd_selectmode_menu add command -label "selected regions if any, otherwise the whole command" -command [list "set_command_select_mode" true true]
  $cmmd_selectmode_menu add command -label "to only use the whole command" -command [list "set_command_select_mode" false true]

  set cur_idx [expr {2 + $idx_offset}]
  if {!$test_command_select_first} {
    incr cur_idx 1
  }
  if {!$test_command_whole_then} {
    incr cur_idx -1
  }
  $cmmd_selectmode_menu entryconfigure $cur_idx -background $menu_selected_background -activebackground "#e0e0ff"

  set label_text "Control + Mouse-3 for execution menu"
  $cmmd_selectmode_menu add command -label $label_text -state "disabled" -background "#e0f0e0" -activebackground "#e0f0e0"

  tk_popup $cmmd_selectmode_menu $use_x $use_y
}

set cmmd_maxima_run_menu ""
proc offer_command_maxima_run_action {x_abs y_abs x_pos y_pos} {
  global topw
  global cmmd_maxima_run_menu
  global in_lisp_mode
  global in_debug_mode
  global default_menu_offset_x
  global default_menu_offset_y

  if {[is_stack_mode]} {
    if {![is_command_field_active]} {
      return
    }
  } {
    if {[is_batch_maxima_running] && (![is_batch_maxima_waiting])} {
      return
    }
  }

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  if {$cmmd_maxima_run_menu!=""} {
    catch {destroy $cmmd_maxima_run_menu}
  }
  set cmmd_maxima_run_menu [menu "$topw.cmmd_maxima_run_menu" -tearoff false]
  $cmmd_maxima_run_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $cmmd_maxima_run_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $cmmd_maxima_run_menu configure -activeborderwidth 1
  $cmmd_maxima_run_menu configure -title "command actions"

  set label_text "Commands: Control/Shift + Enter"
  $cmmd_maxima_run_menu add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  if {[is_stack_mode]} {
    set label_text "Shift+Enter: stay at stack level"
    $cmmd_maxima_run_menu add command -label $label_text -command [list "trigger_command_local_wise"]
    set label_text "Control+Enter: move to stack top"
    $cmmd_maxima_run_menu add command -label $label_text -command [list "trigger_command_stack_wise" false]
    set label_text "Control+Shift+Enter: move+copy to stack top"
    $cmmd_maxima_run_menu add command -label $label_text -command [list "trigger_command_stack_wise" true]
    set label_text "Alt+Enter: check bracket matching"
    $cmmd_maxima_run_menu add command -label $label_text -command [list "trigger_command_check"]

    set label_text "Going among stack levels: Control + p/n/l/j"
    $cmmd_maxima_run_menu add command -label $label_text -state "disabled" -background "#e0f0e0" -activebackground "#e0f0e0"
  } {
    set label_text "Shift+Enter: without change at input"
    $cmmd_maxima_run_menu add command -label $label_text -command [list "default_start_batch_maxima" false false]
    set label_text "Control+Enter: input set as \"last\" and cleaned"
    $cmmd_maxima_run_menu add command -label $label_text -command [list "default_start_batch_maxima" true true]
    set label_text "Control+Shift+Enter: input set as \"last\""
    $cmmd_maxima_run_menu add command -label $label_text -command [list "default_start_batch_maxima" true false]
    set label_text "Alt+Enter: check bracket matching"
    $cmmd_maxima_run_menu add command -label $label_text -command [list "test_start_batch_maxima"]

    set label_text "Taking the last set command: Control + l"
    $cmmd_maxima_run_menu add command -label $label_text -state "disabled" -background "#e0f0e0" -activebackground "#e0f0e0"
  }

  if {[is_batch_mode] || ((!$in_lisp_mode) && (!$in_debug_mode))} {
    set label_text "Update of user-defined names for key-completion:"
    $cmmd_maxima_run_menu add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

    if {[is_stack_mode]} {
      set label_text "Control+u: read user-defined functions"
      $cmmd_maxima_run_menu add command -label $label_text -command [list "ask_for_current_names" "functions"]
    } {
      set label_text "Control+u: read user-defined functions"
      $cmmd_maxima_run_menu add command -label $label_text -command [list "names_start_batch_maxima" true false]
      set label_text "Control+Alt-u: (run input and) read user-defined functions"
      $cmmd_maxima_run_menu add command -label $label_text -command [list "names_start_batch_maxima" true true]
    }
    if {[is_stack_mode]} {
      set label_text "Control+U: read user-defined values"
      $cmmd_maxima_run_menu add command -label $label_text -command [list "ask_for_current_names" "values"]
    } {
      set label_text "Control+U: read user-defined values"
      $cmmd_maxima_run_menu add command -label $label_text -command [list "names_start_batch_maxima" false false]
      set label_text "Control+Alt-U: (run input and) read user-defined values"
      $cmmd_maxima_run_menu add command -label $label_text -command [list "names_start_batch_maxima" false false]
    }
  }

  set label_text "Key-completion via (Control/Shift) Tab"
  $cmmd_maxima_run_menu add command -label $label_text -state "disabled" -background "#e0f0e0" -activebackground "#e0f0e0"

  tk_popup $cmmd_maxima_run_menu $use_x $use_y
}

# currently allowing unlocking here (if it makes sense at all) and select-mode setting
proc command_field_on_right_click {x_abs y_abs x_pos y_pos with_control with_shift} {
  global maxima_pid
  global default_menu_offset_x
  global default_menu_offset_y

  if {$with_control || $with_shift} {
    return false
  }

  if {[is_command_field_active]} {
    offer_command_select_modes $x_abs $y_abs $x_pos $y_pos
    return true
  }

  # text_field may get unlocked unless it is already unlocked
  set can_unlock_text [expr {![is_text_field_active]}]

  # locked command_field may get unlocked if is connected to Maxima;
  # allowing early stages to get unlocking too (i.e. not requiring $shibboleth_onset_read);
  set can_unlock_cmmd [expr {(![is_command_field_active]) && [check_connected true] && ($maxima_pid!="")}]

  if {(!$can_unlock_text) && (!$can_unlock_cmmd)} {
    return false
  }

  offer_field_unlocking_via_cmmd $x_abs $y_abs $x_pos $y_pos $can_unlock_text $can_unlock_cmmd
  return true
}

proc insert_tilde_line {x_abs y_abs x_pos y_pos curr_index} {
  global text_field
  global min_tilde_count

  set line_length $min_tilde_count
  $text_field mark set "insert" $curr_index

  set line_num [lindex [split $curr_index "."] 0]
  append line_insertion [string repeat "~" $line_length] "\n"

  $text_field edit separator
  $text_field insert "$line_num.0" $line_insertion
  $text_field edit separator

  update
  set curr_index [$text_field index "@$x_pos,$y_pos"]
  $text_field mark set "insert" $curr_index

  update_text_properties
  return true
}

proc extract_file_path {line index with_slash with_backslash with_colon upwards_too} {
  set retval [list false]

  set path_start 0
  set path_end -1

  # ~ can be used for home-dir
  set common_allowed_chars [list "." "-" "~"]

  set line_chars [split $line {}]
  set line_len [llength $line_chars]
  set can_start false
  set start_char [lindex $line_chars $index]
  if {[string is wordchar $start_char]} {
    set can_start true
  }
  if {$start_char in $common_allowed_chars} {
    set can_start true
  }
  if {$with_slash && ($start_char=="/")} {
    set can_start true
  }
  if {$with_backslash && ($start_char=="\\")} {
    set can_start true
  }
  if {$with_colon && ($start_char==":")} {
    set can_start true
  }

  if {$can_start} {
    set path_start $index
    set path_end $index
    set curr_idx $index
    incr curr_idx
    while {$curr_idx < $line_len} {
      set curr_char [lindex $line_chars $curr_idx]
      if {[string is wordchar $curr_char]} {
        set path_end $curr_idx
        incr curr_idx
        continue
      }
      if {$curr_char in $common_allowed_chars} {
        set path_end $curr_idx
        incr curr_idx
        continue
      }
      if {$with_slash && ($curr_char=="/")} {
        set path_end $curr_idx
        incr curr_idx
        continue
      }
      if {$with_backslash && ($curr_char=="\\")} {
        set path_end $curr_idx
        incr curr_idx
        continue
      }
      if {$with_colon && ($curr_char==":")} {
        set path_end $curr_idx
        incr curr_idx
        continue
      }
      break
    }
    set curr_idx $index
    incr curr_idx -1
    while {$curr_idx >= 0} {
      set curr_char [lindex $line_chars $curr_idx]
      if {[string is wordchar $curr_char]} {
        set path_start $curr_idx
        incr curr_idx -1
        continue
      }
      if {$curr_char in $common_allowed_chars} {
        set path_start $curr_idx
        incr curr_idx -1
        continue
      }
      if {$upwards_too && $with_slash && ($curr_char=="/")} {
        set path_start $curr_idx
        incr curr_idx -1
        continue
      }
      if {$upwards_too && $with_backslash && ($curr_char=="\\")} {
        set path_start $curr_idx
        incr curr_idx -1
        continue
      }
      if {$upwards_too && $with_colon && ($curr_char==":")} {
        set path_start $curr_idx
        incr curr_idx -1
        continue
      }
      break
    }
  }

  if {$path_start<=$path_end} {
    set retval [list true [join [lrange $line_chars $path_start $path_end] ""] $path_start $path_end]
  }

  return $retval
}

proc excerpt_file_path {line index delimiter} {
  set retval [list false]

  set path_start 0
  set path_end -1

  set line_chars [split $line {}]
  set line_len [llength $line_chars]
  set can_start true
  set start_char [lindex $line_chars $index]
  if {$start_char==$delimiter} {
    set can_start false
  }

  set got_delimiter_upwards false
  set got_delimiter_downwards false
  if {$can_start} {
    set path_start $index
    set path_end $index
    set curr_idx $index
    incr curr_idx
    while {$curr_idx < $line_len} {
      set curr_char [lindex $line_chars $curr_idx]
      if {$curr_char!=$delimiter} {
        set path_end $curr_idx
        incr curr_idx
        continue
      }
      set got_delimiter_downwards true
      break
    }
    set curr_idx $index
    incr curr_idx -1
    while {$curr_idx >= 0} {
      set curr_char [lindex $line_chars $curr_idx]
      if {$curr_char!=$delimiter} {
        set path_start $curr_idx
        incr curr_idx -1
        continue
      }
      set got_delimiter_upwards true
      break
    }
  }

  if {$got_delimiter_downwards && $got_delimiter_upwards} {
    set retval [list true [join [lrange $line_chars $path_start $path_end] ""] $path_start $path_end]
  }

  return $retval
}

proc complete_file_path {path} {
  global save_path
  global last_dir
  global last_dir_first

  if {[regexp "^\[/\]?\[a-zA-Z\]\\:/" $path]!=0} {
    if {[string range $path 0 0]=="/"} {
      set path [string range $path 1 "end"]
    }
    if {![is_wine_like]} {
      set path [file join \
        [file normalize [translate_wine_drive [string range $path 0 0]]] \
        [string trimleft [string range $path 2 "end"] "/"] \
      ]
    }
    return [file nativename $path]
  }

  set file_pathtype [file pathtype $path]

  if {$file_pathtype=="absolute"} {
    return [file nativename $path]
  }

  set rel_dir ""
  if {$save_path!=""} {
    set rel_dir [file dirname $save_path]
  } elseif {$last_dir != ""} {
    set rel_dir $last_dir
  } else {
    set rel_dir $last_dir_first
  }

  if {$file_pathtype=="relative"} {
    return [file nativename [file join $rel_dir $path]]
  }

  # guessing on the "volumerelative" path type
  return [file nativename [join [string range $rel_dir 0 1] $path]]

}

proc open_html_path {html_path} {
  set null_path [get_null_path]
  if {[catch {
    if {[tk windowingsystem]=="aqua"} {
      exec "--" "open" $html_path ">$null_path" "2>$null_path" "&"
    } elseif {[is_wine_like]} {
      if {![is_under_wine]} {
        exec "--" "start" $html_path ">$null_path" "2>$null_path" "&"
      } {
        exec "--" "start" $html_path "&"
      }
    } {
      exec "--" "xdg-open" $html_path ">$null_path" "2>$null_path" "&"
    }
  }]} {
    return false
  }
  return true
}

set last_text_double_click_time 0
set last_text_double_click_posx ""
set last_text_double_click_posy ""

# currently just putting-in tildes (on empty lines) is implemented,
# and a rudimentary way for displaying images / html pages, and plotting gnuplot files;
proc text_field_on_double_click {x_abs y_abs x_pos y_pos with_control with_shift} {
  global text_field
  global last_text_double_click_time
  global last_text_double_click_posx
  global last_text_double_click_posy
  global default_menu_offset_x
  global default_menu_offset_y

  set is_spurious false
  set curr_time [clock milliseconds]
  # it seems that the spurious events come with cca 0.2 delay
  # (at least in the testing environment)
  if {[expr {$curr_time - $last_text_double_click_time}]<400} {
    if {[expr {abs($last_text_double_click_posx-$x_pos)}]<=2} {
      if {[expr {abs($last_text_double_click_posy-$y_pos)}]<=2} {
        set is_spurious true
      }
    }
  }
  if {$is_spurious} {
    return false
  }
  # intetionally setting new values on real events only,
  # to let it start to accept clicks eventually if someone
  # does intentional long-time compulsive clicking
  set last_text_double_click_time $curr_time
  set last_text_double_click_posx $x_pos
  set last_text_double_click_posy $y_pos

  set curr_index [$text_field index "@$x_pos,$y_pos"]
  set line_empty_like false
  if {[string trim [$text_field get -- "$curr_index linestart" "$curr_index lineend"]]==""} {
    set line_empty_like true
  }

  # empty/with-spaces-only lines can get ~-lines inserted
  if {$line_empty_like} {
    # to make lesser overlap with control-based actions,
    # ~-lines inserted when without pressed modifiers only
    if {$with_control || $with_shift} {
      return false
    }
    if {[is_text_field_active]} {
      insert_tilde_line $x_abs $y_abs $x_pos $y_pos $curr_index
      set_anchor_position $text_field $curr_index
      return true
    } {
      return false
    }
  }

  # search for paths and tags with held control only
  if {!$with_control} {
    return false
  }

  set null_path [get_null_path]

  set line_content [$text_field get "$curr_index linestart" "$curr_index lineend"]
  set curr_line [lindex [split $curr_index "."] 0]
  set curr_col [lindex [split $curr_index "."] 1]

  set used_path false

  set file_path ""
  set path_exists false
  set path_info [list]

  # first, trying to find file path without quotation marks
  if {true} {
    set with_slash true
    set with_backslash false
    if {[is_wine_like]} {
      set with_backslash true
    }
    set upwards_too false
    if {$with_shift} {
      set upwards_too true
    }
    foreach use_colon [list true false] {
      set path_info [extract_file_path $line_content $curr_col $with_slash $with_backslash $use_colon $upwards_too]
      if {[lindex $path_info 0]} {
        set file_path [complete_file_path [lindex $path_info 1]]
        if {[file exists $file_path]} {
          set path_exists true
          break
        }
        set test_col [expr {[lindex $path_info 2] - 1}]
        if {($test_col>=0) && ([$text_field get [strcat $curr_line "." $test_col]]=="/")} {
          if {[string range [lindex $path_info 1] 0 0]!="/"} {
            set file_path [complete_file_path [strcat "/" [lindex $path_info 1]]]
            if {[file exists $file_path]} {
              lset path_info 2 $test_col
              set path_exists true
              break
            }
          }
        }
      }
    }
  }

  # second, trying to find file path with quotation marks
  if {!$path_exists} {
    set delimiter_first "'"
    set delimiter_second "\""
    if {$with_shift} {
      set delimiter_first "\""
      set delimiter_second "'"
    }
    foreach one_delimiter [list $delimiter_first $delimiter_second] {
      set path_info [excerpt_file_path $line_content $curr_col $one_delimiter]
      if {[lindex $path_info 0]} {
        set file_path [complete_file_path [lindex $path_info 1]]
        if {[file exists $file_path]} {
          set path_exists true
          break
        }
      }
    }
  }

  # if an existing path was found, trying to use it
  if {$path_exists} {
    set path_suffix [string tolower [lindex [split $file_path "."] "end"]]
    if {$path_suffix in [list "png" "gif" "jpg" "jpeg" "tif" "tiff" "svg" "pdf" "ps"]} {
      set used_path true
      if {[catch {
        if {[tk windowingsystem]=="aqua"} {
          exec "--" "open" $file_path ">$null_path" "2>$null_path" "&"
        } elseif {[is_wine_like]} {
          exec "--" "start" $file_path ">$null_path" "2>$null_path" "&"
        } {
          exec "--" "xdg-open" $file_path ">$null_path" "2>$null_path" "&"
        }
      }]} {
        if {[catch {
          exec "--" "display" $file_path ">$null_path" "2>$null_path" "&"
        }]} {
          catch {
            exec "--" "gm" "display" $file_path ">$null_path" "2>$null_path" "&"
          }
        }
      }
    }
    if {$path_suffix in [list "html" "htm"]} {
      set used_path true
      open_html_path $file_path
    }
    if {$path_suffix in [list "gnu" "plot" "plt" "gpl" "gp"]} {
      set used_path true
      catch {
        exec "--" [get_gnuplot_path] $file_path "-e" "bind 'Close' 'unset output; exit gnuplot;'" "-e" "pause mouse close;" ">$null_path" "2>$null_path" "&"
      }
    }
  }

  if {!$used_path} {
    set html_link ""
    foreach one_html_link_info [html_extract_http_links $line_content] {
      set one_html_link_start [lindex $one_html_link_info 0]
      set one_html_link_end [lindex $one_html_link_info 1]
      if {($curr_col >= $one_html_link_start) && ($curr_col <= $one_html_link_end)} {
        set html_link [string range $line_content $one_html_link_start $one_html_link_end]
        if {$html_link != ""} {
          set path_info [list true $html_link $one_html_link_start $one_html_link_end]
          break
        }
      }
    }
    if {$html_link != ""} {
      set used_path true
      open_html_path $html_link
    }
  }

  if {$used_path} {
    set col_start [lindex $path_info 2]
    set col_end [lindex $path_info 3]
    incr col_end
    $text_field tag remove "sel" "1.0" "end"
    set sel_start_idx "$curr_line.$col_start"
    $text_field tag add "sel" $sel_start_idx "$curr_line.$col_end"
    set_anchor_position $text_field $sel_start_idx
    return true
  }

  return [on_tag_link_try_to_jump $curr_line $curr_col $x_abs $y_abs $x_pos $y_pos $with_control $with_shift]
}

proc place_cursor_on_mouse {wdg x_abs y_abs x_pos y_pos} {
  focus $wdg
  $wdg mark set "insert" [$wdg index "@$x_pos,$y_pos"]
}

proc seek_inner_code_onset {} {
  after 0 seek_inner_code
}

proc seek_inner_code {} {
  global text_field
  global text_inner_codes_tag
  global text_inner_codes_start_tag
  global text_inner_codes_end_tag
  global text_inner_codes_start_vis_tag
  global text_inner_codes_end_vis_tag
  global text_inner_star_line_tag
  global text_inner_hash_line_tag
  global text_inner_equa_line_tag
  global highlight_text_blocks_mode
  global text_inner_code_blocks
  global text_inner_line_parts
  global min_tilde_count
  global min_other_count
  global cursor_was_at_block_border

  set insert_line [lindex [split [$text_field index "insert"] "."] 0]
  set ending_line [lindex [split [$text_field index "end - 1 chars"] "."] 0]

  set text_inner_line_parts [list]
  set text_inner_code_blocks [list]

  $text_field tag remove $text_inner_codes_end_vis_tag "1.0" "end"
  $text_field tag remove $text_inner_codes_start_vis_tag "1.0" "end"
  $text_field tag remove $text_inner_codes_end_tag "1.0" "end"
  $text_field tag remove $text_inner_codes_start_tag "1.0" "end"
  $text_field tag remove $text_inner_codes_tag "1.0" "end"

  set cursor_was_at_block_border false

  foreach one_inner_tag [list $text_inner_star_line_tag $text_inner_hash_line_tag $text_inner_equa_line_tag] {
    $text_field tag remove $one_inner_tag "1.0" "end"
  }

  set tilde_lines [$text_field search -all -regexp "^\[~\]\{$min_tilde_count\}\[~\]*\$" "1.0"]

  set star_lines [$text_field search -all -regexp "^\[\\*\]\{$min_other_count\}\[\\*\]*\$" "1.0"]
  set hash_lines [$text_field search -all -regexp "^\[#\]\{$min_other_count\}\[#\]*\$" "1.0"]
  set equals_lines [$text_field search -all -regexp "^\[=\]\{$min_other_count\}\[=\]*\$" "1.0"]

  set lines_considered [list]

  foreach {one_set one_char} [list $tilde_lines "~" $star_lines "*" $hash_lines "#" $equals_lines "="] {
    foreach one_line $one_set {
      set one_start [lindex [split $one_line "."] 0]
      set one_start_plus $one_start
      incr one_start_plus
      lappend lines_considered [list $one_start $one_char]
    }
  }
  set lines_considered [lsort -integer -index 0 $lines_considered]

  set last_tilde_line -10
  set inside_code false
  set code_start ""
  set code_start_plus ""
  set one_line_info_idx 0
  set one_line_info_len [llength $lines_considered]

  # tcl gives empty results on over-list-border index,
  # thus we can use it safely, getting just "" there;
  set next_line_info [lindex $lines_considered $one_line_info_idx]
  set next_line [lindex $next_line_info 0]
  set next_char [lindex $next_line_info 1]
  while {$one_line_info_idx<$one_line_info_len} {
    set one_line $next_line
    set one_char $next_char
    incr one_line_info_idx
    set next_line_info [lindex $lines_considered $one_line_info_idx]
    set next_line [lindex $next_line_info 0]
    set next_char [lindex $next_line_info 1]
    set one_line_plus $one_line
    incr one_line_plus
    set one_line_minus $one_line
    incr one_line_minus -1

    set next_line_far false
    if {$inside_code} {
      $text_field tag add $text_inner_codes_tag "$code_start.end" "$one_line.0"
      set inside_code false
      if {$one_char=="~"} {
        if {$insert_line==$one_line} {
          set cursor_was_at_block_border true
        }
        lappend text_inner_code_blocks [list $code_start $one_line]
        if {($next_line=="") || ($next_line > $one_line_plus) || ($next_char != "~")} {
          set next_line_far true
        }
        if {($last_tilde_line < $one_line_minus) && $next_line_far && ($insert_line!=$one_line) && ($ending_line!=$one_line)} {
          $text_field tag add $text_inner_codes_end_tag "$one_line.0" "$one_line_plus.0"
        } {
          $text_field tag add $text_inner_codes_end_vis_tag "$one_line.0" "$one_line_plus.0"
        }
        set last_tilde_line $one_line
      } {
        lappend text_inner_code_blocks [list $code_start [expr {$one_line-1}]]
      }
    } {
      if {$one_char=="~"} {
        if {$insert_line==$one_line} {
          set cursor_was_at_block_border true
        }
        if {($next_line=="") || ($next_line > $one_line_plus)} {
          set next_line_far true
        }
        if {($last_tilde_line < $one_line_minus) && $next_line_far && ($insert_line!=$one_line) && ($ending_line!=$one_line)} {
          $text_field tag add $text_inner_codes_start_tag "$one_line.0" "$one_line_plus.0"
        } {
          $text_field tag add $text_inner_codes_start_vis_tag "$one_line.0" "$one_line_plus.0"
        }
        set last_tilde_line $one_line
        set code_start $one_line
        set code_start_plus $one_line_plus
        set inside_code true
      }
    }
    if {$one_char=="*"} {
      lappend text_inner_line_parts $one_line
      if {$highlight_text_blocks_mode!=2} {
        $text_field tag add $text_inner_star_line_tag "$one_line.0" "$one_line_plus.0"
      } {
        $text_field tag add $text_inner_star_line_tag "$one_line.0" "$one_line.end"
      }
    }
    if {$one_char=="="} {
      lappend text_inner_line_parts $one_line
      if {$highlight_text_blocks_mode!=2} {
        $text_field tag add $text_inner_equa_line_tag "$one_line.0" "$one_line_plus.0"
      } {
        $text_field tag add $text_inner_equa_line_tag "$one_line.0" "$one_line.end"
      }
    }
    if {$one_char=="#"} {
      lappend text_inner_line_parts $one_line
      if {$highlight_text_blocks_mode!=2} {
        $text_field tag add $text_inner_hash_line_tag "$one_line.0" "$one_line_plus.0"
      } {
        $text_field tag add $text_inner_hash_line_tag "$one_line.0" "$one_line.end"
      }
    }
  }

  if {$inside_code} {
    $text_field tag add $text_inner_codes_tag "$code_start.end" "end"
    lappend text_inner_code_blocks [list $code_start "end"]
  }

  return
}

bind $text_field <<Modified>> {
  text_got_modified
}

# having zombies after gnuplot exiting
# (that got started by pathname-doubleclick);
# running something trivial once-a-second
# (even when it actually does not exists,
# like on mswin platforms)
# should not hurt that much;
# notice that clipboard enforcer (when used)
# runs with "&" alike, thus it needs this
# cleaning process too;
proc try_cleanup_of_zombies_onset {} {
  after idle "try_cleanup_of_zombies"
}
proc try_cleanup_of_zombies {} {
  catch {
    exec true
  }
  after 10000 "try_cleanup_of_zombies_onset"
}
after 10000 try_cleanup_of_zombies_onset

setting_text_highlight_blocks 0

bind $text_field <Control-r> {
  open_plot_part
  break
}
bind $text_field <Control-Alt-r> {
  open_plot_part $text_field false
  break
}
bind $text_field <Control-R> {
  open_plot_part $text_field true
  break
}

bind $command_field <Control-r> {
  open_plot_part
  break
}
bind $command_field <Control-Alt-r> {
  open_plot_part $command_field false
  break
}
bind $command_field <Control-R> {
  open_plot_part $command_field true
  break
}

set html_tmp_path ""
proc set_html_tmp_path {} {
  global html_tmp_path

  if {$html_tmp_path!=""} {
    return
  }

  if {[catch {
    set tmph_info [get_tempfile "html"]
    set tmph [lindex $tmph_info 0]
    set html_tmp_path [lindex $tmph_info 1]
    if {$tmph != ""} {
      close $tmph
    }
  }]} {
    set html_tmp_path ""
  }

  if {$html_tmp_path==""} {
    return
  }

  if {!([string tolower [file extension $html_tmp_path]] in [list "html" "htm"])} {
    set html_tmp_path_new [join [list [file rootname $html_tmp_path] "html"] "."]
    if {![catch {file rename -force -- $html_tmp_path $html_tmp_path_new}]} {
      set html_tmp_path $html_tmp_path_new
    } {
      catch {file delete -- $html_tmp_path}
      set html_tmp_path ""
    }
  }
}

proc trigger_text_export_tmp {with_tmp {with_clip false}} {
  global html_tmp_path

  if {$with_tmp} {
    if {$html_tmp_path==""} {
      set_html_tmp_path
    }
    if {$html_tmp_path==""} {
      show_text_field_message_error [list [list "can not create temporary file"]]
      return false
    }
  }

  set use_path ""
  if {$with_tmp} {
    set use_path $html_tmp_path
  }
  set all_filled [do_text_export_html $with_clip $html_tmp_path]

  if {$html_tmp_path==""} {
    return true
  }
  if {!$all_filled} {
    return false
  }

  return [open_html_path $html_tmp_path]
}

proc export_html_urlize_document_spec {document_spec} {
  set sug_name [tag_suggested_urlize_document_spec $document_spec]
  if {$sug_name!=""} {
    set sug_name [strcat $sug_name ".html"]
  }
  return $sug_name
}

proc tag_suggested_urlize_document_spec {document_spec} {
  set tag_based_id ""
  foreach one_char [split $document_spec {}] {
    if {([string is alnum -strict $one_char]&&[string is ascii -strict $one_char]) || ($one_char in [list "-" "_"])} {
      append tag_based_id [string tolower $one_char]
      continue
    }
    if {[string is wordchar -strict $one_char]} {
      append tag_based_id "-"
      continue
    }
    append tag_based_id "_"
    continue
  }
  return $tag_based_id
}

proc get_tag_html_spec {tag_spec tag_lim} {
  set link_spec "t"

  foreach one_char [split [string tolower $tag_spec] {}] {
    append link_spec [binary encode hex $one_char] "x"
  }

  if {!$tag_lim} {
    append link_spec "t"
  }

  return $link_spec
}

# returns [to_use(boolean) anchor(string) classes(string) data(list)]
proc get_tag_html_decl {tag_decl_spec tag_rank} {
  if {[string range $tag_decl_spec 0 0] in [list "#" "\\"]} {
    set tag_decl_spec [string range $tag_decl_spec 1 "end"]
  }
  set anchor_spec [get_tag_html_spec $tag_decl_spec false]
  set split_pos [string first ":" $tag_decl_spec]
  set tag_type_spec [get_tag_html_spec [string range $tag_decl_spec 0 [expr {$split_pos - 1}]] false]
  set tag_value_spec [get_tag_html_spec [string range $tag_decl_spec [expr {$split_pos + 1}] "end"] false]
  return [list true "$anchor_spec" "tag_local tag_declared" \
    [list "rank" "$tag_rank" "tagtype" "$tag_type_spec" "tagvalue" "$tag_value_spec"]]
}

# returns [to_use(boolean) link(string) classes(string) data(list)]
proc get_tag_html_link {tag_link_spec tag_rank} {
  global tag_type_document_ids
  global tag_def_lc
  global tag_def_uc
  global env

  set link_parts [tag_link_split_into_dir_glob_tag $tag_link_spec]
  if {![lindex $link_parts 0]} {
    return [list false]
  }

  set is_remote [lindex $link_parts 1]
  set tag_parts [lindex $link_parts 2]

  set tags_info [get_tag1_tag2_props $tag_parts]
  if {![lindex $tags_info 0]} {
    return [list false]
  }
  set has_two_tags [lindex $tags_info 1]
  if {$has_two_tags} {
    return [list false]
  }

  set tag_first_info [lindex $tags_info 2]
  set tag_first_type [lindex $tag_first_info 0]
  set tag_first_type_lim [lindex $tag_first_info 1]
  set tag_first_value [lindex $tag_first_info 2]
  set tag_first_value_lim [lindex $tag_first_info 3]

  if {!$is_remote} {
    # first checking if remote URL embedded in (simple) "[@url:...]" tag link
    if {(!$tag_first_type_lim) && ([string tolower $tag_first_type]=="url")} {
      set rem_link_spec_part_link [string trim [string trim $tag_first_value " "] "\""]
      foreach one_test [list "<" ">" "\"" "\\"] {
        if {[string first $one_test $rem_link_spec_part_link]>-1} {
          return [list false]
        }
      }
      if {[string range $rem_link_spec_part_link 0 1]=="~/"} {
        set rem_link_spec_part_link [file normalize $rem_link_spec_part_link]
      } elseif {[regexp "^\[/\]?\[a-zA-Z\]\\:/" $rem_link_spec_part_link]!=0} {
        if {[string range $rem_link_spec_part_link 0 0]=="/"} {
          set rem_link_spec_part_link [string range $rem_link_spec_part_link 1 "end"]
        }
        if {![is_wine_like]} {
          set rem_link_spec_part_link [file join \
            [file normalize [translate_wine_drive [string range $rem_link_spec_part_link 0 0]]] \
            [string trimleft [string range $rem_link_spec_part_link 2 "end"] "/"] \
          ]
        }
      }
      return [list true [string map [list " " "%20"] $rem_link_spec_part_link] "tag_link tag_link_remote" [list "rank" "$tag_rank"]]
    }

    # it may be an embedded image too, via an "[@img:...]" tag link
    if {(!$tag_first_type_lim) && ([string tolower $tag_first_type]=="img")} {
      set rem_link_spec_part_link [string trim [string trim $tag_first_value " "] "\""]
      foreach one_test [list "<" ">" "\"" "\\"] {
        if {[string first $one_test $rem_link_spec_part_link]>-1} {
          return [list false]
        }
      }
      if {[string range $rem_link_spec_part_link 0 1]=="~/"} {
        set rem_link_spec_part_link [file normalize $rem_link_spec_part_link]
      } elseif {[regexp "^\[/\]?\[a-zA-Z\]\\:/" $rem_link_spec_part_link]!=0} {
        if {[string range $rem_link_spec_part_link 0 0]=="/"} {
          set rem_link_spec_part_link [string range $rem_link_spec_part_link 1 "end"]
        }
        if {![is_wine_like]} {
          set rem_link_spec_part_link [file join \
            [file normalize [translate_wine_drive [string range $rem_link_spec_part_link 0 0]]] \
            [string trimleft [string range $rem_link_spec_part_link 2 "end"] "/"] \
          ]
        }
      }
      return [list false "image" [string map [list " " "%20"] $rem_link_spec_part_link] "tag_link tag_link_image" [list "rank" "$tag_rank"]]
    }

    # local link
    set tag_link_type_use [get_tag_html_spec $tag_first_type $tag_first_type_lim]
    set tag_link_value_use [get_tag_html_spec $tag_first_value $tag_first_value_lim]
    set link_spec [get_tag_html_spec [strcat $tag_first_type ":" $tag_first_value] false]
    return [list true "#$link_spec" "tag_local tag_link tag_link_local" \
      [list "rank" "$tag_rank" "tagtype" "$tag_link_type_use" "tagvalue" "$tag_link_value_use"]]
  }

  if {$tag_first_value_lim} {
    return [list false]
  }

  # these below meaningful only if $is_remote==true;
  set remote_dir_part [lindex $link_parts 3]
  set remote_glob_part [lindex $link_parts 4]
  set up_heights [lindex $link_parts 5]
  set sub_depths [lindex $link_parts 6]

  if {[lindex $up_heights 0]} {
    return [list false]
  }
  if {[lindex $sub_depths 0]} {
    return [list false]
  }

  # remote link
  if {!([string tolower $tag_first_type] in $tag_type_document_ids)} {
    # only dealing with links to documents
    return [list false]
  }

  set doc_spec [string tolower $tag_first_value]
  set doc_spec [export_html_urlize_document_spec $doc_spec]

  # allowing all the absolute dirs here even though
  # they will be probably nonsense in web-wise views;
  # user may want to see it locally, without web servers,
  # and then it makes the same sense as in their tag versions;

  set remote_dir_char0 [string range $remote_dir_part 0 0]
  if {$remote_dir_char0} {
    set remote_dir [strcat [file normalize "~"] [string range $remote_dir_part 1 "end"]]
  } {
    # using "/x:" translation alike at within-minima search;
    set remote_dir_char1 [string range $remote_dir_part 1 1]
    set remote_dir_char2 [string range $remote_dir_part 2 2]
    set remote_dir_char3 [string range $remote_dir_part 3 3]
    if {($remote_dir_char0=="/") && ($remote_dir_char2==":") && ($remote_dir_char3=="/")} {
      if {[string is alpha -strict $remote_dir_char1] && [string is ascii -strict $remote_dir_char1]} {
        set remote_dir [file join \
          [file normalize [translate_wine_drive $remote_dir_char1]] \
          [string trimleft [string range $remote_dir_part 3 "end"] "/"] \
        ]
      }
    }
  }
  set remote_dir [string map [list " " "%20"] $remote_dir]

  set remote_glob_fixed false
  set remote_glob_first_star [string first "*" $remote_glob_part]
  if {$remote_glob_first_star == -1} {
    set remote_glob_fixed true
  }
  set remote_glob [string trim $remote_glob_part "*"]

  if {$doc_spec==""} {
    # empty doc specs are taken as (possibly) links to dirs
    if {$remote_glob==""} {
      return [list true $remote_dir "tag_link tag_link_remote" [list "rank" "$tag_rank"]]
    }
    if {$remote_glob_fixed} {
      return [list true [file join $remote_dir $remote_glob] "tag_link tag_link_remote" [list "rank" "$tag_rank"]]
    }
    # non-fixed non-empty glob considered as not providing a link
    return [list false]
  }

  # non-empty glob should be part of the linked file name
  if {$remote_glob!=""} {
    if {$remote_glob_first_star==0} {
      # looking for even inner occurrence if (original) glob starts by "*"
      if {[string first $remote_glob $doc_spec]==-1} {
        return [list false]
      }
    } {
      if {![string equal -length [string length $remote_glob] $remote_glob $doc_spec]} {
        return [list false]
      }
    }
  }

  return [list true [file join $remote_dir $doc_spec] "tag_link tag_link_remote" [list "rank" "$tag_rank"]]

}

proc get_tag_html_link_action {} {
  # read all elements with class 'tag_declared', check whether has highlight switched-on, switching it off;
  # decoding the tag value from the 'data-spec' string, comparing it with the current searched name;
  # if fits it, switch the highlighting on, incr count of matched tags;
  # if it has zero matches at the end, set link highlighting (to red), and end,
  # otherwise make active and move to the nearest (below) highlighted tag (via setting hash in URL);

  set js_part [strcat "" \
{
  var get_tag_rank = function(el) {
    var attr_value = el.getAttribute("data-rank");
    if ((attr_value===null) || (attr_value=="")) {
      return -1;
    }
    var el_rank = parseFloat(attr_value);
    if (isNaN(el_rank)) {return -1;}
    return el_rank;
  }

  var get_tag_type_spec = function(el) {
    var attr_value = el.getAttribute("data-tagtype");
    if ((attr_value===null) || (attr_value=="")) {
      return "";
    }
    return attr_value;
  }

  var get_tag_value_spec = function(el) {
    var attr_value = el.getAttribute("data-tagvalue");
    if ((attr_value===null) || (attr_value=="")) {
      return "";
    }
    return attr_value;
  }

  var set_active_tag = function (rank_start, to_forward) {
    var ind;
    var active_rank_tmp;
    var active_rank_set = -1;
    var active_rank_set_alt = -1;
    var el_to_set;
    var el_to_set_alt;

    var tags_highlighted = document.getElementsByClassName("tag_highlighted_reg");
    var tags_highlighted_len = tags_highlighted.length;
    if (tags_highlighted_len == 0) {return;}
    if (tags_highlighted_len == 1) {
      if (tags_highlighted[0].className.indexOf("tag_highlighted_active")==-1) {
        tags_highlighted[0].className += " tag_highlighted_active";
      }
      return;
    }
    for (ind = tags_highlighted_len-1; ind>=0; ind--) {
      active_rank_tmp = get_tag_rank(tags_highlighted[ind]);
      if (active_rank_tmp==-1) {continue;}
      if (rank_start==-1) {
        if (active_rank_set==-1) {
          el_to_set = tags_highlighted[ind];
          active_rank_set = active_rank_tmp;
          continue;
        }
        if ((to_forward && (active_rank_tmp < active_rank_set)) || ((!to_forward) && (active_rank_tmp > active_rank_set))) {
          el_to_set = tags_highlighted[ind];
          active_rank_set = active_rank_tmp;
          continue;
        }
        continue;
      }
      if (to_forward) {
        if ((active_rank_set_alt==-1) || (active_rank_tmp < active_rank_set_alt)) {
          active_rank_set_alt = active_rank_tmp;
          el_to_set_alt = tags_highlighted[ind];
        }
        if (active_rank_tmp > rank_start) {
          if ((active_rank_set==-1) || (active_rank_tmp < active_rank_set)) {
            active_rank_set = active_rank_tmp;
            el_to_set = tags_highlighted[ind];
          }
        }
        continue;
      }
      // !to_forward, that is going to lesser ranks
      if ((active_rank_set_alt==-1) || (active_rank_tmp > active_rank_set_alt)) {
        active_rank_set_alt = active_rank_tmp;
        el_to_set_alt = tags_highlighted[ind];
      }
      if (active_rank_tmp < rank_start) {
        if ((active_rank_set==-1) || (active_rank_tmp > active_rank_set)) {
          active_rank_set = active_rank_tmp;
          el_to_set = tags_highlighted[ind];
        }
      }
    }

    if (active_rank_set==-1) {
      if (active_rank_set_alt==-1) {
        return;
      }
      el_to_set = el_to_set_alt;
    }

    el_to_set.className += " tag_highlighted_active";

    if (el_to_set.scrollIntoViewIfNeeded) {
      el_to_set.scrollIntoViewIfNeeded();
    } else {
      if (el_to_set.scrollIntoView) {
        el_to_set.scrollIntoView();
      }
    }
  };

  var find_tag_locations = function (el_rank, tag_type_spec, tag_value_spec) {
    var found_count = 0;
    var ind;

    var link_curr;
    var link_classes;

    var links_all = document.getElementsByClassName("tag_link");
    for (ind = links_all.length-1; ind>=0; ind--) {
      link_curr = links_all[ind];
      link_classes = link_curr.className;
      if (link_classes.indexOf("tag_found")!=-1) {
        // removing previous not-found highlighting, if any was set
        link_curr.className = link_curr.className.replace(/(?:^|\s+)tag_found_none(?!\S)/g, "");
        link_curr.className = link_curr.className.replace(/(?:^|\s+)tag_found_some(?!\S)/g, "");
      }
    }

    var tag_curr;
    var tag_classes;
    var tag_curr_tagtype_spec;
    var tag_curr_tagvalue_spec;

    var tags_light = document.getElementsByClassName("tag_highlighted_reg");
    for (ind = tags_light.length - 1; ind>=0; ind--) {
      tag_curr = tags_light[ind];
      tag_classes = tag_curr.className;
      tag_curr.className = tag_curr.className.replace(/(?:^|\s+)tag_highlighted_reg(?!\S)/g, "");
      tag_curr.className = tag_curr.className.replace(/(?:^|\s+)tag_highlighted_active(?!\S)/g, "");
    }

    var tags_all = document.getElementsByClassName("tag_declared");
    for (ind = tags_all.length - 1; ind>=0; ind--) {
      tag_curr = tags_all[ind];
      tag_curr_tagtype_spec = get_tag_type_spec(tag_curr);
      tag_curr_tagvalue_spec = get_tag_value_spec(tag_curr);
      if (tag_curr_tagtype_spec=="") {
        continue;
      }
      if (tag_curr_tagvalue_spec=="") {
        continue;
      }
      if (tag_curr_tagtype_spec.indexOf(tag_type_spec)!=0) {
        continue;
      }
      if (tag_curr_tagvalue_spec.indexOf(tag_value_spec)!=0) {
        continue;
      }
      tag_curr.className += " tag_highlighted_reg";
      found_count += 1;
    }

    // to set the active tag yet;
    set_active_tag(el_rank, true);

    return found_count;
  };

  var move_to_next_location = function (to_forward) {
    // count the highlighted tags;
    // if none found, end; and if only one found, check whether it is active: if not make it active, and end;
    // if more than one, find the one with the least rank greater than the rank of active tag;
    // make the found tag active and visible;

    var ind;
    var active_rank = -1;
    var active_rank_tmp;

    var tags_active = document.getElementsByClassName("tag_highlighted_active");
    for (ind = tags_active.length - 1; ind>=0; ind--) {
      active_rank_tmp = get_tag_rank(tags_active[ind]);
      // removing previous active highlighting, if any was set
      tags_active[ind].className = tags_active[ind].className.replace(/(?:^|\s+)tag_highlighted_active(?!\S)/g, "");
      if (active_rank_tmp==-1) {continue;}
      active_rank = active_rank_tmp;
    }
    set_active_tag(active_rank, to_forward);
  };

  var traverse_found_tags = function(e) {
    e = e || window.event;
    if (e.ctrlKey) {
      var is_backwards = false;
      var is_forwards = false;
      if (typeof e.key !== "undefined") {
        if (e.key in {",":true, "<":true}) {
          is_backwards = true;
        }
        if (e.key in {".":true, ">":true}) {
          is_forwards = true;
        }
      } else {
        var key_val = e.which || e.keyCode || e.charCode;
        if (key_val == 188) { // prev
          is_backwards = true;
        }
        if (key_val == 190) { // next
          is_forwards = true;
        }
      }
      if (is_backwards) {
        if (e.preventDefault) {e.preventDefault();}
        if (e.stopPropagation) {e.stopPropagation();}
        move_to_next_location(false);
      }
      if (is_forwards) {
        if (e.preventDefault) {e.preventDefault();}
        if (e.stopPropagation) {e.stopPropagation();}
        move_to_next_location(true);
      }
    }
  };

  var find_tag_locations_outer = function (e) {
    if (e.preventDefault) {e.preventDefault();}
    if (e.stopPropagation) {e.stopPropagation();}

    var el_rank = get_tag_rank(this);
    var tag_type_spec = get_tag_type_spec(this);
    var tag_value_spec = get_tag_value_spec(this);

    if (tag_type_spec=="") {
      return;
    }
    if (tag_value_spec=="") {
      return;
    }

    var found_count = find_tag_locations(el_rank, tag_type_spec, tag_value_spec);
    if (found_count == 0) {
      this.className += " tag_found_none";
    } else {
      this.className += " tag_highlighted_reg";
    }

  };

  var register_tag_link_click = function () {
    var ind;
    var link_curr;
    var links_local = document.getElementsByClassName("tag_link_local");
    for (ind = links_local.length - 1; ind>=0; ind--) {
      link_curr = links_local[ind];
      link_curr.onclick = find_tag_locations_outer;
      if (link_curr.captureEvents) {link_curr.captureEvents(Event.CLICK);}
    }
  };

  var initialize_page_state = function() {
    register_tag_link_click();
    document.onkeydown = traverse_found_tags;
  };

  if ((document.readyState == "loading") || (document.readyState == "uninitialized")) {
    if (document.addEventListener) {
      document.addEventListener("DOMContentLoaded", initialize_page_state);
    } else {
      window.onload = initialize_page_state;
    }
  } else {
    initialize_page_state();
  }

} \
  ]

  return $js_part
}

set html_ext_link_regexp [strcat "http" "(s)\{0,1\}\\:\\/\\/" \
  "(\[a-zA-Z0-9-\]+\\.)*\[a-zA-Z0-9-\]+" "(\\:\[0-9\]+)\{0,1\}" \
  "(" "(\\/\[a-zA-Z0-9_\\%\\.\\?\\=\\+\\(\\)\\~-\]*)*" "(\\/\[a-zA-Z0-9_\\%\\.\\?\\=\\+\\(\\)\\~-\]*\[a-zA-Z0-9\])" ")\{0,1\}" "(\\/\{0,1\})" \
  "((\[\\?\\=\\._-\]*)\{0,1\}\\#" "(\[a-zA-Z0-9_\\%\\.\\?\\=\\+\\(\\)\\,\\;-\]*)\[a-zA-Z0-9\]\{0,1\})\{0,1\}"]

proc html_extract_http_links {line} {
  global html_ext_link_regexp

  set link_indices [list]

  set curr_indices [list]
  set curr_offset 0
  while {[regexp -nocase -start $curr_offset -indices -- $html_ext_link_regexp $line curr_indices]!=0} {
    set curr_offset [lindex $curr_indices 1]
    lappend link_indices [list [lindex $curr_indices 0] $curr_offset]
  }

  return $link_indices
}

# tags vs. export to html:
# [#doc:_document_id_string_] used for filename urlized(_file_name_).html
# [#top:_topic_string_] used for title of the html document
# [@img:...] used for images
# [@...:...] used for html links
# [#...:...] used for html anchors

set html_map [list "&" "&amp;" "<" "&lt;" ">" "&gt;" "\"" "&quot;" "'" "&#39;" "\\" "&#92;" \
  "(" "&#40;" ")" "&#41;" "\[" "&#91;" "\]" "&#93;" "{" "&#123;" "}" "&#125;" "|" "&#124;" "~" "&#126;"]
set html_title_min_length 20
proc do_text_export_html {set_clip html_exp_path} {
  global text_inner_code_blocks
  global text_inner_line_parts
  global text_field
  global html_map
  global html_title_min_length
  global min_tilde_count
  global tags_decls_list
  global tags_decls_state_clean
  global tags_links_list
  global tags_links_state_clean
  global tag_type_document_ids
  global tag_type_topic_ids
  global tag_border_size

  if {[$text_field index "end -1 chars"]=="1.0"} {
    show_text_field_message_error [list [list "empty document, no export"]]
    return false
  }

  # taking tags first
  if {!$tags_decls_state_clean} {
    set tags_decls_list [extract_tag_list false]
    set tags_decls_state_clean true
  }
  if {!$tags_links_state_clean} {
    set tags_links_list [extract_tag_list true]
    set tags_links_state_clean true
  }

  set top_id_list_use [list]
  foreach one_type_top $tag_type_topic_ids {
    set one_type_top_part "$one_type_top:"
    lappend top_id_list_use $one_type_top_part [string length $one_type_top_part]
  }

  set tag_based_topic ""

  set got_topic false
  foreach one_tag_info $tags_decls_list {
    if {$got_topic} {
      break
    } {
      set one_tag_string_inner [string range [lindex $one_tag_info 0] "$tag_border_size+1" "end-$tag_border_size"]
      foreach {tag_topic_part tag_topic_part_len} $top_id_list_use {
        if {[string equal -nocase -length $tag_topic_part_len $one_tag_string_inner $tag_topic_part]} {
          set tag_topic_test [string range $one_tag_string_inner $tag_topic_part_len "end"]
          if {$tag_topic_test==""} {
            continue
          }
          set tag_based_topic [string map $html_map $tag_topic_test]
          set got_topic true
          break
        }
      }
    }
  }

  set all_tags [lsort -dictionary -index 1 [concat $tags_decls_list $tags_links_list]]
  set all_tags_len [llength $all_tags]

  set content_text [$text_field get "1.0" "end -1 chars"]

  set exp_txt ""
  append exp_txt "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n"
  append exp_txt "<html>\n<head>\n<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\">\n"

  set html_title $tag_based_topic

  if {$html_title==""} {
    set html_title "TGM Minima export"
    # trying to use first words for title
    set text_start ""
    set left_text_part false
    set to_add_ellipsis false
    foreach one_start_line [split [string range $content_text 0 2000] "\n"] {
      set one_start_line [string trim $one_start_line  " \n\t#=*~"]
      if {$one_start_line != ""} {
        foreach one_start_word [split $one_start_line " \t"] {
          if {$left_text_part} {
            set to_add_ellipsis true
            break
          }
          if {$one_start_word==""} {
            continue
          }
          append text_start $one_start_word
          if {[string length $text_start]>=$html_title_min_length} {
            set left_text_part true
            continue
          }
          append text_start " "
        }
        break
      }
    }
    if {$text_start!=""} {
      set text_start [string totitle $text_start 0 0]
      append html_title " &ndash; " [string map $html_map $text_start]
    }
    if {$to_add_ellipsis} {
      append html_title "&#8230;"
    }
  }

  set use_font_size_common "100%"
  set use_font_size_pline "50%"
  set use_font_size_block "108%"
  append exp_txt "<title>$html_title</title>\n"
  append exp_txt "<base target=\"_top\">\n"
  append exp_txt "<style type=\"text/css\">\n"
  append exp_txt " body { background-color: #ffffff; color: #000000; font-size: $use_font_size_common; }\n"
  append exp_txt " .text_common { color: #000000; font-size: $use_font_size_common; }\n"
  append exp_txt " .text_block { background-color: #e0f0ff; color: #000000; font-size: $use_font_size_block; font-weight: bold; font-family: monospace; }\n"
  append exp_txt " .text_line_hash { background-color: #e4d460; color: #000000; font-size: $use_font_size_pline; }\n"
  append exp_txt " .text_line_equal { background-color: #f0f0b0; color: #000000; font-size: $use_font_size_pline; }\n"
  append exp_txt " .text_line_star { background-color: #f0f0d8; color: #000000; font-size: $use_font_size_pline; }\n"
  append exp_txt " .text_line { background-color: #f0f0d8; color: #000000; font-size: $use_font_size_pline; }\n"
  append exp_txt " a.tag_link_remote:link, a.tag_link_remote:visited, a.tag_link_remote:hover, a.tag_link_remote:active"
  append exp_txt " { color: #c0c030; text-decoration:none; }\n"
  append exp_txt " a.tag_link_local:link, a.tag_link_local:visited, a.tag_link_local:hover, a.tag_link_local:active"
  append exp_txt " { color: #3030c0; text-decoration:none; }\n"
  append exp_txt " a.tag_found_some:link, a.tag_found_some:visited, a.tag_found_some:hover, a.tag_found_some:active"
  append exp_txt " { color: #9090ff; text-decoration:none; }\n"
  append exp_txt " a.tag_found_none:link, a.tag_found_none:visited, a.tag_found_none:hover, a.tag_found_none:active"
  append exp_txt " { color: #f06060; text-decoration:none; }\n"
  append exp_txt " .tag_highlighted_reg { color: #209820; }\n"
  append exp_txt " .tag_highlighted_active { color: #80ff80; background-color: #808080; }\n"
  append exp_txt " a.tag_highlighted_reg:link, a.tag_highlighted_reg:visited, a.tag_highlighted_reg:hover, a.tag_highlighted_reg:active"
  append exp_txt " { color: #209820; text-decoration:none; }\n"
  append exp_txt " a.tag_highlighted_active:link, a.tag_highlighted_active:visited, a.tag_highlighted_active:hover, a.tag_highlighted_active:active"
  append exp_txt " { color: #80ff80; background-color: #808080; text-decoration:none; }\n"
  append exp_txt "</style>\n"
  append exp_txt "<script type=\"text/javascript\">\n" [get_tag_html_link_action] "\n</script>\n"
  append exp_txt "</head>\n<body>\n"

  # putting in (simplified) styles for part-line/blocks;
  set next_pline_idx 0
  set next_pline [lindex $text_inner_line_parts $next_pline_idx]
  incr next_pline_idx

  set next_block_idx 0
  set next_block_info [lindex $text_inner_code_blocks $next_block_idx]
  incr next_block_idx
  set next_block_start [lindex $next_block_info 0]
  set next_block_end [lindex $next_block_info 1]

  set have_to_setup_div true
  set inside_block false

  set has_tag false
  set all_tags_idx 0
  set all_tags_rank 1
  if {$all_tags_len>0} {
    set has_tag true
    set one_tag [lindex $all_tags 0]
    set one_tag_inner [string range [lindex $one_tag 0] $tag_border_size "end-$tag_border_size"]
    set one_tag_inner_first [string range $one_tag_inner 0 0]
    set one_tag_inner [string range $one_tag_inner 1 "end"]
    set one_tag_idx_info [split [lindex $one_tag 1] "."]
    set one_tag_row [lindex $one_tag_idx_info 0]
  }

  set line_idx 0
  foreach one_line [split $content_text "\n"] {
    incr line_idx
    set curr_class ""
    set last_char_un ""

    # looking at tags inside the current line
    set one_line_parsed ""
    set curr_line_offset 0
    while {$has_tag && ($one_tag_row==$line_idx)} {
      set one_tag_col [lindex $one_tag_idx_info 1]
      set one_tag_len [lindex $one_tag 2]
      set curr_line_offset_new [expr {$one_tag_col+$one_tag_len}]
      incr one_tag_col -$curr_line_offset
      set curr_line_offset $curr_line_offset_new
      set part_preced [string range $one_line 0 [expr {$one_tag_col-1}]]
      # before putting the preceding parts to the parsed-line string,
      # looking for external links located in them;
      set ext_links [html_extract_http_links $part_preced]
      if {[llength $ext_links]==0} {
        set last_char_un [string range $part_preced "end" "end"]
        append one_line_parsed [string map $html_map $part_preced]
      } {
        set after_link_start 0
        foreach one_ext_link $ext_links {
          set cur_link_start [lindex $one_ext_link 0]
          set cur_link_end [lindex $one_ext_link 1]
          if {$after_link_start<$cur_link_start} {
            append one_line_parsed [string map $html_map [string range $part_preced $after_link_start [expr {$cur_link_start-1}]]]
          }
          set cur_link_string [string range $part_preced $cur_link_start $cur_link_end]
          append one_line_parsed "<a href=\"$cur_link_string\">$cur_link_string</a>"
          set after_link_start [expr {$cur_link_end+1}]
        }
        set last_char_un [string range $part_preced $after_link_start $after_link_start]
        append one_line_parsed [string map $html_map [string range $part_preced $after_link_start "end"]]
      }
      set one_line [string range $one_line [expr {$one_tag_col+$one_tag_len}] "end"]
      if {$one_tag_inner_first!="@"} {
        # dealing with declared tags
        if {$one_line_parsed!=""} {
          if {!($last_char_un in [list " " "\{" "\[" "("])} {
            append one_line_parsed " "
            set last_char_un " "
          }
        }
        set curr_tag_decl_info [get_tag_html_decl $one_tag_inner $all_tags_rank]
        if {[lindex $curr_tag_decl_info 0]} {
          set curr_tag_decl [lindex $curr_tag_decl_info 1]
          set curr_tag_class [lindex $curr_tag_decl_info 2]
          set curr_tag_data [list]
          foreach {curr_tag_data_name curr_tag_data_value} [lindex $curr_tag_decl_info 3] {
            lappend curr_tag_data "data-$curr_tag_data_name=\"$curr_tag_data_value\""
          }
          set curr_tag_data [join $curr_tag_data " "]
          append one_line_parsed "<a name=\"$curr_tag_decl\">"

          set curr_tag_sep [string first ":" $one_tag_inner]
          set curr_tag_type [string range $one_tag_inner 0 [expr {$curr_tag_sep-1}]]
          if {$curr_tag_type in $tag_type_topic_ids} {
            append one_line_parsed "<b class=\"$curr_tag_class\" $curr_tag_data>" [string map $html_map [string range $one_tag_inner [expr {$curr_tag_sep+1}] "end"]] "</b>"
          } elseif {!($curr_tag_type in $tag_type_document_ids)} {
            append one_line_parsed "<b class=\"$curr_tag_class\" $curr_tag_data>" [string map $html_map $one_tag_inner] "</b>"
          }
          append one_line_parsed "</a>"
          set last_char_un ""
          if {$one_line!=""} {
            if {!([string range $one_line 0 0] in [list " " "\}" "\]" ")" "." "," ";" ":"])} {
              append one_line_parsed " "
              set last_char_un " "
            }
          }
        }
      } {
        # dealing with links to tags
        if {$one_line_parsed!=""} {
          if {!($last_char_un in [list " " "\{" "\[" "("])} {
            append one_line_parsed " "
            set last_char_un " "
          }
        }
        set curr_tag_link_info [get_tag_html_link $one_tag_inner $all_tags_rank]
        if {[lindex $curr_tag_link_info 0]} {
          set curr_tag_link [lindex $curr_tag_link_info 1]
          set curr_tag_class [lindex $curr_tag_link_info 2]
          set curr_tag_data [list]
          foreach {curr_tag_data_name curr_tag_data_value} [lindex $curr_tag_link_info 3] {
            lappend curr_tag_data "data-$curr_tag_data_name=\"$curr_tag_data_value\""
          }
          set curr_tag_data [join $curr_tag_data " "]
          append one_line_parsed "<a href=\"$curr_tag_link\" class=\"$curr_tag_class\" $curr_tag_data title=\"" \
            "@" [string map $html_map $one_tag_inner] "\">&rArr;</a>"
          set last_char_un ""
        } elseif {[lindex $curr_tag_link_info 1]=="image"} {
          set curr_tag_link [lindex $curr_tag_link_info 2]
          set curr_tag_class [lindex $curr_tag_link_info 3]
          set curr_tag_data [list]
          foreach {curr_tag_data_name curr_tag_data_value} [lindex $curr_tag_link_info 4] {
            lappend curr_tag_data "data-$curr_tag_data_name=\"$curr_tag_data_value\""
          }
          set curr_tag_data [join $curr_tag_data " "]
          append one_line_parsed "<img src=\"$curr_tag_link\" class=\"$curr_tag_class\" $curr_tag_data />"
          set last_char_un ""
        }
        if {$one_line!=""} {
          if {!([string range $one_line 0 0] in [list " " "\}" "\]" ")" "." "," ";" ":"])} {
            append one_line_parsed " "
            set last_char_un " "
          }
        }
      }
      incr all_tags_idx
      incr all_tags_rank
      if {$all_tags_idx>=$all_tags_len} {
        set has_tag false
      } {
        set one_tag [lindex $all_tags $all_tags_idx]
        set one_tag_inner [string range [lindex $one_tag 0] $tag_border_size "end-$tag_border_size"]
        set one_tag_inner_first [string range $one_tag_inner 0 0]
        set one_tag_inner [string range $one_tag_inner 1 "end"]
        set one_tag_idx_info [split [lindex $one_tag 1] "."]
        set one_tag_row [lindex $one_tag_idx_info 0]
      }
    }

    # looking for external links located in the left part of the current line;
    # only doing the safe-char passing if some link found,
    # since otherwise we would have empty "one_line" variable for all cases
    # while we later use it for deciding on possible changes of the classes;
    set ext_links [html_extract_http_links $one_line]
    if {[llength $ext_links]!=0} {
      set after_link_start 0
      foreach one_ext_link $ext_links {
        set cur_link_start [lindex $one_ext_link 0]
        set cur_link_end [lindex $one_ext_link 1]
        if {$after_link_start<$cur_link_start} {
          append one_line_parsed [string map $html_map [string range $one_line $after_link_start [expr {$cur_link_start-1}]]]
        }
        set cur_link_string [string range $one_line $cur_link_start $cur_link_end]
        append one_line_parsed "<a href=\"$cur_link_string\">$cur_link_string</a>"
        set last_char_un ""
        set after_link_start [expr {$cur_link_end+1}]
      }
      set one_line [string range $one_line $after_link_start "end"]
    }

    if {$next_block_start==$line_idx} {
      if {!$have_to_setup_div} {
        append exp_txt "</div>\n"
      }
      set curr_class "text_block"
      append exp_txt "<div class=\"$curr_class\">\n"
      append exp_txt "<pre class=\"$curr_class\">\n"
      set have_to_setup_div false
      set inside_block true
    }

    set is_partline_line false
    if {$next_pline==$line_idx} {
      set is_partline_line true
      if {!$have_to_setup_div} {
        append exp_txt "</div>\n"
      }
      # the separation lines should contain neither tags nor external links,
      # thus using directly the (otherwise possibly cut) "one_line" variable here;
      set first_line_char [string range $one_line 0 0]
      set curr_class "text_line"
      if {$first_line_char=="#"} {
        set curr_class "text_line_hash"
      }
      if {$first_line_char=="="} {
        set curr_class "text_line_equal"
      }
      if {$first_line_char=="*"} {
        set curr_class "text_line_star"
      }
      append exp_txt "<div class=\"$curr_class\">\n"
      set have_to_setup_div false
    }

    if {$have_to_setup_div} {
      append exp_txt "<div class=\"text_common\">\n"
      set have_to_setup_div false
    }

    # ~-lines considered (alike split lines without tags/links inside them)
    set is_block_tilde_line false
    if {$inside_block && ($one_line_parsed=="")} {
      if {[string length $one_line]>=$min_tilde_count} {
        if {[string trim $one_line "~"]==""} {
          set is_block_tilde_line true
        }
      }
    }

    if {!$is_block_tilde_line} {
      set one_line [strcat $one_line_parsed [string map $html_map $one_line]]
      if {$is_partline_line} {
        set one_line_width_orig_full 80.0
        set one_line_width [expr {round(100 * [string length $one_line] / $one_line_width_orig_full)}]
        if {$one_line_width>100} {
          set one_line_width 100
        }
        append exp_txt "<hr class=\"$curr_class\" width=\"" $one_line_width "%\" size=\"1\" align=\"left\">"
      } {
        append exp_txt "$one_line"
      }
      if {!$inside_block} {
        append exp_txt "<br>"
      }
      append exp_txt "\n"
    }
    set is_block_tilde_line false

    if {$next_pline==$line_idx} {
      append exp_txt "</div>\n"
      set next_pline [lindex $text_inner_line_parts $next_pline_idx]
      incr next_pline_idx
      set have_to_setup_div true
    }
    if {$next_block_end==$line_idx} {
      append exp_txt "</pre>\n"
      append exp_txt "</div>\n"
      set next_block_info [lindex $text_inner_code_blocks $next_block_idx]
      incr next_block_idx
      set next_block_start [lindex $next_block_info 0]
      set next_block_end [lindex $next_block_info 1]
      set have_to_setup_div true
      set inside_block false
    }
  }

  if {$inside_block} {
    append exp_txt "</pre>\n"
  }
  if {!$have_to_setup_div} {
    append exp_txt "</div>\n"
  }
  append exp_txt "</body>\n</html>\n"

  if {$set_clip} {
    set_clipboard_text $exp_txt
    store_clipboard_onset
  }

  if {$html_exp_path==""} {
    return true
  }

  if {[catch {
    set hfh [open $html_exp_path "w"]
    puts -nonewline $hfh $exp_txt
    flush $hfh
    close $hfh
  }]} {
    return false
  }

  return true
}

set continue_tasks_var 0
proc continue_tasks {} {
  global continue_tasks_var

  if {$continue_tasks_var==0} {
    set continue_tasks_var 1
  } {
    set continue_tasks_var 0
  }
}

proc pause_tasks {time_ms} {
  global continue_tasks_var

  after $time_ms continue_tasks
  vwait continue_tasks_var
}

proc bind_keypad {wdg} {
  global keysyms_moves

  foreach one_move $keysyms_moves {
    catch {
      bind $wdg <KeyPress-KP_$one_move> "event generate $wdg <KeyPress-$one_move>; break"
    }
    catch {
      bind $wdg <KeyRelease-KP_$one_move> "event generate $wdg <KeyRelease-$one_move>; break"
    }
    catch {
      bind $wdg <Control-KeyPress-KP_$one_move> "event generate $wdg <Control-KeyPress-$one_move>; break"
    }
    catch {
      bind $wdg <Control-KeyRelease-KP_$one_move> "event generate $wdg <Control-KeyRelease-$one_move>; break"
    }
    catch {
      bind $wdg <Shift-KeyPress-KP_$one_move> "event generate $wdg <Shift-KeyPress-$one_move>; break"
    }
    catch {
      bind $wdg <Shift-KeyRelease-KP_$one_move> "event generate $wdg <Shift-KeyRelease-$one_move>; break"
    }
    catch {
      bind $wdg <Control-Shift-KeyPress-KP_$one_move> "event generate $wdg <Control-Shift-KeyPress-$one_move>; break"
    }
    catch {
      bind $wdg <Control-Shift-KeyRelease-KP_$one_move> "event generate $wdg <Control-Shift-KeyRelease-$one_move>; break"
    }
  }
}

bind_keypad $text_field
bind_keypad $command_field
bind_keypad $search_entry
if {[is_batch_mode]} {
  bind_keypad $tags_decls_line
  bind_keypad $tags_links_line
}

set plot_input_frame ""
set plot_input_field ""
set plot_put_window ""
set maxima_pid_batch ""
set mxmbatch_input_field ""
set mxmbatch_output_field ""

set_text_tags_view_focused false
set_cmmd_tags_view_focused false

# Gnuplot batch windows

# should put in: comments, strings, brackets
set tag_plot_strings "plot_strings"
set tag_plot_comments "plot_comments"
set tag_plot_brackets_pair "plot_brackets_pair"
set tag_plot_brackets "plot_brackets"

# this is quite similar to the command_field-based case,
# may be to put these into a single function eventually;
# both fields are (relatively) small, with single/overall block;
set plot_bracket_jump_position ""
proc update_plot_bracket_pairing {} {
  global plot_bracket_stack
  global bracket_complements
  global plot_bracket_jump_position
  global tag_plot_brackets_pair
  global tag_plot_brackets
  global plot_input_field

  $plot_input_field tag remove $tag_plot_brackets_pair "1.0" "end"
  $plot_input_field tag remove $tag_plot_brackets "1.0" "end"

  set plot_bracket_jump_position ""

  set insert_pos [split [$plot_input_field index "insert"] "."]
  set insert_line [lindex $insert_pos 0]
  set insert_column [lindex $insert_pos 1]
  set insert_text [$plot_input_field get "insert" "insert + 1 chars"]
  set use_bracket ""
  set use_bracket_line -1
  set use_bracket_column -1

  # for (position of) the seen bracket in the bracket stack
  set curr_idx 0
  set one_bracket [list]
  set bracket_count [llength $plot_bracket_stack]
  set having_bracket false

  if {$insert_column>0} {
    set insert_prev_text [$plot_input_field get "insert - 1 chars" "insert"]
    if {$insert_prev_text in [array names bracket_complements]} {
      set having_bracket true
      set use_bracket $insert_prev_text
      set use_bracket_line $insert_line
      set use_bracket_column [expr {$insert_column-1}]
      # checking whether this bracket is (not) acting as a real bracket
      while {$curr_idx < $bracket_count} {
        set one_bracket [lindex $plot_bracket_stack $curr_idx]
        if {($use_bracket_line==[lindex $one_bracket 0]) && ($use_bracket_column==[lindex $one_bracket 1])} {
          break
        }
        incr curr_idx
      }
      if {$curr_idx >= $bracket_count} {
        set curr_idx 0
        set use_bracket ""
        set use_bracket_line -1
        set use_bracket_column -1
        set having_bracket false
      }
    }
  }
  if {$use_bracket==""} {
    if {$insert_text in [array names bracket_complements]} {
      set having_bracket true
      set use_bracket $insert_text
      set use_bracket_line $insert_line
      set use_bracket_column $insert_column
      # checking whether this bracket is (not) acting as a real bracket
      while {$curr_idx < $bracket_count} {
        set one_bracket [lindex $plot_bracket_stack $curr_idx]
        if {($use_bracket_line==[lindex $one_bracket 0]) && ($use_bracket_column==[lindex $one_bracket 1])} {
          break
        }
        incr curr_idx
      }
      if {$curr_idx >= $bracket_count} {
        set curr_idx 0
        set use_bracket ""
        set use_bracket_line -1
        set use_bracket_column -1
        set having_bracket false
      }
    }
  }

  if {!$having_bracket} {
    return
  }

  set brackets_open [list "\{" "\[" "("]
  set brackets_close [list "\}" "\]" ")"]

  set bracket_index_char [lindex $one_bracket 2]
  set bracket_index_char_opposite $bracket_complements($bracket_index_char)
  set bracket_index_start_main [join [lrange $one_bracket 0 1] "."] 
  set is_paired false
  set local_stack_size 1
  if {$bracket_index_char in $brackets_open} {
    # search forwards
    incr curr_idx
    while {$curr_idx < $bracket_count} {
      set one_bracket [lindex $plot_bracket_stack $curr_idx]
      set one_bracket_char [lindex $one_bracket 2]
      if {$one_bracket_char==$bracket_index_char} {
        incr local_stack_size 1
      } elseif {$one_bracket_char==$bracket_index_char_opposite} {
        incr local_stack_size -1
        if {$local_stack_size==0} {
          break
        }
      }
      incr curr_idx
    }
    if {$curr_idx < $bracket_count} {
      set bracket_index_start [join [lrange $one_bracket 0 1] "."]
      $plot_input_field tag add $tag_plot_brackets_pair $bracket_index_start "$bracket_index_start + 1 char"
      set plot_bracket_jump_position "$bracket_index_start + 1 chars"
      set is_paired true
    }
  } else {
    # search backwards
    incr curr_idx -1
    while {$curr_idx >= 0} {
      set one_bracket [lindex $plot_bracket_stack $curr_idx]
      set one_bracket_char [lindex $one_bracket 2]
      if {$one_bracket_char==$bracket_index_char} {
        incr local_stack_size 1
      } elseif {$one_bracket_char==$bracket_index_char_opposite} {
        incr local_stack_size -1
        if {$local_stack_size==0} {
          break
        }
      }
      incr curr_idx -1
    }
    if {$curr_idx >= 0} {
      set bracket_index_start [join [lrange $one_bracket 0 1] "."]
      $plot_input_field tag add $tag_plot_brackets_pair $bracket_index_start "$bracket_index_start + 1 char"
      set plot_bracket_jump_position "$bracket_index_start + 1 chars"
      set is_paired true
    }
  }
  set tag_name_use [expr {$is_paired ? $tag_plot_brackets_pair : $tag_plot_brackets}]
  $plot_input_field tag add $tag_name_use $bracket_index_start_main "$bracket_index_start_main + 1 char"
}

proc get_comments_strings_brackets_in_gnuplot {parse_text} {
  set found_comments [list]
  set found_strings [list]
  set found_brackets [list]

  set in_string false
  set cur_string_start [list]
  set cur_string_onset ""

  set after_backslash false
  set bracket_list [list "\{" "\}" "\[" "\]" "(" ")"]

  set comment_continues false
  set line_idx 0
  foreach one_line [split $parse_text "\n"] {
    incr line_idx
    set char_idx 0
    set one_line_list [split $one_line {}]
    set one_line_list_length [llength $one_line_list]

    set last_line_char ""
    if {$one_line_list_length > 0} {
      set last_line_char [lindex $one_line_list "end"]
    }

    if {$comment_continues} {
      lappend found_comments [list [list $line_idx 0] [list $line_idx [expr {$one_line_list_length - 1}]]]
      if {$last_line_char != "\\"} {
        set comment_continues false
      }
      continue
    }

    set after_backslash false
    if {$one_line_list_length==0} {
      if {$in_string} {
        # strings end on non-continued line ends, that is on empty lines too;
        set in_string false
        set cur_string_onset ""
        lappend found_strings [list $cur_string_start [list $line_idx 0]]
        set cur_string_start [list]
      }
      continue
    }
    while {$char_idx < $one_line_list_length} {
      set one_char [lindex $one_line_list $char_idx]
      set next_char [lindex $one_line_list [expr {$char_idx + 1}]]
      if {$in_string} {
        # strings end on non-continued line ends;
        if {($next_char=="") && ($one_char!="\\")} {
          set in_string false
          set cur_string_onset ""
          lappend found_strings [list $cur_string_start [list $line_idx $char_idx]]
          set cur_string_start [list]
          set after_backslash false
        } elseif {$after_backslash} {
          set after_backslash false
        } {
          if {$one_char==$cur_string_onset} {
            set in_string false
            set cur_string_onset ""
            lappend found_strings [list $cur_string_start [list $line_idx $char_idx]]
            set cur_string_start [list]
          } elseif {($one_char=="\\") && ($cur_string_onset !="'")} {
            set after_backslash true
          }
        }
        incr char_idx
        continue
      } else {
        if {!$after_backslash} {
          if {$one_char in $bracket_list} {
            lappend found_brackets [list $line_idx $char_idx $one_char]
            incr char_idx
            continue
          }
          if {($one_char=="\"") || ($one_char=="'")} {
            set in_string true
            set cur_string_onset $one_char
            set cur_string_start [list $line_idx $char_idx]
            # strings end on non-continued line ends, thus if string starts on line very end, it ends there too;
            if {$next_char==""} {
              set in_string false
              set cur_string_onset ""
              lappend found_strings [list $cur_string_start [list $line_idx $char_idx]]
              set cur_string_start [list]
            }
            incr char_idx
            continue
          }
          # considering hash-to-end_of_line comments
          # that are continued to other line(s) if last line char is backslash
          if {$one_char=="#"} {
            lappend found_comments [list [list $line_idx $char_idx] [list $line_idx [expr {$one_line_list_length - 1}]]]
            if {$last_line_char == "\\"} {
              set comment_continues true
            }
            break
          }
        }
        if {$after_backslash} {
          set after_backslash false
          incr char_idx
          continue
        }
        if {$one_char == "\\"} {
          set after_backslash true
        }
        incr char_idx
        continue
      }
    }
  }

  if {$in_string} {
    lappend found_strings [list $cur_string_start [list [expr {$line_idx+1}] 0]]
  }

  return [list $found_comments $found_strings $found_brackets]
}

proc get_data_holders_in_gnuplot {} {
  global plot_input_field

  set res_rows [list]

  foreach one_row_idx [$plot_input_field search -regexp -all -- "^\[\\t\\ \]*@@D(I|B|W|A\\,|A\\;)\[\\t\\ \]*\$" "1.0" "end"] {
    set one_row [lindex [split $one_row_idx "."] 0]
    if {$one_row==1} {
      lappend res_rows $one_row
      continue
    }
    set prev_row [expr {$one_row - 1}]
    if {[lindex [split [$plot_input_field index "$prev_row.end"] "."] 1]==0} {
      lappend res_rows $one_row
      continue
    }
    if {[$plot_input_field get "$prev_row.end -1 chars" "$prev_row.end"]!="\\"} {
      lappend res_rows $one_row
      continue
    }
  }

  return $res_rows
}

set tag_plot_data_holders "tag_plot_data_holders_name"

set plot_bracket_stack [list]
proc update_plot_strings_comments {} {
  global plot_input_field
  global plot_field_updates_lines_id
  global tag_plot_strings
  global tag_plot_comments
  global plot_bracket_stack
  global tag_plot_data_holders

  catch {after cancel $plot_field_updates_lines_id}
  set plot_field_updates_lines_id ""

  $plot_input_field tag remove $tag_plot_strings "1.0" "end"
  $plot_input_field tag remove $tag_plot_comments "1.0" "end"

  $plot_input_field tag remove $tag_plot_data_holders "1.0" "end"

  foreach one_row [get_data_holders_in_gnuplot] {
    $plot_input_field tag add $tag_plot_data_holders "$one_row.0" "$one_row.end"
  }

  # do a simple parsing on plot field
  set comments_strings_brackets [get_comments_strings_brackets_in_gnuplot [$plot_input_field get "1.0" "end -1 chars"]]

  foreach one_comment [lindex $comments_strings_brackets 0] {
    set one_start [lindex $one_comment 0]
    set index_start [join $one_start "."]

    set one_end [lindex $one_comment 1]
    set index_end "end"
    if {[lindex $one_end 0]!="end"} {
      set index_end [join $one_end "."]
      append index_end " + 1 chars"
    }
    $plot_input_field tag add $tag_plot_comments $index_start $index_end
  }

  foreach one_string [lindex $comments_strings_brackets 1] {
    set one_start [lindex $one_string 0]
    set index_start [join $one_start "."]

    set one_end [lindex $one_string 1]
    set index_end "end"
    if {[lindex $one_end 0]!="end"} {
      set index_end [join $one_end "."]
      append index_end " + 1 chars"
    }
    $plot_input_field tag add $tag_plot_strings $index_start $index_end
  }

  set plot_bracket_stack [lindex $comments_strings_brackets 2]

  # continue with highlight updates
  update_plot_highlight_lines

  return $comments_strings_brackets
}
proc update_plot_highlight_lines {} {
  global plot_field_updates_lines_id

  catch {after cancel $plot_field_updates_lines_id}
  set plot_field_updates_lines_id ""

  # by now we only deal with brackets
  update_plot_bracket_pairing
}

proc update_plot_highlight_lines_onset {} {
  global plot_field_updates_lines_id

  catch {after cancel $plot_field_updates_lines_id}
  set plot_field_updates_lines_id ""
  after 0 "update_plot_highlight_lines"
}

set plot_field_updates_lines_id ""
set plot_field_updates_props_id ""
proc update_plot_properties {} {
  global plot_field_updates_lines_id
  global plot_field_updates_props_id

  catch {after cancel $plot_field_updates_props_id}
  set plot_field_updates_props_id ""

  catch {after cancel $plot_field_updates_lines_id}
  set plot_field_updates_props_id ""

  after 0 update_plot_clean_state

  # this continues with highlights too
  return [update_plot_strings_comments]
}

set common_spec_chars [list "(" ")" "\[" "\]" "\{" "\}" "\"" "\\" "#"]

proc update_plot_properties_delayed {} {
  global plot_field_updates_props_id
  global plot_input_field
  global common_spec_chars

  catch {after cancel $plot_field_updates_props_id}

  set prev_char [$plot_input_field get "insert -1 chars" "insert"]
  set cur_char [$plot_input_field get "insert" "insert +1 chars"]
  set after_time 100
  if {$cur_char in $common_spec_chars} {
    set after_time 0
  } elseif {$prev_char in $common_spec_chars} {
    set after_time 0
  }
  set plot_field_updates_props_id [after $after_time "update_plot_properties"]

  after 0 update_plot_clean_state
}

proc plot_jump_bracket {{to_select false}} {
  global plot_bracket_jump_position
  global plot_input_field

  if {$plot_bracket_jump_position==""} {
    return
  }

  if {$to_select} {
    set start_pos [$plot_input_field index "insert"]
    set start_pos_info [split $start_pos "."]
    set start_pos_row [lindex $start_pos_info 0]
    set start_pos_col [lindex $start_pos_info 1]
    set_anchor_insert $plot_input_field
  }

  $plot_input_field mark set "insert" $plot_bracket_jump_position
  $plot_input_field see $plot_bracket_jump_position

  if {$to_select} {
    set end_pos [$plot_input_field index "insert"]
    set end_pos_info [split $end_pos "."]
    set end_pos_row [lindex $end_pos_info 0]
    set end_pos_col [lindex $end_pos_info 1]
    if {($start_pos_row<$end_pos_row) || (($start_pos_row==$end_pos_row)&&($start_pos_col<$end_pos_col))} {
      set sel_from $start_pos
      set sel_to $end_pos
    } {
      set sel_from $end_pos
      set sel_to $start_pos
    }
    $plot_input_field tag remove "sel" "1.0" "end"
    $plot_input_field tag add "sel" "$sel_from" "$sel_to"
  } {
    set_anchor_insert $plot_input_field
  }

  after 0 "update_plot_highlight_lines"
}

proc update_plot_edited_state {} {
  global gnuplot_at_output_view
  global plot_input_field
}

proc update_plot_clean_state {} {
  global plot_input_field
  global lgnuplot_killplot
  global lgnuplot_plot

  set has_some_content false
  if {[$plot_input_field index "end -1 chars"]!="1.0"} {
    set has_some_content true
  }

  if {$has_some_content} {
    if {[is_gnuplot_running]} {
      $lgnuplot_killplot configure -state "normal"
    } {
      $lgnuplot_killplot configure -state "disabled"
    }
    $lgnuplot_plot configure -state "normal"
  } {
    $lgnuplot_killplot configure -state "disabled"
    $lgnuplot_plot configure -state "disabled"
  }
}

proc update_gnuplot_warning_label {} {
  global lgnuplot_warning
  global plot_input_warning

  if {$lgnuplot_warning==""} {
    return
  }

  if {$plot_input_warning} {
    $lgnuplot_warning configure -foreground "#c0f0c0" -background "#a0c0a0"
  } {
    $lgnuplot_warning configure -foreground "#c0c0c0" -background "#a0a0a0"
  }
}

proc update_plot_input_warning {} {
  set_gnuplot_title
  update_gnuplot_warning_label
}

proc unset_plot_input_warning {} {
  global plot_input_warning

  set plot_input_warning false
  update_plot_input_warning
}
proc set_plot_input_warning {} {
  global plot_input_warning

  set plot_input_warning true
  update_plot_input_warning
}
proc switch_plot_input_warning {} {
  global plot_input_warning

  set plot_input_warning [expr {$plot_input_warning ? false : true}]
  update_plot_input_warning
}

set gnuplot_at_output_view false

proc is_gnuplot_at_output_empty {} {
  global plot_output_field

  if {$plot_output_field==""} {
    return true
  }

  if {[$plot_output_field index "end -1 chars"]=="1.0"} {
    return true
  }

  return false
}

proc set_gnuplot_title {} {
  global plot_put_window
  global plot_shown_part
  global plot_input_warning

  if {$plot_put_window==""} {
    return
  }

  set title_string "TGM Minima: Gnuplot $plot_shown_part"
  if {$plot_input_warning} {
    append title_string " *"
  }

  wm title $plot_put_window $title_string
}

proc present_gnuplot_close {} {
  close_gnuplot_direct
}

proc present_gnuplot_clean {} {
  global plot_input_field
  global lgnuplot_plot
  global lgnuplot_killplot

  $plot_input_field delete "1.0" "end"
  $plot_input_field edit reset
  $plot_input_field edit modified false
  $lgnuplot_plot configure -state "disabled"
  $lgnuplot_killplot configure -state "disabled"
}

proc present_gnuplot_erase_data {} {
  global plot_data_field

  $plot_data_field delete "1.0" "end"
  $plot_data_field edit reset
  $plot_data_field edit modified false
}

proc set_gnuplot_topmost_props {} {
  global gnuplot_keep_always_topmost
  global plot_put_window
  global lgnuplot_topwise

  if {$plot_put_window==""} {
    return
  }

  wm attributes $plot_put_window -topmost $gnuplot_keep_always_topmost

  $lgnuplot_topwise configure -foreground [expr {$gnuplot_keep_always_topmost ? "#ffffff" : "#f0f0f0"}]
}

proc set_gnuplot_topmost {val_top} {
  global gnuplot_keep_always_topmost
  global plot_put_window

  set gnuplot_keep_always_topmost $val_top

  if {$plot_put_window==""} {
    return
  }

  set_gnuplot_topmost_props
}

proc switch_gnuplot_topmost {} {
  global gnuplot_keep_always_topmost
  set_gnuplot_topmost [expr {!$gnuplot_keep_always_topmost}]
}

set gnuplot_topwise_menu ""
proc offer_gnuplot_topwise_menu {x_abs y_abs x_pos y_pos} {
  global topw
  global lgnuplot_topwise
  global gnuplot_topwise_menu
  global gnuplot_keep_always_topmost
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  if {($x_abs=="") || ($y_abs=="")} {
    set abs_pos [get_gnuplot_menu_position $lgnuplot_topwise]
    set use_x [expr {$x_pos + [lindex $abs_pos 0] + $default_menu_offset_x}]
    set use_y [expr {$y_pos + [lindex $abs_pos 1] + $default_menu_offset_y}]
  } {
    set use_x [expr {$x_abs + $default_menu_offset_x}]
    set use_y [expr {$y_abs + $default_menu_offset_y}]
  }

  if {$gnuplot_topwise_menu!=""} {
    catch {destroy $gnuplot_topwise_menu}
  }
  set gnuplot_topwise_menu [menu "$topw.gnuplot_topwise_menu" -tearoff false]
  $gnuplot_topwise_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $gnuplot_topwise_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $gnuplot_topwise_menu configure -activeborderwidth 1
  $gnuplot_topwise_menu configure -title "keeping topwise"

  $gnuplot_topwise_menu add command -label "Set to stay as a topmost window:" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  $gnuplot_topwise_menu add command -label "no" -command [list "set_gnuplot_topmost" false]
  $gnuplot_topwise_menu add command -label "yes" -command [list "set_gnuplot_topmost" true]

  $gnuplot_topwise_menu entryconfigure [expr {$gnuplot_keep_always_topmost ? 2 : 1}] -background $menu_selected_background -activebackground "#e0e0ff"

  tk_popup $gnuplot_topwise_menu $use_x $use_y
}

set gnuplot_keep_focused_time [lindex [lindex $gnuplot_focusing_menu_options $gnuplot_keep_focused_idx] 1]
set gnuplot_keep_focused_full [lindex [lindex $gnuplot_focusing_menu_options $gnuplot_keep_focused_idx] 2]

proc set_gnuplot_focusing_props {} {
  global gnuplot_keep_focused_time
  global plot_put_window
  global lgnuplot_focus

  if {$plot_put_window==""} {
    return
  }
  $lgnuplot_focus configure -background [expr {($gnuplot_keep_focused_time!=0) ? "#ffffff" : "#f0f0f0"}]
}

proc set_gnuplot_focusing {foc_idx} {
  global gnuplot_focusing_menu_options
  global gnuplot_keep_focused_idx
  global gnuplot_keep_focused_time
  global gnuplot_keep_focused_full
  global gnuplot_keep_focused_last_nonzero_idx

  set gnuplot_keep_focused_idx $foc_idx
  if {$gnuplot_keep_focused_idx != 0} {
    set gnuplot_keep_focused_last_nonzero_idx $gnuplot_keep_focused_idx
  }

  set curr_focus_form [lindex $gnuplot_focusing_menu_options $gnuplot_keep_focused_idx]

  set gnuplot_keep_focused_time [lindex $curr_focus_form 1]
  set gnuplot_keep_focused_full [lindex $curr_focus_form 2]

  set_gnuplot_focusing_props
}

proc switch_gnuplot_focusing {} {
  global gnuplot_keep_focused_idx
  global gnuplot_keep_focused_last_nonzero_idx

  if {$gnuplot_keep_focused_idx != 0} {
    set_gnuplot_focusing 0
  } {
    set_gnuplot_focusing $gnuplot_keep_focused_last_nonzero_idx
  }
}

set gnuplot_focusing_menu ""
proc offer_gnuplot_focusing_menu {x_abs y_abs x_pos y_pos} {
  global topw
  global lgnuplot_focus
  global gnuplot_focusing_menu
  global gnuplot_focusing_menu_options
  global gnuplot_keep_focused_idx
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  if {$gnuplot_focusing_menu!=""} {
    catch {destroy $gnuplot_focusing_menu}
  }
  set gnuplot_focusing_menu [menu "$topw.gnuplot_focusing_menu" -tearoff false]
  $gnuplot_focusing_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $gnuplot_focusing_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $gnuplot_focusing_menu configure -activeborderwidth 1
  $gnuplot_focusing_menu configure -title "keeping focused"

  $gnuplot_focusing_menu add command -label "Trying to stay focused after plotting:" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  set use_idx 0
  foreach one_option $gnuplot_focusing_menu_options {
    set curr_selected false
    if {$use_idx==$gnuplot_keep_focused_idx} {
      set curr_selected true
    }
    set one_option_label [lindex $one_option 0]
    $gnuplot_focusing_menu add command -label $one_option_label -command [list "set_gnuplot_focusing" $use_idx]
    incr use_idx
    if {$curr_selected} {
      $gnuplot_focusing_menu entryconfigure $use_idx -background $menu_selected_background -activebackground "#e0e0ff"
    }
  }

  tk_popup $gnuplot_focusing_menu $use_x $use_y
}

proc set_warning_on_gnuplot_parts {to_warn} {
  global plot_input_warning

  set plot_input_warning $to_warn
  update_plot_input_warning
}

set gnuplot_warning_menu ""
proc offer_gnuplot_warning_menu {x_abs y_abs x_pos y_pos} {
  global topw
  global lgnuplot_warning
  global gnuplot_warning_menu
  global plot_input_warning
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  if {$gnuplot_warning_menu!=""} {
    catch {destroy $gnuplot_warning_menu}
  }

  set gnuplot_warning_menu [menu "$topw.gnuplot_warning_menu" -tearoff false]
  $gnuplot_warning_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $gnuplot_warning_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $gnuplot_warning_menu configure -activeborderwidth 1
  $gnuplot_warning_menu configure -title "warning on non-empty data"

  $gnuplot_warning_menu add command -label "To warn at app closing:" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  $gnuplot_warning_menu add command -label "yes, for non-empty input/data" -command [list "set_warning_on_gnuplot_parts" true]
  $gnuplot_warning_menu add command -label "not for gnuplot parts" -command [list "set_warning_on_gnuplot_parts" false]

  set use_idx 2
  if {$plot_input_warning} {
    set use_idx 1
  }
  $gnuplot_warning_menu entryconfigure $use_idx -background $menu_selected_background -activebackground "#e0e0ff"

  tk_popup $gnuplot_warning_menu $use_x $use_y
}

proc get_plot_processed_command {} {
  set command_text_info [take_plot_command]
  if {([lindex $command_text_info 0]==false) || ([lindex $command_text_info 1]=="")} {
    return [list false ""]
  }
  set command_text [lindex $command_text_info 1]

  set command_text_info [insert_data_into_plot_command $command_text]
  if {([lindex $command_text_info 0]==false) || ([lindex $command_text_info 1]=="")} {
    return [list false ""]
  }
  set command_text [lindex $command_text_info 1]

  if {[string range $command_text "end" "end"] != "\n"} {
    append command_text "\n"
  }

  set command_text [strcat "# processed Gnuplot input\n" $command_text]

  return [list true $command_text]
}

set last_dir_gnuplot ""
set save_path_gnuplot ""
proc export_plot_input_into_file {} {
  global last_dir
  global last_dir_gnuplot
  global save_path
  global save_path_gnuplot
  global spec_file_dialogs
  global plot_output_field

  set command_text_info [get_plot_processed_command]
  if {([lindex $command_text_info 0]==false) || ([lindex $command_text_info 1]=="")} {
    return
  }
  set command_text [lindex $command_text_info 1]

  set use_last_dir $last_dir_gnuplot
  if {$use_last_dir==""} {
    set use_last_dir $last_dir
  }

  if {$save_path_gnuplot!=""} {
    if {[file dirname $save_path_gnuplot]!=$use_last_dir} {
      set save_path_gnuplot ""
    }
  }
  if {$save_path_gnuplot==""} {
    if {$save_path!=""} {
      set save_path_gnuplot [strcat [file rootname $save_path] ".gnuplot"]
      #if {![file exists $save_path_gnuplot]} {
      #  set save_path_gnuplot ""
      #}
    }
  }

  set initial_dir $use_last_dir
  if {$save_path_gnuplot!=""} {
    set initial_dir [file dirname $save_path_gnuplot]
  }

  # not sure whether to try to use tag-based suggested export name here;
  set initial_file ""
  if {$save_path_gnuplot!=""} {
    set initial_file [file tail $save_path_gnuplot]
  }

  set exp_label "Export Gnuplot commands"
  set save_path_tmp ""
  if {$initial_file!=""} {
    set use_common_save_dialog true
    if {$spec_file_dialogs(save) != ""} {
      set use_common_save_dialog false
      set bad [catch {
        set save_path_tmp [$spec_file_dialogs(save) -details true -initialdir $initial_dir -confirmoverwrite true \
          -initialfile $initial_file -defaultextension ".gnuplot" -title $exp_label]
      }]
      if {$bad} {
        set use_common_save_dialog true
        set save_path_tmp ""
      }
    }
    if {$use_common_save_dialog} {
      set save_path_tmp [tk_getSaveFile -initialdir $initial_dir -confirmoverwrite true \
        -initialfile $initial_file -defaultextension ".gnuplot" -title $exp_label]
    }
  } {
    set use_common_save_dialog true
    if {$spec_file_dialogs(save) != ""} {
      set use_common_save_dialog false
      set bad [catch {
        set save_path_tmp [$spec_file_dialogs(save) -details true -initialdir $initial_dir -confirmoverwrite true \
          -defaultextension ".gnuplot" -title $exp_label]
      }]
      if {$bad} {
        set use_common_save_dialog true
        set save_path_tmp ""
      }
    }
    if {$use_common_save_dialog} {
      set save_path_tmp [tk_getSaveFile -initialdir $initial_dir -confirmoverwrite true \
        -defaultextension ".gnuplot" -title $exp_label]
    }
  }

  if {$save_path_tmp==""} {
    return
  }
  set save_path_gnuplot $save_path_tmp

  if {[catch {
    set fh [open $save_path_gnuplot "w"]
  }]} {
    show_plot_field_message_error [list [list "can not open export file for writing"]]
    set save_path_gnuplot ""
    return
  }
  if {[catch {
    puts $fh $command_text
  }]} {
    show_plot_field_message_error [list [list "can not write into the export file"]]
    set save_path_gnuplot ""
    return
  }
  catch {close $fh}

  set last_dir_tmp [file dirname $save_path_gnuplot]
  if {$last_dir_tmp==""} {
    set last_dir_gnuplot ""
  } {
    if {$last_dir_tmp==$last_dir} {
      set last_dir_gnuplot ""
    } {
      set last_dir_gnuplot $last_dir_tmp
    }
  }

  set output_state [$plot_output_field cget -state]
  $plot_output_field configure -state "normal"
  $plot_output_field insert "end" "Gnuplot commands exported into:\n$save_path_gnuplot\n" [list]
  $plot_output_field configure -state $output_state

  present_gnuplot_output
}

set plot_exported_command_tag_name "plot_exported_command_tag"
proc export_plot_input_into_output {to_overwrite} {
  global plot_output_field
  global plot_exported_command_tag_name

  set command_text_info [get_plot_processed_command]
  if {([lindex $command_text_info 0]==false) || ([lindex $command_text_info 1]=="")} {
    return
  }
  set command_text [lindex $command_text_info 1]

  set output_state [$plot_output_field cget -state]
  $plot_output_field configure -state "normal"
  if {!$to_overwrite} {
    if {[lindex [split [$plot_output_field index "end"] "."] 1] != 0} {
      $plot_output_field insert "end" "\n" [list]
    }
    $plot_output_field insert "end" $command_text [list $plot_exported_command_tag_name]
    $plot_output_field see "end"
  } {
    $plot_output_field replace "1.0" "end" $command_text [list $plot_exported_command_tag_name]
    $plot_output_field see "1.0"
  }
  $plot_output_field configure -state $output_state

  present_gnuplot_output
}

proc copy_plot_input_into_text {with_input with_data input_first at_end} {
  global plot_input_field
  global plot_data_field
  global text_field

  if {![is_text_field_active]} {
    show_plot_field_message_error [list [list "main text locked"]]
    return
  }

  set has_input false
  if {$with_input} {
    if {[$plot_input_field index "end -1 chars"]!="1.0"} {
      set has_input true
    }
  }
  set has_data false
  if {$with_data} {
    if {[$plot_data_field index "end -1 chars"]!="1.0"} {
      set has_data true
    }
  }
  if {(!$has_input) && (!$has_data)} {
    show_plot_field_message_error [list [list "no content to copy"]]
    return
  }

  set copy_input_part ""
  if {$has_input} {
    set input_header "# Gnuplot input content"
    if {[$plot_input_field get "1.0" "1.end"]!=$input_header} {
      append copy_input_part $input_header "\n"
    }
    append copy_input_part [$plot_input_field get "1.0" "end -1 chars"]
    if {[string range $copy_input_part "end" "end"] != "\n"} {
      append copy_input_part "\n"
    }
  }

  set copy_data_part ""
  if {$has_data} {
    set data_header "# Gnuplot data content"
    if {[$plot_data_field get "1.0" "1.end"]!=$data_header} {
      append copy_data_part $data_header "\n"
    }
    append copy_data_part [$plot_data_field get "1.0" "end -1 chars"]
    if {[string range $copy_data_part "end" "end"] != "\n"} {
      append copy_data_part "\n"
    }
  }

  set copy_content [expr { \
    $input_first ? [strcat $copy_input_part $copy_data_part] : [strcat $copy_data_part $copy_input_part] \
  }]

  if {!$at_end} {
    set insert_info [prepare_cursor_wise_insertion_point_other false]
  } {
    set insert_info [prepare_end_wise_insertion_point_other]
  }
  set insert_position [lindex $insert_info 0]
  set prepare_insert_info [lindex $insert_info 1]
  if {[llength $prepare_insert_info]==2} {
    $text_field edit separator
    $text_field insert [lindex $prepare_insert_info 0] [lindex $prepare_insert_info 1]
  }
  set real_index [$text_field index [expr {$insert_position=="end" ? "end -1 chars" : $insert_position}]]
  $text_field edit separator
  $text_field insert $insert_position $copy_content
  $text_field edit separator
  $text_field mark set "insert" $real_index
  $text_field see $real_index

  update_text_properties
}

proc set_plot_input_select_mode {do_first do_then} {
  global test_plot_input_select_first
  global test_plot_input_whole_then

  set test_plot_input_select_first $do_first
  set test_plot_input_whole_then $do_then
}

set gnuplot_input_menu ""
proc offer_gnuplot_input_menu {x_abs y_abs x_pos y_pos {with_mode true}} {
  global topw
  global plot_input_field
  global plot_output_field
  global gnuplot_input_menu
  global test_plot_input_select_first
  global test_plot_input_whole_then
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  set is_empty [expr {[$plot_input_field index "end -1 chars"] == "1.0"}]
  if {$is_empty && (!$with_mode)} {
    return
  }

  if {($x_abs=="") || ($y_abs=="")} {
    set abs_pos [get_gnuplot_field_position]
    set use_x [expr {$x_pos + [lindex $abs_pos 0] + $default_menu_offset_x}]
    set use_y [expr {$y_pos + [lindex $abs_pos 1] + $default_menu_offset_y}]
  } {
    set use_x [expr {$x_abs + $default_menu_offset_x}]
    set use_y [expr {$y_abs + $default_menu_offset_y}]
  }

  if {$gnuplot_input_menu!=""} {
    catch {destroy $gnuplot_input_menu}
  }
  set gnuplot_input_menu [menu "$topw.gnuplot_input_menu" -tearoff false]
  $gnuplot_input_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $gnuplot_input_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $gnuplot_input_menu configure -activeborderwidth 1
  $gnuplot_input_menu configure -title "input commands"

  if {$with_mode} {
    set menu_label "Usage of text selections for (export and) input to Gnuplot:"
    $gnuplot_input_menu add command -label $menu_label -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

    $gnuplot_input_menu add command -label "to only use selected regions" -command [list "set_plot_input_select_mode" true false]
    $gnuplot_input_menu add command -label "selected regions if any, otherwise the whole input" -command [list "set_plot_input_select_mode" true true]
    $gnuplot_input_menu add command -label "to only use the whole input" -command [list "set_plot_input_select_mode" false true]

    set idx_offset 0
    set cur_idx [expr {2 + $idx_offset}]
    if {!$test_plot_input_select_first} {
      incr cur_idx 1
    }
    if {!$test_plot_input_whole_then} {
      incr cur_idx -1
    }
    $gnuplot_input_menu entryconfigure $cur_idx -background $menu_selected_background -activebackground "#e0e0ff"
  }

  if {!$is_empty} {
    set menu_label "Export processed (with inserted data) input (part, "
    if {$with_mode} {
      append menu_label "see above"
    } {
      append menu_label "according to selections-use mode"
    }
    append menu_label "):"

    $gnuplot_input_menu add command -label $menu_label -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
    $gnuplot_input_menu add command -label "into a file" -command [list "export_plot_input_into_file"]
    if {[$plot_output_field index "end -1 chars"] == "1.0"} {
      $gnuplot_input_menu add command -label "into output" -command [list "export_plot_input_into_output" false]
    } {
      $gnuplot_input_menu add command -label "into output, append it by the end" -command [list "export_plot_input_into_output" false]
      $gnuplot_input_menu add command -label "into output, overwrite the content" -command [list "export_plot_input_into_output" true]
    }

    if {[is_text_field_active]} {
      set menu_label "Copy verbatim input (the whole content) into the main text:"
      $gnuplot_input_menu add command -label $menu_label -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
      $gnuplot_input_menu add command -label "without data, put by cursor" -command [list "copy_plot_input_into_text" true false true false]
      $gnuplot_input_menu add command -label "without data, put at the end" -command [list "copy_plot_input_into_text" true false true true]
      $gnuplot_input_menu add command -label "followed by data, put by cursor" -command [list "copy_plot_input_into_text" true true true false]
      $gnuplot_input_menu add command -label "followed by data, put at the end" -command [list "copy_plot_input_into_text" true true true true]
    }
  }

  tk_popup $gnuplot_input_menu $use_x $use_y
}

proc show_plot_data_help_at_output {to_overwrite} {
  global plot_output_field
  global plot_data_info_tag_name

  set help_text_list [list \
    "Next embedded data forms are supported, each of them works by replacing a @@DX label" \
    "that is located as a single word on a line, i.e. not among other strings:" \
    "" \
    "* inline data: put @@DI in the input part right after plot commands" \
    "  ...this is for the data ended by 'e' (which is put in by the replacer)" \
    "" \
    "* data blocks, since Gnuplot 5.0: put @@DB anywhere above the data usage" \
    "  ...then \$DB are the named data (delimited alike '<< EOD') for use" \
    "" \
    "* data words: put @@DW anywhere above the data usage" \
    "  ...then DWRn variables (n from 1 to DWRC) contain the data," \
    "     and DWR0 contains names of the DWRn variables;" \
    "  DWRn hold strings where individual columns can be accessed by the \"word\" function," \
    "  DWRS(r) function returns sizes of particular rows," \
    "  and DWC(r,c) function is provided as a helper for data access, use can be alike:" \
    "" \
    "plot sample \\" \
    "    \[x=1:DWRC:1\] '+' using (\$1):(0+DWC(\$1,col)) title \"column \".col with lines, \\" \
    "    \[x=1:DWRC:1\] '+' using (\$1):(0+DWC(\$1,col)):(DWC(\$1,col)) notitle with labels" \
    "" \
    "  where col is a user-defined variable for the used column rank;" \
    "  notice that DWCD(r,c,d) function returns d if r and/or c is out of ranges;" \
    "" \
    "* CSV matrices, since Gnuplot 5.2: put @@DA, or @@DA; (\",\" \";\" are the separators)" \
    "  anywhere above the data usage" \
    "  ...then DARn (n from 1 to DARC) are arrays with the data," \
    "     and DAR0, DARS are arrays with the DARn names, DARn array sizes, respectively;" \
    "  DAC(r,c), DACD(r,c,d) functions are provided alike at the words-based case;" \
    "" \
    "Notice that regarding the DW/DA forms, both rows and columns are counted from 1." \
    "" \
  ]
  set help_text [strcat [join $help_text_list "\n"] "\n"]

  set output_state [$plot_output_field cget -state]
  $plot_output_field configure -state "normal"
  if {!$to_overwrite} {
    if {[lindex [split [$plot_output_field index "end -1 chars"] "."] 1] != 0} {
      $plot_output_field insert "end" "\n" [list]
    }
    if {[lindex [split [$plot_output_field index "end -2 chars"] "."] 1] != 0} {
      $plot_output_field insert "end" "\n"
    }
    $plot_output_field insert "end" $help_text [list $plot_data_info_tag_name]
    $plot_output_field see "end"
  } {
    $plot_output_field replace "1.0" "end" $help_text [list $plot_data_info_tag_name]
    $plot_output_field see "1.0"
  }
  $plot_output_field configure -state $output_state

  present_gnuplot_output
}

proc offer_gnuplot_input_copy_menu_on_key {} {
  global plot_input_field

  set is_empty [expr {[$plot_input_field index "end -1 chars"] == "1.0"}]
  if {$is_empty} {
    return
  }

  set pos_list [$plot_input_field bbox "insert"]
  if {[llength $pos_list]==0} {
    set pos_list [list 0 0]
  }

  offer_gnuplot_input_menu "" "" [lindex $pos_list 0] [lindex $pos_list 1] false
}

proc offer_gnuplot_data_copy_menu_on_key {} {
  global plot_data_field

  set is_empty [expr {[$plot_data_field index "end -1 chars"] == "1.0"}]
  if {$is_empty} {
    return
  }
  if {![is_text_field_active]} {
    return
  }

  set pos_list [$plot_data_field bbox "insert"]
  if {[llength $pos_list]==0} {
    set pos_list [list 0 0]
  }

  offer_gnuplot_data_menu "" "" [lindex $pos_list 0] [lindex $pos_list 1] false
}

set gnuplot_data_menu ""
proc offer_gnuplot_data_menu {x_abs y_abs x_pos y_pos {with_help true}} {
  global topw
  global plot_data_field
  global plot_output_field
  global gnuplot_data_menu
  global default_menu_offset_x
  global default_menu_offset_y

  set is_empty [expr {[$plot_data_field index "end -1 chars"] == "1.0"}]
  if {(!$with_help) && ((![is_text_field_active]) || $is_empty)} {
    return
  }

  if {($x_abs=="") || ($y_abs=="")} {
    set abs_pos [get_gnuplot_field_position]
    set use_x [expr {$x_pos + [lindex $abs_pos 0] + $default_menu_offset_x}]
    set use_y [expr {$y_pos + [lindex $abs_pos 1] + $default_menu_offset_y}]
  } {
    set use_x [expr {$x_abs + $default_menu_offset_x}]
    set use_y [expr {$y_abs + $default_menu_offset_y}]
  }

  if {$gnuplot_data_menu!=""} {
    catch {destroy $gnuplot_data_menu}
  }
  set gnuplot_data_menu [menu "$topw.gnuplot_data_menu" -tearoff false]
  $gnuplot_data_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $gnuplot_data_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $gnuplot_data_menu configure -activeborderwidth 1
  $gnuplot_data_menu configure -title "input commands"

  if {$with_help} {
    set menu_label "Show help on data usage:"
    $gnuplot_data_menu add command -label $menu_label -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
    if {[$plot_output_field index "end -1 chars"] == "1.0"} {
      $gnuplot_data_menu add command -label "at output" -command [list "show_plot_data_help_at_output" false]
    } {
      $gnuplot_data_menu add command -label "at output, append it by the end" -command [list "show_plot_data_help_at_output" false]
      $gnuplot_data_menu add command -label "at output, overwrite the content" -command [list "show_plot_data_help_at_output" true]
    }
  }

  if {[is_text_field_active] && (!$is_empty)} {
    set menu_label "Copy verbatim data (the whole content) into the main text:"
    $gnuplot_data_menu add command -label $menu_label -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
    $gnuplot_data_menu add command -label "without input, put by cursor" -command [list "copy_plot_input_into_text" false true false false]
    $gnuplot_data_menu add command -label "without input, put at the end" -command [list "copy_plot_input_into_text" false true false true]
    $gnuplot_data_menu add command -label "followed by input, put by cursor" -command [list "copy_plot_input_into_text" true true false false]
    $gnuplot_data_menu add command -label "followed by input, put at the end" -command [list "copy_plot_input_into_text" true true false true]
  }

  tk_popup $gnuplot_data_menu $use_x $use_y
}

set gnuplot_output_menu ""
proc offer_gnuplot_output_menu {x_abs y_abs x_pos y_pos} {
  global topw
  global plot_output_field
  global gnuplot_output_menu
  global default_menu_offset_x
  global default_menu_offset_y

  if {($x_abs=="") || ($y_abs=="")} {
    set abs_pos [get_gnuplot_field_position]
    set use_x [expr {$x_pos + [lindex $abs_pos 0] + $default_menu_offset_x}]
    set use_y [expr {$y_pos + [lindex $abs_pos 1] + $default_menu_offset_y}]
  } {
    set use_x [expr {$x_abs + $default_menu_offset_x}]
    set use_y [expr {$y_abs + $default_menu_offset_y}]
  }

  if {$gnuplot_output_menu!=""} {
    catch {destroy $gnuplot_output_menu}
  }
  set gnuplot_output_menu [menu "$topw.gnuplot_output_menu" -tearoff false]
  $gnuplot_output_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $gnuplot_output_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $gnuplot_output_menu configure -activeborderwidth 1
  $gnuplot_output_menu configure -title "output commands"

  $gnuplot_output_menu add command -label "Actions on output text:" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  $gnuplot_output_menu add command -label "sweep the content" -command [list "present_gnuplot_sweep"]

  tk_popup $gnuplot_output_menu $use_x $use_y
}

proc present_gnuplot_sweep {} {
  global plot_output_field
  global lgnuplot_output

  $plot_output_field configure -state "normal"
  $plot_output_field delete "1.0" "end"
  $plot_output_field configure -state "disabled"

  $lgnuplot_output configure -state "disabled"
}

ttk::copyBindings File.TButton CurrentFile.TButton
ttk::style configure CurrentFile.TButton \
  -relief flat

proc present_gnuplot_input {} {
  global plot_input_frame
  global plot_data_frame
  global plot_output_frame
  global lgnuplot_input
  global lgnuplot_data
  global lgnuplot_output
  global plot_input_field
  global plot_shown_part
  global gnuplot_at_data_view
  global gnuplot_at_output_view

  $lgnuplot_input configure -state "disabled" -style CurrentFile.TButton
  $lgnuplot_data configure -state "normal" -style File.TButton
  $lgnuplot_output configure -state [expr {[is_gnuplot_at_output_empty] ? "disabled" : "normal"}] -style File.TButton

  catch {pack forget $plot_output_frame}
  catch {pack forget $plot_data_frame}
  pack $plot_input_frame -side top -expand yes -fill both

  focus $plot_input_field

  set gnuplot_at_data_view false
  set gnuplot_at_output_view false
  set plot_shown_part "input"
  set_gnuplot_title
  update_plot_edited_state
}

set gnuplot_at_data_view false

proc present_gnuplot_data {} {
  global plot_input_frame
  global plot_data_frame
  global plot_output_frame
  global lgnuplot_input
  global lgnuplot_data
  global lgnuplot_output
  global plot_data_field
  global plot_shown_part
  global gnuplot_at_data_view
  global gnuplot_at_output_view

  $lgnuplot_input configure -state "normal" -style File.TButton
  $lgnuplot_data configure -state "disabled" -style CurrentFile.TButton
  $lgnuplot_output configure -state [expr {[is_gnuplot_at_output_empty] ? "disabled" : "normal"}] -style File.TButton

  catch {pack forget $plot_input_frame}
  catch {pack forget $plot_output_frame}
  pack $plot_data_frame -side top -expand yes -fill both

  focus $plot_data_field

  set gnuplot_at_data_view true
  set gnuplot_at_output_view false
  set plot_shown_part "data"
  set_gnuplot_title
  update_plot_edited_state
}

proc present_gnuplot_output {} {
  global plot_input_frame
  global plot_data_frame
  global plot_output_frame
  global lgnuplot_input
  global lgnuplot_data
  global lgnuplot_output
  global plot_output_field
  global plot_shown_part
  global gnuplot_at_data_view
  global gnuplot_at_output_view

  $lgnuplot_input configure -state "normal" -style File.TButton
  $lgnuplot_data configure -state "normal" -style File.TButton
  $lgnuplot_output configure -state "disabled" -style CurrentFile.TButton

  catch {pack forget $plot_input_frame}
  catch {pack forget $plot_data_frame}
  pack $plot_output_frame -side top -expand yes -fill both

  focus $plot_output_field

  set gnuplot_at_data_view false
  set gnuplot_at_output_view true
  set plot_shown_part "output"
  set_gnuplot_title
  update_plot_edited_state
}

proc gnuplot_can_show_input {} {
  return true
}
proc gnuplot_can_show_data {} {
  return true
}
proc gnuplot_can_show_output {} {
  global lgnuplot_output

  if {[string tolower [$lgnuplot_output cget -state]]=="normal"} {
    return true
  }
  return false
}

set gnuplot_about_info_tag_name "gnuplot_about_info_tag"
set plot_data_info_tag_name "plot_data_info_tag"

set plot_data_frame ""
set plot_data_field ""
set plot_menu_holder ""

set lgnuplot_killplot ""
set lgnuplot_plot ""
set lgnuplot_input ""
set lgnuplot_output ""
set lgnuplot_data ""
set lgnuplot_close ""
set lgnuplot_state ""
set lgnuplot_topwise ""
set lgnuplot_focus ""
set lgnuplot_warning ""
proc setup_plot_part {} {
  global plot_put_window
  global scr_width
  global plot_menu_holder
  global plot_put_holder
  global plot_input_frame
  global plot_data_frame
  global plot_output_frame
  global plot_input_field
  global plot_data_field
  global plot_output_field
  global tag_plot_strings
  global tag_plot_comments
  global tag_plot_brackets_pair
  global tag_plot_brackets
  global tag_plot_data_holders
  global spec_base_font
  global command_font
  global lgnuplot_input
  global lgnuplot_output
  global lgnuplot_data
  global lgnuplot_killplot
  global lgnuplot_plot
  global lgnuplot_close
  global lgnuplot_topwise
  global lgnuplot_focus
  global lgnuplot_state
  global lgnuplot_warning
  global gnuplot_at_output_view
  global select_fg_col
  global select_bg_col
  global use_buttons_text
  global plot_exported_command_tag_name
  global plot_data_info_tag_name
  global gnuplot_about_info_tag_name
  global plot_input_shown
  global known_selectforeground_selectbackground_for_tags
  global use_tab_length

  if {$plot_put_window != ""} {
    return
  }

  set plot_put_window [toplevel ".gnuplot_batch_put" -width 600 -height 400]

  set top_pos [get_top_window_position]
  set x_pos [expr {60 + [lindex $top_pos 0]}]
  set y_pos [expr {(($scr_width>1000) ? 40 : 5) + [lindex $top_pos 1]}]
  wm geometry $plot_put_window "+$x_pos+$y_pos"

  set flabels [frame $plot_put_window.flabels -relief sunken -bd 1]
  set ftext_holder [frame $plot_put_window.ftext_holder]
  pack $flabels -side top -expand no -fill x
  pack $ftext_holder -side top -expand yes -fill both
  set plot_put_holder $ftext_holder
  set plot_menu_holder $flabels

  set lgnuplot_input [ttk::button $flabels.lginput -text "Input" -command {present_gnuplot_input} -takefocus 0 -width 6 -style CurrentFile.TButton]
  set lgnuplot_data [ttk::button $flabels.lgdata -text "Data" -command {present_gnuplot_data} -takefocus 0 -width 6 -style File.TButton]
  set lgnuplot_output [ttk::button $flabels.lgoutput -text "Output" -command {present_gnuplot_output} -takefocus 0 -width 6 -style File.TButton]

  set circ_ch ""
  if {!$use_buttons_text} {
    catch {set circ_ch [encoding convertfrom "utf-8" "\xE2\x97\x8F"]}
  }

  set topwise_char [expr {($circ_ch=="") ? "t" : $circ_ch}]
  set lgnuplot_topwise [label $flabels.lgtopwise -text "$topwise_char" -relief "flat" -borderwidth 0 -foreground "#f0f0f0" -background "#808080" -width 3]
  set_gnuplot_topmost_props

  set focus_char [expr {($circ_ch=="") ? "f" : $circ_ch}]
  set lgnuplot_focus [label $flabels.lgfocus -text "$focus_char" -relief "flat" -borderwidth 0 -background "#f0f0f0" -foreground "#808080" -width 3]
  set_gnuplot_focusing_props

  set lgnuplot_plot [ttk::button $flabels.lgplot -text "Plot" -command {exec_plot_command false} -takefocus 0 -width 6 -style File.TButton]
  set lgnuplot_killplot [ttk::button $flabels.lgreplot -text "X/Plot" -command {exec_plot_command true} -takefocus 0 -width 6 -style File.TButton]
  set lgnuplot_close [ttk::button $flabels.lgclose -text "X" -command {present_gnuplot_close} -takefocus 0 -width 2 -style File.TButton]

  pack $lgnuplot_input $lgnuplot_data $lgnuplot_output -side left

  pack $lgnuplot_topwise $lgnuplot_focus -side left -fill y

  pack $lgnuplot_plot $lgnuplot_killplot $lgnuplot_close -side left

  $lgnuplot_input configure -state "disabled"
  $lgnuplot_data configure -state "normal"
  $lgnuplot_output configure -state "disabled"

  $lgnuplot_plot configure -state "disabled"
  $lgnuplot_killplot configure -state "disabled"
  $lgnuplot_close configure -state "disabled"

  set warning_char [expr {($circ_ch=="") ? "m" : $circ_ch}]
  set lgnuplot_warning [label $flabels.lgwarning -text "$warning_char" -relief "flat" -borderwidth 0 \
    -foreground "#c0c0c0" -background "#a0a0a0" -width 3]

  set lgnuplot_about [ttk::button $flabels.lgabout -text "?" -command {present_gnuplot_about} -takefocus 0 -width 2 -style File.TButton]
  pack $lgnuplot_about -side right
  pack $lgnuplot_warning -side right -fill y
  update_gnuplot_warning_label

  set lgnuplot_state [ttk::label $flabels.lgstate -text "" -padding [list 5 0 5 0] -width 8]
  pack $lgnuplot_state -side right

  set ftext_in [frame $ftext_holder.ftext_in]
  set ftext_out [frame $ftext_holder.ftext_out]
  set ftext_data [frame $ftext_holder.ftext_data]
  pack $ftext_in -side top -expand yes -fill both
  set plot_input_frame $ftext_in
  set plot_output_frame $ftext_out
  set plot_data_frame $ftext_data

  set plot_input_field [text $plot_input_frame.plot_input_text -wrap none -undo true -maxundo 10000]
  ttk::scrollbar $plot_input_frame.vscr -orient vertical -command "$plot_input_field yview"
  ttk::scrollbar $plot_input_frame.hscr -orient horizontal -command "$plot_input_field xview"
  $plot_input_field configure -xscroll "$plot_input_frame.hscr set" -yscroll "$plot_input_frame.vscr set"
  ttk::sizegrip $plot_input_frame.size_grip

  grid $plot_input_field $plot_input_frame.vscr -sticky nsew
  grid $plot_input_frame.hscr $plot_input_frame.size_grip -sticky nsew

  grid rowconfigure $plot_input_frame $plot_input_field -weight 1
  grid columnconfigure $plot_input_frame $plot_input_field -weight 1

  if {$spec_base_font(command)==""} {
    $plot_input_field configure -font command_font
  } {
    $plot_input_field configure -font $spec_base_font(command)
  }

  set plot_data_field [text $plot_data_frame.plot_data_text -wrap none -undo true -maxundo 10000]
  ttk::scrollbar $plot_data_frame.vscr -orient vertical -command "$plot_data_field yview"
  ttk::scrollbar $plot_data_frame.hscr -orient horizontal -command "$plot_data_field xview"
  $plot_data_field configure -xscroll "$plot_data_frame.hscr set" -yscroll "$plot_data_frame.vscr set"
  ttk::sizegrip $plot_data_frame.size_grip

  grid $plot_data_field $plot_data_frame.vscr -sticky nsew
  grid $plot_data_frame.hscr $plot_data_frame.size_grip -sticky nsew

  grid rowconfigure $plot_data_frame $plot_data_field -weight 1
  grid columnconfigure $plot_data_frame $plot_data_field -weight 1

  if {$spec_base_font(command)==""} {
    $plot_data_field configure -font command_font
  } {
    $plot_data_field configure -font $spec_base_font(command)
  }

  set plot_output_field [text $plot_output_frame.plot_output_text -wrap none -undo false]
  ttk::scrollbar $plot_output_frame.vscr -orient vertical -command "$plot_output_field yview"
  ttk::scrollbar $plot_output_frame.hscr -orient horizontal -command "$plot_output_field xview"
  $plot_output_field configure -xscroll "$plot_output_frame.hscr set" -yscroll "$plot_output_frame.vscr set"
  ttk::sizegrip $plot_output_frame.size_grip

  grid $plot_output_field $plot_output_frame.vscr -sticky nsew
  grid $plot_output_frame.hscr $plot_output_frame.size_grip -sticky nsew

  grid rowconfigure $plot_output_frame $plot_output_field -weight 1
  grid columnconfigure $plot_output_frame $plot_output_field -weight 1

  if {$spec_base_font(command)==""} {
    $plot_output_field configure -font command_font
  } {
    $plot_output_field configure -font $spec_base_font(command)
  }

  $plot_input_field configure -foreground "#000000"
  $plot_input_field configure -background "#ffffff"
  $plot_input_field configure -selectforeground $select_fg_col
  $plot_input_field configure -selectbackground $select_bg_col
  $plot_input_field configure -inactiveselectbackground $select_bg_col

  $plot_input_field tag configure $tag_plot_strings -foreground "#805020"
  $plot_input_field tag configure $tag_plot_comments -foreground "#604060"
  $plot_input_field tag configure $tag_plot_data_holders -foreground "#105037"
  $plot_input_field tag configure $tag_plot_brackets_pair -foreground "#00a0f0"
  $plot_input_field tag configure $tag_plot_brackets -foreground "#f03000"

  $plot_input_field tag raise "sel"
  $plot_input_field tag raise $tag_plot_strings
  $plot_input_field tag raise $tag_plot_comments
  $plot_input_field tag raise $tag_plot_data_holders
  $plot_input_field tag raise $tag_plot_brackets_pair
  $plot_input_field tag raise $tag_plot_brackets

  if {!$known_selectforeground_selectbackground_for_tags} {
    $plot_input_field tag raise "sel"
  }

  $plot_data_field configure -foreground "#000000"
  $plot_data_field configure -background "#ffffff"
  $plot_data_field configure -selectforeground $select_fg_col
  $plot_data_field configure -selectbackground $select_bg_col
  $plot_data_field configure -inactiveselectbackground $select_bg_col

  $plot_data_field tag raise "sel"

  $plot_output_field configure -foreground "#000000"
  $plot_output_field configure -background "#ffffff"
  $plot_output_field configure -selectforeground $select_fg_col
  $plot_output_field configure -selectbackground $select_bg_col
  $plot_output_field configure -inactiveselectbackground $select_bg_col
  $plot_output_field configure -state "disabled"

  $plot_output_field tag configure $plot_exported_command_tag_name -background "#f0f0ff" -foreground "#101010"
  $plot_output_field tag configure $plot_data_info_tag_name -background "#f0fff0" -foreground "#103020"
  $plot_output_field tag configure $gnuplot_about_info_tag_name -background "#ffffe7" -foreground "#101010"
  $plot_output_field tag raise "sel"

  bind $lgnuplot_topwise <1> {
    remove_torn_menus
    offer_gnuplot_topwise_menu %X %Y %x %y
    break
  }
  bind $lgnuplot_topwise <Control-1> {
    remove_torn_menus
    switch_gnuplot_topmost
    break
  }
  bind $lgnuplot_topwise <Shift-1> {
    break
  }
  bind $lgnuplot_topwise <Alt-1> {
    break
  }

  bind $lgnuplot_focus <1> {
    remove_torn_menus
    offer_gnuplot_focusing_menu %X %Y %x %y
    break
  }
  bind $lgnuplot_focus <Control-1> {
    remove_torn_menus
    switch_gnuplot_focusing
    break
  }
  bind $lgnuplot_focus <Shift-1> {
    break
  }
  bind $lgnuplot_focus <Alt-1> {
    break
  }

  bind $lgnuplot_warning <1> {
    remove_torn_menus
    offer_gnuplot_warning_menu %X %Y %x %y
    break
  }
  bind $lgnuplot_warning <Control-1> {
    break
  }
  bind $lgnuplot_warning <Shift-1> {
    break
  }
  bind $lgnuplot_warning <Alt-1> {
    break
  }

  bind $plot_input_field <FocusIn> {
    $plot_input_field tag configure "sel" -bgstipple ""
    remove_torn_menus
  }
  bind $plot_input_field <FocusOut> {
    $plot_input_field tag configure "sel" -bgstipple gray75
  }

  bind $plot_data_field <FocusIn> {
    $plot_data_field tag configure "sel" -bgstipple ""
    remove_torn_menus
  }
  bind $plot_data_field <FocusOut> {
    $plot_data_field tag configure "sel" -bgstipple gray75
  }

  bind $plot_output_field <FocusIn> {
    $plot_output_field tag configure "sel" -bgstipple ""
    remove_torn_menus
  }
  bind $plot_output_field <FocusOut> {
    $plot_output_field tag configure "sel" -bgstipple gray75
  }

  # amounts of bindings alike for text/command fields
  set other_keys [list "Control-s" "Control-t" "Alt-t" "Control-d" "Alt-d" "Control-e" "Control-E" \
    "Control-k" "Control-o" "Control-B" "Alt-b" "Control-f" "Alt-f" "Control-g"  "Control-h" \
    "Control-p" "Control-n" "Control-l" "Control-j" "Control-comma" "Control-period" "Control-u" \
    "Control-slash" "Control-backslash" "Control-r" "Control-R" "Control-q" "Control-w"]
  foreach keyspressed $other_keys {
    bind $plot_input_field <$keyspressed> {
      break
    }
  }
  foreach keyspressed $other_keys {
    bind $plot_data_field <$keyspressed> {
      break
    }
  }
  foreach keyspressed $other_keys {
    bind $plot_output_field <$keyspressed> {
      break
    }
  }

  bind $plot_input_field <3> {
    offer_gnuplot_input_menu %X %Y %x %y
    break
  }
  bind $plot_input_field <Shift-3> {
    break
  }
  bind $plot_input_field <Control-3> {
    break
  }
  bind $plot_input_field <Alt-3> {
    break
  }

  bind $plot_data_field <3> {
    offer_gnuplot_data_menu %X %Y %x %y
  }
  bind $plot_data_field <Shift-3> {
    break
  }
  bind $plot_data_field <Control-3> {
    break
  }
  bind $plot_data_field <Alt-3> {
    break
  }

  bind $plot_output_field <3> {
    offer_gnuplot_output_menu %X %Y %x %y
  }
  bind $plot_output_field <Shift-3> {
    break
  }
  bind $plot_output_field <Control-3> {
    break
  }
  bind $plot_output_field <Alt-3> {
    break
  }

  bind $plot_input_field <Control-a> {
    $plot_input_field tag remove "sel" "1.0" "end"
    $plot_input_field tag add "sel" "1.0" "end -1 chars"
    $plot_input_field edit separator
    set_anchor_opposite $plot_input_field
    break
  }
  bind $plot_data_field <Control-a> {
    $plot_data_field tag remove "sel" "1.0" "end"
    $plot_data_field tag add "sel" "1.0" "end -1 chars"
    $plot_data_field edit separator
    set_anchor_opposite $plot_data_field
    break
  }
  bind $plot_output_field <Control-a> {
    $plot_output_field tag remove "sel" "1.0" "end"
    $plot_output_field tag add "sel" "1.0" "end -1 chars"
    break
  }

  foreach wdg [list $plot_put_window $plot_input_field $plot_data_field $plot_output_field] {
    bind $wdg <Control-k> {
      present_gnuplot_close
    }
  }

  bind $plot_input_field <Control-x> {
    $plot_input_field edit separator
    set_clipboard $plot_input_field
    store_clipboard_onset
    #tk_textCut $plot_input_field
    delete_selected_text $plot_input_field
    $plot_input_field edit separator
    update_plot_properties
    break
  }
  bind $plot_data_field <Control-x> {
    $plot_data_field edit separator
    set_clipboard $plot_data_field
    store_clipboard_onset
    #tk_textCut $plot_data_field
    delete_selected_text $plot_data_field
    $plot_data_field edit separator
    #update_plot_properties
    break
  }

  bind $plot_input_field <Control-c> {
    $plot_input_field edit separator
    set_clipboard $plot_input_field
    store_clipboard_onset
    $plot_input_field edit separator
    update_plot_properties
    break
  }
  bind $plot_data_field <Control-c> {
    $plot_data_field edit separator
    set_clipboard $plot_data_field
    store_clipboard_onset
    $plot_data_field edit separator
    #update_plot_properties
    break
  }
  bind $plot_output_field <Control-c> {
    set_clipboard $plot_output_field
    store_clipboard_onset
    break
  }

  bind $plot_input_field <Control-V> {
    $plot_input_field edit separator
    tk_textPaste $plot_input_field
    $plot_input_field edit separator
    $plot_input_field see "insert"
    update_plot_properties
    break
  }
  bind $plot_data_field <Control-V> {
    $plot_data_field edit separator
    tk_textPaste $plot_data_field
    $plot_data_field edit separator
    $plot_data_field see "insert"
    #update_plot_properties
    break
  }

  bind $plot_input_field <Control-v> {
    $plot_input_field edit separator
    remove_selected_if_by_cursor $plot_input_field
    $plot_input_field edit separator
    tk_textPaste $plot_input_field
    $plot_input_field edit separator
    $plot_input_field see "insert"
    update_plot_properties
    break
  }
  bind $plot_data_field <Control-v> {
    $plot_data_field edit separator
    remove_selected_if_by_cursor $plot_data_field
    $plot_data_field edit separator
    tk_textPaste $plot_data_field
    $plot_data_field edit separator
    $plot_data_field see "insert"
    #update_plot_properties
    break
  }

  bind $plot_input_field <Control-z> {
    catch {
      $plot_input_field edit undo
      update_plot_properties
    }
    break
  }
  bind $plot_data_field <Control-z> {
    catch {
      $plot_data_field edit undo
      #update_plot_properties
    }
    break
  }

  foreach keyspressed [list "Control-y" "Control-Z"] {
    bind $plot_input_field <$keyspressed> {
      catch {
        $plot_input_field edit redo
        update_plot_properties
      }
      break
    }
  }
  foreach keyspressed [list "Control-y" "Control-Z"] {
    bind $plot_data_field <$keyspressed> {
      catch {
        $plot_data_field edit redo
        #update_plot_properties
      }
      break
    }
  }

  foreach keyspressed [list "Escape"] {
    bind $plot_put_window <$keyspressed> {
      close_plot_part
      break
    }
    bind $plot_input_field <$keyspressed> {
      close_plot_part
      break
    }
    bind $plot_data_field <$keyspressed> {
      close_plot_part
      break
    }
    bind $plot_output_field <$keyspressed> {
      close_plot_part
      break
    }
  }

  # direct feeding of gnuplot: without gnuplot restart
  foreach wdg [list $plot_put_window $plot_input_field $plot_data_field $plot_output_field] {
    bind $wdg <Control-Return> {
      exec_plot_command true
      break
    }
  }
  # direct feeding of gnuplot: with gnuplot restart
  foreach wdg [list $plot_put_window $plot_input_field $plot_data_field $plot_output_field] {
    bind $wdg <Shift-Return> {
      exec_plot_command false
      break
    }
  }
  # testing only
  foreach wdg [list $plot_put_window $plot_input_field $plot_data_field $plot_output_field] {
    bind $wdg <Alt-Return> {
      test_plot_command
      break
    }
  }

  foreach wdg [list $plot_put_window $plot_input_field $plot_data_field $plot_output_field] {
    bind $wdg <Control-Shift-Return> {
      break
    }
  }

  bind $plot_input_field <Prior> {
    topwise_page_up $plot_input_field
    set_anchor_insert $plot_input_field
    update_plot_highlight_lines_onset
  }
  bind $plot_input_field <Shift-Prior> {
  }
  bind $plot_data_field <Prior> {
    topwise_page_up $plot_data_field
    set_anchor_insert $plot_data_field
    #update_plot_highlight_lines_onset
  }
  bind $plot_data_field <Shift-Prior> {
  }

  bind $plot_output_field <Prior> {
    topwise_page_up $plot_output_field
  }
  bind $plot_output_field <Shift-Prior> {
  }

  bind $plot_input_field <Next> {
    bottomwise_page_down $plot_input_field
    set_anchor_insert $plot_input_field
    update_plot_highlight_lines_onset
  }
  bind $plot_input_field <Shift-Next> {
  }
  bind $plot_data_field <Next> {
    bottomwise_page_down $plot_data_field
    set_anchor_insert $plot_data_field
    #update_plot_highlight_lines_onset
  }
  bind $plot_data_field <Shift-Next> {
  }

  bind $plot_output_field <Next> {
    bottomwise_page_down $plot_output_field
  }
  bind $plot_output_field <Shift-Next> {
  }

  bind $plot_input_field <Tab> {
    $plot_input_field edit separator
    $plot_input_field insert "insert" [string repeat " " $use_tab_length]
    $plot_input_field edit separator
    update_plot_properties
    break
  }
  bind $plot_data_field <Tab> {
    $plot_data_field edit separator
    $plot_data_field insert "insert" [string repeat " " $use_tab_length]
    $plot_data_field edit separator
    #update_plot_properties
    break
  }

  # jumping on brackets
  bind $plot_input_field <Control-b> {
    plot_jump_bracket
    break
  }
  bind $plot_data_field <Control-b> {
    break
  }
  bind $plot_input_field <Control-B> {
    plot_jump_bracket true
    break
  }
  bind $plot_data_field <Control-B> {
    break
  }

  bind $plot_input_field <Control-w> {
    offer_gnuplot_input_copy_menu_on_key
    break
  }
  bind $plot_data_field <Control-w> {
    offer_gnuplot_data_copy_menu_on_key
    break
  }

  # setting for quit-warning
  bind $plot_input_field <Control-m> {
    set_plot_input_warning
    break
  }
  bind $plot_data_field <Control-m> {
    set_plot_input_warning
    break
  }
  bind $plot_output_field <Control-m> {
    set_plot_input_warning
    break
  }
  # unsetting from quit-warning
  bind $plot_input_field <Control-M> {
    unset_plot_input_warning
    break
  }
  bind $plot_data_field <Control-M> {
    unset_plot_input_warning
    break
  }
  bind $plot_output_field <Control-M> {
    unset_plot_input_warning
    break
  }
  # switching the quit-warning state
  bind $lgnuplot_warning <Control-1> {
    switch_plot_input_warning
    break
  }

  bind $plot_input_field <Control-d> {
    draw_import
    break
  }
  bind $plot_data_field <Control-d> {
    draw_import
    break
  }
  bind $plot_output_field <Control-d> {
    draw_import
    break
  }
  bind $plot_input_field <Control-bar> {
    if {[is_batch_mode]} {
      draw_batch_maxima
    }
    break
  }
  bind $plot_data_field <Control-bar> {
    if {[is_batch_mode]} {
      draw_batch_maxima
    }
    break
  }
  bind $plot_output_field <Control-bar> {
    if {[is_batch_mode]} {
      draw_batch_maxima
    }
    break
  }

  foreach keyspressed [list "Control-e" "Control-Alt-e" "Control-E"] {
    bind $plot_input_field <$keyspressed> {
      present_gnuplot_clean
      break
    }
  }
  foreach keyspressed [list "Control-e" "Control-Alt-e" "Control-E"] {
    bind $plot_data_field <$keyspressed> {
      present_gnuplot_erase_data
      break
    }
  }
  foreach keyspressed [list "Control-e" "Control-Alt-e" "Control-E"] {
    bind $plot_output_field <$keyspressed> {
      present_gnuplot_sweep
      break
    }
  }

  # wrong default placing (at least for left/middle-button) here too
  bind $plot_input_field <2> {
    insert_primary $plot_input_field %X %Y %x %y
    update_plot_properties_delayed
    break
  }
  bind $plot_input_field <ButtonRelease-2> {
    update_plot_properties
    break
  }
  bind $plot_data_field <2> {
    insert_primary $plot_data_field %X %Y %x %y
    #update_plot_properties_delayed
    break
  }
  bind $plot_data_field <ButtonRelease-2> {
    #update_plot_properties
    break
  }

  bind $plot_input_field <1> {
    place_cursor_on_mouse $plot_input_field %X %Y %x %y
    $plot_input_field tag remove "sel" "1.0" "end"
    $plot_input_field edit separator
    set_anchor_insert $plot_input_field
    update_plot_highlight_lines_onset
    break
  }
  bind $plot_input_field <ButtonRelease-1> {
    update_plot_properties
    break
  }
  bind $plot_data_field <1> {
    place_cursor_on_mouse $plot_data_field %X %Y %x %y
    $plot_data_field tag remove "sel" "1.0" "end"
    $plot_data_field edit separator
    set_anchor_insert $plot_data_field
    #update_plot_highlight_lines_onset
    break
  }
  bind $plot_data_field <ButtonRelease-1> {
    #update_plot_properties
    break
  }

  bind $plot_input_field <Control-1> {
    break
  }
  bind $plot_input_field <Control-ButtonRelease-1> {
    break
  }
  bind $plot_data_field <Control-1> {
    break
  }
  bind $plot_data_field <Control-ButtonRelease-1> {
    break
  }

  bind $plot_input_field <Double-1> {
    field_on_double_click $plot_input_field
    break
  }
  bind $plot_input_field <Control-Double-1> {
    set_anchor_insert $plot_input_field
    break
  }
  bind $plot_input_field <Control-Shift-Double-1> {
    set_anchor_insert $plot_input_field
    break
  }
  bind $plot_data_field <Double-1> {
    field_on_double_click $plot_data_field
    break
  }
  bind $plot_data_field <Control-Double-1> {
    set_anchor_insert $plot_data_field
    break
  }
  bind $plot_data_field <Control-Shift-Double-1> {
    set_anchor_insert $plot_data_field
    break
  }

  foreach keyspressed [list "Triple-1" "Quadruple-1"] {
    bind $plot_input_field <$keyspressed> {
      field_on_multi_click $plot_input_field
      break
    }
  }
  foreach keyspressed [list "Control-Triple-1" "Control-Quadruple-1"] {
    bind $plot_input_field <$keyspressed> {
      break
    }
  }
  foreach keyspressed [list "Triple-1" "Quadruple-1"] {
    bind $plot_data_field <$keyspressed> {
      field_on_multi_click $plot_data_field
      break
    }
  }
  foreach keyspressed [list "Control-Triple-1" "Control-Quadruple-1"] {
    bind $plot_data_field <$keyspressed> {
      break
    }
  }

  bind $plot_output_field <Double-1> {
    field_on_double_click $plot_output_field
    break
  }
  bind $plot_output_field <Control-Double-1> {
    break
  }
  bind $plot_output_field <Control-Shift-Double-1> {
    break
  }
  foreach keyspressed [list "Triple-1" "Quadruple-1"] {
    bind $plot_output_field <$keyspressed> {
      field_on_multi_click $plot_output_field
      break
    }
  }
  foreach keyspressed [list "Control-Triple-1" "Control-Quadruple-1"] {
    bind $plot_output_field <$keyspressed> {
      break
    }
  }

  foreach wdg [list $plot_put_window $plot_input_field $plot_data_field $plot_output_field] {
    bind $wdg <Control-i> {
      close_gnuplot_direct
      break
    }
    bind $wdg <Control-I> {
      close_gnuplot_direct
      break
    }
  }

  bind $plot_input_field <Control-comma> {
    break
  }
  bind $plot_input_field <Control-period> {
    if {[gnuplot_can_show_data]} {
      present_gnuplot_data
    }
    break
  }
  bind $plot_data_field <Control-comma> {
    if {[gnuplot_can_show_input]} {
      present_gnuplot_input
    }
    break
  }
  bind $plot_data_field <Control-period> {
    if {[gnuplot_can_show_output]} {
      present_gnuplot_output
    }
    break
  }
  bind $plot_output_field <Control-comma> {
    if {[gnuplot_can_show_data]} {
      present_gnuplot_data
    }
    break
  }
  bind $plot_output_field <Control-period> {
    break
  }

  bind $plot_input_field <<Modified>> {
    update_plot_edited_state
    break
  }

  # binding highlight updates to general actions
  # can not use <KeyRelease> here, since numerical state of keypad gives non-numerical
  # (that is KP_Left, etc.) events for the KeyRelease event types.
  bind $plot_input_field <KeyPress> "+{update_plot_properties_delayed}"
  bind $plot_input_field <ButtonRelease> "+{update_plot_properties_delayed}"

  bind_keypad $plot_put_window
  bind_keypad $plot_input_field
  bind_keypad $plot_data_field
  bind_keypad $plot_output_field

  bind $plot_input_field <KeyPress-Home> {
    home_key_reasonable $plot_input_field
    update_plot_bracket_pairing
    break
  }
  bind $plot_input_field <Control-KeyPress-Home> {
    after 0 "update_plot_highlight_lines"
    after 0 "update_plot_bracket_pairing"
  }
  bind $plot_input_field <Shift-KeyPress-Home> {
    after 0 "update_plot_highlight_lines"
    after 0 "update_plot_bracket_pairing"
  }
  bind $plot_input_field <Alt-KeyPress-Home> {
    after 0 "update_plot_highlight_lines"
    after 0 "update_plot_bracket_pairing"
  }

  bind $plot_data_field <KeyPress-Home> {
    home_key_reasonable $plot_data_field
    break
  }
  bind $plot_data_field <Control-KeyPress-Home> {
  }
  bind $plot_data_field <Shift-KeyPress-Home> {
  }
  bind $plot_data_field <Alt-KeyPress-Home> {
  }

  bind $plot_output_field <KeyPress-Home> {
    home_key_reasonable $plot_output_field
    break
  }
  bind $plot_output_field <Control-KeyPress-Home> {
  }
  bind $plot_output_field <Shift-KeyPress-Home> {
  }
  bind $plot_output_field <Alt-KeyPress-Home> {
  }

  set plot_input_shown true
}

set plot_put_window_last_geom ""
proc close_plot_part {{force false}} {
  global plot_put_window
  global plot_input_shown
  global plot_put_window_last_geom

  if {(!$force) && (!$plot_input_shown)} {
    return
  }
  set plot_input_shown false

  if {$plot_input_shown} {
    set test_geom ""
    if {![catch {
      set test_geom [wm geometry $plot_put_window]
    }]} {
      set plot_put_window_last_geom $test_geom
    }
  }

  catch {
    wm withdraw $plot_put_window
  }
}

set plot_shown_part "input"
set plot_input_shown false
proc show_plot_part {} {
  global plot_put_window
  global plot_input_shown
  global plot_put_window_last_geom
  global plot_shown_part
  global gnuplot_keep_always_topmost

  catch {wm manage $plot_put_window}

  if {!$plot_input_shown} {
    if {$plot_put_window_last_geom != ""} {
      wm geometry $plot_put_window $plot_put_window_last_geom
    }
  }
  set plot_input_shown true

  wm protocol $plot_put_window WM_DELETE_WINDOW {close_plot_part}
  wm deiconify $plot_put_window

  set_gnuplot_title
  wm minsize $plot_put_window 600 300
  wm resizable $plot_put_window true true

  wm attributes $plot_put_window -topmost $gnuplot_keep_always_topmost

  raise $plot_put_window
}

proc open_plot_part {{wdg ""} {ins_at_end false}} {
  global plot_put_window
  global plot_input_field
  global plot_data_field
  global plot_output_field
  global gnuplot_at_data_view
  global gnuplot_at_output_view

  if {$plot_put_window==""} {
    setup_plot_part
  }
  show_plot_part
  update_plot_input_warning

  set command_text ""
  if {$wdg!=""} {
    set command_text [list]
    foreach {sel_start sel_after} [$wdg tag ranges "sel"] {
      lappend command_text [$wdg get $sel_start $sel_after]
    }
    set command_text [join $command_text "\n"]
  }

  set to_switch_to_input false
  if {$command_text!=""} {
    set to_switch_to_input true
    if {[string range $command_text end end]!="\n"} {
      append command_text "\n"
    }
    if {$ins_at_end} {
      set end_pos [split [$plot_input_field index "end -1 chars"] "."]
      if {[lindex $end_pos 1]!=0} {
        $plot_input_field insert "end" "\n"
      }
      $plot_input_field insert "end" $command_text
      $plot_input_field mark set "insert" "end"
      $plot_input_field see "end linestart"
    } {
      set ins_pos [split [$plot_input_field index "insert"] "."]
      set place_row [lindex $ins_pos 0]
      if {[lindex $ins_pos 1]!=0} {
        incr place_row
      }
      set place_idx "$place_row.0"
      $plot_input_field insert $place_idx $command_text
      $plot_input_field mark set "insert" $place_idx
      $plot_input_field yview end
      $plot_input_field see $place_idx
    }
  }

  if {$to_switch_to_input} {
    if {$gnuplot_at_data_view || $gnuplot_at_output_view} {
      present_gnuplot_input
    }
  }

  set field_to_focus $plot_input_field
  if {$gnuplot_at_data_view} {
    set field_to_focus $plot_data_field
  }
  if {$gnuplot_at_output_view} {
    set field_to_focus $plot_output_field
  }

  focus $field_to_focus
  return
}

set gnuplot_version_read [list]
proc present_gnuplot_about {} {
  global gnuplot_version_read
  global plot_output_field
  global lgnuplot_output
  global gnuplot_at_output_view
  global gnuplot_about_info_tag_name

  $plot_output_field configure -state "normal"
  if {[lindex [split [$plot_output_field index "end -1 chars"] "."] 1] != 0} {
    $plot_output_field insert "end" "\n"
  }
  if {[lindex [split [$plot_output_field index "end -2 chars"] "."] 1] != 0} {
    $plot_output_field insert "end" "\n"
  }
  $plot_output_field insert "end" "Shift+Enter for executing Gnuplot commands from the \"Input\" field\n" [list $gnuplot_about_info_tag_name]
  $plot_output_field insert "end" "Control+Enter for restarting Gnuplot and executing the commands.\n" [list $gnuplot_about_info_tag_name]
  $plot_output_field insert "end" "Control+r to open this window, Control+comma/period to slide within.\n" [list $gnuplot_about_info_tag_name]
  if {[llength $gnuplot_version_read]==0} {
    $plot_output_field insert "end" "(Gnuplot version is read once Gnuplot is started)" [list $gnuplot_about_info_tag_name]
  } {
    $plot_output_field insert "end" "Gnuplot info:\n" [list $gnuplot_about_info_tag_name]
    $plot_output_field insert "end" [join $gnuplot_version_read "\n"] [list $gnuplot_about_info_tag_name]
  }
  $plot_output_field insert "end" "\n\n" [list $gnuplot_about_info_tag_name]
  $plot_output_field yview end
  $plot_output_field configure -state "disabled"
  $lgnuplot_output configure -state "normal"

  if {!$gnuplot_at_output_view} {
    present_gnuplot_output
  }
}

proc consume_gnuplot_onset {} {
  global gnuplot_version_read
  global gnuplot_onset_read
  global gnuplot_shibboleth_onset
  global gnuplot_current_pid
  global gnuplot_version_read_pid_prefix

  set gnuplot_version_read [list]

  foreach one_line $gnuplot_onset_read {
    set one_line [string trim $one_line]
    if {[string first "version" [string tolower $one_line]]>-1} {
      set version_line [list]
      foreach version_item [split $one_line] {
        set version_item [string trim $version_item]
        if {$version_item==""} {
          continue
        }
        lappend version_line $version_item
      }
      lappend gnuplot_version_read [join $version_line " "]
    }
    if {[string first $gnuplot_shibboleth_onset $one_line]>-1} {
      break
    }
  }

  if {$gnuplot_current_pid != ""} {
    lappend gnuplot_version_read "$gnuplot_version_read_pid_prefix $gnuplot_current_pid"
  }

  set gnuplot_onset_read [list]
}

set gnuplot_handle ""
set gnuplot_checks_id ""
proc setup_gnuplot_checks {} {
  global gnuplot_handle
  global gnuplot_checks_id

  if {$gnuplot_checks_id!=""} {
    catch {after cancel $gnuplot_checks_id}
  }
  set gnuplot_checks_id ""

  if {$gnuplot_handle!=""} {
    set gnuplot_checks_id [after 1000 check_gnuplot_connected_after]
  }
}

proc check_gnuplot_connected_after {} {
  global gnuplot_checks_id

  if {$gnuplot_checks_id!=""} {
    catch {after cancel $gnuplot_checks_id}
  }
  set gnuplot_checks_id ""

  set is_connected [check_gnuplot_connected]

  if {$is_connected} {
    set gnuplot_checks_id [after 0 setup_gnuplot_checks]
  }
}

proc check_gnuplot_connected {{soft_check false}} {
  global gnuplot_handle
  if {$gnuplot_handle == ""} {
    return false
  }

  set is_connected false
  if {$gnuplot_handle != ""} {
    if {[llength [file channels $gnuplot_handle]]>0} {
      set is_connected true
    }
    set at_eof 0
    if {$is_connected} {
      if {[catch {set at_eof [eof $gnuplot_handle]}]} {
        set is_connected false
      }
    }
    if {$at_eof} {
      set is_connected false
    }
  }

  if {(!$is_connected) && (!$soft_check)} {
    set_gnuplot_disconnect
  }

  return $is_connected
}

proc set_gnuplot_disconnect {} {
  close_gnuplot_direct
}

proc dispatch_has_read_from_gnuplot {} {
  global gnuplot_got_new_command
  global text_field
  global topw

  if {!$gnuplot_got_new_command} {
    return
  }
  set gnuplot_got_new_command false

  if {([focus] == $text_field) || (([focus] == "") && ([focus -lastfor $text_field] == $text_field))} {
    catch {raise $topw}
    show_text_field_message_error [list [list "output from gnuplot" "gnuplot-output"]]
  }
}

set gnuplot_onset_read [list]
proc read_from_gnuplot {} {
  global gnuplot_handle
  global plot_output_field
  global lgnuplot_output
  global gnuplot_just_started_read
  global gnuplot_shibboleth_onset
  global gnuplot_onset_read

  catch {
    fileevent $gnuplot_handle readable ""
  }

  if {![check_gnuplot_connected]} {
    return
  }

  set read_string [chan read $gnuplot_handle]
  # this if it would work correctly under wine-like systems
  set read_string [string map [list "\r" ""] $read_string]

  if {$gnuplot_just_started_read} {
    set read_string_use [list]
    foreach one_read_line [split $read_string "\n"] {
      if {$gnuplot_just_started_read} {
        lappend gnuplot_onset_read $one_read_line
        if {[string first $gnuplot_shibboleth_onset $gnuplot_onset_read]>-1} {
          set gnuplot_just_started_read false
        }
      } {
        lappend read_string_use $one_read_line
      }
    }
    set read_string [join $read_string_use "\n"]

    if {!$gnuplot_just_started_read} {
      consume_gnuplot_onset
    }
  }

  if {$read_string != ""} {
    $plot_output_field configure -state "normal"
    $plot_output_field insert "end" $read_string [list]
    $plot_output_field yview end
    $plot_output_field configure -state "disabled"
    $lgnuplot_output configure -state "normal"
    dispatch_has_read_from_gnuplot
  }

  catch {
    fileevent $gnuplot_handle readable "read_from_gnuplot"
  }
}

set gnuplot_output_check_id ""
set gnuplot_output_path_last_int_start 10
set gnuplot_output_path_last_int $gnuplot_output_path_last_int_start
proc check_gnuplot_output {} {
  global gnuplot_output_id
  global gnuplot_output_check_id
  global gnuplot_output_path_last_int
  global plot_output_field
  global lgnuplot_output
  global gnuplot_just_started_read
  global gnuplot_shibboleth_onset
  global gnuplot_onset_read

  catch {after cancel $gnuplot_output_check_id}
  set gnuplot_output_check_id ""

  set read_data_max 10000
  set read_data_cur ""
  if {[catch {
    set read_data_cur [read $gnuplot_output_id $read_data_max]
  }]} {
    set read_data_cur ""
  }

  # skipping the output if it is too long;
  # we read at most $read_data_max per $gnuplot_output_path_last_int
  if {[string length $read_data_cur]==$read_data_max} {
    catch {
      seek $gnuplot_output_id 0 "end"
    }
  }

  if {$gnuplot_just_started_read} {
    set read_string_use [list]
    foreach one_read_line [split $read_data_cur "\n"] {
      if {$gnuplot_just_started_read} {
        lappend gnuplot_onset_read $one_read_line
        if {[string first $gnuplot_shibboleth_onset $gnuplot_onset_read]>-1} {
          set gnuplot_just_started_read false
        }
      } {
        lappend read_string_use $one_read_line
      }
    }
    set read_data_cur [join $read_string_use "\n"]

    if {!$gnuplot_just_started_read} {
      consume_gnuplot_onset
    }
  }

  set read_string [string map [list "\r" ""] $read_data_cur]
  if {$read_string != ""} {
    $plot_output_field configure -state "normal"
    $plot_output_field insert "end" $read_string [list]
    $plot_output_field yview end
    $plot_output_field configure -state "disabled"
    $lgnuplot_output configure -state "normal"
    dispatch_has_read_from_gnuplot
  }

  set max_check_int 1000
  if {$gnuplot_output_path_last_int<$max_check_int} {
    set gnuplot_output_path_last_int [expr {2*$gnuplot_output_path_last_int}]
    if {$gnuplot_output_path_last_int>$max_check_int} {
      set gnuplot_output_path_last_int $max_check_int
    }
  }

  set gnuplot_output_check_id [after $gnuplot_output_path_last_int "check_gnuplot_output"]
}

proc is_gnuplot_running {} {
  global gnuplot_is_running
  return $gnuplot_is_running
}

set gnuplot_version_read_pid_prefix "PID:"
proc close_gnuplot_direct {{overall false}} {
  global plot_put_window
  global gnuplot_handle
  global gnuplot_checks_id
  global gnuplot_output_id
  global gnuplot_output_check_id
  global lgnuplot_close
  global lgnuplot_state
  global gnuplot_commands_to_write
  global gnuplot_current_pid
  global gnuplot_tempfile_path
  global gnuplot_output_id_write
  global gnuplot_version_read
  global gnuplot_version_read_pid_prefix
  global gnuplot_is_running
  global plot_output_field

  # if Gnuplot was not setup yet, it has not run either;
  # and some variables used here would not be set too;
  if {$plot_put_window==""} {
    return
  }

  if {$gnuplot_checks_id!=""} {
    catch {after cancel $gnuplot_checks_id}
    set gnuplot_checks_id ""
  }

  if {$gnuplot_handle!=""} {
    catch {fileevent $gnuplot_handle readable ""}
    catch {fileevent $gnuplot_handle writable ""}
    catch {close $gnuplot_handle}
    set gnuplot_handle ""
  }
  if {$gnuplot_current_pid!=""} {
    set wait_time1 25
    set wait_time2 25
    if {![is_wine_like]} {
      if {!$overall} {
        pause_tasks $wait_time1
      }
      catch {exec "kill" "-s" "SIGTERM" $gnuplot_current_pid}
      if {!$overall} {
        pause_tasks $wait_time2
      }
      catch {exec "kill" "-s" "SIGKILL" $gnuplot_current_pid}
    } {
      if {!$overall} {
        pause_tasks $wait_time1
      }
      catch {exec "taskkill" "/pid" $gnuplot_current_pid}
      if {!$overall} {
        pause_tasks $wait_time2
      }
      catch {exec "taskkill" "/f" "/pid" $gnuplot_current_pid}
    }
    set gnuplot_current_pid ""
  }

  if {$gnuplot_output_id_write != ""} {
    catch {close $gnuplot_output_id_write}
  }

  # stopping file-based checking, if in use
  if {$gnuplot_output_check_id!=""} {
    catch {after cancel $gnuplot_output_check_id}
    set gnuplot_output_check_id ""
  }
  if {$gnuplot_output_id!=""} {
    catch {close $gnuplot_output_id}
    set gnuplot_output_id ""
  }

  # deleting temporary file only after all its opens are closed
  if {$gnuplot_tempfile_path != ""} {
    catch {
      file delete -- $gnuplot_tempfile_path
    }
    set gnuplot_tempfile_path ""
  }

  set gnuplot_commands_to_write [list]

  $lgnuplot_close configure -state "disabled"
  $lgnuplot_state configure -text ""

  if {[lindex [split [$plot_output_field index "end -1 chars"] "."] 1] != 0} {
    set output_state [$plot_output_field cget -state]
    $plot_output_field configure -state "normal"
    $plot_output_field insert "end" "\n" [list]
    $plot_output_field configure -state $output_state
  }

  if {$overall} {
    close_plot_part
  } {
    set gnuplot_version_read_last [list]
    set pid_prefix_len [string length $gnuplot_version_read_pid_prefix]
    foreach one_part $gnuplot_version_read {
      if {[string equal -nocase -length $pid_prefix_len $gnuplot_version_read_pid_prefix $one_part]} {
        set one_part "last $one_part"
      }
      lappend gnuplot_version_read_last $one_part
    }
    set gnuplot_version_read $gnuplot_version_read_last
  }

  set gnuplot_is_running false
  update_plot_clean_state
}

set gnuplot_tempfile_path ""
set gnuplot_output_id_write ""
set gnuplot_current_pid ""
set gnuplot_output_id ""
set gnuplot_just_started_read false
set gnuplot_just_started_write false
set gnuplot_is_running false
proc start_gnuplot_direct {} {
  global gnuplot_handle
  global gnuplot_output_id
  global gnuplot_output_check_id
  global gnuplot_output_path_last_int
  global gnuplot_output_path_last_int_start
  global lgnuplot_close
  global lgnuplot_state
  global gnuplot_just_started_read
  global gnuplot_just_started_write
  global gnuplot_commands_to_write
  global gnuplot_onset_read
  global gnuplot_current_pid
  global gnuplot_tempfile_path
  global gnuplot_output_id_write
  global gnuplot_is_running

  # this is for explicit starting of a new Gnuplot session,
  # thus making sure that it gets closed before that;
  close_gnuplot_direct

  set null_path [get_null_path]
  set gnuplot_cmmd [get_gnuplot_path]

  # it seems that read events lead to blocks under wine-like systems
  set can_read true
  if {[is_wine_like]} {
    set can_read false
  }

  if {$can_read} {
    catch {
      set gnuplot_handle [open "|$gnuplot_cmmd 2>@1" "w+"]
    }
  } {
    if {$gnuplot_output_id==""} {
      set gnuplot_tempfile_path ""
      set gnuplot_tmp_info [get_tempfile]
      set gnuplot_tmp_id [lindex $gnuplot_tmp_info 0]
      set gnuplot_tempfile_path [lindex $gnuplot_tmp_info 1]
      catch {
        close $gnuplot_tmp_id
      }
      if {$gnuplot_tempfile_path==""} {
        return false
      }
      if {[catch {
        set gnuplot_output_id_write [open $gnuplot_tempfile_path "w"]
      }]} {
        catch {close $gnuplot_output_id_write}
        catch {file delete -- $gnuplot_tempfile_path}
        return false
      }
      if {[catch {
        set gnuplot_output_id [open $gnuplot_tempfile_path "r"]
      }]} {
        catch {close $gnuplot_output_id_write}
        catch {close $gnuplot_output_id}
        catch {file delete -- $gnuplot_tempfile_path}
        return false
      }
    }
    catch {
      set gnuplot_handle [open "|$gnuplot_cmmd >&@ $gnuplot_output_id_write" "w"]
      set gnuplot_output_path_last_int $gnuplot_output_path_last_int_start
      set gnuplot_output_check_id [after $gnuplot_output_path_last_int "check_gnuplot_output"]
    }
  }

  catch {
    chan configure $gnuplot_handle -blocking false
    chan configure $gnuplot_handle -encoding "utf-8"
    chan configure $gnuplot_handle -translation lf
    # we do explicit flushing when writing to gnuplot
    #chan configure $gnuplot_handle -buffering line
  }

  if {[catch {
    set gnuplot_current_pid [lindex [pid $gnuplot_handle] 0]
  }]} {
    set gnuplot_current_pid ""
  }

  setup_gnuplot_checks

  if {$can_read} {
    catch {
      fileevent $gnuplot_handle readable "read_from_gnuplot"
    }
  }

  set gnuplot_just_started_read true
  set gnuplot_just_started_write true
  $lgnuplot_close configure -state "normal"
  $lgnuplot_state configure -text "started"
  set gnuplot_commands_to_write [list]
  set gnuplot_onset_read [list]

  catch {fileevent $gnuplot_handle writable "write_to_gnuplot"}

  set gnuplot_is_running true
  update_plot_clean_state
}

set gnuplot_commands_to_write [list]
set gnuplot_shibboleth_onset ""
proc write_to_gnuplot {} {
  global gnuplot_commands_to_write
  global gnuplot_handle
  global gnuplot_just_started_write
  global gnuplot_shibboleth_onset

  catch {fileevent $gnuplot_handle writable ""}

  set is_just_started $gnuplot_just_started_write
  set gnuplot_just_started_write false

  if {$is_just_started} {
    set gnuplot_shibboleth_onset "gso"
    append gnuplot_shibboleth_onset [expr {int(10000*rand()+1000)}]
    set bad [catch {
      puts $gnuplot_handle "show version;\nprint \"$gnuplot_shibboleth_onset\";\n"
      flush $gnuplot_handle
    }]
    if {[llength $gnuplot_commands_to_write]>0} {
      catch {fileevent $gnuplot_handle writable "write_to_gnuplot"}
    }
    return
  }

  set bad [catch {
    puts $gnuplot_handle [join $gnuplot_commands_to_write "\n"]
    flush $gnuplot_handle
  }]
  if {$bad} {
    show_plot_field_message_error [list [list "can not write command to gnuplot"]]
  }

  set gnuplot_commands_to_write [list]
}

proc translate_line_from_csv_to_words {line sep} {
  # trying to deal with any line, regardless of its CSV compliance
  set use_delm "\""
  set res_list [list]
  if {$line==""} {
    return $res_list
  }

  set is_regular true
  set interim_part ""
  foreach one_part [split $line $sep] {
    # if anyone would put spaces in, even though it is not according to CSV standard
    set one_part_trimmed [string trimleft $one_part]
    if {$is_regular && ($one_part_trimmed=="")} {
      lappend res_list ""
      continue
    }
    # counting starting delimiters (again after trimming)
    set delm_start 0
    while {[string range $one_part_trimmed $delm_start $delm_start]==$use_delm} {
      incr delm_start
    }
    set delm_start_odd [expr {($delm_start%2)==1}]
    # counting ending delimiters (again after trimming)
    set delm_end 0
    while {[string range $one_part_trimmed "end-$delm_end" "end-$delm_end"]==$use_delm} {
      incr delm_end 1
    }
    set delm_end_odd [expr {($delm_end%2)==1}]
    # delimiters should be generally halved
    set one_part [string map [list "\"\"" "\""] $one_part]
    # deciding on separate vs. joined parts
    if {$is_regular} {
      if {!$delm_start_odd} {
        lappend res_list $one_part_trimmed
      } {
        if {$delm_end_odd} {
          lappend res_list [string range $one_part_trimmed 1 "end-1"]
        } {
          set is_regular false
          # can not trim the inner (here ending) parts
          set interim_part [string range [string trimleft $one_part] 1 "end"]
        }
      }
      continue
    }
    if {$delm_end_odd} {
      # can not trim the inner (here starting) parts
      append interim_part $sep [string range [string trimright $one_part] 0 "end-1"]
      lappend res_list $interim_part
      set interim_part ""
      set is_regular true
      continue
    }
    # can not trim the inner (here throughout it) parts
    append interim_part $sep $one_part
  }

  if {!$is_regular} {
    lappend res_list $interim_part
  }

  return $res_list
}

set plot_data_block_separator [strcat "EOD" \
  [expr {int(10000*rand()+1000)}] \
  [expr {int(10000*rand()+1000)}] \
  [expr {int(10000*rand()+1000)}] \
]

proc insert_data_into_plot_command {command} {
  global plot_data_field
  global plot_data_block_separator

  if {$command==""} {
    return [list false ""]
  }

  set data_content [string trim [$plot_data_field get "1.0" "end -1 chars"]]

  set data_lines [list]
  foreach one_data_line [split $data_content "\n"] {
    set trimmed_data_line [string trim $one_data_line]
    if {$trimmed_data_line==""} {
      continue
    }
    if {[string range $trimmed_data_line 0 0]=="#"} {
      continue
    }
    lappend data_lines $one_data_line
  }

  append data_content "\n"

  set command_res ""
  set next_is_continuation false
  foreach one_line [split $command "\n"] {
    set line_is_continuation $next_is_continuation
    # the end-line continuation backslash does not get escaped by preceding backslash for Gnuplot;
    set next_is_continuation [expr {[string range $one_line "end" "end"]=="\\"}]
    set one_line_trimmed [string trim $one_line]
    if {$line_is_continuation} {
      append command_res "$one_line\n"
    } elseif {$one_line_trimmed=="@@DB"} {
      append command_res "\$DB << $plot_data_block_separator\n"
      append command_res $data_content
      append command_res "$plot_data_block_separator\n"
    } elseif {$one_line_trimmed=="@@DI"} {
      append command_res $data_content
      append command_res "e\n"
    } elseif {$one_line_trimmed=="@@DW"} {
      set dl_idx 0
      set all_data_line_names [list]
      foreach one_data_line $data_lines {
        incr dl_idx
        lappend all_data_line_names "DWR$dl_idx"
        append command_res "DWR$dl_idx = \"" [string map [list "\\" "\\\\" "\"" "\\\""] $one_data_line] "\"\n"
      }
      append command_res "DWR0 = \"" [join $all_data_line_names " "] "\"\n"
      append command_res "DWRC = $dl_idx\n"
      append command_res "DWC(r,c) = word(value(sprintf(\"DWR%d\",r)),c)\n"
      append command_res "DWRS(r) = words(value(sprintf(\"DWR%d\",r)))\n"
      append command_res "DWCD(r,c,d) = ((r<1)||(r>DWRC)||(c<1)||(c>DWRS(r))) ? d : DWC(r,c)\n"
    } elseif {($one_line_trimmed=="@@DA,") || ($one_line_trimmed=="@@DA;")} {
      set sep_char [string range $one_line_trimmed "end" "end"]
      set dl_idx 0
      set all_data_line_names [list]
      set all_data_line_sizes [list]
      foreach one_data_line $data_lines {
        set arr_cols [list]
        foreach one_column [translate_line_from_csv_to_words $one_data_line $sep_char] {
          set one_column [string trim $one_column]
          if {($one_column=="NaN") || [string is double -strict $one_column] || [string is entier -strict $one_column]} {
            lappend arr_cols $one_column
          } {
            lappend arr_cols [strcat "\"" [string map [list "\\" "\\\\" "\"" "\\\""] $one_column] "\""]
          }
        }
        set arr_row_len [llength $arr_cols]
        if {$arr_row_len==0} {
          continue
        }
        incr dl_idx
        append command_res [strcat "array DAR$dl_idx\[" $arr_row_len "\] = \[" [join $arr_cols ", "] "\]\n"]
        lappend all_data_line_names "DAR$dl_idx"
        lappend all_data_line_sizes $arr_row_len
      }
      if {$dl_idx>0} {
        append command_res "array DAR0\[$dl_idx\] = \[\"" [join $all_data_line_names "\", \""] "\"\]\n"
        append command_res "array DARS\[$dl_idx\] = \[" [join $all_data_line_sizes ", "] "\]\n"
      }
      append command_res "DARC = $dl_idx\n"
      append command_res "DAC(r,c) = value(sprintf(\"DAR%d\",r))\[c\]\n"
      append command_res "DACD(r,c,d) = ((r<1)||(r>DARC)||(c<1)||(c>DARS\[r\])) ? d : DAC(r,c)\n"
    } {
      append command_res "$one_line\n"
    }
  }

  return [list true [strcat [string trimright $command_res] "\n\n"]]
}

proc take_plot_command {} {
  global plot_input_field
  global test_plot_input_select_first
  global test_plot_input_whole_then
  global gnuplot_at_data_view
  global gnuplot_at_output_view

  set switch_on_error false
  if {$gnuplot_at_data_view || $gnuplot_at_output_view} {
    set switch_on_error true
  }

  if {$plot_input_field==""} {
    return [list false ""]
  }

  set take_part_only false
  set command_text ""

  # if a part is selected, only that part is used;
  # in such a case, errors reported without the accompanied pointings;
  if {$test_plot_input_select_first} {
    set command_text [list]
    foreach {sel_start sel_after} [$plot_input_field tag ranges "sel"] {
      lappend command_text [$plot_input_field get $sel_start $sel_after]
    }
    set command_text [join $command_text "\n"]
    if {$command_text != ""} {
      set take_part_only true
    }
  }
  if {$test_plot_input_whole_then} {
    if {!$take_part_only} {
      set command_text [$plot_input_field get "1.0" "end -1 chars"]
    }
  }

  if {$command_text==""} {
    show_plot_field_message_error [list [list "no commands provided for gnuplot"]]
    return [list false ""]
  }

  set plot_comments_strings_brackets [get_comments_strings_brackets_in_gnuplot $command_text]

  # check if sane text content
  set brackets_use [lindex $plot_comments_strings_brackets 2]
  set err_messages [test_brackets_sane $brackets_use [expr {!$take_part_only}]]

  if {[llength $err_messages]>0} {
    if {$switch_on_error} {
      present_gnuplot_input
      update
    }
    show_plot_field_message_error $err_messages
    return [list false ""]
  }

  return [list true $command_text]
}

proc test_plot_command {} {
  set command_text_info [take_plot_command]
  if {([lindex $command_text_info 0]!=false) && ([lindex $command_text_info 1]!="")} {
    show_plot_field_message_error [list [list "simple parsing: correct"]]
  }
}

set gnuplot_last_plot_time 0
set gnuplot_keep_focused_id ""

# a simple interface to gnuplot
proc exec_plot_command {to_restart} {
  global plot_put_window
  global gnuplot_handle
  global plot_input_field
  global plot_bracket_stack
  global gnuplot_commands_to_write
  global gnuplot_last_plot_time
  global gnuplot_keep_focused_id
  global gnuplot_keep_always_topmost
  global gnuplot_keep_focused_time
  global gnuplot_got_new_command

  if {$plot_input_field==""} {
    return
  }

  set command_text_info [take_plot_command]
  if {([lindex $command_text_info 0]==false) || ([lindex $command_text_info 1]=="")} {
    return
  }
  set command_text [lindex $command_text_info 1]

  set command_text_info [insert_data_into_plot_command $command_text]
  if {([lindex $command_text_info 0]==false) || ([lindex $command_text_info 1]=="")} {
    show_plot_field_message_error [list [list "can not use data for plotting"]]
    return
  }
  set command_text [lindex $command_text_info 1]

  # this for visual purposes only;
  update_plot_properties

  set gnuplot_got_new_command true

  if {![check_gnuplot_connected]} {
    set to_restart true
  }
  if {$to_restart} {
    # starting closes it too, thus without explicit closing here;
    start_gnuplot_direct
  }

  if {![check_gnuplot_connected]} {
    show_plot_field_message_error [list [list "can not start gnuplot"]]
    return
  }

  lappend gnuplot_commands_to_write $command_text
  set bad [catch {
    fileevent $gnuplot_handle writable "write_to_gnuplot"
  }]
  if {$bad} {
    show_plot_field_message_error [list [list "error at gnuplot running"]]
  }

  if {$gnuplot_keep_focused_id!=""} {
    catch {after cancel $gnuplot_keep_focused_id}
    set gnuplot_keep_focused_id ""
  }

  set gnuplot_last_plot_time [clock milliseconds]
  if {$gnuplot_keep_focused_time!=0} {
    set gnuplot_keep_focused_id [after 100 "focus_plot_input_field"]
  }
}

proc focus_plot_input_field {} {
  global plot_put_window
  global gnuplot_keep_focused_id
  global gnuplot_last_plot_time
  global gnuplot_keep_always_topmost
  global gnuplot_keep_focused_time
  global gnuplot_keep_focused_full

  if {$gnuplot_keep_focused_id!=""} {
    catch {after cancel $gnuplot_keep_focused_id}
    set gnuplot_keep_focused_id ""
  }

  if {[focus -displayof $plot_put_window]==""} {
    set wdg [focus -lastfor $plot_put_window]
    focus -force $wdg
    if {!$gnuplot_keep_focused_full} {
      return
    }
  }

  if {$gnuplot_keep_focused_time!=0} {
    if {[clock milliseconds] < [expr {$gnuplot_last_plot_time + 100*$gnuplot_keep_focused_time}]} {
      set gnuplot_keep_focused_id [after 100 "focus_plot_input_field"]
      return
    }
  }
}

# Maxima batch windows

set maxima_batch_pid_present ""
proc take_pid_maxima_batch {} {
  global maxima_batch_onset_string
  global batch_maxima_ids
  global maxima_batch_pid_present
  global maxima_pid_batch
  global maxima_batch_is_continuation
  global priority_wine_common
  global priority_wine_idle_set
  global minima_uses_twapi
  # the PID that we got by inner proc read belongs to wineserver (not to Maxima),
  # thus getting it via a (more involved) outer way;
  #global maxima_batch_pid_outer

  if {$maxima_batch_is_continuation} {
    return
  }

  foreach one_line [split $maxima_batch_onset_string "\n"] {
    foreach one_part [split $one_line] {
      if {$one_part==""} {
        continue
      }
      set one_part [split [string trim $one_part] "="]
      if {[llength $one_part]==2} {
        if {[string tolower [lindex $one_part 0]]=="pid"} {
          set cur_pid [lindex $one_part 1]
          if {!($cur_pid in $batch_maxima_ids)} {
            lappend batch_maxima_ids $cur_pid
            # have to reset cpu-usage info here, since a new value means that we have been tracking the enveloping batch
            if {![is_under_wine]} {
              reset_cpu_usage
            }
          }
        }
      }
    }
  }

  set maxima_batch_pid_present [lindex $batch_maxima_ids "end"]
  set maxima_pid_batch [expr {[is_batch_maxima_running] ? $maxima_batch_pid_present : ""}]

  if {[is_wine_like] && ($maxima_pid_batch!="") && $minima_uses_twapi} {
    catch {
      twapi::set_priority_class $maxima_pid_batch $priority_wine_common
      set priority_wine_idle_set false
    }
  }
}

set maxima_batch_start_row 0
proc take_onset_maxima_batch {} {
  global maxima_batch_onset_string
  global info_tag_start
  global info_tag_middle
  global info_tag_end
  global maxima_info_version
  global maxima_info_lisp_name
  global maxima_batch_start_row
  global mxmbatch_output_field
  global maxima_batch_is_continuation

  if {$maxima_batch_is_continuation} {
    return
  }

  take_pid_maxima_batch

  set info_part_first ""
  set info_part_second ""

  set loc_tag_start [string last $info_tag_start $maxima_batch_onset_string]
  set loc_tag_middle [string last $info_tag_middle $maxima_batch_onset_string]
  set loc_tag_end [string last $info_tag_end $maxima_batch_onset_string]

  if {($loc_tag_start>-1) && ($loc_tag_middle>-1) && ($loc_tag_start < $loc_tag_middle)} {
    set info_part_first [string range $maxima_batch_onset_string [expr {
      $loc_tag_start + [string length $info_tag_start]
    }] [expr {$loc_tag_middle - 1}]]
  }

  if {($loc_tag_middle>-1) && ($loc_tag_end>-1) && ($loc_tag_middle < $loc_tag_end)} {
    set info_part_second [string range $maxima_batch_onset_string [expr {
      $loc_tag_middle + [string length $info_tag_middle]
    }] [expr {$loc_tag_end - 1}]]
  }

  set parts_reduced [list]
  foreach one_part [split $info_part_first] {
    set one_part [string trim $one_part]
    if {$one_part==""} {
      continue
    }
    lappend parts_reduced $one_part
  }
  set info_part_first [join $parts_reduced " "]

  set parts_reduced [list]
  foreach one_part [split $info_part_second] {
    set one_part [string trim $one_part]
    if {$one_part==""} {
      continue
    }
    lappend parts_reduced $one_part
  }
  set info_part_second [join $parts_reduced " "]

  if {[string length $info_part_first]>0} {
    set maxima_info_version $info_part_first
  }
  if {[string length $info_part_second]>0} {
    set maxima_info_lisp_name $info_part_second
  }

  set maxima_batch_start_row [lindex [split [$mxmbatch_output_field index "end"] "."] 0]
}

proc is_preceded_count_during_asking_within_limits {} {
  global batch_maxima_asks_shibboleth
  global batch_maxima_asks_received
  global maxima_batch_max_row_count

  set shbb_loc [string first $batch_maxima_asks_shibboleth $batch_maxima_asks_received]
  if {$shbb_loc!=-1} {
    # if already started to read the names, not killing it regardless of the count of preceded rows;
    return true
  }

  if {[llength [split $batch_maxima_asks_received "\n"]]>$maxima_batch_max_row_count} {
    return false
  }

  return true
}

set maxima_batch_asked_to_write_to_text false
set maxima_batch_wrote_row_count_at_text 0

proc read_maxima_batch_write_to_text {use_mxmout} {
  global mxmbatch_text_insert_mark_name
  global text_field
  global maxima_batch_max_row_count_for_text
  global maxima_batch_wrote_row_count_at_text
  global mxmbatch_text_command_is_one_line
  global mxmbatch_text_command_one_line_string
  global maxima_batch_asked_to_write_to_text

  if {!$maxima_batch_asked_to_write_to_text} {
    return
  }

  if {$maxima_batch_wrote_row_count_at_text>=$maxima_batch_max_row_count_for_text} {
    return
  }

  set max_line_idx_to_write [expr {$maxima_batch_max_row_count_for_text - ($maxima_batch_wrote_row_count_at_text + 1)}]
  set use_mxmout_list_for_text [split $use_mxmout "\n"]

  if {$mxmbatch_text_command_is_one_line && ($maxima_batch_wrote_row_count_at_text==0)} {
    set first_read_line [lindex $use_mxmout_list_for_text 0]
    if {[string range $first_read_line "end" "end"] in [list ";"]} {
      set first_read_line [string range $first_read_line 0 "end-1"]
    }
    if {$first_read_line==$mxmbatch_text_command_one_line_string} {
      set use_mxmout_list_for_text [lrange $use_mxmout_list_for_text 1 "end"]
      set mxmbatch_text_command_is_one_line false
    }
  }

  if {[llength $use_mxmout_list_for_text]==0} {
    return
  }

  if {!($mxmbatch_text_insert_mark_name in [$text_field mark names])} {
    # first writing in, or file (re)loaded and/or emptied, though we wanna write in there;
    set insert_info [prepare_cursor_wise_insertion_point_other true]
    set mark_insert_position [lindex $insert_info 0]
    set prepare_insert_info [lindex $insert_info 1]
    if {[llength $prepare_insert_info]==2} {
      $text_field edit separator
      $text_field insert [lindex $prepare_insert_info 0] [lindex $prepare_insert_info 1]
      $text_field edit separator
    }
    $text_field mark set $mxmbatch_text_insert_mark_name $mark_insert_position
  }

  $text_field edit separator
  $text_field insert $mxmbatch_text_insert_mark_name \
    [join [lrange $use_mxmout_list_for_text 0 $max_line_idx_to_write] "\n"] [list]
  $text_field edit separator
  $text_field see $mxmbatch_text_insert_mark_name
  incr maxima_batch_wrote_row_count_at_text [expr {[llength $use_mxmout_list_for_text] - 1}]
  update_text_properties
}

set maxima_batch_to_end false
set maxima_batch_end_reached false
set batch_maxima_end_shibboleth ""
set print_logging_maxima_output_last false
set maxima_batch_last_output_line_string ""

set maxima_batch_onset_string ""
set maxima_batch_onset_string_row_count 0
set maxima_batch_first_time_info false
set maxima_batch_time_tag "maxima_batch_time_tag"
set maxima_batch_output_tag "maxima_batch_output_tag"
proc read_maxima_batch {} {
  global batch_maxima_handle
  global maxima_batch_just_started_read
  global maxima_batch_onset_string
  global maxima_batch_onset_string_row_count
  global mxmbatch_output_field
  global mxmbatch_sweep
  global maxima_batch_first_time_info
  global time_elapsed_pattern
  global time_elapsed_regexps
  global maxima_batch_time_tag
  global maxima_batch_output_tag
  global batch_maxima_using_alt_display
  global batch_maxima_asks_received
  global batch_maxima_start_shibboleth
  global maxima_output_to_split
  global info_line_size
  global info_line_long
  global info_line_inner
  global info_line_tex
  global time_elapsed_regexps_start
  global batch_maxima_asks_names
  global maxima_batch_start_row
  global maxima_batch_max_row_count
  global mxmbatch_text_insert_mark_name
  global text_field
  global maxima_batch_max_row_count_for_text
  global maxima_batch_wrote_row_count_at_text
  global maxima_batch_asked_to_write_to_text
  global batch_maxima_end_automatic
  global batch_maxima_end_shibboleth
  global maxima_batch_to_end
  global maxima_batch_end_reached
  global maxima_batch_is_continuation
  global print_logging_maxima_output_last
  global print_logging_maxima_output
  global print_logging_maxima_input
  global maxima_batch_last_output_line_string

  set write_to_text false
  if {$maxima_batch_asked_to_write_to_text} {
    if {$maxima_batch_wrote_row_count_at_text<$maxima_batch_max_row_count_for_text} {
      set write_to_text true
    }
  }

  if {$maxima_output_to_split && (!$batch_maxima_asks_names)} {
    set maxima_output_to_split_len [lindex $info_line_size 0]
    if {(!$info_line_long) || $info_line_inner || $info_line_tex} {
      # the issue with spurious line breaks seems to be solved
      #incr maxima_output_to_split_len 20
    }
  } {
    set maxima_output_to_split_len 0
  }

  if {$maxima_batch_just_started_read} {
    set maxima_batch_first_time_info true
  }

  catch {fileevent $batch_maxima_handle readable ""}
  set mxmout ""
  if {[catch {
    set mxmout [read $batch_maxima_handle]
    if {$print_logging_maxima_output} {
      if {$print_logging_maxima_input && (!$print_logging_maxima_output_last)} {
        set print_logging_maxima_output_last true
        print_logging "\nXXX output\n"
      }
      print_logging $mxmout
    }
  }]} {
    set mxmout ""
  }

  if {[is_batch_maxima_waiting]} {
    # this occasionally happens for outputs from the if commands around the batch;
    return
  }

  set to_setup_waiting false

  # using clisp under wine leads to getting the output in small chunks;
  # it may possibly occur for some other cases too, thus dealing with it here;
  # eventually should use this for simplifying the field-writing stuff below,
  # with using parts of the non-full line when it gets too large (in 1D cases)
  # (that is not waiting to its filling if not necessary);
  set last_nl_pos [string last "\n" $mxmout]
  if {$last_nl_pos==-1} {
    append maxima_batch_last_output_line_string $mxmout
    set mxmout ""
  } {
    set last_part_after_nl [string trimleft [string range $mxmout [expr {$last_nl_pos + 1}] "end"] "\r"]
    set mxmout [strcat $maxima_batch_last_output_line_string [string range $mxmout 0 $last_nl_pos]]
    set maxima_batch_last_output_line_string $last_part_after_nl
  }

  set use_mxmout ""
  if {$mxmout!=""} {
    set end_pos [string first $batch_maxima_end_shibboleth $mxmout]
    if {$end_pos>-1} {
      set mxmout [string range $mxmout 0 [expr {$end_pos - 1}]]
      set end_pos [string last "\n" $mxmout]
      if {$end_pos>-1} {
        set mxmout [string range $mxmout 0 [expr {$end_pos - 0}]]
      } {
        set mxmout ""
      }
      if {!$maxima_batch_end_reached} {
        if {!$batch_maxima_end_automatic} {
          set to_setup_waiting true
        }
      }
      set maxima_batch_end_reached true
    }
    if {$batch_maxima_end_automatic} {
      if {$maxima_batch_end_reached} {
        set maxima_batch_to_end true
      }
    }

    if {$maxima_batch_just_started_read} {
      set use_mxmout [list]
      foreach one_read_part [split $mxmout "\n"] {
        set one_read_part [string trim $one_read_part "\r"]
        if {$maxima_batch_just_started_read} {
          incr maxima_batch_onset_string_row_count
          append maxima_batch_onset_string $one_read_part "\n"
          if {[string first $batch_maxima_start_shibboleth $maxima_batch_onset_string] != -1} {
            if {!$maxima_batch_is_continuation} {
              take_onset_maxima_batch
            }
            set maxima_batch_just_started_read false
          } elseif {(!$maxima_batch_is_continuation) && ([string first "pid=" $one_read_part] != -1)} {
            # early taking of PID only from the separate lines, otherwise would be taking it amount of times,
            # plus the "take_onset_maxima_batch" function does it again (on the whole read part);
            take_pid_maxima_batch
          } {
            # checking if the onset does not run away (like if an issue with prepend line, when being put into the ini stuff);
            # since some Lisp implementations tend to output superfluous debugging info, putting the limit somewhat up here;
            if {[expr {(1000 + $maxima_batch_max_row_count)}] < $maxima_batch_onset_string_row_count} {
              after 0 [list "show_batch_maxima_message_error" [list [list "Maxima stopped due to exceeded initialization count of rows"]]]
              stop_batch_maxima
              return
            }
          }
          continue
        }
        if {$maxima_batch_first_time_info} {
          if {$one_read_part!=""} {
            set maxima_batch_first_time_info false
          }
          if {[string match -nocase $time_elapsed_pattern $one_read_part]} {
            continue
          }
        }
        lappend use_mxmout $one_read_part
      }
      set use_mxmout [join $use_mxmout "\n"]
    } {
      set mxmout [string map [list "\r" ""] $mxmout]
      if {$maxima_batch_first_time_info} {
        set use_mxmout [list]
        foreach one_read_part [split $mxmout "\n"] {
          if {$maxima_batch_first_time_info} {
            if {$one_read_part!=""} {
              set maxima_batch_first_time_info false
            }
            if {[string match -nocase $time_elapsed_pattern $one_read_part]} {
              continue
            }
          }
          lappend use_mxmout $one_read_part
        }
        set use_mxmout [join $use_mxmout "\n"]
      } {
        set use_mxmout $mxmout
      }
    }

    if {($use_mxmout!="") && ([is_batch_maxima_asking])} {
      append batch_maxima_asks_received $use_mxmout
    }

    if {($use_mxmout!="") && (![is_batch_maxima_asking])} {
      set cur_state [$mxmbatch_output_field cget -state]
      $mxmbatch_output_field configure -state "normal"
      # (more) below splitting into shorter lines
      if {$maxima_output_to_split_len==0} {
        $mxmbatch_output_field insert "end" $use_mxmout [list]
        if {$write_to_text} {
          read_maxima_batch_write_to_text $use_mxmout
        }
      } {
        set use_mxmout_list [split $use_mxmout "\n"]
        set use_mxmout_list_len [llength $use_mxmout_list]
        # dealing with the first (currently) read line where have to consider its (possibly) previously read parts
        set end_pos [$mxmbatch_output_field index "end -1 chars"]
        set end_pos_info [split $end_pos "."]
        set end_col [lindex $end_pos_info 1]
        set can_add [expr {$maxima_output_to_split_len - $end_col}]
        set first_read_line [lindex $use_mxmout_list 0]
        if {$can_add<0} {
          # already had decided to not spilt this (i.e. taking it as time-eval info)
          $mxmbatch_output_field insert "end" $first_read_line [list]
          if {$write_to_text} {
            read_maxima_batch_write_to_text $first_read_line
          }
        } {
          set curr_read_len [string length $first_read_line]
          $mxmbatch_output_field insert "end" [string range $first_read_line 0 [expr {$can_add - 1}]] [list]
          if {$write_to_text} {
            read_maxima_batch_write_to_text [string range $first_read_line 0 [expr {$can_add - 1}]]
          }
          if {$curr_read_len>$can_add} {
            # if it seems to be eval-time info, putting it there without splitting
            if {[regexp -- $time_elapsed_regexps_start [$mxmbatch_output_field get "$end_pos linestart" "$end_pos lineend"]]} {
              $mxmbatch_output_field insert "end" [string range $first_read_line $can_add "end"] [list]
              if {$write_to_text} {
                read_maxima_batch_write_to_text [string range $first_read_line $can_add "end"]
              }
            } {
              set curr_line_offset $maxima_output_to_split_len
              while {$curr_line_offset<$curr_read_len} {
                $mxmbatch_output_field insert "end" "\\\n" [list]
                $mxmbatch_output_field insert "end" [string range $first_read_line $curr_line_offset [expr {$curr_line_offset + $maxima_output_to_split_len - 1}]] [list]
                if {$write_to_text} {
                  read_maxima_batch_write_to_text "\\\n"
                  read_maxima_batch_write_to_text [string range $first_read_line $curr_line_offset [expr {$curr_line_offset + $maxima_output_to_split_len - 1}]]
                }
                incr curr_line_offset $maxima_output_to_split_len
              }
            }
          }
        }
        # the other currently read lines
        set curr_line_idx 0
        while {true} {
          incr curr_line_idx
          if {$curr_line_idx==$use_mxmout_list_len} {
            break
          }
          $mxmbatch_output_field insert "end" "\n" [list]
          if {$write_to_text} {
            read_maxima_batch_write_to_text "\n"
          }
          set curr_read_line [lindex $use_mxmout_list $curr_line_idx]
          set curr_read_len [string length $curr_read_line]
          if {$curr_read_len<=$maxima_output_to_split_len} {
            $mxmbatch_output_field insert "end" $curr_read_line [list]
            if {$write_to_text} {
              read_maxima_batch_write_to_text $curr_read_line
            }
            continue
          }
          if {[regexp -- $time_elapsed_regexps_start $curr_read_line]} {
            $mxmbatch_output_field insert "end" $curr_read_line [list]
            if {$write_to_text} {
              read_maxima_batch_write_to_text $curr_read_line
            }
            continue
          }
          set first_line_part [string range $curr_read_line 0 [expr {$maxima_output_to_split_len - 1}]]
          $mxmbatch_output_field insert "end" $first_line_part [list]
          if {$write_to_text} {
            read_maxima_batch_write_to_text $first_line_part
          }
          set curr_line_offset $maxima_output_to_split_len
          while {$curr_line_offset<$curr_read_len} {
            $mxmbatch_output_field insert "end" "\\\n" [list]
            $mxmbatch_output_field insert "end" [string range $curr_read_line $curr_line_offset [expr {$curr_line_offset + $maxima_output_to_split_len - 1}]] [list]
            if {$write_to_text} {
              read_maxima_batch_write_to_text "\\\n"
              read_maxima_batch_write_to_text [string range $curr_read_line $curr_line_offset [expr {$curr_line_offset + $maxima_output_to_split_len - 1}]]
            }
            incr curr_line_offset $maxima_output_to_split_len
          }
        }
      }
      $mxmbatch_output_field configure -state $cur_state
      $mxmbatch_output_field see "end -1 chars linestart"
      $mxmbatch_output_field tag remove $maxima_batch_time_tag "1.0" "end"
      set tei_count_var [list]
      set tei_pos_var [$mxmbatch_output_field search -regexp -all -count tei_count_var $time_elapsed_regexps "1.0" "end"]
      set tei_len [llength $tei_pos_var]
      set tei_idx 0
      while {$tei_idx<$tei_len} {
        set tei_start [lindex $tei_pos_var $tei_idx]
        set tei_offset [lindex $tei_count_var $tei_idx]
        $mxmbatch_output_field tag add $maxima_batch_time_tag $tei_start "$tei_start +$tei_offset chars"
        incr tei_idx
      }
      $mxmbatch_output_field tag remove $maxima_batch_output_tag "1.0" "end"
      $mxmbatch_output_field tag add $maxima_batch_output_tag "1.0" "end"
      $mxmbatch_sweep configure -state "normal"
    }
  }

  if {$to_setup_waiting} {
    setup_waiting_batch_maxima
  }

  if {$maxima_batch_to_end} {
    stop_batch_maxima
    return
  }

  if {[catch {
    if {[eof $batch_maxima_handle]} {
      stop_batch_maxima
      return
    }
  }]} {
    stop_batch_maxima
    return
  }

  if {($use_mxmout!="") && (![is_batch_maxima_asking])} {
    if {[expr {$maxima_batch_start_row + $maxima_batch_max_row_count}] < [lindex [split [$mxmbatch_output_field index "end"] "."] 0]} {
      after 0 [list "show_batch_maxima_message_error" [list [list "Maxima stopped due to exceeded batch count of rows"]]]
      stop_batch_maxima
      return
    }
  }

  if {($use_mxmout!="") && ([is_batch_maxima_asking])} {
    if {![is_preceded_count_during_asking_within_limits]} {
      after 0 [list "show_batch_maxima_message_error" [list [list "Maxima stopped due to exceeded batch count of rows"]]]
      stop_batch_maxima
      return
    }
  }

  catch {fileevent $batch_maxima_handle readable "read_maxima_batch"}
}

proc is_batch_maxima_asking {} {
  global batch_maxima_asks_names
  return $batch_maxima_asks_names
}

set batch_maxima_asks_received ""
proc names_start_batch_maxima_process {} {
  global batch_maxima_asks_functions
  global batch_maxima_asks_received
  global batch_maxima_asks_shibboleth

  set pos_from [string last $batch_maxima_asks_shibboleth $batch_maxima_asks_received]
  if {$pos_from==-1} {
    # this may happen on a killed question (e.g. when wrong prepend-command provided)
    return
  }
  set shibboleth_len [string length $batch_maxima_asks_shibboleth]
  set batch_maxima_asks_received [string range $batch_maxima_asks_received [expr {$pos_from + $shibboleth_len}] "end"]

  if {$batch_maxima_asks_functions} {
    set_functions_info $batch_maxima_asks_received
  } {
    set_values_info $batch_maxima_asks_received
  }

  set batch_maxima_asks_received ""
}

set batch_maxima_asks_names false
set batch_maxima_asks_functions false
set batch_maxima_asks_with_command false
proc names_start_batch_maxima {ask_functions with_command} {
  global batch_maxima_asks_names
  global batch_maxima_asks_functions
  global batch_maxima_asks_with_command
  global test_command_select_first
  global test_command_whole_then
  global text_field
  global mxmbatch_text_insert_mark_name
  global maxima_batch_asked_to_write_to_text

  if {![is_mxmbatch_input_field_active]} {
    return
  }
  if {[is_batch_maxima_running] && (![is_batch_maxima_waiting])} {
    return
  }
  set use_trigger [expr {[is_batch_maxima_waiting] ? "continue_batch_maxima" : "start_batch_maxima"}]

  set batch_maxima_asks_functions $ask_functions
  set batch_maxima_asks_with_command $with_command
  set batch_maxima_asks_names true
  $text_field mark unset $mxmbatch_text_insert_mark_name
  set maxima_batch_asked_to_write_to_text false

  if {![$use_trigger $test_command_select_first $test_command_whole_then]} {
    set batch_maxima_asks_names false
    return
  }
  state_batch_maxima
}

proc mxmbatch_input_field_set_active {} {
  global mxmbatch_input_field

  $mxmbatch_input_field configure -state "normal"
  $mxmbatch_input_field configure -foreground "#000000"
}
proc mxmbatch_input_field_set_locked {} {
  global mxmbatch_input_field

  $mxmbatch_input_field configure -state "disabled"
  $mxmbatch_input_field configure -foreground "#404080"
}

proc test_start_batch_maxima {} {
  global test_command_select_first
  global test_command_whole_then

  set curr_command [command_batch_maxima $test_command_select_first $test_command_whole_then]
  if {![lindex $curr_command 0]} {
    return false
  }

  show_batch_maxima_message_error [list [list "simple parsing: correct"]]
  return true
}

set batch_maxima_last_command ""
proc default_start_batch_maxima {{set_last_command true} {clean_field false}} {
  global test_command_select_first
  global test_command_whole_then
  global batch_maxima_asks_names
  global batch_maxima_last_command
  global mxmbatch_input_field
  global text_field
  global mxmbatch_text_insert_mark_name
  global maxima_batch_asked_to_write_to_text

  if {![is_mxmbatch_input_field_active]} {
    return
  }
  if {[is_batch_maxima_running] && (![is_batch_maxima_waiting])} {
    return
  }
  set use_trigger [expr {[is_batch_maxima_waiting] ? "continue_batch_maxima" : "start_batch_maxima"}]

  if {$set_last_command} {
    set current_command [get_command_content false]
  }

  set batch_maxima_asks_names false
  $text_field mark unset $mxmbatch_text_insert_mark_name
  set maxima_batch_asked_to_write_to_text false

  if {![$use_trigger $test_command_select_first $test_command_whole_then]} {
    return
  }

  if {$set_last_command} {
    set batch_maxima_last_command $current_command
  }
  if {$clean_field} {
    $mxmbatch_input_field delete "1.0" "end"
  }
}

set batch_maxima_asks_shibboleth ""
proc command_batch_maxima_question {} {
  global batch_maxima_asks_names
  global batch_maxima_asks_functions
  global batch_maxima_asks_shibboleth
  global info_tag_start
  global info_tag_end

  set shbb01 [strcat "so" [expr {int(10000*rand()+1000)}]]
  set shbb02 [strcat "so" [expr {int(10000*rand()+1000)}]]
  set shbb03 [strcat "so" [expr {int(10000*rand()+1000)}]]

  set batch_maxima_asks_shibboleth [strcat $shbb01 $shbb02 $shbb03]

  set form "values"
  if {$batch_maxima_asks_functions} {
    set form "functions"
  }
  # avoiding use of simplode, since its package need not be loaded
  set form_part "tree_reduce(lambda(\[x,y\],sconcat(x,\";\",y)),$form,\"\")"

  set ask_command "(leftjust:true,display2d:false,linel:1000000,ttyoff:false,"
  append ask_command "grind(concat(\"$shbb01\",\"$shbb02\",\"$shbb03\",\"$info_tag_start\",$form_part,\"$info_tag_end\")),"
  append ask_command "kill(labels))\$"

  return $ask_command
}

proc command_batch_maxima_prepend {{even_empty_like false}} {
  global mxmbatch_input_prepend_use
  global mxmbatch_input_prepend
  global allow_maxima_system_variables
  global batch_maxima_asks_names

  set mssg_onset "onset line:"

  set command_onset ""

  if {!$mxmbatch_input_prepend_use} {
    return [list true ""]
  }

  set command_onset [string trim $mxmbatch_input_prepend]
  if {$command_onset==""} {
    return [list true ""]
  }

  set mxmbatch_comments_strings_brackets [get_comments_strings_brackets $command_onset]

  # check if sane text content
  set brackets_use [lindex $mxmbatch_comments_strings_brackets 2]
  set err_messages [test_brackets_sane $brackets_use false [strcat $mssg_onset " "]]

  set string_last [lindex [lindex $mxmbatch_comments_strings_brackets 1] "end"]
  if {[llength $string_last]>1} {
    if {[lindex [lindex $string_last 1] 0]=="end"} {
      lappend err_messages [list "$mssg_onset unfinished string"]
    }
  }

  # checking open comments too, since they should be closed for Maxima
  set comment_last [lindex [lindex $mxmbatch_comments_strings_brackets 0] "end"]
  if {[llength $comment_last]>1} {
    if {[lindex [lindex $comment_last 1] 0]=="end"} {
      lappend err_messages [list "$mssg_onset unfinished comment"]
    }
  }

  if {[llength $err_messages]>0} {
    show_batch_maxima_message_error $err_messages
    return [list false]
  }

  if {($batch_maxima_asks_names || (!$allow_maxima_system_variables)) && ($command_onset != "")} {
    set disallowed_warns [check_setting_presence_in_command $command_onset $mxmbatch_comments_strings_brackets false]
    if {[llength $disallowed_warns]>0} {
      show_batch_maxima_message_error $disallowed_warns
      return [list false]
    }
    # since stuff could be set/executed via :lisp(...) constructs
    set disallowed_warns [check_setting_presence_in_command $command_onset $mxmbatch_comments_strings_brackets true]
    if {[llength $disallowed_warns]>0} {
      show_batch_maxima_message_error $disallowed_warns
      return [list false]
    }
  }

  # if "[lindex $end_check 0]==false" the string is effectively empty;
  # that is not an issue for the prepend line though;
  set end_check [check_command_ending false $command_onset [lindex $mxmbatch_comments_strings_brackets 0]]
  if {[lindex $end_check 0]==false} {
    if {$even_empty_like} {
      return [list true [strcat $command_onset "\n"]]
    } {
      return [list true ""]
    }
  }

  if {([llength $end_check]>1) && ([lindex $end_check 1]!="")} {
    set command_onset [lindex $end_check 1]
  }

  return [list true [strcat $command_onset "\n"]]
}

# notice that the prepend line is served by a separate function;
proc command_batch_maxima {{with_selected true} {with_whole true}} {
  global mxmbatch_input_field
  global batch_maxima_asks_names
  global batch_maxima_asks_with_command
  global make_line_endings
  global allow_maxima_system_variables

  if {$mxmbatch_input_field==""} {
    return [list false]
  }

  set name_question ""
  if {$batch_maxima_asks_names} {
    set name_question [command_batch_maxima_question]
    if {!$batch_maxima_asks_with_command} {
      return [list true $name_question]
    }
  }

  set take_part_only false

  if {$with_selected} {
    # if a part is selected, only that part is used;
    # in such a case the positions at error messages do not fit;
    set command_text [list]
    foreach {sel_start sel_after} [$mxmbatch_input_field tag ranges "sel"] {
      lappend command_text [$mxmbatch_input_field get $sel_start $sel_after]
    }
    set command_text [string trim [join $command_text "\n"]]
    if {$command_text != ""} {
      set take_part_only true
    }
  }

  if {$with_whole && (!$take_part_only)} {
    set command_text [string trim [$mxmbatch_input_field get "1.0" "end -1 chars"]]
  }

  if {$command_text==""} {
    show_batch_maxima_message_error [list [list "no command provided for Maxima"]]
    return [list false]
  }

  set mxmbatch_comments_strings_brackets [get_comments_strings_brackets $command_text]

  # check if sane text content
  set brackets_use [lindex $mxmbatch_comments_strings_brackets 2]
  set err_messages [test_brackets_sane $brackets_use [expr {!$take_part_only}]]

  set string_last [lindex [lindex $mxmbatch_comments_strings_brackets 1] "end"]
  if {[llength $string_last]>1} {
    if {[lindex [lindex $string_last 1] 0]=="end"} {
      if {!$take_part_only} {
        lappend err_messages [list "unfinished string" [list [join [lindex $string_last 0] "."] "end"]]
      } {
        lappend err_messages [list "unfinished string"]
      }
    }
  }

  # checking open comments too, since they should be closed for Maxima
  set comment_last [lindex [lindex $mxmbatch_comments_strings_brackets 0] "end"]
  if {[llength $comment_last]>1} {
    if {[lindex [lindex $comment_last 1] 0]=="end"} {
      if {!$take_part_only} {
        lappend err_messages [list "unfinished comment" [list [join [lindex $comment_last 0] "."] "end"]]
      } {
        lappend err_messages [list "unfinished comment"]
      }
    }
  }

  if {[llength $err_messages]>0} {
    show_batch_maxima_message_error $err_messages
    return [list false]
  }

  if {($batch_maxima_asks_names || (!$allow_maxima_system_variables)) && ($command_text != "")} {
    set disallowed_warns [check_setting_presence_in_command $command_text $mxmbatch_comments_strings_brackets false]
    if {[llength $disallowed_warns]>0} {
      show_batch_maxima_message_error $disallowed_warns
      return [list false]
    }
    # since stuff could be set/executed via :lisp(...) constructs
    set disallowed_warns [check_setting_presence_in_command $command_text $mxmbatch_comments_strings_brackets true]
    if {[llength $disallowed_warns]>0} {
      show_batch_maxima_message_error $disallowed_warns
      return [list false]
    }
  }

  if {!$make_line_endings} {
    set end_check [check_command_ending false $command_text [lindex $mxmbatch_comments_strings_brackets 0]]
    if {[lindex $end_check 0]==false} {
      # taking error message in the result (if any), to trigger its display;
      set err_mssg [lindex $end_check 1]
      if {$err_mssg!=""} {
        show_batch_maxima_message_error [list [list $err_mssg]]
      }
      return [list false]
    }
    if {([llength $end_check]>1) && ([lindex $end_check 1]!="")} {
      set command_text [lindex $end_check 1]
    }
  } {
    set command_string_info [put_maxima_command_ends_to_lines $command_text \
      [lindex $mxmbatch_comments_strings_brackets 0] \
      [lindex $mxmbatch_comments_strings_brackets 1] \
      [lindex $mxmbatch_comments_strings_brackets 2] \
    ]
    if {(![lindex $command_string_info 0]) || ([lindex $command_string_info 1]=="")} {
      show_batch_maxima_message_error [list [list "void command"]]
      return [list false]
    }
    set command_text [lindex $command_string_info 1]
  }

  if {$batch_maxima_asks_names} {
    append command_text "\n" $name_question
  }

  return [list true $command_text]
}

proc is_batch_maxima_waiting {} {
  global batch_maxima_running
  global maxima_batch_end_reached

  # had issues when using (!$batch_maxima_end_automatic) in here too;
  if {$batch_maxima_running && $maxima_batch_end_reached} {
    return true
  }

  return false
}

proc is_batch_maxima_running {} {
  global batch_maxima_running
  return $batch_maxima_running
}

proc check_line_valid_for_maxima {line} {
  if {$line==""} {
    return [list false [list [list "empty line"]]]
  }

  set mxmbatch_comments_strings_brackets [get_comments_strings_brackets $line]

  # check if sane text content
  set brackets_use [lindex $mxmbatch_comments_strings_brackets 2]
  set err_messages [test_brackets_sane $brackets_use false]

  set string_last [lindex [lindex $mxmbatch_comments_strings_brackets 1] "end"]
  if {[llength $string_last]>1} {
    if {[lindex [lindex $string_last 1] 0]=="end"} {
      lappend err_messages [list "unfinished string"]
    }
  }

  # checking open comments too, since they should be closed for Maxima
  set comment_last [lindex [lindex $mxmbatch_comments_strings_brackets 0] "end"]
  if {[llength $comment_last]>1} {
    if {[lindex [lindex $comment_last 1] 0]=="end"} {
      lappend err_messages [list "unfinished comment"]
    }
  }

  if {[llength $err_messages]>0} {
    return [list false $err_messages]
  }

  set end_check [check_command_ending false $line [lindex $mxmbatch_comments_strings_brackets 0]]
  if {[lindex $end_check 0]==false} {
    return [list false "effectively empty line"]
  }

  if {([llength $end_check]>1) && ([lindex $end_check 1]!="")} {
    set line [lindex $end_check 1]
  }

  return [list true [strcat $line "\n"]]
}

set mxmbatch_insert_line ""
set mxmbatch_insert_line_info_shown false
proc insert_line_to_batch_maxima {{destroy_insertion true}} {
  global mxmbatch_insert_line
  global batch_maxima_handle
  global maxima_batch_to_end
  global maxima_batch_end_reached
  global mxmbatch_insert_line_info
  global mxmbatch_insert_line_info_shown
  global mxmbatch_insert_line_info_holder
  global print_logging_maxima_input
  global print_logging_maxima_output_last
  global mxmbatch_insert_line_append
  global mxmbatch_output_field

  set line_checked [check_line_valid_for_maxima $mxmbatch_insert_line]
  if {![lindex $line_checked 0]} {
    set err_msgs [list]
    foreach one_err [lindex $line_checked 1] {
      lappend err_msgs [lindex $one_err 0]
    }
    $mxmbatch_insert_line_info configure -text [join $err_msgs "\n"]
    if {!$mxmbatch_insert_line_info_shown} {
      set mxmbatch_insert_line_info_shown true
      pack $mxmbatch_insert_line_info_holder -side top
    }
    return
  }

  $mxmbatch_insert_line_info configure -text ""
  if {$mxmbatch_insert_line_info_shown} {
    pack forget $mxmbatch_insert_line_info_holder
    set mxmbatch_insert_line_info_shown false
  }

  set line_use [lindex $line_checked 1]

  if {$destroy_insertion} {
    destroy_batch_insertion
  }

  if {![is_batch_maxima_running]} {
    show_batch_maxima_message_error [list [list "Batch has already stopped"]]
    return
  }
  if {[is_batch_maxima_asking]} {
    show_batch_maxima_message_error [list [list "Current batch is for names update"]]
    return
  }
  if {$maxima_batch_to_end || $maxima_batch_end_reached} {
    show_batch_maxima_message_error [list [list "Batch has already finished"]]
    return
  }

  if {$mxmbatch_insert_line_append} {
    set curr_state [$mxmbatch_output_field cget -state]
    $mxmbatch_output_field configure -state "normal"
    set append_line_index "end"
    set to_replace_by_line false
    if {[$mxmbatch_output_field get "end -2 chars"] != "\n"} {
      $mxmbatch_output_field insert "end" "\n"
    } {
      if {[$mxmbatch_output_field get "end -3 chars"] == "\n"} {
        set append_line_index "end -1 chars"
        set to_replace_by_line true
      }
    }
    set add_line_char ""
    if {[string range $line_use "end" "end"] != "\n"} {
      set add_line_char "\n"
    }
    if {$to_replace_by_line} {
      $mxmbatch_output_field replace $append_line_index "end" [strcat $line_use $add_line_char]
    } {
      $mxmbatch_output_field insert $append_line_index [strcat $line_use $add_line_char]
    }
    $mxmbatch_output_field mark set "insert" "end"
    $mxmbatch_output_field configure -state $curr_state
    $mxmbatch_output_field edit separator
  }

  set print_logging_maxima_output_last false
  catch {
    puts $batch_maxima_handle $line_use
    flush $batch_maxima_handle
    if {$print_logging_maxima_input} {
      print_logging [strcat "\nXXX insert" "\n"]
      print_logging [strcat $line_use "\n"]
    }
  }

  return
}

proc destroy_batch_insertion {} {
  global mxmbatch_insert_display
  global mxmbatch_output_field

  catch {grab release $mxmbatch_insert_display}
  catch {wm withdraw $mxmbatch_insert_display}
  catch {focus $mxmbatch_output_field}
}

set mxmbatch_insert_display ""
set mxmbatch_insert_display_entry ""
set mxmbatch_insert_line_info ""
set mxmbatch_insert_line_info_holder ""
set mxmbatch_insert_line_append true
proc present_insert_to_batch_maxima {{x_pos ""} {y_pos ""}} {
  global mxmbatch_insert_line
  global mxmbatch_insert_display
  global mxmbatch_insert_display_entry
  global spec_base_font
  global maxima_batch_to_end
  global maxima_batch_end_reached
  global mxmbatch_insert_line_info
  global mxmbatch_insert_line_info_holder
  global mxmbatch_insert_line_append

  if {![is_batch_maxima_running]} {
    show_batch_maxima_message_error [list [list "Batch has already stopped"]]
    return
  }
  if {[is_batch_maxima_asking]} {
    show_batch_maxima_message_error [list [list "Current batch is for names update"]]
    return
  }
  if {$maxima_batch_to_end || $maxima_batch_end_reached} {
    show_batch_maxima_message_error [list [list "Batch has already finished"]]
    return
  }

  remove_torn_menus

  if {$mxmbatch_insert_display==""} {

    set mxmbatch_insert_display [toplevel ".mxmbatch_insert_display" -width 200 -height 200]
    wm title $mxmbatch_insert_display "Line insertion to batch"
    wm minsize $mxmbatch_insert_display 300 50
    wm resizable $mxmbatch_insert_display false false

    if {($x_pos=="") || ($x_pos=="")} {
      set top_pos [get_maxima_batch_output_position]
      set x_pos [expr {15 + [lindex $top_pos 0]}]
      set y_pos [expr {15 + [lindex $top_pos 1]}]
    }
    wm geometry $mxmbatch_insert_display "+$x_pos+$y_pos"

    set lframe [labelframe $mxmbatch_insert_display.labelframe \
      -text "aimed at answering Maxima questions only" -bd 1 -padx 10 -pady 15]
    pack $lframe -expand yes -fill both

    set mxmbatch_insert_display_entry [entry $lframe.entry -width 35 -bd 1 -textvariable mxmbatch_insert_line]
    base_bindings_for_entry $mxmbatch_insert_display_entry "mxmbatch_insert_line"

    if {$spec_base_font(search)==""} {
      $mxmbatch_insert_display_entry configure -font search_font
    } {
      $mxmbatch_insert_display_entry configure -font $spec_base_font(search)
    }

    pack $mxmbatch_insert_display_entry -side top -expand yes -fill x

    set mxmbatch_insert_display_insert_frame [frame $lframe.insert_frame]
    set mxmbatch_insert_display_insert_append [ttk::checkbutton $mxmbatch_insert_display_insert_frame.insert_append \
      -variable mxmbatch_insert_line_append -onvalue true -offvalue false -takefocus 1 -style BatchLine.TCheckbutton -width 0 \
      -text " should append the line into the output? "]
    pack $mxmbatch_insert_display_insert_append -side left
    pack $mxmbatch_insert_display_insert_frame -side top -expand yes -fill x

    frame $lframe.infos -pady 5
    set mxmbatch_insert_line_info_holder $lframe.infos

    set mxmbatch_insert_line_info [label $lframe.infos.line_errs -text ""]
    pack $mxmbatch_insert_line_info -side top

  } {
    destroy_batch_insertion
    catch {wm manage $mxmbatch_insert_display}

    wm deiconify $mxmbatch_insert_display
    raise $mxmbatch_insert_display
  }

  bind $mxmbatch_insert_display <Escape> {
    destroy_batch_insertion
  }
  bind $mxmbatch_insert_display_entry <Return> {
    insert_line_to_batch_maxima true
  }

  wm attributes $mxmbatch_insert_display -topmost true
  wm protocol $mxmbatch_insert_display WM_DELETE_WINDOW {destroy_batch_insertion}

  catch {grab $mxmbatch_insert_display}
  focus $mxmbatch_insert_display_entry
}

set mxmbatch_active_actions_menu ""
proc offer_mxmbatch_active_actions {x_abs y_abs x_pos y_pos} {
  global topw
  global mxmbatch_active_actions_menu
  global maxima_batch_to_end
  global maxima_batch_end_reached
  global default_menu_offset_x
  global default_menu_offset_y

  if {![is_batch_maxima_running]} {
    return
  }
  if {[is_batch_maxima_asking]} {
    return
  }

  if {($x_abs=="") || ($y_abs=="")} {
    set abs_pos [get_maxima_batch_output_position]
    set use_x [expr {$x_pos + [lindex $abs_pos 0] + $default_menu_offset_x}]
    set use_y [expr {$y_pos + [lindex $abs_pos 1] + $default_menu_offset_y}]
  } {
    set use_x [expr {$x_abs + $default_menu_offset_x}]
    set use_y [expr {$y_abs + $default_menu_offset_y}]
  }

  if {$mxmbatch_active_actions_menu!=""} {
    catch {destroy $mxmbatch_active_actions_menu}
  }
  set mxmbatch_active_actions_menu [menu "$topw.mxmbatch_active_actions_menu" -tearoff false]
  $mxmbatch_active_actions_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $mxmbatch_active_actions_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $mxmbatch_active_actions_menu configure -activeborderwidth 1
  $mxmbatch_active_actions_menu configure -title "active batch menu"

  if {(!$maxima_batch_to_end) && (!$maxima_batch_end_reached)} {
    set label_text "if Maxima asks a question (e.g. on sign of a variable)"
    $mxmbatch_active_actions_menu add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
    $mxmbatch_active_actions_menu add command -label "push line to the current batch" -command [list "present_insert_to_batch_maxima" [expr {$use_x - 50}] [expr {$use_y - 15}]]
  }

  set label_text "if Maxima batch does not stop"
  $mxmbatch_active_actions_menu add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
  if {![is_batch_maxima_waiting]} {
    $mxmbatch_active_actions_menu add command -label "stop the current Maxima batch" -command "stop_batch_maxima"
  } {
    $mxmbatch_active_actions_menu add command -label "stop the current Maxima process" -command "stop_batch_maxima"
  }

  tk_popup $mxmbatch_active_actions_menu $use_x $use_y
}

proc update_batch_maxima_prepend_mode {to_display} {
  global command_onset_use_along_commands
  set command_onset_use_along_commands $to_display
}

set mxmbatch_prepend_mode_menu ""
proc offer_prepend_mode_batch_maxima {x_abs y_abs x_pos y_pos} {
  global topw
  global mxmbatch_prepend_mode_menu
  global command_onset_use_along_commands
  global mxmbatch_input_prepend_use
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  if {$mxmbatch_prepend_mode_menu!=""} {
    catch {destroy $mxmbatch_prepend_mode_menu}
  }
  set mxmbatch_prepend_mode_menu [menu "$topw.mxmbatch_prepend_mode_menu" -tearoff false]
  $mxmbatch_prepend_mode_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $mxmbatch_prepend_mode_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $mxmbatch_prepend_mode_menu configure -activeborderwidth 1
  $mxmbatch_prepend_mode_menu configure -title "prepend command mode"

  set label_text [strcat "The prepend command" [expr {$mxmbatch_input_prepend_use ? "" : ", once is enabled"}] ":"]
  $mxmbatch_prepend_mode_menu add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  $mxmbatch_prepend_mode_menu add command -label "to be used within initialization stuff only" -command [list "update_batch_maxima_prepend_mode" false]
  $mxmbatch_prepend_mode_menu add command -label "to be used (and displayed) along other commands" -command [list "update_batch_maxima_prepend_mode" true]

  set used_idx [expr {$command_onset_use_along_commands ? 2 : 1}]
  $mxmbatch_prepend_mode_menu entryconfigure $used_idx -background $menu_selected_background -activebackground "#e0e0ff"

  $mxmbatch_prepend_mode_menu add command -label "hot-start batches use it in the \"along\" case only" -state "disabled" -background "#e0f0e0" -activebackground "#e0f0e0"

  tk_popup $mxmbatch_prepend_mode_menu $use_x $use_y
}

proc reset_batch_line_counters {to_text} {
  global mxmbatch_output_field
  global maxima_batch_onset_string
  global maxima_batch_onset_string_row_count
  global maxima_batch_start_row
  global maxima_batch_wrote_row_count_at_text
  global batch_maxima_asks_received
  global maxima_batch_asked_to_write_to_text

  set maxima_batch_onset_string ""
  set maxima_batch_onset_string_row_count 0
  set maxima_batch_start_row [lindex [split [$mxmbatch_output_field index "end"] "."] 0]
  set maxima_batch_wrote_row_count_at_text 0
  set batch_maxima_asks_received ""

  if {$to_text} {
    set maxima_batch_asked_to_write_to_text false
  }
}

proc set_bt_kill_for_interrupt {} {
  global bt_kill
  global use_buttons_text
  #global icon_image_mm
  global icon_image_hg

  if {$use_buttons_text} {
    #$bt_kill configure -text "X"
    $bt_kill configure -text "W"
  } {
    #$bt_kill configure -image $icon_image_mm
    $bt_kill configure -image $icon_image_hg
  }
}

proc set_bt_kill_for_stop {} {
  global bt_kill
  global use_buttons_text
  global icon_image_bm

  if {$use_buttons_text} {
    $bt_kill configure -text "K"
  } {
    $bt_kill configure -image $icon_image_bm
  }
}

proc setup_waiting_batch_maxima {} {
  global state_label
  global mxmbatch_start
  global mxmbatch_stop
  global mxmbatch_break
  global batch_maxima_asks_names
  global should_display_cpu_usage
  global could_display_cpu_usage
  global maxima_batch_last_output_line_string

  set maxima_batch_last_output_line_string ""

  if {[is_batch_maxima_asking]} {
    names_start_batch_maxima_process
  }
  set batch_maxima_asks_names false

  reset_batch_line_counters true
  reset_cpu_usage
  set should_display_cpu_usage false
  set could_display_cpu_usage true

  set_state_label_text "prepared"

  $mxmbatch_start configure -state "normal"
  $mxmbatch_stop configure -state "normal"
  $mxmbatch_break configure -state "disabled"

  set_bt_kill_for_stop

  state_batch_maxima
}

proc fill_batch_batch_maxima {{command_onset ""} {command_string ""}} {
  global batch_maxima_input_tmpfile_mac
  global info_line_left
  global info_line_long
  global info_line_1dim
  global info_line_inner
  global info_line_tex
  global info_tag_start
  global info_tag_middle
  global info_tag_end
  global with_labels
  global bind_labels
  global info_line_size
  global show_time_elapsed
  global batch_maxima_start_shibboleth
  global command_onset_use_along_commands
  global batch_maxima_end_shibboleth
  global batch_maxima_asks_names
  global print_logging_maxima_input
  global print_logging_maxima_output_last
  global asked_temporary_path

  if {$batch_maxima_input_tmpfile_mac==""} {
    return false
  }

  set use_line_left $info_line_left
  set use_line_long $info_line_long
  set use_line_1dim $info_line_1dim
  set use_line_inner $info_line_inner
  set use_line_tex $info_line_tex

  if {$info_line_1dim} {
    set use_line_left true
    # some forms get split by Maxima, some are dealt with Minima;
    #set use_line_long true
  } {
    set use_line_long false
  }

  if {$info_line_inner || $info_line_tex} {
    set use_line_1dim true
    set use_line_left true
    set use_line_long true
  }

  if {$batch_maxima_asks_names} {
    set use_line_1dim true
    set use_line_left true
    set use_line_long true
    set use_line_inner false
    set use_line_tex false
  }

  set batch_maxima_input_inner [list]

  # these are done at the outer parts:
  # "reset_displays()\$"
  # ":lisp(setf *display-labels-p* ...)"

  # notice that SBCL breaks on "?..." names in batches
  # (due to a SBCL bug in concatenated streams: #690408),
  # thus generally have to use ":lisp(...)" ways here;
  # and along with that, ":lisp(...)" lines should be
  # without ending semicolons for the batch-wise cases;

  lappend batch_maxima_input_inner "linel:90\$"

  if {$asked_temporary_path!=""} {
    lappend batch_maxima_input_inner [strcat \
      "maxima_tempdir:\"" [string map [list "\\" "\\\\" "\"" "\\\""] $asked_temporary_path] "\";" \
    ]
  }

  set nolabels_value [expr {$bind_labels ? "false" : "true"}]
  lappend batch_maxima_input_inner "nolabels:$nolabels_value\$"

  set leftjust_value [expr {$use_line_left ? "true" : "false"}]
  lappend batch_maxima_input_inner "leftjust:$leftjust_value\$"

  set display2d_value [expr {$use_line_1dim ? "false" : "true"}]
  lappend batch_maxima_input_inner "display2d:$display2d_value\$"

  lappend batch_maxima_input_inner [strcat "linel:" [lindex $info_line_size [expr {$use_line_long ? 1 : 0}]] "\$"]

  set showtime_value [expr {$show_time_elapsed ? "true" : "false"}]
  lappend batch_maxima_input_inner "showtime:$showtime_value\$"

  set batch_maxima_start_shibboleth_a "so"
  append batch_maxima_start_shibboleth_a [expr {int(10000*rand()+1000)}]
  set batch_maxima_start_shibboleth_b "so"
  append batch_maxima_start_shibboleth_b [expr {int(10000*rand()+1000)}]
  set batch_maxima_start_shibboleth_c "so"
  append batch_maxima_start_shibboleth_c [expr {int(10000*rand()+1000)}]

  set batch_maxima_start_shibboleth [strcat \
    $batch_maxima_start_shibboleth_a \
    $batch_maxima_start_shibboleth_b \
    $batch_maxima_start_shibboleth_c]
  set shibboleth_cmmd [strcat \
    "grind" "(concat" "(\"$batch_maxima_start_shibboleth_a\"," \
    "\"$batch_maxima_start_shibboleth_b\"," \
    "\"$batch_maxima_start_shibboleth_c\"))" \
  ]

  set tex_disp_mode_wl "(lambda (form) (format t \"(~a) \$\$~a\$\$~%\" (string-downcase (subseq (format nil \"~a\" (cadr form)) 1)) (\$tex1 (caddr form))))"
  set tex_disp_mode_wol "(lambda (form) (format t \"\$\$~a\$\$~%\" (\$tex1 (caddr form))))"

  if {$use_line_tex} {
    if {$with_labels} {
      lappend batch_maxima_input_inner [strcat ":lisp(setf *alt-display1d* " $tex_disp_mode_wl ")"]
    } {
      lappend batch_maxima_input_inner [strcat ":lisp(setf *alt-display1d* " $tex_disp_mode_wol ")"]
    }
  } elseif {$use_line_inner} {
    set alt_disp_mode_wl "(setf *alt-display1d* (lambda (form) (format t \"(~a) ~a~%\" (string-downcase (subseq (format nil \"~a\" (cadr form)) 1)) (caddr form))))"
    set alt_disp_mode_wol "(setf *alt-display1d* (lambda (form) (format t \"~a~%\" (caddr form))))"
    set alt_disp_mode [expr {$with_labels ? $alt_disp_mode_wl : $alt_disp_mode_wol}]
    lappend batch_maxima_input_inner [strcat ":lisp" $alt_disp_mode]
  }

  if {!$command_onset_use_along_commands} {
    if {$command_onset!=""} {
      lappend batch_maxima_input_inner $command_onset
    }
  }

  lappend batch_maxima_input_inner "($shibboleth_cmmd,kill(labels))\$"

  if {$command_onset_use_along_commands} {
    if {$command_onset!=""} {
      lappend batch_maxima_input_inner $command_onset
    }
  }

  lappend batch_maxima_input_inner $command_string

  set batch_maxima_end_shibboleth "so"
  append batch_maxima_end_shibboleth [expr {int(10000*rand()+1000)}]
  append batch_maxima_end_shibboleth "so"
  append batch_maxima_end_shibboleth [expr {int(10000*rand()+1000)}]
  append batch_maxima_end_shibboleth "so"
  append batch_maxima_end_shibboleth [expr {int(10000*rand()+1000)}]

  lappend batch_maxima_input_inner ""
  lappend batch_maxima_input_inner [strcat "\"" $batch_maxima_end_shibboleth "\";"]

  # considering that the ending-option may be changed before the batch finishes;
  #if {$batch_maxima_end_automatic} {
  #  lappend batch_maxima_input_inner "quit()\$"
  #  lappend batch_maxima_input_inner "quit()\$"
  #  lappend batch_maxima_input_inner "quit()\$"
  #}

  set print_logging_maxima_output_last false
  if {[catch {
    set tmp_id [open $batch_maxima_input_tmpfile_mac "w"]
    puts $tmp_id [join $batch_maxima_input_inner "\n"]
    flush $tmp_id
    close $tmp_id

    if {$print_logging_maxima_input} {
      print_logging [strcat "\nXXX batch" "\n"]
      print_logging [strcat [join $batch_maxima_input_inner "\n"] "\n"]
    }
  }]} {
    catch {file delete -- $batch_maxima_input_tmpfile_mac}
    show_batch_maxima_message_error [list [list "can not fill batch file for Maxima"]]
    return false
  }

  return true
}

# not trying to deal with error messages that get buried within Maxima by now
#set mxm_batch_error_message_part ""
#set err_tag_start "|\[\[\["
#set err_tag_end "\]\]\]|"

set maxima_batch_is_continuation false
proc continue_batch_maxima {with_selected with_whole {use_command ""}} {
  global batch_maxima_end_shibboleth
  global batch_maxima_start_shibboleth
  global command_onset_use_along_commands
  global maxima_batch_to_end
  global maxima_batch_end_reached
  global batch_maxima_handle
  global batch_maxima_input_tmpfile_mac
  global with_labels
  global maxima_batch_wrote_row_count_at_text
  global state_label
  global bt_kill
  global batch_maxima_asks_names
  global command_history
  global mxmbatch_start
  global mxmbatch_stop
  global mxmbatch_break
  global mxmbatch_output_field
  global maxima_batch_just_started_read
  global maxima_batch_is_continuation
  global priority_wine_common
  global priority_wine_idle_set
  global minima_uses_twapi
  global maxima_pid_batch
  global should_display_cpu_usage
  global could_display_cpu_usage
  global batch_maxima_end_cleaning
  global print_logging_maxima_input
  global print_logging_maxima_output_last
  global mxm_display_exp_dstar
  global mxm_display_sorted_ratsimp
  global maxima_batch_last_output_line_string

  if {![is_batch_maxima_waiting]} {
    return false
  }
  if {$batch_maxima_handle==""} {
    return false
  }

  set maxima_batch_last_output_line_string ""

  # onset only used in continuation, if put along commands;
  set command_onset ""
  if {($command_onset_use_along_commands)} {
    set command_onset_info [command_batch_maxima_prepend false]
    if {![lindex $command_onset_info 0]} {
      return false
    }
    set command_onset [lindex $command_onset_info 1]
  }
  set command_onset_info ""

  set command_string ""
  if {$with_selected || $with_whole} {
    set command_string_info [command_batch_maxima $with_selected $with_whole]
    if {![lindex $command_string_info 0]} {
      return false
    }
    set command_string [lindex $command_string_info 1]
    set command_string_info ""
  }
  if {($command_string=="") && ($use_command=="")} {
    # such a call should not be done;
    return false
  }

  reset_batch_line_counters false

  if {($command_string!="") && ($use_command!="")} {
    append command_string "\n" $use_command
  } elseif {$command_string==""} {
    set command_string $use_command
  }

  if {![fill_batch_batch_maxima $command_onset $command_string]} {
    return false
  }

  catch {
    fileevent $batch_maxima_handle readable ""
  }

  set batch_maxima_input_outer [list]
  lappend batch_maxima_input_outer "1;/*"
  lappend batch_maxima_input_outer "(to-maxima)"
  lappend batch_maxima_input_outer "?*/1;"
  lappend batch_maxima_input_outer "linel:200\$"
  lappend batch_maxima_input_outer [strcat ":lisp" "(setf *alt-display1d* nil)" ""]
  lappend batch_maxima_input_outer [strcat ":lisp" "(setf *alt-display2d* nil)" ""]
  lappend batch_maxima_input_outer "ttyoff:false\$"
  if {$batch_maxima_end_cleaning} {
    lappend batch_maxima_input_outer "kill(all)\$"
    lappend batch_maxima_input_outer "for cont in contexts do killcontext(cont)\$"
    lappend batch_maxima_input_outer "remlet(all)\$"
    # Maxima apparently does not have a way how to reasonably remove all the rules, properties, etc.
  }

  lappend batch_maxima_input_outer [strcat ":lisp" "(setf *display-labels-p* " [expr {$with_labels ? "t" : "nil"}] ")"]

  if {$batch_maxima_end_cleaning} {
    if {[is_wine_like]} {
      lappend batch_maxima_input_outer ":lisp(if (string-equal *maxima-lispname* \"sbcl\")(setf (getf *plot-options* :plot_format) '\$gnuplot_pipes))"
    }
  }

  lappend batch_maxima_input_outer ":lisp(defun add-lineinfo (lis)(append lis (list 'batch)))"

  lappend batch_maxima_input_outer ":lisp(if (or (string-equal *maxima-lispname* \"sbcl\")(string-equal *maxima-lispname* \"cmucl\"))(setf *query-io* *standard-input*))"

  if {$mxm_display_exp_dstar} {
    lappend batch_maxima_input_outer ":lisp(setf (get 'mexpt 'grind) 'msize-infix)"
  } {
    lappend batch_maxima_input_outer ":lisp(setf (get 'mexpt 'grind) 'msz-mexpt)"
  }

  if {$mxm_display_sorted_ratsimp} {
    # while common Maxima throw-actions use 'mcatch (caught explicitly below),
    # rat functions use 'rat-err (that is caught implicitly, by rat-error-to-merror);
    lappend batch_maxima_input_outer [strcat ":lisp(defun linear-displa (form) " \
      "(fresh-line *standard-output*)" \
      "(mgrind " \
      "  (handler-case " \
      "    (let ((\$ratprint nil)(\$keepfloat t)(mcatch (cons bindlist loclist))(res nil)(aux nil))" \
      "      (progn" \
      "        (catch 'mcatch" \
      "          (rat-error-to-merror (progn" \
      "            (setf res (\$ratsimp form))" \
      "            (setf aux t)" \
      "          ))" \
      "        )" \
      "        (if aux res form)" \
      "      )" \
      "    )" \
      "    (condition () form)(error () form)" \
      "  )" \
      "*standard-output*)" \
      "(mterpri))" ]
  } {
    lappend batch_maxima_input_outer ":lisp(defun linear-displa (form) (fresh-line *standard-output*)(mgrind form *standard-output*)(mterpri))"
  }

  set batch_file_name_write $batch_maxima_input_tmpfile_mac
  set ending_unit [strcat "\"" $batch_maxima_end_shibboleth "\",?force\\-output(),?\\*alt\\-display1d\\*:false,?\\*alt\\-display2d\\*:false,ttyoff:false,linel:80,?clear\\-input(),?clear\\-input(?\\*query\\-io\\*)," "\"" $batch_maxima_end_shibboleth "\""]

  lappend batch_maxima_input_outer [strcat \
    "if (\{if errcatch(catch(batch(\"$batch_file_name_write\")))=\[\] then " \
    "($ending_unit) else ($ending_unit)\}=false) then ($ending_unit) else ($ending_unit);"]

  set print_logging_maxima_output_last false
  if {[catch {
    puts $batch_maxima_handle [strcat [join $batch_maxima_input_outer "\n"]]
    flush $batch_maxima_handle
    if {$print_logging_maxima_input} {
      print_logging [strcat "\nXXX input" "\n"]
      print_logging [strcat [join $batch_maxima_input_outer "\n"] "\n"]
    }
  }]} {
    stop_batch_maxima
    show_batch_maxima_message_error [list [list "can not start new Maxima batch"]]
    return false
  }

  set maxima_batch_is_continuation true
  set maxima_batch_just_started_read true
  set maxima_batch_end_reached false

  if {[catch {
    fileevent $batch_maxima_handle readable "read_maxima_batch"
  }]} {
    stop_batch_maxima
    show_batch_maxima_message_error [list [list "can not read from the new Maxima batch"]]
    return false
  }

  if {[lindex [split [$mxmbatch_output_field index "end -1 chars"] "."] 1]!="0"} {
    set cur_state [$mxmbatch_output_field cget -state]
    $mxmbatch_output_field configure -state "normal"
    $mxmbatch_output_field insert "end" "\n" [list]
    $mxmbatch_output_field configure -state $cur_state
  }

  $mxmbatch_start configure -state "disabled"
  $mxmbatch_stop configure -state "normal"
  $mxmbatch_break configure -state "normal"

  if {!$batch_maxima_asks_names} {
    set curr_time [clock format [clock seconds] -gmt true -format "%Y-%m-%d %H:%M:%S"]
    lappend command_history [list "$curr_time GMT" $command_string]
  }

  set maxima_batch_wrote_row_count_at_text 0
  $bt_kill configure -state "normal"
  if {[has_common_ctrlc]} {
    set_bt_kill_for_interrupt
  }

  reset_cpu_usage
  set_ignore_cpu_usage false

  if {[is_batch_maxima_asking]} {
    set should_display_cpu_usage false
    set could_display_cpu_usage true
    set_state_label_text "updating"
  } {
    set should_display_cpu_usage true
    set could_display_cpu_usage true
    set_state_label_text "batch running"
  }

  if {[is_wine_like] && ($maxima_pid_batch!="") && $minima_uses_twapi} {
    catch {
      twapi::set_priority_class $maxima_pid_batch $priority_wine_common
      set priority_wine_idle_set false
    }
  }

  state_batch_maxima
  return true
}

set batch_maxima_running false
set batch_maxima_handle ""
set batch_maxima_ids [list]
set batch_maxima_input_tmpfile_mac ""
set batch_maxima_output_tmpfile_tmp ""
set batch_maxima_output_id_write ""
set batch_maxima_output_id_read ""
set maxima_batch_just_started_read false
set maxima_batch_ever_started false
set batch_maxima_output_path_last_int_start 10
set batch_maxima_output_path_last_int $batch_maxima_output_path_last_int_start
set batch_maxima_output_check_id ""
set batch_maxima_state_check_id ""
set batch_maxima_using_alt_display false
set batch_maxima_start_shibboleth ""
proc start_batch_maxima {{with_selected false} {with_whole true} {use_command ""}} {
  global mxmbatch_output_field
  global batch_maxima_running
  global maxima_path
  global maxima_version_to_use
  global batch_maxima_ids
  global batch_maxima_handle
  global maxima_batch_pid_present
  global maxima_pid_batch
  global mxmbatch_start
  global mxmbatch_stop
  global mxmbatch_break
  global batch_maxima_input_tmpfile_mac
  global batch_maxima_output_tmpfile_tmp
  global batch_maxima_output_id_write
  global batch_maxima_output_id_read
  global maxima_batch_just_started_read
  global maxima_batch_ever_started
  global batch_maxima_output_check_id
  global batch_maxima_state_check_id
  global info_tag_start
  global info_tag_middle
  global info_tag_end
  global lisp_flavor_to_use
  global maxima_userdir_to_set
  global maxima_initname_to_set
  global test_suffixes_maxima
  global test_suffixes_lisp
  global with_labels
  global batch_maxima_using_alt_display
  global batch_maxima_window
  global batch_maxima_asks_received
  global batch_maxima_start_shibboleth
  global batch_maxima_asks_names
  global command_history
  global command_onset_use_along_commands
  global maxima_batch_wrote_row_count_at_text
  global bt_kill
  global state_label
  global batch_maxima_end_shibboleth
  global maxima_batch_to_end
  global maxima_batch_end_reached
  global maxima_batch_is_continuation
  global priority_wine_common
  global priority_wine_idle_set
  global minima_uses_twapi
  global should_display_cpu_usage
  global could_display_cpu_usage
  global print_logging_maxima_input
  global print_logging_maxima_output_last
  global mxm_display_exp_dstar
  global mxm_display_sorted_ratsimp
  global outer_rnd_part_identifier
  global maxima_batch_last_output_line_string

  if {$batch_maxima_window==""} {
    return false
  }

  update_batch_maxima_properties

  # if a job is running, not continuing;
  # should stop a job explicitly, if want it;
  if {[is_batch_maxima_running]} {
    return false
  }

  set maxima_batch_is_continuation false

  # not putting comments in, since first, they are not visible in output of batches,
  # and second, they may cause issues when e.g. :lisp(.....) command follows;
  set command_onset_info [command_batch_maxima_prepend false]
  if {![lindex $command_onset_info 0]} {
    return false
  }
  set command_onset [lindex $command_onset_info 1]
  set command_onset_info ""

  set command_string ""
  if {$with_selected || $with_whole} {
    set command_string_info [command_batch_maxima $with_selected $with_whole]
    if {![lindex $command_string_info 0]} {
      return false
    }
    set command_string [lindex $command_string_info 1]
    set command_string_info ""
  }
  if {($command_string=="") && ($use_command=="")} {
    # such a call should not be done;
    return false
  }

  if {($command_string!="") && ($use_command!="")} {
    append command_string "\n" $use_command
  } elseif {$command_string==""} {
    set command_string $use_command
  }

  set maxima_batch_last_output_line_string ""

  set maxima_batch_to_end false
  set maxima_batch_end_reached false
  set batch_maxima_asks_received ""

  # using (simple) input-to-read approach here
  set batch_maxima_input_tmpfile_mac ""
  set batch_maxima_output_tmpfile_tmp ""
  set batch_maxima_output_id_write ""
  set batch_maxima_output_id_read ""
  set batch_maxima_ids [list]
  set batch_maxima_output_check_id ""
  set batch_maxima_state_check_id ""
  set outer_rnd_part_identifier ""

  catch {
    set tmp_id_info [get_tempfile "mac"]
    set tmp_id [lindex $tmp_id_info 0]
    set batch_maxima_input_tmpfile_mac [lindex $tmp_id_info 1]
    if {$tmp_id != ""} {
      close $tmp_id
    }
  }
  if {$batch_maxima_input_tmpfile_mac==""} {
    show_batch_maxima_message_error [list [list "can not create batch file for Maxima"]]
    return false
  }

  if {![file exists $batch_maxima_input_tmpfile_mac]} {
    set batch_maxima_input_tmpfile_mac ""
    show_batch_maxima_message_error [list [list "could not create batch file for Maxima"]]
    return false
  }

  if {[file extension $batch_maxima_input_tmpfile_mac]!=".mac"} {
    catch {file delete -- $batch_maxima_input_tmpfile_mac}
    show_batch_maxima_message_error [list [list "could not name batch file for Maxima"]]
    return false
  }

  if {![fill_batch_batch_maxima $command_onset $command_string]} {
    return false
  }

  set batch_maxima_input_outer [list]
  lappend batch_maxima_input_outer "1;/*"
  lappend batch_maxima_input_outer "(to-maxima)"
  lappend batch_maxima_input_outer "?*/1;"
  lappend batch_maxima_input_outer "linel:200\$"
  lappend batch_maxima_input_outer [strcat ":lisp" "(setf *alt-display1d* nil)" ""]
  lappend batch_maxima_input_outer [strcat ":lisp" "(setf *alt-display2d* nil)" ""]
  lappend batch_maxima_input_outer "ttyoff:false\$"

  lappend batch_maxima_input_outer ":lisp(\$disp (\$concat \" pid=\" (getpid) \" \"))"
  lappend batch_maxima_input_outer [strcat ":lisp" "(setf *display-labels-p* " [expr {$with_labels ? "t" : "nil"}] ")"]

  if {[is_wine_like]} {
    lappend batch_maxima_input_outer ":lisp(if (string-equal *maxima-lispname* \"sbcl\") (setf (getf *plot-options* :plot_format) '\$gnuplot_pipes))"
  }
  lappend batch_maxima_input_outer ":lisp(defun add-lineinfo (lis)(append lis (list 'batch)))"

  lappend batch_maxima_input_outer "grind(concat(\"$info_tag_start\",build_info()@version,\"$info_tag_middle\",build_info()@lisp_name,\"$info_tag_end\"))\$"
  lappend batch_maxima_input_outer ":lisp(if (or (string-equal *maxima-lispname* \"sbcl\")(string-equal *maxima-lispname* \"cmucl\"))(setf *query-io* *standard-input*))"

  if {$mxm_display_exp_dstar} {
    lappend batch_maxima_input_outer ":lisp(setf (get 'mexpt 'grind) 'msize-infix)"
  } {
    lappend batch_maxima_input_outer ":lisp(setf (get 'mexpt 'grind) 'msz-mexpt)"
  }

  if {$mxm_display_sorted_ratsimp} {
    # while common Maxima throw-actions use 'mcatch (caught explicitly below),
    # rat functions use 'rat-err (that is caught implicitly, by rat-error-to-merror);
    lappend batch_maxima_input_outer [strcat ":lisp(defun linear-displa (form) " \
      "(fresh-line *standard-output*)" \
      "(mgrind " \
      "  (handler-case " \
      "    (let ((\$ratprint nil)(\$keepfloat t)(mcatch (cons bindlist loclist))(res nil)(aux nil))" \
      "      (progn" \
      "        (catch 'mcatch" \
      "          (rat-error-to-merror (progn" \
      "            (setf res (\$ratsimp form))" \
      "            (setf aux t)" \
      "          ))" \
      "        )" \
      "        (if aux res form)" \
      "      )" \
      "    )" \
      "    (condition () form)(error () form)" \
      "  )" \
      "*standard-output*)" \
      "(mterpri))" ]
  } {
    lappend batch_maxima_input_outer ":lisp(defun linear-displa (form) (fresh-line *standard-output*)(mgrind form *standard-output*)(mterpri))"
  }

  set ending_unit [strcat "\"" $batch_maxima_end_shibboleth "\",?force\\-output(),?\\*alt\\-display1d\\*:false,?\\*alt\\-display2d\\*:false,ttyoff:false,linel:80,?clear\\-input(),?clear\\-input(?\\*query\\-io\\*)," "\"" $batch_maxima_end_shibboleth "\""]

  set batch_file_name_write $batch_maxima_input_tmpfile_mac

  lappend batch_maxima_input_outer [strcat \
    "if (\{if errcatch(catch(batch(\"$batch_file_name_write\")))=\[\] then " \
    "($ending_unit) else ($ending_unit)\}=false) then ($ending_unit) else ($ending_unit);"]

  set batch_maxima_output_tmpfile_tmp ""
  catch {
    set tmp_id_info [get_tempfile]
    set tmp_id [lindex $tmp_id_info 0]
    set batch_maxima_output_tmpfile_tmp [lindex $tmp_id_info 1]
    if {$tmp_id != ""} {
      close $tmp_id
    }
  }
  if {$batch_maxima_output_tmpfile_tmp==""} {
    catch {file delete -- $batch_maxima_input_tmpfile_mac}
    show_batch_maxima_message_error [list [list "can not create output file for Maxima"]]
    return false
  }

  if {[catch {
    set batch_maxima_output_id_write [open $batch_maxima_output_tmpfile_tmp "w"]
  }]} {
    catch {file delete -- $batch_maxima_input_tmpfile_mac}
    catch {file delete -- $batch_maxima_output_tmpfile_tmp}
    show_batch_maxima_message_error [list [list "can not prepare output file for Maxima"]]
    return false
  }

  if {[catch {
    set batch_maxima_output_id_read [open $batch_maxima_output_tmpfile_tmp "r"]
  }]} {
    catch {file delete -- $batch_maxima_input_tmpfile_mac}
    catch {file delete -- $batch_maxima_output_tmpfile_tmp}
    catch {close $batch_maxima_output_id_write}
    show_batch_maxima_message_error [list [list "can not prepare output file for Maxima"]]
    return false
  }

  # the flash-error mode is only available for the stack mode;
  # here redirecting Maxima stderr into the null device;
  set print_logging_maxima_output_last false
  if {[catch {
    set null_path [get_null_path]

    set maxima_startup_info [get_maxima_startup_base]
    set maxima_startup_direct [lindex $maxima_startup_info 0]
    set maxima_startup_base [lindex $maxima_startup_info 1]

    if {[is_under_wine]} {
      set outer_rnd_part_identifier [expr {int(10000*(rand()+1.0))}]
    }

    set batch_maxima_start_string [strcat \
      $maxima_startup_base \
      [expr {($outer_rnd_part_identifier!="") ? " --rnd-identifier $outer_rnd_part_identifier" : ""} ] \
      [expr {[is_wine_like] ? "" : " 2>$null_path"}] \
      [expr {$maxima_startup_direct ? " &" : ""}] \
    ]
    set batch_maxima_handle [open $batch_maxima_start_string "r+"]
    if {$print_logging_maxima_input} {
      print_logging [strcat "\nXXX start" "\n"]
      print_logging [strcat $batch_maxima_start_string "\n"]
    }

    set batch_maxima_ids [pid $batch_maxima_handle]
    set maxima_batch_pid_present [lindex $batch_maxima_ids "end"]
    puts $batch_maxima_handle [strcat [join $batch_maxima_input_outer "\n"]]
    flush $batch_maxima_handle
    if {$print_logging_maxima_input} {
      print_logging [strcat "\nXXX input" "\n"]
      print_logging [strcat [join $batch_maxima_input_outer "\n"] "\n"]
    }
  }]} {
    set outer_rnd_part_identifier ""
    stop_batch_maxima
    show_batch_maxima_message_error [list [list "can not start Maxima batch"]]
    return false
  }

  if {[is_wine_like] && $minima_uses_twapi} {
    foreach one_pid $batch_maxima_ids {
      catch {
        twapi::set_priority_class $one_pid $priority_wine_common
        set priority_wine_idle_set false
      }
    }
  }

  catch {
    chan configure $batch_maxima_handle -blocking false
    chan configure $batch_maxima_handle -encoding "utf-8"
    chan configure $batch_maxima_handle -translation lf
    # we do explicit flushing when writing to maxima
  }

  fileevent $batch_maxima_handle readable "read_maxima_batch"

  set maxima_batch_just_started_read true
  set maxima_batch_ever_started true

  if {[lindex [split [$mxmbatch_output_field index "end -1 chars"] "."] 1]!="0"} {
    set cur_state [$mxmbatch_output_field cget -state]
    $mxmbatch_output_field configure -state "normal"
    $mxmbatch_output_field insert "end" "\n" [list]
    $mxmbatch_output_field configure -state $cur_state
  }

  $mxmbatch_start configure -state "disabled"
  $mxmbatch_stop configure -state "normal"
  $mxmbatch_break configure -state "normal"

  if {!$batch_maxima_asks_names} {
    set curr_time [clock format [clock seconds] -gmt true -format "%Y-%m-%d %H:%M:%S"]
    lappend command_history [list "$curr_time GMT" $command_string]
  }

  set maxima_batch_wrote_row_count_at_text 0
  $bt_kill configure -state "normal"
  if {[has_common_ctrlc]} {
    set_bt_kill_for_interrupt
  }

  reset_cpu_usage
  set_ignore_cpu_usage false

  if {![is_under_wine]} {
    start_cpu_usage_updates
  } {
    start_cpu_usage_updates_under_wine
  }

  if {[is_batch_maxima_asking]} {
    set should_display_cpu_usage false
    set could_display_cpu_usage true
    set_state_label_text "updating"
  } {
    set should_display_cpu_usage true
    set could_display_cpu_usage true
    set_state_label_text "batch running"
  }

  set maxima_pid_batch $maxima_batch_pid_present
  set batch_maxima_running true
  state_batch_maxima
  return true
}

proc break_batch_maxima {} {
  if {[is_batch_maxima_waiting]} {
    return
  }

  if {[has_common_ctrlc]} {
    do_kill_maxima false
  }
}

proc stop_batch_maxima {{overall false}} {
  global batch_maxima_window
  global mxmbatch_start
  global mxmbatch_stop
  global mxmbatch_break
  global batch_maxima_running
  global batch_maxima_handle
  global batch_maxima_ids
  global batch_maxima_output_check_id
  global batch_maxima_state_check_id
  global batch_maxima_input_tmpfile_mac
  global batch_maxima_output_tmpfile_tmp
  global batch_maxima_output_id_write
  global batch_maxima_output_id_read
  global maxima_batch_onset_string
  global maxima_batch_onset_string_row_count
  global maxima_pid_batch
  global batch_maxima_asks_names
  global batch_maxima_last_command
  global mxmbatch_input_field
  global bt_kill
  global state_label
  global maxima_batch_asked_to_write_to_text
  global maxima_batch_end_reached
  global priority_wine_idle
  global priority_wine_idle_set
  global minima_uses_twapi
  global should_display_cpu_usage
  global could_display_cpu_usage
  global print_logging_maxima_input
  global print_logging_maxima_output_last
  global maxima_batch_last_output_line_string

  # have to go through the function even when apparently not running,
  # since it can be in a partially-successful attempt to make a start;

  # if the batch_maxima stuff was not setup yet, it does not run either;
  if {$batch_maxima_window==""} {
    return
  }

  if {[is_wine_like] && $minima_uses_twapi} {
    if {$maxima_pid_batch!=""} {
      catch {
        twapi::set_priority_class $maxima_pid_batch $priority_wine_idle
        set priority_wine_idle_set true
      }
    }
    foreach one_pid $batch_maxima_ids {
      catch {
        twapi::set_priority_class $one_pid $priority_wine_idle
        set priority_wine_idle_set true
      }
    }
  }

  set maxima_batch_last_output_line_string ""

  reset_batch_line_counters true

  set maxima_pid_batch ""

  if {$batch_maxima_output_check_id != ""} {
    catch {after cancel $batch_maxima_output_check_id}
    set batch_maxima_output_check_id ""
  }
  if {$batch_maxima_state_check_id != ""} {
    catch {after cancel $batch_maxima_state_check_id}
    set batch_maxima_state_check_id ""
  }

  foreach one_pid [lreverse $batch_maxima_ids] {
    if {$one_pid==""} {
      continue
    }
    if {(!$overall) && ($batch_maxima_handle != "")} {
      set print_logging_maxima_output_last false
      set batch_maxima_stop_string "quit();\n"
      catch {puts $batch_maxima_handle $batch_maxima_stop_string}
      if {$print_logging_maxima_input} {
        print_logging [strcat "\nXXX stop" "\n"]
        print_logging [strcat $batch_maxima_stop_string "\n"]
      }
    }
    set wait_time1 10
    set wait_time2 10
    if {![is_wine_like]} {
      if {!$overall} {
        pause_tasks $wait_time1
      }
      catch {exec "kill" "-s" "SIGTERM" $one_pid}
      if {!$overall} {
        pause_tasks $wait_time2
      }
      catch {exec "kill" "-s" "SIGKILL" $one_pid}
    } {
      if {!$overall} {
        pause_tasks $wait_time1
      }
      catch {exec "taskkill" "/pid" $one_pid}
      if {!$overall} {
        pause_tasks $wait_time2
      }
      catch {exec "taskkill" "/f" "/pid" $one_pid}
    }
  }
  set batch_maxima_ids [list]

  if {$batch_maxima_handle != ""} {
    catch {close $batch_maxima_handle}
    set batch_maxima_handle ""
  }

  if {$batch_maxima_output_id_write != ""} {
    catch {close $batch_maxima_output_id_write}
    set batch_maxima_output_id_write ""
  }
  if {$batch_maxima_output_id_read != ""} {
    catch {close $batch_maxima_output_id_read}
    set batch_maxima_output_id_read ""
  }

  if {$batch_maxima_input_tmpfile_mac != ""} {
    catch {file delete -- $batch_maxima_input_tmpfile_mac}
    set batch_maxima_input_tmpfile_mac ""
  }
  if {$batch_maxima_output_tmpfile_tmp != ""} {
    catch {file delete -- $batch_maxima_output_tmpfile_tmp}
    set batch_maxima_output_tmpfile_tmp ""
  }

  if {[is_batch_maxima_asking]} {
    names_start_batch_maxima_process
  }
  set batch_maxima_asks_names false

  $mxmbatch_start configure -state "normal"
  $mxmbatch_stop configure -state "disabled"
  $mxmbatch_break configure -state "disabled"

  mxmbatch_input_field_set_active

  set maxima_batch_end_reached false

  $bt_kill configure -state "disabled"
  set_bt_kill_for_stop

  reset_cpu_usage
  set_ignore_cpu_usage true
  stop_cpu_usage_updates

  set should_display_cpu_usage false
  set could_display_cpu_usage false
  set_state_label_text "batch mode"
  set batch_maxima_running false
  state_batch_maxima
}

proc is_mxmbatch_input_field_active {} {
  global mxmbatch_input_field

  if {$mxmbatch_input_field==""} {
    return false
  }
  if {[$mxmbatch_input_field cget -state]!="normal"} {
    return false
  }
  return true
}

proc batch_maxima_set_last_command {} {
  global batch_maxima_last_command
  global mxmbatch_input_field

  if {$mxmbatch_input_field==""} {
    return false
  }
  if {![is_mxmbatch_input_field_active]} {
    return false
  }
  if {[get_command_content false] == $batch_maxima_last_command} {
    return true
  }

  set cur_state [$mxmbatch_input_field cget -state]
  $mxmbatch_input_field configure -state "normal"
  $mxmbatch_input_field edit separator
  $mxmbatch_input_field replace "1.0" "end" $batch_maxima_last_command [list]
  $mxmbatch_input_field edit separator
  $mxmbatch_input_field configure -state $cur_state

  return true
}

set batch_maxima_version_read_pid_prefix "PID:"
proc state_batch_maxima {} {
  global batch_maxima_running
  global maxima_batch_pid_label
  global batch_maxima_version_read_pid_prefix
  global maxima_batch_pid_present
  global batch_maxima_window
  global should_display_cpu_usage
  global could_display_cpu_usage
  global last_cpu_usage_info
  global last_cpu_usage_high

  if {$batch_maxima_window==""} {
    return
  }

  set info_display ""
  if {$should_display_cpu_usage || ($could_display_cpu_usage && $last_cpu_usage_high)} {
    set info_display $last_cpu_usage_info
  }
  if {$info_display!=""} {
    $maxima_batch_pid_label configure -text "$info_display $batch_maxima_version_read_pid_prefix"
    return
  }

  if {(!$batch_maxima_running) && ($maxima_batch_pid_present !="")} {
    $maxima_batch_pid_label configure -text "last $batch_maxima_version_read_pid_prefix"
  } elseif {$batch_maxima_running} {
    if {[is_batch_maxima_waiting]} {
      $maxima_batch_pid_label configure -text "prepared $batch_maxima_version_read_pid_prefix"
    } elseif {[is_batch_maxima_asking]} {
      $maxima_batch_pid_label configure -text "updating $batch_maxima_version_read_pid_prefix"
    } {
      $maxima_batch_pid_label configure -text "running $batch_maxima_version_read_pid_prefix"
    }
  } {
    $maxima_batch_pid_label configure -text "$batch_maxima_version_read_pid_prefix"
  }
}

proc close_batch_maxima {} {
  global batch_maxima_window
  global mxmbatch_input_field

  catch {focus $mxmbatch_input_field}
  catch {wm withdraw $batch_maxima_window}
}

proc title_batch_maxima {} {
  global batch_maxima_window
  global batch_maxima_input_warning
  global batch_maxima_output_warning

  if {$batch_maxima_window==""} {
    return
  }

  set title_string "TGM Minima: Maxima batches"
  if {$batch_maxima_input_warning || $batch_maxima_output_warning} {
    append title_string " *"
  }

  wm title $batch_maxima_window $title_string
}

proc draw_batch_maxima {} {
  global batch_maxima_window
  global mxmbatch_input_field

  if {$batch_maxima_window==""} {
    setup_batch_maxima
  }

  remove_torn_menus
  update_mxmbatch_warning_label

  catch {wm manage $batch_maxima_window}

  wm protocol $batch_maxima_window WM_DELETE_WINDOW {close_batch_maxima}
  title_batch_maxima
  wm deiconify $batch_maxima_window
  wm minsize $batch_maxima_window 640 300
  wm resizable $batch_maxima_window true true

  raise $batch_maxima_window
  focus $mxmbatch_input_field
}

proc sweep_batch_maxima {} {
  global mxmbatch_output_field
  global mxmbatch_sweep
  global maxima_batch_start_row

  set cur_state [$mxmbatch_output_field cget -state]
  $mxmbatch_output_field configure -state "normal"
  $mxmbatch_output_field delete "1.0" "end"
  $mxmbatch_output_field configure -state $cur_state

  set maxima_batch_start_row 0

  $mxmbatch_sweep configure -state "disabled"
}

proc set_max_count_value_batch_maxima_text {{to_change false} {value 10}} {
  global maxima_batch_max_row_count_for_text
  global mxmbatch_info_label_maxcount_text

  if {$to_change} {
    set maxima_batch_max_row_count_for_text $value
  }

  if {$mxmbatch_info_label_maxcount_text==""} {
    return
  }
  $mxmbatch_info_label_maxcount_text configure -text [strcat "(" $maxima_batch_max_row_count_for_text ")"]
}

set mxmbatch_label_maxcount_text_menu ""
proc offer_info_maxcount_stop_batch_maxima_text {x_abs y_abs x_pos y_pos} {
  global topw
  global maxima_batch_max_row_count_for_text
  global mxmbatch_info_label_maxcount_text
  global mxmbatch_label_maxcount_text_menu
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  if  {![is_batch_mode]} {
    return
  }
  if {$mxmbatch_info_label_maxcount_text==""} {
    return
  }

  set use_x [expr {$x_abs - 100}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  if {$mxmbatch_label_maxcount_text_menu!=""} {
    catch {destroy $mxmbatch_label_maxcount_text_menu}
  }
  set mxmbatch_label_maxcount_text_menu [menu "$topw.mxmbatch_label_maxcount_text_menu" -tearoff false]
  $mxmbatch_label_maxcount_text_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $mxmbatch_label_maxcount_text_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $mxmbatch_label_maxcount_text_menu configure -activeborderwidth 1

  $mxmbatch_label_maxcount_text_menu add command -label "Maximal (total) row count per batch put to text:" \
    -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  set counts_offered [list 10 20 40 80]
  set cnt_idx 0
  foreach one_cnt $counts_offered {
    incr cnt_idx
    $mxmbatch_label_maxcount_text_menu add command -label "$one_cnt" \
      -command [list "set_max_count_value_batch_maxima_text" true $one_cnt]
    if {$one_cnt==$maxima_batch_max_row_count_for_text} {
      $mxmbatch_label_maxcount_text_menu entryconfigure $cnt_idx -background $menu_selected_background -activebackground "#e0e0ff"
    }
  }

  tk_popup $mxmbatch_label_maxcount_text_menu $use_x $use_y
}
set_max_count_value_batch_maxima_text false

bind $mxmbatch_info_label_maxcount_text <1> {
  offer_info_maxcount_stop_batch_maxima_text %X %Y %x %y
}

set maxima_batch_max_row_count_exp 3
set maxima_batch_max_row_count 0
set mxmbatch_info_label_maxcount ""

proc set_max_count_label_batch_maxima {} {
  global mxmbatch_info_label_maxcount
  global maxima_batch_max_row_count_exp

  if {$mxmbatch_info_label_maxcount==""} {
    return
  }

  $mxmbatch_info_label_maxcount configure -text "(10^$maxima_batch_max_row_count_exp)"
}

proc set_max_count_value_batch_maxima {} {
  global maxima_batch_max_row_count_exp
  global maxima_batch_max_row_count

  set maxima_batch_max_row_count [expr {10**$maxima_batch_max_row_count_exp}]

  set_max_count_label_batch_maxima
}
set_max_count_value_batch_maxima

proc change_max_count_exp_value_batch_maxima {increase} {
  global maxima_batch_max_row_count_exp
  global maxima_batch_max_row_count

  incr maxima_batch_max_row_count_exp $increase
  if {$maxima_batch_max_row_count_exp>5} {
    set maxima_batch_max_row_count_exp 2
  }
  if {$maxima_batch_max_row_count_exp<2} {
    set maxima_batch_max_row_count_exp 5
  }

  set_max_count_value_batch_maxima
}

proc set_max_count_exp_value_batch_maxima {change {exp_value 2}} {
  global maxima_batch_max_row_count_exp
  global maxima_batch_max_row_count

  if {$change} {
    set maxima_batch_max_row_count_exp $exp_value
  }

  set_max_count_value_batch_maxima
}

set mxmbatch_label_maxcount_menu ""
proc offer_info_maxcount_stop_batch_maxima {x_abs y_abs x_pos y_pos} {
  global topw
  global maxima_batch_max_row_count_exp
  global maxima_batch_max_row_count
  global mxmbatch_info_label_maxcount
  global mxmbatch_label_maxcount_menu
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  if  {![is_batch_mode]} {
    return
  }
  if {$mxmbatch_info_label_maxcount==""} {
    return
  }

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  catch {destroy $mxmbatch_label_maxcount_menu}
  set mxmbatch_label_maxcount_menu [menu "$topw.mxmbatch_label_maxcount_menu" -tearoff false]
  $mxmbatch_label_maxcount_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $mxmbatch_label_maxcount_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $mxmbatch_label_maxcount_menu configure -activeborderwidth 1

  $mxmbatch_label_maxcount_menu add command -label "Maximal (total) row count per batch" \
    -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"
  $mxmbatch_label_maxcount_menu add command -label "100" \
    -command [list "set_max_count_exp_value_batch_maxima" true 2]
  $mxmbatch_label_maxcount_menu add command -label "1,000" \
    -command [list "set_max_count_exp_value_batch_maxima" true 3]
  $mxmbatch_label_maxcount_menu add command -label "10,000" \
    -command [list "set_max_count_exp_value_batch_maxima" true 4]
  $mxmbatch_label_maxcount_menu add command -label "100,000" \
    -command [list "set_max_count_exp_value_batch_maxima" true 5]

  set cur_idx $maxima_batch_max_row_count_exp
  incr cur_idx -1

  $mxmbatch_label_maxcount_menu entryconfigure $cur_idx -background $menu_selected_background -activebackground "#e0e0ff"

  tk_popup $mxmbatch_label_maxcount_menu $use_x $use_y
}

proc update_prepend_state_batch_maxima {} {
  global mxmbatch_input_prepend_use
  global mxmbatch_input_entry

  if {$mxmbatch_input_entry==""} {
    return
  }

  $mxmbatch_input_entry configure -state [expr {$mxmbatch_input_prepend_use ? "normal" : "disabled"}]
}

set maxima_batch_about_info_tag "maxima_batch_about_info_tag"
proc about_batch_maxima {} {
  global mxmbatch_output_field
  global maxima_batch_about_info_tag
  global mxmbatch_sweep

  if {$mxmbatch_output_field==""} {
    return
  }

  set about_info [strcat \
    "Maxima batches can be run both from this \"batch window\"\n" \
    "and from the main text (if set at the Mouse-3 menu there).\n" \
    "Internally done via calling the \"batch\" Maxima function. \n" \
    "\n" \
    "Control+| to open this window, Control+comma/period to slide within.\n" \
    "\n" \
  ]

  set curr_state [$mxmbatch_output_field cget -state]
  $mxmbatch_output_field configure -state "normal"
  if {[lindex [split [$mxmbatch_output_field index "end -1 chars"] "."] 1]!=0} {
    $mxmbatch_output_field insert "end" "\n" [list]
  }
  $mxmbatch_output_field insert "end" $about_info [list $maxima_batch_about_info_tag]
  $mxmbatch_output_field configure -state $curr_state
  $mxmbatch_output_field see "end"

  $mxmbatch_sweep configure -state "normal"
}

ttk::copyBindings TCheckbutton Batch.TCheckbutton
ttk::style configure Batch.TCheckbutton \
  -indicatorcolor "#c0c0c0" -indicatorrelief sunken -padding [list 2 0 0 0] \
  -indicatormargin [list 0 0 0 0] -indicatordiameter 10 -borderwidth 1
ttk::style map Batch.TCheckbutton -indicatorcolor \
  [list pressed "#f0f0ff" selected "#8080c0"]

ttk::copyBindings TCheckbutton BatchLine.TCheckbutton
ttk::style configure BatchLine.TCheckbutton \
  -indicatorcolor "#f0f0f0" -indicatorrelief sunken -padding [list 2 0 0 0] \
  -indicatormargin [list 0 0 0 0] -indicatordiameter 10 -borderwidth 1
ttk::style map BatchLine.TCheckbutton -indicatorcolor \
  [list pressed "#c0c0c0" selected "#8080c0"]

proc is_batch_maxima_set {} {
  global batch_maxima_window
  return [expr {($batch_maxima_window=="") ? false : true}]
}

set mxmbatch_start ""
set mxmbatch_stop ""
set mxmbatch_break ""
set mxmbatch_sweep ""
set mxmbatch_output_frame ""
set mxmbatch_input_frame ""
set mxmbatch_pane_holder ""
set maxima_batch_pid_label ""
set mxmbatch_input_entry ""
set mxmbatch_input_entry_holder ""
set maxima_batch_warning_label ""
set mxmbatch_info_label_stopping ""
proc setup_batch_maxima {} {
  global batch_maxima_window
  global scr_width
  global scr_height
  global mxmbatch_input_field
  global mxmbatch_output_field
  global spec_base_font
  global mxmbatch_start
  global mxmbatch_stop
  global mxmbatch_break
  global mxmbatch_sweep
  global maxima_batch_time_tag
  global maxima_batch_output_tag
  global maxima_batch_pid_present
  global mxmbatch_output_frame
  global mxmbatch_input_frame
  global mxmbatch_pane_holder
  global maxima_batch_pid_label
  global batch_maxima_version_read_pid_prefix
  global tag_command_strings
  global tag_command_comments
  global tag_command_bracket_pair
  global tag_command_bracket
  global select_fg_col
  global select_bg_col
  global with_labels
  global info_labels_foreground
  global bind_labels
  global show_time_elapsed
  global mxmbatch_fmenu
  global mxmbatch_info_label_display
  global mxmbatch_info_label_verb
  global mxmbatch_info_label_bind
  global mxmbatch_info_label_time
  global mxmbatch_info_label_adjust
  global mxmbatch_info_label_maxcount
  global mxmbatch_input_prepend
  global mxmbatch_input_prepend_use
  global mxmbatch_input_entry
  global mxmbatch_input_entry_holder
  global maxima_batch_warning_label
  global use_buttons_text
  global maxima_batch_about_info_tag
  global mxmbatch_info_label_endings
  global info_label_endings_char
  global make_line_endings
  global mxmbatch_info_label_stopping
  global known_selectforeground_selectbackground_for_tags

  if {$batch_maxima_window!=""} {
    return
  }

  set batch_maxima_window [toplevel ".maxima_batch" -width [expr {($scr_width>1000) ? 900 : 720}] -height 400]

  set top_pos [get_top_window_position]
  set x_pos [expr {40 + [lindex $top_pos 0]}]
  set y_pos [expr {(($scr_width>1000) ? 50 : 15) + [lindex $top_pos 1]}]
  wm geometry $batch_maxima_window "+$x_pos+$y_pos"

  set flabels [frame $batch_maxima_window.flabels -relief sunken -bd 1]
  pack $flabels -side top -expand no -fill x
  set ftext_holder [panedwindow $batch_maxima_window.ftext_holder -orient vertical -handlesize 1 -borderwidth 0 -sashpad 0 \
    -width [expr {($scr_width>1000) ? 880 : 700}]]
  pack $ftext_holder -side top -expand yes -fill both
  set mxmbatch_pane_holder $ftext_holder
  set mxmbatch_fmenu $flabels

  set mxmbatch_start [ttk::button $flabels.lmstart -text "Start" -command {default_start_batch_maxima} -takefocus 0 -width 6 -style File.TButton]
  set mxmbatch_break [ttk::button $flabels.lmbreak -text "Break" -command {break_batch_maxima} -takefocus 0 -width 6 -style File.TButton]
  set mxmbatch_stop [ttk::button $flabels.lmstop -text "Stop" -command {stop_batch_maxima} -takefocus 0 -width 6 -style File.TButton]
  set mxmbatch_sweep [ttk::button $flabels.lmsweep -text "Sweep" -command {sweep_batch_maxima} -takefocus 0 -width 6 -style File.TButton]
  pack $mxmbatch_start -side left
  if {[has_common_ctrlc]} {
    pack $mxmbatch_break -side left
  }
  pack $mxmbatch_stop $mxmbatch_sweep -side left

  $mxmbatch_start configure -state "normal"
  $mxmbatch_break configure -state "disabled"
  $mxmbatch_stop configure -state "disabled"
  $mxmbatch_sweep configure -state "disabled"

  set label_spacer_i [label $flabels.label_spacer_i -text " " -bd 0 -width 1]

  set mxmbatch_info_label_display [label $flabels.maximadisplay -text " " -bd 0 -width 3]

  set mxmbatch_info_label_verb [label $flabels.withlabels -text "%" -bd 0 -width 1]
  if {!$with_labels} {
    $mxmbatch_info_label_verb configure -foreground [lindex $info_labels_foreground 0]
  }
  if {[is_wine_like]} {
    $mxmbatch_info_label_verb configure -width 2 -justify "right" -anchor "e"
  }

  set mxmbatch_info_label_bind [label $flabels.labelsbound -text "b" -bd 0 -width 1 -justify "center"]
  if {!$bind_labels} {
    $mxmbatch_info_label_bind configure -foreground [lindex $info_labels_foreground 0]
  }
  if {[is_wine_like]} {
    $mxmbatch_info_label_bind configure -justify "left" -anchor "w"
  }

  set mxmbatch_info_label_adjust [label $flabels.displayadjust -text [get_display_adjust_char] -bd 0 -width 2]
  $mxmbatch_info_label_adjust configure -justify "center"

  set mxmbatch_info_label_time [label $flabels.timeelapsed -text "t" -bd 0 -width 2 -justify "center"]
  if {!$show_time_elapsed} {
    $mxmbatch_info_label_time configure -foreground [lindex $info_labels_foreground 0]
  }

  set mxmbatch_info_label_endings [label $flabels.lineendings -text $info_label_endings_char -bd 0 -width 2 -justify "left" -anchor "w"]
  if {!$make_line_endings} {
    $mxmbatch_info_label_endings configure -foreground [lindex $info_labels_foreground 0]
  }

  set mxmbatch_info_label_stopping [label $flabels.batchstopping -text [get_batch_ending_label_char] -bd 0 -width 2 -justify "left" -anchor "w"]

  pack $label_spacer_i $mxmbatch_info_label_display $mxmbatch_info_label_verb $mxmbatch_info_label_bind \
    $mxmbatch_info_label_adjust $mxmbatch_info_label_time $mxmbatch_info_label_endings $mxmbatch_info_label_stopping -side left

  bind $mxmbatch_info_label_display <1> {
    offer_info_display_request %X %Y %x %y
    break
  }
  mxmbatch_update_info_display_request false

  bind $mxmbatch_info_label_verb <1> {
    remove_torn_menus
    offer_info_labels_shown_bound %X %Y %x %y
    break
  }

  bind $mxmbatch_info_label_bind <1> {
    remove_torn_menus
    offer_info_labels_shown_bound %X %Y %x %y
    break
  }

  bind $mxmbatch_info_label_adjust <1> {
    remove_torn_menus
    mxmbatch_offer_info_display_adjust %X %Y %x %y true
    break
  }
  bind $mxmbatch_info_label_adjust <Shift-1> {
    remove_torn_menus
    change_info_display_adjust_request true false
    break
  }
  bind $mxmbatch_info_label_adjust <Control-1> {
    remove_torn_menus
    change_info_display_adjust_request false true
    break
  }
  bind $mxmbatch_info_label_adjust <Control-Shift-1> {
    remove_torn_menus
    change_info_display_adjust_request true true
    break
  }

  bind $mxmbatch_info_label_time <1> {
    remove_torn_menus
    mxmbatch_offer_info_show_time %X %Y %x %y
    break
  }
  bind $mxmbatch_info_label_time <Control-1> {
    remove_torn_menus
    change_info_show_time_request
    break
  }

  bind $mxmbatch_info_label_endings <1> {
    remove_torn_menus
    offer_info_make_lineendings %X %Y %x %y true
  }
  bind $mxmbatch_info_label_endings <Control-1> {
    remove_torn_menus
    change_info_make_lineendings_request
    break
  }

  bind $mxmbatch_info_label_stopping <1> {
    remove_torn_menus
    mxmbatch_offer_info_batch_stopping %X %Y %x %y true
  }
  bind $mxmbatch_info_label_stopping <Control-1> {
    remove_torn_menus
    change_info_batch_stopping_request
    break
  }

  set mxmbatch_info_label_maxcount [label $flabels.rowmaxcount -text " " -bd 0 -width 6 -justify "left" -fg "#45456a"]
  set_max_count_label_batch_maxima
  pack $mxmbatch_info_label_maxcount -side left

  bind $mxmbatch_info_label_maxcount <1> {
    remove_torn_menus
    offer_info_maxcount_stop_batch_maxima %X %Y %x %y
  }

  set mxmbatch_about [ttk::button $flabels.lgabout -text "?" -command {about_batch_maxima} -takefocus 0 -width 2 -style File.TButton]
  pack $mxmbatch_about -side right

  set circ_ch ""
  if {!$use_buttons_text} {
    catch {set circ_ch [encoding convertfrom "utf-8" "\xE2\x97\x8F"]}
  }
  set warning_char [expr {($circ_ch=="") ? "m" : $circ_ch}]
  set maxima_batch_warning_label [label $flabels.lmwarning -text "$warning_char" -relief "flat" -borderwidth 0 \
    -foreground "#c0c0c0" -background "#a0a0a0" -width 3]
  pack $maxima_batch_warning_label -side right -fill y
  update_mxmbatch_warning_label

  bind $maxima_batch_warning_label <1> {
    remove_torn_menus
    offer_mxmbatch_warning_menu %X %Y %x %y
    break
  }
  bind $maxima_batch_warning_label <Control-1> {
    remove_torn_menus
    switch_batch_maxima_puts_warning
    break
  }

  set pid_entry [entry $flabels.lmpid_entry -textvariable maxima_batch_pid_present -width 8 -justify "center" -state "readonly" \
    -relief flat -borderwidth 0 -takefocus 1 -exportselection true]
  base_bindings_for_entry $pid_entry "maxima_batch_pid_present"
  pack $pid_entry -side "right"

  set maxima_batch_pid_label [ttk::label $flabels.lmpid_label -text "$batch_maxima_version_read_pid_prefix" \
    -width 13 -padding [list 10 0 0 0] -justify "right" -anchor "e"]
  pack $maxima_batch_pid_label -side "right"

  set mxmbatch_output_frame [ttk::frame $ftext_holder.output_frame -relief groove -borderwidth 2]
  set mxmbatch_input_frame [ttk::frame $ftext_holder.input_frame -relief groove -borderwidth 2]
  $ftext_holder add $mxmbatch_output_frame -minsize 100 -stretch always
  $ftext_holder add $mxmbatch_input_frame -minsize 120 -stretch never

  set mxmbatch_output_field [text $mxmbatch_output_frame.mxmbatch_output_field -wrap char -undo false \
    -height 16]
  ttk::scrollbar $mxmbatch_output_frame.vscr -orient vertical -command "$mxmbatch_output_field yview"
  $mxmbatch_output_field configure -yscroll "$mxmbatch_output_frame.vscr set"

  grid $mxmbatch_output_field $mxmbatch_output_frame.vscr -sticky nsew

  grid rowconfigure $mxmbatch_output_frame $mxmbatch_output_field -weight 1
  grid columnconfigure $mxmbatch_output_frame $mxmbatch_output_field -weight 1

  if {$spec_base_font(command)==""} {
    $mxmbatch_output_field configure -font command_font
  } {
    $mxmbatch_output_field configure -font $spec_base_font(command)
  }

  $mxmbatch_output_field configure -foreground "#000000"
  $mxmbatch_output_field configure -background "#ffffff"
  $mxmbatch_output_field configure -selectforeground $select_fg_col
  $mxmbatch_output_field configure -selectbackground $select_bg_col
  $mxmbatch_output_field configure -inactiveselectbackground $select_bg_col
  $mxmbatch_output_field configure -state "disabled"

  $mxmbatch_output_field tag configure $maxima_batch_about_info_tag -background "#ffffe7" -foreground "#101010"
  $mxmbatch_output_field tag configure $maxima_batch_time_tag -foreground "#909090"
  $mxmbatch_output_field tag configure $maxima_batch_output_tag -lmargin1 0 -lmargin2 5
  catch {
    $mxmbatch_output_field tag configure $maxima_batch_output_tag -lmargincolor "#000000"
  }
  $mxmbatch_output_field tag raise "sel"

  set prepend_holder [frame $mxmbatch_input_frame.prepend_holder]
  set mxmbatch_input_entry_holder $prepend_holder
  set mxmbatch_input_entry [entry $prepend_holder.mxmbatch_input_entry -textvariable mxmbatch_input_prepend -takefocus 1]
  set mxmbatch_input_prepend_cb [ttk::checkbutton $prepend_holder.mxmbatch_input_prepend_cb \
    -variable mxmbatch_input_prepend_use -onvalue true -offvalue false -takefocus 0 -command {update_prepend_state_batch_maxima} \
    -style Batch.TCheckbutton -width 0]
  update_prepend_state_batch_maxima
  base_bindings_for_entry $mxmbatch_input_entry "mxmbatch_input_prepend"
  pack $mxmbatch_input_entry -side left -expand yes -fill both
  pack $mxmbatch_input_prepend_cb -side right -expand no

  bind $mxmbatch_input_entry <Tab> {
    entry_action_on_tab $mxmbatch_input_entry "mxmbatch_input_prepend" false
    break
  }
  foreach keyspressed [list "Shift-Key-Tab" "Shift-Key-KP_Tab" "Shift-Key-ISO_Left_Tab" "Key-ISO_Left_Tab"] {
    catch {
      bind $mxmbatch_input_entry <$keyspressed> {
        entry_action_on_tab $mxmbatch_input_entry "mxmbatch_input_prepend" true
        break
      }
    }
  }

  set mxmbatch_input_field [text $mxmbatch_input_frame.mxmbatch_input_field -wrap none -undo true -maxundo 10000 \
    -height 8]
  ttk::scrollbar $mxmbatch_input_frame.vscr -orient vertical -command "$mxmbatch_input_field yview"
  ttk::scrollbar $mxmbatch_input_frame.hscr -orient horizontal -command "$mxmbatch_input_field xview"
  $mxmbatch_input_field configure -xscroll "$mxmbatch_input_frame.hscr set" -yscroll "$mxmbatch_input_frame.vscr set"
  ttk::sizegrip $mxmbatch_input_frame.size_grip

  grid $prepend_holder -sticky nsew -columnspan 2
  grid $mxmbatch_input_field $mxmbatch_input_frame.vscr -sticky nsew
  grid $mxmbatch_input_frame.hscr $mxmbatch_input_frame.size_grip -sticky nsew

  grid rowconfigure $mxmbatch_input_frame $mxmbatch_input_field -weight 1
  grid columnconfigure $mxmbatch_input_frame $mxmbatch_input_field -weight 1

  if {$spec_base_font(command)==""} {
    $mxmbatch_input_field configure -font command_font
    $mxmbatch_input_entry configure -font command_font
  } {
    $mxmbatch_input_field configure -font $spec_base_font(command)
    $mxmbatch_input_entry configure -font $spec_base_font(command)
  }

  $mxmbatch_input_field configure -foreground "#000000"
  $mxmbatch_input_field configure -background "#ffffff"
  $mxmbatch_input_field configure -selectforeground $select_fg_col
  $mxmbatch_input_field configure -selectbackground $select_bg_col
  $mxmbatch_input_field configure -inactiveselectbackground $select_bg_col

  $mxmbatch_input_field tag configure $tag_command_strings -foreground "#805020"
  $mxmbatch_input_field tag configure $tag_command_comments -foreground "#605060"
  $mxmbatch_input_field tag configure $tag_command_bracket_pair -foreground "#00a0f0"
  $mxmbatch_input_field tag configure $tag_command_bracket -foreground "#f03000"

  $mxmbatch_input_field tag raise "sel"
  $mxmbatch_input_field tag raise $tag_command_strings
  $mxmbatch_input_field tag raise $tag_command_comments
  $mxmbatch_input_field tag raise $tag_command_bracket_pair
  $mxmbatch_input_field tag raise $tag_command_bracket

  if {!$known_selectforeground_selectbackground_for_tags} {
    $mxmbatch_input_field tag raise "sel"
  }

  bind $mxmbatch_input_field <FocusIn> {
    $mxmbatch_input_field tag configure "sel" -bgstipple ""
  }
  bind $mxmbatch_input_field <FocusOut> {
    $mxmbatch_input_field tag configure "sel" -bgstipple gray75
  }

  bind $mxmbatch_output_field <FocusIn> {
    $mxmbatch_output_field tag configure "sel" -bgstipple ""
    remove_torn_menus
  }
  bind $mxmbatch_output_field <FocusOut> {
    $mxmbatch_output_field tag configure "sel" -bgstipple gray75
  }

  bind $mxmbatch_output_field <1> {
    remove_torn_menus
    focus $mxmbatch_output_field
  }
  bind $mxmbatch_output_field <3> {
    remove_torn_menus
    focus $mxmbatch_output_field
    catch {
      if {[is_batch_maxima_running] && (![is_batch_maxima_asking])} {
        offer_mxmbatch_active_actions %X %Y %x %y
      }
    }
    break
  }

  # first generally removing bindings
  set other_keys [list "Control-s" "Control-t" "Alt-t" "Control-d" "Alt-d" "Control-e" "Control-E" \
    "Control-k" "Control-o" "Control-B" "Alt-b" "Control-f" "Alt-f" "Control-g"  "Control-h" \
    "Control-p" "Control-n" "Control-l" "Control-j" "Control-comma" "Control-period" "Control-u" \
    "Control-slash" "Control-backslash" "Control-r" "Control-R" "Control-q" "Control-w"]
  foreach keyspressed $other_keys {
    bind $mxmbatch_input_field <$keyspressed> {
      break
    }
  }
  foreach keyspressed $other_keys {
    bind $mxmbatch_output_field <$keyspressed> {
      break
    }
  }

  foreach wdg [list $batch_maxima_window $mxmbatch_input_field $mxmbatch_output_field $mxmbatch_input_entry] {
    bind $wdg <Escape> {
      close_batch_maxima
      break
    }

    bind $wdg <Alt-Return> {
      test_start_batch_maxima
      break
    }
    bind $wdg <Control-Return> {
      if {(![is_batch_maxima_running]) || [is_batch_maxima_waiting]} {
        default_start_batch_maxima true true
      }
      break
    }
    bind $wdg <Shift-Return> {
      if {(![is_batch_maxima_running]) || [is_batch_maxima_waiting]} {
        default_start_batch_maxima false false
      }
      break
    }
    bind $wdg <Control-Shift-Return> {
      if {(![is_batch_maxima_running]) || [is_batch_maxima_waiting]} {
        default_start_batch_maxima true false
      }
      break
    }
    bind $wdg <Control-Alt-Return> {
      break
    }
    bind $wdg <Shift-Alt-Return> {
      break
    }
    bind $wdg <Control-Shift-Alt-Return> {
      break
    }

    bind $wdg <Control-i> {
      if {[has_common_ctrlc]} {
        break_batch_maxima
      } {
        stop_batch_maxima
      }
      break
    }
    bind $wdg <Control-I> {
      stop_batch_maxima
      break
    }
    bind $wdg <Control-k> {
      stop_batch_maxima
      break
    }

    bind $wdg <Control-e> {
      sweep_batch_maxima
      break
    }
    bind $wdg <Control-Alt-e> {
      sweep_batch_maxima
      break
    }
    bind $wdg <Control-E> {
      sweep_batch_maxima
      break
    }

    bind $wdg <Control-u> {
      names_start_batch_maxima true false
      break
    }
    bind $wdg <Control-Alt-u> {
      names_start_batch_maxima true true
      break
    }
    bind $wdg <Control-U> {
      names_start_batch_maxima false false
      break
    }
    bind $wdg <Control-Alt-U> {
      names_start_batch_maxima false true
      break
    }

    bind $wdg <Control-l> {
      batch_maxima_set_last_command
      break
    }

    bind $wdg <Alt-plus> {
      change_info_batch_stopping_request
      break
    }
    bind $wdg <Alt-semicolon> {
      change_info_make_lineendings_request
      break
    }
    bind $wdg <Alt-t> {
      change_info_show_time_request
      break
    }
    bind $wdg <Alt-r> {
      change_info_display_adjust_request false true
      break
    }
    bind $wdg <Alt-asciicircum> {
      change_info_display_adjust_request true false
      break
    }
    bind $wdg <Alt-b> {
      mxmbatch_update_info_labels_bound_request true
      break
    }
    bind $wdg <Alt-percent> {
      mxmbatch_update_info_labels_request true
      break
    }
    bind $wdg <Alt-k> {
      change_max_count_exp_value_batch_maxima 1
      break
    }
    bind $wdg <Alt-K> {
      change_max_count_exp_value_batch_maxima -1
      break
    }

    bind $wdg <Control-quoteright> {
      draw_history
      break
    }

    bind $wdg <Control-bar> {
      if {[is_batch_maxima_running] && (![is_batch_maxima_asking])} {
        present_insert_to_batch_maxima
      }
      break
    }
  }

  bind $mxmbatch_input_field <Control-a> {
    $mxmbatch_input_field tag remove "sel" "1.0" "end"
    $mxmbatch_input_field tag add "sel" "1.0" "end -1 chars"
    $mxmbatch_input_field edit separator
    set_anchor_opposite $mxmbatch_input_field
    break
  }
  bind $mxmbatch_output_field <Control-a> {
    $mxmbatch_output_field tag remove "sel" "1.0" "end"
    $mxmbatch_output_field tag add "sel" "1.0" "end -1 chars"
    break
  }

  bind $mxmbatch_input_field <Control-x> {
    $mxmbatch_input_field edit separator
    set_clipboard $mxmbatch_input_field
    store_clipboard_onset
    #tk_textCut $mxmbatch_input_field
    delete_selected_text $mxmbatch_input_field
    $mxmbatch_input_field edit separator
    update_batch_maxima_properties
    break
  }

  bind $mxmbatch_input_field <Control-c> {
    $mxmbatch_input_field edit separator
    set_clipboard $mxmbatch_input_field
    store_clipboard_onset
    $mxmbatch_input_field edit separator
    update_batch_maxima_properties
    break
  }
  bind $mxmbatch_output_field <Control-c> {
    set_clipboard $mxmbatch_output_field
    store_clipboard_onset
    break
  }

  bind $mxmbatch_input_field <Control-V> {
    $mxmbatch_input_field edit separator
    tk_textPaste $mxmbatch_input_field
    $mxmbatch_input_field edit separator
    $mxmbatch_input_field see "insert"
    update_batch_maxima_properties
    break
  }
  bind $mxmbatch_input_field <Control-v> {
    $mxmbatch_input_field edit separator
    remove_selected_if_by_cursor $mxmbatch_input_field
    $mxmbatch_input_field edit separator
    tk_textPaste $mxmbatch_input_field
    $mxmbatch_input_field edit separator
    $mxmbatch_input_field see "insert"
    update_batch_maxima_properties
    break
  }

  bind $mxmbatch_input_field <Control-z> {
    catch {
      $mxmbatch_input_field edit undo
      update_batch_maxima_properties
    }
    break
  }
  bind $mxmbatch_input_field <Control-y> {
    catch {
      $mxmbatch_input_field edit redo
      update_batch_maxima_properties
    }
    break
  }
  bind $mxmbatch_input_field <Control-Z> {
    catch {
      $mxmbatch_input_field edit redo
      update_batch_maxima_properties
    }
    break
  }

  # eventually should provide menu for the batch-to-text copying actions
  bind $mxmbatch_input_field <Control-w> {
    if {[is_text_field_active]} {
      command_to_text_do_copy false true
    }
    break
  }
  bind $mxmbatch_input_field <Control-Alt-w> {
    if {[is_text_field_active]} {
      #command_to_text_do_paste true
      command_to_text_do_insert true
    }
    break
  }
  bind $mxmbatch_input_field <Control-W> {
    if {[is_text_field_active]} {
      command_to_text_do_copy true true
    }
    break
  }
  bind $mxmbatch_input_field <Control-Alt-W> {
    if {[is_text_field_active]} {
      command_to_text_do_append true
    }
    break
  }

  bind $mxmbatch_input_field <Prior> {
    topwise_page_up $mxmbatch_input_field
    set_anchor_insert $mxmbatch_input_field
    update_batch_maxima_highlights_onset
  }
  bind $mxmbatch_input_field <Shift-Prior> {
  }
  bind $mxmbatch_output_field <Prior> {
    topwise_page_up $mxmbatch_output_field
  }
  bind $mxmbatch_output_field <Shift-Prior> {
  }

  bind $mxmbatch_input_field <Next> {
    bottomwise_page_down $mxmbatch_input_field
    set_anchor_insert $mxmbatch_input_field
    update_batch_maxima_highlights_onset
  }
  bind $mxmbatch_input_field <Shift-Next> {
  }
  bind $mxmbatch_output_field <Next> {
    bottomwise_page_down $mxmbatch_output_field
  }
  bind $mxmbatch_output_field <Shift-Next> {
  }

  # jumping on brackets
  bind $mxmbatch_input_field <Control-b> {
    batch_maxima_jump_bracket
    break
  }
  bind $mxmbatch_input_field <Control-B> {
    batch_maxima_jump_bracket true
    break
  }

  # setting for quit-warning
  bind $mxmbatch_input_field <Control-m> {
    set_batch_maxima_input_warning
    break
  }
  bind $mxmbatch_input_entry <Control-m> {
    set_batch_maxima_input_warning
    break
  }
  bind $mxmbatch_output_field <Control-m> {
    set_batch_maxima_output_warning
    break
  }
  bind $pid_entry <Control-m> {
    set_batch_maxima_output_warning
    break
  }

  # unsetting from quit-warning
  bind $mxmbatch_input_field <Control-M> {
    unset_batch_maxima_input_warning
    break
  }
  bind $mxmbatch_input_entry <Control-M> {
    unset_batch_maxima_input_warning
    break
  }
  bind $mxmbatch_output_field <Control-M> {
    unset_batch_maxima_output_warning
    break
  }
  bind $pid_entry <Control-M> {
    unset_batch_maxima_output_warning
    break
  }

  bind $mxmbatch_input_entry <3> {
    remove_torn_menus
    catch {
      offer_prepend_mode_batch_maxima %X %Y %x %y
    }
    break
  }
  bind $mxmbatch_input_entry <ButtonRelease-3> {
    break
  }

  bind $mxmbatch_input_field <3> {
    remove_torn_menus
    catch {
      focus $mxmbatch_input_field
      offer_command_select_modes %X %Y %x %y
    }
    break
  }
  bind $mxmbatch_input_field <Control-3> {
    remove_torn_menus
    if {(![is_batch_maxima_running]) || ([is_batch_maxima_waiting])} {
      catch {
        focus $mxmbatch_input_field
        offer_command_maxima_run_action %X %Y %x %y
      }
    }
    break
  }
  bind $mxmbatch_input_field <Shift-3> {
    break
  }
  bind $mxmbatch_input_field <Alt-3> {
    break
  }

  bind $mxmbatch_input_field <2> {
    insert_primary $mxmbatch_input_field %X %Y %x %y
    update_batch_maxima_properties_delayed
    remove_torn_menus
    break
  }
  bind $mxmbatch_input_field <ButtonRelease-2> {
    update_batch_maxima_properties_onset
    break
  }
  bind $mxmbatch_input_field <1> {
    place_cursor_on_mouse $mxmbatch_input_field %X %Y %x %y
    $mxmbatch_input_field tag remove "sel" "1.0" "end"
    set_anchor_insert $mxmbatch_input_field
    remove_torn_menus
    update_batch_maxima_highlights_onset
    break
  }
  bind $mxmbatch_input_field <ButtonRelease-1> {
    update_batch_maxima_properties_onset
    break
  }
  bind $mxmbatch_input_field <Control-1> {
    break
  }
  bind $mxmbatch_input_field <Control-ButtonRelease-1> {
    break
  }

  bind $mxmbatch_input_entry <1> {
    focus $mxmbatch_input_entry
    remove_torn_menus
  }

  bind $mxmbatch_input_field <Double-1> {
    field_on_double_click $mxmbatch_input_field
    break
  }
  bind $mxmbatch_input_field <Control-Double-1> {
    set_anchor_insert $mxmbatch_input_field
    break
  }
  bind $mxmbatch_input_field <Control-Shift-Double-1> {
    set_anchor_insert $mxmbatch_input_field
    break
  }
  foreach keyspressed [list "Triple-1" "Quadruple-1"] {
    bind $mxmbatch_input_field <$keyspressed> {
      field_on_multi_click $mxmbatch_input_field
      break
    }
  }
  foreach keyspressed [list "Control-Triple-1" "Control-Quadruple-1"] {
    bind $mxmbatch_input_field <$keyspressed> {
      break
    }
  }

  bind $mxmbatch_input_field <Double-1> {
    field_on_double_click $mxmbatch_input_field
    break
  }
  bind $mxmbatch_input_field <Control-Double-1> {
    break
  }
  bind $mxmbatch_input_field <Control-Shift-Double-1> {
    break
  }
  foreach keyspressed [list "Triple-1" "Quadruple-1"] {
    bind $mxmbatch_input_field <$keyspressed> {
      field_on_multi_click $mxmbatch_input_field
      break
    }
  }
  foreach keyspressed [list "Control-Triple-1" "Control-Quadruple-1"] {
    bind $mxmbatch_input_field <$keyspressed> {
      break
    }
  }

  bind $batch_maxima_window <Control-comma> {
    focus $mxmbatch_input_field
    break
  }
  bind $mxmbatch_input_field <Control-comma> {
    focus $mxmbatch_input_entry
    break
  }
  bind $mxmbatch_input_entry <Control-comma> {
    focus $mxmbatch_output_field
    break
  }
  bind $mxmbatch_output_field <Control-comma> {
    focus $mxmbatch_output_field
    break
  }

  bind $batch_maxima_window <Control-period> {
    focus $mxmbatch_output_field
    break
  }
  bind $mxmbatch_input_field <Control-period> {
    focus $mxmbatch_input_field
    break
  }
  bind $mxmbatch_input_entry <Control-period> {
    focus $mxmbatch_input_field
    break
  }
  bind $mxmbatch_output_field <Control-period> {
    focus $mxmbatch_input_entry
    break
  }

  bind $mxmbatch_input_field <Control-r> {
    open_plot_part
    break
  }
  bind $mxmbatch_input_field <Control-Alt-r> {
    open_plot_part $mxmbatch_input_field false
    break
  }
  bind $mxmbatch_input_field <Control-R> {
    open_plot_part $mxmbatch_input_field true
    break
  }
  bind $mxmbatch_output_field <Control-r> {
    open_plot_part
    break
  }
  bind $mxmbatch_output_field <Control-Alt-r> {
    open_plot_part $mxmbatch_output_field false
    break
  }
  bind $mxmbatch_output_field <Control-R> {
    open_plot_part $mxmbatch_output_field true
    break
  }

  bind $mxmbatch_input_entry <Control-r> {
    open_plot_part
    break
  }

  bind $mxmbatch_input_field <Control-d> {
    draw_import
    break
  }
  bind $mxmbatch_output_field <Control-d> {
    draw_import
    break
  }
  bind $mxmbatch_input_entry <Control-d> {
    draw_import
    break
  }

  bind $mxmbatch_input_field <Tab> {
    string_action_on_tab $mxmbatch_input_field false false
    update_batch_maxima_properties
    break
  }

  catch {
    bind $mxmbatch_input_field <Control-Tab> {
      show_menu_key_names $mxmbatch_input_field "get_maxima_batch_input_position" false
      break
    }
  }
  foreach keyspressed [list "Shift-Key-Tab" "Shift-Key-KP_Tab" "Shift-Key-ISO_Left_Tab" "Key-ISO_Left_Tab"] {
    catch {
      bind $mxmbatch_input_field <$keyspressed> {
        string_action_on_tab $mxmbatch_input_field false true
        update_batch_maxima_properties
        break
      }
    }
    catch {
      bind $mxmbatch_input_field <Control-$keyspressed> {
        show_menu_key_names $mxmbatch_input_field "get_maxima_batch_input_position" true
        break
      }
    }
  }

  bind $mxmbatch_input_field <<Modified>> {
    update_batch_maxima_properties
    break
  }

  # binding highlight updates to general actions
  # can not use <KeyRelease> here, since numerical state of keypad gives non-numerical
  # (that is KP_Left, etc.) events for the KeyRelease event types.
  bind $mxmbatch_input_field <KeyPress> "+{update_batch_maxima_properties_delayed}"
  bind $mxmbatch_input_field <ButtonRelease> "+{update_batch_maxima_properties_delayed}"

  bind_keypad $batch_maxima_window
  bind_keypad $mxmbatch_input_field
  bind_keypad $mxmbatch_output_field

  bind $mxmbatch_input_field <KeyPress-Home> {
    home_key_reasonable $mxmbatch_input_field
    update_batch_maxima_highlights_onset
    break
  }
  bind $mxmbatch_input_field <Control-KeyPress-Home> {
    update_batch_maxima_highlights_onset
  }
  bind $mxmbatch_input_field <Shift-KeyPress-Home> {
    update_batch_maxima_highlights_onset
  }
  bind $mxmbatch_input_field <Alt-KeyPress-Home> {
    update_batch_maxima_highlights_onset
  }

  bind $mxmbatch_output_field <KeyPress-Home> {
    home_key_reasonable $mxmbatch_output_field
    break
  }
  bind $mxmbatch_output_field <Control-KeyPress-Home> {
  }
  bind $mxmbatch_output_field <Shift-KeyPress-Home> {
  }
  bind $mxmbatch_output_field <Alt-KeyPress-Home> {
  }
}

set update_batch_maxima_highlights_id ""
proc update_batch_maxima_highlights_onset {} {
  global update_batch_maxima_highlights_id

  if {$update_batch_maxima_highlights_id!=""} {
    catch {after cancel $update_batch_maxima_highlights_id}
    set update_batch_maxima_highlights_id ""
  }

  set update_batch_maxima_highlights_id [after 0 "update_cmmd_bracket_pairing"]
}

proc update_batch_maxima_properties_onset {} {
  global update_batch_maxima_properties_id

  if {$update_batch_maxima_properties_id!=""} {
    catch {after cancel $update_batch_maxima_properties_id}
    set update_batch_maxima_properties_id ""
  }

  set update_batch_maxima_properties_id [after 0 "update_cmmd_strings_comments"]
}

proc update_batch_maxima_properties {} {
  global update_batch_maxima_properties_id

  remove_torn_menus

  if {$update_batch_maxima_properties_id!=""} {
    catch {after cancel $update_batch_maxima_properties_id}
    set update_batch_maxima_properties_id ""
  }

  update_cmmd_strings_comments
}

set update_batch_maxima_properties_id ""
proc update_batch_maxima_properties_delayed {} {
  global update_batch_maxima_properties_id
  global mxmbatch_input_field
  global common_spec_chars

  remove_torn_menus

  if {$update_batch_maxima_properties_id!=""} {
    catch {after cancel $update_batch_maxima_properties_id}
    set update_batch_maxima_properties_id ""
  }

  set prev_char [$mxmbatch_input_field get "insert -1 chars" "insert"]
  set cur_char [$mxmbatch_input_field get "insert" "insert +1 chars"]
  set after_time 100
  if {$cur_char in $common_spec_chars} {
    set after_time 0
  } elseif {$prev_char in $common_spec_chars} {
    set after_time 0
  }

  set update_batch_maxima_properties_id [after $after_time "update_cmmd_strings_comments"]
}

proc update_mxmbatch_warning {} {
  update_mxmbatch_warning_label
  title_batch_maxima
}

proc update_mxmbatch_warning_label {} {
  global maxima_batch_warning_label
  global batch_maxima_input_warning

  if {$maxima_batch_warning_label==""} {
    return
  }

  if {$batch_maxima_input_warning} {
    $maxima_batch_warning_label configure -foreground "#c0f0c0" -background "#a0c0a0"
  } {
    $maxima_batch_warning_label configure -foreground "#c0c0c0" -background "#a0a0a0"
  }
}

set mxmbatch_warning_menu ""
proc offer_mxmbatch_warning_menu {x_abs y_abs x_pos y_pos} {
  global topw
  global maxima_batch_warning_label
  global mxmbatch_warning_menu
  global batch_maxima_input_warning
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  if {$mxmbatch_warning_menu!=""} {
    catch {destroy $mxmbatch_warning_menu}
  }

  set mxmbatch_warning_menu [menu "$topw.mxmbatch_warning_menu" -tearoff false]
  $mxmbatch_warning_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $mxmbatch_warning_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $mxmbatch_warning_menu configure -activeborderwidth 1
  $mxmbatch_warning_menu configure -title "warning on non-empty data"

  $mxmbatch_warning_menu add command -label "To warn at app closing:" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  $mxmbatch_warning_menu add command -label "yes, for non-empty input/output" -command [list "set_batch_maxima_input_warning"]
  $mxmbatch_warning_menu add command -label "not for maxima batch parts" -command [list "unset_batch_maxima_input_warning"]

  set use_idx 2
  if {$batch_maxima_input_warning} {
    set use_idx 1
  }
  $mxmbatch_warning_menu entryconfigure $use_idx -background $menu_selected_background -activebackground "#e0e0ff"

  tk_popup $mxmbatch_warning_menu $use_x $use_y
}

proc set_batch_maxima_input_warning {} {
  global batch_maxima_input_warning
  global batch_maxima_output_warning

  set batch_maxima_input_warning true
  set batch_maxima_output_warning true

  update_mxmbatch_warning
}
proc set_batch_maxima_output_warning {} {
  global batch_maxima_input_warning
  global batch_maxima_output_warning

  set batch_maxima_input_warning true
  set batch_maxima_output_warning true

  update_mxmbatch_warning
}
proc unset_batch_maxima_input_warning {} {
  global batch_maxima_input_warning
  global batch_maxima_output_warning

  set batch_maxima_input_warning false
  set batch_maxima_output_warning false

  update_mxmbatch_warning
}
proc unset_batch_maxima_output_warning {} {
  global batch_maxima_input_warning
  global batch_maxima_output_warning

  set batch_maxima_input_warning false
  set batch_maxima_output_warning false

  update_mxmbatch_warning
}
proc switch_batch_maxima_puts_warning {} {
  global batch_maxima_input_warning
  global batch_maxima_output_warning

  if {($batch_maxima_input_warning)&&($batch_maxima_output_warning)} {
    set batch_maxima_input_warning false
    set batch_maxima_output_warning false
  } elseif {(!$batch_maxima_input_warning)&&(!$batch_maxima_output_warning)} {
    set batch_maxima_input_warning true
    set batch_maxima_output_warning true
  } else {
    set batch_maxima_input_warning true
    set batch_maxima_output_warning true
  }

  update_mxmbatch_warning
}

# base stuff for (links to) tags

set tag_regexp_definition_tilde_line ""
proc get_til_regexp {} {
  global tag_regexp_definition_tilde_line
  return $tag_regexp_definition_tilde_line
}
proc get_til_regexp_prepare {} {
  global tag_regexp_definition_tilde_line
  global min_tilde_count

  set reg_parts [list]
  foreach one_char [list "~"] {
    lappend reg_parts [strcat "(?:" "^\\$one_char\{$min_tilde_count,\}\$" ")"]
  }
  set tag_regexp_definition_tilde_line [join $reg_parts "|"]
}
get_til_regexp_prepare

set tag_regexp_definition_section ""
proc get_sec_regexp {} {
  global tag_regexp_definition_section
  return $tag_regexp_definition_section
}

proc get_sec_regexp_prepare {} {
  global tag_regexp_definition_section
  global min_other_count

  set reg_parts [list]
  foreach one_char [list "*" "=" "#"] {
    lappend reg_parts [strcat "(?:" "^\\$one_char\{$min_other_count,\}\$" ")"]
  }
  set tag_regexp_definition_section [join $reg_parts "|"]
}
get_sec_regexp_prepare

set tag_regexp_definition_declared ""
set tag_regexp_definition_link ""
proc get_tag_regexp {link} {
  global tag_regexp_definition_declared
  global tag_regexp_definition_link

  return [expr {$link ? $tag_regexp_definition_link : $tag_regexp_definition_declared}]
}

set tag_def_lc "*"
set tag_def_uc "="
set tag_def_lc_re "\\*"
set tag_def_uc_re "\\="
set tag_type_document_ids [list "doc" "document"]
set tag_type_topic_ids [list "top" "topic"]
proc get_tag_regexp_prepare {} {
  global tag_regexp_definition_declared
  global tag_regexp_definition_link
  global tag_def_lc_re
  global tag_def_uc_re
  global tag_border_size

  # using for the simplest reasonable way (incl. without escaping the tags);
  # always taken case insensitive;
  # ...should at some point add (transient) highlighting by direct regexp searches;

  # brackets
  set lbr "\\\["
  set rbr "\\\]"
  # backslash
  set bks "\\\\"
  # escaped chars
  set hsh "\\#"
  set ats "\\@"
  set cln "\\:"
  # braces
  set lbc "\\\{"
  set rbc "\\\}"
  # parentheses
  set lpr "\\("
  set rpr "\\)"
  # (double)quote char
  set dqt "\""
  # considering tag types with ASCII alphanum and _- chars only;
  # tag type char
  set ttc "a-zA-Z0-9_-"
  # comparison mode
  set cmp [strcat "\[" $tag_def_lc_re $tag_def_uc_re "\]?"]
  set cmp_uc [strcat "\[" $tag_def_uc_re "\]?"]

  set border_left [string repeat $lbr $tag_border_size]
  set border_right [string repeat $rbr $tag_border_size]

  set valc [strcat "\[^" $lbr $rbr $bks $lbc $rbc $lpr $rpr $dqt "\]"]
  set paired_brackets [strcat "(?:$lbr" "(?!\[$ats$hsh\])" "$valc*$rbr$valc*)"]
  set paired_parentheses "(?:$lpr$valc*$rpr$valc*)"
  set paired_quotes "(?:$dqt$valc*$dqt$valc*)"

  # requiring non-empty tag type, though allowing empty tag value;
  set tag_re_base [strcat "$cln$valc*" "(?:$paired_brackets|$paired_parentheses|$paired_quotes)*"]
  set tag_re_base_decl [strcat "\[$ttc\]+" $tag_re_base]
  set tag_re_base_link [strcat "(?:(?:\[$ttc\]+)|(?:\[$ttc\]*$tag_def_lc_re))" $tag_re_base]

  set tag_re_decl [strcat "$border_left$hsh" $tag_re_base_decl "$border_right"]

  set space "\\ \\t"
  # requiring at least one "/" for dir part;
  set dir_trans [strcat "(?:$cln" "\\d*)"]
  set remote_dir [strcat "(?:" $dir_trans "|(?:\\~)|(?:\\.\{1,2\})|(?:/\[a-zA-Z\]\\:)|" \
    "(?:\[\\.$space$ttc\]*\[$space$ttc\]+\[\\.$space$ttc\]*))?" "(?:/\[\\.$space$ttc\]*)*" "/"]
  # allowing empty glob part;
  set remote_glob [strcat "\[\\*$space$ttc\]*$cmp_uc" "$dir_trans?"]

  # should allow more than one decl-tag part here (probably just one/two by now);
  set tag_re_link [strcat "$border_left$ats" "(?:" $remote_dir $remote_glob "$bks)?" \
    $tag_re_base_link "(?:$bks" "\{1,2\}" $tag_re_base_link ")?" "$cmp$border_right"]

  set tag_regexp_definition_declared $tag_re_decl
  set tag_regexp_definition_link $tag_re_link
  return

  # [#type:value] ...declared tag;
  #
  # type: simple word-like
  # value: a more-free form, without backslash and braces,
  #        and only allowing non-nested matched prentheses, brackets, and double-quotes;
  # ...and without any nesting of (links to) tags either;
  #
  # [@type:value] ...local link to a tag,
  # ...alike declared tags, with possible */= specifiers for limited/complete comparisons;
  #    and with the tag-type part possibly ending by the * specifier too;
  #
  # [@.dir./.glob.\type:value] ...remote link to a tag,
  # ...alike local links, with "\" in front of seeked tag, and (possibly) with dir and/or glob specs;
  #
  #...here only dealing with flat tags (should be simpler for other dealings within TGM Minima);
  #...going this way (and only later we can see whether to add nested links
  #   or remote links to (local there) links at all);
  #
  # may be later to add tag combinations in links:
  # [@ ... \ tag1 (\|\\) tag2 (\|\\) ... (\|\\) tagN]
  # [@./\sec:asdf\sec:asdf/zxcv\exp:1234]
  # [@./\sec:asdf\sec:asdf/zxcv\\exp:1234]
  #
  # ...regarding single backslash:
  # tags considered as required to be present between the line separators (#=*~) along with the preceding ones;
  # all tags required to be there, and there may be line separators of lower power (# > = > * > ~) in between them
  # than is the first line separator above the preceding tag;
  # ...if no line separator above (possibly e.g. for file-tags), no restriction by line separators there then;
  #
  # ...regarding double backslash:
  # tags considered as required to be present anywhere within the current file;
  #
  # ...this should be implemented at some later time (by now just reserving the # character out of tags);
  #
  # remote searches at html export can only be done for file tags
  # (without limited comparisons), probably forced lowercase;
  # ...other remote modes would require auxiliary server-side stuff;
  # ...may be to do some AJAX stuff for that eventually (not now);
  # local searches should be more flexible even for html export
  # (possibly with the help of javascript);

}
get_tag_regexp_prepare

proc extract_tag_list {links} {
  global text_field

  set tags_found [list]

  set tag_reg [expr {$links ? [get_tag_regexp true] : [get_tag_regexp false]}]
  set len_var [list]
  set pos_var [$text_field search -regexp -count len_var -all -overlap $tag_reg "1.0"]

  set tag_idx 0
  set tag_count [llength $pos_var]
  while {$tag_idx<$tag_count} {
    set curr_tag_pos [lindex $pos_var $tag_idx]
    set curr_tag_pos_info [split $curr_tag_pos "."]
    set curr_tag_len [lindex $len_var $tag_idx]
    set curr_tag [$text_field get $curr_tag_pos "$curr_tag_pos + $curr_tag_len chars"]
    lappend tags_found [list $curr_tag $curr_tag_pos $curr_tag_len]
    incr tag_idx
  }

  return $tags_found
}

proc set_tags_shown_state {value} {
  global tags_decls_state_clean
  global tags_links_state_clean

  if {($value==true) || ($value==1) || ($value=="t")} {
    set tags_decls_state_clean true
    set tags_links_state_clean true
  } {
    set tags_decls_state_clean false
    set tags_links_state_clean false
  }
}

proc clear_tag_info {} {
  global tags_decls_id
  global tags_decls_pos
  global tags_decls_size
  global tags_decls_show
  global tags_decls_show_default_text
  global tags_decls_mark_name
  global tags_decls_state_clean
  global tags_decls_list
  global tags_links_id
  global tags_links_pos
  global tags_links_size
  global tags_links_show
  global tags_links_show_default_text
  global tags_links_mark_name
  global tags_links_state_clean
  global tags_links_list

  set tags_decls_id -1
  set tags_decls_pos ""
  set tags_decls_size 0
  set tags_decls_list [list]
  set tags_decls_state_clean false
  set tags_decls_show $tags_decls_show_default_text
  catch {$text_field mark unset $tags_decls_mark_name}

  set tags_links_id -1
  set tags_links_pos ""
  set tags_links_size 0
  set tags_links_list [list]
  set tags_links_state_clean false
  set tags_links_show $tags_links_show_default_text
  catch {$text_field mark unset $tags_links_mark_name}
}

set tags_decls_id -1
set tags_decls_pos ""
set tags_decls_size 0
set tags_decls_list [list]
set tags_decls_mark_name "tags_decls_current"
set tags_decls_state_clean false
set tags_links_id -1
set tags_links_pos ""
set tags_links_size 0
set tags_links_list [list]
set tags_links_mark_name "tags_links_current"
set tags_links_state_clean false
proc slide_tags_inner {form change} {
  global text_field
  global tags_decls_id
  global tags_decls_pos
  global tags_decls_size
  global tags_decls_show
  global tags_decls_show_default_text
  global tags_decls_mark_name
  global tags_decls_state_clean
  global tags_decls_list
  global tags_links_id
  global tags_links_pos
  global tags_links_size
  global tags_links_show
  global tags_links_show_default_text
  global tags_links_mark_name
  global tags_links_state_clean
  global tags_links_list
  global tag_border_size

  if {$form=="decls"} {
    if {$tags_decls_state_clean} {
      set current_tags_use $tags_decls_list
    } {
      set current_tags_use [extract_tag_list false]
      set tags_decls_list $current_tags_use
    }
    set tags_use_id $tags_decls_id
    set tags_use_pos $tags_decls_pos
    set tags_use_size $tags_decls_size
    set tags_use_show $tags_decls_show
    set tags_use_show_default_text $tags_decls_show_default_text
    set tags_use_mark_name $tags_decls_mark_name
    set tags_use_state_clean $tags_decls_state_clean
  } elseif {$form=="links"} {
    if {$tags_links_state_clean} {
      set current_tags_use $tags_links_list
    } {
      set current_tags_use [extract_tag_list true]
      set tags_links_list $current_tags_use
    }
    set tags_use_id $tags_links_id
    set tags_use_pos $tags_links_pos
    set tags_use_size $tags_links_size
    set tags_use_show $tags_links_show
    set tags_use_show_default_text $tags_links_show_default_text
    set tags_use_mark_name $tags_links_mark_name
    set tags_use_state_clean $tags_links_state_clean
  } {
    return false
  }

  set current_tags_use_len [llength $current_tags_use]

  if {$current_tags_use_len==0} {
    set tags_use_id -1
    set tags_use_pos ""
    set tags_use_size 0
    set tags_use_show $tags_use_show_default_text
    catch {$text_field mark unset $tags_use_mark_name}
    return [list $tags_use_id $tags_use_pos $tags_use_size $tags_use_show]
  }

  if {$change==0} {
    # if no-sliding action and without a (link to) tag selected,
    # remaining in the no-selected state (just the tag list got read here);
    if {$tags_use_id==-1} {
      set tags_use_pos ""
      set tags_use_size 0
      set tags_use_show $tags_use_show_default_text
      catch {$text_field mark unset $tags_use_mark_name}
      return [list $tags_use_id $tags_use_pos $tags_use_size $tags_use_show]
    }
    # if no-sliding action and (supposing that) no outer change had been done;
    # we just re-read the tag info according to the given id;
    if {$tags_use_state_clean} {
      set tags_info [lindex $current_tags_use $tags_use_id]
      set tags_use_pos [lindex $tags_info 1]
      set tags_use_size [lindex $tags_info 2]
      set tags_use_show [string range [lindex $tags_info 0] $tag_border_size "end-$tag_border_size"]
      catch {$text_field mark set $tags_use_mark_name $tags_use_pos}
      return [list $tags_use_id $tags_use_pos $tags_use_size $tags_use_show]
    }
  }

  if {$tags_use_state_clean || ($tags_use_id==-1)} {
    set tags_use_state_clean true
    # if sure that text_field has not changed its content
    incr tags_use_id $change
    if {$tags_use_id>=$current_tags_use_len} {
      set tags_use_id 0
    }
    if {$tags_use_id<0} {
      set tags_use_id [expr {$current_tags_use_len - 1}]
    }
    set tags_use_show [string range [lindex [lindex $current_tags_use $tags_use_id] 0] $tag_border_size "end-$tag_border_size"]
    set tags_use_pos [lindex [lindex $current_tags_use $tags_use_id] 1]
    set tags_use_size [lindex [lindex $current_tags_use $tags_use_id] 2]
    $text_field mark set $tags_use_mark_name "$tags_use_pos"
    return [list $tags_use_id $tags_use_pos $tags_use_size $tags_use_show]
  } {
    # otherwise a sot-of fuzzy search for the current tag position/content via:
    # marked_string: from the mark,
    # presented string: at the entry (plus borders),
    # slide_string: from the extracted list of (links to) tags;
    set tag_changed false
    set tag_found false
    set curr_slide_id -1
    set mark_pos ""
    set marked_string ""
    catch {
      # taking where the tag is located currently, since it might had moved
      set mark_pos [$text_field index $tags_use_mark_name]
      # taking the tag string (as if the tag itself was not edited)
      set marked_string [$text_field get $mark_pos "$mark_pos +$tags_use_size chars"]
    }
    set presented_string [string repeat "\[" $tag_border_size]
    append presented_string $tags_use_show [string repeat "\]" $tag_border_size]
    set slide_string [lindex [lindex $current_tags_use $tags_use_id] 0]
    if {($mark_pos!="") && ($marked_string!="") && ($marked_string==$presented_string)} {
      # if the stuff that is hown in the entry corresponds to the stuff by the mark
      if {$slide_string==$presented_string} {
        # and if the count of tags before the current tag was not changed
        set curr_slide_id $tags_use_id
        set tag_found true
      } {
        # otherwise looking for the index of the tag within the list of the extracted (links to) tags
        set curr_idx 0
        while {$curr_idx<$current_tags_use_len} {
          if {[lindex [lindex $current_tags_use $curr_idx] 1]==$mark_pos} {
            set curr_slide_id $curr_idx
            set tag_found true
            break
          }
          incr curr_idx
        }
      }
    } {
      # if presented string is not the marked string, but if having the mark set at first
      if {(!$tag_found) && ($mark_pos!="")} {
        set curr_idx 0
        while {$curr_idx<$current_tags_use_len} {
          if {[lindex [lindex $current_tags_use $curr_idx] 1]==$mark_pos} {
            set curr_slide_id $curr_idx
            set tag_found true
            if {[lindex [lindex $current_tags_use $curr_idx] 0]!=$presented_string} {
              set tag_changed true
            }
            break
          }
          incr curr_idx
        }
      }
      # if the mark does not help here
      if {!$tag_found} {
        set curr_dist ""
        set curr_idx 0
        set use_pos [expr {($mark_pos!="") ? $mark_pos : $tags_use_pos}]
        # looking for an occurrence closest to a (sort-of) expected position
        while {$curr_idx<$current_tags_use_len} {
          if {[lindex [lindex $current_tags_use $curr_idx] 0]==$presented_string} {
            set test_dist [lindex [$text_field count -chars $use_pos [lindex [lindex $current_tags_use $curr_idx] 1]] 0]
            if {($curr_dist=="") || ($curr_dist>=$test_dist)} {
              set curr_slide_id $curr_idx
              set curr_dist $test_dist
              set tag_found true
            }
          }
          incr curr_idx
        }
      }
    }
    # next to deal with the possibly changed stuff
    if {$tag_found} {
      set tags_use_id $curr_slide_id
      # if the tag content changed itself, remaining on it (to show it at first)
      if {!$tag_changed} {
        incr tags_use_id $change
        if {$tags_use_id>=$current_tags_use_len} {
          set tags_use_id 0
        }
        if {$tags_use_id<0} {
          set tags_use_id [expr {$current_tags_use_len - 1}]
        }
      }
      set tags_use_pos [lindex [lindex $current_tags_use $tags_use_id] 1]
      set tags_use_size [lindex [lindex $current_tags_use $tags_use_id] 2]
      set tags_use_show [string range [lindex [lindex $current_tags_use $tags_use_id] 0] $tag_border_size "end-$tag_border_size"]
      catch {$text_field mark set $tags_use_mark_name $tags_use_pos}
      set tags_use_state_clean true
      return [list $tags_use_id $tags_use_pos $tags_use_size $tags_use_show]
    } {
      # if the current tag was not found, showing the default text
      set tags_use_id -1
      set tags_use_pos ""
      set tags_use_size 0
      set tags_use_show $tags_use_show_default_text
      catch {$text_field mark unset $tags_use_mark_name}
      set tags_use_state_clean true
      return [list $tags_use_id $tags_use_pos $tags_use_size $tags_use_show]
    }
  }
}

proc slide_tags_declared {change {pause_at_top false}} {
  global text_field
  global tags_decls_id
  global tags_decls_pos
  global tags_decls_size
  global tags_decls_show
  global tags_decls_show_default_text
  global tags_decls_mark_name
  global tags_decls_state_clean
  global tags_decls_list

  if {$pause_at_top} {
    set show_default_text false
    if {$tags_decls_state_clean} {
      if {($change==1) && ($tags_decls_id == [expr {[llength $tags_decls_list] - 1}])} {
        set show_default_text true
      }
      if {($change==-1) && ($tags_decls_id == 0)} {
        set show_default_text true
      }
    }
    if {$show_default_text} {
      set tags_decls_id -1
      set tags_decls_pos "1.0"
      set tags_decls_size 0
      set tags_decls_show $tags_decls_show_default_text
      $text_field mark set $tags_decls_mark_name "1.0"
      return
    }

    if {$tags_decls_state_clean} {
      if {$tags_decls_id==-1} {
        if {$change==1} {
          set change 0
          set tags_decls_id 0
          $text_field mark set $tags_decls_mark_name [lindex [lindex $tags_decls_list $tags_decls_id] 1]
        } elseif {$change==-1} {
          set change 0
          set tags_decls_id [expr {[llength $tags_decls_list] - 1}]
          $text_field mark set $tags_decls_mark_name [lindex [lindex $tags_decls_list $tags_decls_id] 1]
        }
      }
    }
  }

  set slide_res [slide_tags_inner "decls" $change]
  set tags_decls_id [lindex $slide_res 0]
  set tags_decls_pos [lindex $slide_res 1]
  set tags_decls_size [lindex $slide_res 2]
  set tags_decls_show [lindex $slide_res 3]
  set tags_decls_state_clean true
}

proc slide_tags_links {change {pause_at_top false}} {
  global text_field
  global tags_links_id
  global tags_links_pos
  global tags_links_size
  global tags_links_show
  global tags_links_show_default_text
  global tags_links_mark_name
  global tags_links_state_clean
  global tags_links_list

  if {$pause_at_top} {
    set show_default_text false
    if {$tags_links_state_clean} {
      if {($change==1) && ($tags_links_id == [expr {[llength $tags_links_list] - 1}])} {
        set show_default_text true
      }
      if {($change==-1) && ($tags_links_id == 0)} {
        set show_default_text true
      }
    }
    if {$show_default_text} {
      set tags_links_id -1
      set tags_links_pos "1.0"
      set tags_links_size 0
      set tags_links_show $tags_links_show_default_text
      $text_field mark set $tags_links_mark_name "1.0"
      return
    }

    if {$tags_links_state_clean} {
      if {$tags_links_id==-1} {
        if {$change==1} {
          set change 0
          set tags_links_id 0
          $text_field mark set $tags_links_mark_name [lindex [lindex $tags_links_list $tags_links_id] 1]
        } elseif {$change==-1} {
          set change 0
          set tags_links_id [expr {[llength $tags_links_list] - 1}]
          $text_field mark set $tags_links_mark_name [lindex [lindex $tags_links_list $tags_links_id] 1]
        }
      }
    }
  }

  set slide_res [slide_tags_inner "links" $change]
  set tags_links_id [lindex $slide_res 0]
  set tags_links_pos [lindex $slide_res 1]
  set tags_links_size [lindex $slide_res 2]
  set tags_links_show [lindex $slide_res 3]
  set tags_links_state_clean true
}

proc show_all_sections {} {
  global text_field
  global search_tag_all

  set reg_parts [list [strcat "(?:" [get_sec_regexp] ")"]]
  set use_reg [join $reg_parts "|"]

  # according to current definitions, the "-overlap" switch is not required here;
  # and since not dealing with ~-lines here, neither the "-elide" switch is used;
  set search_idx_starts [$text_field search -all -count search_idx_lengths -regexp -- $use_reg "1.0" "end"]

  unset_search_highlights $text_field $search_tag_all
  if {[llength $search_idx_starts]==0} {
    return true
  }

  set_search_highlights $text_field $search_tag_all $search_idx_starts $search_idx_lengths
  focus $text_field
  seek_next

  return true
}

proc show_all_tags {declared {with_other false} {to_focus true} {to_seek true}} {
  global text_field
  global search_tag_all
  global tag_border_size

  set reg_parts [list]

  if {$declared} {
    lappend reg_parts [strcat "(?:" [get_tag_regexp false] ")"]
    if {$with_other} {
      lappend reg_parts [strcat "(?:" [get_tag_regexp true] ")"]
    }
  } {
    lappend reg_parts [strcat "(?:" [get_tag_regexp true] ")"]
    if {$with_other} {
      lappend reg_parts [strcat "(?:" [get_tag_regexp false] ")"]
    }
  }

  set use_reg [join $reg_parts "|"]

  # according to current definitions, the "-overlap" switch is not required here;
  # and since not dealing with ~-lines here, neither the "-elide" switch is used;
  set search_idx_starts [$text_field search -all -count search_idx_lengths -regexp -- $use_reg "1.0" "end"]

  unset_search_highlights $text_field $search_tag_all

  if {[llength $search_idx_starts]==0} {
    set message ""
    if {$with_other} {
      set message "no (links to) tags "
    } elseif {$declared} {
      set message "no tags "
    } {
      set message "no links to tags "
    }
    append message "in document"

    show_text_field_message_error [list [list $message]]
    return true
  }

  set use_lengths [list]
  foreach one_len $search_idx_lengths {
    lappend use_lengths [expr {$one_len - 2*$tag_border_size}]
  }

  set use_starts [list]
  foreach one_pos $search_idx_starts {
    lappend use_starts [strcat $one_pos " +" $tag_border_size " chars"]
  }

  set_search_highlights $text_field $search_tag_all $use_starts $use_lengths
  if {$to_focus} {
    focus $text_field
    if {$to_seek} {
      seek_next
    }
  }

  return true
}

# with checking whether the tag is still located there,
# and adjusting it if located nearby;
# otherwise (if not found) at all, presenting the default text,
# to display that the tag was not found;
proc see_tag_entry {declared to_focus to_select} {
  global text_field
  global tags_decls_pos
  global tags_links_pos
  global tags_decls_state_clean
  global tags_links_state_clean
  global tags_decls_size
  global tags_links_size
  global tag_border_size
  global tags_decls_id
  global tags_links_id

  if {$declared} {
    if {!$tags_decls_state_clean} {
      slide_tags_declared 0
    }
    if {$tags_decls_id==-1} {
      return false
    }
    if {$tags_decls_pos==""} {
      return false
    }
    catch {
      if {!$to_select} {
        set tag_pos_see "$tags_decls_pos +$tag_border_size chars"
        $text_field mark set "insert" $tag_pos_see
        $text_field see $tag_pos_see
        set_anchor_insert $text_field
      } {
        set tag_pos_see "$tags_decls_pos +$tags_decls_size chars"
        $text_field tag remove "sel" "1.0" "end"
        set_anchor_position $text_field $tags_decls_pos
        $text_field mark set "insert" $tag_pos_see
        $text_field tag add "sel" "$tags_decls_pos" "$tag_pos_see"
        $text_field see $tag_pos_see
      }
      if {$to_focus} {
        focus $text_field
      }
    }
    return true
  } {
    if {!$tags_links_state_clean} {
      slide_tags_links 0
    }
    if {$tags_links_id==-1} {
      return false
    }
    if {$tags_links_pos==""} {
      return false
    }
    catch {
      if {!$to_select} {
        set tag_pos_see "$tags_links_pos +$tag_border_size chars"
        $text_field mark set "insert" $tag_pos_see
        $text_field see $tag_pos_see
        set_anchor_insert $text_field
      } {
        set tag_pos_see "$tags_links_pos +$tags_links_size chars"
        $text_field tag remove "sel" "1.0" "end"
        set_anchor_position $text_field $tags_links_pos
        $text_field mark set "insert" $tag_pos_see
        $text_field tag add "sel" "$tags_links_pos" "$tag_pos_see"
        $text_field see $tag_pos_see
      }
      if {$to_focus} {
        focus $text_field
      }
    }
    return true
  }
}

proc highligh_current_tag {declared} {
  global text_field
  global tags_decls_pos
  global tags_links_pos
  global tags_decls_state_clean
  global tags_links_state_clean
  global tags_decls_size
  global tags_links_size
  global tag_border_size
  global tags_decls_id
  global tags_links_id
  global search_tag_all
  global tag_border_size

  unset_search_highlights $text_field $search_tag_all

  if {$declared} {
    if {$tags_decls_id!=-1} {
      see_tag_entry $declared false false
      set_search_highlights $text_field $search_tag_all \
        [list "$tags_decls_pos +$tag_border_size chars"] [list [expr {$tags_decls_size - 2*$tag_border_size}]]
    } {
      return false
    }
  } {
    if {$tags_links_id!=-1} {
      see_tag_entry $declared false false
      set_search_highlights $text_field $search_tag_all \
        [list "$tags_links_pos +$tag_border_size chars"] [list [expr {$tags_links_size - 2*$tag_border_size}]]
    } {
      return false
    }
  }
  update_text_highlight_lines_onset
  update_text_bracket_pairing

  return true
}

proc find_closest_tag {declared} {
  global text_field
  global tags_decls_state_clean
  global tags_links_state_clean
  global tags_decls_list
  global tags_links_list

  set use_list [list]

  if {$declared} {
    if {!$tags_decls_state_clean} {
      set tags_decls_list [extract_tag_list false]
      set tags_decls_state_clean true
    }
    set use_list $tags_decls_list
  } {
    if {!$tags_links_state_clean} {
      set tags_links_list [extract_tag_list true]
      set tags_links_state_clean true
    }
    set use_list $tags_links_list
  }

  if {[llength $use_list]==0} {
    return [list false]
  }

  set cursor_pos [$text_field index "insert"]
  set cursor_info [split $cursor_pos "."]
  set cursor_row [lindex $cursor_info 0]
  set cursor_col [lindex $cursor_info 1]

  set prev_tag ""
  set prev_tag_row ""
  set prev_tag_col ""
  set prev_tag_idx -1

  set next_tag ""
  set next_tag_row ""
  set next_tag_col ""
  set next_tag_idx -1

  set one_tag_idx -1
  foreach one_tag $use_list {
    incr one_tag_idx
    set one_tag_pos_info [split [lindex $one_tag 1] "."]
    set one_tag_pos_row [lindex $one_tag_pos_info 0]
    set one_tag_pos_col [lindex $one_tag_pos_info 1]
    if {($one_tag_pos_row<$cursor_row) || (($one_tag_pos_row==$cursor_row) && ($one_tag_pos_col<$cursor_col))} {
      set prev_tag $one_tag
      set prev_tag_row $one_tag_pos_row
      set prev_tag_col $one_tag_pos_col
      set prev_tag_idx $one_tag_idx
      continue
    }
    set next_tag $one_tag
    set next_tag_row $one_tag_pos_row
    set next_tag_col $one_tag_pos_col
    set next_tag_idx $one_tag_idx
    break
  }

  set use_tag ""
  set use_tag_row ""
  set use_tag_col ""
  set use_tag_idx ""

  if {$prev_tag!=""} {
    if {($prev_tag_row==$cursor_row) && ([expr {$prev_tag_col + [lindex $prev_tag 2]}]>=$cursor_col)} {
      set use_tag $prev_tag
      set use_tag_row $prev_tag_row
      set use_tag_col $prev_tag_col
      set use_tag_idx $prev_tag_idx
    }
  }

  if {$use_tag==""} {
    if {($prev_tag!="") && ($next_tag=="")} {
      set use_tag $prev_tag
      set use_tag_row $prev_tag_row
      set use_tag_col $prev_tag_col
      set use_tag_idx $prev_tag_idx
    }
    if {($prev_tag=="") && ($next_tag!="")} {
      set use_tag $next_tag
      set use_tag_row $next_tag_row
      set use_tag_col $next_tag_col
      set use_tag_idx $next_tag_idx
    }
  }

  if {$use_tag==""} {
    set prev_end [strcat [lindex $prev_tag 1] " +" [lindex $prev_tag 2] " chars"]
    set prev_dist [lindex [$text_field count -displaychars [$text_field index $prev_end] $cursor_pos] 0]

    set next_dist [lindex [$text_field count -displaychars $cursor_pos [lindex $next_tag 1]] 0]

    if {$prev_dist<=$next_dist} {
      set use_tag $prev_tag
      set use_tag_row $prev_tag_row
      set use_tag_col $prev_tag_col
      set use_tag_idx $prev_tag_idx
    } {
      set use_tag $next_tag
      set use_tag_row $next_tag_row
      set use_tag_col $next_tag_col
      set use_tag_idx $next_tag_idx
    }
  }

  return [list true $use_tag_idx $use_tag $use_tag_row $use_tag_col]
}

proc set_entry_on_closest_tag {declared} {
  global text_field
  global tag_border_size
  global tags_decls_id
  global tags_links_id
  global tags_decls_mark_name
  global tags_links_mark_name

  set found_tag_info [find_closest_tag $declared]
  if {![lindex $found_tag_info 0]} {
    return
  }

  set found_tag_idx [lindex $found_tag_info 1]
  set found_tag [lindex $found_tag_info 2]
  set found_tag_pos [lindex $found_tag 1]

  if {$declared} {
    set tags_decls_id $found_tag_idx
    $text_field mark set $tags_decls_mark_name $found_tag_pos
    $text_field mark set "insert" "$found_tag_pos +$tag_border_size chars"
    slide_tags_declared 0
  } {
    set tags_links_id $found_tag_idx
    $text_field mark set $tags_links_mark_name $found_tag_pos
    $text_field mark set "insert" "$found_tag_pos +$tag_border_size chars"
    slide_tags_links 0
  }
}

proc check_have_some_tags {declared {to_slip false}} {
  global tags_decls_state_clean
  global tags_links_state_clean
  global tags_decls_list
  global tags_links_list

  if {$declared} {
    if {!$tags_decls_state_clean} {
      set tags_decls_list [extract_tag_list false]
      set tags_decls_state_clean true
    }
    set use_list $tags_decls_list
    if {$to_slip} {
      slide_tags_declared 0
    }
  } {
    if {!$tags_links_state_clean} {
      set tags_links_list [extract_tag_list true]
      set tags_links_state_clean true
    }
    set use_list $tags_links_list
    if {$to_slip} {
      slide_tags_links 0
    }
  }

  return [expr {[llength $use_list]>0}]
}

bind $tags_decls_line <Right> {
  slide_tags_declared 1
  break
}
bind $tags_decls_line <Left> {
  slide_tags_declared -1
  break
}

bind $tags_links_line <Right> {
  slide_tags_links 1
  break
}
bind $tags_links_line <Left> {
  slide_tags_links -1
  break
}

bind $tags_decls_line <Tab> {
  slide_tags_declared 1
  highligh_current_tag true
  break
}
bind $tags_links_line <Tab> {
  slide_tags_links 1
  highligh_current_tag false
  break
}

foreach keyspressed [list "Shift-Key-Tab" "Shift-Key-KP_Tab" "Shift-Key-ISO_Left_Tab" "Key-ISO_Left_Tab"] {
  catch {
    bind $tags_decls_line <$keyspressed> {
      slide_tags_declared -1
      highligh_current_tag true
      break
    }
  }
  catch {
    bind $tags_links_line <$keyspressed> {
      slide_tags_links -1
      highligh_current_tag false
      break
    }
  }
}

bind $tags_decls_line <Return> {
  if {[check_have_some_tags true true]} {
    highligh_current_tag true
  }
  update_text_highlight_lines_onset
  break
}
bind $tags_decls_line <Shift-Return> {
  set_entry_on_closest_tag true
  if {[check_have_some_tags true false]} {
    highligh_current_tag true
  }
  update_text_highlight_lines_onset
  break
}
bind $tags_decls_line <Control-Return> {
  if {[check_have_some_tags true true]} {
    highligh_current_tag true
    see_tag_entry true true false
  }
  update_text_highlight_lines_onset
  break
}
bind $tags_decls_line <Control-Shift-Return> {
  if {[check_have_some_tags true true]} {
    highligh_current_tag true
    see_tag_entry true true true
  }
  update_text_highlight_lines_onset
  break
}
bind $tags_decls_line <Alt-Return> {
  if {[check_have_some_tags true true]} {
    show_all_tags true false true true
  }
  update_text_highlight_lines_onset
  break
}
bind $tags_decls_line <Control-Alt-Return> {
  break
}
bind $tags_decls_line <Shift-Alt-Return> {
  break
}
bind $tags_decls_line <Control-Shift-Alt-Return> {
  break
}

bind $tags_links_line <Return> {
  if {[check_have_some_tags false true]} {
    highligh_current_tag false
  }
  update_text_highlight_lines_onset
  break
}
bind $tags_links_line <Shift-Return> {
  set_entry_on_closest_tag false
  if {[check_have_some_tags false false]} {
    highligh_current_tag false
  }
  update_text_highlight_lines_onset
  break
}
bind $tags_links_line <Control-Return> {
  if {[check_have_some_tags false true]} {
    highligh_current_tag false
    see_tag_entry false true false
  }
  update_text_highlight_lines_onset
  break
}
bind $tags_links_line <Control-Shift-Return> {
  if {[check_have_some_tags false true]} {
    highligh_current_tag false
    see_tag_entry false true true
  }
  update_text_highlight_lines_onset
  break
}
bind $tags_links_line <Alt-Return> {
  if {[check_have_some_tags false true]} {
    show_all_tags false false true true
  }
  update_text_highlight_lines_onset
  break
}
bind $tags_links_line <Control-Alt-Return> {
  break
}
bind $tags_links_line <Shift-Alt-Return> {
  break
}
bind $tags_links_line <Control-Shift-Alt-Return> {
  break
}

proc show_tags_declared {current} {
  # to show menu with (a subset of) all the declared tags
  global tags_decls_state_clean
  global text_field
  global tags_decls_id
  global tags_decls_list

  if {!$tags_decls_state_clean} {
    slide_tags_declared 0
  }

  if {[is_stack_mode]} {
    set tags_decls_id -1
  }
  if {$current==""} {
    set current false
    set ins_pos_info [split [$text_field index "insert"] "."]
    set ins_pos_row [lindex $ins_pos_info 0]
    set ins_pos_col [lindex $ins_pos_info 1]
    set tag_idx -1
    foreach one_tag_info $tags_decls_list {
      incr tag_idx
      set one_tag_pos_info [split [lindex $one_tag_info 1] "."]
      set one_tag_row [lindex $one_tag_pos_info 0]
      if {$ins_pos_row>$one_tag_row} {
        continue
      }
      if {$ins_pos_row<$one_tag_row} {
        break
      }
      set one_tag_len [lindex $one_tag_info 2]
      set one_tag_col_start [lindex $one_tag_pos_info 1]
      set one_tag_col_end [expr {$one_tag_col_start + $one_tag_len}]
      if {($ins_pos_col<$one_tag_col_start) || ($ins_pos_col>$one_tag_col_end)} {
        continue
      }
      set tags_decls_id $tag_idx
      slide_tags_declared 0
      set current true
      break
    }
  }

  show_tags_inner "decls" $current
}
proc show_tags_links {current} {
  # to show menu with (a subset of) all the links to tags
  global tags_links_state_clean
  global text_field
  global tags_links_id
  global tags_links_list

  if {!$tags_links_state_clean} {
    slide_tags_links 0
  }

  if {[is_stack_mode]} {
    set tags_links_id -1
  }
  if {$current==""} {
    set current false
    set ins_pos_info [split [$text_field index "insert"] "."]
    set ins_pos_row [lindex $ins_pos_info 0]
    set ins_pos_col [lindex $ins_pos_info 1]
    set tag_idx -1
    foreach one_tag_info $tags_links_list {
      incr tag_idx
      set one_tag_pos_info [split [lindex $one_tag_info 1] "."]
      set one_tag_row [lindex $one_tag_pos_info 0]
      if {$ins_pos_row>$one_tag_row} {
        continue
      }
      if {$ins_pos_row<$one_tag_row} {
        break
      }
      set one_tag_len [lindex $one_tag_info 2]
      set one_tag_col_start [lindex $one_tag_pos_info 1]
      set one_tag_col_end [expr {$one_tag_col_start + $one_tag_len}]
      if {($ins_pos_col<$one_tag_col_start) || ($ins_pos_col>$one_tag_col_end)} {
        continue
      }
      set tags_links_id $tag_idx
      slide_tags_links 0
      set current true
      break
    }
  }

  show_tags_inner "links" $current
}

foreach wdg [list $topw $text_field $command_field $search_entry $tags_decls_line $tags_links_line] {
  bind $wdg <Control-bracketleft> {
    if {[is_batch_mode]} {
      show_tags_declared true
    } {
      if {[is_text_field_active]} {
        show_tags_declared ""
      }
    }
    break
  }
  bind $wdg <Control-Alt-bracketleft> {
    break
  }
  bind $wdg <Control-bracketright> {
    if {[is_batch_mode]} {
      show_tags_links true
    } {
      if {[is_text_field_active]} {
        show_tags_links ""
      }
    }
    break
  }
  bind $wdg <Control-Alt-bracketright> {
    break
  }
}

bind $text_field <Control-Alt-bracketleft> {
  if {[is_text_field_active]} {
    insert_tag_empty_declared
    clear_temporary_highlights
  }
  break
}
bind $text_field <Control-Alt-bracketright> {
  if {[is_text_field_active]} {
    insert_tag_empty_link false
    clear_temporary_highlights
  }
  break
}

bind $tags_decls_line <Control-Tab> {
  show_tags_declared true
  break
}
bind $tags_links_line <Control-Tab> {
  show_tags_links true
  break
}

foreach keyspressed [list "Shift-Key-Tab" "Shift-Key-KP_Tab" "Shift-Key-ISO_Left_Tab" "Key-ISO_Left_Tab"] {
  catch {
    bind $tags_decls_line <Control-$keyspressed> {
      show_tags_declared false
      break
    }
  }
  catch {
    bind $tags_links_line <Control-$keyspressed> {
      show_tags_links false
      break
    }
  }
}

proc remove_tags_menu_shown {} {
  global tags_menu_shown_tears
  global tags_menu_shown_torn
  global tags_menu_shown

  if {!$tags_menu_shown_tears} {
    return
  }
  set tags_menu_shown_tears false

  if {$tags_menu_shown_torn!=""} {
    catch {destroy $tags_menu_shown_torn}
    set tags_menu_shown_torn ""
  }
  if {$tags_menu_shown!=""} {
    catch {destroy $tags_menu_shown}
    set tags_menu_shown ""
  }

  catch {focus $text_field}
}

proc tags_select_from_menu {idx} {
  global text_field
  global tags_menu_shown_form
  global tags_decls_id
  global tags_decls_list
  global tags_decls_state_clean
  global tags_links_id
  global tags_links_list
  global tags_links_state_clean
  global tag_border_size

  if {($idx=="none") || ($idx==-1)} {
    return
  }
  catch {remove_tags_menu_shown}
  if {![is_text_field_active]} {
    return
  }

  # here we expect already read tag list,
  # though re-reading it if some (late) answer came from Maxima
  # (in such a case it may lead to a wrong tag if we got some another
  # as an answer part, though it should be generally rare);
  if {$tags_menu_shown_form=="decls"} {
    if {$tags_decls_state_clean} {
      set use_list $tags_decls_list
    } {
      set use_list [extract_tag_list false]
      set tags_decls_list $use_list
      set tags_decls_state_clean true
    }
    set with_links false
  } elseif {$tags_menu_shown_form=="links"} {
    if {$tags_links_state_clean} {
      set use_list $tags_links_list
    } {
      set use_list [extract_tag_list true]
      set tags_links_list $use_list
      set tags_links_state_clean true
    }
    set use_list $tags_links_list
    set with_links true
  } {
    return
  }

  set use_list_len [llength $use_list]
  if {$use_list_len==0} {
    return
  }

  set curr_tag_info [lindex $use_list $idx]

  if {true} {
    if {$with_links} {
      set tags_links_id $idx
      slide_tags_links 0
      highligh_current_tag false
    } {
      set tags_decls_id $idx
      slide_tags_declared 0
      highligh_current_tag true
    }
  }

  if {true} {
    $text_field tag remove "sel" "1.0" "end"
    set tag_pos [lindex $curr_tag_info 1]
    set tag_pos_see "$tag_pos +$tag_border_size chars"
    $text_field see $tag_pos_see
    $text_field mark set "insert" $tag_pos_see
    update_text_highlight_lines_onset
    set_anchor_insert $text_field
    focus $text_field
  }
}

proc update_tags_menu_entries {forwards step} {
  global max_size_of_tags_menu
  global tags_menu_offset
  global tags_menu_shown
  global tags_menu_shown_form
  global tags_decls_id
  global tags_decls_list
  global tags_links_id
  global tags_links_list
  global tags_menu_label_start
  global tag_border_size
  global menu_selected_background

  if {$tags_menu_shown==""} {
    return
  }

  if {$tags_menu_shown_form=="decls"} {
    set use_id $tags_decls_id
    set use_list $tags_decls_list
  } elseif {$tags_menu_shown_form=="links"} {
    set use_id $tags_links_id
    set use_list $tags_links_list
  } {
    return
  }

  set use_list_len [llength $use_list]
  if {$use_list_len==0} {
    return
  }

  if {$forwards} {
    set offset_idx $tags_menu_offset
    if {$offset_idx>=$use_list_len} {
      set offset_idx 0
    }
  } {
    set offset_idx [expr {$tags_menu_offset - (2*$step)}]
    if {$offset_idx<0} {
      set offset_idx [expr {$use_list_len - ($use_list_len % $max_size_of_tags_menu)}]
    }
  }

  set label_text $tags_menu_label_start
  append label_text "[expr {1 + $offset_idx}] - [expr {min($use_list_len, $max_size_of_tags_menu + $offset_idx)}]"
  append label_text " / $use_list_len"
  $tags_menu_shown entryconfigure 0 -label $label_text

  set curr_idx 0
  set list_idx $offset_idx
  while {$list_idx<$use_list_len} {
    set one_tag_info [lindex $use_list $list_idx]
    set one_name [string range [lindex $one_tag_info 0] $tag_border_size "end-$tag_border_size"]
    incr curr_idx
    $tags_menu_shown entryconfigure $curr_idx -label $one_name -command [list "tags_select_from_menu" $list_idx]
    $tags_menu_shown entryconfigure $curr_idx -background "#ffffff" -state "normal"
    incr list_idx
    if {$curr_idx==$max_size_of_tags_menu} {
      break
    }
  }

  while {$curr_idx<$max_size_of_tags_menu} {
    incr curr_idx
    $tags_menu_shown entryconfigure $curr_idx -label "   " -state disabled -background "#f0f0f0" -command [list "tags_select_from_menu" "none"]
  }

  set id_to_highlight -1
  if {$use_id>=$offset_idx} {
    if {$use_id<[expr {$offset_idx + $max_size_of_tags_menu}]} {
      set id_to_highlight [expr {$use_id - $offset_idx}]
    }
  }

  if {$id_to_highlight>-1} {
    $tags_menu_shown entryconfigure [expr {$id_to_highlight+1}] -background $menu_selected_background -activebackground "#e0e0ff"
  }

  set tags_menu_offset [expr {$offset_idx + $curr_idx}]
}

set max_size_of_tags_menu [get_max_menu_entry_count]
set tags_menu_offset 0
set tags_menu_shown ""
set tags_menu_shown_form ""
set tags_menu_label_start ""
set tags_menu_shown_torn ""
set tags_menu_shown_tears false
proc show_tags_inner {form positioned} {
  global topw
  global max_size_of_tags_menu
  global tags_menu_offset
  global tags_menu_shown
  global tags_menu_shown_form
  global tags_decls_id
  global tags_decls_list
  global tags_links_id
  global tags_links_list
  global tags_menu_label_start
  global tags_menu_shown_tears
  global tags_menu_shown_torn
  global tag_border_size
  global menu_selected_background

  if {$form=="decls"} {
    set use_id $tags_decls_id
    set use_list $tags_decls_list
    set tags_menu_shown_form "decls"
    set with_links false
  } elseif {$form=="links"} {
    set use_id $tags_links_id
    set use_list $tags_links_list
    set tags_menu_shown_form "links"
    set with_links true
  } {
    return
  }

  set use_list_len [llength $use_list]
  if {$use_list_len==0} {
    return
  }

  set abs_pos [get_text_field_position]
  set pos_x [expr {5 + [lindex $abs_pos 0]}]
  set pos_y [expr {5 + [lindex $abs_pos 1]}]

  if {$tags_menu_shown_torn!=""} {
    catch {destroy $tags_menu_shown_torn}
    set tags_menu_shown_torn ""
  }
  catch {destroy $tags_menu_shown}
  set tags_menu_shown [menu "$topw.tagsmenushown" -tearoff false]
  $tags_menu_shown configure -relief solid -borderwidth 1 -background "#f0f0f0" -disabledforeground "#000000"
  $tags_menu_shown configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f8ffc0"
  $tags_menu_shown configure -relief "solid" -borderwidth 1 -activeborderwidth 1

  set shown_by_parts false
  set tags_menu_shown_tears false
  if {$use_list_len>$max_size_of_tags_menu} {
    set shown_by_parts true
    set tags_menu_shown_tears true
  }
  if {[is_linux_like]} {
    set tags_menu_shown_tears false
  }

  if {$shown_by_parts} {
    set backwards_list [list "2" "Control-Left" "Control-P" "Left"]
    foreach keyspressed $backwards_list {
      bind $tags_menu_shown <$keyspressed> {
        catch {
          update_tags_menu_entries false $max_size_of_tags_menu
        }
        break
      }
    }
    set forwards_list [list "3" "Control-Right" "Control-N" "Right"]
    foreach keyspressed $forwards_list {
      bind $tags_menu_shown <$keyspressed> {
        catch {
          update_tags_menu_entries true $max_size_of_tags_menu
        }
        break
      }
    }
  }

  set id_to_highlight $use_id
  if {$shown_by_parts && ($use_id>=$max_size_of_tags_menu)} {
    set id_to_highlight -1
  }

  set offset_idx 0
  if {$positioned && ($use_id>=0)} {
    set offset_idx [expr {$use_id - ($use_id % $max_size_of_tags_menu)}]
    set id_to_highlight [expr {$use_id % $max_size_of_tags_menu}]
  }

  set tags_menu_label_start ""
  set label_text [expr {$with_links ? "Links, " : "Tags, "}]
  if {$shown_by_parts} {
    append label_text "listed part: "
    set tags_menu_label_start $label_text
    append label_text "[expr {1 + $offset_idx}] - [expr {min($use_list_len, $max_size_of_tags_menu + $offset_idx)}]"
    append label_text " / $use_list_len"
  } {
    append label_text "all listed: "
    append label_text "1 - $use_list_len"
  }
  $tags_menu_shown add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  set curr_idx 0
  set list_idx $offset_idx
  while {$list_idx<$use_list_len} {
    set one_tag_info [lindex $use_list $list_idx]
    set one_name [string range [lindex $one_tag_info 0] $tag_border_size "end-$tag_border_size"]
    $tags_menu_shown add command -label $one_name -command [list "tags_select_from_menu" $list_idx]
    incr curr_idx
    incr list_idx
    $tags_menu_shown entryconfigure $curr_idx -background "#ffffff"
    if {$curr_idx==$max_size_of_tags_menu} {
      break
    }
  }

  # to finish a column if it is at an offset part
  if {$shown_by_parts} {
    while {$curr_idx<$max_size_of_tags_menu} {
      $tags_menu_shown add command -label "   " -state disabled -background "#f0f0f0" -command [list "tags_select_from_menu" "none"]
      incr curr_idx
    }
  }

  if {$id_to_highlight>-1} {
    $tags_menu_shown entryconfigure [expr {$id_to_highlight+1}] -background $menu_selected_background -activebackground "#e0e0ff"
  }

  set tags_menu_offset [expr {$offset_idx + $curr_idx}]

  if {$tags_menu_shown_tears} {
    if {$with_links} {
      append titl_string ""
    }
    $tags_menu_shown configure -title [expr {$with_links ? "Local links" : "Local tags"}]
    set tags_menu_shown_torn [::tk::TearOffMenu $tags_menu_shown $pos_x $pos_y]
    catch {wm resizable $tags_menu_shown false false}
    catch {wm resizable $tags_menu_shown_torn false false}
    catch {focus $tags_menu_shown_torn}
    catch {
      bind $tags_menu_shown_torn <Escape> {
        remove_tags_menu_shown
        break
      }
    }
  } {
    tk_popup $tags_menu_shown $pos_x $pos_y
  }
}

# (most of) the scrape/import stuff;
# notice that (most of) regexps used here should be improved eventually
# (they work on simple cases, though not robust enough for involved cases);

# at some next versions, data in import should allow more (and inline) ways
# of quotations, and allowing it to get exported into gnuplot-data window;

proc update_import_warning_label {} {
  global import_warning_label
  global import_input_warning

  if {$import_warning_label==""} {
    return
  }

  if {$import_input_warning} {
    $import_warning_label configure -foreground "#c0f0c0" -background "#a0c0a0"
  } {
    $import_warning_label configure -foreground "#c0c0c0" -background "#a0a0a0"
  }
}

proc update_import_input_warning {} {
  update_import_warning_label
  set_import_title
}

proc unset_import_input_warning {} {
  global import_input_warning

  set import_input_warning false
  update_import_input_warning
}

proc set_import_input_warning {} {
  global import_input_warning

  set import_input_warning true
  update_import_input_warning
}

proc switch_import_input_warning {} {
  global import_input_warning

  set import_input_warning [expr {$import_input_warning ? false : true}]
  update_import_input_warning
}

set import_warning_menu ""
proc offer_import_warning_menu {x_abs y_abs x_pos y_pos} {
  global topw
  global import_warning_label
  global import_warning_menu
  global import_input_warning
  global default_menu_offset_x
  global default_menu_offset_y
  global menu_selected_background

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  if {$import_warning_menu!=""} {
    catch {destroy $import_warning_menu}
  }

  set import_warning_menu [menu "$topw.import_warning_menu" -tearoff false]
  $import_warning_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $import_warning_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $import_warning_menu configure -activeborderwidth 1
  $import_warning_menu configure -title "warning on non-empty data"

  $import_warning_menu add command -label "To warn at app closing:" -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  $import_warning_menu add command -label "yes, for non-empty input/output" -command [list "set_import_input_warning"]
  $import_warning_menu add command -label "not for import parts" -command [list "unset_import_input_warning"]

  set use_idx 2
  if {$import_input_warning} {
    set use_idx 1
  }
  $import_warning_menu entryconfigure $use_idx -background $menu_selected_background -activebackground "#e0e0ff"

  tk_popup $import_warning_menu $use_x $use_y
}

set import_numer_output true
proc update_import_numer_output {} {
  global import_output_numer_label
  global import_numer_output
  global info_labels_foreground

  if {$import_output_numer_label==""} {
    return
  }

  set fgcol [lindex $info_labels_foreground [expr {$import_numer_output ? 1 : 0}]]
  $import_output_numer_label configure -foreground $fgcol

  # using it for new translated stuff, thus no direct data update here
  #update_import_output_data
}

proc unset_import_numer_output {} {
  global import_numer_output

  set import_numer_output false
  update_import_numer_output
}

proc set_import_numer_output {} {
  global import_numer_output

  set import_numer_output true
  update_import_numer_output
}

proc switch_import_numer_output {} {
  global import_numer_output

  set import_numer_output [expr {$import_numer_output ? false : true}]
  update_import_numer_output
}

proc is_import_numer_output {} {
  global import_numer_output
  return $import_numer_output
}

proc sweep_import_enable {} {
  global bt_import_output_sweep

  if {$bt_import_output_sweep==""} {
    return
  }

  $bt_import_output_sweep configure -state "normal"
}
proc sweep_import_disable {} {
  global bt_import_output_sweep

  if {$bt_import_output_sweep==""} {
    return
  }

  $bt_import_output_sweep configure -state "disabled"
}

proc sweep_import_output {} {
  global import_output_field

  if {$import_output_field==""} {
    return
  }

  set curr_state [$import_output_field cget -state]
  $import_output_field configure -state "normal"
  $import_output_field delete "1.0" "end"
  $import_output_field configure -state $curr_state

  sweep_import_disable
}

set rdarr_char "=>"
catch {
  set rdarr_char [encoding convertfrom "utf-8" "\xE2\x87\x92"]
}

proc present_import_about {} {
  global import_output_field
  global rdarr_char
  global import_output_tag
  global import_help_tag

  if {$import_output_field==""} {
    return
  }

  set curr_state [$import_output_field cget -state]
  $import_output_field configure -state "normal"

  if {[lindex [split [$import_output_field index "end -1 chars"] "."] 1]!=0} {
    $import_output_field insert "end" "\n" [list $import_output_tag]
  }
  if {[$import_output_field index "end -1 chars"]!="1.0"} {
    $import_output_field insert "end" "\n" [list $import_output_tag]
  }

  set help_string ""
  append help_string "\".tex\" action (Control+Enter) scans a given LaTeX file for tabular data,\n"
  append help_string "extracts any found, and puts them into the (lower) input text;\n"
  append help_string "it translates the found tables into Maxima fields too.\n"
  append help_string "\n"
  append help_string "\"\\\\$rdarr_char\[\]\" action (Shift+Enter) translates LaTeX tabular data given\n"
  append help_string "in the (lower) input text, with tables (fields) considered\n"
  append help_string "separated by empty lines; tables assumed to be given without\n"
  append help_string "any headers (for start/end, lines, etc.) there.\n"
  append help_string "\n"
  append help_string "\".num\" option sets whether numerical cells should be put in as numbers,\n"
  append help_string "that is, not as strings with quotes.\n"
  append help_string "\n"
  append help_string "Control+d to open this window, Control+comma/period to slide within.\n"
  append help_string "\n"

  set last_end_pos [$import_output_field index "end -1 chars"]

  $import_output_field insert "end" $help_string [list $import_output_tag $import_help_tag]
  $import_output_field configure -state $curr_state

  $import_output_field see "end"
  $import_output_field see $last_end_pos

  focus $import_output_field
  sweep_import_enable
}

set import_input_action_menu ""
proc offer_import_input_menu {x_abs y_abs x_pos y_pos} {
  global topw
  global import_input_action_menu
  global default_menu_offset_x
  global default_menu_offset_y

  set use_x [expr {$x_abs + $default_menu_offset_x}]
  set use_y [expr {$y_abs + $default_menu_offset_y}]

  if {$import_input_action_menu!=""} {
    catch {destroy $import_input_action_menu}
  }
  set import_input_action_menu [menu "$topw.import_input_action_menu" -tearoff false]
  $import_input_action_menu configure -relief solid -borderwidth 1 -background "#ffffff" -disabledforeground "#000000"
  $import_input_action_menu configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f0f0f0"
  $import_input_action_menu configure -activeborderwidth 1
  $import_input_action_menu configure -title "import actions"

  set label_text "Import artions: Control/Shift + Enter"
  $import_input_action_menu add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  set label_text "Control+Enter: scrapes TeX file for tabular data"
  $import_input_action_menu add command -label $label_text -command [list "import_scrape_data" "latex"]
  set label_text "Shift+Enter: translates TeX tabular data (from here)"
  $import_input_action_menu add command -label $label_text -command [list "import_translate_data_latex"]

  tk_popup $import_input_action_menu $use_x $use_y
}

set bt_import_translate_wiki ""
set bt_import_translate_tex ""
set bt_import_scrape_tex ""
set bt_import_output_sweep ""
set import_path ""
set import_window ""
set import_fmenu ""
set import_pane_holder ""
set import_output_frame ""
set import_input_frame ""
set import_output_field ""
set import_input_field ""
set import_warning_label ""
set import_output_tag "import_output_tag"
set import_help_tag "import_help_tag"
proc setup_import {} {
  global import_window
  global scr_width
  global import_fmenu
  global import_pane_holder
  global import_output_frame
  global import_input_frame
  global import_output_field
  global import_input_field
  global bt_import_translate_wiki
  global bt_import_translate_tex
  global bt_import_scrape_tex
  global bt_import_output_sweep
  global import_path
  global spec_base_font
  global select_fg_col
  global select_bg_col
  global import_output_tag
  global import_help_tag
  global tag_command_bracket_pair
  global tag_command_bracket
  global import_output_numer_label
  global rdarr_char
  global use_buttons_text
  global import_warning_label
  global known_selectforeground_selectbackground_for_tags

  if {$import_window!=""} {
    return
  }

  set import_window [toplevel ".import" -width 600 -height 400]

  set top_pos [get_top_window_position]
  set x_pos [expr {20 + [lindex $top_pos 0]}]
  set y_pos [expr {(($scr_width>1000) ? 60 : 25) + [lindex $top_pos 1]}]
  wm geometry $import_window "+$x_pos+$y_pos"

  set flabels [frame $import_window.flabels -relief sunken -bd 1]
  set import_fmenu $flabels
  pack $flabels -side top -expand no -fill x
  set ftext_holder [panedwindow $import_window.ftext_holder -orient vertical -handlesize 1 -borderwidth 0 -sashpad 0]
  pack $ftext_holder -side top -expand yes -fill both
  set import_pane_holder $ftext_holder

  set transl_label ""
  append transl_label "\\\\" $rdarr_char "\[\]"

  set bt_import_scrape_tex [ttk::button $flabels.biscrapetex -text ".tex" -command {import_scrape_data "latex"} -takefocus 0 -width 6 -style File.TButton]
  set bt_import_translate_tex [ttk::button $flabels.bitranslatetex -text $transl_label -command {import_translate_data_latex} -takefocus 0 -width 6 -style File.TButton]
  set bt_import_output_sweep [ttk::button $flabels.bioutputsweep -text "Sweep" -command {sweep_import_output} -takefocus 0 -width 6 -style File.TButton]
  pack $bt_import_scrape_tex $bt_import_translate_tex $bt_import_output_sweep -side left

  set import_output_numer_label [ttk::label $flabels.bioutputnumer -text "num" -padding [list 5 0 5 0]]
  pack $import_output_numer_label -side left
  update_import_numer_output

  set limport_about [ttk::button $flabels.ldimport -text "?" -command {present_import_about} -takefocus 0 -width 2 -style File.TButton]
  pack $limport_about -side right

  set circ_ch ""
  if {!$use_buttons_text} {
    catch {set circ_ch [encoding convertfrom "utf-8" "\xE2\x97\x8F"]}
  }
  set warning_char [expr {($circ_ch=="") ? "m" : $circ_ch}]
  set import_warning_label [label $flabels.ldwarning -text "$warning_char" -relief "flat" -borderwidth 0 \
    -foreground "#c0c0c0" -background "#a0a0a0" -width 3]
  pack $import_warning_label -side right -fill y
  update_import_warning_label

  bind $import_warning_label <1> {
    remove_torn_menus
    offer_import_warning_menu %X %Y %x %y
    break
  }
  bind $import_warning_label <Control-1> {
    remove_torn_menus
    switch_import_input_warning
    break
  }

  $bt_import_scrape_tex configure -state "normal"
  $bt_import_translate_tex configure -state "normal"
  $bt_import_output_sweep configure -state "disabled"

  set import_output_frame [ttk::frame $ftext_holder.output_frame -relief groove -borderwidth 2]
  set import_input_frame [ttk::frame $ftext_holder.input_frame -relief groove -borderwidth 2]
  $ftext_holder add $import_output_frame -minsize 100 -stretch always
  $ftext_holder add $import_input_frame -minsize 120 -stretch always

  set import_output_field [text $import_output_frame.import_output_field -wrap char -undo false -height 12]
  ttk::scrollbar $import_output_frame.vscr -orient vertical -command "$import_output_field yview"
  $import_output_field configure -yscroll "$import_output_frame.vscr set"

  $import_output_field configure -state "disabled"

  grid $import_output_field $import_output_frame.vscr -sticky nsew

  grid rowconfigure $import_output_frame $import_output_field -weight 1
  grid columnconfigure $import_output_frame $import_output_field -weight 1

  if {$spec_base_font(command)==""} {
    $import_output_field configure -font command_font
  } {
    $import_output_field configure -font $spec_base_font(command)
  }

  $import_output_field configure -foreground "#000000"
  $import_output_field configure -background "#ffffff"
  $import_output_field configure -selectforeground $select_fg_col
  $import_output_field configure -selectbackground $select_bg_col
  $import_output_field configure -inactiveselectbackground $select_bg_col

  $import_output_field tag configure $import_help_tag -background "#ffffe7" -foreground "#101010"
  $import_output_field tag configure $import_output_tag -lmargin1 0 -lmargin2 5
  catch {
    $import_output_field tag configure $import_output_tag -lmargincolor "#9090b0"
  }

  $import_output_field tag raise $import_help_tag
  $import_output_field tag raise "sel"
  $import_output_field tag raise $import_output_tag
  if {!$known_selectforeground_selectbackground_for_tags} {
    $import_output_field tag raise "sel"
  }

  set import_input_field [text $import_input_frame.import_input_field -wrap none -undo true -maxundo 10000 -height 12]
  ttk::scrollbar $import_input_frame.vscr -orient vertical -command "$import_input_field yview"
  ttk::scrollbar $import_input_frame.hscr -orient horizontal -command "$import_input_field xview"
  $import_input_field configure -xscroll "$import_input_frame.hscr set" -yscroll "$import_input_frame.vscr set"
  ttk::sizegrip $import_input_frame.size_grip

  grid $import_input_field $import_input_frame.vscr -sticky nsew
  grid $import_input_frame.hscr $import_input_frame.size_grip -sticky nsew

  grid rowconfigure $import_input_frame $import_input_field -weight 1
  grid columnconfigure $import_input_frame $import_input_field -weight 1

  if {$spec_base_font(command)==""} {
    $import_input_field configure -font command_font
  } {
    $import_input_field configure -font $spec_base_font(command)
  }

  $import_input_field configure -foreground "#000000"
  $import_input_field configure -background "#ffffff"
  $import_input_field configure -selectforeground $select_fg_col
  $import_input_field configure -selectbackground $select_bg_col
  $import_input_field configure -inactiveselectbackground $select_bg_col

  $import_input_field tag configure $tag_command_bracket_pair -foreground "#00a0f0"
  $import_input_field tag configure $tag_command_bracket -foreground "#f03000"

  $import_input_field tag raise "sel"
  $import_input_field tag raise $tag_command_bracket_pair
  $import_input_field tag raise $tag_command_bracket

  if {!$known_selectforeground_selectbackground_for_tags} {
    $import_input_field tag raise "sel"
  }

  bind $import_input_field <FocusIn> {
    $import_input_field tag configure "sel" -bgstipple ""
    remove_torn_menus
  }
  bind $import_input_field <FocusOut> {
    $import_input_field tag configure "sel" -bgstipple gray75
  }

  bind $import_output_field <FocusIn> {
    $import_output_field tag configure "sel" -bgstipple ""
    remove_torn_menus
  }
  bind $import_output_field <FocusOut> {
    $import_output_field tag configure "sel" -bgstipple gray75
  }

  bind $import_output_field <1> {
    focus $import_output_field
  }

  bind $import_input_field <3> {
    offer_import_input_menu %X %Y %x %y
    break
  }
  bind $import_input_field <Shift-3> {
    break
  }
  bind $import_input_field <Control-3> {
    break
  }
  bind $import_input_field <Alt-3> {
    break
  }

  # first generally removing bindings
  set other_keys [list "Control-s" "Control-t" "Alt-t" "Control-d" "Alt-d" "Control-e" "Control-E" \
    "Control-k" "Control-o" "Control-B" "Alt-b" "Control-f" "Alt-f" "Control-g"  "Control-h" \
    "Control-p" "Control-n" "Control-l" "Control-j" "Control-comma" "Control-period" "Control-u" \
    "Control-slash" "Control-backslash" "Control-r" "Control-R" "Control-q" "Control-w"]
  foreach keyspressed $other_keys {
    bind $import_input_field <$keyspressed> {
      break
    }
  }
  foreach keyspressed $other_keys {
    bind $import_output_field <$keyspressed> {
      break
    }
  }

  foreach wdg [list $import_window $import_input_field $import_output_field] {
    bind $wdg <Escape> {
      close_import
      break
    }

    bind $wdg <Control-e> {
      sweep_import_output
      break
    }
    bind $wdg <Control-Alt-e> {
      sweep_import_output
      break
    }
    bind $wdg <Control-E> {
      sweep_import_output
      break
    }
  }

  # calling the loading/translation actions
  foreach wdg [list $import_window $import_input_field $import_output_field] {
    bind $wdg <Control-Return> {
      import_scrape_data "latex"
      break
    }
    bind $wdg <Shift-Return> {
      import_translate_data_latex
      break
    }
    bind $wdg <Control-Shift-Return> {
      break
    }
    bind $wdg <Alt-Return> {
      break
    }
  }

  bind $import_input_field <Control-a> {
    $import_input_field tag remove "sel" "1.0" "end"
    $import_input_field tag add "sel" "1.0" "end -1 chars"
    $import_input_field edit separator
    set_anchor_opposite $import_input_field
    break
  }
  bind $import_output_field <Control-a> {
    $import_output_field tag remove "sel" "1.0" "end"
    $import_output_field tag add "sel" "1.0" "end -1 chars"
    break
  }

  bind $import_input_field <Control-x> {
    $import_input_field edit separator
    set_clipboard $import_input_field
    store_clipboard_onset
    #tk_textCut $import_input_field
    delete_selected_text $import_input_field
    $import_input_field edit separator
    break
  }

  bind $import_input_field <Control-c> {
    $import_input_field edit separator
    set_clipboard $import_input_field
    store_clipboard_onset
    $import_input_field edit separator
    break
  }
  bind $import_output_field <Control-c> {
    set_clipboard $import_output_field
    store_clipboard_onset
    break
  }

  bind $import_input_field <Control-V> {
    $import_input_field edit separator
    tk_textPaste $import_input_field
    $import_input_field edit separator
    $import_input_field see "insert"
    break
  }
  bind $import_input_field <Control-v> {
    $import_input_field edit separator
    remove_selected_if_by_cursor $import_input_field
    $import_input_field edit separator
    tk_textPaste $import_input_field
    $import_input_field edit separator
    $import_input_field see "insert"
    break
  }

  bind $import_input_field <Control-z> {
    catch {
      $import_input_field edit undo
    }
    break
  }
  bind $import_input_field <Control-y> {
    catch {
      $import_input_field edit redo
    }
    break
  }
  bind $import_input_field <Control-Z> {
    catch {
      $import_input_field edit redo
    }
    break
  }

  bind $import_input_field <Prior> {
    topwise_page_up $import_input_field
    set_anchor_insert $import_input_field
  }
  bind $import_input_field <Shift-Prior> {
  }
  bind $import_output_field <Prior> {
    topwise_page_up $import_output_field
  }
  bind $import_output_field <Shift-Prior> {
  }

  bind $import_input_field <Next> {
    bottomwise_page_down $import_input_field
    set_anchor_insert $import_input_field
  }
  bind $import_input_field <Shift-Next> {
  }
  bind $import_output_field <Next> {
    bottomwise_page_down $import_output_field
  }
  bind $import_output_field <Shift-Next> {
  }

  # without bracket jumping here,
  # since the considered tex data are not based on that;
  bind $import_input_field <Control-b> {
    break
  }
  bind $import_input_field <Control-B> {
    break
  }

  # setting for quit-warning
  bind $import_input_field <Control-m> {
    set_import_input_warning
    break
  }
  bind $import_output_field <Control-m> {
    set_import_input_warning
    break
  }
  # unsetting from quit-warning
  bind $import_input_field <Control-M> {
    unset_import_input_warning
    break
  }
  bind $import_output_field <Control-M> {
    unset_import_input_warning
    break
  }

  # setting the numer-output state
  bind $import_input_field <Control-n> {
    set_import_numer_output
    break
  }
  bind $import_output_field <Control-n> {
    set_import_numer_output
    break
  }
  # unsetting the numer-output state
  bind $import_input_field <Control-N> {
    unset_import_numer_output
    break
  }
  bind $import_output_field <Control-N> {
    unset_import_numer_output
    break
  }
  # switching the numer-output state
  bind $import_output_numer_label <1> {
    switch_import_numer_output
    break
  }

  bind $import_input_field <2> {
    insert_primary $import_input_field %X %Y %x %y
    break
  }
  bind $import_input_field <ButtonRelease-2> {
    break
  }
  bind $import_input_field <1> {
    place_cursor_on_mouse $import_input_field %X %Y %x %y
    $import_input_field tag remove "sel" "1.0" "end"
    set_anchor_insert $import_input_field
    break
  }
  bind $import_input_field <ButtonRelease-1> {
    break
  }
  bind $import_input_field <Control-1> {
    break
  }
  bind $import_input_field <Control-ButtonRelease-1> {
    break
  }

  bind $import_input_field <Double-1> {
    field_on_double_click $import_input_field
    break
  }
  bind $import_input_field <Control-Double-1> {
    set_anchor_insert $import_input_field
    break
  }
  bind $import_input_field <Control-Shift-Double-1> {
    set_anchor_insert $import_input_field
    break
  }
  foreach keyspressed [list "Triple-1" "Quadruple-1"] {
    bind $import_input_field <$keyspressed> {
      field_on_multi_click $import_input_field
      break
    }
  }
  foreach keyspressed [list "Control-Triple-1" "Control-Quadruple-1"] {
    bind $import_input_field <$keyspressed> {
      break
    }
  }

  bind $import_input_field <Double-1> {
    field_on_double_click $import_input_field
    break
  }
  bind $import_input_field <Control-Double-1> {
    break
  }
  bind $import_input_field <Control-Shift-Double-1> {
    break
  }
  foreach keyspressed [list "Triple-1" "Quadruple-1"] {
    bind $import_input_field <$keyspressed> {
      field_on_multi_click $import_input_field
      break
    }
  }
  foreach keyspressed [list "Control-Triple-1" "Control-Quadruple-1"] {
    bind $import_input_field <$keyspressed> {
      break
    }
  }

  foreach wdg [list $import_window $import_input_field $import_output_field] {
    bind $wdg <Control-comma> {
      focus $import_output_field
      break
    }
    bind $wdg <Control-period> {
      focus $import_input_field
      break
    }
  }

  bind $import_input_field <Control-r> {
    open_plot_part
    break
  }
  bind $import_input_field <Control-Alt-r> {
    open_plot_part $import_input_field false
    break
  }
  bind $import_input_field <Control-R> {
    open_plot_part $import_input_field true
    break
  }
  bind $import_output_field <Control-r> {
    open_plot_part
    break
  }
  bind $import_output_field <Control-Alt-r> {
    open_plot_part $import_output_field false
    break
  }
  bind $import_output_field <Control-R> {
    open_plot_part $import_output_field true
    break
  }

  bind $import_input_field <Control-bar> {
    if {[is_batch_mode]} {
      draw_batch_maxima
    }
    break
  }
  bind $import_output_field <Control-bar> {
    if {[is_batch_mode]} {
      draw_batch_maxima
    }
    break
  }

  bind $import_input_field <Tab> {
    string_action_on_tab_simple $import_input_field
    break
  }

  catch {
    bind $import_input_field <Control-Tab> {
      break
    }
  }
  foreach keyspressed [list "Shift-Key-Tab" "Shift-Key-KP_Tab" "Shift-Key-ISO_Left_Tab" "Key-ISO_Left_Tab"] {
    catch {
      bind $import_input_field <$keyspressed> {
        string_action_on_tab_simple $import_input_field
        break
      }
    }
    catch {
      bind $import_input_field <Control-$keyspressed> {
        break
      }
    }
  }

  bind $import_input_field <<Modified>> {
    break
  }

  bind_keypad $import_window
  bind_keypad $import_input_field
  bind_keypad $import_output_field

  bind $import_input_field <KeyPress-Home> {
    home_key_reasonable $import_input_field
    break
  }
  bind $import_input_field <Control-KeyPress-Home> {
  }
  bind $import_input_field <Shift-KeyPress-Home> {
  }
  bind $import_input_field <Alt-KeyPress-Home> {
  }

  bind $import_output_field <KeyPress-Home> {
    home_key_reasonable $import_output_field
    break
  }
  bind $import_output_field <Control-KeyPress-Home> {
  }
  bind $import_output_field <Shift-KeyPress-Home> {
  }
  bind $import_output_field <Alt-KeyPress-Home> {
  }
}

set last_dir_data $last_dir
set extracted_tables [list]

set import_latex_re_round_brackets "\\(\[^\\(\\)\]*\\)"
set import_latex_re_square_brackets "\\\[\[^\\\[\\\]\]*\\\]"
set import_latex_re_curly_brackets "\\\{\[^\\\{\\\}\]*(\\\{\[^\\\{\\\}\]*\\\}\[^\\\{\\\}\]*)*\[^\\\{\\\}\]*\\\}"

proc import_latex_purge_tabular_start {line} {
  global import_latex_re_square_brackets
  global import_latex_re_curly_brackets

  # here the comparisons go by lines, thus can use "\\s" that would otherwise match newlines too;
  # testing the "tabu"-based table onset, then "tabular*"-based table onset;
  set cmp_re1 "^\[\\s\]*to\[\\s\]*\\\\linewidth\[\\s\]*$import_latex_re_curly_brackets"
  # [] pairs without [,] inside; {} pairs that can have matched {} pairs inside;
  set cmp_re2 "^($import_latex_re_square_brackets)*($import_latex_re_curly_brackets)+"
  # not checking for "\\*" form since the left "*" can indicate that the (next) string continues from above;
  #set cmp_re3 "^\\*"

  set match_pos [list]
  if {[regexp -indices -- $cmp_re1 $line match_pos]!=0} {
    set line [string range $line [expr {[lindex $match_pos 1] + 1}] "end"]
    return [string trim $line]
  }

  while {true} {
    set match_pos [list]
    if {[regexp -indices -- $cmp_re2 $line match_pos]==0} {
      break
    }
    set line [string range $line [expr {[lindex $match_pos 1] + 1}] "end"]
  }
  return [string trim $line]
}

proc import_latex_purge_newline_specs {line} {
  global import_latex_re_square_brackets

  set cmp_re_off "^$import_latex_re_square_brackets"
  set match_pos_off [list]
  if {[regexp -indices -- $cmp_re_off $line match_pos_off]>0} {
    # removing [...] part
    set line [string range $line [expr {[lindex $match_pos_off 1] + 1}] "end"]
    return $line
  }
  return $line
}

proc import_latex_translate_bordered_part {line} {
  global import_latex_re_square_brackets
  global import_latex_re_curly_brackets
  global import_latex_re_round_brackets

  # regarding the other/non-specific cases, we only translate (inner here) tabel-row separators into " "
  foreach test_case [list "\[\\w\]+"] {
    set cmp_re_off "^$test_case"
    append cmp_re_off "($import_latex_re_round_brackets)?($import_latex_re_square_brackets)?($import_latex_re_curly_brackets)+"
    set match_pos_off [list]
    if {[regexp -indices -- $cmp_re_off $line match_pos_off]>0} {
      set translated_part [string map [list "\\tabularnewline" " " "\\\\" " "] [string range $line 0 [lindex $match_pos_off 1]]]
      return [list true $translated_part [string range $line [expr {[lindex $match_pos_off 1] + 1}] "end"]]
    }
  }
  return [list false]
}

proc import_latex_extract_bordered_part {line} {
  global import_latex_re_square_brackets
  global import_latex_re_curly_brackets
  global import_latex_re_round_brackets

  # purging various tabular-related cases: exact
  foreach test_case [list "newline" "\\begin\{verbatim\}" "\\end\{verbatim\}" "nodata"] {
    set test_case_len [string length $test_case]
    if {[string equal -nocase -length $test_case_len $line $test_case]} {
      return [list true " " [string range $line $test_case_len "end"]]
    }
  }

  # purging various tabular-related cases: possibly with brackets
  foreach test_case [list "\[\\w\]*line\[\\w\]*" "\[\\w\]*rule\[\\w\]*" "rowfont" "noalign" "tablenotemark"] {
    set cmp_re_off "^$test_case"
    append cmp_re_off "($import_latex_re_round_brackets)?($import_latex_re_square_brackets)?($import_latex_re_curly_brackets)*"
    set match_pos_off [list]
    if {[regexp -indices -- $cmp_re_off $line match_pos_off]>0} {
      return [list true " " [string range $line [expr {[lindex $match_pos_off 1] + 1}] "end"]]
    }
  }

  # eventually should deal with the "multirow"-count too;
  foreach test_case [list "multirow"] {
    set cmp_re_off "^$test_case"
    append cmp_re_off "($import_latex_re_curly_brackets)\{3\}"
    set match_pos_off [list]
    if {[regexp -indices -- $cmp_re_off $line match_pos_off]>0} {
      set matched_part [string range $line 0 [lindex $match_pos_off 1]]
      # not dealing with the count by now
      #set match_pos_aux [list]
      #set cmp_re_aux "^$test_case($import_latex_re_curly_brackets)\{1\}"
      #regexp -indices -- $cmp_re_aux $line match_pos_aux
      #set thecount [string range $line [expr {[string length $test_case]+1}] [lindex $match_pos_aux 1]]
      set match_pos_aux [list]
      set cmp_re_aux "^($import_latex_re_curly_brackets)\{1\}"
      set matched_rev [string map [list "\{" "\}" "\}" "\{"] [string reverse $matched_part]]
      regexp -indices -- $cmp_re_aux $matched_rev match_pos_aux
      set thecontent [string reverse [string range $matched_rev 1 [expr {[lindex $match_pos_aux 1]-1}]]]
      set thecontent [string map [list "\\tabularnewline" " " "\\\\" " " "\{" "\}" "\}" "\{"] $thecontent]
      set thecontent " $thecontent "
      return [list true $thecontent [string range $line [expr {[lindex $match_pos_off 1] + 1}] "end"]]
    }
  }

  # extracting from last part of: word([...])?({...})*{...}
  foreach test_case [list "parbox" "math\[\\w\]*" "text\[\\w\]*" "multicolumn" "multirow"] {
    set cmp_re_off "^$test_case"
    append cmp_re_off "($import_latex_re_square_brackets)?($import_latex_re_curly_brackets)+"
    set match_pos_off [list]
    if {[regexp -indices -- $cmp_re_off $line match_pos_off]>0} {
      set matched_part [string range $line 0 [lindex $match_pos_off 1]]
      set match_pos_aux [list]
      set cmp_re_aux "^($import_latex_re_curly_brackets)\{1\}"
      set matched_rev [string map [list "\{" "\}" "\}" "\{"] [string reverse $matched_part]]
      regexp -indices -- $cmp_re_aux $matched_rev match_pos_aux
      set thecontent [string reverse [string range $matched_rev 1 [expr {[lindex $match_pos_aux 1]-1}]]]
      set thecontent [string map [list "\\tabularnewline" " " "\\\\" " " "\{" "\}" "\}" "\{"] $thecontent]
      set thecontent " $thecontent "
      return [list true $thecontent [string range $line [expr {[lindex $match_pos_off 1] + 1}] "end"]]
    }
  }

  # not dealing with general cases here
  return [list false]
}

proc import_latex_take_without_backslash {line} {
  set match_pos_use [list]
  set cmp_re_use "^\[^\\\\\]*"
  if {[regexp -indices -- $cmp_re_use $line match_pos_use]>0} {
    set part_use [string range $line 0 [expr {[lindex $match_pos_use 1] - 0}]]
    set part_rest [string range $line [expr {[lindex $match_pos_use 1] + 1}] "end"]
    return [list $part_use $part_rest]
  }
  return [list "" $line]
}

proc import_latex_clean_one_column {column} {
  set use_column [list]
  foreach one_word [split $column] {
    set one_word [string trim $one_word]
    if {$one_word==""} {
      continue
    }
    if {[string range $one_word "end" "end"]=="\\"} {
      append one_word " "
    }
    lappend use_column $one_word
  }
  return [join $use_column " "]
}

proc import_from_latex_parse_table {table} {
  set tabularnewline "tabularnewline"
  set tabularnewline_len [string length $tabularnewline]

  set cur_tab_parsed [list]
  set cur_tab_line [list]
  set cur_tab_line_part ""
  # going over the read lines (that need not be table rows);
  # since braces-delimited parts can span over sveral written rows,
  # having concatenated (all) those here;
  foreach one_tab_part $table {
    if {$cur_tab_line_part!=""} {
      lappend cur_tab_line $cur_tab_line_part
      set cur_tab_line_part ""
    }
    while {true} {
      set one_tab_part [string trim $one_tab_part]
      if {$one_tab_part==""} {
        if {$cur_tab_line_part!=""} {
          lappend cur_tab_line $cur_tab_line_part
          set cur_tab_line_part ""
        }
        break
      }
      if {[string range $one_tab_part 0 0]=="\\"} {
        set next_char [string range $one_tab_part 1 1]
        set newline_here false
        set newline_length 0
        if {$next_char=="\\"} {
          set newline_here true
          set newline_length 1
        } elseif {[string range $one_tab_part 1 $tabularnewline_len]==$tabularnewline} {
          set newline_here true
          set newline_length $tabularnewline_len
        }
        if {$newline_here} {
          if {$cur_tab_line_part!=""} {
            lappend cur_tab_line $cur_tab_line_part
            set cur_tab_line_part ""
          }
          if {[llength $cur_tab_line]>0} {
            lappend cur_tab_parsed [join $cur_tab_line " "]
            set cur_tab_line [list]
          }
          set one_tab_part [import_latex_purge_newline_specs [string range $one_tab_part [expr {$newline_length+1}] "end"]]
          continue
        }
        # should take \xy with x being one (any) char, and y being non-word char,
        # since it is commonly \%, \$, diacritics, etc.
        set next_char_plus [string range $one_tab_part 2 2]
        if {![string is wordchar -strict $next_char_plus]} {
          append cur_tab_line_part "\\" $next_char
          set one_tab_part [string range $one_tab_part 2 "end"]
          continue
        }
        if {[string is alnum -strict $next_char] && [string is ascii -strict $next_char]} {
          # to take content of the last part if known form here
          set take_rest [import_latex_extract_bordered_part [string range $one_tab_part 1 "end"]]
          if {[lindex $take_rest 0]} {
            append cur_tab_line_part [lindex $take_rest 1]
            set one_tab_part [lindex $take_rest 2]
            continue
          }
          # trying to translate inner parts if any
          set take_rest [import_latex_translate_bordered_part [string range $one_tab_part 1 "end"]]
          if {[lindex $take_rest 0]} {
            append cur_tab_line_part "\\" [lindex $take_rest 1]
            set one_tab_part [lindex $take_rest 2]
            continue
          }
        }
        # if nothing else deals with backslash, taking it together with string up to next backslash
        set take_rest [import_latex_take_without_backslash [string range $one_tab_part 1 "end"]]
        append cur_tab_line_part "\\" [lindex $take_rest 0]
        set one_tab_part [lindex $take_rest 1]
        continue
      } {
        # to take (longest) part without backslash
        set take_rest [import_latex_take_without_backslash $one_tab_part]
        if {[lindex $take_rest 0]!=""} {
          append cur_tab_line_part [lindex $take_rest 0]
        }
        set one_tab_part [lindex $take_rest 1]
        continue
      }
    }
  }

  if {$cur_tab_line_part!=""} {
    lappend cur_tab_line $cur_tab_line_part
    set cur_tab_line_part ""
  }
  if {[llength $cur_tab_line]>0} {
    lappend cur_tab_parsed [join $cur_tab_line " "]
    set cur_tab_line [list]
  }

  # here the lines should be table rows,
  # having to split them into columns yet
  set cur_tab_rows_columns [list]
  foreach one_row $cur_tab_parsed {
    set one_row [string trim $one_row]
    if {$one_row==""} {
      continue
    }

    set row_colums [list]
    set one_column ""

    set after_backslash false

    set one_row_list [split $one_row {}]
    set one_row_list_len [llength $one_row_list]
    set one_row_idx -1
    while {true} {
      incr one_row_idx
      if {$one_row_idx>=$one_row_list_len} {
        break
      }
      set curr_char [lindex $one_row_list $one_row_idx]
      if {$after_backslash} {
        append one_column $curr_char
        set after_backslash false
        continue
      }
      if {$curr_char=="\\"} {
        append one_column $curr_char
        set after_backslash true
        continue
      }
      if {$curr_char!="&"} {
        append one_column $curr_char
        continue
      }
      lappend row_colums [import_latex_clean_one_column $one_column]
      set one_column ""
    }
    lappend row_colums [import_latex_clean_one_column $one_column]
    set one_column ""

    lappend cur_tab_rows_columns $row_colums

  }

  return $cur_tab_rows_columns
}

proc import_from_latex_file_inner {path} {
  global import_input_field

  if {$import_input_field==""} {
    return false
  }

  set comment_block_start "\\begin\{comment\}"
  set comment_block_start_len [string length $comment_block_start]
  set comment_block_end "\\end\{comment\}"
  set comment_block_end_len [string length $comment_block_end]

  set tabular_block_start_re "(\\\\begin\\\{tabu\[^\\\{\\\}\]*\\\})|(\\\\startdata)"
  set tabular_block_end_re "(\\\\end\\\{tabu\[^\\\{\\\}\]*\\\})|(\\\\enddata)"

  if {[catch {
    set fh [open $path "r"]
  }]} {
    show_import_input_message_error [list [list "Can not open file for read"]]
    return false
  }

  set tables [list]
  set curr_table [list]

  set at_comment false
  set at_table false

  # first reading the whole text, since gonna deal with it via (global) regexp searches
  set one_line ""
  while {true} {
    if {[catch {
      set one_count [gets $fh one_line]
    }]} {
      catch {close $fh}
      show_import_input_message_error [list [list "Can not read from the file"]]
      return false
    }
    if {$one_count<0} {
      break
    }

    set one_line [string trim $one_line]

    # a simplified way for cutting off the %-based comments
    set one_line_list [split $one_line {}]
    set one_line_len [llength $one_line_list]
    set char_idx -1
    set after_backslash false
    set comment_from -1
    while {$char_idx<$one_line_len} {
      incr char_idx
      if {$after_backslash} {
        set after_backslash false
        continue
      }
      set curr_char [lindex $one_line_list $char_idx]
      if {$curr_char=="\\"} {
        set after_backslash true
        continue
      }
      if {$curr_char=="%"} {
        set comment_from $char_idx
        break
      }
    }
    if {$comment_from>-1} {
      incr comment_from -1
      set one_line [string range $one_line 0 $comment_from]
    }
    if {[string length $one_line]==0} {
      continue
    }

    # dealing with block comments
    set line_parts ""
    set comment_start_pos -1
    if {$at_comment} {
      set comment_start_pos 0
    } {
      set comment_start_pos [string first $comment_block_start $one_line]
      if {$comment_start_pos>-1} {
        set at_comment true
        append line_parts [string range $one_line 0 [expr {$comment_start_pos-1}]]
        set one_line [string range $one_line [expr {$comment_start_pos + $comment_block_start_len}] "end"]
      }
    }

    # taking line parts if outside comments on a line with comment start/end statements
    if {$at_comment} {
      while {true} {
        set comment_end_pos [string first $comment_block_end $one_line]
        if {$comment_end_pos==-1} {
          break
        }
        set at_comment false
        set one_line [string range $one_line [expr {$comment_end_pos + $comment_block_end_len}] "end"]
        set comment_start_pos [string first $comment_block_start $one_line]
        if {$comment_start_pos==-1} {
          append line_parts $one_line
          break
        }
        set at_comment true
        append line_parts [string range $one_line 0 [expr {$comment_start_pos-1}]]
        set one_line [string range $one_line [expr {$comment_start_pos + $comment_block_start_len}] "end"]
      }

      if {[string length $line_parts]==0} {
        continue
      }

      set one_line $line_parts
    }

    set tabular_start_pos -1
    set use_line ""
    if {$at_table} {
      set tabular_start_pos 0
      set use_line $one_line
    } {
      set tabular_match [list]
      if {[regexp -indices -- $tabular_block_start_re $one_line tabular_match]>0} {
        set tabular_start_pos [expr {[lindex $tabular_match 1] + 1}]
        set use_line [string range $one_line $tabular_start_pos "end"]
        set use_line [import_latex_purge_tabular_start $use_line]
        set at_table true
        # start new table
        if {[llength $curr_table]>0} {
          lappend tables $curr_table
        }
        set curr_table [list]
      }
    }

    if {$at_table} {
      while {true} {
        set tabular_match [list]
        if {[regexp -indices -- $tabular_block_end_re $use_line tabular_match]>0} {
          set tab_part [string range $use_line 0 [expr {[lindex $tabular_match 0] - 1}]]
          set tab_part [string trim $tab_part]
          if {$tab_part!=""} {
            lappend curr_table $tab_part
          }
          # to finish the current table
          if {[llength $curr_table]>0} {
            lappend tables $curr_table
          }
          set curr_table [list]
          set at_table false
          set use_line [string range $use_line [expr {[lindex $tabular_match 1] + 1}] "end"]

          set tabular_match [list]
          if {[regexp -indices -- $tabular_block_start_re $use_line tabular_match]>0} {
            set tabular_start_pos [expr {[lindex $tabular_match 1] + 1}]
            set use_line [string range $one_line $tabular_start_pos "end"]
            set use_line [import_latex_purge_tabular_start $use_line]
            set at_table true
            # start new table
            set curr_table [list]
          } {
            break
          }
        } {
          lappend curr_table $use_line
          break
        }
      }
    }
  }

  catch {close $fh}

  if {[llength $curr_table]>0} {
    lappend tables $curr_table
    set curr_table [list]
  }

  set tables_parsed [list]
  # simple parsing of the tables
  foreach one_table $tables {
    set one_table [list [join $one_table " "]]
    lappend tables_parsed [import_from_latex_parse_table $one_table]
  }

  set curr_state [$import_input_field cget -state]
  $import_input_field configure -state "normal"
  $import_input_field delete "1.0" "end"

  set is_first_table true
  foreach table $tables_parsed {
    set table_string ""
    if {$is_first_table} {
      set is_first_table false
    } {
      append table_string "\n"
    }

    set is_first_table_row true
    foreach table_row $table {
      if {[llength $table_row]==0} {
        continue
      }
      if {$is_first_table_row} {
        set is_first_table_row false
      } {
        append table_string " \\\\\n"
      }
      set is_first_table_row_column true
      foreach table_column $table_row {
        if {$is_first_table_row_column} {
          set is_first_table_row_column false
        } {
          append table_string " & "
        }
        append table_string $table_column
      }
    }
    append table_string "\n"

    $import_input_field insert "end" $table_string [list]
  }

  $import_input_field configure -state $curr_state

  return true
}

proc import_from_latex_file {path} {
  global import_input_field
  global import_output_field
  global import_path
  global last_dir_data
  global import_output_tag

  if {$import_input_field==""} {
    return
  }

  set imp_res [import_from_latex_file_inner $path]
  if {!$imp_res} {
    return
  }
  $import_input_field mark set "insert" "1.0"
  $import_input_field see "1.0"

  set import_path $path
  set last_dir_data [file dirname $path]

  set last_end_pos [$import_output_field index "end"]

  # putting the header in there, so that the file name is visible
  set onset_string ""
  if {[$import_output_field index "end -1 chars"]!="1.0"} {
    append onset_string "\n"
  }
  # if $path contains "/*" or "*/", it breaks the comment here,
  # though it is just decorative here, thus not dealing with it;
  append onset_string "/* " $path " */\n"

  set curr_state [$import_output_field cget -state]
  $import_output_field configure -state "normal"
  $import_output_field insert "end" $onset_string [list $import_output_tag]
  $import_output_field configure -state $curr_state

  # if something got, translating it and putting it into the output too;
  if {[$import_input_field index "end -1 chars"]!="1.0"} {
    import_translate_data_latex_inner
  }

  $import_output_field see "end"
  $import_output_field see $last_end_pos

  if {[$import_output_field index "end -1 chars"]!="1.0"} {
    sweep_import_enable
  }
}

proc import_scrape_data {form} {
  global import_input_field
  global last_dir_data
  global spec_file_dialogs

  if {$import_input_field==""} {
    return
  }
  set form [string tolower $form]
  if {$form=="latex"} {
    set data_form "LaTeX"
  } else {
    return
  }

  set open_path ""
  if {true} {
    set use_common_open_dialog true
    if {$spec_file_dialogs(open) != ""} {
      set use_common_open_dialog false
      set bad [catch {
        set open_path [$spec_file_dialogs(open) -details true -initialdir $last_dir_data -multiple false -title "Extract from $data_form"]
      }]
      if {$bad} {
        set use_common_open_dialog true
        set open_path ""
      }
    }
    if {$use_common_open_dialog} {
      set open_path [tk_getOpenFile -initialdir $last_dir_data -multiple false -title "Extract from $data_form"]
    }
  }

  if {$open_path==""} {
    return
  }

  if {$form=="latex"} {
    import_from_latex_file $open_path
  }
}

proc import_translate_data_latex {} {
  global import_output_field
  global import_output_tag
  global import_input_field

  if {$import_input_field==""} {
    return
  }

  if {[$import_input_field index "end -1 chars"]=="1.0"} {
    show_import_input_message_error [list [list "Empty input text"]]
    return
  }

  set onset_string ""
  if {[$import_output_field index "end -1 chars"]!="1.0"} {
    append onset_string "\n"
  }
  append onset_string "/* translated tabular data */\n"

  set last_end_pos [$import_output_field index "end"]

  set curr_state [$import_output_field cget -state]
  $import_output_field configure -state "normal"
  $import_output_field insert "end" $onset_string [list $import_output_tag]
  $import_output_field configure -state $curr_state

  import_translate_data_latex_inner

  $import_output_field see "end"
  $import_output_field see $last_end_pos

  if {[$import_output_field index "end -1 chars"]!="1.0"} {
    sweep_import_enable
  }
}

proc import_translate_data_latex_inner {} {
  global import_input_field
  global import_output_field
  global import_output_tag

  if {$import_input_field==""} {
    return
  }

  set use_numer [is_import_numer_output]

  set columns_safe [list "\\" "\\\\" "\"" "\\\""]

  set tables [list]
  set curr_table ""
  set first_table_row true
  foreach one_row [split [$import_input_field get "1.0" "end -1 chars"] "\n"] {
    set one_row [string trim $one_row]
    if {$one_row==""} {
      if {$curr_table!=""} {
        lappend tables $curr_table
        set curr_table ""
        set first_table_row true
      }
    }
    if {$first_table_row} {
      set first_table_row false
    } {
      append curr_table " "
    }
    append curr_table $one_row
  }
  if {$curr_table!=""} {
    lappend tables $curr_table
    set $curr_table ""
  }

  set tables_parsed [list]
  foreach one_table $tables {
    lappend tables_parsed [import_from_latex_parse_table [list $one_table]]
  }

  set curr_state [$import_output_field cget -state]
  $import_output_field configure -state "normal"

  set is_first_table true
  foreach table $tables_parsed {
    set table_string ""
    if {$is_first_table} {
      set is_first_table false
    } {
      append table_string "/* */\n"
    }
    append table_string "\["

    set is_first_table_row true
    foreach table_row $table {
      if {[llength $table_row]==0} {
        continue
      }
      if {$is_first_table_row} {
        set is_first_table_row false
      } {
        append table_string ",\n"
      }
      append table_string "\["
      set is_first_table_row_column true
      foreach table_column $table_row {
        set table_column [import_latex_clean_one_column $table_column]
        if {$is_first_table_row_column} {
          set is_first_table_row_column false
        } {
          append table_string ","
        }
        set col_numer false
        if {$use_numer} {
          if {[string is entier -strict $table_column] || [string is double -strict $table_column]} {
            set col_numer true
          }
        }
        if {$col_numer} {
          append table_string $table_column
        } {
          append table_string "\"" [string map $columns_safe $table_column] "\""
        }
      }
      append table_string "\]"
    }
    append table_string "\];\n"

    $import_output_field insert "end" $table_string [list $import_output_tag]
  }

  $import_output_field configure -state $curr_state
}

# since many texts are only available at pdf (that is without TeX sources),
# should try to extract from pdf too, probably via use of the pdftotext tool;
# a curious developer can provide a patch for that;

# spelling services (needs aspell/hunspell/ispell)

proc spell_get_checker_paths {} {
  global spell_path

  if {$spell_path!=""} {
    return [list [strcat "\"" $spell_path "\""]]
  }

  if {![is_wine_like]} {
    set use_list [list]
    foreach one_test [list "aspell" "hunspell" "ispell"] {
      if {[auto_execok $one_test]!=""} {
        lappend use_list $one_test
      }
    }
    return $use_list
  }

  set use_list [list]
  foreach one_test [list "aspell.exe" "hunspell.exe" "ispell.exe"] {
    if {[auto_execok $one_test]!=""} {
      lappend use_list $one_test
    }
  }
  if {[llength $use_list]>0} {
    return $use_list
  }

  set use_list [list]
  foreach one_path_drive [list "C" "D" "Z"] {
    foreach one_path_base [list \
      "/" \
      "/Program Files/" \
      "/Program Files (x86)/" \
    ] {
      set one_path [strcat $one_path_drive ":" $one_path_base "Aspell/bin/aspell.exe"]
      if {[file exists $one_path] && [file isfile $one_path]} {
        lappend use_list "\"$one_path\""
      }
    }
  }
  return $use_list
}

proc spell_get_selected_words {wdg ascii_only} {
  set sel_words [list]
  set word_count_max 40000
  set word_count_use 0
  set too_many_words false

  foreach {sel_start sel_end} [$wdg tag ranges "sel"] {
    set start_info [split $sel_start "."]
    set start_row [lindex $start_info 0]
    set start_col [lindex $start_info 1]
    set end_info [split $sel_end "."]
    set end_row [lindex $end_info 0]
    set end_col [lindex $end_info 1]

    set cur_row [expr {$start_row - 1}]
    while {true} {
      incr cur_row
      if {$cur_row > $end_row} {
        break
      }
      set cur_col -1
      if {$cur_row == $start_row} {
        set cur_col [expr {$start_col - 1}]
      }
      set stop_col [lindex [split [$wdg index "$cur_row.end"] "."] 1]
      if {$cur_row == $end_row} {
        set stop_col $end_col
      }
      set word_start -1
      set last_apostrophe false
      while {true} {
        incr cur_col
        if {$cur_col>=$stop_col} {
          break
        }
        set cur_char [$wdg get "$cur_row.$cur_col"]
        set at_word [string is alpha -strict $cur_char]
        if {$at_word} {
          if {$ascii_only && (![string is ascii -strict $cur_char])} {
            set at_word false
          } {
            set last_apostrophe false
          }
        }
        if {(!$at_word) && ($word_start != -1) && (!$last_apostrophe) && ($cur_char=="'")} {
          set at_word true
          set last_apostrophe true
        }
        if {!$at_word} {
          if {$word_start>-1} {
            set word_end $cur_col
            if {$last_apostrophe} {
              incr word_end -1
            }
            set cur_word [$wdg get -displaychars -- "$cur_row.$word_start" "$cur_row.$word_end"]
            set use_the_word true
            if {$cur_word == ""} {
              set use_the_word false
            } elseif {([string length $cur_word]==1) && [string is ascii -strict $cur_word]} {
              set use_the_word false
            }
            if {$use_the_word} {
              incr word_count_use
              if {$word_count_use > $word_count_max} {
                set too_many_words true
                break
              }
              lappend sel_words [list $cur_word $cur_row $word_start $word_end]
            }
            set word_start -1
            set last_apostrophe false
          }
        } {
          if {$word_start==-1} {
            set word_start $cur_col
          }
        }
      }
      if {$word_start>-1} {
        set word_end $cur_col
        if {$last_apostrophe} {
          incr word_end -1
        }
        set cur_word [$wdg get -displaychars -- "$cur_row.$word_start" "$cur_row.$word_end"]
        set use_the_word true
        if {$cur_word == ""} {
          set use_the_word false
        } elseif {([string length $cur_word]==1) && [string is ascii -strict $cur_word]} {
          set use_the_word false
        }
        if {$use_the_word} {
          incr word_count_use
          if {$word_count_use > $word_count_max} {
            set too_many_words true
            break
          }
          lappend sel_words [list $cur_word $cur_row $word_start $word_end]
        }
        set word_start -1
        set last_apostrophe false
      }
    }
    if {$too_many_words} {
      break
    }
  }

  if {$too_many_words} {
    return [list false "selected region with too many words"]
  }

  return [list true $sel_words]
}

set spellcheck_hnd ""
set spellcheck_pids ""
proc spell_check_close_checker {} {
  global spellcheck_hnd
  global spellcheck_pids

  if {$spellcheck_hnd != ""} {
    catch {close $spellcheck_hnd}
    set spellcheck_hnd ""
  }

  if {$spellcheck_pids != ""} {
    foreach one_pid $spellcheck_pids {
      if {![is_wine_like]} {
        catch {exec "kill" "-s" "SIGKILL" $one_pid}
      } {
        catch {exec "taskkill" "/f" "/pid" $one_pid}
      }
    }
    set spellcheck_pids ""
  }
}

proc remove_spch_words_menu_shown {} {
  global spch_repl_words_menu_shown_tears
  global spch_repl_words_menu_shown_torn
  global spch_repl_words_menu_shown

  if {!$spch_repl_words_menu_shown_tears} {
    return
  }
  set spch_repl_words_menu_shown_tears false

  if {$spch_repl_words_menu_shown_torn!=""} {
    catch {destroy $spch_repl_words_menu_shown_torn}
    set spch_repl_words_menu_shown_torn ""
  }
  if {$spch_repl_words_menu_shown!=""} {
    catch {destroy $spch_repl_words_menu_shown}
    set spch_repl_words_menu_shown ""
  }

  catch {focus $text_field}
}

proc spell_check_offer_replacement_update_menu_entries {forwards step} {
  global spch_repl_words_count_show
  global spch_repl_words_menu_offset
  global spch_repl_words_menu_shown
  global spch_repl_words_menu_label_start
  global spch_last_found_repl_words

  if {$spch_repl_words_menu_shown==""} {
    return
  }

  set use_list $spch_last_found_repl_words
  set use_list_len [llength $use_list]
  if {$use_list_len==0} {
    return
  }

  if {$forwards} {
    set offset_idx $spch_repl_words_menu_offset
    if {$offset_idx>=$use_list_len} {
      set offset_idx 0
    }
  } {
    set offset_idx [expr {$spch_repl_words_menu_offset - (2*$step)}]
    if {$offset_idx<0} {
      set offset_idx [expr {$use_list_len - ($use_list_len % $spch_repl_words_count_show)}]
    }
  }

  set label_text $spch_repl_words_menu_label_start
  append label_text "[expr {1 + $offset_idx}] - [expr {min($use_list_len, $spch_repl_words_count_show + $offset_idx)}]"
  append label_text " / $use_list_len"
  $spch_repl_words_menu_shown entryconfigure 0 -label $label_text

  set curr_idx 0
  set list_idx $offset_idx
  while {$list_idx<$use_list_len} {
    set show_word [lindex $use_list $list_idx]
    # would be probably too strange to deal with word parts here,
    # though not being sure if the suggested words can not be too long;
    if {false} {
      if {[string length $show_word] > 55} {
        set show_word_tmp [string range $show_word "0" "15"]
        append show_word_tmp "..."
        append show_word_tmp [string range $show_word "end-35" "end"]
        set show_word $show_word_tmp
      }
    }
    incr curr_idx
    $spch_repl_words_menu_shown entryconfigure $curr_idx -label $show_word -command [list "spell_check_make_replacement" $list_idx]
    $spch_repl_words_menu_shown entryconfigure $curr_idx -background "#ffffff" -state "normal"
    incr list_idx
    if {$curr_idx==$spch_repl_words_count_show} {
      break
    }
  }

  while {$curr_idx<$spch_repl_words_count_show} {
    incr curr_idx
    $spch_repl_words_menu_shown entryconfigure $curr_idx -label "   " -state disabled -background "#f0f0f0" -command [list "spell_check_make_replacement" "none"]
  }

  set spch_repl_words_menu_offset [expr {$offset_idx + $curr_idx}]
}

proc spell_check_make_replacement {word_idx} {
  global text_field
  global spch_last_single_orig_word
  global spch_last_found_repl_words

  if {![is_text_field_active]} {
    return
  }

  remove_spch_words_menu_shown

  if {$word_idx in [list "" "none"]} {
    return
  }

  set word_repl ""
  if {[catch {
    set word_repl [lindex $spch_last_found_repl_words $word_idx]
  }]} {
    set word_repl ""
  }

  if {$word_repl == ""} {
    return
  }

  set word_row [lindex $spch_last_single_orig_word 1]
  set word_col1 [lindex $spch_last_single_orig_word 2]
  set word_col2 [lindex $spch_last_single_orig_word 3]

  # checking whether the orig word is present there yet
  if {[$text_field get -displaychars [strcat $word_row "." $word_col1] [strcat $word_row "." $word_col2]] != [lindex $spch_last_single_orig_word 0]} {
    show_text_field_message_error [list [list "the original word has changed meanwhile"]]
    return
  }

  catch {
    $text_field edit separator
    $text_field replace [strcat $word_row "." $word_col1] [strcat $word_row "." $word_col2] $word_repl
    $text_field edit separator
  }
  update_text_properties

  catch {focus $text_field}
}

set spch_repl_words_count_show 5
set spch_repl_words_menu_offset 0
set spch_repl_words_menu_shown ""
set spch_repl_words_menu_label_start ""
set spch_repl_words_menu_shown_torn ""
set spch_repl_words_menu_shown_tears false
set spch_last_single_orig_word [list]
set spch_last_found_repl_words [list]
proc spell_check_offer_replacement {x_abs y_abs x_pos y_pos {y_alt ""}} {
  global topw
  global spch_repl_words_count_show
  global spch_repl_words_menu_offset
  global spch_repl_words_menu_shown
  global spch_repl_words_menu_label_start
  global spch_repl_words_menu_shown_torn
  global spch_repl_words_menu_shown_tears
  global spch_last_single_orig_word
  global spch_last_found_repl_words
  global default_menu_offset_x
  global default_menu_offset_y

  if {![is_text_field_active]} {
    return
  }

  set spch_repl_words_menu_offset 0

  if {[llength $spch_last_single_orig_word]==0} {
    return
  }

  set use_list $spch_last_found_repl_words
  set use_list_len [llength $use_list]
  if {$use_list_len==0} {
    return
  }

  if {($x_abs=="") || ($y_abs=="")} {
    set abs_pos [get_text_field_position]
    set use_x [expr {$x_pos + [lindex $abs_pos 0] + $default_menu_offset_x}]
    set use_y [expr {$y_pos + [lindex $abs_pos 1] + $default_menu_offset_y}]
    if {$y_alt!=""} {
      set alt_y [expr {$y_alt + [lindex $abs_pos 1] + $default_menu_offset_y}]
    } {
      set alt_y $use_y
    }
  } {
    set use_x [expr {$x_abs + $default_menu_offset_x}]
    set use_y [expr {$y_abs + $default_menu_offset_y}]
    set alt_y $use_y
  }

  if {$spch_repl_words_menu_shown_torn!=""} {
    catch {destroy $spch_repl_words_menu_shown_torn}
    set spch_repl_words_menu_shown_torn ""
  }
  catch {destroy $spch_repl_words_menu_shown}
  set spch_repl_words_menu_shown [menu "$topw.spchreplwordsmenushown" -tearoff false]
  $spch_repl_words_menu_shown configure -relief solid -borderwidth 1 -background "#f0f0f0" -disabledforeground "#000000"
  $spch_repl_words_menu_shown configure -activeforeground "#000000" -foreground "#000000" -activebackground "#f8ffc0"
  $spch_repl_words_menu_shown configure -relief "solid" -borderwidth 1 -activeborderwidth 1

  set shown_by_parts false
  set spch_repl_words_menu_shown_tears false
  if {$use_list_len>$spch_repl_words_count_show} {
    set shown_by_parts true
    set spch_repl_words_menu_shown_tears true
  }
  if {[is_linux_like]} {
    set spch_repl_words_menu_shown_tears false
  }

  if {$shown_by_parts} {
    set backwards_list [list "2" "Control-Left" "Control-P" "Left"]
    foreach keyspressed $backwards_list {
      bind $spch_repl_words_menu_shown <$keyspressed> {
        catch {
          spell_check_offer_replacement_update_menu_entries false $spch_repl_words_count_show
        }
        break
      }
    }
    set forwards_list [list "3" "Control-Right" "Control-N" "Right"]
    foreach keyspressed $forwards_list {
      bind $spch_repl_words_menu_shown <$keyspressed> {
        catch {
          spell_check_offer_replacement_update_menu_entries true $spch_repl_words_count_show
        }
        break
      }
    }
  }

  set offset_idx 0
  set label_text "Found words, "
  if {$shown_by_parts} {
    append label_text "shown part: "
    set spch_repl_words_menu_label_start $label_text
    append label_text "[expr {1 + $offset_idx}] - [expr {min($use_list_len, $spch_repl_words_count_show + $offset_idx)}]"
    append label_text " / $use_list_len"
  } {
    append label_text "all shown: "
    append label_text "1 - $use_list_len"
  }
  $spch_repl_words_menu_shown add command -label $label_text -state "disabled" -background "#e0e0e0" -activebackground "#e0e0e0"

  set curr_idx 0
  set list_idx $offset_idx
  while {$list_idx<$use_list_len} {
    set show_word [lindex $use_list $list_idx]
    # would be probably too strange to deal with word parts here,
    # though not being sure if the suggested words can not be too long;
    if {false} {
      if {[string length $show_word] > 55} {
        set show_word_tmp [string range $show_word "0" "15"]
        append show_word_tmp "..."
        append show_word_tmp [string range $show_word "end-35" "end"]
        set show_word $show_word_tmp
      }
    }
    $spch_repl_words_menu_shown add command -label $show_word -command [list "spell_check_make_replacement" $list_idx]
    incr curr_idx
    incr list_idx
    $spch_repl_words_menu_shown entryconfigure $curr_idx -background "#ffffff"
    if {$curr_idx==$spch_repl_words_count_show} {
      break
    }
  }

  # to finish a column if it is at an offset part;
  # this case should not happen, since only using one-column menus here;
  if {$shown_by_parts} {
    while {$curr_idx<$spch_repl_words_count_show} {
      $spch_repl_words_menu_shown add command -label "   " -state disabled -background "#f0f0f0" -command [list "spell_check_make_replacement" "none"]
      incr curr_idx
    }
  }

  set spch_repl_words_menu_offset [expr {$offset_idx + $curr_idx}]

  set current_menu $spch_repl_words_menu_shown
  set y_offset_part 20
  if {(!$spch_repl_words_menu_shown_tears) && [is_wine_like]} {
    set y_offset_part 2
  }
  set pos_vert [expr {$alt_y - ([$current_menu yposition [$current_menu index last]] + $y_offset_part)}]
  set pos_vert_min [expr {24 + [winfo vrooty $current_menu]}]
  if {$pos_vert < $pos_vert_min} {
    set pos_vert $use_y
  }

  if {$spch_repl_words_menu_shown_tears} {
    $spch_repl_words_menu_shown configure -title "Found words"
    set spch_repl_words_menu_shown_torn [::tk::TearOffMenu $spch_repl_words_menu_shown $use_x $pos_vert]
    catch {wm resizable $spch_repl_words_menu_shown false false}
    catch {wm resizable $spch_repl_words_menu_shown_torn false false}
    catch {focus $spch_repl_words_menu_shown_torn}
    catch {
      bind $spch_repl_words_menu_shown_torn <Escape> {
        remove_spch_words_menu_shown
        break
      }
    }
  } {
    tk_popup $spch_repl_words_menu_shown $use_x [expr {[is_wine_like] ? $pos_vert : $use_y}]
  }
}

proc set_spell_env_vars {} {
  global env

  set env_name "ISPELL_CHARSET"
  set env($env_name) "utf8"
}
set_spell_env_vars

set spch_in_work false
proc spell_check_selected_text {allow_concats ascii_only} {
  global spch_in_work

  if {$spch_in_work} {
    return
  }
  set spch_in_work true

  set got_issue false
  if {[catch {
    spell_check_selected_text_inner $allow_concats $ascii_only
  }]} {
    set got_issue true
  }

  catch {
    spell_check_close_checker
  }

  set spch_in_work false

  if {$got_issue} {
    # issues should be caught at the inner function, thus this should not occur
    show_text_field_message_error [list [list "an issue during spell-checking"]]
  }
}

proc spell_check_selected_text_inner {allow_concats ascii_only} {
  global text_field
  global search_tag_all
  global spellcheck_hnd
  global spellcheck_pids
  global spch_last_single_orig_word
  global spch_last_found_repl_words

  set sel_words [spell_get_selected_words $text_field $ascii_only]
  if {![lindex $sel_words 0]} {
    show_text_field_message_error [list [list [lindex $sel_words 1]]]
    return
  }
  set sel_words [lindex $sel_words 1]

  if {[llength $sel_words]==0} {
    show_text_field_message_error [list [list "no words selected"]]
    return
  }

  spell_check_close_checker

  set use_spell_paths [spell_get_checker_paths]
  if {[llength $use_spell_paths]==0} {
    set err_mssgs [list [list "no spell-checker found, needs aspell/hunspell/ispell"]]
    if {[is_wine_like]} {
      lappend err_mssgs [list "aspell can be downloaded from http://aspell.net/win32/"]
    }
    show_text_field_message_error $err_mssgs
    return
  }

  foreach one_path $use_spell_paths {
    set checker_startup [strcat "|" $one_path " -a"]
    if {$allow_concats} {
      set to_add_concat_option false
      set one_path_lower [string tolower $one_path]
      if {[string first "aspell" $one_path_lower]>-1} {
        set to_add_concat_option true
      }
      if {[string first "ispell" $one_path_lower]>-1} {
        set to_add_concat_option true
      }
      if {[string first "hunspell" $one_path_lower]>-1} {
        set to_add_concat_option false
      }
      if {$to_add_concat_option} {
        append checker_startup " -C"
      }
    }

    if {[catch {
      set spellcheck_hnd [open $checker_startup "r+"]
      chan configure $spellcheck_hnd -blocking false
      set spellcheck_pids [pid $spellcheck_hnd]
    }]} {
      spell_check_close_checker
    } {
      break
    }
  }

  if {$spellcheck_hnd==""} {
    show_text_field_message_error [list [list "can not start spell checker"]]
    return
  }

  set got_err false
  set got_unk false
  if {[catch {
    set start_time [clock milliseconds]
    while {true} {
      set onset_line [gets $spellcheck_hnd]
      if {$onset_line != ""} {
        if {[string first "spell" [string tolower $onset_line]]==-1} {
          set got_unk true
        }
        break
      }
      if {[expr {$start_time + 2000}] < [clock milliseconds]} {
        set got_err true
        break
      }
      after 1
    }
  }]} {
    set got_err true
  }

  if {$got_unk} {
    spell_check_close_checker
    show_text_field_message_error [list [list "unknown form of spell checker"]]
    return
  }
  if {$got_err} {
    spell_check_close_checker
    show_text_field_message_error [list [list "can not communicate with checker"]]
    return
  }

  foreach one_word_info $sel_words {
    set one_str [lindex $one_word_info 0]
    if {[catch {
      puts $spellcheck_hnd $one_str
      flush $spellcheck_hnd
    }]} {
      spell_check_close_checker
      set got_err true
      break
    }
  }

  if {$got_err} {
    show_text_field_message_error [list [list "could not write to spell checker"]]
    return
  }

  # suggestions only used if a single non-fit word found
  set can_add_nears true

  # it should be better to do the reading via events
  # (with still checking the times to avoid stalled situations);
  # though it is more involved then doing a direct loop;
  set unk_words [list]
  set got_err false
  foreach one_word_info $sel_words {
    set line_set [list]
    set line_one ""
    if {true} {
      set start_time [clock milliseconds]
      while {true} {
        if {[catch {
          set cnt_read [gets $spellcheck_hnd line_one]
        }]} {
          set got_err true
        }
        if {$got_err} {
          break
        }
        if {$cnt_read==-1} {
          if {[expr {$start_time + 1000}] < [clock milliseconds]} {
            set got_err true
            break
          }
          update
          after 1
          continue
        }
        if {$line_one==""} {
          # line processing should end with an empty line
          break
        }
        lappend line_set $line_one
      }
      if {$got_err} {
        break
      }
      if {([llength $line_set]==1) && ([string range [lindex $line_set 0] 0 0]=="-")} {
        if {!$allow_concats} {
          lappend unk_words $one_word_info
          set can_add_nears false
        }
        continue
      }
      if {([llength $line_set]==1) && ([string range [lindex $line_set 0] 0 0]=="+")} {
        if {$can_add_nears} {
          lappend one_word_info [list [string trim [string range [lindex $line_set 0] 1 "end"]]]
        }
        lappend unk_words $one_word_info
        set can_add_nears false
        continue
      }
      foreach line1 $line_set {
        # the word could had been split by the spell checker, thus has to deal with it;
        set first_line_char [string range $line1 0 0]
        if {$first_line_char=="*"} {
          continue
        }
        if {$allow_concats} {
          if {$first_line_char=="-"} {
            continue
          }
        }
        set one_str_orig [lindex $one_word_info 0]
        set one_str ""
        set match0 ""
        set one_cnt ""
        set one_offset ""
        set one_nears_all ""
        set got_nears false
        if {[regexp "^(?:\\&|\\?)\\s+(\[^\\s\]+)\\s+(\\d+)\\s+(\\d+)\\:\\s+(.*)\$" $line1 match0 one_str one_cnt one_offset one_nears_all]==1} {
          set got_nears true
        }
        set exp_str ""
        if {$got_nears} {
          set exp_str [string range $one_str_orig $one_offset [expr {$one_offset + [string length $one_str] - 1}]]
          set exp_fits true
          if {$exp_str != $one_str} {
            # ispell tends to have issues with utf
            set exp_fits false
            set aux_offset $one_offset
            while {[string length [encoding convertto utf-8 [string range $one_str_orig 0 $aux_offset]]]>$one_offset} {
              if {$aux_offset<=0} {
                break
              }
              incr aux_offset -1
            }
            if {$aux_offset<0} {
              set aux_offset 0
            }
            incr aux_offset
            set exp_str [string range $one_str_orig $aux_offset [expr {$aux_offset + [string length $one_str] - 1}]]
            if {$exp_str == $one_str} {
              set one_offset $aux_offset
              set exp_fits true
            }
          }
          if {$exp_fits && ($one_str != "")} {
            # if can not fit the claimed string, can not show it, nor use the nears;
            set one_word_unk [list \
              $one_str \
              [lindex $one_word_info 1] \
              [expr {[lindex $one_word_info 2] + $one_offset}] \
              [expr {[lindex $one_word_info 2] + $one_offset + [string length $one_str]}] \
            ]
            set fits_by_concat false
            set one_near_use [list]
            # hunspell (contrary to aspell) does not allow concatenated words to be accepted via a startup ooption;
            # and along with that, hunspell apparently does not output the "-" lines at more cases than aspell;
            if {$allow_concats || $can_add_nears} {
              foreach one_near [split $one_nears_all ","] {
                set one_near [string trim $one_near]
                if {$one_near==""} {
                  continue
                }
                if {$allow_concats} {
                  if {[join [split $one_near "-"] ""]==$one_str} {
                    set fits_by_concat true
                    break
                  }
                  if {[join [split $one_near " "] ""]==$one_str} {
                    set fits_by_concat true
                    break
                  }
                }
                if {$can_add_nears} {
                  lappend one_near_use $one_near
                }
              }
            }
            if {!$fits_by_concat} {
              # this only if not fitting by concat
              if {[llength $one_near_use]>0} {
                lappend one_word_unk $one_near_use
              }
              lappend unk_words $one_word_unk
              set can_add_nears false
            }
          }
          continue
        }
        if {!$got_nears} {
          if {!($first_line_char in [list "-" "+" "&" "?" "#"])} {
            set got_err true
            break
          }
        }
        set match0 ""
        set one_str ""
        set one_offset ""
        if {[regexp "^(?:\\#)\\s+(\[^\\s\]+)\\s+(\\d+)(?:\[\\d\]+.*)?\$" $line1 match0 one_str one_offset]==1} {
          set exp_str [string range $one_str_orig $one_offset [expr {$one_offset + [string length $one_str] - 1}]]
          set exp_fits true
          if {$exp_str != $one_str} {
            # ispell tends to have issues with utf
            set exp_fits false
            set aux_offset $one_offset
            while {[string length [encoding convertto utf-8 [string range $one_str_orig 0 $aux_offset]]]>$one_offset} {
              if {$aux_offset<=0} {
                break
              }
              incr aux_offset -1
            }
            if {$aux_offset<0} {
              set aux_offset 0
            }
            incr aux_offset
            set exp_str [string range $one_str_orig $aux_offset [expr {$aux_offset + [string length $one_str] - 1}]]
            if {$exp_str == $one_str} {
              set one_offset $aux_offset
              set exp_fits true
            }
          }
          if {$exp_fits && ($one_str != "")} {
            # if can not fit the claimed string, can not show it, nor use the nears;
            set one_word_unk [list \
              $one_str \
              [lindex $one_word_info 1] \
              [expr {[lindex $one_word_info 2] + $one_offset}] \
              [expr {[lindex $one_word_info 2] + $one_offset + [string length $one_str]}] \
            ]
            lappend unk_words $one_word_unk
            set can_add_nears false
          }
        }
      }
      if {$got_err} {
        break
      }
    } {
      set got_err true
      break
    }
    if {$got_err} {
      break
    }
  }

  spell_check_close_checker

  if {$got_err} {
    show_text_field_message_error [list [list "could not use spell checker"]]
    return
  }

  unset_search_highlights $text_field $search_tag_all

  set tags_found_start [list]
  set tags_found_len [list]
  set has_some false
  set words_changed false
  foreach one_word_info $unk_words {
    set has_some true
    if {[lindex $one_word_info 0] != [$text_field get -displaychars \
      [strcat [lindex $one_word_info 1] "." [lindex $one_word_info 2]] \
      [strcat [lindex $one_word_info 1] "." [lindex $one_word_info 3]]]} {
      set words_changed true
      break
    }
    lappend tags_found_start [strcat [lindex $one_word_info 1] "." [lindex $one_word_info 2]]
    lappend tags_found_len [expr {[lindex $one_word_info 3] - [lindex $one_word_info 2]}]
  }

  # checking whether the words are still there;
  if {$words_changed} {
    show_text_field_message_error [list [list "checked text changed meanwhile"]]
    return
  }

  if {!$has_some} {
    return
  }

  $text_field tag remove "sel" "1.0" "end"

  if {true} {
    set_search_highlights $text_field $search_tag_all $tags_found_start $tags_found_len
    seek_next
  }

  if {![is_text_field_active]} {
    return
  }

  set single_near [list]
  if {[llength $unk_words]==1} {
    set single_near [lindex [lindex $unk_words 0] 4]
  }
  if {[llength $single_near]==0} {
    return
  }

  set spch_last_single_orig_word [lindex $unk_words 0]
  set spch_last_found_repl_words $single_near

  set word_row [lindex $spch_last_single_orig_word 1]
  set word_col [lindex $spch_last_single_orig_word [expr {[is_wine_like] ? 3 : 2}]]
  if {[is_wine_like]} {
    incr word_col -1
  }
  set word_ind [strcat $word_row "." $word_col]
  set word_pos [$text_field bbox $word_ind]
  if {$word_pos==[list]} {
    $text_field see $word_ind
    update
    set word_pos [$text_field bbox $word_ind]
    if {$word_pos==[list]} {
      return
    }
  }
  set line_height [lindex $word_pos 3]
  if {$line_height<5} {
    set line_height 20
  }
  set col_offset [lindex $word_pos 0]
  if {[is_wine_like]} {
    incr col_offset [lindex $word_pos 2]
    incr col_offset 5
  }

  after 0 [list \
    "spell_check_offer_replacement" "" "" $col_offset [expr {[lindex $word_pos 1] + $line_height}] [lindex $word_pos 1] \
  ]
}

# common startup

proc force_events_update {} {
  update
  after 200 "force_events_update"
}
if {[is_wine_like]} {
  after 1000 "force_events_update"
  # an attempt to be/get over the console(s)
  wm attributes $topw -topmost true
  update
  wm attributes $topw -topmost false
  update
}

# starting the stuff
update_stack_label
setup_maxima

if {$editor_batch_state} {
  if {$file_path_to_open!=""} {
    load_from_file $file_path_to_open
  }
}
